/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998, Emmanuel CURIS
   Cration : le 27 novembre 1998

   Module de cration d'un modle structural
   Gestion des molcules

   Une molcule est une runion de fragments (chaque fragment tant un ensemble d'atomes
   lis entre eux par un point commun logique).
   On peut dfinir un ensemble d'atomes absorbeurs dans cette molcule.
   Absorbeurs & fragments sont forms  partir des atomes de la liste chane dans la molcule

   Exemples de constructions logiques :
    1) molcule simple, 1 fragment = la molcule
    2) protine, 1 fragment par acide amin
          _ou_ 1 fragment par chane peptidique (? pour complexes protiques...)
    3) complexe organomtallique, 1 fragment = absorbeur + 1 fragment par ligand
    4) acide nuclique, 1 fragment par base (ou paire de base pour l'ADN ?)
    5) oligosaccharide, 1 fragment par sucre

    Note : cette structure est assez peu adapte  la chimie du solide
    (solides ioniques ou mtalliques tels NaCl) -> prfrer alors une description
    en terme de Motif + lments de symtrie... Encore  faire

    Il faudrait d'ailleurs pouvoir dire que Motif = Molecule de faon  utiliser
    les symtries aussi pour les espces cristallises (complexes essentiellement).
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "systeme.h"
#include "graphes.h"

#include "atomes.h"
#include "cristallo.h" /* Pour la maille */
#include "exafs.h"
#include "fichiers.h" /* Pour sauvegarde, quand effacement */
#include "interface.h" /* DEVRA DISPARATRE  TERME --- Pour I_MOL_Sauver() */
#include "modeles.h"  /* Pour effacement des modles */
#include "science.h"  /* Au cas o M_PI n'est pas dfini */

/* ~~~~~~~~ Variables locales  ce module ~~~~~~~~ */

static const char *titre_tout_effacer = "Effacer toutes les mol"e_AIGU"cules en m"e_AIGU"moire";

/* ~~~~~~~~ Fonctions locales  ce module ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique du module ~~~~~~~ */

/* -------- Cration d'une nouvelle molcule ---------------------------\
 |   Renvoie un pointeur sur la molcule cre                          |
 |          NULL en cas d'erreur  la cration                          |
 |   Initialement, la molcule n'est constitu d'aucun fragment         |
 |                                                                      |
 | nom : le nom  donner  la molcule                                  |
 \--------------------------------------------------------------------- */

Molecule *MOL_creer(const char *nom)
{
  Molecule *molecule;

  molecule = (Molecule *) malloc( sizeof(Molecule) );
  if (molecule == NULL) {
    ERR_Memoire( "Structure Molecule", sizeof( Molecule ) );
    return NULL;
  }

  /* On prpare la molcule */
  molecule->fragments = NULL;  /* Aucun fragment initialement */
  molecule->nbr_fragments = 0;

  molecule->nbr_atomes = 0; /* Aucun atome pour l'instant */
  molecule->premier = NULL;

  molecule->masse_molaire = 0.0;
  molecule->pos_barycentre.x = 0.0;
  molecule->pos_barycentre.y = 0.0;
  molecule->pos_barycentre.z = 0.0;

  molecule->maille = NULL; /* Pas d'informations cristallographiques maille */

  /* On nomme la molcule */
  molecule->nom = NULL;
  if ( nom != NULL ) {
    MOL_RenommerMolecule( molecule, nom );
  } else {
    MOL_RenommerMolecule( molecule, "<Sans nom>" );
  }

  /* Pour les liaisons : on ne suppose rien... donc aucune liaison a priori */
  molecule->nbr_liaisons = 0;
  molecule->liaisons = NULL;

  /* Cas des absorbeurs & modles */
  molecule->absorbeurs = NULL;
  molecule->nbr_absorbeurs = 0;
  molecule->absorbeur_courant = 0;
  molecule->modele = NULL;

  /* Quelques drapeaux de base */
  molecule->drapeaux.sauve   = FAUX;
  molecule->drapeaux.marque  = FAUX;
  molecule->drapeaux.protege = FAUX;
  molecule->drapeaux.visible = VRAI;

  /* On s'occupe du chanage des molcules (liste simplement chane) */
  molecule->suivante = NULL;
  if ( premiere_molecule != NULL ) { /* Il y a dj des molcules existantes
					-> on ajoute la nouvelle  la fin de la liste */
    Molecule *m = premiere_molecule;

    while ( m->suivante )
      m = m->suivante;
    m->suivante = molecule;

    if ( fen_vue_3d != NULL ) { /* On adapte l'ascenseur comme il faut */
      float taille;

      taille = 1000.0 / (float) (nombre_molecules + 1);
      /* position = 1000.0 * ((float) (num_molecule - 1)) / (float) (nombre_molecules + 1)); */
      FEN_ascenseur(fen_vue_3d, ASC_HORIZONTAL, -1, (int) taille);
    }
  } else {
    premiere_molecule = molecule;
  }
  nombre_molecules++; /* Une molcule de plus */

/*   MOL_PrepareListe(); */
  return molecule;
}

int MOL_RenommerMolecule(Molecule *molecule, const char *nouveau_nom)
{
  if ( NULL == molecule ) return 0;
  if ( NULL == nouveau_nom ) return -1;

  /* On libre l'ancien nom */
  free( molecule->nom ); molecule->nom = NULL;

  /* On _copie_ le nouveau,  librer avec free() */
  molecule->nom = strdup( nouveau_nom );
  return 0;
}

/* -------- Effacer toutes les molcules en mmoire ----------------- */

boolean MOL_tteffacer(void)
{
  int reponse;
  boolean tt_sauve = FAUX, tt_perdre = FAUX;
  Molecule *molecule;

  if (premiere_molecule == NULL) return VRAI;

  met_chaine(ARB_QUITTER, FIN_PRG_TITRE, 0,
	     "Fin du programme - Attention, mol"e_AIGU"cule non sauv"e_AIGU"e", TXT_REMPLACE);
  OBJ_cache(ARB_QUITTER, QUIT_NUM_ATOM, FORCE_OUI); /* Pour les atomes... */
  OBJ_cache(ARB_QUITTER, QUIT_X_ATOME, FORCE_OUI); /* Pour les atomes... */
  OBJ_cache(ARB_QUITTER, QUIT_Y_ATOME, FORCE_OUI); /* Pour les atomes... */
  OBJ_cache(ARB_QUITTER, QUIT_Z_ATOME, FORCE_OUI); /* Pour les atomes... */
#ifndef _DEVELOPPE_
  OBJ_active(ARB_QUITTER, QUIT_TTSAUVER, FORCE_NON);
#endif

  /* On efface toutes les molcules, sans se proccuper des graphes */
  while (premiere_molecule != NULL)
  {
    molecule = premiere_molecule;

    if (tt_perdre == VRAI) /* On fait comme si elle tait sauve */
      molecule->drapeaux.sauve = VRAI;

    if ( molecule->drapeaux.sauve == FAUX ) /* Molcule non-sauve */
    {
      if (tt_sauve == VRAI) /* On sauve sans demander */
      {
      } else {
	met_chaine(ARB_QUITTER, QUIT_NOM_GRAPHE, 0, molecule->nom, TXT_REMPLACE);
	reponse = ARB_Bloquant( ARB_QUITTER, titre_tout_effacer, principale, 0 );
	switch (reponse)
	{
	case QUIT_RESTER: return FAUX; /* On annule */
	case QUIT_TTSAUVER:
	  tt_sauve = VRAI;
	case QUIT_SAUVER:
	  I_MOL_Sauver( molecule );
	  break;

	case QUIT_TTPERDRE:
	  tt_perdre = VRAI;
	}
      }
      molecule->drapeaux.sauve = VRAI; /* De toute fano, ici, on a sauv */
    }

    MOL_effacer(molecule, VRAI, VRAI);
  }

  return VRAI;
}

/* -------- Effacer une molcule -------------------------------------\
 |   Renvoie 0 si tout va bien, <0 en cas d'erreur                    |
 |                                                                    |
 | molecule : la molcule que l'on veut effacer                       |
 | sauver   : VRAI si on force l'effacement si elle n'est pas sauve  |
 | forcer   : VRAI si on force l'effacement quoi qu'il arrive         |
 \------------------------------------------------------------------- */

int MOL_effacer(Molecule *molecule, boolean sauver, boolean forcer)
{
  unsigned int i;

  if (molecule == NULL) return 0;
  if ( (NULL == premiere_molecule) || /* Molcule hors chanage : arrive quand on efface le dernier atome de la molcule
					 ==> efface la molcule, mais si dj en cours d'effacement... */
       ( (molecule->suivante == NULL ) && (molecule != premiere_molecule) ) ) {
    return -1;
  }

  if ( ( molecule->drapeaux.protege == VRAI ) /* Molcule en cours d'utilisation */
    && ( forcer == FAUX ) ) { /* Et on s'en proccupe */
    (void) fprintf(stderr, "MOL_effacer : molcule en cours d'utilisation, protge...\n");
    return -2;
  }

  if ( ( molecule->drapeaux.sauve == FAUX ) /* Molcule non sauve */
    && ( sauver == FAUX ) ) /* On s'en proccupe donc */
  {
    int reponse;
                              /* v-----------------------------v */
    reponse = form_alert(1, "[2][---- Molcule non sauv"e_AIGU"e ----"
			       "|  Voulez-vous la sauver avant"
			       "|de l'effacer, ou annuler"
       			       "|l'effacement ?][Effacer| Sauver | Annuler]");
    if (reponse == 3) return FAUX;
    if (reponse == 2) {
      if ( I_MOL_Sauver( molecule ) == FAUX)
	return FAUX; /* Sauvegarde a chou : on annule l'effacement */
    }
  }

  /* On limine les liaisons intra-molculaires, en premier car n'efface pas les atomes
     mais vite ensuite pas mal de vrifications -> devrait faire gagner du temps
     AVANT d'extraire du chanage, pour pouvoir utiliser effacer_liaison correctement
  */
  while (molecule->liaisons != NULL)
    effacer_liaison(molecule->liaisons, VRAI);

  /* On extrait la molcule  effacer du chanage */
  if (molecule == premiere_molecule) {/* On efface la premire molcule */
    premiere_molecule = molecule->suivante; /* Nouvelle premire molcule */
  } else {
    Molecule *avant;

    avant = premiere_molecule;
    while (avant->suivante != molecule) {
      avant = avant->suivante;
      if (NULL == avant) {
	(void) fprintf(stderr, "MOL_effacer : molcule hors chanage ! [%s]\n", molecule->nom);
#ifdef _DEVELOPPE_
	MOL_Controle_Affiche_Structure(molecule, 80);
	abort();
#else
	return -10;
#endif
      }
    }

    avant->suivante = molecule->suivante; /* Extraction hors du chanage ralise */
    molecule->suivante = NULL; /* Pour qu'elle soit vraiment hors chanage */
  }
  nombre_molecules--; /* Puisqu'elle est hors chanage, elle ne compte plus */

  /* Si c'est la dernire molcule, on ferme la vue 3D (avant d'effacer les atomes, car vite des redessins continuels !) */
  if (premiere_molecule == NULL) { /* Dernire molcule */
    Message message;

    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = -1; message.fenetre.y = -1;
    message.fenetre.larg = -1; message.fenetre.haut = -1; /* Fermeture force... */

    if (fen_vue_3d != NULL) { /* On ferme proprement... */
      (fen_vue_3d->message)(fen_vue_3d, &message);
    }
    mol_courante = NULL; /* Plus de molcule courante */
  } else { /* Taille de l'ascenseur de la fentre 3D */
    if (fen_vue_3d != NULL) {
      float taille = 1000.0 / (float) (nombre_molecules);

      FEN_ascenseur(fen_vue_3d, ASC_HORIZONTAL, -1, (int) taille);
    }

    if (molecule == mol_courante) {
      if (molecule->suivante)
	MOL_aller(molecule->suivante);
      else
	MOL_aller(premiere_molecule); /* Il faudrait en fait celle juste avant, mais bon... */
    }
  }

  /* --- On efface maintenant rellement la molcule --- */
  /* On efface le(s) modle(s) */
  while ( molecule->modele ) {
    (void) MODELE_EffacerModele( &(molecule->modele), molecule->modele, FAUX );
  }

  /* On efface les informations sur les absorbeurs
     En premier car comme a limine les chemins sans effacer les atomes,
     donc moins de vrifications en effaant les fragments : devrait tre plus rapide et plus fiable */
#ifdef _DEVELOPPE_
  printf("Molcule %s : effacement des absorbeurs\n", molecule->nom);
#endif
  for (i = molecule->nbr_absorbeurs ; i > 0 ; i--)
    ABS_oter(molecule, i - 1, FAUX); /* On n'efface pas l'atome associ */
#ifdef _DEVELOPPE_
  printf("              ... on libre donc %p\n", (void *) molecule->absorbeurs);
#endif
  free( molecule->absorbeurs );
  molecule->absorbeurs = NULL;

  /* On limine les fragments
     On ne peut pas compter sur la correction automatique du chanage,
     car la molcule n'apparat plus dans la chane et est donc invisible
     On doit donc suivre le chanage  la main...
  */
#ifdef _DEVELOPPE_
  printf("Molcule %s : effacement des fragments\n", molecule->nom);
  i = 0;
#endif
  while ( molecule->fragments != NULL ) { /* while ncessaire, car molecule->nbr_fragments change avec l'effacement ! */
#ifdef _DEVELOPPE_
    i++;
    printf("  mol_effacer, effacement du fragment %d demand\n", i);
#endif
    FRG_effacer(molecule, molecule->fragments, VRAI, FAUX ); /* On n'efface pas les atomes */
  }

  /* On efface les atomes. Devrait tre rapide, car plus rien dans la molcule ! */
  while ( molecule->premier ) {
    ATM_efface(molecule, molecule->premier, VRAI);
  }

  /* On efface les informations cristallographiques */
  CRIS_EffacerMaille( molecule->maille );
  molecule->maille = NULL;

  /* On efface le nom de la molcule */
  free( molecule->nom ); molecule->nom = NULL;

  /* On efface la molcule elle-mme */
  free(molecule);
  MOL_PrepareListe();
  return 0; /* Tout s'est bien pass */
}

/* -------- Ajouter un fragment  une molcule dj existante --------\
 |   Renvoie 0 si tout s'est bien pass, <0 sinon                     |
 |  Les fragments sont ajouts au dbut de la liste de fragments      |
 |  Les fragments ajouts ne doivent pas dj appartenir  une liste! |
 |                                                                    |
 | molecule      : la molcule o ajouter le fragment                 |
 | nbr_fragments : combien de fragments faut-il ajouter ?             |
 | fragments     : la liste des fragments  ajouter                   |
 \------------------------------------------------------------------- */

int MOL_ajoute_fragment(Molecule *molecule, unsigned int nbr_fragments, Fragment **fragments)
{
  unsigned int i;

  if (molecule == NULL) return 0;

  if ( (fragments == NULL) || (nbr_fragments == 0) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "Liste de fragments passe nulle !\n");
#endif
    return -1;
  }

  /* On se place sur le dernier fragment existant de la molcule */
  for (i = 0 ; i < nbr_fragments ; i++) {
    if ( fragments[i] == NULL ) {
      (void) fprintf(stderr, "MOL_ajoute_fragment : un des fragments de la liste n'existe pas\n");
    } else {
      if ( fragments[i]->suivant == NULL ) {
	fragments[i]->suivant = molecule->fragments;
	molecule->fragments = fragments[i];
	molecule->nbr_fragments++; /* Un fragment en plus */
      }
#ifdef _DEVELOPPE_
      else {
	printf("Fragment pass : dj contenu dans une liste de fragments\n");
	printf("En effet, il pointe dj sur un fragment ultrieur !\n");
	printf("Fragment : %s\n", fragments[i]->nom);
      }
#endif
    }
  }

#ifdef _DEVELOPPE_
  {
    Fragment *frg;
    printf("Nouveau chanage des fragments dans la molcule %s\n", molecule->nom);
    printf("  Total attendu = %u fragments\n", molecule->nbr_fragments);
    for (i = 0, frg = molecule->fragments ; frg != NULL ; i++, frg = frg->suivant) {
      printf("   fragment %d : %s\n", i, frg->nom);
    }
    printf("  Total rel = %u fragments\n", i);
  }
#endif

  return 0; /* Pas de problme */
}

int MOL_efface_fragment(Molecule *molecule, Fragment *fragment)
{
  if (molecule == NULL) return 0;
  if (fragment == NULL) return 0;
  if (molecule->fragments == NULL) return -1;

  if (molecule->fragments == fragment) /* On te le premier fragment */
    molecule->fragments = fragment->suivant;
  else {
    Fragment *avant = molecule->fragments;

    while (avant->suivant != fragment) {
      avant = avant->suivant;
      if (avant == NULL) return -1; /* Mais trange ! */
    }
    avant->suivant = fragment->suivant;
  }
  fragment->suivant = NULL; /* Fragment entirement hors-chanage */

  molecule->nbr_fragments--;
  return 0;
}

/* -------- Enlever les marques internes (marques->mk_1 et marques->mk_2) pour tous les atomes d'une molcule -------- */

void MOL_tt_demarquer(Molecule *molecule)
{
  Fragment *fragment;
  unsigned int i;

  if (molecule == NULL) return;

  /* On te la marque (marque mk_1) des absorbeurs */
  for (i = 0 ; i < molecule->nbr_absorbeurs ; i++) {
    molecule->absorbeurs[i].absorbeur->marques.mk_2 = FAUX;
    molecule->absorbeurs[i].absorbeur->marques.mk_1 = FAUX;
  }

  /* On te la marque des atomes des fragments */
  for (fragment = molecule->fragments ; fragment != NULL ; fragment = fragment->suivant ) {
    for (i = 0 ; i < fragment->nbr_atomes ; i++) {
      fragment->atomes[i]->marques.mk_2 = FAUX;
      fragment->atomes[i]->marques.mk_1 = FAUX;
    }
  }
}

/* -------- Recherche du barycentre de la molcule --------
    Calcule aussi les barycentres partiels (de chaque fragment)

    Si centrer = VRAI, on cherche  recentrer la molcule aprs
*/

int MOL_cherche_barycentre(Molecule *molecule, boolean centrer)
{
  double masse = 0.0, x_g = 0.0, y_g = 0.0, z_g = 0.0;
  Atome *atome;

  if (molecule == NULL) return -1;

  for ( atome = molecule->premier ; atome != NULL ; atome = atome->suivant ) {
    double m_atm;

    m_atm = (double) ATM_Z_vers_masse(atome->num_atomique, ATM_MASSE_MOLAIRE);
    masse += m_atm;
    x_g += m_atm * atome->position.x;
    y_g += m_atm * atome->position.y;
    z_g += m_atm * atome->position.z;
  }

  if (masse <= 0.0) {
    (void) fprintf(stderr, "Molcule de masse nulle ! Impossible !\n");
    return -1;
  }

  molecule->pos_barycentre.x = x_g / masse; /* Coordoonnes du barycentre */
  molecule->pos_barycentre.y = y_g / masse;
  molecule->pos_barycentre.z = z_g / masse;

  if (centrer == VRAI) {
    /* Pour centrer, on ralise une translation */
    for ( atome = molecule->premier ; atome != NULL ; atome = atome->suivant ) {
      atome->position.x -= molecule->pos_barycentre.x;
      atome->position.y -= molecule->pos_barycentre.y;
      atome->position.z -= molecule->pos_barycentre.z;
    }

    molecule->pos_barycentre.x = 0.0; /* Coordoonnes du barycentre : origine du systme */
    molecule->pos_barycentre.y = 0.0;
    molecule->pos_barycentre.z = 0.0;
  }

#ifdef _DEVELOPPE_
/*   printf("MOL_cherche_barycentre : barycentre en (%f, %f, %f)\n", molecule->x_g, molecule->y_g, molecule->z_g); */
#endif
  return 0;
}

int MOL_initialise_vue(Molecule *molecule)
{
  double x_min = 50000.0, y_min = 50000.0, z_min = 50000.0,
    x_max = -50000.0, y_max = -50000.0, z_max = -50000.0,
    d_x, d_y, d_z;
  Atome *atome;

  if ( ( NULL == molecule ) || ( NULL == molecule->premier ) ) {
    /* Pas d'atome dans la molcule : pas de bote englobante ! */
    return -1;
  }

  /* Le premier atome sert  dfinir la bote */
  x_min = molecule->premier->position.x; x_max = x_min;
  y_min = molecule->premier->position.y; y_max = y_min;
  z_min = molecule->premier->position.z; z_max = z_min;

  /* On tient compte du rayon, s'il n'est pas absurde ! */
  if ( molecule->premier->rayon > 0.0 ) {
    double rayon;

    rayon = molecule->premier->rayon / 100.0;  /* rayon en pm, 100 pm = 1 A, coordonnes en A */
    x_min -= rayon; x_max += rayon;
    y_min -= rayon; y_max += rayon;
    z_min -= rayon; z_max += rayon;
  }

  /* On cherche le pav droit, align sur les axes du repre, englobant la molcule au mieux */
  for ( atome = molecule->premier->suivant ; atome != NULL ; atome = atome->suivant ) {

    /* On tient compte du rayon, s'il n'est pas absurde ! */
    if ( atome->rayon > 0.0 ) {
      double rayon = atome->rayon / 100.0;

      if ( ( atome->position.x - rayon ) < x_min ) x_min = atome->position.x - rayon;
      if ( ( atome->position.y - rayon ) < y_min ) y_min = atome->position.y - rayon;
      if ( ( atome->position.z - rayon ) < z_min ) z_min = atome->position.z - rayon;
      if ( ( atome->position.x + rayon ) > x_max ) x_max = atome->position.x + rayon;
      if ( ( atome->position.y + rayon ) > y_max ) y_max = atome->position.y + rayon;
      if ( ( atome->position.z + rayon ) > z_max ) z_max = atome->position.z + rayon;
    } else {
      if ( atome->position.x < x_min ) x_min = atome->position.x;
      if ( atome->position.y < y_min ) y_min = atome->position.y;
      if ( atome->position.z < z_min ) z_min = atome->position.z;
      if ( atome->position.x > x_max ) x_max = atome->position.x;
      if ( atome->position.y > y_max ) y_max = atome->position.y;
      if ( atome->position.z > z_max ) z_max = atome->position.z;
    }
  }

  /* On s'arrange pour que la maille soit visible aussi
     En effet, il semble que parfois, sans cela, on puisse avoir de mchants plantages
     si la maille dborde trop de la vue ( VRIFIER - observ avec Mesa/linux)
     Apparemment, ce n'est pas cela. Qu'est-ce alors ??
  */
  if ( molecule->maille != NULL ) { /* Il y a bien une maille */
    Vecteur3D v_a, v_b, v_c;
    double x_100, x_010, y_010, x_001, y_001, z_001;

    /* Coordonnes des coins */
    CRIS_VecteursMaille( molecule->maille, &v_a, &v_b, &v_c );
    x_100 = v_a.x;
    x_010 = v_b.x; y_010 = v_b.y;
    x_001 = v_c.x; y_001 = v_c.y; z_001 = v_c.z;

    if ( 0 < x_min) x_min = 0; if (0 > x_max) x_max = 0; /* Coin Bas, Avant, Gauche = origine*/
    if ( 0 < y_min) y_min = 0; if (0 > y_max) y_max = 0;
    if ( 0 < z_min) z_min = 0; if (0 > z_max) z_max = 0;
    if ( x_100 < x_min ) x_min = x_100; if ( x_100 > x_max ) x_max = x_100; /* Coin B A Droit (x_100, 0, 0) */
    if ( x_010 < x_min ) x_min = x_010; if ( x_010 > x_max ) x_max = x_010; /* Coin B aRrire G (x_010, y_010, 0) */
    if ( y_010 < y_min ) y_min = y_010; if ( y_010 > y_max ) y_max = y_010;
    if ( x_001 < x_min ) x_min = x_001; if ( x_001 > x_max ) x_max = x_001; /* Coin Haut A D (x_001, y_001, z_001) */
    if ( y_001 < y_min ) y_min = y_001; if ( y_001 > y_max ) y_max = y_001;
    if ( z_001 < z_min ) z_min = z_001; if ( z_001 > z_max ) z_max = z_001;
    if ( x_100 + x_010 < x_min ) x_min = x_100 + x_010; if ( x_100 + x_010 > x_max ) x_max = x_100 + x_010; /* Coin B R D */
    if ( x_100 + x_001 < x_min ) x_min = x_100 + x_001; if ( x_100 + x_001 > x_max ) x_max = x_100 + x_001; /* Coin H A D */
    if ( x_010 + x_001 < x_min ) x_min = x_010 + x_001; if ( x_010 + x_001 > x_max ) x_max = x_010 + x_001; /* Coin B R D */
    if ( y_010 + y_001 < y_min ) y_min = y_010 + y_001; if ( y_010 + y_001 > y_max ) y_max = y_010 + y_001;
    if (x_100+x_010+x_001 < x_min) x_min = x_100+x_010+x_001; if (x_100+x_010+x_001 > x_max) x_max = x_100+x_010+x_001; /* HRD */
#ifdef _DEVELOPPE_
    printf("Init vue : la maille a t prise en compte\n");
#endif
  }

#ifdef _DEVELOPPE_ 
  printf("Bbox [(%f, %f, %f) (%f, %f, %f)]\n", x_min, y_min, z_min, x_max, y_max, z_max);
#endif

  /* On adapte la vue en consquence.
     Si une dimension est nulle (molcule linaire ou plane), on force une taille minimale
     pour tre sr de voir quelque chose */
  d_x = x_max - x_min; if (d_x < 1.0) d_x = 5.0;
  d_y = y_max - y_min; if (d_y < 1.0) d_y = 5.0;
  d_z = z_max - z_min; if (d_z < 1.0) d_z = 5.0;

  molecule->point_vise.x = (x_min + x_max) / 2.0;
  molecule->point_vise.y = (y_min + y_max) / 2.0;
  molecule->point_vise.z = (z_min + z_max) / 2.0;
  molecule->angle1 = 20.0; /* Ouverture de 20 degrs - OK selon POV */

  molecule->pos_oeil.x = (x_min + x_max) / 2.0; /* Au centre du pav droit entourant la molcule */
  molecule->pos_oeil.y = (y_min + y_max) / 2.0;
  molecule->pos_oeil.z = (z_min + z_max) / 2.0;

  if ( (d_z > d_x) && (d_z > d_y) ) /* Axe Z le plus utilis -> horizontal sur l'cran */
  {
    molecule->z_h = 0.0;
    if (d_y > d_x) /* Axe Y second axe, vertical  l'cran : oeil sur l'axe X */
    {
      molecule->pos_oeil.x += d_z / (2 * tan( molecule->angle1 * M_PI / 360.0 ) ) + d_x / 2.0;
      molecule->x_h = 0.0;
      molecule->y_h = 1.0;
      molecule->angle2 = 2 * atan( d_y/d_z * tan( molecule->angle1 * M_PI / 360.0 ) ) * 180.0 / M_PI;
    }
    else /* Axe X second axe, vertical  l'cran : oeil sur l'axe Y */
    {
      molecule->pos_oeil.y += d_z / (2 * tan( molecule->angle1 * M_PI / 360.0 ) ) + d_y / 2.0;
      molecule->x_h = 1.0;
      molecule->y_h = 0.0;
      molecule->angle2 = 2 * atan( d_x/d_z * tan( molecule->angle1 * M_PI / 360.0 ) ) * 180.0 / M_PI;
    }
  }
  else if ( (d_y > d_z) && (d_y > d_x) ) /* Axe Y le plus utilis, horizontal sur l'cran */
  {
    molecule->y_h = 0;
    if (d_z > d_x) /* oeil sur l'axe des X */
    {
      molecule->pos_oeil.x += d_y / (2 * tan( molecule->angle1 * M_PI / 360.0 ) ) + d_x / 2.0;
      molecule->x_h = 0.0;
      molecule->z_h = 1.0;
      molecule->angle2 = 2 * atan( d_x/d_y * tan( molecule->angle1 * M_PI / 360.0 ) ) * 180.0 / M_PI;
    }
    else /* oeil sur l'axe des Z */
    {
      molecule->pos_oeil.z += d_y / (2 * tan( molecule->angle1 * M_PI / 360.0 ) ) + d_z / 2.0;
      molecule->x_h = 1.0;
      molecule->z_h = 0.0;
      molecule->angle2 = 2 * atan( d_z/d_y * tan( molecule->angle1 * M_PI / 360.0 ) ) * 180.0 / M_PI;
    }
  } else { /* Axe X le plus utilis */
    printf( "%s : Axe X le plus utilis...\n", __FUNCTION__ );
    molecule->x_h = 0;
    if ( d_z > d_y ) { /* Axe Z second axe principal : oeil sur l'axe des Y */
      printf( "%s : puis l'axe Z.\n", __FUNCTION__ );
      molecule->pos_oeil.y += d_x / (2 * tan( molecule->angle1 * M_PI / 360.0 ) ) + d_y / 2.0;
      molecule->y_h = 0;
      molecule->z_h = 1.0;
      molecule->angle2 = 2 * atan( d_y/d_z * tan( molecule->angle1 * M_PI / 360.0 ) ) * 180.0 / M_PI;
    } else { /* Axe Y second axe principal : oeil sur l'axe des Z */
      printf( "%s : puis l'axe X.\n", __FUNCTION__ );
      molecule->pos_oeil.z += d_x / (2 * tan( molecule->angle1 * M_PI / 360.0 ) ) + d_z / 2.0;
      molecule->y_h = 1.0;
      molecule->z_h = 0;
      molecule->angle2 = 2 * atan( d_y/d_x * tan( molecule->angle1 * M_PI / 360.0 ) ) * 180.0 / M_PI;
    }
  }

  printf( "Caractristiques retenues : Oeil en (%g, %g, %g) ; Vecteur vertical (%g, %g, %g)\n", 
	  molecule->pos_oeil.x, molecule->pos_oeil.y, molecule->pos_oeil.z, molecule->x_h, molecule->y_h, molecule->z_h );

  return 0;
}

/* ~~~~~~~~ Fonctions de type cristallographie ~~~~~~~~ */

/* -------- Cration d'une maille supplmentaire dans la molcule ------
     Renvoie le fragment contenant cette nouvelle maille (NULL en cas d'erreur)
     Elle vrifie que les atomes ainsi crs n'existent pas dj

     La molcule indique doit possder les paramtres de maille et
     (dans l'tat actuel) une maille centrale, (0, 0, 0) qui est copie

 molecule : la molcule  laquelle ajouter la maille
 i, j, k  : position (x, y, z) de la maille  ajouter
 -------- */

Fragment *MOL_creer_maille(Molecule *molecule, int i, int j, int k)
{
  char titre[LG_MAX_NOM_FRAGMENT + 100];
  Fragment *nv_maille = NULL, *maille_centre;
  double c_a, c_b, c_g, s_g;
  double x_100, x_010, y_010, x_001, y_001, z_001;
  unsigned int compteur;

  if (molecule == NULL) return NULL;
  if ( NULL == molecule->maille ) {
    (void) fprintf(stderr, "%s,%s l.%u : pas de paramtres cristallins connus !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  if ( (i == 0) && (j == 0) && (k == 0) ) {
    fprintf(stderr, "MOL_creer_maille : la maille (0, 0, 0) existe dj !\n");
    return NULL;
  }

  maille_centre = MOL_centre(molecule);
  if (maille_centre == NULL) return NULL;

  /* On cre le nouveau fragment */
  sprintf(titre, "Maille (%d, %d, %d)", i, j, k);
  nv_maille = FRG_creer(molecule, NULL, titre, maille_centre->nbr_atomes, NULL, NULL);
  if (nv_maille == NULL)
    return NULL;

  /* Calcul des composantes des vecteurs lmentaires de translation */
  c_a = cos( molecule->maille->alpha );
  c_b = cos( molecule->maille->beta );
  c_g = cos( molecule->maille->gamma ); s_g = sin( molecule->maille->gamma );
  x_100 = molecule->maille->a;
  x_010 = c_g * molecule->maille->b;
  y_010 = molecule->maille->b * s_g;
  x_001 = c_b * molecule->maille->c;
  y_001 = molecule->maille->c * ( c_a - c_b * c_g ) / s_g;
  z_001 = molecule->maille->c * sqrt( fabs(1 - c_a*c_a - c_b*c_b - c_g*c_g + 2 * c_a * c_b * c_g) ) / s_g;

  for (compteur = 0 ; compteur < maille_centre->nbr_atomes ; compteur++) {
    Atome *atome = maille_centre->atomes[compteur], *nv_atome;
    Vecteur3D nv_position;

    nv_position.x = atome->position.x + x_100 * (double) i + x_010 * (double) j + x_001 * (double) k;
    nv_position.y = atome->position.y +                      y_010 * (double) j + y_001 * (double) k;
    nv_position.z = atome->position.z +                                           z_001 * (double) k;
    nv_atome = ATM_chercher( &nv_position, atome->num_atomique, molecule);
    if (nv_atome == NULL) /* Atome inexistant : on le cre */
      nv_atome = ATM_creer(molecule, NULL, atome->num_atomique, &nv_position, atome->charge, atome->rayon);
#ifdef _DEVELOPPE_
    else
      printf("Atome dj existant...\n");
#endif

    FRG_ajoute_atome(nv_maille, 1, &nv_atome);
  }

  /* Il faudrait aussi recopier le schma de liaisons... */

  /* On recalcule ce qui est ncessaire */
  MOL_cherche_barycentre(molecule, FAUX);
  MOL_initialise_vue(molecule);
  return nv_maille;
}

/* ------ Renvoie le fragment contenant la maille centrale, s'il existe ------ */

Fragment *MOL_centre(Molecule *molecule)
{
  Fragment *fragment;
  if (molecule == NULL) return NULL;

  for (fragment = molecule->fragments ; fragment != NULL ; fragment = fragment->suivant)
    if ( (fragment->drapeaux.maille == VRAI) && (fragment->drapeaux.origine == VRAI) )
      return fragment; /* On a trouv la maille centrale */

  return NULL; /* Ici : aucun fragment ne convient */
}

#ifdef _DEVELOPPE_

void MOL_Controle_Affiche_Structure(Molecule *molecule, int niveau)
{
  Fragment *fragment;
  unsigned long i;

  printf("\n-------- Contrle et affichage de la structure d'une molcule --------\n");

  if (NULL == molecule) {
    printf("** Molcule passe inexistante (NULL) **\n");
    return;
  }

  /* Gnralits */
  printf("Molcule : %s\n", molecule->nom);
  printf("Contient : %u fragments  [%p]\n", molecule->nbr_fragments , (void *) molecule->fragments );
  printf("           %u liaisons   [%p]\n", molecule->nbr_liaisons  , (void *) molecule->liaisons  );
  printf("           %u absorbeurs [%p]\n", molecule->nbr_absorbeurs, (void *) molecule->absorbeurs);
  if ( NULL == molecule->premier ) {
    printf("           0 = %lu atomes\n", molecule->nbr_atomes);
  } else {
    printf("           %lu atomes [%p]\n", molecule->nbr_atomes, (void *) molecule->premier);
  }

  if (niveau < 1) return;
  printf("\n");

  /* Informations chimiques  */
  printf("Masse molaire : %f g/mol\n", molecule->masse_molaire);
  if ( molecule->maille != NULL ) {
    printf("Maille        : a = %f A, b = %f A, c = %f A, alpha = %f, beta = %f, gamma = %f\n",
	   molecule->maille->a, molecule->maille->b, molecule->maille->c,
	   molecule->maille->alpha * 180.0 / M_PI, molecule->maille->beta * 180.0 / M_PI, molecule->maille->gamma * 180.0 / M_PI);
/*     printf("                groupe d'espace = %d\n", molecule->maille.grp_esp); */
  }
  printf("Barycentre    : (%f, %f, %f)\n", molecule->pos_barycentre.x, molecule->pos_barycentre.y, molecule->pos_barycentre.z);

  /* Vue en 3D */
  printf("OEil en    (%f, %f, %f)\n", molecule->pos_oeil.x, molecule->pos_oeil.y, molecule->pos_oeil.z);
  printf("regarde en (%f, %f, %f)\n", molecule->point_vise.x, molecule->point_vise.y, molecule->point_vise.z);

  if (niveau < 2) return;
  printf("\n");

  /* Infos sur les fragments */
  for (i = 0, fragment = molecule->fragments ; fragment != NULL ; i++, fragment = fragment->suivant) {
    printf("Fragment %lu : %s\n", i + 1, fragment->nom);
    printf("Contient : %u atomes (sur %u rservs) [%p]\n", fragment->nbr_atomes, fragment->max_atomes, (void *) fragment->atomes);
    printf("           %u liaisons [%p]\n", fragment->nbr_liaisons, (void *) fragment->liaisons);

    if (niveau < 3) continue;
  }
  printf("==> Total : %lu fragments rels\n\n", i);

  /* Infos sur les absorbeurs */
  for (i = 0 ; i < molecule->nbr_absorbeurs ; i++) {
    Absorbeur *absorbeur = molecule->absorbeurs + i;

    printf("Absorbeur : %s en (%f, %f, %f)\n", ATM_Z_vers_symbole(absorbeur->absorbeur->num_atomique),
	   absorbeur->absorbeur->position.x, absorbeur->absorbeur->position.y, absorbeur->absorbeur->position.z);
/*    printf("Contient %u chemins [%p]\n", absorbeur->nbr_chemins, absorbeur->chemins); */
  }
}

#endif
