/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2001, Emmanuel CURIS
   Création du fichier : 21 décembre 2001

   Tout ce qui concerne l'extraction des oscillations EXAFS

   Pondération d'un spectre par k^n (« magnification ») [Récupéré de EXAFS.c]
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h> /* Pour isspace et isdigit */

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"
#include "extrait.h"
#include "matrices.h"

#include "exf_ponderer.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#ifdef _TOS_ /* Codage des accents TOS */
# define AU_CARRE 'ý'
# define AU_CUBE  'þ'
#endif

#if defined(_UNIX_) || defined(_WINDOWS_) /* Codage des accents en Latin-1 */
# ifdef CODAGE_LATIN1_
#  define AU_CARRE '\262' /* ², en latin-1, octal = 262 */
#  define AU_CUBE  '\263' /* ³, en latin-1, octal = 263 */
# endif
# ifdef CODAGE_UTF8_
#  define AU_CARRE '²'
#  define AU_CUBE  '³'
# endif
#endif

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int EXF_PondereComplexe(Graphe *graphe, int degre);
static int EXF_PondereReel    (Graphe *graphe, int degre);

static char *KPN_adapte_legende(const char *legende, int degre); /* Adaptation de la légende en fonction de la pondération par k^n */

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -~-~-~-~-~-~-~-~ Calcul effectif des pondérations -~-~-~-~-~-~-~-~ */

/* -------- Cas d'un graphe à valeurs réelles -------- */

static int EXF_PondereReel(Graphe *graphe, int degre)
{
  Graphe *erreur = NULL;
  double *x, *y, *y_err = NULL;
  unsigned long i;
  int sortie = 0;

  if ( NULL == graphe ) return -1;

  x = graphe->x;
  y = graphe->y;

  erreur = graphe->erreur_y;
  if ( erreur != NULL )  /* Il y a des erreurs en Y, on doit les adapter */
    y_err = erreur->y;

  if ( degre > 0 ) { /* Aucune vérification à faire */
    unsigned long i;
    unsigned short int j;

    for ( i = 0 ; i < graphe->nombre_point ; i++, x++, y++ ) {
      double x_p_n;

      x_p_n = *x;
      for ( j = 1 ; j < degre ; j++ )
	x_p_n *= *x;

      *y *= x_p_n;

      if ( erreur != NULL ) { /* On modifie aussi l'erreur */
	*y_err *= fabs( x_p_n );
	y_err++;
      }

      if ( graphe->covariance != NULL ) { /* On modifie aussi la matrice de covariance */
	unsigned int n_col;

	/* Si W est la matrice de pondération, on a S' = W S W'
	   comme W est diagonale, revient à multiplier la ligne i & la colonne i par la valeur (w_i,i)
	   Du fait de la symétrie de S, on se limite au début de ligne et on finit sur la colonne */

	/* On multiplie la ligne i de la matrice
	   1. moitié inférieure: ligne directement */
	for ( n_col = 0 ; n_col <= i ; n_col++ ) {
	  graphe->covariance->valeurs[ MAT_SYM_element(i, n_col) ] *= x_p_n;
	}
	/* 2. moitié supérieure, par symétrie revient à remplir la colonne i */
	for ( ; n_col < graphe->nombre_point ; n_col++ ) {
	  graphe->covariance->valeurs[ MAT_SYM_element(n_col, i) ] *= x_p_n;
	}
      }
    }

    return 0;
  }

  /* Pondération par 1/x^n : il faut vérifier que x != 0 */
  for ( i = 0 ; i < graphe->nombre_point ; i++, x++, y++, y_err++ ) {
    if ( *x == 0.0) { /* On diviserait par zéro */
      sortie = DIV_ZERO; /* On ne modifie pas le y : aucune piste pour savoir quoi choisir ! */
    } else {
      double x_p_n;

      x_p_n = pow( *x, (double) -degre );
      *y /= x_p_n;

      if ( erreur != NULL ) /* On modifie aussi l'erreur */
	*y_err /= fabs( x_p_n );

      if ( graphe->covariance != NULL ) { /* On modifie aussi la matrice de covariance */
	unsigned int n_col;

	/* On multiplie la ligne i de la matrice (cf plus haut...)
	   1. moitié inférieure: ligne directement */
	for ( n_col = 0 ; n_col <= i ; n_col++ ) {
	  graphe->covariance->valeurs[ MAT_SYM_element(i, n_col) ] *= x_p_n;
	}
	/* 2. moitié supérieure, par symétrie revient à remplir la colonne i */
	for ( ; n_col < graphe->nombre_point ; n_col++ ) {
	  graphe->covariance->valeurs[ MAT_SYM_element(n_col, i) ] *= x_p_n;
	}
      }
    }
  }

  return sortie;
}

/* -------- Cas d'un graphe à valeurs complexes -------- */

static int EXF_PondereComplexe(Graphe *graphe, int degre)
{
  Graphe *erreur;
  register double *x;
  register Complexe *y, *y_erreur = NULL;
  unsigned long i;
  int sortie = 0;

  x = graphe->x;
  y = (Complexe *) (graphe->y);
  erreur = graphe->erreur_y;
  if ( erreur != NULL )
    y_erreur = (Complexe *) (erreur->y);

  if (degre < 0) { /* Risque de division par zéro */
    degre = -degre;
    for (i = 0 ; i < graphe->nombre_point ; i++, x++, y++) {
      int j;
      double x_p_n;

      x_p_n = *x;
      for (j = 1 ; j < degre ; j++)
	x_p_n *= *x;

      if ( x_p_n == 0.0 )
	sortie = DIV_ZERO;
      else {
	y->re /= x_p_n;
	y->im /= x_p_n;

	if (erreur != NULL) {
	  x_p_n = fabs(x_p_n);
	  y_erreur->re /= x_p_n;
	  y_erreur->im /= x_p_n;
	  y_erreur++;
	}

	if (graphe->covariance != NULL) { /* On modifie aussi la matrice de covariance */
	  (void) fprintf(stderr, "EXF_PondereComplexe : n < 0, Matrice de covariance, À FAIRE\n");
	}
      }
    }

    return sortie;
  }

  /* Aucun risque de division par zéro */
  for (i = 0 ; i < graphe->nombre_point ; i++, x++, y++) {
    int j;
    double x_p_n;

    x_p_n = *x;
    for (j = 1 ; j < degre ; j++)
      x_p_n *= *x;

    y->re *= x_p_n;
    y->im *= x_p_n;
    if (erreur != NULL) {
      x_p_n = fabs(x_p_n);
      y_erreur->re *= x_p_n;
      y_erreur->im *= x_p_n;
      y_erreur++;
    }

    if ( graphe->covariance != NULL ) { /* On modifie aussi la matrice de covariance */
      OPTION_A_FAIRE( "Pondération d'un graphe complexe, n > 0, adaptation de la matrice de covariance" );
    }
  }

  return 0;
}

/* -------- Modification du texte de la légende pour tenir compte des pondérations

   Renvoie le texte de la nouvelle légende dans une nouvelle chaîne - NULL en cas d'erreur.
   Ne modifie pas la chaîne d'origine
 -------- */

static char *KPN_adapte_legende(const char *legende, int degre)
{
  char *nv_legende;
  size_t lg_legende;

  /* (void) fprintf( stderr, "*** Adaptation de la légende après pondération ***\n" ); */

  if ( NULL == legende ) return NULL;
  if (    0 == degre   ) return NULL; /* Pondération sans effet... */

  /* (void) fprintf( stderr, "%s,%s l.%u : Légende à adapter = %s\n", */
  /* 		  __FILE__, __FUNCTION__, __LINE__, legende ); */

  if ( *legende == 'k' ) { /* On suppose qu'il y a déjà eu une multiplication */
    unsigned long decalage = 0;
    int exposant_existant = 0, avec_exposant = FAUX;
    const char *debut = legende;

    /* On passe au caractère suivant */
    legende++;

    /* On recherche l'éventuel exposant */
    exposant_existant = TXT_DecoderExposant( legende, &decalage );
    if ( decalage > 0 ) {
      /* Des caractères ont bien été utilisés pour calculer un exposant */
      avec_exposant = VRAI;

      legende += decalage;  	  /* on les saute pour la suite */
    } else {
      /* Pas d'exposant, mais peut être exposant = 1 si k * ou k* ou équivalent */
      if ( isspace( *legende ) ) { exposant_existant = 1 ; avec_exposant = VRAI; }
      if ( '*'    == *legende  ) { exposant_existant = 1 ; avec_exposant = VRAI; }
      if ( '\xD7' == *legende  ) { exposant_existant = 1 ; avec_exposant = VRAI; } /* ×, codage latin-1/windows-1252 */
      if ( ( '\xC3' == *legende ) && ( '\x97' == *(legende + 1) ) ) {              /* ×, codage UTF-8 */
	exposant_existant = 1 ; avec_exposant = VRAI;
      }
    }

    if ( VRAI == avec_exposant ) {
      degre += exposant_existant; /* On corrige l'exposant */

      /* On saute les éventuels espaces suivant l'exposant */
      while ( isspace( *legende ) )
	legende++;

      /* On saute l'éventuel symbole multiplicatif
	 Remarque : code UTF-8 C3 97 = Ã? en latin-1 (où ? = indéfini) => conflit peu probable
	                             = Ã— en windows-1252              => conflit peu probable
	                             = בù en TOS                       => conflit peu probable
       */
      if      ( '*'    == *legende ) legende++; /* k *... et k*... */
      else if ( '\xD7' == *legende ) legende++; /* k ×... et k×..., codage latin-1/windows-1252 */
      else if ( ( '\xC3' == *legende ) && ( '\x97' == *(legende + 1) ) ) legende += 2; /* idem, codage UTF-8 */

      /* On saute les éventuels espaces qui suivraient le symbole multiplicatif */
      while ( isspace( *legende ) )
	legende++;
    } else /* Autre cas : on annule car a priori le k n'est pas un k de pondération */
      legende = debut;
  }

  /* Calcul de la longueur nécessaire en ajoutant le k^n × ... */
  lg_legende = strlen( legende ) + 1; /* 0 final = +1 */
#ifdef _utf8_
  if ( degre != 0 ) {
    lg_legende++;		/* Le k */

    if ( degre < 0 ) lg_legende += 3; /* 3 octets pour signe − en exposant (⁻) : E2 81 BB en UTF-8 */

    if ( degre != 1 ) {
      unsigned int n_chiffres;
      
      n_chiffres = 1 + (unsigned int) log10( (double) -degre );
      lg_legende += 3 * n_chiffres; /* 3 octets [max] par chiffre en exposant en UTF-8 */
    }
  }
#else
  if ( degre < 0 ) {
    lg_legende += 7 + (size_t) log10( (double) -degre ); /* +1 : signe - ; +3 : " * " ; +2 : "k^" */
  } else if ( degre > 0 ) {
    lg_legende += 4; /* +1 : k, +3 : " * " */
    if ( degre > 3 )
      lg_legende += 2 + (size_t) log10( (double) degre); /* ^+ nbr de chiffres de l'exposant, (int) log10(a) = n si 10^n <= a < 10^(n + 1) */
    else if ( degre > 1 )
      lg_legende++; /* Caractère ² ou ^3 */
  }
#endif

  nv_legende = (char *) calloc( lg_legende, sizeof(char) );
  if ( NULL == nv_legende ) {
    ERR_Memoire( "L"e_AIGU"gende adapt"e_AIGU"e", lg_legende * sizeof( char ) );
    return NULL;
  }

#ifdef _DEVELOPPE_
  (void) printf( "%s,%s l.%u : Degré réel à utiliser = %d ; légende conservée = %s\n",
		 __FILE__, __FUNCTION__, __LINE__, degre, legende );

#endif

  if ( degre == 1 )
    sprintf( nv_legende, "k "MULTIPLIER" %s", legende );
  else if ( degre == 0 ) /* Seule chance : il y avait un k^... avant */
    strcpy( nv_legende, legende );
  else if ( degre == 2 )
    sprintf( nv_legende, "k"DEUX_EXPOSANT" "MULTIPLIER" %s", legende );
  else if ( degre == 3 )
    sprintf( nv_legende, "k"TROIS_EXPOSANT" "MULTIPLIER" %s", legende );    
  else
    sprintf( nv_legende, "k^%d "MULTIPLIER" %s", degre, legende );

  return nv_legende;
}

/* -------- Version avec interface graphique -------- */

Graphe *EXF_ponderer(Graphe *source, int proposition_exposant)
{
  int exposant; /* Valeur de n à utiliser */
  int reponse;
  Graphe *destination;
  struct infos_demande v_exposant = { BI_CHAMP_DEMANDE_ENTIER_S, "n =", NULL,
				      BI_EDITABLE_ENTIER, 0.0, 0.0, 0.0,
				      0, NULL, { 0 }, { 0 }, { 0 } };

  /* On demande la pondération à utiliser */
  v_exposant.valeur_initiale.entier_S = proposition_exposant;
  v_exposant.valeur_defaut.entier_S   = proposition_exposant;
  reponse = BI_DemandeValeur( principale, &Q_ExposantPonderation, &v_exposant );
  if ( 3 == reponse ) return source; /* Annulation */
  if ( 2 == reponse ) { /* On veut un nouveau graphe */
    destination = GRF_Copier( source, NULL, FAUX );
    if ( NULL == destination ) return source; /* Pas pu créer le graphe */
    GRF_titre( destination, source->nom ); /* Même nom pour la copie */
  } else {
    destination = source;
  }

  exposant = (int) v_exposant.valeur_donnee.entier_S;
  if ( 0 == exposant ) {
    return destination; /* Multiplication par k^0 = 1 */
  }

  if ( EXF_k_p_n( destination, exposant, VRAI ) == DIV_ZERO ) {
    BI_ERR_alerte( NULL, &ERR_KPN_DivisionParZero, VRAI );
  }
  
  GRF_modifier( destination, destination, FAUX ); /* Le nombre de points n'a pas changé */
  GRF_aller( destination );

  return destination; /* Le graphe après magnification */
}

/* -------------- Réalisation de l'opération elle-même --------------------------------\
 |   Renvoie 0 si tout va bien, un numéro d'erreur sinon                               |
 |                                                                                     |
 | Suppose que le graphe est de type réel                                              |
 | Travaille en place ; s'occupe de modifier les barres d'erreur en Y en même temps    |
 |                                                                                     |
 | spectre    : le spectre sur lequel réaliser l'opération                             |
 | degre      : le degre à utiliser (on aura Y ---> Y * X^degre)                       |
 | historique : VRAI si il faut inscrire dans l'historique                             |
 \------------------------------------------------------------------------------------ */

int EXF_k_p_n(Graphe *spectre, int degre, boolean historique)
{
  int sortie = 0;
  char *legende;

  if ( NULL == spectre ) return E_V_K_PAS_GRAPHE;
  if ( 0    == degre   ) return 0; /* Rien à faire si on multiplie par X^0 = 1... */

  if ( GRF_FCT_REELLE == spectre->genre_donnees )
    sortie = EXF_PondereReel(spectre, degre);
  else if ( GRF_FCT_COMPLEXE == spectre->genre_donnees )
    sortie = EXF_PondereComplexe(spectre, degre);
  else {
    (void) fprintf( stderr, "%s,%s l.%u : Pondération impossible / graphe de type %d\n",
		    __FILE__, __FUNCTION__, __LINE__, spectre->genre_donnees );
    OPTION_A_FAIRE( "Pondération, graphe autre que fonction réelle ou complexe" );
    return -10;
  }

  GRF_bornes( spectre, NE_PAS_FAIRE, DESORDONNE ); /* Bornes en Y ont changé... pas en X */
  GRF_bornes( spectre->erreur_y, NE_PAS_FAIRE, DESORDONNE );

  if ( spectre->affichage != NULL ) {
    const char *unite_X;

    /* On modifie l'unité en Y */
    unite_X = GRF_QuelleUnite( spectre, AXE_X );
    if ( unite_X != NULL ) {
      const char *unite_Y;
      char *nv_unite_Y;

      unite_Y = GRF_QuelleUnite( spectre, AXE_Y );
      nv_unite_Y = TXT_ProduitUnites( unite_Y, unite_X, degre );
      GRF_ChangerUnite( spectre, AXE_Y, nv_unite_Y );

      /* Plus besoin de la nouvelle légende, copiée */
      free( nv_unite_Y ); nv_unite_Y = NULL;
    }

    /* On s'occupe maintenant de modifier la légende en Y */
    (void) fprintf( stderr, "%s,%s l.%u : On va adapter la légende en Y [\"%s\", degré = %d]...\n",
		    __FILE__, __FUNCTION__, __LINE__, GRF_QuelleLegende( spectre, AXE_Y ), degre );
    legende = KPN_adapte_legende( GRF_QuelleLegende( spectre, AXE_Y ), degre );
    GRF_ChangerLegende( spectre, AXE_Y, legende, FAUX );
  }

  /* Log dans l'historique */
  if ( VRAI == historique ) {
    HIST_Titre(1, "Pond"e_AIGU"ration par k"n_EXPOSANT, "Pond");
    HIST_Texte(1, 0, "Graphe : %s\n", spectre->nom);
    HIST_Texte(3, 0, "Degr"e_AIGU"  : n = %d\n", degre);
  } else {
    HIST_Texte(5, 2, "pond"e_AIGU"ration par k"n_EXPOSANT", n = %d\n", degre);
  }

  /* Mise à jour éventuelle de l'affichage */
  GRF_MiseAJour( spectre, VRAI );

  return sortie;
}
