/* LASE - Logiciel d'analyse des spectres EXAFS
   (C) Emmanuel CURIS, 1998 - 2003

   Sauvegarde d'un graphe au format postscript
   L'unit de base est le 1/100 de mm = taille d'un point donc 
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "fichiers.h" /* Pour export_ps() */
#include "ps.h"
#include "graphes.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define MAX_PS_POINTS (1000) /* Nombre maximal de points dans un chemin */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* 1er niveau */
static void PS_GrapheComplexe_Re (FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
				  int couleur, int style_ligne, int style_marque);
static void PS_GrapheComplexe_Im (FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
				  int couleur, int style_ligne, int style_marque);
static void PS_GrapheComplexe_Mod(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
				  int couleur, int style_ligne, int style_marque);

static void PS_GrapheReel_XY(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
			     int couleur, int style_ligne, int style_marque);

static void PS_GrapheSerie(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
			   int couleur, int style_ligne, int style_marque);

/* 2e niveau : dessin brut */
static void PS_cmpIm_lignes (FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle);
static void PS_cmpIm_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque);
static void PS_cmpRe_lignes (FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle);
static void PS_cmpRe_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque);
static void PS_cmpMod_lignes (FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle);
static void PS_cmpMod_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque);

static void PS_err_Im(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas);
static void PS_err_Re(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas);
static void PS_err_X (FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas);
static void PS_err_Y (FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas);

static void PS_reel_lignes (FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle);
static void PS_reel_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque);

static void PS_serie_lignes (FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle);
static void PS_serie_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque);

/* ~~~~~~~~ Ralisation pratique ~~~~~~~~ */

/* -------- Fonction d'appel, visible de l'extrieur : se charge d'appeler la bonne fonction  chaque fois --------
   Renvoie 0 si le trac a pu se faire correctement

   Le graphe est trac en tenant compte de la rgion & des chelles indiqu dans le cadre.

   fichier : le fichier PS dans lequel faire la sauvegarde   
 -------------------------------------------------------------------------------- */

int PS_Graphe(FILE *fichier, struct infos_ps *infos, struct rect_ps *cadre, struct graphe *graphe,
	      int erreurs, int couleur, int style_ligne, int style_marque)
{
  int methode = 0;
  int retour;

  /* Contrle prliminaire */
  if ( ( NULL == fichier ) || ( NULL == infos ) || ( NULL == cadre ) || ( NULL == graphe ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : L'un des paramtres pass est incorrect\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
#endif
    return -1;
  }
  retour = AXE_ValiderBornes( &(cadre->cadre.axe_X) ) | AXE_ValiderBornes( &(cadre->cadre.axe_Y) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Certaines bornes sont absurdes [X : %g  %g ; Y : %g  %g]\n",
		   __FILE__, __FUNCTION__, __LINE__,
		   cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max);
    return -2;
  }

  /* Choix de la couleur & du style de trac */
  if ( couleur < 0 ) {
    if ( graphe->affichage != NULL ) {
      couleur = graphe->affichage->couleur;
    } else {
      couleur = NOIR;
    }
  }
  if ( style_ligne < 0 ) {
    if ( graphe->affichage != NULL ) {
      style_ligne = graphe->affichage->style;
    } else {
      style_ligne = 0;
    }
  }
  if ( style_marque < 0 ) {
    if ( graphe->affichage != NULL ) {
      style_marque = graphe->affichage->marque;
    } else {
      style_marque = 0;
    }
  }

  /* Choix de la mthode de trac */
  if ( graphe->affichage != NULL ) {
    methode = graphe->affichage->methode;
  } else {
    if ( GRF_FCT_COMPLEXE == graphe->genre_donnees )
      methode = GRF_C_IM_MOD; /* Car pas 0 */
  }

  /* En fonction de la nature des donnes, on appelle ce qu'il faut */
  switch ( graphe->genre_donnees )
  {
  case GRF_FCT_REELLE:
    if ( GRF_R_CARTESIEN == methode ) {
      PS_GrapheReel_XY( fichier, cadre, graphe, erreurs, couleur, style_ligne, style_marque);
    } else if ( GRF_R_POLAIRE == methode ) {
      OPTION_A_FAIRE( "graphe en coordonnes polaires" );
    } else {
      (void) fprintf(stderr, "TRC_Graphe : mthode inconnue, graphes rels [%d]\n", methode);
      return +1;
    }
    return 0;

  case GRF_FCT_COMPLEXE:
    if ( GRF_C_IM_MOD == methode ) {
      PS_GrapheComplexe_Im (fichier, cadre, graphe, erreurs, couleur, style_ligne, style_marque);
      erreurs &= ~AXE_X; /* Dj traces : inutile de le refaire */
      PS_GrapheComplexe_Mod(fichier, cadre, graphe, erreurs, couleur + 1, style_ligne, style_marque);
    } else if ( GRF_C_RE_MOD == methode ) {
      PS_GrapheComplexe_Re (fichier, cadre, graphe, erreurs, couleur, style_ligne, style_marque);
      erreurs &= ~AXE_X; /* Dj traces : inutile de le refaire */
      PS_GrapheComplexe_Mod(fichier, cadre, graphe, erreurs, couleur + 1, style_ligne, style_marque);
    } else if ( GRF_C_RE_IM == methode ) {
      PS_GrapheComplexe_Re (fichier, cadre, graphe, erreurs, couleur, style_ligne, style_marque);
      erreurs &= ~AXE_X; /* Dj traces : inutile de le refaire */
      PS_GrapheComplexe_Im (fichier, cadre, graphe, erreurs, couleur + 1, style_ligne, style_marque);
    } else {
      (void) fprintf(stderr, "TRC_Graphe : mthode inconnue, graphes complexes [%d]\n", methode);
      return +1;
    }
    return 0;

  case GRF_SERIE:
    if ( graphe->x != NULL ) {
      PS_GrapheReel_XY( fichier, cadre, graphe, erreurs, couleur, style_ligne, style_marque);
      return 0;
    }
    PS_GrapheSerie( fichier, cadre, graphe, erreurs, couleur, style_ligne, style_marque);
    return 0;

  default:
    (void) fprintf(stderr, "TRC_Graphe : Type de graphe inconnu ou non adapt [%d]\n", graphe->genre_donnees);
    break;
  }

  return -1;
}

/* -------- Graphe complexe cartsien (Z=f(X)) -------- */

/* Partie imaginaire */
static void PS_GrapheComplexe_Im(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
				 int couleur, int style_ligne, int style_marque)
{
  double x_echelle, y_echelle;

  /* On met la bonne couleur */
  PS_Couleur( fichier, couleur );

  /* On contrle que l'on a bien quelque chose  afficher */
  if ( ( graphe->x_min > cadre->cadre.axe_X.max ) || ( graphe->x_max < cadre->cadre.axe_X.min ) ||
       ( graphe->y_min > cadre->cadre.axe_Y.max ) || ( graphe->y_max < cadre->cadre.axe_Y.min ) ) {
    (void) fprintf( fichier, "%% Graphe hors de la zone reprsente\n");
    return;
  }

  /* On calcule les chelles [on a dj vrifi que les bornes du cadre sont plausibles, normalement] */
  x_echelle = cadre->larg / ( cadre->cadre.axe_X.max - cadre->cadre.axe_X.min );
  y_echelle = cadre->haut / ( cadre->cadre.axe_Y.max - cadre->cadre.axe_Y.min );

  /* On trace le graphe proprement dit */
  if ( style_ligne > 0 ) {
    (void) PS_StyleLigne( fichier, style_ligne );
    PS_cmpIm_lignes( fichier, graphe, cadre, x_echelle, y_echelle );
  }
  if ( style_marque > 0 ) {
    /* On annule le pointill */
    (void) fprintf( fichier, "[] 0 setdash\n" );
    PS_cmpIm_marques( fichier, graphe, cadre, x_echelle, y_echelle, style_marque );
  }

  /* Affichage des erreurs [en X] */
  if ( erreurs & AXE_X ) {
    PS_err_X( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
  if ( erreurs & AXE_Y ) {
    PS_err_Im( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
}

/* Module */
static void PS_GrapheComplexe_Mod(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
				  int couleur, int style_ligne, int style_marque)
{
  double x_echelle, y_echelle;

  /* On met la bonne couleur */
  PS_Couleur( fichier, couleur );

  /* On contrle que l'on a bien quelque chose  afficher */
  if ( ( graphe->x_min > cadre->cadre.axe_X.max ) || ( graphe->x_max < cadre->cadre.axe_X.min ) ||
       ( graphe->y_min > cadre->cadre.axe_Y.max ) || ( graphe->y_max < cadre->cadre.axe_Y.min ) ) {
    (void) fprintf( fichier, "%% Graphe hors de la zone reprsente\n");
    return;
  }

  /* On calcule les chelles [on a dj vrifi que les bornes du cadre sont plausibles, normalement] */
  x_echelle = cadre->larg / ( cadre->cadre.axe_X.max - cadre->cadre.axe_X.min );
  y_echelle = cadre->haut / ( cadre->cadre.axe_Y.max - cadre->cadre.axe_Y.min );

  /* On trace le graphe proprement dit */
  if ( style_ligne > 0 ) {
    (void) PS_StyleLigne( fichier, style_ligne );
    PS_cmpMod_lignes( fichier, graphe, cadre, x_echelle, y_echelle );
  }
  if ( style_marque > 0 ) {
    /* On annule le pointill */
    (void) fprintf( fichier, "[] 0 setdash\n" );
    PS_cmpMod_marques( fichier, graphe, cadre, x_echelle, y_echelle, style_marque );
  }

  /* Affichage des erreurs [en X ou en Y] */
  if ( erreurs & AXE_X ) {
    PS_err_X( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
  if ( erreurs & AXE_Y ) {
    OPTION_A_FAIRE( "Tracer les erreurs sur le module" );
  }
}

/* Partie relle */
static void PS_GrapheComplexe_Re(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
				 int couleur, int style_ligne, int style_marque)
{
  double x_echelle, y_echelle;

  /* On met la bonne couleur */
  PS_Couleur( fichier, couleur );

  /* On contrle que l'on a bien quelque chose  afficher */
  if ( ( graphe->x_min > cadre->cadre.axe_X.max ) || ( graphe->x_max < cadre->cadre.axe_X.min ) ||
       ( graphe->y_min > cadre->cadre.axe_Y.max ) || ( graphe->y_max < cadre->cadre.axe_Y.min ) ) {
    (void) fprintf( fichier, "%% Graphe hors de la zone reprsente\n");
    return;
  }

  /* On calcule les chelles [on a dj vrifi que les bornes du cadre sont plausibles, normalement] */
  x_echelle = cadre->larg / ( cadre->cadre.axe_X.max - cadre->cadre.axe_X.min );
  y_echelle = cadre->haut / ( cadre->cadre.axe_Y.max - cadre->cadre.axe_Y.min );

  /* On trace le graphe proprement dit */
  if ( style_ligne > 0 ) {
    (void) PS_StyleLigne( fichier, style_ligne );
    PS_cmpRe_lignes( fichier, graphe, cadre, x_echelle, y_echelle );
  }
  if ( style_marque > 0 ) {
    /* On annule le pointill */
    (void) fprintf( fichier, "[] 0 setdash\n" );
    PS_cmpRe_marques( fichier, graphe, cadre, x_echelle, y_echelle, style_marque );
  }

  /* Affichage des erreurs [en X ou en Y] */
  if ( erreurs & AXE_X ) {
    PS_err_X( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
  if ( erreurs & AXE_Y ) {
    PS_err_Re( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
}

/* -------- Graphe rel, cartsien (Y=f(X)) -------- */

static void PS_GrapheReel_XY(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
			     int couleur, int style_ligne, int style_marque)
{
  double x_echelle, y_echelle;

  /* On met la bonne couleur */
  PS_Couleur( fichier, couleur );

  /* On contrle que l'on a bien quelque chose  afficher */
  if ( ( graphe->x_min > cadre->cadre.axe_X.max ) || ( graphe->x_max < cadre->cadre.axe_X.min ) ||
       ( graphe->y_min > cadre->cadre.axe_Y.max ) || ( graphe->y_max < cadre->cadre.axe_Y.min ) ) {
    (void) fprintf( fichier, "%% Graphe hors de la zone reprsente\n");
    return;
  }

  /* On calcule les chelles [on a dj vrifi que les bornes du cadre sont plausibles, normalement] */
  x_echelle = cadre->larg / ( cadre->cadre.axe_X.max - cadre->cadre.axe_X.min );
  y_echelle = cadre->haut / ( cadre->cadre.axe_Y.max - cadre->cadre.axe_Y.min );

  /* On trace le graphe proprement dit */
  if ( style_ligne > 0 ) {
    (void) PS_StyleLigne( fichier, style_ligne );
    PS_reel_lignes( fichier, graphe, cadre, x_echelle, y_echelle );
  }
  if ( style_marque > 0 ) {
    /* On annule le pointill */
    (void) fprintf( fichier, "[] 0 setdash\n" );
    PS_reel_marques( fichier, graphe, cadre, x_echelle, y_echelle, style_marque );
  }

  /* Affichage des erreurs [en X ou en Y] */
  if ( erreurs & AXE_X ) {
    PS_err_X( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
  if ( erreurs & AXE_Y ) {
    PS_err_Y( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
}

/* -------- Graphe rel, srie de valeurs [X implicite = numro de point] -------- */

static void PS_GrapheSerie(FILE *fichier, struct rect_ps *cadre, Graphe *graphe, int erreurs,
			   int couleur, int style_ligne, int style_marque)
{
  double x_echelle, y_echelle;

  /* On met la bonne couleur */
  PS_Couleur( fichier, couleur );

  /* On contrle que l'on a bien quelque chose  afficher */
  if ( ( graphe->x_min > cadre->cadre.axe_X.max ) || ( graphe->x_max < cadre->cadre.axe_X.min ) ||
       ( graphe->y_min > cadre->cadre.axe_Y.max ) || ( graphe->y_max < cadre->cadre.axe_Y.min ) ) {
    (void) fprintf( fichier, "%% Graphe hors de la zone reprsente\n");
    return;
  }

  /* On calcule les chelles [on a dj vrifi que les bornes du cadre sont plausibles, normalement] */
  x_echelle = cadre->larg / ( cadre->cadre.axe_X.max - cadre->cadre.axe_X.min );
  y_echelle = cadre->haut / ( cadre->cadre.axe_Y.max - cadre->cadre.axe_Y.min );

  /* On trace le graphe proprement dit */
  if ( style_ligne > 0 ) {
    (void) PS_StyleLigne( fichier, style_ligne );
    PS_serie_lignes( fichier, graphe, cadre, x_echelle, y_echelle );
  }
  if ( style_marque > 0 ) {
    /* On annule le pointill */
    (void) fprintf( fichier, "[] 0 setdash\n" );
    PS_serie_marques( fichier, graphe, cadre, x_echelle, y_echelle, style_marque );
  }

  /* Affichage des erreurs [possibles en Y seulement !] */
  if ( erreurs & AXE_Y ) {
    PS_err_Y( fichier, graphe, cadre, x_echelle, y_echelle, 0 );
  }
}

/* ~-~-~-~-~-~-~-~-~ Trac brut ~-~-~-~-~-~-~-~-~ */

/* -------- Partie imaginaire d'une fonction complexe Z=f(X), trac en lignes -------- */

static void PS_cmpIm_lignes(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle)
{
  unsigned long i = 0;
  unsigned int compteur = 0;
  register double *x = graphe->x;
  Complexe *y = (Complexe *) graphe->y;

  while ( i < graphe->nombre_point ) { /* Tant qu'il reste des points */
    /* On passe tous les points qui sont avant la zone intressante */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, y->im ) == FAUX ) {
      i++; x++; y++;

      if (i >= graphe->nombre_point) return; /* On s'arrte */
    }

    /* Ici, le point i convient mais pas le i -1 */
    if ( i > 0 ) { /* Le point (i - 1) existe => on trace un segment de droite sur la partie intressante */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     *x, y->im, *(x - 1), (y - 1)->im, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->im - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 1;
    } else {
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->im - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 0; /* Premier point ne compte pas ! */
    }
    if ( i == ( graphe->nombre_point - 1) ) break;

    /* On passe au point suivant */
    i++;
    x++; y++;

    /* On trace tous les points qui sont  l'intrieur de la zone */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, y->im ) == VRAI ) {
      /* Le prcdent tait dedans -> segment de droite simple */
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->im - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++;
      if ( compteur >= MAX_PS_POINTS ) { /* Pour viter de dborder de la pile : pile dite pleine */
	(void) fprintf( fichier, "stroke\n" );
	(void) fprintf( fichier, "%g %g moveto ",
			cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
			cadre->y_BG + ( y->im - cadre->cadre.axe_Y.min ) * y_echelle );
	compteur = 0;
      }

      /* On passe au point suivant */
      x++; y++;
      i++;
      if ( i >= graphe->nombre_point ) break;
    }

    /* Ici : soit on a fini par manque de point, soit car on sort de la zone */
    if ( i < graphe->nombre_point ) { /* On sort de la zone */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     *(x - 1), (y - 1)->im, *x, y->im, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++; /* Ce point compte ! */
    }

    /* On trace s'il reste des points */
    if ( compteur > 0 ) {
      (void) fprintf( fichier, "stroke\n" );
    }
  }
}

/* -------- Partie imaginaire d'une fonction complexe Z=f(X), trac en marques -------- */

static void PS_cmpIm_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque)
{
  unsigned long i;
  double *x;
  Complexe *y;

  for ( i = 0, x = graphe->x, y = (Complexe *) graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    if ( VRAI == TRC_EstDansRegion2D( &(cadre->cadre), *x, y->im ) ) {
      (void) fprintf( fichier, "%g %g %s ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->im - cadre->cadre.axe_Y.min ) * y_echelle, PS_Marque( marque ) );
    }
  }

  (void) fputc( '\n', fichier );
}

/* -------- Module d'une fonction complexe Z=f(X), trac en lignes -------- */

static void PS_cmpMod_lignes(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle)
{
  unsigned long i = 0;
  unsigned int compteur = 0;
  register double *x = graphe->x, module;
  Complexe *y = (Complexe *) graphe->y;

  module = sqrt( y->re * y->re + y->im * y->im );
  while ( i < graphe->nombre_point ) { /* Tant qu'il reste des points */
    /* On passe tous les points qui sont avant la zone intressante */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, module ) == FAUX ) {
      i++; x++; y++;

      if (i >= graphe->nombre_point) return; /* On s'arrte */
      module = sqrt( y->re * y->re + y->im * y->im );
    }

    /* Ici, le point i convient mais pas le i -1 */
    if ( i > 0 ) { /* Le point (i - 1) existe => on trace un segment de droite sur la partie intressante */
      double x_vrai, y_vrai;
      double mod_avant;

      mod_avant = sqrt( (y - 1)->re * (y - 1)->re + (y - 1)->im * (y - 1)->im );
      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     *x, module, *(x - 1), mod_avant, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( module - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 1;
    } else {
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( module - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 0; /* Premier point ne compte pas ! */
    }
    if ( i == ( graphe->nombre_point - 1) ) break;

    /* On passe au point suivant */
    i++;
    x++; y++;
    module = sqrt( y->re * y->re + y->im * y->im );

    /* On trace tous les points qui sont  l'intrieur de la zone */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, y->re ) == VRAI ) {
      /* Le prcdent tait dedans -> segment de droite simple */
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( module - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++;
      if ( compteur >= MAX_PS_POINTS ) { /* Pour viter de dborder de la pile : pile dite pleine */
	(void) fprintf( fichier, "stroke\n" );
	(void) fprintf( fichier, "%g %g moveto ",
			cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
			cadre->y_BG + ( module - cadre->cadre.axe_Y.min ) * y_echelle );
	compteur = 0;
      }

      /* On passe au point suivant */
      i++;
      if ( i >= graphe->nombre_point ) break;
      x++; y++;
      module = sqrt( y->re * y->re + y->im * y->im );
    }

    /* Ici : soit on a fini par manque de point, soit car on sort de la zone */
    if ( i < graphe->nombre_point ) { /* On sort de la zone */
      double x_vrai, y_vrai;
      double mod_avant;

      mod_avant = sqrt( (y - 1)->re * (y - 1)->re + (y - 1)->im * (y - 1)->im );
      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     *(x - 1), mod_avant, *x, module, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++; /* Ce point compte ! */
    }

    /* On trace s'il reste des points */
    if ( compteur > 0 ) {
      (void) fprintf( fichier, "stroke\n" );
    }
  }
}

/* -------- Module d'une fonction complexe Z=f(X), trac en marques -------- */

static void PS_cmpMod_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque)
{
  unsigned long i;
  double *x, module;
  Complexe *y;

  for ( i = 0, x = graphe->x, y = (Complexe *) graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    module = sqrt( y->re * y->re + y->im * y->im );
    if ( VRAI == TRC_EstDansRegion2D( &(cadre->cadre), *x, module ) ) {
      (void) fprintf( fichier, "%g %g %s ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( module - cadre->cadre.axe_Y.min ) * y_echelle, PS_Marque( marque ) );
    }
  }

  (void) fputc( '\n', fichier );
}

/* -------- Partie relle d'une fonction complexe Z=f(X), trac en lignes -------- */

static void PS_cmpRe_lignes(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle)
{
  unsigned long i = 0;
  unsigned int compteur = 0;
  register double *x = graphe->x;
  Complexe *y = (Complexe *) graphe->y;

  while ( i < graphe->nombre_point ) { /* Tant qu'il reste des points */
    /* On passe tous les points qui sont avant la zone intressante */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, y->re ) == FAUX ) {
      i++; x++; y++;

      if (i >= graphe->nombre_point) return; /* On s'arrte */
    }

    /* Ici, le point i convient mais pas le i -1 */
    if ( i > 0 ) { /* Le point (i - 1) existe => on trace un segment de droite sur la partie intressante */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     *x, y->re, *(x - 1), (y - 1)->re, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->re - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 1;
    } else {
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->re - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 0; /* Premier point ne compte pas ! */
    }
    if ( i == ( graphe->nombre_point - 1) ) break;

    /* On passe au point suivant */
    i++;
    x++; y++;

    /* On trace tous les points qui sont  l'intrieur de la zone */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, y->re ) == VRAI ) {
      /* Le prcdent tait dedans -> segment de droite simple */
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->re - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++;
      if ( compteur >= MAX_PS_POINTS ) { /* Pour viter de dborder de la pile : pile dite pleine */
	(void) fprintf( fichier, "stroke\n" );
	(void) fprintf( fichier, "%g %g moveto ",
			cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
			cadre->y_BG + ( y->re - cadre->cadre.axe_Y.min ) * y_echelle );
	compteur = 0;
      }

      /* On passe au point suivant */
      x++; y++;
      i++;
      if ( i >= graphe->nombre_point ) break;
    }

    /* Ici : soit on a fini par manque de point, soit car on sort de la zone */
    if ( i < graphe->nombre_point ) { /* On sort de la zone */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     *(x - 1), (y - 1)->re, *x, y->re, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++; /* Ce point compte ! */
    }

    /* On trace s'il reste des points */
    if ( compteur > 0 ) {
      (void) fprintf( fichier, "stroke\n" );
    }
  }
}

/* -------- Partie relle d'une fonction complexe Z=f(X), trac en marques -------- */

static void PS_cmpRe_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque)
{
  unsigned long i;
  double *x;
  Complexe *y;

  for ( i = 0, x = graphe->x, y = (Complexe *) graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    if ( VRAI == TRC_EstDansRegion2D( &(cadre->cadre), *x, y->re ) ) {
      (void) fprintf( fichier, "%g %g %s ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y->re - cadre->cadre.axe_Y.min ) * y_echelle, PS_Marque( marque ) );
    }
  }

  (void) fputc( '\n', fichier );
}

/* -------- Erreurs sur la partie imaginaire d'un graphe complexe -------- */

static void PS_err_Im(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas)
{
  unsigned long i, idx_err;
  register double *x = graphe->x;
  register Complexe *y = (Complexe *) graphe->y, *y_err;
  double points[ 4 ];

  if ( NULL == graphe->erreur_y ) return; /* Pas d'erreur associe */
  y_err = (Complexe *) graphe->erreur_y->y;

  /* Bouche en chaque point */
  for ( i = 0, idx_err = 0 ; i < graphe->nombre_point ; i++, idx_err++, y++, y_err++ ) {
    if ( ( pas > 1 ) && ( ( idx_err % pas ) != 0 ) ) continue; /* Incertitudes sur certains points, multiples de pas */

    points[ 0 ] = cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle;
    if ( ( points[ 0 ] < cadre->x_BG ) || ( points[ 0 ] > ( cadre->x_BG + cadre->larg ) ) ) continue;
    points[ 2 ] = points[ 0 ];

    points[ 1 ] = cadre->y_BG + ( ( y->im - y_err->im ) - cadre->cadre.axe_Y.min ) * y_echelle;
    if ( points[ 1 ] > ( cadre->y_BG + cadre->haut ) ) continue;

    points[ 3 ] = cadre->y_BG + ( ( y->im + y_err->im ) - cadre->cadre.axe_Y.min ) * y_echelle;
    if ( points[ 3 ] < cadre->y_BG ) continue;

    if ( points[ 1 ] < cadre->y_BG ) points[ 1 ] = cadre->y_BG;
    if ( points[ 3 ] > ( cadre->y_BG + cadre->haut ) ) points[ 3 ] = cadre->y_BG + cadre->haut;

    (void) fprintf( fichier, "%g %g moveto %g %g lineto ", points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
  }
  /* Termin */
  (void) fprintf( fichier, "stroke\n" );
}

/* -------- Erreurs sur la partie relle d'un graphe complexe -------- */

static void PS_err_Re(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas)
{
  unsigned long i, idx_err;
  register double *x = graphe->x;
  register Complexe *y = (Complexe *) graphe->y, *y_err;
  double points[ 4 ];

  if ( NULL == graphe->erreur_y ) return; /* Pas d'erreur associe */
  y_err = (Complexe *) graphe->erreur_y->y;

  /* Bouche en chaque point */
  for ( i = 0, idx_err = 0 ; i < graphe->nombre_point ; i++, idx_err++, y++, y_err++ ) {
    if ( ( pas > 1 ) && ( ( idx_err % pas ) != 0 ) ) continue; /* Incertitudes sur certains points, multiples de pas */

    points[ 0 ] = cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle;
    if ( ( points[ 0 ] < cadre->x_BG ) || ( points[ 0 ] > ( cadre->x_BG + cadre->larg ) ) ) continue;
    points[ 2 ] = points[ 0 ];

    points[ 1 ] = cadre->y_BG + ( ( y->re - y_err->re ) - cadre->cadre.axe_Y.min ) * y_echelle;
    if ( points[ 1 ] > ( cadre->y_BG + cadre->haut ) ) continue;

    points[ 3 ] = cadre->y_BG + ( ( y->re + y_err->re ) - cadre->cadre.axe_Y.min ) * y_echelle;
    if ( points[ 3 ] < cadre->y_BG ) continue;

    if ( points[ 1 ] < cadre->y_BG ) points[ 1 ] = cadre->y_BG;
    if ( points[ 3 ] > ( cadre->y_BG + cadre->haut ) ) points[ 3 ] = cadre->y_BG + cadre->haut;

    (void) fprintf( fichier, "%g %g moveto %g %g lineto ", points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
  }
  /* Termin */
  (void) fprintf( fichier, "stroke\n" );
}

/* -------- Erreurs sur les X d'un graphe (rel) -------- */

static void PS_err_X(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas)
{
  unsigned long i, idx_err;
  register double *x = graphe->x, *y = graphe->y;
  double points[ 4 ];

  if ( NULL == graphe->erreur_x ) return; /* Pas d'erreur associe */

  /* Bouche en chaque point */
  for ( i = 0, idx_err = 0 ; i < graphe->nombre_point ; i++, idx_err++, y++ ) {
    if ( ( pas > 1 ) && ( ( idx_err % pas ) != 0 ) ) continue; /* Incertitudes sur certains points, multiples de pas */

    points[ 0 ] = cadre->x_BG + ( ( *x - graphe->erreur_x->y[ i ] ) - cadre->cadre.axe_X.min ) * x_echelle;
    if ( points[ 0 ] > ( cadre->x_BG + cadre->larg ) ) continue;

    points[ 2 ] = cadre->x_BG + ( ( *x + graphe->erreur_x->y[ i ] ) - cadre->cadre.axe_X.min ) * x_echelle;
    if ( points[ 2 ] < cadre->x_BG ) continue;

    points[ 1 ] = cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle;
    if ( ( points[ 1 ] < cadre->y_BG ) || ( points[ 1 ] > ( cadre->y_BG + cadre->haut ) ) ) continue;
    points[ 3 ] = points[ 0 ];

    if ( points[ 0 ] < cadre->x_BG ) points[ 0 ] = cadre->x_BG;
    if ( points[ 2 ] > ( cadre->x_BG + cadre->larg ) ) points[ 2 ] = cadre->x_BG + cadre->larg;

    (void) fprintf( fichier, "%g %g moveto %g %g lineto ", points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
  }
  /* Termin */
  (void) fprintf( fichier, "stroke\n" );
}

/* -------- Erreurs sur les Y d'un graphe (rel ou srie) -------- */

static void PS_err_Y(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, unsigned int pas)
{
  unsigned long i, idx_err;
  double x_vrai;
  register double *y = graphe->y;
  double points[ 4 ];

  if ( NULL == graphe->erreur_y ) return; /* Pas d'erreur associe */

  /* Bouche en chaque point */
  for ( i = 0, idx_err = 0 ; i < graphe->nombre_point ; i++, idx_err++, y++ ) {
    if ( ( pas > 1 ) && ( ( idx_err % pas ) != 0 ) ) continue; /* Incertitudes sur certains points, multiples de pas */

    if ( NULL == graphe->x )
      x_vrai = (double) i;
    else {
      x_vrai = graphe->x[ i ];
    }
    points[ 0 ] = cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle;
    if ( ( points[ 0 ] < cadre->x_BG ) || ( points[ 0 ] > ( cadre->x_BG + cadre->larg ) ) ) continue;
    points[ 2 ] = points[ 0 ];

    points[ 1 ] = cadre->y_BG + ( ( *y - graphe->erreur_y->y[ i ] ) - cadre->cadre.axe_Y.min ) * y_echelle;
    if ( points[ 1 ] > ( cadre->y_BG + cadre->haut ) ) continue;

    points[ 3 ] = cadre->y_BG + ( ( *y + graphe->erreur_y->y[ i ] ) - cadre->cadre.axe_Y.min ) * y_echelle;
    if ( points[ 3 ] < cadre->y_BG ) continue;

    if ( points[ 1 ] < cadre->y_BG ) points[ 1 ] = cadre->y_BG;
    if ( points[ 3 ] > ( cadre->y_BG + cadre->haut ) ) points[ 3 ] = cadre->y_BG + cadre->haut;

    (void) fprintf( fichier, "%g %g moveto %g %g lineto ", points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
  }
  /* Termin */
  (void) fprintf( fichier, "stroke\n" );
}

/* -------- Fonction Y=f(X) relle cartsienne, trac en lignes -------- */

static void PS_reel_lignes(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle)
{
  unsigned long i = 0;
  unsigned int compteur = 0;
  register double *x = graphe->x, *y = graphe->y;

  while ( i < graphe->nombre_point ) { /* Tant qu'il reste des points */
    /* On passe tous les points qui sont avant la zone intressante */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, *y ) == FAUX ) {
      i++; x++; y++;

      if (i >= graphe->nombre_point) return; /* On s'arrte */
    }

    /* Ici, le point i convient mais pas le i -1 */
    if ( i > 0 ) { /* Le point (i - 1) existe => on trace un segment de droite sur la partie intressante */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_Y.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.max,
			     *x, *y, *(x - 1), *(y - 1), &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 1;
    } else {
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 0; /* Premier point ne compte pas ! */
    }
    if ( i == ( graphe->nombre_point - 1) ) break;

    /* On passe au point suivant */
    i++;
    x++; y++;

    /* On trace tous les points qui sont  l'intrieur de la zone */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), *x, *y ) == VRAI ) {
      /* Le prcdent tait dedans -> segment de droite simple */
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++;
      if ( compteur >= MAX_PS_POINTS ) { /* Pour viter de dborder de la pile : pile dite pleine */
	(void) fprintf( fichier, "stroke\n" );
	(void) fprintf( fichier, "%g %g moveto ",
			cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
			cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
	compteur = 0;
      }

      /* On passe au point suivant */
      x++; y++;
      i++;
      if ( i >= graphe->nombre_point ) break;
    }

    /* Ici : soit on a fini par manque de point, soit car on sort de la zone */
    if ( i < graphe->nombre_point ) { /* On sort de la zone */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_Y.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.max,
			     *(x - 1), *(y - 1), *x, *y, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++; /* Ce point compte ! */
    }

    /* On trace s'il reste des points */
    if ( compteur > 0 ) {
      (void) fprintf( fichier, "stroke\n" );
    }
  }
}

/* -------- Fonction Y=f(X) relle cartsienne, trac avec une marque par point -------- */

static void PS_reel_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque)
{
  unsigned long i;
  double *x, *y;

  for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    if ( VRAI == TRC_EstDansRegion2D( &(cadre->cadre), *x, *y ) ) {
      (void) fprintf( fichier, "%g %g %s ",
		      cadre->x_BG + ( *x - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle, PS_Marque( marque ) );
    }
  }

  (void) fputc( '\n', fichier );
}

/* -------- Srie relle cartsienne, trac en lignes -------- */

static void PS_serie_lignes(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle)
{
  unsigned long i = 0;
  unsigned int compteur = 0;
  register double *y = graphe->y;

  while ( i < graphe->nombre_point ) { /* Tant qu'il reste des points */
    /* On passe tous les points qui sont avant la zone intressante */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), (double) i, *y ) == FAUX ) {
      i++; y++;

      if ( i >= graphe->nombre_point ) return; /* On s'arrte */
    }

    /* Ici, le point i convient mais pas le i -1 */
    if ( i > 0 ) { /* Le point (i - 1) existe => on trace un segment de droite sur la partie intressante */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     (double) i, *y, (double) (i - 1), *(y - 1), &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( ((double) i) - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 1;
    } else {
      (void) fprintf( fichier, "%g %g moveto ",
		      cadre->x_BG + ( ((double) i) - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur = 0; /* Premier point ne compte pas ! */
    }
    if ( i == ( graphe->nombre_point - 1) ) break;

    /* On passe au point suivant */
    i++; y++;

    /* On trace tous les points qui sont  l'intrieur de la zone */
    while ( TRC_EstDansRegion2D( &(cadre->cadre), (double) i, *y ) == VRAI ) {
      /* Le prcdent tait dedans -> segment de droite simple */
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( ((double) i) - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++;
      if ( compteur >= MAX_PS_POINTS ) { /* Pour viter de dborder de la pile : pile dite pleine */
	(void) fprintf( fichier, "stroke\n" );
	(void) fprintf( fichier, "%g %g moveto ",
			cadre->x_BG + ( ((double) i) - cadre->cadre.axe_X.min ) * x_echelle,
			cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle );
	compteur = 0;
      }

      /* On passe au point suivant */
      i++; y++;
      if ( i >= graphe->nombre_point ) break;
    }

    /* Ici : soit on a fini par manque de point, soit car on sort de la zone */
    if ( i < graphe->nombre_point ) { /* On sort de la zone */
      double x_vrai, y_vrai;

      TRC_PointIntersection( cadre->cadre.axe_X.min, cadre->cadre.axe_X.max, cadre->cadre.axe_Y.min, cadre->cadre.axe_Y.max,
			     (double) (i - 1), *(y - 1), (double) i, *y, &x_vrai, &y_vrai );
      (void) fprintf( fichier, "%g %g lineto ",
		      cadre->x_BG + ( x_vrai - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( y_vrai - cadre->cadre.axe_Y.min ) * y_echelle );
      compteur++; /* Ce point compte ! */
    }

    /* On trace s'il reste des points */
    if ( compteur > 0 ) {
      (void) fprintf( fichier, "stroke\n" );
    }
  }
}

/* -------- Srie relle cartsienne, trac avec une marque par point -------- */

static void PS_serie_marques(FILE *fichier, Graphe *graphe, struct rect_ps *cadre, double x_echelle, double y_echelle, int marque)
{
  unsigned long i;
  double *y;

  for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
    if ( VRAI == TRC_EstDansRegion2D( &(cadre->cadre), (double) i, *y ) ) {
      (void) fprintf( fichier, "%g %g %s ",
		      cadre->x_BG + ( ((double) i) - cadre->cadre.axe_X.min ) * x_echelle,
		      cadre->y_BG + ( *y - cadre->cadre.axe_Y.min ) * y_echelle, PS_Marque( marque ) );
    }
  }

  (void) fputc( '\n', fichier );
}
