/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2003 Emmanuel CURIS
   Création : le 7 mars 2003

   Récupération des données expérimentales pour obtenir µ total

   Routines génériques pour les formats les plus courants : données ASCII organisées en colonnes
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "serie.h"

#include "import_xas.h"
#include "esrf.h" /* Pour compter le nombre de lignes */

#include "ext_generique.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

static int           EXTRAIT_TrouverDebut   (FILE *fichier, int type_ligne, int commentaire );

static unsigned long EXTRAIT_CompterLignes  (FILE *fichier, int commentaire, boolean multi_spectres, int ligne, unsigned int col_x,
					     unsigned int nbr_successifs);

int EXTRAIT_Generique(const char *nom, FILE *fichier, Infos_Extraction *infos, EXT_ListeGraphes *graphes, int ligne,
		      unsigned int nbr_spectres, struct serie *serie, unsigned int num_spectre)
{
  int retour;
  unsigned long nbr_points;

  ENTREE_FONCTION("EXTRAIT_Generique", "nom=%s fichier=%p infos=%p graphes=%p ligne=%d nbr_spectres=%u serie=%p num_spectre=%u",
		  nom, fichier, infos, graphes, ligne, nbr_spectres, serie, num_spectre);

  /* On trouve le début des données [et, éventuellement, on détecte quelques informations] */
  retour = EXTRAIT_TrouverDebut( fichier, ligne, infos->commentaire );
  if ( retour != 0 ) {
    if ( EOF == retour ) { /* Fin de fichier intempestive */
      (void) BI_ERR_alerte( NULL, &ERR_FichierInterrompu, VRAI, nom );
    }
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans EXTRAIT_TrouverDebut [code %d ; fichier %s ; spectre %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour, nom, num_spectre);
    SORTIE_FONCTION("EXTRAIT_Generique", "ERREUR, code %d", -1);
    return -128; /* < -100 : il n'y aura pas d'autre message */
  }

  /* Faut-il détecter le nombre de colonnes ? */
  if ( 0 == infos->nbr_colonnes ) {
    infos->nbr_colonnes = compte_colonnes( fichier, 0 );
    if ( 0 == infos->nbr_colonnes ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans compte_colonnes [fichier %s]\n",
		     __FILE__, __FUNCTION__, __LINE__, nom );
      SORTIE_FONCTION("EXTRAIT_Generique", "ERREUR, code %d", -4);
      return -4;
    }
    AVERTISSEMENT ( "%u colonnes détectées dans %s\n", infos->nbr_colonnes, nom );
    HIST_Texte(4, 0, "            %u d"e_AIGU"t"e_AIGU"ecteur(s) en plus de I0\n", infos->nbr_colonnes - 2);
  }

  /* On regarde combien de lignes il faut charger... */
  nbr_points = EXTRAIT_CompterLignes( fichier, infos->commentaire, ( infos->proprietes & DETECT_MULTI_SPECTRE ), ligne, infos->col_x, nbr_spectres - num_spectre);
  if ( 0 == nbr_points ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans EXTRAIT_CompterLignes [fichier %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom );
    SORTIE_FONCTION("EXTRAIT_Generique", "ERREUR, code %d (aucun point)", -5);
    return -5;
  } else if ( nbr_points < 3 ) {
    AVERTISSEMENT( "Spectre ignoré [Nom=%s, numero=%d/%d] (trop peu de points : %lu)", nom, nbr_spectres - num_spectre, nbr_spectres, nbr_points );
    (void) BI_ERR_alerte( NULL, &ERR_TropPeuPoints, VRAI );
    /* On ignore simplement ce spectre */
    sauter_en_tete( fichier, nbr_points, infos->commentaire );
    SORTIE_FONCTION("EXTRAIT_Generique", "spectre ignoré [%d] (trop peu de points)", 0);
    return 0;
  }

  /* Si on veut bien conserver ce spectre */
  if ( graphes != NULL ) {
    unsigned long nbr_graphes;

    if ( ( DETECT_MULTI_MOYENNER == infos->MD_traitement ) ||
	 ( DETECT_MULTI_MOY_LOG  == infos->MD_traitement ) ) {
      nbr_graphes = 3; /* 1 par méthode : fluo, transmission, électrons */
    } else{
      nbr_graphes = infos->nbr_colonnes;
    }

    /* On prépare les graphes en conséquence */
    retour = EXTRAIT_PreparerGraphes( infos, nbr_graphes, nbr_points, nom, graphes, serie, num_spectre, num_spectre + 1 );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "EXTRAIT_Generique : ERREUR dans EXTRAIT_PreparerGraphes [%d, \"%s\"]\n", retour, nom);
      SORTIE_FONCTION("EXTRAIT_Generique", "ERREUR, code %d", -11);
      return -11;
    }

    /* On charge, en fonction du mode demandé */
    retour = EXTRAIT_ChargerSpectres( fichier, nbr_points, infos, graphes, nom, num_spectre + 1 );
    if ( retour > 0 ) {
      int c __attribute__((__unused__));

      /* Ignorer les lignes du spectre qui pourraient rester...
       Elles peuvent aussi inclure des commentaires, surtout si on a interrompu pour une coupure de faisceau ! */
      c = sauter_en_tete( fichier, nbr_points - retour, infos->commentaire );
    }
  } else {
    /* On saute tout ! */
    sauter_en_tete( fichier, nbr_points, infos->commentaire );

    retour = 0;
  }

  /* Tout s'est bien passé si l'on est là */
  SORTIE_FONCTION("EXTRAIT_Generique", "RAS --- %d", retour);
  return retour;
}

/* -------- Trouver le début de la région intéressante dans le fichier -------- */

static int EXTRAIT_TrouverDebut( FILE *fichier, int type_ligne, int commentaire )
{
  int trouve = FAUX, c;

  ENTREE_FONCTION("EXTRAIT_TrouverDebut", "fichier=%p type_ligne=%d commentaire=%d",
		  fichier, type_ligne, commentaire);

  switch ( type_ligne )
  {
  case LIGNE_LURE_AQUVISU:
  case LIGNE_ESRF_SNBL:
  case LIGNE_SOLEIL_SAMBA:
    /* On saute les commentaires initiaux [les lignes après commencent par des espaces] */
    c = fgetc( fichier );
    ungetc( c, fichier ); /* S'il n'y a pas de commentaire ! */
    while ( c == commentaire ) {
      c = ignore_ligne( fichier );
      if ( EOF == c ) {
	(void) fprintf( stderr, "%s,%s l.%u : fin de fichier atteinte / LURE Acquvisu [type %d]\n",
			__FILE__, __FUNCTION__, __LINE__, type_ligne );
	SORTIE_FONCTION("EXTRAIT_TrouverDebut", "ERREUR, code %d (fin de fichier atteinte)", EOF);
	return EOF;
      }
    }
    break;

  case LIGNE_LURE_D21:
    /* Les données sont au début du fichier, puisque l'en-tête a dû être passé avant... */
    break;

  case LIGNE_HASYLAB_DORIS:
    /* On cherche le code %d en début de ligne */
    c = fgetc( fichier );
    ungetc( c, fichier ); /* Pour être compatible avec ignore_ligne dans la suite */
    do {
      if ( EOF == c ) {
	(void) fprintf(stderr, "EXTRAIT_TrouverDebut : fin de fichier atteinte / Hasylab, DORIS\n");
	SORTIE_FONCTION("EXTRAIT_TrouverDebut", "ERREUR, code %d (fin de fichier atteinte)", EOF);
	return EOF;
      }
      if ( '%' == c ) {
	(void) fgetc( fichier ); /* On fait sauter le % initial */
	c = fgetc( fichier );
	if ( 'd' == c ) {
	  trouve = VRAI;
	} else if ( EOF == c ) {
	  (void) fprintf(stderr, "EXTRAIT_TrouverDebut : fin de fichier atteinte / Hasylab, DORIS [fichier fini par %%]\n");
	  SORTIE_FONCTION("EXTRAIT_TrouverDebut", "ERREUR, code %d (fin de fichier atteinte)", EOF);
	  return EOF;
	}
      } else if ( ( c != commentaire ) && ( c != ' ') ) {
	(void) fprintf(stderr, "EXTRAIT_TrouverDebut : code de début de ligne inconnu [%c], chargement hasardeux\n", c);
      }
      c = ignore_ligne( fichier );
    } while ( FAUX == trouve );

    /* On saute ensuite la partie qui décompte les colonnes */
    trouve = FAUX;
    do {
      if ( EOF == c ) {
	(void) fprintf(stderr, "EXTRAIT_TrouverDebut : fin de fichier atteinte / Hasylab, DORIS [fichier fini par %%d]\n");
	SORTIE_FONCTION("EXTRAIT_TrouverDebut", "ERREUR, code %d (fin de fichier atteinte)", EOF);
	return EOF;
      }

      (void) fgetc( fichier ); /* 1er caractère : devrait toujours être un espace ! */
      c = fgetc( fichier );
      if ( ' ' == c ) {
	trouve = VRAI;
      } else if ( EOF == c ) {
	(void) fprintf(stderr, "EXTRAIT_TrouverDebut : fin de fichier atteinte / Hasylab, DORIS [aucune donnée après %%d]\n");
	SORTIE_FONCTION("EXTRAIT_TrouverDebut", "ERREUR, code %d (fin de fichier atteinte)", EOF);
	return EOF;
      } else if ( 'C' != c ) {
	(void) fprintf(stderr, "EXTRAIT_TrouverDebut : code de début de données inconnu [%c], chargement hasardeux\n", c);
      }
      c = ignore_ligne( fichier );
    } while ( FAUX == trouve );
    break;

  case LIGNE_ESRF_ID21:
    if ( EOF == ID21_TrouverDebut( fichier, "z3escan" ) ) {
      SORTIE_FONCTION( __FUNCTION__, "ERREUR, code %d (fin de fichier atteinte)", EOF);
      return EOF;
    }
    break;

  case LIGNE_ESRF_ID22:
    if ( EOF == ID22_TrouverDebut( fichier, "energy" ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Ligne ID22, fin de fichier atteinte\n",
		     __FILE__, __FUNCTION__, __LINE__);
      SORTIE_FONCTION( __FUNCTION__, "ERREUR, code %d (fin de fichier atteinte)", EOF);
      return EOF;
    }
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : type de ligne inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, type_ligne);
  }

  /* Tout s'est bien passé */
  SORTIE_FONCTION("EXTRAIT_TrouverDebut", "RAS --- %d", 0);
  return 0;
}

/* -------- Compter combien de lignes il faut prendre en compte
            donne le nombre de points pour la création des graphes

   Tient compte des éventuels spectres successifs à fondre ou moyenner !
   -------- */

static unsigned long EXTRAIT_CompterLignes( FILE *fichier, int commentaire, boolean multi_spectres, int ligne, unsigned int col_x,
					    unsigned int nbr_successifs)
{
  unsigned long nbr_lignes = 0;

  if ( FAUX == multi_spectres ) { /* Fichier unique : toutes les lignes servent */
    return compte_lignes( fichier, commentaire );
  }

  /* Remarque : pour la fusion, on pourrait imaginer de faire un seul spectre, mais rendrait délicate la
     détection des interruptions notées par des commentaires [type ESRF]... */

  switch ( ligne )
  {
  case LIGNE_LURE_AQUVISU:
    AVERTISSEMENT( "Appel pour Acquvisu > 1 spectre [%u spectres] !\n", nbr_successifs );
    return compte_lignes( fichier, commentaire );

  case LIGNE_LURE_D21:
    /* On suppose des blocs homogènes. Pas garanti, mais le plus probable ! */
    nbr_lignes = compte_lignes( fichier, commentaire );
    if ( ( nbr_lignes % nbr_successifs ) != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : %lu (lignes) n'est pas un multiple de %u (spectres)\n",
		      __FILE__, __FUNCTION__, __LINE__, nbr_lignes, nbr_successifs );
      return 0;
    }
    nbr_lignes = nbr_lignes % nbr_successifs;
    break;

  case LIGNE_HASYLAB_DORIS:
    OPTION_A_FAIRE( "Spectres successifs Hasylab" );
    break;

  case LIGNE_ESRF_SNBL:
    nbr_lignes = SNBL_CompterLignes( fichier, col_x );
    break;

  case LIGNE_ESRF_ID21:
  case LIGNE_ESRF_ID22:
    nbr_lignes = ESRF_CompterLignesScan( fichier );
    break;

#ifdef _DEVELOPPE_
  default:
    (void) fprintf(stderr, "%s,%s l.%u : Type de ligne inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, ligne);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return 0;
#endif
  }

  return nbr_lignes;
}

/* -~-~-~-~-~-~-~-~ Chargement proprement dit des spectres -~-~-~-~-~-~-~-~ */

/* -------- Chargement proprement dit, pour un fichier multi-spectres en colonnes --------
     Renvoie : 0 si le chargement s'est déroulé sans histoire
              <0 en cas d'erreur grave [à charge de l'appelant d'effacer éventuellement les graphes]
	      >0 en cas de chargement partiel [à charge de l'appelant de tronquer les graphes]

     À charge de l'appelant de s'occuper ensuite des bornes ou de faire un traitement ultérieur...
 */

long EXTRAIT_ChargerSpectres(FILE *fichier, unsigned long nbr_lignes, Infos_Extraction *infos,
			     EXT_ListeGraphes *graphes, const char *nom_fichier, unsigned int passage)
{
  unsigned long i;
  double *valeurs;
  int *erreurs = NULL, err_i0 = FAUX;
  int (*analyse)(unsigned int, double *, Intervalle_I *, unsigned long, int *, int, double, unsigned long,
		 Graphe **, Graphe **, const char *, Infos_Extraction *);

  if ( ( DETECT_MULTI_MOYENNER == infos->MD_traitement ) ||
       ( DETECT_MULTI_MOY_LOG  == infos->MD_traitement ) ) {
    analyse = EXTRAIT_AnalyseLigneSommer;
  } else {
    analyse = EXTRAIT_AnalyseLigneSeparer;
  }

  /* Vérifications préliminaires */
  if ( ( NULL == fichier ) || ( NULL == infos ) || ( NULL == graphes ) ) {
    (void) fprintf( stderr, "%s,%s l.%u : paramètre incorrect (fichier=%p, %lu lignes, infos=%p)\n",
		    __FILE__, __FUNCTION__, __LINE__, (void *) fichier, nbr_lignes, (void *) infos );
    return -1;
  }

  if ( NULL == graphes->x ) { /* On ne conserve rien */
#ifdef _DEVELOPPE_
    (void) fprintf( stderr, "%s,%s l.%u : on se contente de sauter le bloc de %lu lignes\n",
		    __FILE__, __FUNCTION__, __LINE__, nbr_lignes );
#endif
    for ( i = 0 ; i < nbr_lignes ; i++ ) {
      (void) ignore_ligne( fichier );
    }

    return 0;
  }

  /* On réserve pour les valeurs lues, ligne à ligne */
  valeurs = (double *) malloc( sizeof(double) * infos->nbr_colonnes );
  if ( NULL == valeurs ) {
    ERR_Memoire( "Table des valeurs", sizeof(double) * infos->nbr_colonnes );
    return -2;
  }

  /* On réserve pour les erreurs éventuelles */
  if ( graphes->spectres != NULL ) { /* Sinon, le contrôle est sans intérêt...*/
    erreurs = (int *) malloc( sizeof( int ) * infos->nbr_colonnes );
    if ( erreurs != NULL ) {
      unsigned int j;

      for ( j = 0 ; j < infos->nbr_colonnes ; j++ ) {
	erreurs[ j ] = FAUX; /* Il n'y a eu pour l'instant aucune erreur sur ce spectre */
      }
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : imposible de réserver les erreurs --- Aucun message !\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }
  }

  /* Pour chaque ligne (= énergie expérimentale) */
  for ( i = 0 ; i < nbr_lignes ; i++ ) {
    int c, retour;
    double i0;

    c = fgetc( fichier );
    ungetc( c, fichier ); /* Au cas où la ligne commence directement par un chiffre... */
    if ( c == infos->commentaire ) {
      int retour;

      retour = BI_ERR_alerte( NULL, &MSG_CommentaireDansSpectre, VRAI, nom_fichier, passage );
      if ( ARRETER_CHARGEMENT == retour ) {
	free( valeurs ); valeurs = NULL;
	free( erreurs ); erreurs = NULL;
	return i;
      }
      if ( ANNULER_CHARGEMENT == retour ) {
	free( valeurs ); valeurs = NULL;
	free( erreurs ); erreurs = NULL;
	return -50;
      }

      /* Commentaire au milieu du spectre : coupure de faisceau probable... */
      while( c == infos->commentaire ) {
	c = ignore_ligne( fichier );
	if ( c == EOF ) {
	  (void) fprintf(stderr, "%s,%s l.%u : fin de fichier atteinte [%s]\n",
			 __FILE__, __FUNCTION__, __LINE__, nom_fichier);
	  free( valeurs ); valeurs = NULL;
	  free( erreurs ); erreurs = NULL;
	  return i; /* On s'arrête, en tronquant le graphe */
	}
      }
    }

    /* On commence par lire entièrement la ligne */
    lit_valeurs( fichier, infos->nbr_colonnes, 0, valeurs, NULL );

    /* Ensuite, on construit ce qu'il faut
       1. les X */
    if ( ( infos->col_x > 0 ) && ( infos->col_x <= infos->nbr_colonnes ) ) { /* X dans le fichier */
      graphes->x [ i ] = valeurs[ infos->col_x - 1 ];
    } else { /* On utilise, arbitrairement, le numéro de ligne comme X */
      graphes->x [ i ] = (double) i;
    }

    /* 2. les I0, si nécessaire */
    i0 = valeurs[ infos->col_i0 ];
    if ( graphes->i_0 != NULL ) {
      graphes->i_0->y[ i ] = i0;
    }

    /* Faut-il construire des spectres ? Si oui, on contrôle que I0 est correct !*/
    if ( graphes->spectres != NULL ) {
      if ( i0 <= 0.0 ) {
	if ( FAUX == err_i0 ) {
	  int reponse;
	  unsigned int j;

	  reponse = BI_ERR_alerte( NULL, &ERR_CoupureFaisceau, VRAI, nom_fichier );
	  if ( ANNULER_CHARGEMENT == reponse ) {
	    free( valeurs ); valeurs = NULL;
	    free( erreurs ); erreurs = NULL;
	    return -50;
	  }
	  if ( ARRETER_CHARGEMENT == reponse ) {
	    free( valeurs ); valeurs = NULL;
	    free( erreurs ); erreurs = NULL;
	    return i;
	  }
	  
	  for ( j = 0 ; j < infos->nbr_colonnes ; j++ )
	    erreurs[ j ] = VRAI;
	}
	err_i0 = VRAI;
      }
    }

    /* 3. les spectres en transmission */
    retour = analyse( infos->nbr_colonnes, valeurs, infos->colonnes_trans, infos->nbr_I_trans, erreurs,
		      DETECT_TRANSMISSION, i0, i, graphes->spectres, graphes->i_1, nom_fichier, infos );
    if ( retour != 0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans analyse / transmission [%d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
#endif
      free( valeurs ); valeurs = NULL;
      free( erreurs ); erreurs = NULL;

      if ( ANNULER_CHARGEMENT == retour ) return -50; /* On annule le chargement de ce spectre */
      if ( ARRETER_CHARGEMENT == retour ) return i; /* On tronque le chargement au nombre de points considéré */
      return retour;
    }

    /* 4. les spectres en fluorescence */
    retour = analyse( infos->nbr_colonnes, valeurs, infos->colonnes_fluo, infos->nbr_I_fluo, erreurs,
		      DETECT_FLUORESCENCE, i0, i, graphes->spectres, graphes->i_1, nom_fichier, infos );
    if ( retour != 0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans analyse / fluorescence [%d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
#endif
      free( valeurs ); valeurs = NULL;
      free( erreurs ); erreurs = NULL;

      if ( ANNULER_CHARGEMENT == retour ) return -50; /* On annule le chargement de ce spectre */
      if ( ARRETER_CHARGEMENT == retour ) return i; /* On tronque le chargement au nombre de points considéré */
      return retour;
    }

    /* 5. les spectres en détection d'électrons */
    retour = analyse( infos->nbr_colonnes, valeurs, infos->colonnes_det, infos->nbr_I_det, erreurs,
		      DETECT_ELECTRONS, i0, i, graphes->spectres, graphes->i_1, nom_fichier, infos );
    if ( retour != 0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans analyse / détection d'électrons [%d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
#endif
      free( valeurs ); valeurs = NULL;
      free( erreurs ); erreurs = NULL;

      if ( ANNULER_CHARGEMENT == retour ) return -50; /* On annule le chargement de ce spectre */
      if ( ARRETER_CHARGEMENT == retour ) return i; /* On tronque le chargement au nombre de points considéré */
      return retour;
    }
  }

  /* Plus besoin de la table des colonnes, ni de celle des erreurs */
  free( valeurs ); valeurs = NULL;
  free( erreurs ); erreurs = NULL;

  /* Tout s'est bien passé */
  return 0;
}
