/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Cration de ce fichier : 15 septembre 1998

   Module de filtrage des spectres (essentiellement par transforme de Fourier)

   Corrections de phase et d'amplitude (avant / aprs calcul de la TF / TFI)
   Source original rcupr de ../FOURIER.c et ../AJUSTE.c
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"
#include "extrait.h" /* Pour EXF_k_p_n */
#include "filtrage.h"
#include "lissage.h"
#include "science.h" /* Si jamais M_PI n'avait pas t dfini dans <math.h> */

#include "tf_corrige.txt"

/* ******** Variables locales ************************ */

#define LG_MAX_NOM_LPM 34

/* ******** Fonctions locales ************************ */

static void prepare_amplitude(void);
static int GER_amplitude(unsigned int objet, int x, int y);

/* ******** Ralisation pratique des fonctions ******* */

/* ~~~~~~~~ Correction en phase et amplitude d'une TF ~~~~~~~~ */

/* -------- Correction en amplitude du spectre dont on veut la TF ---------------\
 |   Renvoie <0 si il y a eu un problme lors de la correction                   |
 |                                                                               |
 | graphe  : le graphe dont on veut raliser la correction                       |
 | correct : le graphe contenant les donnes d'amplitude servant  corriger      |
 \------------------------------------------------------------------------------ */

boolean TF_coramp(Graphe *graphe, Graphe *correct)
{
  double *b = NULL, *c = NULL, *d = NULL; /* Tableaux pour la spline */
  double amp;
  register double *x;
  int retour = 0;
  unsigned long i;

  if ( (correct == NULL) || (graphe == NULL) ) return ERR_GRF_NULL;

  if ( (correct->x_min >= graphe->x_max ) || (correct->x_max <= graphe->x_min) ) {
    (void) EXF_ERREUR(ERR_DISJOINTS);
    return ERR_GRF_DISJOINTS;
  }

  /* On calcule la spline qui permettra ensuite la correction */
  retour = SPL_tt_prep( correct, &b, &c, &d);
  if (retour < 0) {
    (void) LIS_ERREUR( retour, correct->nom );
    return FAUX; /* Problme pour le calcul de la SPLINE */
  }

  x = graphe->x;
  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    register Complexe *y = (Complexe *) graphe->y;

    for (i = 0 ; i < graphe->nombre_point ; i++, x++, y++) {
      amp = SPL_tt_calc( correct, *x, b, c, d);

      if (amp != 0.0) { /* Pas de risque de division par zro */
	y->re /= amp;
	y->im /= amp;
      } else {
	retour = ERR_GRF_DIV_ZERO;
      }
    }
  } else if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
    register double *y = graphe->y;

    for (i = 0 ; i < graphe->nombre_point ; i++, x++, y++) {
      amp = SPL_tt_calc( correct, *x, b, c, d);
      if (amp != 0.0) { /* Pas de risque de division par zro */
	*y /= amp;
      } else {
	retour = ERR_GRF_DIV_ZERO;
      }
    }
  } else {
  }

  if ( ERR_GRF_DIV_ZERO == retour ) {
    EXF_ERREUR(DIV_ZERO, "Amplitude nulle", txt_err_val_inchangee);
  }

  if ( graphe->erreur_y != NULL ) {
    OPTION_A_FAIRE( "Propager les incertitudes" );
  }

  free(b); free(c); free(d); /* On libre la mmoire */
  return VRAI;
}

/* -------- Correction de phase avant de faire la TF ---------------------------------------------------\
 |  Renvoie FAUX en cas de problme                                                                     |
 |  Le graphe  corriger (dont on va faire la TF) DOIT tre de type complexe,                           |
 |       mme si la fonction est relle                                                                 |
 |                                                                                                      |
 | graphe  : le graphe  corriger                                                                       |
 | correct : graphe contenant les phases servant  la correction                                        |
 | reel    : VRAI = la partie imaginaire, non initialise, est suppose nulle (graphe rel  l'origine) |
 \----------------------------------------------------------------------------------------------------- */

boolean TF_corpha(Graphe *graphe, Graphe *correct)
{
  register Complexe *y;
  register double *x;
  double *b,*c,*d;
  double pha, reelle, imag;
  double cosinus, sinus;
  unsigned long i;
  int retour;

  if ( (graphe == NULL) || (correct == NULL) )  return ERR_GRF_NULL;

  if ( (correct->x_min >= graphe->x_max ) || (correct->x_max <= graphe->x_min) ) {
    (void) EXF_ERREUR(ERR_DISJOINTS);
    return ERR_GRF_DISJOINTS;
  }

  retour = GRF_ChangerGenre( graphe, GRF_FCT_COMPLEXE );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de convertir en complexe [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return FAUX;
  }

  retour = SPL_tt_prep( correct, &b, &c, &d);
  if ( retour < 0 ) { /* On prpare la Spline */
    (void) LIS_ERREUR( retour, correct->nom );
    return FAUX;
  }

  x = graphe->x;
  y = (Complexe *) graphe->y;

  for (i = 0 ; i < graphe->nombre_point ; i++) {
    pha = SPL_tt_calc( correct, *x, b, c, d);
    cosinus = cos(pha); sinus = sin(pha);
    reelle = y->re; imag = y->im;

    y->re = reelle * cosinus + imag   * sinus;
    y->im = imag   * cosinus - reelle * sinus;

    x++; y++;
  }

  free(b); free(c); free(d); /* On libre la mmoire */
  return VRAI;
}

/* ~~~~~~~~ Correction des phases et amplitudes extraites par TFI ~~~~~~~~ *
 |                                                                         |
 | Correspond en fait au passage amplitude totale <-> fonction amplitude   |
 |                               phase totale     <-> dphasage            |
 |                                                                         |
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ------------- Correction de la phase ---------------------------------------------------\
 |  Renvoie le graphe corrig, si la correction s'est faite, NULL sinon                    |
 |                                                                                         |
 | Principe : aprs TF, on a khi(k) = Re( Amp * exp(i * Phase) ) = Amp * cos (Phase)       |
 |            or on veut la phase lectronique, phase. La relation entre les deux est :    |
 |                                                                                         |
 | Phase = 2 k r + (phase lectronique) - pi/2 [le pi/2 est pour passer de cos  sin...]   |
 | sens : < 0 = conversion lectronique -> total                                           |
 |          0 = demander  l'utilisateur                                                   |
 |        > 0 = conversion total -> lectronique                                           |
 \---------------------------------------------------------------------------------------- */

Graphe *TFI_CorrigerPhase(Graphe *graphe, int sens, double longueur)
{
  unsigned long i;
  double *x, *y;
  double decalage = 0.0;

  if (graphe == NULL) return NULL; /* N'a pas de sens sinon */

  if ( FAUX == GRF_EstFonctionReelle( graphe ) ) {
    EXF_ERREUR(GRAPHE_COMPLEXE);
    return NULL;
  }

  /* Doit-on demander le sens et les valeurs  l'utilisateur ? */
  if ( 0 == sens) {
    int reponse;
    struct infos_demande valeurs[ 2 ] = {
      { BI_CHAMP_DEMANDE_REEL    , "R =", ANGSTROM, BI_EDITABLE_MIN|BI_EDITABLE_UNITE, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
      { BI_CHAMP_DEMANDE_ENTIER_S, "n =", NULL, BI_EDITABLE_ENTIER                   , 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } }
    };

    /* On demande donc  l'utilisateur les valeurs pour la conversion */
    reponse = BI_DemandeMultiple( NULL, &Q_CorrectionPhase, 2, valeurs );
    if ( 2 == reponse ) sens = -1;
    else if ( 1 == reponse ) sens = +1;
    else return NULL; /* Correction annule */

    /* On lit les valeurs */
    longueur = valeurs[ 0 ].valeur_donnee.reel;
    decalage = ( 2.0 * (double) valeurs[ 1 ].valeur_donnee.entier_S ) * M_PI;
  }

  longueur *= 2.0;
  decalage -= M_PI / 2.0;
  if ( sens > 0 ) {
    HIST_Titre(1, "Conversion phase totale -> phase "e_AIGU"lectronique", "TFI");

    /* total -> lectronique : f = F + pi/2 - 2 * k * r */
    for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
      *y -= longueur * *x + decalage;
    }

    /* Lgende axe des Y */
    GRF_ChangerLegende( graphe, AXE_Y, (char *) legende_phase_electron, VRAI );
  } else {
    HIST_Titre(1, "Conversion phase "e_AIGU"lectronique -> phase totale", "TFI");

    /* lectronique -> total : F = f - pi/2 + 2 * k * r */
    for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
      *y += longueur * *x + decalage;
    }

    /* Lgende axe des Y */
    GRF_ChangerLegende( graphe, AXE_Y, (char *) legende_phase_total, VRAI );
  }

  /* Les bornes en Y ont pu changer, pas celles en X  */
  GRF_bornes( graphe, NE_PAS_FAIRE, DESORDONNE );

  /* Log. dans l'historique */
  HIST_Texte(1, 0, "Graphe trait"e_AIGU" : %s\n", graphe->nom);
  HIST_Texte(3, 0, "Distance        : %f "ANGSTROM"\n", longueur);
  HIST_Texte(5, 0, "D"e_AIGU"calage : %f\n", decalage);
	
  return graphe;
}

/* ------------ Correction de l'amplitude --------------------------------------------------\
 |    Renvoie le graphe si la correction a t effectue                                    |
 | Postulat  : le graphe contient une amplitude correspondant  une seule couche EXAFS      |
 |                                                                                          |
 | Principe de la correction : on connat khi(k) = Amp * cos( Phase )                       |
 |                             on cherche l'amplitude lectronique, amp, dfinie par :      |
 |                                                                                          |
 |                     2  2                                                                 |
 |         n   N   -2 s  k   -2 R / lambda(k)                                               |
 |  Amp = k  ---- e         e                 * (amplitude lectronique)                    |
 |              2                                                                           |
 |           k R                                                                            |
 \----------------------------------------------------------------------------------------- */

Graphe *corr_amplitude(Graphe *graphe, double n_voisin, double longueur, double sigma)
{
  int type_lpm = CA_LPM_AUCUN, n = 0;
  double gamma = 0.0, eta = 0.0, r_carre;
  Graphe *libre_parcours = NULL;

  if (graphe == NULL) return NULL;

  if ( graphe->genre_donnees != GRF_FCT_REELLE ) {
    EXF_ERREUR(GRAPHE_COMPLEXE);
    return NULL;
  }

  if ( (n_voisin <= 0.0) || (sigma < 0.0) || (longueur <= 0.0) ) { /* Valeurs irralistes */
    int retour;
    extern const char *txt_aucun_choisi, *txt_aucun_graphe;
    boolean lpm_existe;

    /* Existe-t-il des graphes de type libre parcours ou supposs tels */
    lpm_existe = GRF_type_existe(GR_LIBREP, FAUX, NULL);
    OBJ_active(CHOIX_AMPLITUDE, CORAMP_GRAPHE , lpm_existe);
    OBJ_active(CHOIX_AMPLITUDE, CORAMP_LPM_GRF, lpm_existe);
    if (lpm_existe)
      met_chaine(CHOIX_AMPLITUDE, CORAMP_GRAPHE, LG_MAX_NOM_LPM, txt_aucun_choisi, TXT_COPIER);
    else
      met_chaine(CHOIX_AMPLITUDE, CORAMP_GRAPHE, LG_MAX_NOM_LPM, txt_aucun_graphe, TXT_COPIER);

    if (arbres[CHOIX_AMPLITUDE].drapeaux.prepare == FAUX)
      prepare_amplitude();

    do {
      retour = ARB_Bloquant( CHOIX_AMPLITUDE, titre_arbre_amplitude, principale, 0 );
      if (retour == AMP_ANNULE) return NULL; /* On annule le calcul */
    } while (retour != AMP_OK);

    n_voisin = OBJ_LireNombre( CHOIX_AMPLITUDE, VOISINS, NULL );
    sigma    = OBJ_LireNombre( CHOIX_AMPLITUDE, AMP_DEBYE, NULL );
    longueur = OBJ_LireNombre( CHOIX_AMPLITUDE, AMP_LONGUEUR, NULL );
    n = OBJ_LireEntier_U( CHOIX_AMPLITUDE, CA_PONDERATION );
  }

  if ( OBJ_choisit(CHOIX_AMPLITUDE, CORAMP_LPM_CT, QUESTION) ) {
    type_lpm = CORAMP_LPM_CT;
    gamma = OBJ_LireNombre( CHOIX_AMPLITUDE, CORAMP_LPM_VAL, NULL );
  } else if ( OBJ_choisit(CHOIX_AMPLITUDE, CORAMP_LPM_VAR, QUESTION) ) {
    type_lpm = CORAMP_LPM_VAR;
    gamma = OBJ_LireNombre( CHOIX_AMPLITUDE, CORAMP_LPM_VAL, NULL );
    eta  = OBJ_LireNombre( CHOIX_AMPLITUDE, CORAMP_VAL_ETA, NULL );
  } else if ( OBJ_choisit(CHOIX_AMPLITUDE, CORAMP_LPM_GRF, QUESTION) == VRAI ) {
    type_lpm = CORAMP_LPM_GRF;
    libre_parcours = (Graphe *) ARB_LireComplement( CHOIX_AMPLITUDE );
    if (libre_parcours == NULL)
      type_lpm = CA_LPM_AUCUN;
  } else
    type_lpm = CA_LPM_AUCUN;

  if (n_voisin == 0.0) {
    EXF_ERREUR(DIV_ZERO, "Nombre de voisins nul", txt_err_calc_annule);
    return NULL;
  }

  /* Calcul de la correction */
  r_carre = longueur * longueur; /* On prend le carr */
  if ( 0.0 == r_carre ) {
    EXF_ERREUR(DIV_ZERO, "Distance nulle", txt_err_calc_annule);
    return NULL;
  }

  /* Conversion effective amplitude totale -> amplitude lectronique */
  /* 1. Terme constant : R/N */
  GRF_FacteurMultiplicatif( graphe, r_carre / n_voisin, AXE_Y, ANGSTROM""DEUX_EXPOSANT );

  /* 2. Pondration par k^n, terme en 1/k ==> pondration par k^(1 - n) */
  EXF_k_p_n(graphe, 1 - n, FAUX);

  /* 3. Terme de Debye-Waller, s'il existe */
  if ( sigma != 0.0 ) {
    unsigned long i;
    double *x, *y;

    sigma *= 2.0;

    for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
      double DW;

      DW = sigma * (*x) * (*x);
      DW = exp(DW);
      *y *= DW;
    }

    if ( graphe->erreur_y != NULL ) {
      (void) fprintf(stderr, "corr_amplitude : Debye-Waller, traiter les erreurs en Y !\n");
    }
    if ( graphe->covariance != NULL ) {
      (void) fprintf(stderr, "corr_amplitude : Debye-Waller, traiter la matrice de covariance !\n");
    }
  }

  /* 4. Terme de libre parcours moyen, s'il existe */
  if ( type_lpm != CA_LPM_AUCUN ) {
    if ( CORAMP_LPM_CT == type_lpm ) { /* Constant, exp( 2 * R / lambda ) avec lambda = 1 / Gamma */
      GRF_FacteurMultiplicatif( graphe, exp( 2.0 * longueur * gamma ), AXE_Y, NULL );
    } else if ( CORAMP_LPM_VAR == type_lpm ) { /* lambda(k) = */
      unsigned long i;
      double *x, *y, tmp;

      tmp = longueur * 2.0 * gamma;
      for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
	double lpm;

	if ( *x == 0.0 ) {
	  /* Division par zro ! */
	  continue;
	}
	lpm = eta / *x;
	lpm *= lpm;
	lpm *= lpm; /* Contient donc (eta / k)^4 */

	lpm = exp( tmp * (*x + lpm) ); /* 2 R G (k + (n/k)^4 */
	*y *= lpm;
      }

      if ( graphe->erreur_y != NULL ) {
	(void) fprintf(stderr, "corr_amplitude : LPM / eta, traiter les erreurs en Y !\n");
      }
      if ( graphe->covariance != NULL ) {
	(void) fprintf(stderr, "corr_amplitude : LPM / eta, traiter la matrice de covariance !\n");
      }
    } else if ( CORAMP_LPM_GRF == type_lpm ) { /* lambda(k) dans un graphe */
      double *b = NULL, *c = NULL, *d = NULL;
      int retour;

      retour = SPL_tt_prep( libre_parcours, &b, &c, &d);
      if ( retour != 0 ) { /* Erreur dans la Spline */
	(void) LIS_ERREUR( retour, libre_parcours->nom );
      } else {
	double *x, *y, tmp;
	unsigned long i;

	tmp = 2.0 * longueur;
	for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
	  double lpm;

	  lpm = SPL_tt_calc(libre_parcours, *x, b, c, d);
	  if ( lpm == 0.0 ) {
	    /* Division par zro ! */
	    continue;
	  }

	  lpm = exp( tmp / lpm );
	  *y *= lpm;
	}

	if ( graphe->erreur_y != NULL ) {
	  (void) fprintf(stderr, "corr_amplitude : LPM / spline, traiter les erreurs en Y !\n");
	}
	if ( graphe->covariance != NULL ) {
	  (void) fprintf(stderr, "corr_amplitude : LPM / spline, traiter la matrice de covariance !\n");
	}

	free(b); free(c); free(d);
      }
    } else {
      (void) fprintf(stderr, "corr_amplitude : type de LPM inconnu !\n");
    }
  }
  
  GRF_bornes(graphe, ORDONNE, DESORDONNE);

  /* On met une lgende en Y plausible */
  GRF_ChangerLegende( graphe, AXE_Y, (char *) legende_amplitude, VRAI );

  /* Log. dans l'historique */
  HIST_Titre(1, "Conversion amplitude totale -> amplitude "e_AIGU"lectronique", "TFI");
  HIST_Texte(1, 0, "Graphe trait"e_AIGU" : %s\n", graphe->nom);
  HIST_Texte(3, 0, "Nombre de voisins       : %f\n", n_voisin);
  HIST_Texte(3, 0, "Distance                : %f\n", longueur);
  HIST_Texte(3, 0, "Facteur de Debye-Waller : %f\n", sigma);
  HIST_Texte(2, 0, "Libre parcours moyen    : ");
  if (type_lpm == CA_LPM_AUCUN)
    HIST_Texte(2, 0, "non pris en compte");
  else if (type_lpm == CORAMP_LPM_CT) {
    HIST_Texte(2, 0, "constant");
    HIST_Texte(3, 0, " ("GAMMA_MAJUSCULE" = %f)", gamma);
  } else if (type_lpm == CORAMP_LPM_VAR) {
    HIST_Texte(2, 0, "variable");
    HIST_Texte(3, 0, " ("GAMMA_MAJUSCULE" = %f, eta = %f)", gamma, eta);
  } else if (type_lpm == CORAMP_LPM_GRF) {
    HIST_Texte(2, 0, "variable");
    HIST_Texte(3, 0, "(graphe %s)", libre_parcours->nom);
  }
  HIST_Texte(2, 0, "\n");

  return graphe; /* Conversion a t effectue */
}

/* ------ Gestion de l'arbre associ  la correction d'amplitude extraite ------ */

static int GER_amplitude(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  Graphe *graphe;

  switch (objet)
  {
  case CA_LPM_AUCUN:   /* Pas de correction du libre parcours moyen */
  case CORAMP_LPM_CT:  /* On demande correction de libre parcours moyen, constante */
  case CORAMP_LPM_VAR: /* On demande correction de libre parcours moyen, avec formule  deux variables */
  case CORAMP_LPM_GRF: /* On demande correction de libre parcours moyen, dj calcul dans un graphe */
    OPTION_A_FAIRE( "Activation convenable des champs complmentaires, si besoin est !" );
    OBJ_choisit(CHOIX_AMPLITUDE, objet, FORCE_OUI);
    return BI_EVENEMENT_TRAITE;

  case CORAMP_GRAPHE: /* Quel graphe de libre parcours utiliser pour la correction ? */
    graphe = GRF_choisir(GR_LIBREP, VRAI, VRAI, NULL);
    ARB_MettreComplement( CHOIX_AMPLITUDE, graphe );
    met_chaine( CHOIX_AMPLITUDE, CORAMP_GRAPHE, 30, graphe->nom, TXT_COPIER );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

static void prepare_amplitude(void)
{
  if ( FAUX == arbres[ CHOIX_AMPLITUDE ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( CHOIX_AMPLITUDE, "menu_filtre.html#corr_amplitude", GER_amplitude, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }

  OBJ_deplace(CHOIX_AMPLITUDE, AMP_LONGUEUR, 0, 1, DEPLACE_PIXELS);
  OBJ_deplace(CHOIX_AMPLITUDE, AMP_DEBYE, 0, 2, DEPLACE_PIXELS);
  OBJ_deplace(CHOIX_AMPLITUDE, CA_PONDERATION, 0, 3, DEPLACE_PIXELS);

#ifdef _ENGLISH_
  met_chaine(CHOIX_AMPLITUDE, VOISINS     , 0, "Degeneracy         : N = __.____ neighbours", TXT_REMPLACE);
  met_chaine(CHOIX_AMPLITUDE, AMP_LONGUEUR, 0, "Path length        : R = __.____ "ANGSTROM, TXT_REMPLACE);
  met_chaine(CHOIX_AMPLITUDE, AMP_DEBYE   , 0, "Debye-Waller factor: "SIGMA""DEUX_EXPOSANT" = _._______ "ANGSTROM""DEUX_EXPOSANT, TXT_REMPLACE);
  met_chaine(CHOIX_AMPLITUDE, CA_PONDERATION, 0, "Weighting by k^__", TXT_REMPLACE);
  met_chaine(CHOIX_AMPLITUDE, CA_TXT_LPM, 0, "Mean free path:", TXT_REMPLACE);
  met_chaine(CHOIX_AMPLITUDE, AMP_ANNULE, 0, "Cancel", TXT_REMPLACE);
  met_chaine(CHOIX_AMPLITUDE, CA_TXT_AUCUN, 0, "Keep", TXT_REMPLACE);
#endif

  arbres[CHOIX_AMPLITUDE].drapeaux.prepare = VRAI;
}
