/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2003, Emmanuel CURIS

   Recherche des bornes d'un graphe, de type quelconque
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "graphes.h"
#include "science.h" /* pour M_PI, au cas o */

#include "stats_utils.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GRF_Bornes_3D(Graphe *graphe, int en_x, int en_y, int en_z);
static int GRF_Bornes_C (Graphe *graphe, int en_y);
static int GRF_Bornes_R (Graphe *graphe, int en_x, int en_y);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Appel depuis l'extrieur ~-~-~-~-~-~-~-~- */

/* -------- Calculer les bornes d'un graphe ------------------\
 |                                                            |
 | graphe : le graphe  traiter                               |
 | en_x   : NE_PAS_FAIRE = on ne les cherche pas              |
 |          ORDONNE      = ils sont tris par ordre croissant |
 |          DESORDONNE   = ils ne sont pas tris              |
 | en_y   : mme chose mais pour les Y                        |
 \----------------------------------------------------------- */

void GRF_bornes(Graphe *graphe, int en_x, int en_y)
{
  int retour = 0;

  if ( NULL == graphe ) return; /* Pas de graphe */

  if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
    retour = GRF_Bornes_R( graphe, en_x, en_y );
  } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    /* On s'occupe d'abord des X, facile */
    if ( NULL == graphe->x ) {
      (void) fprintf(stderr, "GRF_Bornes_R : graphe incomplet, ERREUR GRAVE [x = %p, y = %p]\n",
		     (void *) graphe->x, (void *) graphe->y);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return;
    }

    if ( DESORDONNE == en_x ) {
      STAT_ChercherExtrema_D( graphe->x, graphe->nombre_point, &(graphe->x_min), &(graphe->x_max) );
    } else if ( ORDONNE == en_x ) {
      graphe->x_min = graphe->x[0];
      graphe->x_max = graphe->x[graphe->nombre_point - 1];
    }

    /* On s'occupe maintenant des Y */
    retour = GRF_Bornes_C( graphe, en_y );
  } else if ( GRF_COURBE_ESPACE == graphe->genre_donnees ) {
    retour = GRF_Bornes_3D( graphe, en_x, en_y, DESORDONNE );
  } else if ( GRF_SERIE == graphe->genre_donnees ) {
    if ( NULL == graphe->x ) {
      if ( DESORDONNE == en_y ) {
	STAT_ChercherExtrema_D( graphe->y, graphe->nombre_point, &(graphe->y_min), &(graphe->y_max) );
      } else if ( en_y != NE_PAS_FAIRE ) {
	graphe->y_min = graphe->y[ 0 ];
	graphe->y_max = graphe->y[ graphe->nombre_point - 1 ];
      }
    } else {
      /* On peut faire exactement comme pour un graphe rel, puisqu'il y a une table des X */
      retour = GRF_Bornes_R( graphe, en_x, en_y );
    }
  } else if ( GRF_ERREUR_R == graphe->genre_donnees ) {
    if ( DESORDONNE == en_y ) {
      STAT_ChercherExtrema_D( graphe->y, graphe->nombre_point, &(graphe->y_min), &(graphe->y_max) );
    } else if ( en_y != NE_PAS_FAIRE ) {
      graphe->y_min = graphe->y[ 0 ];
      graphe->y_max = graphe->y[ graphe->nombre_point - 1 ];
    }
  } else if ( GRF_ERREUR_C == graphe->genre_donnees ) {
    retour = GRF_Bornes_C( graphe, en_y );
  } else {
    retour = -5;
    (void) fprintf(stderr, "GRF_bornes : impossible de calculer les bornes pour ce type de graphe [%d]\n", graphe->genre_donnees);
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "GRF_bornes : erreur lors de la recherche des bornes [code : %d]\n", retour);
  }

#ifdef _DEVELOPPE_
  /* Vrifications post-recherche */
  if ( ( graphe->x_min == graphe->x_max ) && ( en_x != NE_PAS_FAIRE ) ) {
    (void) fprintf(stderr, "GRF_bornes : graphe  X constants [%g, %lu points / \"%s\"]\n", graphe->x_min, graphe->nombre_point, graphe->nom);
    /* Peut se produire lors de chargements partiels... */
  }
  if ( ( graphe->y_min == graphe->y_max ) && ( en_y != NE_PAS_FAIRE ) ) {
    (void) fprintf(stderr, "GRF_bornes : graphe  Y constants [%g, %lu points / \"%s\"]\n", graphe->y_min, graphe->nombre_point, graphe->nom);
  }
#endif

  GRF_MiseAJour( graphe, VRAI );
}

/* -------- Indiquer l'intervalle en X utilis par la plage en Y donne -------- */

int GRF_QuelsXPourY(Graphe *graphe, double y_min, double y_max, double *x_min, double *x_max)
{
  double *x;
  double min, max;
  unsigned long i;

  if ( NULL == graphe ) return -1;

  if ( ( graphe->y_min >= y_max ) || ( graphe->y_max <= y_min ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s : le graphe %s n'a pas de valeurs en Y entre %g et %g\n", __FUNCTION__, graphe->nom, y_min, y_max);
#endif
    return +1; /* Pas de valeur dans x_min, x_max */
  }

  /* Valeurs absurdes pour le dbut */
  min = graphe->x_max;
  max = graphe->x_min;

  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    Complexe *y = (Complexe *) (graphe->y);
    double module;

    for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++, y++) {
      module = sqrt( y->re * y->re + y->im * y->im );
      if ( ( -module >= y_min ) && ( module <= y_max ) ) { /* On est bon en Y */
	if ( *x < min ) min = *x;
	if ( *x > max ) max = *x;
      }
    }
  } else {
    double *y = graphe->y;

    for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++, y++) {
      if ( (*y >= y_min) && (*y <= y_max) ) { /* On est bon en Y */
	if (*x > max) max = *x;
	if (*x < min) min = *x;
      }
    }
  }

  *x_min = min; *x_max = max;
  return 0;
}

/* -------- Indiquer l'intervalle en Y utilis par la plage en X donne -------- */

int GRF_QuelsYPourX(Graphe *graphe, double x_min, double x_max, double *y_min, double *y_max)
{
  double *x;
  double min = 0.0, max = 0.0;
  unsigned long i;

  if ( NULL == graphe ) return -1;

  if ( ( graphe->x_min >= x_max ) || ( graphe->x_max <= x_min ) ) return +1; /* Pas de valeur dans x_min, x_max */

  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    Complexe *y = (Complexe *) (graphe->y);
    double r_min, r_max, i_min, i_max, m_max;
    int methode = GRF_C_IM_MOD;

    if ( graphe->affichage != NULL ) {
      methode = graphe->affichage->methode;
    }

    r_min = graphe->y_max; r_max = graphe->y_min; /* On est sr que les Y sont dans cet intervalle... */
    i_min = graphe->y_max; i_max = graphe->y_min;
    m_max = 0.0; /* Module : on travaille sur le carr */

    /* Boucle complte car rien ne prouve que les X soient tris */
    for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++, y++) {
      if ( (*x >= x_min) && (*x <= x_max) ) { /* On est bon en X */
	double module;
	if (y->re < r_min) r_min = y->re;
	if (y->re > r_max) r_max = y->re;
	if (y->im < i_min) i_min = y->im;
	if (y->im > i_max) i_max = y->im;
	module = y->re * y->re + y->im * y->im;
	if (module > m_max) m_max = module;
      }
    }

    /* Suivant l'affichage, on choisit ce qu'il faut, de toute faon rarement m_min */
    if ( GRF_C_IM_MOD == methode ) { /* Im + module */
      min = i_min; max = sqrt(m_max); /* On a forcment |y| >= Im(y) */
    } else if ( GRF_C_RE_MOD == methode ) { /* Re + module */
      min = r_min; max = sqrt(m_max);
    } else if ( GRF_C_RE_IM  == methode ) { /* Re + Im */
      (r_min < i_min) ? (min = r_min) : (min = i_min);
      (r_max > i_max) ? (max = r_max) : (max = i_max);
    }
  } else {
    double *y = graphe->y;
    min = graphe->y_max; max = graphe->y_min; /* On est sr que les Y sont dans cet intervalle... */

    /* Boucle complte car rien ne prouve que les X soient tris */
    for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++, y++) {
      if ( (*x >= x_min) && (*x <= x_max) ) { /* On est bon en X */
	if (*y > max) max = *y;
	if (*y < min) min = *y;
      }
    }
  }

  *y_min = min; *y_max = max;
  return 0;
}

/* ~-~-~-~-~-~-~-~- Routines internes ~-~-~-~-~-~-~-~- */

/* -------- Cas d'un graphe (x,y)  valeurs relles -------- */

static int GRF_Bornes_R(Graphe *graphe, int en_x, int en_y)
{
  if ( ( graphe->x == NULL ) || ( NULL == graphe->y ) ) {
    (void) fprintf(stderr, "GRF_Bornes_R : graphe incomplet, ERREUR GRAVE [x = %p, y = %p]\n",
		   (void *) graphe->x, (void *) graphe->y);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

/* Vu que les deux tables sont distinctes, et vu la vitesse des machines actuelles,
   il est aussi simple de faire deux fois une boucle 0->nbr_points... */
  if ( DESORDONNE == en_x ) {
    STAT_ChercherExtrema_D( graphe->x, graphe->nombre_point, &(graphe->x_min), &(graphe->x_max) );
  } else if ( ORDONNE == en_x ) {
    graphe->x_min = graphe->x[0];
    graphe->x_max = graphe->x[graphe->nombre_point - 1];
  }

  if ( DESORDONNE == en_y ) {
    STAT_ChercherExtrema_D( graphe->y, graphe->nombre_point, &(graphe->y_min), &(graphe->y_max) );
  } else if ( ORDONNE == en_y ) {
    graphe->y_min = graphe->y[0];
    graphe->y_max = graphe->y[graphe->nombre_point - 1];
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Cas d'un graphe (x,y)  valeurs complexes -------- */

static int GRF_Bornes_C(Graphe *graphe, int en_y)
{
  Complexe *y;
  double re_min = 0.0, im_min = 0.0, mod_min = 0.0;
  double re_max = 0.0, im_max = 0.0, mod_max = 0.0;

  if ( NULL == graphe->y ) {
    (void) fprintf(stderr, "GRF_Bornes_C : graphe incomplet, ERREUR GRAVE [x = %p, y = %p]\n",
		   (void *) graphe->x, (void *) graphe->y);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  /* On s'occupe maintenant des Y */
  if ( DESORDONNE == en_y ) {
    unsigned long i;
    double module;

    y = (Complexe *) graphe->y;
    re_min = y->re; im_min = y->im;
    mod_min = re_min * re_min + im_min * im_min;
    re_max = re_min; im_max = im_min; mod_max = mod_min;
    for ( i = 1, y++ ; i < graphe->nombre_point ; i++, y++ ) {
      module = y->re * y->re + y->im * y->im;

      if ( y->re < re_min ) re_min = y->re;
      if ( y->im < im_min ) im_min = y->im;
      if ( module < mod_min ) mod_min = module;

      if ( y->re > re_max ) re_max = y->re;
      if ( y->im > im_max ) im_max = y->im;
      if ( module > mod_max ) mod_max = module;

#ifdef _DEVELOPPE_
      if ( is_nan( y->re ) || ( is_nan( y->im ) ) || ( is_nan( module ) ) ) {
	(void) fprintf(stderr, "%s,%s l.%u : Points i=%lu [0  %lu], NaN : Re=%g, Im=%g, |z|=%g [X=%g ; graphe %s]\n",
		       __FILE__, __FUNCTION__, __LINE__, i, graphe->nombre_point - 1, y->re, y->im, module,
		       ( graphe->x != NULL ) ? ( graphe->x[ i ] ) : (0.0), graphe->nom );
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	return -1;
      }
#endif
    }
  } else if ( ORDONNE == en_y ) {
    /* Minimum : premier point */
    y = (Complexe *) graphe->y;
    re_min = y->re; im_min = y->im;
    mod_min = re_min * re_min + im_min * im_min;

    /* Maximum : dernier point */
    y = (Complexe *) (graphe->y + graphe->nombre_point - 1);
    re_max = y->re; im_max = y->im;
    mod_max = re_max * re_max + im_max * im_max;
  }

  if ( en_y != NE_PAS_FAIRE ) {
    int methode = GRF_C_IM_MOD;

    if ( graphe->affichage != NULL ) {
      methode = graphe->affichage->methode;
    }
    if ( GRF_C_IM_MOD == methode ) {
      graphe->y_min = im_min;
      graphe->y_max = sqrt( mod_max );
    } else if ( GRF_C_RE_MOD == methode ) {
      graphe->y_min = re_min;
      graphe->y_max = sqrt( mod_max );
    } else {
      graphe->y_min = ( re_min < im_min ) ? ( re_min ) : ( im_min );
      graphe->y_max = ( re_max > im_max ) ? ( re_max ) : ( im_max );
    }
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Cas d'un graphe 3D (x,y,z)  valeurs relles -------- */

static int GRF_Bornes_3D(Graphe *graphe, int en_x, int en_y, int en_z)
{
  double z_min, z_max;

  if ( ( graphe->x == NULL ) || ( NULL == graphe->y ) || ( NULL == graphe->z.z ) ) {
    (void) fprintf(stderr, "GRF_Bornes_3D : graphe incomplet, ERREUR GRAVE [x = %p, y = %p, 3D = %p]\n",
		   (void *) graphe->x, (void *) graphe->y, (void *) graphe->z.z);
    return -1;
  }

  /* On s'occupe d'abord des X, facile */
  if ( DESORDONNE == en_x ) {
    STAT_ChercherExtrema_D( graphe->x, graphe->nombre_point, &(graphe->x_min), &(graphe->x_max) );
  } else if ( ORDONNE == en_x ) {
    graphe->x_min = graphe->x[0];
    graphe->x_max = graphe->x[graphe->nombre_point - 1];
  }

  /* Mme chose pour les Y */
  if ( DESORDONNE == en_y ) {
    STAT_ChercherExtrema_D( graphe->y, graphe->nombre_point, &(graphe->y_min), &(graphe->y_max) );
  } else if ( ORDONNE == en_y ) {
    graphe->y_min = graphe->y[0];
    graphe->y_max = graphe->y[graphe->nombre_point - 1];
  }

  /* Mme chose pour les Z */
  if ( DESORDONNE == en_z ) {
    STAT_ChercherExtrema_D( graphe->z.z, graphe->nombre_point, &z_min, &z_max );
  } else if ( ORDONNE == en_z ) {
    z_min = graphe->z.z[0];
    z_max = graphe->z.z[graphe->nombre_point - 1];
  }

  /* On construit la projection 3D  partir de ces rsultats */
  if ( graphe->affichage != NULL ) {
    double dx, dy, dz;
    struct grf_infos_affichage *affichage;

    affichage = graphe->affichage;
    if ( NULL == affichage->complement.infos_3d ) {
      (void) fprintf(stderr, "GRF_Bornes_3D : Graphe 3D, affichage, mais pas d'infos 3D !\n");
      return 0;
    }
    affichage->complement.infos_3d->z_min = z_min;
    affichage->complement.infos_3d->z_max = z_max;

    affichage->complement.infos_3d->x_vu = (graphe->x_min + graphe->x_max) / 2.0;
    affichage->complement.infos_3d->y_vu = (graphe->y_min + graphe->y_max) / 2.0;
    affichage->complement.infos_3d->z_vu = (z_min + z_max) / 2.0;

    affichage->complement.infos_3d->x_oeil = affichage->complement.infos_3d->x_vu;
    affichage->complement.infos_3d->y_oeil = affichage->complement.infos_3d->y_vu;
    affichage->complement.infos_3d->z_oeil = affichage->complement.infos_3d->z_vu;

    dx = graphe->x_max - graphe->x_min; if (dx < 1.0e-3) dx = 1.0e-3;
    dy = graphe->y_max - graphe->y_min; if (dy < 1.0e-3) dy = 1.0e-3;
    dz = affichage->complement.infos_3d->z_max  - affichage->complement.infos_3d->z_min ; if (dz < 1.0e-3) dz = 1.0e-3;

    if ( (dz > dx) && (dz > dy) ) { /* Axe Z le plus utilis -> horizontal sur l'cran */
      affichage->complement.infos_3d->z_h = 0.0;
      if ( dy > dx ) { /* Axe Y second axe, vertical  l'cran : oeil sur l'axe X */
	affichage->complement.infos_3d->x_oeil += dz / (2 * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) + dx / 2.0;
	affichage->complement.infos_3d->x_h = 0.0;
	affichage->complement.infos_3d->y_h = 1.0;
	affichage->complement.infos_3d->angle_h = 2 * atan( dy/dz * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) * 180.0 / M_PI;
      } else { /* Axe X second axe, vertical  l'cran : oeil sur l'axe Y */
	affichage->complement.infos_3d->y_oeil += dz / (2 * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) + dy / 2.0;
	affichage->complement.infos_3d->x_h = 1.0;
	affichage->complement.infos_3d->y_h = 0.0;
	affichage->complement.infos_3d->angle_h = 2 * atan( dx/dz * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) * 180.0 / M_PI;
      }
    } else if ( (dy > dz) && (dy > dx) ) { /* Axe Y le plus utilis, horizontal sur l'cran */
      affichage->complement.infos_3d->y_h = 0;
      if (dz > dx) { /* oeil sur l'axe des X */
	affichage->complement.infos_3d->x_oeil += dy / (2 * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) + dx / 2.0;
	affichage->complement.infos_3d->x_h = 0.0;
	affichage->complement.infos_3d->z_h = 1.0;
	affichage->complement.infos_3d->angle_h = 2 * atan( dx/dy * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) * 180.0 / M_PI;
      } else { /* oeil sur l'axe des Z */
	affichage->complement.infos_3d->z_oeil += dy / (2 * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) + dz / 2.0;
	affichage->complement.infos_3d->x_h = 1.0;
	affichage->complement.infos_3d->z_h = 0.0;
	affichage->complement.infos_3d->angle_h = 2 * atan( dz/dy * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) * 180.0 / M_PI;
      }
    } else { /* Axe X le plus utilis */
      affichage->complement.infos_3d->x_h = 0;
      if (dz > dy) { /* Axe Z second axe principal : oeil sur l'axe des Y */
	affichage->complement.infos_3d->y_oeil += dx / (2 * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) + dy / 2.0;
	affichage->complement.infos_3d->y_h = 0;
	affichage->complement.infos_3d->z_h = 1.0;
	affichage->complement.infos_3d->angle_h = 2 * atan( dy/dz * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) * 180.0 / M_PI;
      } else { /* Axe Y second axe principal : oeil sur l'axe des Z */
	affichage->complement.infos_3d->z_oeil += dx / (2 * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) + dz / 2.0;
	affichage->complement.infos_3d->y_h = 1.0;
	affichage->complement.infos_3d->z_h = 0;
	affichage->complement.infos_3d->angle_h = 2 * atan( dy/dx * tan( affichage->complement.infos_3d->angle_v * M_PI / 360.0 ) ) * 180.0 / M_PI;
      }
    }
  }

  /* Tout s'est bien pass */
  return 0;
}
