/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS

   Opérations diverses, très simples, sur le contenu des graphes
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "lissage.h"
#include "matrices.h"
#include "serie.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GRF_ReelVersComplexe(Graphe *graphe);

/* ~~~~~~~~ Réalisation ~~~~~~~~~~~~~~ */

/* ~-~-~-~-~-~-~-~ Informations sur les propriétés du graphe ~-~-~-~-~-~-~-~ */

inline boolean GRF_EstIsole(const Graphe *graphe)
{
  if ( NULL == graphe ) return VRAI;
  if ( VRAI == graphe->drapeaux.isole ) return VRAI;
  return FAUX;
}

inline boolean GRF_EstProtege(const Graphe *graphe)
{
  if ( NULL == graphe ) return VRAI;
  if ( VRAI == graphe->drapeaux.protege ) return VRAI;
  return FAUX;
}

inline void GRF_DeprotegerGraphe(Graphe *graphe)
{
  if ( NULL == graphe ) return;
  graphe->drapeaux.protege = FAUX;
}

inline void GRF_ProtegerGraphe(Graphe *graphe)
{
  if ( NULL == graphe ) return;
  graphe->drapeaux.protege = VRAI;
}

/* ~-~-~-~-~-~-~-~ Informations sur les données du graphe ~-~-~-~-~-~-~-~ */

/* -------- Le graphe contient-il des valeurs complexes ? -------- */

inline boolean GRF_EstComplexe(const Graphe *graphe)
{
  if ( NULL == graphe ) return FAUX;
  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) return VRAI;
  if ( GRF_ERREUR_C == graphe->genre_donnees ) return VRAI;

  return FAUX;
}

/* -------- Le graphe contient-il une fonction, Y=f(X) ou Z=f(X) ? ----- */

inline boolean GRF_EstFonction(const Graphe *graphe)
{
  if ( VRAI == GRF_EstFonctionReelle( graphe ) ) {
    return VRAI;
  }

  if ( NULL == graphe ) return FAUX;

  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees )
    return VRAI;

  return FAUX;
}

/* -------- Le graphe contient-il une fonction Y=f(X), réelle ? -------- */

inline boolean GRF_EstFonctionReelle(const Graphe *graphe)
{
  if ( NULL == graphe ) return FAUX;

  if ( GRF_FCT_REELLE == graphe->genre_donnees ) return VRAI;
  if ( GRF_SERIE == graphe->genre_donnees ) {
    if ( NULL == graphe->x ) return FAUX; /* Pas de table des X => pas utilisable */

    /* Table des X => utilisable comme si c'était un graphe à valeurs réelles */
    return VRAI;
  }

  /* Autre cas : ce n'est pas une fonction à valeurs réelles */
  return FAUX;
}

/* -------- Le graphe contient-il des valeurs réelles ? ----- */

inline boolean GRF_EstReel(const Graphe *graphe)
{
  if ( NULL == graphe ) return FAUX;

  if ( GRF_FCT_REELLE   == graphe->genre_donnees ) return VRAI;
  if ( GRF_SERIE        == graphe->genre_donnees ) return VRAI;
  if ( GRF_COURBE_PLANE == graphe->genre_donnees ) return VRAI;
  if ( GRF_ERREUR_R     == graphe->genre_donnees ) return VRAI;

  /* Autre cas : ce n'est pas un graphe à valeurs réelles */
  return FAUX;
}

/* -------- Le graphe contient-il un spectre EXAFS ? -------- */

inline boolean GRF_EstSpectreEXAFS(const Graphe *graphe, boolean strict)
{
  /* Un spectre EXAFS est forcément à valeurs réelles
     Mais peut être une série avec pseudo-X : tirages MC par exemple ! */
  if ( FAUX == GRF_EstFonctionReelle( graphe ) ) return FAUX;

  if ( GR_EXAFS == graphe->type ) return VRAI; /* Spectre EXAFS normalisé */
  if ( VRAI == graphe->drapeaux.exafs ) return VRAI; /* Flag mis. Certains types peuvent être ou non EXAFS suivant les circonstances :
							modèles, TFI,... */

  if ( FAUX == strict ) { /* Les spectres non-normalisés intéressent aussi */
    if ( GR_EXAFS1 == graphe->type ) return VRAI;
    if ( GR_EXAFS2 == graphe->type ) return VRAI;
  }

  /* Ce n'est pas un spectre EXAFS */
  return FAUX;
}

/* ~-~-~-~-~-~-~-~- Opérations sur un seul graphe ~-~-~-~-~-~-~-~- */

/* -------- Permuter les X et les Y -------- */

int GRF_PermuterXY(Graphe *graphe)
{
  double *tmp_pd, tmp_d;
  char *legende_X = NULL, *legende_Y = NULL;
  const char *uX, *uY;
  char *unite_X= NULL, *unite_Y = NULL;
  int retour;

  /*  Controles préliminaires */
  if ( NULL == graphe ) {
    return -2;
  }
  if ( ( graphe->genre_donnees != GRF_FCT_REELLE ) && ( graphe->genre_donnees != GRF_COURBE_PLANE ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Opération interdite sur ce type de graphe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On mémorise les légendes & unités, pour ne pas les perdre en changeant de type */
  if ( graphe->affichage != NULL ) {
    legende_X = strdup( GRF_QuelleLegende( graphe, AXE_X ) );
    uX = GRF_QuelleUnite( graphe, AXE_X );
    if ( uX != NULL ) {
      unite_X = strdup( uX );
    }

    legende_Y = strdup( GRF_QuelleLegende( graphe, AXE_Y ) );
    uY = GRF_QuelleUnite( graphe, AXE_Y );
    if ( uY != NULL ) {
      unite_Y = strdup( uY );
    }
  }

  /* On perd le type du graphe, qui devient « autre » */
  retour = GRF_ChangerType( graphe, GR_AUTRETYPE );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Changement de type impossible [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -3;
  }

  /* Permutation des tables X,Y */
  tmp_pd = graphe->x;
  graphe->x = graphe->y;
  graphe->y = tmp_pd;

  /* Permutation des bornes en X,Y */
  tmp_d = graphe->x_min;
  graphe->x_min = graphe->y_min;
  graphe->y_min = tmp_d;

  tmp_d = graphe->x_max;
  graphe->x_max = graphe->y_max;
  graphe->y_max = tmp_d;

  /* Si nécessaire : permutation de l'affichage */
  /* Les légendes sont déjà dans de la mémoire à part : pas besoin d'en refaire une copie ! */
  if ( graphe->affichage != NULL ) {
    GRF_ChangerLegende( graphe, AXE_X, legende_Y, FAUX );
    GRF_ChangerLegende( graphe, AXE_Y, legende_X, FAUX );

    GRF_ChangerUnite( graphe, AXE_X, unite_Y );
    free( unite_Y ); unite_Y = NULL; /* Une copie est faite dans GRF_ChangerUnite */
    GRF_ChangerUnite( graphe, AXE_Y, unite_X );
    free( unite_X ); unite_X = NULL; /* Une copie est faite dans GRF_ChangerUnite */
  }

  /* S'il y a des erreurs, on les permute aussi */
  if ( ( graphe->erreur_x != NULL ) || ( graphe->erreur_y != NULL ) ) {
    Graphe *tmp_pG;
    int tmp_i;

    /* Permutation */
    tmp_pG = graphe->erreur_x;
    graphe->erreur_x = graphe->erreur_y;
    graphe->erreur_y = tmp_pG;

    tmp_i = graphe->drapeaux.erreur_x;
    graphe->drapeaux.erreur_x = graphe->drapeaux.erreur_y;
    graphe->drapeaux.erreur_y = tmp_i;

    /* On leur remet des X cohérents */
    if ( graphe->erreur_x != NULL ) {
      graphe->erreur_x->x_min = graphe->x_min;
      graphe->erreur_x->x_max = graphe->x_max;
    }
    if ( graphe->erreur_y != NULL ) {
      graphe->erreur_y->x_min = graphe->x_min;
      graphe->erreur_y->x_max = graphe->x_max;
    }
  }

  /* Log dans l'historique */
  HIST_Texte(4, 0, "\nX et Y du graphe %s "e_AIGU"chang"e_AIGU"s\n", graphe_courant->nom);

  /* Mise à jour si nécessaire */
  GRF_MiseAJour( graphe, VRAI );

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Ajoute une constante à l'un des axes indiqué -------- */

void GRF_FacteurAdditif(Graphe *graphe, double constante, double err_constante, int axe)
{
  unsigned long i;
  double *valeur;

  if ( NULL == graphe ) return;

  if ( AXE_X == axe ) {
    if ( ( GRF_SERIE == graphe->genre_donnees ) || ( GRF_MULTI_SERIE == graphe->genre_donnees ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Opération interdite, car X implicites ou référencés ailleurs [graphe %s, C=%g ± %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, graphe->nom, constante, err_constante);
#ifdef _DEVELOPPE_
      abort();
#endif
      return;
    }

    /* On ajoute la valeur */
    for ( i = 0, valeur = graphe->x ; i < graphe->nombre_point ; i++, valeur++ ) {
      *valeur += constante;
    }

    /* S'il y a une erreur associée à la constante */
    if ( err_constante > 0.0 ) {
      (void) fprintf(stderr, "[graphe %s, C=%g ± %g]", graphe->nom, constante, err_constante);
      OPTION_A_FAIRE( "Erreur associée à la constante ajoutée en X" );
    }

    /* Mise à jour des bornes */
    graphe->x_min += constante;
    graphe->x_max += constante;
  }

  if ( AXE_Y == axe ) {
    if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
      Complexe *y;

      /* On ajoute la valeur à la partie réelle */
      for ( i = 0, y = (Complexe *) graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
	y->re += constante; 
      }

      if ( err_constante > 0.0 ) { /* Il y a une erreur associée */
	OPTION_A_FAIRE( "Incertitude sur la constante ajoutée [graphe complexe]" );
      }
    } else {
      /* On ajoute la valeur */
      for ( i = 0, valeur = graphe->y ; i < graphe->nombre_point ; i++, valeur++ ) {
	*valeur += constante; 
      }

      if ( err_constante > 0.0 ) { /* Il y a une erreur associée */
	double variance, var_constante;
	Graphe *erreur;

	erreur = graphe->erreur_y;
	if ( NULL == erreur ) { /* Pas d'erreur auparavant */
	  erreur = GRF_creer_erreurs( graphe, AXE_Y );
	  if ( NULL == erreur ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer les erreurs\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    return;
	  }
	}

	var_constante = err_constante * err_constante;

	/* On suppose l'erreur sur la constante indépendante des erreurs sur les Y
	   => les carrés des erreurs (variances) sont additifs */
	for ( i = 0, valeur = erreur->y ; i < erreur->nombre_point ; i++, valeur++ ) {
	  variance = *valeur * *valeur;
	  *valeur = sqrt( var_constante + variance );
	}

	/* On calcule les bornes */
	GRF_bornes( erreur, NE_PAS_FAIRE, DESORDONNE );
      }

      /* On adapte les bornes */
      graphe->y_min += constante;
      graphe->y_max += constante;
    }
  }

  /* Si ce que l'on modifie est le graphe courant : on le redessine ! */
  GRF_MiseAJour( graphe, VRAI );
}

/* -------- Multiplie les X / les Y par la valeur indiquée -------- */

void GRF_FacteurMultiplicatif(Graphe *graphe, double valeur, int axe, const char *unite)
{
  if (NULL == graphe) {
    /* Normal dans le traitement des barres d'erreurs -> ni message, ni arret brutal */
    return;
  }

  if ( axe & AXE_X ) {
    double *x;
    unsigned long i;

    if ( ( NULL == graphe->x ) || ( GRF_SERIE == graphe->genre_donnees ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas de X utilisable !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return;
    }

    /* Les valeurs */
    for (x = graphe->x, i = 0 ; i < graphe->nombre_point ; i++, x++)
      *x *= valeur;

    /* Les bornes */
    if (valeur > 0.0) {
      graphe->x_min *= valeur;
      graphe->x_max *= valeur;
    } else {
      double tmp = graphe->x_min;
      graphe->x_min = valeur * graphe->x_max;
      graphe->x_max = valeur * tmp;
    }

    /* Les barres d'erreur, si nécessaire [X implicites] */
    if ( graphe->erreur_x != NULL ) {
      graphe->erreur_x->x_min = graphe->x_min;
      graphe->erreur_x->x_max = graphe->x_max;

      GRF_FacteurMultiplicatif( graphe->erreur_x, fabs( valeur ), AXE_Y, NULL );
    }
    if ( graphe->erreur_y != NULL ) {
      graphe->erreur_y->x_min = graphe->x_min;
      graphe->erreur_y->x_max = graphe->x_max;
    }

    /* L'unité, si nécessaire */
    if ( unite != NULL ) {
      char *unite_X;

      unite_X = TXT_ProduitUnites( GRF_QuelleUnite( graphe, AXE_X ), unite, 1 );
      GRF_ChangerUnite( graphe, AXE_X, unite_X );
    }
  }

  if ( axe & AXE_Y ) {
    if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
      Complexe *y;
      unsigned long i;

      /* Les valeurs */
      for (y = (Complexe *) graphe->y, i = 0 ; i < graphe->nombre_point ; i++, y++) {
	y->re *= valeur;
	y->im *= valeur;
      }

      /* Les bornes */
      GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE); /* Bornes plus délicates à estimer car dépend du module... */

      /* Matrice de covariance ! */
      if ( graphe->covariance ) {
	double *valeurs;
	double coef = valeur * valeur;
	unsigned long i, j;

	valeurs = graphe->covariance->valeurs;

	for (i = 0 ; i < 2 * graphe->nombre_point ; i++) {
	  for (j = 0 ; j <= i ; j++) {
	    *valeurs *= coef;
	    valeurs++;
	  }
	}
      }
    } else {
      double *y;
      unsigned long i;

      /* Les valeurs */
      for (y = graphe->y, i = 0 ; i < graphe->nombre_point ; i++, y++)
	*y *= valeur;

      /* Les bornes */
      if (valeur > 0.0) {
	graphe->y_min *= valeur;
	graphe->y_max *= valeur;
      } else {
	double tmp = graphe->y_min;
	graphe->y_min = valeur * graphe->y_max;
	graphe->y_max = valeur * tmp;
      }

      /* Matrice de covariance ! */
      if ( graphe->covariance ) {
	double *valeurs;
	double coef = valeur * valeur;
	unsigned long i, j;

	valeurs = graphe->covariance->valeurs;

	for (i = 0 ; i < graphe->nombre_point ; i++) {
	  for (j = 0 ; j <= i ; j++) {
	    *valeurs *= coef;
	    valeurs++;
	  }
	}
      }
    }
    /* Les barres d'erreur, si nécessaire */
    GRF_FacteurMultiplicatif( graphe->erreur_y, fabs(valeur), AXE_Y, NULL );

    /* L'unité, si nécessaire */
    if ( unite != NULL ) {
      char *unite_Y;

      unite_Y = TXT_ProduitUnites( GRF_QuelleUnite( graphe, AXE_Y ), unite, 1 );
      GRF_ChangerUnite( graphe, AXE_Y, unite_Y );
      free( unite_Y ); unite_Y = NULL;
    }
  }

  /* Si ce que l'on modifie est le graphe courant : on le redessine ! */
  GRF_MiseAJour( graphe, VRAI );
}

/* -------- Mettre une valeur constante -------- */

void GRF_MettreConstante(Graphe *graphe, double valeur, unsigned int axe)
{
  unsigned int i;

  if ( NULL == graphe ) return;

  if ( axe & AXE_X ) {
    double *x;

    if ( GRF_SERIE == graphe->genre_donnees ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de changer les X dans un graphe de type série\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
    for ( i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++ ) {
      *x = valeur;
    }
  }

  if ( axe & AXE_Y ) {
    if ( ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) || ( GRF_ERREUR_C == graphe->genre_donnees ) ) {
      Complexe *y;

      for ( i = 0, y = (Complexe *) graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
	y->re = valeur;
	y->im = 0.0;
      }
    } else if ( ( GRF_EstFonctionReelle( graphe ) ) || ( GRF_ERREUR_R == graphe->genre_donnees ) ) {
      double *y;

      for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
	*y = valeur;
      }
    } else {
      AVERTISSEMENT( "Genre de graphe non géré en Y : %d", graphe->genre_donnees );
    }
  }
}

/* -------- Changer le type d'un graphe -------- */

int GRF_ChangerType(Graphe *graphe, int nouveau_type)
{
  boolean delier = VRAI;
  const char *unite_X, *unite_Y;

  if ( NULL == graphe ) return -1;

  if ( nouveau_type == graphe->type ) return 0; /* Rien à faire : on est déjà du bon type ! */

  if ( GR_DEVIATION == nouveau_type ) {
    if ( graphe->z.ddl != NULL ) {
      if ( ( GRF_ERREUR_R == graphe->genre_donnees ) || ( GRF_ERREUR_C == graphe->genre_donnees ) ) {
	/* La table des ddl existe déjà, il suffit donc juste de mettre le nouveau type */
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Il y a déjà des informations en Z, impossible de mémoriser les ddl [%s]\n",
		       __FILE__, __FUNCTION__, __LINE__, graphe->nom);
	return -2;
      }
    } else {
      graphe->z.ddl = (int *) calloc( graphe->nombre_point, sizeof( int ) );
      if ( NULL == graphe->z.ddl ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de mémoriser les ddl\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -3;
      }
    }
  }

  if ( graphe->pointeur.autre != NULL ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : ATTENTION, on perd les infos complémentaires / graphe %s\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->nom);
#endif

    free( graphe->pointeur.autre );
    graphe->pointeur.autre = NULL;
  }
  if ( GR_DEVIATION == graphe->type ) {
    if ( ( GRF_ERREUR_R == graphe->genre_donnees ) || ( GRF_ERREUR_C == graphe->genre_donnees ) ) {
      /* Les ddl sont toujours présents... */
    } else {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : ATTENTION, on perd les d.d.l. / graphe %s\n",
		     __FILE__, __FUNCTION__, __LINE__, graphe->nom);
#endif
      free( graphe->z.ddl );
      graphe->z.ddl = NULL;
    }
  }

  /* Drapeau d'exafs */
  if ( ( GR_EXAFS == nouveau_type ) || ( ( GR_TFI == nouveau_type ) && ( GRF_FCT_REELLE == graphe->genre_donnees ) ) ) {
    if ( VRAI == GRF_EstSpectreEXAFS( graphe, FAUX ) )
      delier = FAUX; /* Pas nécessaire de délier */
    graphe->drapeaux.exafs = VRAI;
  } else {
    graphe->drapeaux.exafs = FAUX;
  }

  /* On isole le graphe des structures qui pourraient l'utiliser
     En effet, avec le changement de type, le lien n'a plus de sens
     et risquerait, en plus, de ne pas être défait à l'effacement du graphe -> plantage */
  if ( VRAI == delier ) {
    GRF_DelierGraphe( GRF_QuelleSerie( graphe ), graphe );
  }

  /* On mémorise les anciennes unités, pour éviter de perdre des unités par défaut */
  unite_X = GRF_QuelleUnite( graphe, AXE_X );
  unite_Y = GRF_QuelleUnite( graphe, AXE_Y );

  /* On change le type */
  graphe->type = nouveau_type;

  /* On met les unités correctes, uniquement si nécessaire
     S'il y avait une unité par défaut, unite_X != NULL => elle est conservé
     S'il n'y en avait pas, et pas d'unité, le résultat ne doit toujours pas en avoir -> "" est l'idéal car on est sûr de ne pas avoir le défaut...
  */
  if ( NULL == unite_X ) {
    GRF_ChangerUnite( graphe, AXE_X, "" );
  } else {
    GRF_ChangerUnite( graphe, AXE_X, unite_X );
  }
  if ( NULL == unite_X ) {
    GRF_ChangerUnite( graphe, AXE_Y, "" );
  } else {
    GRF_ChangerUnite( graphe, AXE_Y, unite_Y );
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Changer le genre des données contenues dans un graphe -------- */

int GRF_ChangerGenre(Graphe *graphe, int nouveau_genre)
{
  if ( NULL == graphe ) return -1;

  if ( nouveau_genre == graphe->genre_donnees ) return 0; /* Rien à faire */

  /* En fonction du genre original... */
  switch ( graphe->genre_donnees )
  {
  case GRF_FCT_REELLE:
    if ( GRF_FCT_COMPLEXE == nouveau_genre ) {
      return GRF_ReelVersComplexe( graphe );
    } else if ( GRF_SERIE == nouveau_genre ) {
      /* Il n'y a plus de table des X */
      free( graphe->x ); graphe->x = NULL;
      graphe->x_min = 0;
      graphe->x_max = graphe->nombre_point - 1;

      /* On n'a plus d'erreurs en X, forcément */
      GRF_eff_isole( graphe->erreur_x );
      graphe->erreur_x = NULL;

      /* On définit le genre correctement */
      graphe->genre_donnees = GRF_SERIE;
      return 0;
    } else if ( GRF_ERREUR_R == nouveau_genre ) {
      /* Réserver les ddl */
      if ( GR_DEVIATION != graphe->type ) {
	graphe->z.ddl = (int *) calloc( graphe->nombre_point, sizeof( int ) );
	if ( NULL == graphe->z.ddl ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -1;
	}
      }

      /* Les X n'existent pas */
      free( graphe->x );
      graphe->x = NULL;

      /* On définit le genre correctement */
      graphe->genre_donnees = GRF_ERREUR_R;
      return 0;
    } else if ( GRF_HISTOGRAMME == nouveau_genre ) {
      unsigned long *effectifs, i;

      /* On réserve la table des effectifs */
      effectifs = (unsigned long *) calloc( graphe->nombre_point + 1, sizeof( unsigned long ) );
      if ( NULL == effectifs ) {
	(void) fprintf(stderr, "%s, %s l.%u : Impossible d'allouer la dernière classe\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -1;
      }

      /* On prépare les effectifs de chaque classe */
      for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
	if ( graphe->y[ i ] <= 0.0 ) {
	  effectifs[ i ] = 0;
	} else {
	  effectifs[ i ] = (unsigned long) ( graphe->y[ i ] );
	}
      }

      /* On met un effectif nul dans la dernière classe, par défaut */
      effectifs[ graphe->nombre_point ] = 0;

      /* On installe la table */
      graphe->z.effectifs = effectifs;

      /* On change le genre */
      graphe->genre_donnees = GRF_HISTOGRAMME;
      if ( graphe->affichage != NULL ) {
	/* Affichage en densité de fréquence par défaut */
	graphe->affichage->methode = GRF_H_DP;
      }

      /* On libère les tables */
      free( graphe->y ); graphe->y = NULL;
      return 0;
    }
    break;

  case GRF_FCT_COMPLEXE: /* Non-transformables simplement : que garder ? */
    if ( GRF_ERREUR_C == nouveau_genre ) {
      /* Réserver les ddl */
      if ( GR_DEVIATION != graphe->type ) {
	graphe->z.ddl = (int *) calloc( graphe->nombre_point, sizeof( int ) );
	if ( NULL == graphe->z.ddl ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -1;
	}
      }

      /* Les X n'existent pas */
      free( graphe->x );
      graphe->x = NULL;

      /* On définit le genre correctement */
      graphe->genre_donnees = GRF_ERREUR_C;
      return 0;
    }
    break;

  case GRF_SERIE:
    if ( GRF_FCT_REELLE == nouveau_genre ) {
      double *nvlle_table;

      nvlle_table = (double *) malloc( graphe->nombre_point * sizeof( double ) );
      if ( NULL == nvlle_table ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la nouvelle table des X\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -5;
      }

      if ( NULL == graphe->x ) {
	unsigned long i;

	for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
	  nvlle_table[ i ] = (double) i;
	}
      } else {
	memcpy( nvlle_table, graphe->x, sizeof( double ) * graphe->nombre_point );
	/* Pas besoin de recalculer les bornes en X, normalement */	   
      }
      graphe->x = nvlle_table; /* Les X précédents éventuels doivent être libérés ailleurs ! */

      graphe->genre_donnees = GRF_FCT_REELLE;
      return 0;
    } else if ( GRF_FCT_COMPLEXE == nouveau_genre ) {
      int retour;

      /* On commence par faire apparaître les X : conversion en fonction réelle */
      retour = GRF_ChangerGenre( graphe, GRF_FCT_REELLE );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : La conversion en réel a échoué [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	return -28;
      }

      /* On fait ensuite la conversion en complexes */
      retour = GRF_ChangerGenre( graphe, GRF_FCT_COMPLEXE );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : La conversion en complexe a échoué [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	return -29;
      }

      /* Conversion réussie */
      return 0;
    }
    break;

  case GRF_ERREUR_R:
    if ( GRF_ERREUR_C == nouveau_genre ) {
      double *vx_y = NULL;
      Complexe *y_comp;
      unsigned long i;

      /* On reserve la nouvelle table de valeurs */
      y_comp = (Complexe *) malloc( sizeof(Complexe) * graphe->nombre_point );
      if ( NULL == y_comp ) {
	(void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return ERR_GRF_MEMOIRE;
      }

      for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
	y_comp[ i ].re = (graphe->y)[ i ];
	y_comp[ i ].im = 0.0;
      }

      vx_y = graphe->y;
      graphe->y = (double *) y_comp;
      graphe->genre_donnees = GRF_ERREUR_C;

      free(vx_y); vx_y = NULL;	/* Plus besoin des anciens Y */
    }
    break;
  }

#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "%s,%s l.%u : Conversion interdite : %d -> %d [%s]\n",
		 __FILE__, __FUNCTION__, __LINE__, graphe->genre_donnees, nouveau_genre, graphe->nom);
#endif
  return -20;
}

/* -------- Transformer un graphe à valeurs réelles en graphe à valeurs complexes -------- */

static int GRF_ReelVersComplexe(Graphe *graphe)
{
  double *vx_y;
  int retour;
  Complexe *y_comp;
  unsigned long i;

  y_comp = (Complexe *) malloc( sizeof(Complexe) * graphe->nombre_point );
  if ( NULL == y_comp ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return ERR_GRF_MEMOIRE;
  }

  for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
    y_comp[ i ].re = (graphe->y)[ i ];
    y_comp[ i ].im = 0.0;
  }

  vx_y = graphe->y;
  graphe->genre_donnees = GRF_FCT_COMPLEXE;
  graphe->y = (double *) y_comp;

  free(vx_y); /* Plus besoin des anciens Y */
  vx_y = NULL;

  /* On met un affichage en (Re, Im) pour être tranquille */
  if ( graphe->affichage != NULL ) {
    graphe->affichage->methode = GRF_C_RE_IM;
  }

  /* S'il y a des erreurs en Y, il faut les transformer aussi ! */
  if ( graphe->erreur_y != NULL ) {
    retour = GRF_ChangerGenre( graphe->erreur_y, GRF_ERREUR_C );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur en convertissant les incertitudes\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -5;
    }
  }

  return 0; /* Tout s'est bien passé */
}

/* -------- Lire une valeur dans un graphe, avec interpolation -------- */

double GRF_LireValeur(const Graphe *graphe, double x, int mode)
{
  double valeur = 0.0;

  /* Vérifications préliminaires */
  if ( NULL == graphe ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "GRF_LireValeur : Appel avec graphe=NULL\n");
#endif
    return 0.0;
  }
  if ( GRF_FCT_REELLE != graphe->genre_donnees ) {
    /* Une série avec X extérieurs est considérée comme un graphe réel */
    if ( ( GRF_SERIE != graphe->genre_donnees ) || ( NULL == graphe->x ) ) {
      (void) fprintf(stderr, "GRF_LireValeur : le graphe n'est pas à valeurs réelles [%s]\n", graphe->nom);
      return 0.0;
    }
  }

  /* Est-on bien dans l'intervalle du graphe ? */
  if ( x < graphe->x_min ) {
    (void) fprintf( stderr, "GRF_LireValeur : ATTENTION, on fait une extrapolation [x=%g, min=%g]\n", x, graphe->x_min );
    return graphe->y[ 0 ]; /* Suppose implicitement un graphe trié par X croissants */
  }
  if ( x > graphe->x_max ) {
    (void) fprintf( stderr, "GRF_LireValeur : ATTENTION, on fait une extrapolation [x=%g, min=%g]\n", x, graphe->x_max );
    return graphe->y[ graphe->nombre_point - 1 ]; /* Suppose implicitement un graphe trié par X croissants */
  }

  /* Actuellement : seulement une interpolation linéaire */
  switch ( mode )
  {
  case INTERPOLE_LINEAIRE:
    valeur = GRF_InterpoleLineaire( graphe, x, NULL );
    break;

  case INTERPOLE_PARABOLE:
    printf("GRF_LireValeur : À FAIRE --- interpolation parabolique\n");
    break;

  case INTERPOLE_CUBIQUE:
    printf("GRF_LireValeur : À FAIRE --- interpolation cubique\n");
    break;

  case INTERPOLE_SPLINE:
    printf("GRF_LireValeur : À FAIRE --- interpolation par spline\n");
    break;

  default:
    (void) fprintf(stderr, "GRF_LireValeur : Méthode d'interpolation inconnue [%d]\n", mode);
    break;
  }

  return valeur;
}

/* -------- Ôter le point i [0..(N-1)] d'un graphe -------- */

int GRF_OterPoint(Graphe *graphe, unsigned long i)
{
  unsigned long j;

  if ( NULL == graphe ) return 0;
  if ( i >= graphe->nombre_point ) {
    AVERTISSEMENT( "Numéro de point absurde : %lu [0 à %lu possible]", i, graphe->nombre_point - 1 );
    return -1;
  }

  /* On se contente de décaler les points suivants d'un cran... */
  switch ( graphe->genre_donnees )
  {
  case GRF_FCT_REELLE:
    for ( j = i ; j < (graphe->nombre_point - 1) ; j++ ) {
      graphe->x[ j ] = graphe->x[ j + 1 ];
      graphe->y[ j ] = graphe->y[ j + 1 ];
    }
    break;

  case GRF_FCT_COMPLEXE:
    {
      Complexe *z;

      for ( j = i, z = (Complexe *) graphe->y ; j < (graphe->nombre_point - 1) ; j++ ) {
	graphe->x[ j ] = graphe->x[ j + 1 ];
	z[ j ] = z[ j + 1 ];
      }
    }
    break;

  case GRF_SERIE:
    /* On garde la correspondance originale numéro de ligne / valeur... ou avec les X indiqués ==> on doit convertir en fonction réelle... */
    GRF_ChangerGenre( graphe, GRF_FCT_REELLE );

    /* Et on peut maintenant faire l'élimination */
    return GRF_OterPoint( graphe, i );
    break;

  case GRF_COURBE_PLANE:
    OPTION_A_FAIRE( "Courbe plane" );
    return -10;

  case GRF_COURBE_ESPACE:
    OPTION_A_FAIRE( "Courbe dans l'espace" );
    return -10;

  case GRF_SURFACE:
    OPTION_A_FAIRE( "Surface Z=f(X,Y)" );
    return -10;

  case GRF_HISTOGRAMME:
    /* Impossible d'ôter une classe à un histogramme sans perdre la cohérence ! */
    return 50;

  case GRF_FAISCEAU:
    OPTION_A_FAIRE( "Faisceau de graphes" );
    return -10;

  case GRF_MULTI_SERIE:
    OPTION_A_FAIRE( "multi-série" );
    return -10;

  case GRF_ERREUR_R:
    for ( j = i ; j < (graphe->nombre_point - 1) ; j++ ) {
      graphe->y[ j ] = graphe->y[ j + 1 ];
      graphe->z.ddl[ j ] = graphe->z.ddl[ j + 1 ];
    }
    break;

  case GRF_ERREUR_C:
    {
      Complexe *z;

      for ( j = i, z = (Complexe *) graphe->y ; j < (graphe->nombre_point - 1) ; j++ ) {
	z[ j ] = z[ j + 1 ];
	graphe->z.ddl[ j ] = graphe->z.ddl[ j + 1 ];
      }
    }

  default:
    OPTION_A_FAIRE( "Genre de graphe pas encore géré" );
  }

  /* On retire le point correspondant des barres d'erreur */
  GRF_OterPoint( graphe->erreur_x, i );
  GRF_OterPoint( graphe->erreur_y, i );

  /* On a enlevé un point */
  graphe->nombre_point--;

  /* On le retire, si nécessaire, de la matrice de covariance */
  if ( graphe->covariance != NULL ) {
    OPTION_A_FAIRE( "Éliminer la colonne et la ligne correspondantes de la matrice de covariance !" );
  }

  /* ATTENTION ! On ne cherche pas à recalculer les bornes.
     En effet, appelé quand on ôte des valeurs manquantes, donc le graphe peut avoir des NaN s'il reste d'autres valeurs manquantes */

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Ôter les valeurs manquantes [codées par NaN] du graphe indiqué --------
     Renvoie le nombre de valeurs manquantes ôtées
   -------- */

unsigned long GRF_OterValeursManquantes(Graphe *graphe)
{
  unsigned long i, nombre_otees = 0;

  if ( NULL == graphe ) return 0;

  for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
    /* Contrôle sur les X */
    if ( graphe->x != NULL ) {
      if ( is_nan( graphe->x[ i ] ) ) {
	OPTION_A_FAIRE( "Valeur manquante en X" );
      }
    }

    /* Contrôle en Y */
    if ( ( GRF_FCT_REELLE == graphe->genre_donnees ) || ( GRF_SERIE == graphe->genre_donnees ) ) {
      if ( is_nan( graphe->y[ i ] ) ) {
	GRF_OterPoint( graphe, i );
	nombre_otees++;
	i--;			/* Car maintenant, on a en position i le point qui était avant en (i + 1) ! */
      }
    } else {
      OPTION_A_FAIRE( "Test des valeurs manquantes en Y, autre que réel !" );
    }
  }

  /* Terminé */
  return nombre_otees;
}

/* ~-~-~-~-~-~-~-~- Opérations entre deux graphes ~-~-~-~-~-~-~-~- */

/* -------- Somme de deux graphes « en place » ---------------\
 |  Renvoie 0   si tout s'est bien passé,                     |
 |          !=0 en cas d'erreur                               |
 |  Suppose que les X des deux graphes sont cohérents         |
 |                                                            |
 | Calcule reference + soustrait, résultat dans reference     |
 | covariances : Cov(Y_ref, Y_soustrait), si erreur dans les 2|
 \----------------------------------------------------------- */

int GRF_AjouterGraphe(Graphe *reference, Graphe *soustrait, boolean traite_erreurs, double *covariances)
{
  if ( (NULL == reference) || (NULL == soustrait) ) return -1;

  if ( reference->nombre_point != soustrait->nombre_point ) {
    (void) fprintf(stderr, "%s,%s l.%u : Nombre de points différents entre les deux graphes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }
  if ( FAUX == GRF_EstFonction( reference ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : La référence n'est pas une fonction réelle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  if ( VRAI == GRF_EstFonctionReelle( reference ) ) {
    unsigned long i;
    double *y_ref, *y_ote;

    if ( FAUX == GRF_EstFonctionReelle( soustrait ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Référence réelle (ou assimilable), 2nd graphe non réel\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -4; /* À un graphe réel, on ne peut ajouter en place qu'un graphe réel */
    }

    /* Addition en place */
    for (i = 0, y_ref = reference->y, y_ote = soustrait->y ; i < reference->nombre_point ; i++, y_ref++, y_ote++) {
      *y_ref += *y_ote;
    }

    /* Contrôle des bornes */
    GRF_bornes(reference, NE_PAS_FAIRE, DESORDONNE);

    /* Traitement des erreurs en Y */
    if ( FAUX == traite_erreurs ) return 0; /* Terminé */
    if ( soustrait->erreur_y != NULL ) {
      Graphe *err_ref;

      err_ref = reference->erreur_y;
      if ( err_ref != NULL ) { /* Erreurs sur les deux => Erreur totale : somme */
	for (i = 0, y_ref = err_ref->y, y_ote = soustrait->erreur_y->y ;
	     i < reference->nombre_point ; i++, y_ref++, y_ote++) {
	  double erreur;

	  /* Ce sont les écarts-types qui sont stockés, mais les variances qui sont additives.
	     Suppose qu'il y a indépendance... */
	  erreur = *y_ref * *y_ref + *y_ote * *y_ote;
	  if ( covariances != NULL ) { /* Ce n'est pas le cas : terme correctif */
	    erreur += 2.0 * *covariances; /* V(A-B) = V(A) + V(B) - 2 cov(A, B) */
	    covariances++;
	  }
	  *y_ref = sqrt(erreur);
	}

	GRF_bornes(reference->erreur_y, DESORDONNE, DESORDONNE);
      } else {
	/* Les erreurs viennent juste du 2e graphe : inchangées */
	reference->erreur_y = GRF_Copier( soustrait->erreur_y, NULL, VRAI );
      }
    }

    return 0;
  }

  OPTION_A_FAIRE( "Cas d'un graphe complexe");
  return -215;
}

/* -------- Soustraction de deux graphes « en place » --------\
 |  Renvoie 0 si tout s'est bien passé, != 0 si erreur        |
 |  Suppose que les X des deux graphes sont cohérents         |
 |                                                            |
 | Calcule reference - soustrait, résultat dans reference     |
 | covariances : Cov(Y_ref, Y_soustrait), si erreur dans les 2|
 \----------------------------------------------------------- */

int GRF_SoustraireGraphe(Graphe *reference, Graphe *soustrait, boolean traite_erreurs, double *covariances)
{
  if ( (NULL == reference) || (NULL == soustrait) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Appel avec un graphe null\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return -1;
  }

  if ( reference->nombre_point != soustrait->nombre_point ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Nombre de points différent\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return -2;
  }

  if ( GRF_FCT_REELLE == reference->genre_donnees ) {
    unsigned long i;
    double *y_ref, *y_ote;

    if ( ( GRF_FCT_REELLE != soustrait->genre_donnees ) && ( GRF_SERIE != soustrait->genre_donnees ) ) {
      (void) fprintf(stderr, "GRF_SoustraireGraphe : Impossible de soustraire en place un graphe complexe à un graphe réel\n");
      return -3; /* À un graphe réel, on ne peut soustraire en place qu'un graphe réel */
    }

    /* Soustraction en place */
    for (i = 0, y_ref = reference->y, y_ote = soustrait->y ; i < reference->nombre_point ; i++, y_ref++, y_ote++) {
      *y_ref -= *y_ote;
    }

    /* Contrôle des bornes */
    GRF_bornes(reference, NE_PAS_FAIRE, DESORDONNE);

    /* Traitement des erreurs en Y */
    if ( FAUX == traite_erreurs ) return 0; /* Terminé, tout s'est bien passé */
    if ( soustrait->erreur_y != NULL ) {
      Graphe *err_ref;

      err_ref = reference->erreur_y;
      if ( err_ref != NULL ) { /* Erreurs sur les deux => Erreur totale : somme */
	for (i = 0, y_ref = reference->erreur_y->y, y_ote = soustrait->erreur_y->y ;
	     i < reference->nombre_point ; i++, y_ref++, y_ote++) {
	  double erreur;

	  /* Ce sont les écarts-types qui sont stockés, mais les variances qui sont additives.
	     Suppose qu'il y a indépendance... */
	  erreur = *y_ref * *y_ref + *y_ote * *y_ote;
	  if ( covariances != NULL ) { /* Ce n'est pas le cas : terme correctif */
	    erreur -= 2.0 * *covariances; /* V(A-B) = V(A) + V(B) - 2 cov(A, B) */
	    covariances++;
	  }
	  *y_ref = sqrt(erreur);
	}
      } else {
	/* Les erreurs viennent juste du 2e graphe : inchangées */
	reference->erreur_y = GRF_Copier( soustrait->erreur_y, NULL, VRAI );
      }
    }

    return 0;
  }

  OPTION_A_FAIRE( "Cas complexe" );
  return -10;
}
