/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2002, Emmanuel CURIS
   Création : 12 février 2002

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Tentative de correction des effets de l'épaisseur de coupe - modèle simple
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"

#include "exafs.h"
#include "images.h"
#include "interface.h"
#include "lissage.h" /* Pour le lissage linéaire & les splines */
#include "science.h" /* Pour M_PI */
#include "tri.h"   /* Pour le tri */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean traite_ligne(Graphe *profil_observe, Graphe *profil_reel, double d_x, double d_z, unsigned int nbr_tranches);

static Image *IMG_ligne_graphe(Image *source, double epais_tranche, double epais_totale, double angle_faisceau);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ------ Modele simple : axe d'empilement des coupes perpendiculaire au plan de l'echantillon
    ==> effet en X seulement, on peut raisonner en lignes, plus simple...

    Methode directe par recurrence ("explicite")
        ==> Il y a de fortes instabilites numeriques qui perturbent... Voir une methode "implicite"

    -------- */

Image *IMG_epais_lignes(Image *source, double epais_tranche, double epais_totale, 
			double angle_faisceau, double inclinaison, double rotation)
{
  Image *resultat;
  double d_x, d_y, d_z;
  double *num_cases;
  unsigned int i, j, nbr_tranches;
  unsigned int larg_image;

  if (source == NULL) return NULL;
  if ( source->genre != IMG_GENRE_REELLE ) return NULL;

  /* Un essai : epaisseur des tranches optimale
       ==> il faut, en projetant, un pas en x égal au pas dans l'image d'origine
       Ce pas en x est pente * d_z = epais_tranche * sin(theta) = source->l_reel
       ==> d_z = source->l_reel / sin(theta);
     Paraît pas terrible... On annule donc
  */
#ifdef optimiser
  d_x = sin(angle_faisceau) + cos(rotation) * sin(inclinaison);
  epais_tranche = source->l_reel / d_x;
  nbr_tranches = 1 + (unsigned int) (epais_totale / epais_tranche);
  epais_totale = nbr_tranches * epais_tranche;
  printf("Epaisseur de tranche optimale : %f microns => %d tranches pour un total de %f microns\n",
	 epais_tranche, nbr_tranches, epais_totale);
#endif

  /* Combien de tranches ? */
  nbr_tranches = (unsigned int) rint(epais_totale / epais_tranche);
  if (nbr_tranches < 2.0) { /* Trop peu de tranches, calcul sans interet */
    return NULL;
  }

  num_cases = (double *) calloc(nbr_tranches, sizeof(double));
  if (num_cases == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Images/epaisseur.c", "IMG_epais_lignes", "Table des cases");
#ifdef _DEVELOPPE_
    fprintf(stderr, "IMG_epais_lignes : Impossible d'allouer la table des cases\n");
    fprintf(stderr, "   Taille demandee : %lu", (unsigned long) nbr_tranches * sizeof(double));
#endif
  }

  resultat = IMG_CreerImage( NULL, source, source->genre, source->type, source->larg, source->haut);
  if (resultat == NULL) {
    return NULL;
  }
  resultat->l_reel = source->l_reel;
  resultat->h_reel = source->h_reel;
/*  resultat->x_centre = source->x_centre;
  resultat->y_centre = source->y_centre; */

  /* Epaisseur d'une tranche */
  d_z = epais_totale / (double) nbr_tranches;

  /* Increments unitaires */
  d_x = /* sin(angle_faisceau) + cos(rotation) * sin(inclinaison);*/ tan(angle_faisceau);
  d_y = sin(rotation) * sin(inclinaison);
  if (d_y != 0.0) { /* Decalage en Y non nul => interferences entre les lignes, pas gere */
    IMG_EffacerImage( resultat, VRAI );
    return NULL;
  }
  /*  printf("Increments : %f en x, %f en y - epaisseur de coupe %f\n", d_x, d_y, d_z); */

  /* Init : la premiere colonne - valeurs les plus a gauches */
  for (j = 0 ; j < resultat->haut ; j++) {
    resultat->valeurs.R[j * resultat->larg] = source->valeurs.R[j * source->larg] /* / */ ;
  }

  /* Correction ligne a ligne, car on suppose l'axe d'empilement normal au plan de l'echantillon */
  for (i = 1 ; i < resultat->larg ; i++) {
    unsigned int k;
    double x_reel  __attribute__((__unused__))/*, y_reel*/;

    /* Position du point dans l'image superposition */
    x_reel = i * source->l_reel/* - source->x_centre*/;
/*    y_reel = -resultat->y_centre;*/ /* De toute facon, identique d'une ligne a l'autre ! */

    /*    printf("-- Colonne %d --\n", i);*/
    /* On regarde quelles cases on rencontre et on met le numero dans la table */
    for (k = 0 ; k < nbr_tranches ; k++) {
      /*      double x_tranche, y_tranche;*/ /* Position dans la nouvelle tranche */
      double x_case/*, y_case*/;

      /*      x_tranche = x_reel + d_x * (double) k * d_z; */
      /*      y_tranche = y_reel + d_y * (double) k * d_z; */
      x_case = ((double) i) - ( d_x * d_z * (double) k / resultat->l_reel );
      /* y_case = (int) floor( (y_tranche + resultat->y_centre) / (resultat->h_reel) ); */

      if (x_case > i) {
	printf("GROS PROBLEME : case pas encore calculee necessaire... [case %f demandee, %u calculee\n", x_case, i);
      }

      num_cases[ k ] = x_case;
      /*      printf(" tranche %u -> case %f\n", k, x_case); */
    }

    /* On corrige ensuite pour chaque ligne */
    for (j = 0 ; j < resultat->haut ; j++) {
      double somme = 0.0, total;
      unsigned int contrib_case = 0, k;

      total = source->valeurs.R[ j * source->larg + i ]; /* Resultat de la somme */
      /* On fait la somme des valeurs connues
	 et on calcule le coefficient de la case inconnue */
      for (k = 0 ; k < nbr_tranches ; k++) {
	if ( num_cases[ k ] < 0.0 ) {
	  somme += resultat->valeurs.R[ j * resultat->larg ]/*0.0*/; /* Faut-il prendre 0 ou la valeur dans la case 0 ? */
	} else if ( num_cases[ k ] < i - 1 ) { /* Case deja connue normalement
					      Interpolation lineaire entre les valeurs deja calculees */
	  double valeur;
	  int n_case = (int) num_cases[ k ];

	  /*	  printf("Case : %f -> entre %d et %d\n", num_cases[ k ], n_case, n_case + 1); */
	  valeur = LIS_lineaire((double) n_case, (double) n_case + 1.0,
				resultat->valeurs.R[ n_case + j * resultat->larg ],
				resultat->valeurs.R[ n_case + 1 + j * resultat->larg ], num_cases[ k ]);
	  somme += valeur;
	} else if ( num_cases[ k ] < i ) { /* Interpolation lineaire impossible => on met la valeur... */
	  somme += resultat->valeurs.R[ j * resultat->larg + (int) num_cases[k]];
	} else if ( num_cases[ k ] < i + 1.0 ) { /* Case cherchee */
	  contrib_case++;
	} 
      }

      /* Ici, on a donc : total = contrib_case * valeur_case_i + somme */
      /*      printf(" equation ligne %u : %f = %u * V[i] + %f\n", j, total, contrib_case, somme); */
      if (contrib_case > 0) {
	double nv_valeur = (total - somme) / (double) contrib_case;
	if (nv_valeur < 0.0 )
	  nv_valeur = 0.0; 
	resultat->valeurs.R[ i + j * resultat->larg ] = nv_valeur;
      } else {
	printf("PROBLEME : CASE CHERCHEE NE CONTRIBUE PAS A ELLE-MEME !\n");
      }
    }
  }

  /* Ici, normalement, on a l'image corrigee... */
  IMG_bornes(resultat);

  /* -------- METHODE DIRECTE MAIS AVEC UN PAS "OPTIMISE", POUR COMPARAISON -------- */

  /* Le pas a utiliser en x est epaisseur / (nbr_tranches * pente)
     on doit calculer pour x compris entre 0 et l_reel * larg_image_source
     ==> on en deduit le nombre de points pour l'image resultante
   */
  larg_image = (unsigned int) ( (source->l_reel * (double) source->larg) / d_x );
  resultat = IMG_CreerImage( NULL, resultat, source->genre, source->type, larg_image, source->haut);
  if (resultat == NULL) {
    return NULL;
  }
  resultat->l_reel = d_x;
  resultat->h_reel = source->h_reel;
/*  resultat->x_centre = source->x_centre;
  resultat->y_centre = source->y_centre;*/

  /* Correction ligne a ligne, car on suppose l'axe d'empilement normal au plan de l'echantillon
     Comme on est en pas optimal, on peut raisonner directement sur les cases... */

  /* On initialise le profil [valeurs en 0] */
  for (j = 0 ; j < resultat->haut ; j++) {
    double valeur;

    /* Modele : profil nul en dehors de l'image */
    valeur = 2.0 * source->valeurs.R[j * source->larg] / d_z;

    /* Modele : profil constant en-dehors de l'image */
    /*    valeur = source->valeurs.R[j * source->larg] / epais_totale; */

    resultat->valeurs.R[j * resultat->larg] = valeur;
  }

  for (i = 1 ; i < resultat->larg ; i++) {
    unsigned int k;
    double x_reel, y_reel;

    /* Position du point dans le profil lineaire d'origine */
    x_reel = i * resultat->l_reel/* - resultat->x_centre*/;

    /*    printf("-- Colonne %d --\n", i); */
    /* On corrige pour chaque ligne */
    for (j = 0 ; j < resultat->haut ; j++) {
      double somme = 0.0, total, valeur;
      int x;

      y_reel = j * resultat->h_reel/* - resultat->y_centre*/;

      total = (double) IMG_lit_valeur(source, x_reel, y_reel, IMG_INTERPOLE_BILINEAIRE, NULL); /* Resultat de l'integration */
      /* On fait la somme des valeurs connues
       Comme on est en pas optimal, on descend d'une case en x a chaque fois que l'on monte d'un cran en k */
      for (k = 1, x = i - 1 ; k < nbr_tranches ; k++, x--) {
	if ( x < 0 ) { /* Hors de l'image. Faut-il prendre 0 ou la valeur dans la case 0 ? */
	  somme += 0.0;
	  /*	  somme += resultat->valeurs[ j * resultat->larg ];*/
	} else {
	  somme += resultat->valeurs.R[ j * resultat->larg + x ];
	}
      }

      /* Ici, on a donc :
	                [ nbr_tranches               ]
        valeur[i] = 2 * [ ------------ total - somme ] - valeur[x - epaisseur / pente]
                        [  epaisseur                 ]
      */
      valeur = 2.0 * (total / d_z - somme);
      if ( x  < 0 ) {
	/* Valeur nulle... */
	/*	valeur -= resultat->valeurs[j * resultat->larg ];*/ /* Ou 0.0 ? */
      } else {
	valeur -= resultat->valeurs.R[j * resultat->larg + x ];
      }
      if (valeur < 0.0) valeur = 0.0;
      resultat->valeurs.R[j * resultat->larg + i] = (double) valeur;
    }
  }

  /* Ici, normalement, on a l'image corrigée... */
  IMG_bornes(resultat);
  return resultat;
}

/* -------- Correction des effets d'épaisseur, effet sur les X uniquement ----------------\
 |   Renvoie l'image corrigée, NULL en cas de problème                                    |
 |                                                                                        |
 |  Modèle : au point x_i de l'image, on mesure F(x_i) lié à la valeur dans le profil par |
 |                                                                                        |
 |          /e                                                                            |
 | F(x_i) = |   f(x_i - pente * z) dz                                                     |
 |          /0                                                                            |
 |                                                                                        |
 | On estime l'intégrale par la méthode des trapèzes avec un pas h = épaisseur de tranche |
 | ==> on a besoin des valeurs de f en x_i, x_i - p*h, x_i - 2*p*h,..., x_i - n*p*h = x_i - p*e
 |   on suppose les valeurs aux x précédents connues =>
 |       on en déduit donc f en x_i, qui pourra servir pour les autres par récurrence     |
 |
 | ==> démarrage de la récurrence : on postule f(x) = 0 pour x négatif ("hors-échantillon");
 |   Cela demande donc, pour chaque x_i, de remonter jusqu'à des x < 0 => il faut en fait plus de n points
 |
 |   Cela suppose aussi que l'on connaisse F pour d'autres valeurs que les x_i, si la grille n'est pas optimale =>
 |      on passe par un graphe => lissage par spline et travail en double
 \--------------------------------------------------------------------------------------- */

/* Traitement d'une seule ligne */

static boolean traite_ligne(Graphe *profil_observe, Graphe *profil_reel, double d_x, double d_z, unsigned int nbr_tranches)
{
  unsigned long i;
  double *b, *c, *d; /* Spline pour le profil observé */
  double F_negatif, F_x;
  register double *x, *y;

  if ( SPL_tt_prep(profil_observe, &b, &c, &d) <= 0.0 ) {
    return FAUX;
  }

  /* Valeurs pour x négatif : on doit avoir F(0) = f(0) puisque l'on traverse une seule case */
  F_x = SPL_tt_calc(profil_observe, profil_reel->x[0], b, c, d);
  F_negatif = (2.0 / d_z - 1.0) * F_x / (double) (2 * nbr_tranches - 1);
  *(profil_reel->y) = F_x;

/*  printf("X calculés :\n"); */
  for ( i = 1, x = profil_reel->x + 1, y = profil_reel->y + 1 ; i < profil_reel->nombre_point ; i++, x++, y++) {
    double F_x;
    double valeur, somme = 0.0;
    double x_reel;
    unsigned long k; /* Déplacements en z */

    F_x = SPL_tt_calc(profil_observe, *x, b, c, d); /* Valeur observée en ce point */
/*    printf("%.20g ", *x); */

    for (k = 1 ; k < nbr_tranches ; k++) {
/*       register double *x_obs; */
/*       unsigned long i; */

      x_reel = *x - d_x * (double) k; /* On remonte de k tranches */
/*      printf("%.20g ", x_reel); */

      if (x_reel < 0.0) /* X négatif, hors profil => supposé nul, donc il faut F_negatif */
	somme += F_negatif;
      else
	somme += GRF_InterpoleLineaire(profil_reel, x_reel, NULL);
    }

    valeur  = 2.0 * ( (F_x / d_z) - somme );
    x_reel = *x - d_x * (double) nbr_tranches;
/*    printf("%.20g\n", x_reel); */
    if (x_reel >= 0.0)
      valeur -= GRF_InterpoleLineaire(profil_reel, x_reel, NULL);
    else
      valeur -= F_negatif;
    
    *y = valeur;
  }

  free(b); free(c); free(d);
  return VRAI;
}

static Image *IMG_ligne_graphe(Image *source, double epais_tranche, double epais_totale, double angle_faisceau)
{
  Image *resultat = NULL;
  Graphe *profil_observe, *profil_reel;
  unsigned long nbr_points, i;
  unsigned int j;
  unsigned int nbr_tranches;
  double pente, d_x, x_reel;
  double *b, *c, *d;
  register double *x/*, *y*/;

  if (source == NULL) return NULL;

  if (cos(angle_faisceau) == 0.0) return source;  /* Faisceau perpendiculaire à l'échantillon -> pas d'effet */
/*  pente = tan(angle_faisceau); */
  pente = 1.0; /* 45°, pour commencer */
  if (pente == 0.0) return NULL; /* Faisceau tangent à l'échantillon => impossible */
  d_x = pente * epais_tranche;   /* Déplacement en X quand on change de tranche */

  /* Combien de tranches ? = nombre de pas pour la méthode des trapèzes */
  nbr_tranches = (unsigned int) rint(epais_totale / epais_tranche);
  if (nbr_tranches < 2) { /* Trop peu de tranches, calcul sans interet */
    return NULL;
  }

  profil_observe = GRF_nouveau( NULL, source->larg, GR_INTERNE, GR_NORMAL, "Profil horizontal observé" );
  if ( profil_observe == NULL ) return NULL;

/* On cherche a calculer f(x_i) pour chacun des x du profil initial.
   Pour cela, il faut remonter par pas de h = epaisseur_tranche * pente jusqu'aux valeurs négatives
   c'est-à-dire pour k allant de 0 à x_i/h inclus, soit k + 1 point

   Comme ce pas h n'est pas forcément égal au pas en x de l'image, la grille doit être faite pour chaque x de l'inage originale
*/
  for ( i = 0, x = profil_observe->x, nbr_points = 0 ; i < source->larg ; i++, x++) {
    *x = source->l_reel * (double) i; /* On en profite pour faire les X du profil initial */
    nbr_points += 1 + (unsigned int) (*x / d_x);
  }
  profil_observe->x_min = 0.0;
  profil_observe->x_max = (double) source->l_reel * (double) (source->larg - 1);

  profil_reel = GRF_nouveau( NULL, nbr_points, GR_INTERNE, GR_NORMAL, "Profil original" );
  if (profil_reel == NULL) {
    GRF_efface(profil_observe, VRAI, VRAI);
    return NULL;
  }

  /* On réalise l'image contenant le résultat */
  resultat = IMG_CreerImage( NULL, source, source->genre, source->type, source->larg, source->haut);
  if (resultat == NULL) {
    GRF_efface(profil_observe, VRAI, VRAI);
    GRF_efface(profil_reel, VRAI, VRAI);
    return NULL;
  }
  resultat->l_reel = source->l_reel;
  resultat->h_reel = source->h_reel;
/*  resultat->x_centre = source->x_centre;
  resultat->y_centre = source->y_centre;*/

  /* On calcule ces x_i, dans l'ordre */
  for (i = 0, x = profil_reel->x, nbr_points = 0 ; i < source->larg ; i++) {
    x_reel = (double) i * (double) source->l_reel;
    for (j = 0 ; x_reel - j * d_x > 0.0 ; j++, nbr_points++, x++) {
      if (nbr_points >= profil_reel->nombre_point) {
	printf("GROS PROBLEME sur le nombre de points\n");
	break;
      }
      *x = x_reel - d_x * (double) j;
    }
  }
  profil_reel->x_min = 0.0;
  profil_reel->x_max = profil_observe->x_max;

  /* On réordonne pour limiter le nombre de calculs, en éliminant les doublons éventuels */
  GRF_Trier( profil_reel, GRF_TRI_RAPIDE );
  printf("AVANT : %lu points -", profil_reel->nombre_point);
  /* Des points plus proches que 0.001 microns sont confondus */
  GRF_TraiterDoublons( profil_reel, GRF_DOUBLONS_SOUPLE, 0.001, FAUX, FAUX );
  printf("APRES : %lu points\n", profil_reel->nombre_point);
  GRF_bornes(profil_reel, ORDONNE, NE_PAS_FAIRE);

  /* On travaille ligne à ligne */
/*  for (j = 0 ; j < source->haut ; j++) */ {
    register double *y;

    /* Valeurs observées : profil initial */
    j = 6;
    for (i = 0, y = profil_observe->y ; i < profil_observe->nombre_point ; i++, y++) {
      *y = (double) source->valeurs.R[i + j * source->larg];
    }

    traite_ligne(profil_observe, profil_reel, d_x, epais_tranche, nbr_tranches);
  }

    /* Interpolation pas spline pour le profil final */
    for (j = 0 ; j < source->haut ; j++) {
      if ( SPL_tt_prep(profil_reel, &b, &c, &d) <= 0.0 ) {
      /* Calcul interrompu, impossible de continuer... */
/*      break; */
    }
   
    /* On recopie dans l'image résultat */
    for (i = 0 ; i < resultat->larg ; i++) {
      x_reel = (double) i * (double) resultat->l_reel;
      resultat->valeurs.R[i + j * resultat->larg] = SPL_tt_calc(profil_reel, x_reel, b, c, d);
    }
    free(b); free(c); free(d); /* Libération des tableaux */
  }

  GRF_bornes(profil_reel, DESORDONNE, DESORDONNE);
  GRF_bornes(profil_observe, DESORDONNE, DESORDONNE);
/*  GRF_efface(profil_reel, VRAI, VRAI); */
/*  GRF_efface(profil_observe, VRAI, VRAI); */
  printf("Calcul de l'image corrigée fini\n");
  IMG_bornes(resultat);
  return resultat;
}

/* -------- Interface -------- */

Image *IMG_corrige_epaisseur(Image *source)
{
  double epais_tranche, epais_totale;
  double angle_faisceau, inclinaison, rotation;
  int retour;
  Image *resultat;

  if (source == NULL) return NULL;

  retour = ARB_Bloquant( IMG_EFFETS_3D, "Correction des effets d'"e_AIGU"paisseur", principale, 0 );
  if (retour == IE3D_ANNULER)
    return NULL;

  /* Lecture des paramètres */
  epais_tranche = OBJ_LireNombre( IMG_EFFETS_3D, IE3D_EPAIS_COUPE, NULL );
  epais_totale  = OBJ_LireNombre( IMG_EFFETS_3D, IE3D_EPAIS_ECH, NULL );

  angle_faisceau = OBJ_LireAngle( IMG_EFFETS_3D, IE3D_ANGLE_FAIS, NULL );
  inclinaison = OBJ_LireAngle( IMG_EFFETS_3D, IE3D_ANGLE_NORM, NULL );
  rotation = OBJ_LireAngle( IMG_EFFETS_3D, IE3D_ANGLE_HOR, NULL );

  resultat = IMG_epais_lignes(source, epais_tranche, epais_totale, angle_faisceau, inclinaison, rotation);
  IMG_AllerImage( resultat );

  /* Méthode violente en calculs */
  resultat = IMG_ligne_graphe(source, epais_tranche, epais_totale, angle_faisceau);
  IMG_AllerImage( resultat );    

  return resultat;
}
