/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2002, Emmanuel CURIS
   Cration : 23 fvrier 2002

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Rotation des images (angle quelconque)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "systeme.h"

#include "images.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static inline int IMG_MiroirH_R(Image *image);
static inline int IMG_MiroirH_C(Image *image);
static inline int IMG_MiroirH_G(Image *image);

static inline int IMG_MiroirV_R(Image *image);
static inline int IMG_MiroirV_C(Image *image);
static inline int IMG_MiroirV_G(Image *image);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

Image *IMG_Miroir_Horizontal(Image *source, double position)
{
  Image *resultat;
  int retour;

  if (NULL == source) return NULL;

  /* On copie l'image,  la suite de la source */
  resultat = IMG_CopierImage( source, NULL, FAUX );
  if (NULL == resultat) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de copier l'image\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Miroir effectif : en place sur cette copie */
  retour = IMG_PseudoMiroir( resultat, IMG_MIROIR_HORIZONTAL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_PseudoMiroir() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    IMG_EffacerImage( resultat, FAUX );
    return NULL;
  }

  /* On met  jour les bornes / repre */
  resultat->x_min = 2.0 * position - (source->x_min + (source->larg - 1) * source->l_reel);

  /* S'il y a une coupe */
  if ( ( source->infos_affichage != NULL ) && ( source->infos_affichage->coupe != NULL ) ) {
    resultat->infos_affichage->xp_1 = 2.0 * position - source->infos_affichage->xp_1;
    resultat->infos_affichage->xp_2 = 2.0 * position - source->infos_affichage->xp_2;
  }

  return resultat;
}

Image *IMG_Miroir_Vertical(Image *source, double position)
{
  Image *resultat;
  int retour;

  if (NULL == source) return NULL;

  /* On copie l'image,  la suite de la source */
  resultat = IMG_CopierImage( source, NULL, FAUX );
  if (NULL == resultat) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de copier l'image\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Miroir effectif : en place sur cette copie */
  retour = IMG_PseudoMiroir( resultat, IMG_MIROIR_VERTICAL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_PseudoMiroir() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    IMG_EffacerImage( resultat, FAUX );
    return NULL;
  }

  /* On met  jour les bornes / repre */
  resultat->y_min = 2.0 * position - (source->y_min + (source->haut - 1) * source->h_reel);

  /* S'il y a une coupe */
  if ( ( source->infos_affichage != NULL ) && ( source->infos_affichage->coupe != NULL ) ) {
    resultat->infos_affichage->yp_1 = 2.0 * position - source->infos_affichage->yp_1;
    resultat->infos_affichage->yp_2 = 2.0 * position - source->infos_affichage->yp_2;
  }

  return resultat;
}

/* ~-~-~-~-~-~-~-~- Miroirs en place : pour avoir les bons axes au chargement ~-~-~-~-~-~- */

/* -------- Appel gnrique
   Miroir, mais sans modifier les coordonnes ni le repre
   -------- */

int IMG_PseudoMiroir(Image *image, int sens)
{
  int retour = 0;

  if (NULL == image) return -1;

  switch ( image->genre )
  {
  case IMG_GENRE_REELLE:
    if ( sens & IMG_MIROIR_HORIZONTAL )
      retour += IMG_MiroirH_R( image );
    if ( sens & IMG_MIROIR_VERTICAL )
      retour += IMG_MiroirV_R( image );
    break;

  case IMG_GENRE_COMPLEXE:
    if ( sens & IMG_MIROIR_HORIZONTAL )
      retour += IMG_MiroirH_C( image );
    if ( sens & IMG_MIROIR_VERTICAL )
      retour += IMG_MiroirV_C( image );
    break;

  case IMG_GENRE_GRAPHE:
    if ( sens & IMG_MIROIR_HORIZONTAL )
      retour += IMG_MiroirH_G( image );
    if ( sens & IMG_MIROIR_VERTICAL )
      retour += IMG_MiroirV_G( image );
    break;

  case IMG_GENRE_DIAGRAMME:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE / diagramme\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case IMG_GENRE_PHOTO:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE / photo\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Genre inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, image->genre);
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le miroir !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }


  /* Si l'image a des erreurs : elle subit le mme traitement */
  if ( image->erreurs != NULL ) {
    retour = IMG_PseudoMiroir( image->erreurs, sens );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de faire le miroir pour les erreurs\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }

    return +1; /* Erreur mineure */
  }

  /* Tout s'est bien pass */
  return 0;

  if ( sens & IMG_MIROIR_VERTICAL ) {
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Miroir H, image  valeurs relles -------- */

static inline int IMG_MiroirH_R(Image *image)
{
  unsigned long x, y;
  unsigned int demi_hauteur;
  unsigned long pt_1, pt_2;
  double valeur;

  demi_hauteur = image->haut >> 1;
  for (x = 0 ; x < image->larg ; x++) {
    for (y = 0 ; y < demi_hauteur ; y++) {
      pt_1 = y * image->larg + x;
      pt_2 = ( (image->haut - 1) - y ) * image->larg + x;

      valeur = image->valeurs.R[ pt_1 ];
      image->valeurs.R[ pt_1 ] = image->valeurs.R[ pt_2 ];
      image->valeurs.R[ pt_2 ] = valeur;
    }
  }

  /* Tout s'est bien pass */
  return 0;
}

static inline int IMG_MiroirH_C(Image *image)
{
  unsigned long x, y;
  unsigned int demi_hauteur;
  unsigned long pt_1, pt_2;
  Complexe valeur;

  demi_hauteur = image->haut >> 1;
  for (x = 0 ; x < image->larg ; x++) {
    for (y = 0 ; y < demi_hauteur ; y++) {
      pt_1 = y * image->larg + x;
      pt_2 = ( (image->haut - 1) - y ) * image->larg + x;

      valeur = image->valeurs.C[ pt_1 ];
      image->valeurs.C[ pt_1 ] = image->valeurs.C[ pt_2 ];
      image->valeurs.C[ pt_2 ] = valeur;
    }
  }

  /* Tout s'est bien pass */
  return 0;
}

static inline int IMG_MiroirH_G(Image *image)
{
  unsigned long x, y;
  unsigned int demi_hauteur;
  unsigned long pt_1, pt_2;
  struct graphe *valeur;

  demi_hauteur = image->haut >> 1;
  for (x = 0 ; x < image->larg ; x++) {
    for (y = 0 ; y < demi_hauteur ; y++) {
      pt_1 = y * image->larg + x;
      pt_2 = ( (image->haut - 1) - y ) * image->larg + x;

      valeur = image->valeurs.G[ pt_1 ];
      image->valeurs.G[ pt_1 ] = image->valeurs.G[ pt_2 ];
      image->valeurs.G[ pt_2 ] = valeur;
    }
  }

  /* Tout s'est bien pass */
  return 0;
}

static inline int IMG_MiroirV_R(Image *image)
{
  unsigned long x, y, offset = 0;
  unsigned int demi_largeur;
  double valeur;

  demi_largeur = image->larg >> 1;
  for (y = 0 ; y < image->haut ; y++, offset += image->larg) {
    for (x = 0 ; x < demi_largeur ; x++) {
      valeur = image->valeurs.R[ offset + x ];
      image->valeurs.R[ offset + x ] = image->valeurs.R[ offset + (image->larg - 1) - x ];
      image->valeurs.R[ offset + (image->larg - 1) - x ] = valeur;
    }
  }

  /* Tout s'est bien pass */
  return 0;
}

static inline int IMG_MiroirV_C(Image *image)
{
  unsigned long x, y, offset = 0;
  unsigned int demi_largeur;
  Complexe valeur;

  demi_largeur = image->larg >> 1;
  for (y = 0 ; y < image->haut ; y++, offset += image->larg) {
    for (x = 0 ; x < demi_largeur ; x++) {
      valeur = image->valeurs.C[ offset + x ];
      image->valeurs.C[ offset + x ] = image->valeurs.C[ offset + (image->larg - 1) - x ];
      image->valeurs.C[ offset + (image->larg - 1) - x ] = valeur;
    }
  }

  /* Tout s'est bien pass */
  return 0;
}

static inline int IMG_MiroirV_G(Image *image)
{
  unsigned long x, y, offset = 0;
  unsigned int demi_largeur;
  struct graphe *valeur;

  demi_largeur = image->larg >> 1;
  for (y = 0 ; y < image->haut ; y++, offset += image->larg) {
    for (x = 0 ; x < demi_largeur ; x++) {
      valeur = image->valeurs.G[ offset + x ];
      image->valeurs.G[ offset + x ] = image->valeurs.G[ offset + (image->larg - 1) - x ];
      image->valeurs.G[ offset + (image->larg - 1) - x ] = valeur;
    }
  }

  /* Tout s'est bien pass */
  return 0;
}
