/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2002, Emmanuel CURIS
   Cration : 23 fvrier 2002

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Rotation des images (angle quelconque)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "txt_utils.h"
#include "systeme.h"

#include "exafs.h"

#include "graphes.h"
#include "images.h"
#include "interface.h"
#include "science.h" /* Pour M_PI */

#include "rotation.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GER_arbre_rotation(unsigned int objet, int x, int y);
static void prepare_arbre_rotation(void);

static Image *IMG_rotation_90 (Image *source);
static Image *IMG_rotation_180(Image *source);
static Image *IMG_rotation_270(Image *source);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Rotation de 90 ---------- */

static Image *IMG_rotation_90(Image *source)
{
  Image *resultat;
  unsigned long i, j;
  unsigned long case_source;

  if (NULL == source) return NULL;

  resultat = IMG_CreerImage( NULL, source, source->genre, source->type, source->haut, source->larg );
  if (NULL == resultat) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_CreerImage()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  resultat->l_reel = source->h_reel;
  resultat->h_reel = source->l_reel;
  resultat->x_min = -(source->y_min + (source->haut - 1) * source->h_reel);
  resultat->y_min = source->x_min;

  switch ( source->genre )
  {
  case IMG_GENRE_REELLE:
    {
      double *valeurs;

      for (j = 0, valeurs = resultat->valeurs.R ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source, j, (source->haut - 1 - i), &case_source );
	  *valeurs = source->valeurs.R[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_COMPLEXE:
    {
      Complexe *valeurs;

      for (j = 0, valeurs = resultat->valeurs.C ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source, j, (source->haut - 1 - i), &case_source );
	  *valeurs = source->valeurs.C[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_GRAPHE:
    {
      struct graphe **valeurs;

      for (j = 0, valeurs = resultat->valeurs.G ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source, j, (source->haut - 1 - i), &case_source );
	  *valeurs = source->valeurs.G[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_DIAGRAMME:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- diagramme\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case IMG_GENRE_PHOTO:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- photo\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : genre inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, source->genre);
  }

  /* Les bornes ne changent pas ! */
  resultat->val_min = source->val_min;
  resultat->val_max = source->val_max;

  return resultat;
}

/* -------- Rotation de 180 ---------- */

static Image *IMG_rotation_180(Image *source)
{
  Image *resultat;
  unsigned long i, j;
  unsigned long case_source;

  if (NULL == source) return NULL;

  resultat = IMG_CreerImage( NULL, source, source->genre, source->type, source->larg, source->haut );
  if (NULL == resultat) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_CreerImage()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  resultat->l_reel = source->l_reel;
  resultat->h_reel = source->h_reel;
  resultat->x_min = -(source->x_min + (source->larg - 1) * source->l_reel);
  resultat->y_min = -(source->y_min + (source->haut - 1) * source->h_reel);

  switch ( source->genre )
  {
  case IMG_GENRE_REELLE:
    {
      double *valeurs;

      for (j = 0, valeurs = resultat->valeurs.R ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source,  (source->larg - 1 - i), (source->haut - 1 - j), &case_source );
	  *valeurs = source->valeurs.R[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_COMPLEXE:
    {
      Complexe *valeurs;

      for (j = 0, valeurs = resultat->valeurs.C ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source,  (source->larg - 1 - i), (source->haut - 1 - j), &case_source );
	  *valeurs = source->valeurs.C[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_GRAPHE:
    {
      struct graphe **valeurs;

      for (j = 0, valeurs = resultat->valeurs.G ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source,  (source->larg - 1 - i), (source->haut - 1 - j), &case_source );
	  *valeurs = source->valeurs.G[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_DIAGRAMME:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- diagramme\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case IMG_GENRE_PHOTO:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- photo\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : genre inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, source->genre);
  }

  resultat->val_min = source->val_min;
  resultat->val_max = source->val_max;

  return resultat;
}

/* -------- Rotation de 270 = -90 ---------- */

static Image *IMG_rotation_270(Image *source)
{
  Image *resultat;
  unsigned long i, j;
  unsigned long case_source;

  if (NULL == source) return NULL;

  resultat = IMG_CreerImage( NULL, source, source->genre, source->type, source->haut, source->larg );
  if (NULL == resultat) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_CreerImage()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  resultat->l_reel = source->h_reel;
  resultat->h_reel = source->l_reel;
  resultat->x_min = source->y_min;
  resultat->y_min = -(source->x_min + (source->larg - 1) * source->l_reel);


  switch ( source->genre )
  {
  case IMG_GENRE_REELLE:
    {
      double *valeurs;

      for (j = 0, valeurs = resultat->valeurs.R ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source,  (source->larg - 1 - j), i, &case_source );
	  *valeurs = source->valeurs.R[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_COMPLEXE:
    {
      Complexe *valeurs;

      for (j = 0, valeurs = resultat->valeurs.C ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source,  (source->larg - 1 - j), i, &case_source );
	  *valeurs = source->valeurs.C[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_GRAPHE:
    {
      struct graphe **valeurs;

      for (j = 0, valeurs = resultat->valeurs.G ; j < resultat->haut ; j++) {
	for (i = 0 ; i < resultat->larg ; i++, valeurs++) {
	  IMG_IJ_v_N( source,  (source->larg - 1 - j), i, &case_source );
	  *valeurs = source->valeurs.G[ case_source ];
	}
      }
    }
    break;

  case IMG_GENRE_DIAGRAMME:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- diagramme\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case IMG_GENRE_PHOTO:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- photo\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : genre inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, source->genre);
  }

  resultat->val_min = source->val_min;
  resultat->val_max = source->val_max;

  return resultat;
}

/* -------- Rotation d'un angle quelconque --------------------\
 | Renvoie l'image rsultat                                    |
 | La rotation se fait autour du centre du repre de l'image   |
 |        (dfini dans image->x_centre, image->y_centre)       |
 |                                                             |
 | source = l'image  tourner                                  |
 | angle  = angle de rotation, en radians                      |
 | pas_x, pas_y : taille des cases dans l'image tourne        |
 \------------------------------------------------------------ */

Image *IMG_rotation_quelconque(Image *source, double angle, double pas_x, double pas_y)
{
  Image *resultat = NULL;
  double c_angle, s_angle; /* sinus et cosinus */
  double lg_image, ht_image;
  double x_min, y_min, x_max, y_max; /* De l'image originale */
  double x1, y1, x2, y2; /* Du rectangle entourant l'image tourne */
  double *valeurs;
  unsigned int larg, haut, i, j;

  if (source == NULL) return NULL;
  if ( source->genre != IMG_GENRE_REELLE ) {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Rotation quelconque d'une image de genre %d\n",
		   __FILE__, __FUNCTION__, __LINE__, source->genre);
    return NULL;
  }

  if ( (pas_x <= 0.0) || (pas_y <= 0.0) ) { /* Pas absurde ! */
    (void) fprintf(stderr, "IMG_rotation_quelconque : Pas absurde, dx = %g, dy = %g\n", pas_x, pas_y);
    return NULL;
  }

  ENTREE_FONCTION("IMG_rotation_quelconque", "source=%p, angle=%g, pas_x=%g, pas_y=%g", source, angle, pas_x, pas_y);

  c_angle = cos(angle); s_angle = sin(angle);
  x_min = source->x_min; y_min = source->y_min;
  x_max = x_min + source->l_reel * (double) (source->larg - 1);
  y_max = y_min + source->h_reel * (double) (source->haut - 1);

  /* ---- Recherche des dimensions & repre de l'image tourne ---- */
  {
    double x_hg, x_hd, x_bg, x_bd; /* Coordonnes des sommets aprs rotation */
    double y_hg, y_hd, y_bg, y_bd;

    /* Coin suprieur gauche */
    x_hg = c_angle * x_min - s_angle * y_max;
    y_hg = s_angle * x_min + c_angle * y_max;
    /* Coin suprieur droit */
    x_hd = c_angle * x_max - s_angle * y_max;
    y_hd = s_angle * x_max + c_angle * y_max;
    /* Coin infrieur droit */
    x_bd = c_angle * x_max - s_angle * y_min;
    y_bd = s_angle * x_max + c_angle * y_min;
    /* Coin infrieur gauche */
    x_bg = c_angle * x_min - s_angle * y_min;
    y_bg = s_angle * x_min + c_angle * y_min;
    /* Rectangle englobant */
    x1 = x_hg; if (x1 > x_hd) x1 = x_hd; if (x1 > x_bd) x1 = x_bd; if (x1 > x_bg) x1 = x_bg;
    y1 = y_hg; if (y1 > y_hd) y1 = y_hd; if (y1 > y_bd) y1 = y_bd; if (y1 > y_bg) y1 = y_bg;
    x2 = x_hg; if (x2 < x_hd) x2 = x_hd; if (x2 < x_bd) x2 = x_bd; if (x2 < x_bg) x2 = x_bg;
    y2 = y_hg; if (y2 < y_hd) y2 = y_hd; if (y2 < y_bd) y2 = y_bd; if (y2 < y_bg) y2 = y_bg;

#ifdef _DEVELOPPE_
    printf("Coins aprs rotation : (%f, %f) - (%f, %f) - (%f, %f) - (%f, %f)\n",
      x_hg, y_hg, x_hd, y_hd, x_bg, y_bg, x_bd, y_bd);
    printf("Rectangle englobant : (%f, %f) - (%f, %f)\n",
      x1, y1, x2, y2);
#endif
  }

  lg_image = x2 - x1;
  ht_image = y2 - y1;
  larg = 1 + (unsigned int) lg_image / pas_x; /* Nombre de cases dans l'image */
  haut = 1 + (unsigned int) ht_image / pas_y;

  resultat = IMG_CreerImage( NULL, source, source->genre, source->type, larg, haut );
  if (resultat == NULL) {
    (void) fprintf(stderr, "IMG_rotation_quelconque : ERREUR dans creer_image()\n");
    SORTIE_FONCTION("IMG_rotation_quelconque", "%p [ERREUR]", NULL);
    return NULL;
  }
  resultat->l_reel = pas_x; /*(resultat->larg - 1) / lg_image; */
  resultat->h_reel = pas_y; /*(resultat->haut - 1) / ht_image; */

  resultat->x_min = x1;
  resultat->y_min = y1;

  /* Rotation effective
     Travail par ligne -> dans l'ordre de stockage des valeurs ! */
  for ( i = 0, valeurs = resultat->valeurs.R ; i < resultat->larg ; i++, valeurs++ ) {
    double x;

    x = resultat->x_min + resultat->l_reel * (double) i; 
    for ( j = 0 ; j < resultat->haut ; j++, valeurs++ ) {
      double y, x_av, y_av;
      double valeur;

      y = resultat->y_min + resultat->h_reel * (double) j;
      x_av =  c_angle * x + s_angle * y;
      y_av = -s_angle * x + c_angle * y;

      valeur = IMG_lit_valeur(source, x_av, y_av, IMG_INTERPOLE_BILINEAIRE, NULL);
      *valeurs = valeur;
    }
  }

  IMG_bornes(resultat);
  SORTIE_FONCTION("IMG_rotation_quelconque", "%p", resultat);
  return resultat;
}

/* -~-~-~-~ Interface graphique -~-~-~-~ */

Image *IMG_tourner(Image *source)
{
  Image *resultat = NULL;
  char *titre;
  int reponse;
  double angle;
  double pas_x, pas_y;

  if (source == NULL) return NULL;

  if ( FAUX == arbres[ROTATION_IMAGE].drapeaux.prepare)
    prepare_arbre_rotation();

  ARB_MettreComplement( ROTATION_IMAGE, source );
  reponse = ARB_Bloquant( ROTATION_IMAGE, titre_arbre_rotation, principale, 0 );
  ARB_MettreComplement( ROTATION_IMAGE, NULL );
  if ( RI_ANNULER == reponse )
    return NULL;

  /* Les cas spciaux */
  if ( OBJ_choisit(ROTATION_IMAGE, RI_90, QUESTION) == VRAI ) {
    angle = M_PI_2;
    resultat = IMG_rotation_90(source);
    titre = TXT_genere_chaine(texte_masque_nom_90, source->nom);
  } else if ( OBJ_choisit(ROTATION_IMAGE, RI_180, QUESTION) == VRAI ) {
    angle = M_PI;
    resultat = IMG_rotation_180(source);
    titre = TXT_genere_chaine(texte_masque_nom_180, source->nom);
  } else if ( OBJ_choisit(ROTATION_IMAGE, RI_MOINS_90, QUESTION) == VRAI ) {
    angle = -M_PI_2;
    resultat = IMG_rotation_270(source);
    titre = TXT_genere_chaine(texte_masque_nom_270, source->nom);
  } else { /* Cas gnrique */
    angle = OBJ_LireAngle( ROTATION_IMAGE, RI_ANGLE, NULL );
    /* On se place dans 0-2*pi (on est dj dans [-2pi, 2pi]) */
    while (angle < 0.0)
      angle += M_PI;

    if ( angle < 0.01 ) return source; /* Tourn de moins de 1/100 de radian = 0,6 degrs environ : ignor */

    titre = TXT_genere_chaine(texte_masque_nom_autre, source->nom, angle * 180.0 / M_PI );

    pas_x = OBJ_LireNombre( ROTATION_IMAGE, RI_PAS_X, NULL );
    pas_y = OBJ_LireNombre( ROTATION_IMAGE, RI_PAS_Y, NULL );

    resultat = IMG_rotation_quelconque(source, angle, pas_x, pas_y);
  }

  IMG_Titre(resultat, titre);
  free(titre);

  /* Veut-on garder le profil de coupe pour l'image tourne ? */
  if ( ( source->infos_affichage != NULL ) && ( source->infos_affichage->coupe != NULL ) ) {
    if ( OBJ_choisit(ROTATION_IMAGE, RI_GARDER_PROFIL, QUESTION) == VRAI ) {
      resultat->infos_affichage->coupe = GRF_Copier( source->infos_affichage->coupe, NULL, VRAI ) ;

      resultat->infos_affichage->xp_1 = source->infos_affichage->xp_1 * cos(angle) - source->infos_affichage->yp_1 * sin(angle);
      resultat->infos_affichage->yp_1 = source->infos_affichage->xp_1 * sin(angle) + source->infos_affichage->yp_1 * cos(angle);
      resultat->infos_affichage->xp_2 = source->infos_affichage->xp_2 * cos(angle) - source->infos_affichage->yp_2 * sin(angle);
      resultat->infos_affichage->yp_2 = source->infos_affichage->xp_2 * sin(angle) + source->infos_affichage->yp_2 * cos(angle);
    }
  }

  return resultat;
}

/* -------- Gestion de l'arbre -------- */

static int GER_arbre_rotation(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  Image *source;

  source = (Image *) ARB_LireComplement( ROTATION_IMAGE );

  switch (objet)
  {
  case RI_PROPOSER_PAS:
    OBJ_choisit(ROTATION_IMAGE, RI_PROPOSER_PAS, FORCE_OUI);
    {
      double i, angle, pas;

      angle = OBJ_LireAngle( ROTATION_IMAGE, RI_ANGLE, NULL );

      pas = source->l_reel * fabs(cos(angle));
      i = source->h_reel * fabs(sin(angle));
      if (i < pas) pas = i;
      OBJ_MettreNombre( ROTATION_IMAGE, RI_PAS_X, pas, 0.0 );
    
      pas = source->h_reel * fabs(cos(angle));
      i   = source->l_reel * fabs(sin(angle));
      if (i < pas) pas = i;
      OBJ_MettreNombre( ROTATION_IMAGE, RI_PAS_Y, pas, 0.0 );
    }
    OBJ_choisit(ROTATION_IMAGE, RI_PROPOSER_PAS, FORCE_NON);
    return VRAI;

  case RI_MOINS_90:
  case RI_180:
  case RI_90:
    OBJ_choisit(ROTATION_IMAGE, objet, FORCE_OUI);
    OBJ_active(ROTATION_IMAGE, RI_ANGLE, FORCE_NON);
    OBJ_active(ROTATION_IMAGE, RI_PROPOSER_PAS, FORCE_NON);
    OBJ_active(ROTATION_IMAGE, RI_PAS_X, FORCE_NON);
    OBJ_active(ROTATION_IMAGE, RI_PAS_Y, FORCE_NON);
    return VRAI;

  case RI_AUTRE_ANGLE:
    OBJ_choisit(ROTATION_IMAGE, objet, FORCE_OUI);
    OBJ_active(ROTATION_IMAGE, RI_ANGLE, FORCE_OUI);
    OBJ_active(ROTATION_IMAGE, RI_PROPOSER_PAS, FORCE_OUI);
    OBJ_active(ROTATION_IMAGE, RI_PAS_X, FORCE_OUI);
    OBJ_active(ROTATION_IMAGE, RI_PAS_Y, FORCE_OUI);
    return VRAI;
  }

  return FAUX;
}

/* -------- Prparation de l'arbre -------- */

static void prepare_arbre_rotation(void)
{

#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte(ROTATION_IMAGE, RI_TXT_ANGLE, txt_arbre_ri_angle);
  OBJ_TraduireTexte(ROTATION_IMAGE, RI_AUTRE_ANGLE, txt_arbre_ri_autre);
  OBJ_TraduireTexte(ROTATION_IMAGE, RI_TXT_PAS_X, txt_arbre_ri_pas_x);
  OBJ_TraduireTexte(ROTATION_IMAGE, RI_TXT_PAS_Y, txt_arbre_ri_pas_y);
  OBJ_TraduireTexte(ROTATION_IMAGE, RI_PROPOSER_PAS, txt_arbre_ri_propose);
  OBJ_TraduireTexte(ROTATION_IMAGE, RI_TXT_GARDER, txt_arbre_ri_garder);
  OBJ_TraduireTexte(ROTATION_IMAGE, RI_OK, txt_arbre_ri_ok);
#endif

  arbres[ROTATION_IMAGE].fonction.arbre = GER_arbre_rotation;
  arbres[ROTATION_IMAGE].drapeaux.prepare = VRAI;
}
