/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2001, Emmanuel CURIS
   Cration : 11 avril 2001

   Imports/export, des fichiers exprimentaux entre autres
   Import d'un fichier enregistre sur la ligne BM-01 B de l'ESRF
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "txt_utils.h"
#include "systeme.h"

#include "exafs.h"

#include "esrf.h"
#include "fichiers.h"
#include "serie.h"

#include "esrf_bm01.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static unsigned long bm01_compter_lignes(FILE *fichier, int commentaire, boolean successifs);

/* ~~~~~~~~ Realisation pratique ~~~~~ */

/* -~-~-~-~-~-~-~-~ Lecture d'un enregistrement EXAFS -~-~-~-~-~-~-~-~ */

int bm01b_lit_exafs(FILE *fichier, int numero_scan, const char *nom_fichier)
{
  unsigned int nbr_colonnes, colonne_i0, passage = 0;
  int n_lu, reponse;
  char *cln_trans, *cln_fluo;
  Graphe **graphes;
  Intervalle_I *transmission, *fluorescence;
  unsigned long nbr_trans, nbr_fluo;

  if (NULL == fichier) return -1;

  /* On saute les lignes de commentaires */
  ignore_ligne(fichier);
  do {
    int c;

    c = fgetc(fichier);
    if (c == '#') {/* Commentaire */
      c = fgetc(fichier);
      if (c == 'N') {/* Suivi du nombre de colonnes */
	fscanf(fichier, "%u", &nbr_colonnes);
      } else if (c == 'L') {
	fprintf(stderr, "bm01b_lit_exafs : il faudrait lire les noms de colonnes\n");
      }
      ignore_ligne(fichier);
    } else {
      ungetc(c, fichier);
      break;
    }
  } while ( ! feof(fichier) );

  graphes = (Graphe **) calloc( nbr_colonnes, sizeof(Graphe *));
  if ( NULL == graphes ) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Import/esrf_bm01.c", "", "graphes");
    (void) fprintf(stderr, "bm01b_lit_exafs : impossible de creer la table des %u graphes\n", nbr_colonnes);
    return -3;
  }

  /* On demande les colonnes  conserver */
  reponse = ARB_Bloquant( IMP_ESRF_EXAFS, titre_arbre_colonnes, principale, 0 );
  if (IEE_ANNULER == reponse) {
#ifdef _DEVELOPPE_
    (void) printf("bm01b_lit_exafs(fichier=%p [%s],scan=%d) : calcul annul\n", (void *) fichier, nom_fichier, numero_scan);
#endif
    return 1; /* Calcul annul */
  }

  colonne_i0 = OBJ_LireEntier_U( IMP_ESRF_EXAFS, IEE_COL_I0 );
  cln_trans = lit_chaine(IMP_ESRF_EXAFS, IEE_COL_TRANS, 0, NULL, TXT_REMPLACE);
  cln_fluo  = lit_chaine(IMP_ESRF_EXAFS, IEE_COL_FLUO , 0, NULL, TXT_REMPLACE);

  transmission = TXT_genere_intervalles( cln_trans, &nbr_trans );
  if ( NULL == transmission ) {
    (void) fprintf(stderr, "bm01b_lit_exafs : impossible de crer les intervalles en transmission\n");
    return -2;
  }

  fluorescence = TXT_genere_intervalles( cln_fluo, &nbr_fluo );
  if ( NULL == fluorescence ) {
    (void) fprintf(stderr, "bm01b_lit_exafs : impossible de crer les intervalles en fluorescence\n");
    free(transmission);
    return -2;
  }

  /* On compte le nombre de lignes  conserver */
  do {
    unsigned long i, nbr_points;
    long pos_debut;
    double x;
    int c;

    passage++;
    nbr_points = bm01_compter_lignes(fichier, '#', VRAI);
    if (nbr_points < 2) {
      (void) fprintf(stderr, "bm01b_lit_exafs : trop peu de valeurs ! On ignore ce scan [passage %u]", passage);
      break;
    }
    c = '0'; /* On ne peut pas commencer par un commentaire, car ignores... */

    /* On prpare les graphes */
    for (i = 1 ; i <= nbr_colonnes ; i++) {
      if ( ( TXT_est_dans_intervalle(i, transmission, nbr_trans) ) ||
	   ( TXT_est_dans_intervalle(i, fluorescence, nbr_fluo) ) ) {
	char *titre;

	titre = TXT_genere_chaine("%F - scan %d, colonne %l, passage %u", nom_fichier, numero_scan, i + 1, passage);
	graphes[ i - 1 ] = GRF_nouveau(NULL,nbr_points, GR_BRUT, GR_NORMAL, titre );
	if ( NULL == graphes[ i-1 ] ) {
	  (void) fprintf(stderr, "bm01b_lit_exafs : impossible de creer le graphe / colonne %lu\n", i);
	}
	free(titre);
      }
    }

    /* On lit les lignes */
    for (i = 0 ; i < nbr_points ; i++) {
      double energie, i0;
      unsigned int j;
      boolean message = FAUX;

      while ( '#' == c ) { /* Commentaire a ignorer */
	if (FAUX == message) {
	  (void) BI_ERR_alerte(NULL, &INF_SpectreInterrompu, VRAI, passage);
	  message = VRAI; /* Pour ne mettre le message qu'une fois par passage */
	}
	c = ignore_ligne(fichier);
#ifdef _DEVELOPPE_
	printf("On saute une ligne de commentaire... c = %c\n", c);
#endif
	if (EOF == c) {
	  (void) fprintf(stderr, "bm01b_lit_exafs : fin de fichier apres un commentaire, "
			 "mais il devrait rester des lignes [i = %lu, nbr_points = %lu]\n", i, nbr_points);
	  break;
	}
      }

      /* Lecture de l'nergie */
      fscanf(fichier, "%lf", &energie);
/*       printf("Energie : %f eV\n", energie); */
      for (j = 0 ; j < nbr_colonnes ; j++) {
	if ( graphes[j] != NULL )
	  graphes[j]->x[i] = energie;
      }

      /* Lecture des intensits */
      for (j = 2 ; j <= nbr_colonnes ; j++) {
	if (j == colonne_i0) {
	  fscanf(fichier, "%lf", &i0);
	  if (0.0 >= i0) {
	    (void) fprintf(stderr, "bm01b_lit_exafs : I0 <= 0, probleme faisceau, a faire\n");
	  }
	} else if ( TXT_est_dans_intervalle(j, transmission, nbr_trans) ) {
	  double i_t;

	  fscanf(fichier, "%lf", &i_t);
	  if ( graphes[ j-1 ] != NULL ) {
	    graphes[j-1]->y[i] = log( i0 / i_t );
	  }
	} else if ( TXT_est_dans_intervalle(j, fluorescence, nbr_fluo) ) {
	  double i_f;

	  fscanf(fichier, "%lf", &i_f);
	  if ( graphes[ j-1 ] != NULL ) {
	    graphes[j-1]->y[i] = i_f / i0;
	  }
	} else { /* On ignore la colonne */
	  fscanf(fichier, "%*s");
	}
      }

      c = ignore_ligne(fichier); /* Normalement, juste les fins de ligne... */
    }

    /* Couleurs, styles,... des graphes */    
    for (i = 0 ; i < nbr_colonnes ; i++) {
      if ( NULL != graphes[ i ] ) {
	GRF_bornes(graphes[i], ORDONNE, DESORDONNE);
	GRF_couleur( graphes[i], (i % 15) + 1, (i / 15 + 1) % 16, 0 );
      }

      graphes[i] = NULL; /* Pour ne pas etre gene au tour suivant */
    }

    /* On verifie si la ligne suivante commence par une valeur ou non : spectres successifs */
    pos_debut = ftell(fichier);
    if ( '#' == c ) { /* Ligne de commentaire pour la suite */
      boolean ignorable = VRAI;

      while ('#' == c) { /* Si coupure de faisceau pile entre deux iterations, ne pas arreter... */
	(void) fgetc(fichier);
	c = fgetc(fichier);
	if ( 'M' == c ) { /* Ligne de commentaire ignorable */
	  c = ignore_ligne(fichier);
	} else {
	  ignorable = FAUX;
	  break;
	}
      }

      if (FAUX == ignorable) { /* Commentaire de changement de scan */
	fseek(fichier, pos_debut, SEEK_SET); /* On se remet au debut pour l'analyser correctement */
	break; /* On sort de la boucle */
      } else {
	printf("bm01b_lit_exafs : coupure de faisceau entre deux spectres\n");
	pos_debut = ftell(fichier); /* On se place au debut des lignes apres ces commentaires */
      }
    }
    n_lu = fscanf(fichier, "%lf", &x);
    fseek(fichier, pos_debut, SEEK_SET);

  } while (n_lu == 1);

  free(graphes);
  free(transmission);
  free(fluorescence);

  return 0; /* Tout s'est bien passe */
}

/* -------- Compter le nombre de lignes, sachant que l'on peut avoir des graphes non-spars ! -------- */

static unsigned long bm01_compter_lignes(FILE *fichier, int commentaire, boolean successifs)
{
  unsigned long nbr_lignes = 0;
  long pos_debut;

  /* On memorise le debut */
  pos_debut = ftell(fichier);
  if (FAUX == successifs) { /* On charge tout en une fois */
    int c;

    c = fgetc(fichier);
    while ( (c != commentaire) && (c != EOF) ) { /* Commentaire -> le scan suivant commence */
      nbr_lignes++; /* Une ligne de plus... */
      if (ignore_ligne(fichier) == EOF) /* On l'ignore */
	c = EOF;
      else
	c = fgetc(fichier); /* Et on passe  la suivante */
    }
  } else { /* On ne compte que tant que l'on est en X croissants */
    double x_lu;
    int c, n_lu;

    n_lu = fscanf(fichier, "%lf", &x_lu);
    if ( n_lu != 1 ) {
      (void) fprintf(stderr, "bm01_compter_lignes : la premire ligne ne commence pas par une nergie !\n");
      (void) SYS_ERREUR(ERREUR_LECTURE, NULL, fichier, NULL);
      return 0;
    }
    c = ignore_ligne(fichier);
    if ( EOF != c ) { /* Une seule ligne ! */
      double x_avant;

      do {
	while (c == commentaire) {
	  (void) fgetc(fichier); /* Car ignore_ligne n'a pas lu le caractre ! */
	  c = fgetc(fichier);
#ifdef _DEVELOPPE_
	  printf("bm01_compter_lignes : commentaire au milieu !\n");
#endif
	  if ( 'M' == c ) { /* #M : interruption pour remplissage de l'anneau : on peut ignorer ! */
	    c = ignore_ligne(fichier);
	    if (EOF == c) { /* Le fichier se finit par un commentaire */
#ifdef _DEVELOPPE_
	      printf("bm01_compter_lignes : le fichier finit par un commentaire\n");
#endif
	      fseek(fichier, pos_debut, SEEK_SET); /* On se remet au dbut des donnes */
	      return nbr_lignes;
	    }
	  } else { /* On change de scan ou commentaire inconnu */
#ifdef _DEVELOPPE_
	    printf("bm01_compter_lignes : commentaire de changement de scan [%c] !\n", c);
#endif
	    break;
	  }
	}

	x_avant = x_lu;
	nbr_lignes++; /* La ligne compte */
	
	/* On attaque la suivante */
	n_lu = fscanf(fichier, "%lf", &x_lu);
	if ( n_lu != 1 ) {
	  (void) SYS_ERREUR(ERREUR_LECTURE, NULL, fichier, NULL);
	  return 0;
	}
	c = ignore_ligne(fichier);
      } while ( (x_lu > x_avant) && (c != EOF) );
    } else {
      nbr_lignes = 1;
    }
  }

  fseek(fichier, pos_debut, SEEK_SET); /* On se remet au dbut des donnes */
  return nbr_lignes;
}
