/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modlisation de dtecteurs
*/

#ifndef DETECTEUR_H_
# define DETECTEUR_H_ /* Protection contre l'inclusion multiple */

# include "utils_3d.h" /* Description d'objets dans l'espace */

struct faisceau;
struct graphe;
struct image;
struct matiere;

/* -------- Nature du dtecteur -------- */

#define DETECTEUR_CHAMBRE 1 /* Chambre d'ionisation */
#define DETECTEUR_DIODE   2 /* Photodiode */
#define DETECTEUR_FLUO    3 /* Dtecteur de fluorescence classique : Si/Li, Ge,... */

/* -------- Proprits du dtecteur -------- */

#define DETECTEUR_ABSORBANT    ( 1U << 0 ) /* Bit mis : le dtecteur absorbe */
#define DETECTEUR_NON_LINEAIRE ( 1U << 1 ) /* Bit mis : le dtecteur n'est pas linaire */
#define DETECTEUR_INHOMOGENE   ( 1U << 2 ) /* Bit mis : le dtecteur n'a pas la mme rponse sur toute la surface */
#define DETECTEUR_RESOLU       ( 1U << 3 ) /* Bit mis : le dtecteur spare les nergies */
#define DETECTEUR_NRJ_VARIABLE ( 1U << 4 ) /* Bit mis : le dtecteur n'a pas la mme rponse  toutes les nergies */
#define DETECTEUR_A_L_INFINI   ( 1U << 5 ) /* Bit mis : on peut supposer le dtecteur  l'infini
					               ==> l'angle solide est le meme en tout point de l'chantillon */
#define DETECTEUR_ENTIER       ( 1U << 6 ) /* Bit mis : le dtecteur ne peut donner qu'une valeur entire */
#define DETECTEUR_SURFACE      ( 1U << 7 ) /* Bit mis : le dtecteur n'est pas ponctuel */
#define DETECTEUR_BRUITE       ( 1U << 8 ) /* Bit mis : le dtecteur est bruit / Poisson */
#define DETECTEUR_A_VIDE       ( 1U << 9 ) /* Bit mis : le dtecteur donne un signal  vide */

/* -------- Dfinition de ce qu'est un dtecteur -------- */

typedef struct detecteur {
  struct detecteur *suivant; /* Liste simplement chane des dtecteurs disponibles */
  char *nom;

  int type; /* Type de dtecteur : joue sur les imperfections ventuelles */
  unsigned int proprietes;  /* Proprits : imperfections par exemple */

  Orientation3D orientation;/* Position & orientation du dtecteur */
  double angle_solide;      /* Angle solide intercept par le dtecteur */

  double largeur, hauteur;  /* Dimensions de la surface active, si non-ponctuel */

  struct matiere *composition; /* Si absorbant */
  double longueur;             /* Longueur traverse, en cm */

  double coups_a_vide;       /* Nombre de coups en l'absence de faisceau */

  struct graphe *r_signal;   /* Rponse du dtecteur en fonction de l'intensit */
  struct graphe *r_energie;  /* Rponse du dtecteur en fonction de l'nergie */
  struct image  *r_surface;  /* Rponse en fonction de la surface [image dans le plan du dtecteur] */
} Detecteur;

/* Gestion des dtecteurs */

/* -------- Cration, effacement -------- */

Detecteur *DET_CreerDetecteur  (int type, int proprietes);
int        DET_ChainerDetecteur(Detecteur *avant, Detecteur *detecteur);
int        DET_DechainerDetecteur(Detecteur *detecteur);

int        DET_NommerDetecteur(Detecteur *detecteur, const char *nom);

int        DET_EffacerDetecteur(Detecteur *detecteur, boolean forcer);

/* Version pour le slecteur universel */
int DETECTEUR_Charger_I (unsigned long n_detecteur);
int DETECTEUR_Copier_I  (unsigned long n_original);
int DETECTEUR_Effacer_I (unsigned long n_original);
int DETECTEUR_Modifier_I(unsigned long n_original);
int DETECTEUR_Nouveau_I (unsigned long n_avant);
int DETECTEUR_Sauver_I  (unsigned long n_detecteur);

unsigned long DETECTEUR_Compter(void);
unsigned long DETECTEUR_D_Vers_N(Detecteur *premier, Detecteur *detecteur);
Detecteur    *DETECTEUR_N_Vers_D(Detecteur *premier, unsigned long n);
const char   *DETECTEUR_TrouverNom(unsigned int arbre, unsigned long objet);

Detecteur *DETECTEUR_Charger(FILE *fichier, int version, const char *nom_fichier, struct matiere **table_compositions);
int        DETECTEUR_Sauver (FILE *fichier, Detecteur *detecteur);

int DET_ControlerOptions(Detecteur *detecteur);
int DETECTEUR_Parametrer(Detecteur *detecteur);

/* -------- Manipulation des donnes internes -------- */

/* -------- Utilisation (simulation du fonctionnement) -------- */

double DETECTEUR_FaisceauTraverse_I(Detecteur *detecteur, struct faisceau *faisceau, double energie, struct faisceau *f_sortie);

struct graphe *DETECTEUR_CalculerSpectre(Detecteur *detecteur, struct faisceau *faisceau);

#endif /* DETECTEUR_H_ */
