/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - cration du fichier : 3 juin 1998

   Toutes les fonctions qui interviennent pour l'interpolation
   et le lissage de courbes
*/

#ifndef LISSAGE_H_
# define LISSAGE_H_    /* Protection contre l'inclusion multiple */
# include "complexe.h" /* Nombres complexes */

struct graphe;			/* Dfinie dans graphes.h  */
struct matrice;			/* Dfinie dans matrices.h */
struct serie;			/* Dfinie dans serie.h    */

/* ~~~~~~~~ Codes d'erreur ~~~~~~~~ */

#define ERR_LIS_DOUBLON         (-1)
#define ERR_LIS_TROP_PEU_POINTS (-2)
#define ERR_LIS_DIV_ZERO        (-3)
#define ERR_LIS_MEMOIRE         (-4)
#define ERR_LIS_NON_TRIE        (-5)
#define ERR_LIS_NON_INVERSIBLE  (-6)
#define ERR_LIS_AUCUN_GRAPHE      (-7)
#define ERR_LIS_METHODE_INADAPTEE (-8)
#define ERR_LIS_GRAPHES_DISJOINTS (-9)
#define ERR_LIS_AUTRE_ERREUR    (-100)

int LIS_ERREUR(int num_erreur, ...); /* Dans Lissage/erreurs.c */

/* ~~~~~~~~ Mthodes d'interpolation ~~~~~~~~ */

#define INTERPOLE_LINEAIRE   0  /* Polynmes simples */
#define INTERPOLE_PARABOLE   1
#define INTERPOLE_CUBIQUE    2  /* 3 = sparateur */
#define INTERPOLE_SPLINE     4  /* Splines cubiques */
#define INTERPOLE_SPLINE_AUC 5	/* Splines, en conservant l'aire sous la courbe */

#define NBR_METHODES_INTERPOLATION 6

const char *INT_NomMethode(unsigned int numero_methode);

#define INTERPOLE_BORDS_0_0 0 /* Si on dborde, on prend 0 arbitrairement */
#define INTERPOLE_BORDS_0_Y 1 /* On prend 0 avant, la dernire valeur disponible aprs */
#define INTERPOLE_BORDS_Y_0 2 /* On prend la dernire valeur disponible avant, 0 aprs */
#define INTERPOLE_BORDS_Y_Y 3 /* On prend la dernire valeur disponible avant et aprs */

/* -------- Les paramtres de lissage -------- */

struct infos_lissage {
  int methode;
  boolean table_fournie;	/* VRAI : on utilise une table des X, sinon x_min, x_max et le pas */
  double x_min, x_max, pas;
  unsigned long nbr_points;
  double *table_x;
  int code_erreur;	/* Contiendra le code de l'erreur en cas d'erreur, 0 si aucun problme, <0 si grave, >0 si avertissement */
};

/* -------- Fonctions trs gnriques -------- */

struct graphe *GRF_Interpoler     (struct graphe *source, struct infos_lissage *comment);
int            LIS_InterpolerSerie(struct serie *serie, struct infos_lissage *infos);

/* -------- Gnralits -------- */

struct graphe *GRF_CalerSurGrille(const struct graphe *source, int methode, int bords,
				  double x_min, double x_max, unsigned long n_points, int *code_erreur);
struct graphe *GRF_CalerSurGraphe(const struct graphe *source, int methode, int bords, const struct graphe *reference, int *code_erreur);
struct graphe *GRF_CalerSurTable (const struct graphe *graphe, unsigned long nbr_pts, double *x, int methode, int bords, int *code_erreur);

int            LIS_RecalerGraphe (const struct graphe *original, struct graphe *resultat, int methode, int bords);

unsigned long LIS_RepererPointAvant(double *valeurs, unsigned long nbr_valeurs, double x, unsigned long i_depart);

/* -------- Interpolations polynomiales simples -------- */

double LIS_lineaire(double x1, double x2, double y1, double y2, double x);
double LIS_planaire(double x1, double y1, double z1, double x2, double y2, double z2,
		    double x3, double y3, double z3, double x, double y);

double GRF_LireValeur(const struct graphe *graphe, double x, int mode);

double GRF_InterpoleLineaire       (const struct graphe *graphe, double x, unsigned long *depart);
double GRF_InterpoleLineaire_Erreur(const struct graphe *graphe, double x, unsigned long *depart);

/* ****** Mthode des splines _pour les spectres rels_ ********* */

/* -------- Splines cubiques -------- */

/* ------ Prparations des tableaux ----------------------------- */

int SPL_prepare(const struct graphe *graphe, char nbr_zones, double *zones, 
                   double **b, double **c, double **d, double **tableau,
		   unsigned long *nbr_points);
int SPL_tt_prep(const struct graphe *graphe, double **b, double **c, double **d);

/* ------ Calculs des points avec ces tableaux ------------------ */

double SPL_calcule  (double x, unsigned long nbr_pts, double *tableau,
		     double *b, double *c, double *d);
double SPL_tt_calc  (const struct graphe *graphe, double x, double *b, double *c, double *d);
double SPL_tt_derive(const struct graphe *graphe, double x, double *b, double *c, double *d, double *derivee);

/* ------ Pour calculer en une fois tout un spectre liss ------- */

struct graphe *lisse_reel(struct graphe *source, unsigned long nv_points) __attribute__((__deprecated__));

/* ****** Mthode des splines _pour les spectres complexes_ ***** */

/* ------ Prparations des tableaux ----------------------------- */

double CSPL_prepare(const struct graphe *graphe, int nbr_zones, double *zones, 
	 	    Complexe **b, Complexe **c, Complexe **d, double **tab_x, Complexe **tab_y,
		    unsigned long *nbr_points);
double CSPL_tt_prep(const struct graphe *graphe, Complexe **b, Complexe **c, Complexe **d);

/* ------ Calculs des points avec ces tableaux ------------------ */

Complexe CSPL_calcule(double x, unsigned long nbr_pts, double *tab_x, Complexe *tab_y,
		      Complexe *b, Complexe *c, Complexe *d);
Complexe CSPL_tt_calc(const struct graphe *graphe, double x, Complexe *b, Complexe *c, Complexe *d);

/* ------ Pour calculer en une fois tout un spectre liss ------- */

struct graphe *lisse_complexe(struct graphe *source, unsigned long nv_points);

/* ****** Mthode des moindres carrs *************************** */

int Gauss(int degre, double *matrice, double *coefs);

int MC_Lineaire_Erreurs   (struct matrice *mat_MDC, struct matrice *v_constant, struct matrice **erreurs);
int MC_Lineaire_Resolution(struct matrice *mat_MDC, struct matrice *v_constant, struct matrice *v_parametres);

/* ------ Polynme simple, degr indiffrent -------------------- */

double MC_Polynome_PreparationMatrices(unsigned int degre, double *table_X, double *table_Y, unsigned long nbr_points,
				       struct matrice *mat_MDC, struct matrice *v_constant);
double MC_Polynome_SurPartieGraphe    (const struct graphe *graphe, int degre, int nbr_zones, double *zones, double *poly);

/* ------ Succession de polynmes avec plusieurs points d'ancrage [juris] */

double MC_Polynome_SurPartieGraphe_Juris(const struct graphe *graphe, int degre, int nbr_zones, double *zones, double *coefs);

#endif /* LISSAGE_H_ */

