/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS
   Cration : mars 2006

   Interface graphique avec l'utilisateur

   Bruitage
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "serie.h"
#include "interface.h"
#include "statistiques.h"

/* -------- Variables locales -------- */

#include "int_bruiter.txt"

/* -------- Fonctions locales -------- */

static int  I_GRF_ParametresBruitage   (struct infos_bruitage *infos, Graphe *source, boolean *copie);
static void I_GRF_PreparerArbreBruitage(struct infos_bruitage *infos, Graphe *source, boolean copie);

/* ~-~-~-~-~-~-~-~-~ Bruiter un graphe [valeurs relles] ~-~-~-~-~-~-~-~-~ */

Graphe *I_GRF_Bruiter(Graphe *graphe)
{
  Graphe *copie;
  boolean nouveau = FAUX;
  int reponse;
  struct infos_bruitage infos = { STAT_BRUITER_CONSTANTE, DP_GAUSSIENNE, 0.0, NULL };

  /* Contrles prliminaires */
  if ( NULL == graphe ) return NULL;

  if ( FAUX == graphe->drapeaux.theorique ) {
    int reponse;

    reponse = BI_ERR_alerte( NULL, &ERR_GrapheExperimental, VRAI );
    if ( reponse != 1 ) return NULL;
  }

  /* Choix de la mthode de bruitage */
  reponse = I_GRF_ParametresBruitage( &infos, graphe, &nouveau );
  if ( reponse != 0 ) return NULL; /* Annulation ou problme */

  if ( VRAI == nouveau ) { /* On veut normaliser une copie, sans toucher  l'original */
    copie = GRF_Copier( graphe, NULL, VRAI );
    if ( NULL == copie ) {
/*       (void) BI_ERR_alerte( principale, &ERR_CopieImpossible, VRAI ); */
      return NULL;
    }
  } else {
    copie = graphe;
  }

  /* On ralise le bruitage _de la copie_ */
  reponse = STAT_GRF_Bruiter( copie, &infos );
  if ( reponse != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans le bruitage [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, reponse);
    if ( copie != graphe ) {
      /* Aucune raison de garder la copie ! */
      GRF_eff_isole( copie );
    }
    return NULL;
  }

  /* Si la normalisation a russi, on insre la copie aprs l'original [si copie != original !] */
  if ( copie != graphe ) {
    char *titre = NULL;

    reponse = GRF_InsererDansSerie( copie, NULL, graphe );
    if ( reponse != 0 ) {
      OPTION_A_FAIRE( "Insertion dans une autre srie en cas d'chec : demander la srie ou en crer une nouvelle" );
    }

    /* On adapte les paramtres d'affichage : axes inchangs */
    GRF_CopierAffichage( copie, graphe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_DESSIN );

    /* On donne un titre */
    if ( STAT_BRUITER_CONSTANTE == infos.methode ) {
      titre = TXT_genere_chaine( "%s bruit"e_AIGU" ["SIGMA" = %*f]", graphe->nom, infos.sigma, GRF_QuelleUnite( graphe, AXE_Y ) );
    } else if ( STAT_BRUITER_PROPORTIONNELLE == infos.methode ) {
      titre = TXT_genere_chaine( "%s bruit"e_AIGU" ["SIGMA" = %P de Y]", graphe->nom, infos.sigma );
    }

    GRF_titre( copie, titre );
    free( titre ); titre = NULL;
  }

  /* On indique, dans l'historique, ce que l'on a fait */
  HIST_Titre(1, "Bruitage", "BRUI");
  HIST_Texte(1, 0, "Graphe bruit"e_AIGU" : %s\n", graphe->nom);

  /* Tout s'est bien pass */
  return copie;
}

/* -------- Arbre de paramtrage du bruitage -------- */

static int I_GRF_ParametresBruitage(struct infos_bruitage *infos, Graphe *source, boolean *copie)
{
  int reponse;

  /* On prpare l'arbre */
  I_GRF_PreparerArbreBruitage( infos, source, *copie );

  /* On l'appelle */
  reponse = ARB_Bloquant( BRUITER_GRAPHE, titre_arbre_bruiter_graphe, principale, 0 );
  if ( reponse != BG_OK ) {
    /* Annulation ou problme */
    return -1;
  }

  /* On lit les valeurs indiques par l'utilisateur */
  infos->loi_proba = OBJ_OptionMenu( BRUITER_GRAPHE, BG_FAMILLE, 0, QUESTION );
  if ( VRAI == OBJ_choisit( BRUITER_GRAPHE, BG_CONSTANTE, QUESTION ) ) {
    /* Erreur constante avec X */
    infos->methode = STAT_BRUITER_CONSTANTE;
    infos->sigma = OBJ_LireNombre( BRUITER_GRAPHE, BG_S_CONSTANTE, NULL );
  } else if ( VRAI == OBJ_choisit( BRUITER_GRAPHE, BG_PROPORTION, QUESTION ) ) {
    /* Erreur proportionnelle  X */
    infos->methode = STAT_BRUITER_PROPORTIONNELLE;
    infos->sigma = OBJ_LireNombre( BRUITER_GRAPHE, BG_S_PROP, NULL );
  } else {
    /* Erreur quelconque, un graphe donne l'erreur en fonction de X */
    infos->methode = STAT_BRUITER_QUELCONQUE;
    OPTION_A_FAIRE( "Choix du graphe contenant s=f(X)" );
    return +1;
  }

  *copie   = OBJ_choisit( BRUITER_GRAPHE, BG_NOUVEAU, QUESTION );

  /* On a termin sans souci */
  return 0;
}

static void I_GRF_PreparerArbreBruitage(struct infos_bruitage *infos, Graphe *graphe, boolean copie)
{
  if ( FAUX == arbres[ BRUITER_GRAPHE ].drapeaux.prepare ) {
    int retour;

    /* On installe le slecteur de mthodes */
    OBJ_ChangerType( BRUITER_GRAPHE, BG_FAMILLE, G_SSMENU );
    OBJ_MettreMenu ( BRUITER_GRAPHE, BG_FAMILLE, DP_NOMBRE_LOIS, txt_loi_probas_continues, NULL, NULL );

    /* Initialisation classique */
    retour = ARB_Initialiser( BRUITER_GRAPHE, NULL, NULL, NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur en voulant initialiser l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }

    /* On amliore le placement des objets */
    OBJ_deplace( BRUITER_GRAPHE, BG_GRAPHE, 0, +2, DEPLACE_PIXELS );

    /* On traduit si ncessaire */
#ifdef AVEC_TRADUCTION_
#endif

    /* Valeurs initiales uniquement pour le premier appel */
    OBJ_MettreNombre( BRUITER_GRAPHE, BG_S_CONSTANTE, ( graphe->y_max - graphe->y_min ) * 0.001, 0.0 );
    OBJ_MettreNombre( BRUITER_GRAPHE, BG_S_PROP, 10.0, 0.0 );

    /* Adaptation termine */
    arbres[ BRUITER_GRAPHE ].drapeaux.prepare = VRAI;
  }

  /* On installe les valeurs initiales */
  OBJ_OptionMenu( BRUITER_GRAPHE, BG_FAMILLE, infos->loi_proba, METTRE );

  OBJ_OptionsEditable( BRUITER_GRAPHE, BG_S_CONSTANTE, BI_EDITABLE_MIN | BI_EDITABLE_UNITE, 0.0, 0.0, GRF_QuelleUnite( graphe, AXE_Y ) );

  OBJ_choisit( BRUITER_GRAPHE, BG_NOUVEAU, VRAI == copie );
}
