/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2010, Emmanuel CURIS

   Généralités du programme LASE (interface graphique)

   Gestion des chemins dans un modèle d'ajustement EXAFS par chemins
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "systeme.h"

#include "exafs.h"

#include "atomes.h"
#include "couches.h"
#include "feff.h"
#include "graphes.h"
#include "interface.h"
#include "modeles.h"

#include "int_chemins.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void I_CHEM_PreparerArbreCombinaison(void);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Combinaison de chemins géométriquement (ou chimiquement) proches ~-~-~-~-~-~-~-~-~ */

/* -------- Appel par l'utilisateur --------
    Ne renvoie rien

    modele = le modèle qui contient les chemins à combiner [NULL == choisir le modèle courant]
 */

void I_CHEM_CombinerChemins(struct modele_ajuste *modele)
{
  int reponse, retour;
  unsigned long nombre_chemins_initial = 0;
  struct infos_combiner_couches parametres = { COMBINER_COUCHES_DISTANCE | COMBINER_COUCHES_ANGLE | COMBINER_COUCHES_N_CONSTANT,
					       0.05, 10.0, 0UL, 0UL };
  Fenetre *fenetre_liste_couches = NULL;

  /* On recherche le modèle courant */
  if ( NULL == modele ) {
    if ( NULL == mol_courante ) {
      (void) BI_ERR_alerte( fen_vue_3d, &ERR_AucuneMolecule, VRAI );
      return;
    }
    if ( NULL == mol_courante->absorbeurs ) {
      (void) BI_ERR_alerte( fen_vue_3d, &ERR_MoleculeSansAbsorbeur, VRAI );
      return;
    }
    modele = mol_courante->absorbeurs[ mol_courante->absorbeur_courant ].modele_actif;
    if ( NULL == modele ) {
      (void) BI_ERR_alerte( fen_vue_3d, &ERR_AbsorbeurSansModele, VRAI );
      return;
    }
  }

  /* Controles sur le modèle */
  if ( modele->type != MODELE_EXAFS_CHEMINS ) {
    (void) BI_ERR_alerte( fen_vue_3d, &ERR_TypeModeleIncorrect, VRAI );
    return;
  }
  if ( ( NULL == modele->infos_modele.feff ) || ( NULL == modele->infos_modele.feff->couches ) ) {
    (void) BI_ERR_alerte( fen_vue_3d, &ERR_PasDeChemin, VRAI );
    return;
  }
  if ( 1 == modele->infos_modele.feff->nbr_couches ) {
    (void) BI_ERR_alerte( fen_vue_3d, &ERR_CheminUnique, VRAI );
    return;
  }

  /* Préparation de l'arbre */
  if ( FAUX == arbres[ COMBINE_CHEMINS ].drapeaux.prepare ) {
    I_CHEM_PreparerArbreCombinaison();
  }

  /* Appel de l'arbre */
  reponse = ARB_Bloquant( COMBINE_CHEMINS, titre_arbre_combiner_chemins, fen_vue_3d, 0 );
  if ( CC_ANNULER == reponse ) {
    return;
  }

  /* Lecture des réponses de l'arbre */
  /* a. Mode de combinaison */
  if ( OBJ_choisit( COMBINE_CHEMINS, CC_DD_TOTAL, QUESTION ) == VRAI ) {
    parametres.drapeaux |= COMBINER_COUCHES_TOTAL;
  }

  /* b. Paramètres, limites,... */
  parametres.distance_limite = OBJ_LireNombre( COMBINE_CHEMINS, CC_DIFF_DIST , NULL );
  parametres.angle_limite    = OBJ_LireNombre( COMBINE_CHEMINS, CC_DIFF_ANGLE, NULL ); /* Angle affiché & mémorisé en degrés : nombre normal... */

  if( OBJ_choisit( COMBINE_CHEMINS, CC_ESTIME_DW, QUESTION ) == VRAI ) {
    parametres.drapeaux |= COMBINER_COUCHES_ESTIMER_DW;
  }

  /* Réalisation des calculs */
  nombre_chemins_initial = modele->infos_modele.feff->nbr_couches;
  retour = NC_CombinerCouches( modele->infos_modele.feff, &parametres );
  if ( retour != 0 ) {
    return;
  }

  /* Affichage des résultats */
  (void) BI_ERR_alerte( fen_vue_3d, &MSG_ResultatsCombinaison, FAUX,
			nombre_chemins_initial, modele->infos_modele.feff->nbr_couches, parametres.nombre_uniques, parametres.nombre_groupes );

  /* On protège contre des erreurs éventuelles d'affichage */
  /* a. Couche éditée ? */
  fenetre_liste_couches = ARB_QuelleFenetre( arbres + AJ_EXAFS_CHEMINS, FAUX );
  if ( fenetre_liste_couches != NULL ) { /* Éditeur de couches ouvert */
    struct modele_feff *modele_fenetre;

    modele_fenetre = (struct modele_feff *) FEN_LireComplement( fenetre_liste_couches, FEN_COMPLEMENT_1 );
    if ( modele->infos_modele.feff == modele_fenetre ) { /* sur le modèle modifié ! */
      couche_editee = NULL; /* On ne sauve pas l'ancienne couche : rien ne prouve qu'elle existe encore ! */
      COUCHE_Aller( modele->infos_modele.feff, modele->infos_modele.feff->couches ); /* On se place (arbitrairement) sur la première couche */
    }
  }

  /* b. Absorbeur en cours d'affichage ? */
  if ( modele == mol_courante->absorbeurs[ mol_courante->absorbeur_courant ].modele_actif ) {
    VUE3D_ChoisirChemin();
  }
}

/* -------- Préparation de l'arbre -------- */

static void I_CHEM_PreparerArbreCombinaison(void)
{
  /* On initialise l'arbre */

  /* Au besoin, traduction */
#ifdef AVEC_TRADUCTION_
#endif

  /* Terminé */
  arbres[ COMBINE_CHEMINS ].drapeaux.prepare = VRAI;
}
