/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 2006, Emmanuel CURIS
   Création : 2 juin 2006

   Interface graphique

   Décomposition d'un graphe sur une base de 2 [ou plus à terme] graphes
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "automatise.h"
#include "decomposition.h"	/* Les modèles spécifiques pour les décompositions */
#include "interface.h"
#include "lissage.h"		/* Devrait disparaître après réorganisation de la décomposition ! */
#include "modeles.h"		/* La décomposition se fait avec les modèles ! */
#include "serie.h"
#include "tracer.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LG_MAX_NOM_METHODE 25
#define LG_MAX_NOM_SPECTRE 50

#define METHODE_PAR_DEFAUT MODELE_DECOMPOSITION_MOINDRES_CARRES

/* static unsigned int methode_de_decomposition = METHODE_PAR_DEFAUT;  /\* Pour mémoriser la méthode d'un appel à l'autre ! *\/ */

static ModeleAjustement *premier_modele = NULL;	 /* Premier modèle de la liste
						    Liste des modèles de décomposition, accessible de l'extérieur
						    en utilisant I_ModeleDecomposition_Premier()
						  */

static unsigned long i_vecteur_actif = 0;

#include "int_decomposition.txt"  /* Les traductions */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int  CLV_decompose(Fenetre *fenetre, unsigned int touche, int etat);

static void I_GRF_Decomposition_ChangerDeMethode (unsigned int arbre, unsigned int objet, unsigned long option_choisie);
static int  I_GRF_Decomposition_GererArbre       (unsigned int numero_objet, int x, int y);
static void I_GRF_Decomposition_MettreModele     (ModeleAjustement *modele);
static void I_GRF_Decomposition_MettreVecteurBase(struct element_base *vecteur_base);
static int  I_GRF_Decomposition_PreparerArbre    (unsigned int numero_arbre);

static int      I_GRF_Decomposition_FermerFenetre  (Fenetre *fenetre, int mode_fermeture);
static int      I_GRF_Decomposition_Messages       (Fenetre *fenetre, Message *message);
static Fenetre *I_GRF_Decomposition_RappelerFenetre(Graphe *graphe);
static void     I_GRF_Decomposition_RedessinFenetre(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);
static int      I_GRF_Decomposition_ClicFenetre    (Fenetre *fenetre, int bouton, int n_clic, int etat, int x, int y);

static void     I_GRF_Decomposition_LibererInfos(ModeleAjustement *infos);

static void     I_GRF_Decomposition_Historique(Graphe *source, ModeleAjustement *infos);

static ModeleAjustement *I_MOD_N_V_Modele(unsigned long numero_modele);

static int I_GRF_Decomposition_LancerAjustement(ModeleAjustement *modele, int afficher_resultats);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Création d'un modèle de décomposition ~-~-~-~-~-~-~-~-~ */

struct modele_ajuste *I_ModeleDecomposition_Nouveau(unsigned long numero_avant, struct graphe *graphe_a_decomposer)
{
  ModeleAjustement *modele_avant, *nouveau_modele;
  const char *nom_modele;
  struct graphe **references;
  unsigned long nombre_references = 0;

  if ( NULL == graphe_a_decomposer ) {
    graphe_a_decomposer = graphe_courant;
  }

  /* On demande le nom du modèle à créer */
  nom_modele = BI_Demande_Texte( principale, &Q_NomModele, txt_nouveau_modele );
  if ( NULL == nom_modele ) {
    /* On annule, d'une façon ou d'une autre */
    return NULL;
  }

  /* On repère celui à mettre avant */
  modele_avant = I_MOD_N_V_Modele( numero_avant );

  /* On crée le modèle */
  nouveau_modele = MODELE_CreerModele( NULL, nom_modele, MODELE_DECOMPOSITION, graphe_a_decomposer, VRAI );
  if ( NULL == nouveau_modele ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de créer un nouveau modèle !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return NULL;
  }

  /* On fait le chaînage */
  if ( NULL == modele_avant ) {	/* C'est que cela devient le premier ! */
    nouveau_modele->suivant = premier_modele;
    premier_modele = nouveau_modele;
  } else {
    nouveau_modele->suivant = modele_avant->suivant;
    modele_avant->suivant = nouveau_modele;
  }

  /* On choisit les graphes qui serviront de vecteur de base... */
  references = GRF_mchoisir( graphe_a_decomposer->type, 2, 0, &nombre_references );
  if ( ( references != NULL ) && ( nombre_references > 0 ) ) {
    unsigned long i;
    struct graphe *graphe;

    for ( i = 0 ; i < nombre_references ; i++ ) {
      int retour;

      graphe = references[ i ];
      /* printf( "On ajoute le graphe n° %lu / %lu [%p]\n", i, nombre_references, graphe ); */
      /* printf( "Graphe : %p [%s]\n", graphe, graphe->nom ); */
      retour = MODELE_Decomposition_AjouterVecteurDeBase( nouveau_modele->infos_modele.decomposition,
      							  graphe );
      if ( retour != 0 ) {
	OPTION_A_FAIRE( "Avertir d'un problème dans la génération de la base de décomposition..." );
      }
    }
  }

  /* Terminé sans erreur */
  return nouveau_modele;
}

/* -------- Conversion numéro de modèle -> modèle lui-même -------- */

static ModeleAjustement *I_MOD_N_V_Modele(unsigned long numero_modele)
{
  unsigned long i;
  ModeleAjustement *modele;

  /* S'il n'y a aucun modèle, de toute façon quel que soit le numéro, sans intérêt... */
  if ( NULL == premier_modele ) {
    if ( numero_modele > 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : Numéro de modèle incorrect [%lu]\n",
		      __FILE__, __FUNCTION__, __LINE__, numero_modele );
    }
    return NULL;
  }

  for ( modele = premier_modele, i = 0 ; modele != NULL ;
	i++, modele = modele->suivant ) {
    if ( i == numero_modele ) return modele;
  }

  (void) fprintf( stderr, "%s,%s l.%u : Numéro de modèle incorrect [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, numero_modele );
  return NULL;
}

/* -------- Lancement d'une décomposition -------- */

static int I_GRF_Decomposition_LancerAjustement(ModeleAjustement *modele, boolean afficher_resultats)
{
  int retour;

  /* Décomposition : les résultats ont peu d'intérêt, on les efface sans demander... */
  retour = MODELE_AjustementComplet( modele, NULL, FAUX, FAUX, Q_MODELE_ANCIENS_RESULTATS_EFFACER );
  if ( retour != PAS_DE_PROBLEME ) {
    (void) fprintf( stderr, "%s,%s l.%u : Erreur en voulant faire la décomposition\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return retour;
  }

  /* Si pas de souci : les résultats ont potentiellement changé, on affiche et on redessine */
  I_GRF_Decomposition_MettreVecteurBase( modele->infos_modele.decomposition->base + i_vecteur_actif );
  FEN_redessin( ARB_QuelleFenetre( arbres + DECOMP_GRAPHE, FAUX ), BI_RDS_UTIL );

  /* Veut-on en plus afficher les résultats ? */
  if ( VRAI == afficher_resultats ) {
    BI_Boite_Alerte MSG_AfficherResultats = {
      ERR_ICONE_INFO,
      titre_message_resultats, NULL,
      1, 1, 1, { texte_bouton_compris, NULL, NULL, NULL, NULL } };
    char *message = NULL;
    unsigned long i;

    /* On construit le message avec les coefficients de chaque graphe */
    message = strdup( texte_en_tete_message_resultats );
    for ( i = 0 ; i < modele->infos_modele.decomposition->nombre_bases ; i++ ) {
      message = TXT_AjouteTexte( message, TXT_genere_chaine( "Graphe %lu : %e\n", i + 1, 
							     VAR_LireValeur( &( modele->infos_modele.decomposition->base[ i ].coefficient ),
									     VARIABLE_LIRE_VALEUR_AJUSTEE ),
							     modele->infos_modele.decomposition->base[ i ].erreur ) );
    }
    MSG_AfficherResultats.texte = message;

    BI_ERR_alerte( NULL, &MSG_AfficherResultats, VRAI );

    free( message ); message = NULL; MSG_AfficherResultats.texte = NULL;
  }

  return retour;
}

/* ~-~-~-~-~-~-~-~-~ Paramétrage d'un modèle de décomposition ~-~-~-~-~-~-~-~-~ */

/* -------- Appel par l'utilisateur -------- */

int I_GRF_Decomposer(struct graphe *source)
{
  ModeleAjustement *modele_decomposition;
  Fenetre *fenetre = NULL;

  /* Contrôle sur le graphe */
  if ( NULL == source ) return -1;  /* Pas de graphe indiqué */

  /* Est-ce que l'arbre est déjà ouvert ? */
  if ( VRAI == ARB_EstOuvert( DECOMP_GRAPHE ) ) {
    fenetre = I_GRF_Decomposition_RappelerFenetre( source );
    if ( NULL == fenetre ) {
      /* On ne souhaite pas mettre à jour la fenêtre avec le graphe indiqué -> terminé ! */
      return -2;
    }
  }

  /* Existe-t-il des modèles de décomposition ? */
  modele_decomposition = I_ModeleDecomposition_ChoisirCourant( source );
  if ( NULL == modele_decomposition ) {
    (void) fprintf( stderr, "%s, %s l.%u : Impossible de trouver ou créer un modèle de décomposition !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FEN_Detruire( fenetre );	/* La fenêtre peut encore exister, si était associée à un ancien graphe... */
    return -3;
  }
  I_GRF_Decomposition_MettreModele( modele_decomposition );

  /* Ici : fenêtre à initialiser, voire à créer */
  if ( NULL == fenetre ) {
    /* Il faut bien créer la fenêtre, pas seulement la réinitialiser */
    fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, titre_fenetre, NULL,
				  DECOMP_GRAPHE, BI_FEN_ARBRE_HAUT, I_GRF_Decomposition_PreparerArbre,
				  I_GRF_Decomposition_RedessinFenetre, CLV_decompose,
				  I_GRF_Decomposition_ClicFenetre, I_GRF_Decomposition_Messages, RCT_pointeur,
				  ( int (*)(Fenetre *) ) NULL );
    if ( NULL == fenetre ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la fenêtre pour l'interface !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      MODELE_EffacerModele( NULL, modele_decomposition, VRAI ); modele_decomposition = NULL;
      return -4; /* Rien à faire... */
    }
    fenetre->drapeaux.redessin    = VRAI;
    fenetre->msg_masque.fermeture = VRAI; /* Messages traités par MSG_decompose */
  }

  /* On protège le graphe à modéliser */
  source->drapeaux.protege = VRAI;

  /* On associe le graphe à la fenêtre */
  FEN_GRF_LierGraphe( fenetre, source );

  /* On associe les informations à la fenêtre & à l'arbre */
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, modele_decomposition );
  ARB_MettreComplement( DECOMP_GRAPHE, modele_decomposition );

  /* On ouvre la fenêtre */
  FEN_Ouvrir( fenetre );

  /* On fait le calcul */
  I_GRF_Decomposition_LancerAjustement( modele_decomposition, FAUX );

  /* Terminé sans erreur */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Aides pour la gestion de la fenêtre ~-~-~-~-~-~-~-~-~ */

/* ------ Fermeture de la fenêtre (avec ou sans calcul) -------- */

static int I_GRF_Decomposition_FermerFenetre(Fenetre *fenetre, int mode_fermeture)
{
  int reponse = BI_FEN_FERMER_FORCER_ANNULE;  /* Par défaut, on annule */

  Graphe *experience;
  ModeleAjustement *modele;

  /* On récupère les informations */
  experience = (Graphe *)     FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  modele     = (ModeleAjustement *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( ( NULL == experience ) || ( NULL == modele ) ) {
    mode_fermeture = BI_FEN_FERMER_FORCER_ANNULE;
  }

  /* Suivant le type de fermeture, on adapte le calcul à faire*/
  if ( BI_FEN_FERMER_FORCER_NORMAL == mode_fermeture ) { /* On demande ce qu'il faut faire */
    reponse = BI_ERR_alerte( fenetre, &Q_GRF_DCMP_QuitterModule, VRAI );
  } else if ( BI_FEN_FERMER_FORCER_ANNULE == mode_fermeture ) {
    /* On force la fermeture en annulant les calculs */
    reponse = I_GRF_DCMP_FIN_MODULE_FERMER;
  } else if ( BI_FEN_FERMER_FORCER_OK == mode_fermeture ) {
    /* On force la fermeture en faisant les calculs [l'idéal serait de demander, si opportun, si on le fait aussi sur la série] */
    reponse = I_GRF_DCMP_FIN_MODULE_RESULTAT;
  }

  if ( I_GRF_DCMP_FIN_MODULE_ANNULER == reponse ) { /* On annule la fermeture */
    return BI_EVENEMENT_TRAITE;
  }

  /* De toute façon, on ferme ==> on délie les infos de la fenêtre ; évite les problèmes lors de l'effacement des graphes */
  /* On dissocie les infos de la fenêtre.
     Obligatoire, sinon : si on efface (p. ex.) infos->c_lineaire
     on reçoit le message de fermeture -> on demande d'effacer
     infos->c_lineaire -> on reçoit la fermeture ->... : stack overflow !
				  
     Comme ça : si on efface, on reçoit fermeture -> on enlève les infos, on
     demande d'effacer infos->c_lineaire, non trouvé associé à la fenêtre :
     on s'arrête aussitôt - les autres spectres à effacer sont de toute façon
     effacés après, donc devrait tourner

     [18 mai 2008 : c_lineaire et residu étant devenus isolés, ne devrait plus pouvoir se produire]
  */
        
  ARB_MettreComplement( DECOMP_GRAPHE, NULL );
  FEN_GRF_LierGraphe( fenetre, NULL );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );
  FEN_Detruire( fenetre );

  if ( I_GRF_DCMP_FIN_MODULE_FERMER == reponse ) { /* On ferme sans conserver les résultats */
    /* On informe que le graphe résultat va disparaître */
/*     GRF_modifier( infos->c_lineaire, NULL, VRAI ); */

    /* Et l'on fait l'effacement */
/*     I_GRF_Decomposition_LibererInfos( infos ); infos = NULL; */

/*     GRF_DeprotegerGraphe( experience ); */

    return BI_EVENEMENT_FEN_FERMEE;
  }

  if ( I_GRF_DCMP_FIN_MODULE_RESULTAT == reponse ) {
    char *titre;
    struct serie *serie_resultats = NULL;

    /* On veut juste mémoriser les résultats : on les place si possible juste à la suite de l'expérience */
    serie_resultats = I_GRF_InsererDansSerie( NULL, experience, modele->modele );
    if ( NULL == serie_resultats ) {
      (void) fprintf( stderr, "%s, %s l.%u : ERREUR, impossible de placer le graphe dans une nouvelle série !\n",
		      __FILE__, __FUNCTION__, __LINE__ );
    }
    serie_resultats = I_GRF_InsererDansSerie( serie_resultats, modele->modele, modele->residu );

    /* On adapte les légendes, unités,... */
    GRF_CopierAffichage( modele->modele, experience,
			 GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y );
    GRF_CopierAffichage( modele->modele, experience,
			 GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y );

    /* On adapte le titre */
    titre = TXT_genere_chaine( titre_modele, experience->nom, 2 );
    GRF_titre( modele->modele, titre );
    free( titre ); titre = NULL;

    titre = TXT_genere_chaine( titre_residu, experience->nom, 2 );
    GRF_titre( modele->residu, titre );
    free( titre ); titre = NULL;

    /* On laisse une trace dans l'historique */
    I_GRF_Decomposition_Historique( experience, modele );

    /* Terminé */
    GRF_DeprotegerGraphe( modele->modele ); modele->modele = NULL; /* Pour ne pas l'effacer en fermant les infos ! */
    GRF_DeprotegerGraphe( modele->residu ); modele->residu = NULL; /* Pour ne pas l'effacer en fermant les infos ! */
    GRF_DeprotegerGraphe( experience );

    I_GRF_Decomposition_LibererInfos( modele );

    return BI_EVENEMENT_FEN_FERMEE;
  }

  if ( I_GRF_DCMP_FIN_MODULE_SERIE == reponse ) {
    /* Si on est là : on veut décomposer pour toute la série... */
    /* Les coefficients sont imprimés, par contre les graphes calculés ne sont pas intéressants dans ce cas */
    Serie_Graphes *serie;

    OPTION_A_FAIRE( "Demander si l'on veut le modèle de chaque graphe ou un suivi « cinétique »" );

    /* À quelle série appartient le graphe actuellement traité ? */
    serie = GRF_QuelleSerie( experience );
    if ( NULL == serie ) {
      (void) fprintf(stderr, "%s, %s l.%u : PROBLÈME, graphe décomposé isolé, comment a-t-on pu arriver là ?\n",
		     __FILE__, __FUNCTION__, __LINE__);
    } else {
      Graphe *graphe;

      OPTION_A_FAIRE( "Mémorisation dans un graphe de chaque coefficient ; affichage mieux ; ..." );
      for ( graphe = serie->premier_graphe ; graphe != NULL ; graphe = graphe->suivant) {
	if ( graphe->type == experience->type ) {
	  /* Les deux graphes sont compatibles */
/* 	  GRF_DecomposerSurBase( graphe, modele ); */
/* 	  printf( "%f %f pour %s\n", infos->coef_1,infos->coef_2, graphe->nom ); */
	}
      }
    }

    /* Terminé */
    GRF_DeprotegerGraphe( modele->modele );
    GRF_DeprotegerGraphe( modele->residu );
    GRF_DeprotegerGraphe( experience );

    I_GRF_Decomposition_LibererInfos( modele );

    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "On ne devrait jamais arriver ici ! reponse = %d", reponse );

  return BI_EVENEMENT_ERREUR;
}

/* ----------------- Traitement des messages (décomposition d'un spectre) -------------- */

static int I_GRF_Decomposition_Messages(Fenetre *fenetre, Message *message)
{
  int retour = BI_EVENEMENT_NON_TRAITE;

  switch ( message->type )
  {
  case MSG_GRAPHE_MODIFIE:
    OPTION_A_FAIRE( "Gérer une modification du graphe expérimental ou des graphes de référence" );
    break;

  case WM_CLOSED:
    retour = I_GRF_Decomposition_FermerFenetre( fenetre, message->fenetre.haut );
    break;

  default:
    AVERTISSEMENT( "Message ignoré [%d]", message->type );
  }

  /* On renvoie ce que l'on a eu comme réponse */
  return retour;
}

/* ----------------- Redessin de la fenêtre d'interface (décomposition d'un spectre) --- */

static void I_GRF_Decomposition_RedessinFenetre(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  ModeleAjustement *modele;
  Graphe *graphe;
  struct decomposition *decomposition;
  struct element_base *base;

  double echelle_X, echelle_Y, x_min, x_max, y_min, y_max;
  int x_barre;
  unsigned long i;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  /* On récupère les infos. */
  graphe = (Graphe *)           FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  modele = (ModeleAjustement *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( ( NULL == graphe ) || ( NULL == modele ) ) return;

  decomposition = modele->infos_modele.decomposition;

  /* On calcule l'échelle globale - pour avoir une meilleure superposition */
  x_min = graphe->x_min;  x_max = graphe->x_max;
  y_min = graphe->y_min;  y_max = graphe->y_max;

  if ( decomposition != NULL ) {
    for ( i = 0, base = decomposition->base ; i < decomposition->nombre_bases ; i++, base++ ) {
      if ( NULL == base->graphe ) continue;

      if ( base->graphe->x_min < x_min) x_min = base->graphe->x_min;
      if ( base->graphe->x_max > x_max) x_max = base->graphe->x_max;
      if ( base->graphe->y_min < y_min) y_min = base->graphe->y_min;
      if ( base->graphe->y_max > y_max) y_max = base->graphe->y_max;
    }
  }

  TRC_Echelles( fenetre->w_larg, fenetre->w_haut, x_min, x_max, y_min, y_max, &echelle_X, &echelle_Y );

  /* On trace le graphe que l'on traite */
  TRC_Graphe( fenetre, graphe, NOIR, 1, 0, 0, 0, fenetre->w_haut - 1,
	      x_min, x_max, y_min, y_max, echelle_X, echelle_Y );

  /* On met les deux index de l'intervalle utilisé */
  x_barre = TRC_ConvertitX_PvE( modele->x_min, x_min, 0, echelle_X );
  TRC_Couleur( ROUGE ); /* Barre rouge : borne basse de l'intervalle pris en compte */
  TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );

  x_barre = TRC_ConvertitX_PvE( modele->x_max, x_min, 0, echelle_X );
  TRC_Couleur( VERT ); /* Barre verte : borne haute de l'intervalle pris en compte */
  TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );

  /* On trace les spectres de la base - en pointillé, couleur incrémentée */
  if ( decomposition != NULL ) {
    unsigned int couleur = NOIR;

    for ( i = 0, base = decomposition->base ; i < decomposition->nombre_bases ; i++, base++ ) {
      if ( NULL == base->graphe ) continue;

      TRC_Graphe( fenetre, base->graphe, couleur, 2, 0, 0, 0, fenetre->w_haut - 1,
		  x_min, x_max, y_min, y_max, echelle_X, echelle_Y );
      couleur++;
    }
  }

  /* On trace le graphe combinaison linéaire des deux spectres de base - en rouge */
  if ( modele->modele != NULL ) {
    TRC_Graphe( fenetre, modele->modele, ROUGE, 1, 0, 0, 0, fenetre->w_haut - 1,
		x_min, x_max, y_min, y_max, echelle_X, echelle_Y );
  }

  /* On trace le résidu - en bleu */
  if ( modele->residu != NULL ) {
    TRC_Graphe( fenetre, modele->residu, BLEU, 1, 0, 0, 0, fenetre->w_haut - 1,
		x_min, x_max, y_min, y_max, echelle_X, echelle_Y );
  }
}

/* -------- Rappeler la fenêtre si elle est déjà ouverte -------- */

static Fenetre *I_GRF_Decomposition_RappelerFenetre(Graphe *source)
{
  int reponse;

  ModeleAjustement *modele = NULL;
  Fenetre *fenetre;
  Graphe *graphe;

  /* Quelle est la fenêtre associée à l'arbre ? */
  fenetre = ARB_QuelleFenetre( &(arbres[ DECOMP_GRAPHE ]), FAUX );

  /* Quelles sont les informations associées à cette fenêtre ? */
  graphe = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  modele = (ModeleAjustement *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  /* Est-ce que le graphe est le même ? */
  if ( graphe == source ) {
    /* On travaille sur le même graphe : on se contente de faire repasser au premier plan */
    FEN_premierplan( fenetre );
    return NULL;
  }

  /* On pose la question : faut-il rappeler,... ? */
  reponse = BI_ERR_alerte( fenetre, &Q_ReactiverDecomposition, VRAI, graphe->nom, source->nom );
  switch ( reponse )
  {
  case I_GRF_DCMP_RAPPELER:
    FEN_premierplan( fenetre );
    return NULL;

  case I_GRF_DCMP_RELANCER:
    /* On termine les calculs pour l'ancien graphe : ferme la fenêtre et libère tout */

    /* On dissocie les infos de la fenêtre, pour éviter des soucis... */
    FEN_GRF_LierGraphe( fenetre, NULL );
    FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

    /* On efface les graphes inutiles */
    GRF_efface( modele->residu, VRAI, VRAI );
    GRF_efface( modele->modele, VRAI, VRAI );
    GRF_DeprotegerGraphe( graphe );

    /* On rappelera donc une nouvelle fenêtre propre... */
    FEN_premierplan( fenetre );
    return fenetre;

  case I_GRF_DCMP_ANNULER:
    /* On ne fait rien de spécial */
    return NULL;

  default:
    (void) fprintf( stderr, "%s,%s l.%u : Erreur en posant la question, ou réponse inconnue [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, reponse );
    return fenetre;
  }
}

/* ~-~-~-~-~-~-~-~-~ Aides pour la gestion de l'arbre ~-~-~-~-~-~-~-~-~ */

/* -------- Gestion de l'arbre proprement dite -------- */

static int I_GRF_Decomposition_GererArbre(unsigned int num_objet, int x, int y)
{
  Fenetre *fenetre;
  Graphe  *spectre;
  ModeleAjustement *modele;
  int reponse;
  struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };

  fenetre = ARB_QuelleFenetre( arbres + DECOMP_GRAPHE, FAUX ); /* Fenêtre actuelle */
  spectre = (Graphe *)     FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  modele  = (ModeleAjustement *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  /* Gestion de l'ascenseur */
  if ( ( DG_GRF_AVANT  == num_objet ) || ( DG_FOND_ASC  == num_objet ) ||
       ( DG_GRF_BOUTON == num_objet ) || ( DG_GRF_APRES == num_objet ) ) {
    unsigned long nouveau_vecteur;

    nouveau_vecteur = ARB_ascenseur( DECOMP_GRAPHE, DG_GRF_AVANT, DG_FOND_ASC, DG_GRF_BOUTON, DG_GRF_APRES, ASC_VERTICAL,
				     modele->infos_modele.decomposition->nombre_bases, 1, i_vecteur_actif,
				     num_objet, x, y, ECHANGE );
    if ( nouveau_vecteur != i_vecteur_actif ) {
      i_vecteur_actif = nouveau_vecteur;
      I_GRF_Decomposition_MettreVecteurBase( modele->infos_modele.decomposition->base + i_vecteur_actif );
    }

    return BI_EVENEMENT_TRAITE;
  }

  switch( num_objet )
  {
  case DG_XMIN:
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "Xmin =";
    nv_valeur.min = spectre->x_min;
    nv_valeur.max = modele->x_max;
    nv_valeur.valeur_initiale.reel = modele->x_min;

    reponse = BI_DemandeValeur( fenetre, &Q_ValeurXmin, &nv_valeur );
    if ( Q_ValeurXmin.bouton_annuler == (unsigned int) reponse ) return BI_EVENEMENT_TRAITE; /* On annule */

    modele->x_min = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( DECOMP_GRAPHE, DG_XMIN, modele->x_min, 0.0 );

    /* On fait le calcul [se charge d'afficher...] */
    I_GRF_Decomposition_LancerAjustement( modele, FAUX );
    return BI_EVENEMENT_TRAITE;

  case DG_XMAX:
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "Xmax =";
    nv_valeur.min = modele->x_min;
    nv_valeur.max = spectre->x_max;
    nv_valeur.valeur_initiale.reel = modele->x_max;

    reponse = BI_DemandeValeur( fenetre, &Q_ValeurXmax, &nv_valeur );
    if ( Q_ValeurXmax.bouton_annuler == (unsigned int) reponse ) return BI_EVENEMENT_TRAITE; /* On annule */

    modele->x_max = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( DECOMP_GRAPHE, DG_XMAX, modele->x_max, 0.0 );

    /* On fait le calcul [se charge d'afficher...] */
    I_GRF_Decomposition_LancerAjustement( modele, FAUX );
    return BI_EVENEMENT_TRAITE;

  case DG_GRAPHE:		/* On veut changer le graphe de référence */
    {
      struct graphe *nv_graphe;

      nv_graphe = GRF_choisir( modele->experience->type, FAUX, VRAI, NULL );
      if ( ( nv_graphe != modele->experience ) &&
	   ( nv_graphe != modele->infos_modele.decomposition->base[ i_vecteur_actif ].graphe ) ) {
	int retour;

	/* On mémorise ce nouveau graphe de référence */
	retour = MODELE_Decomposition_RemplacerVecteurDeBase( modele->infos_modele.decomposition, i_vecteur_actif, nv_graphe );
	if ( PAS_DE_PROBLEME == retour ) {
	  /* On relance l'ajustement... */
	  I_GRF_Decomposition_LancerAjustement( modele, VRAI );
	}
      }
    }
    return BI_EVENEMENT_TRAITE;

  case DG_AJOUTER_GRF:		/* Ajouter un graphe à ceux déjà utilisés */
    {
      struct graphe *nv_graphe;

      nv_graphe = GRF_choisir( modele->experience->type, FAUX, VRAI, NULL );
      if ( ( nv_graphe != modele->experience ) &&
	   ( nv_graphe != modele->infos_modele.decomposition->base[ i_vecteur_actif ].graphe ) ) {
	int retour;

	/* On mémorise ce nouveau graphe de référence */
	retour = MODELE_Decomposition_AjouterVecteurDeBase( modele->infos_modele.decomposition, nv_graphe );
	if ( PAS_DE_PROBLEME == retour ) {
	  /* On relance l'ajustement... */
	  I_GRF_Decomposition_LancerAjustement( modele, VRAI );
	}
      }
    }
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Changer la méthode d'ajustement -------- */

static void I_GRF_Decomposition_ChangerDeMethode(unsigned int numero_arbre, unsigned int numero_objet __attribute__((__unused__)), unsigned long option_choisie)
{
  DCMP_infos *infos_decomposition;
  Fenetre *fenetre;
  ModeleAjustement *modele;

  /* On récupère le modèle associé à l'arbre */
  fenetre = ARB_QuelleFenetre( arbres + numero_arbre, VRAI );
  modele  = (ModeleAjustement *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == modele ) {
    (void) fprintf( stderr, "%s, %s l.%u : Appel sans modèle associé à l'arbre !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  infos_decomposition = modele->infos_modele.decomposition;

  /* A-t-on changé de méthode ? */
  if ( option_choisie == infos_decomposition->methode ) {
    /* Rien à faire ! */
    return;
  }

  /* On fait le changement */
  infos_decomposition->methode = option_choisie;
  I_GRF_Decomposition_LancerAjustement( modele, VRAI );

  /* On met le texte dans l'objet [nécessaire ?] */
}

/* -------- Afficher les informations pour le modèle indiqué -------- */

static void I_GRF_Decomposition_MettreModele(ModeleAjustement *modele)
{
  DCMP_infos *infos_decomposition;

  if ( NULL == modele ) {
    OBJ_MettreIntervalle( DECOMP_GRAPHE, DG_XMIN, DG_XMAX, 0, 0, 0.0, 0.0, 0.0, 0 );
    OBJ_MettreEntier_U( DECOMP_GRAPHE, DG_NBR_BASE, 0 );
    return;
  }
  infos_decomposition = modele->infos_modele.decomposition;

  /* Bornes du modèle */
  OBJ_MettreIntervalle( DECOMP_GRAPHE, DG_XMIN, DG_XMAX, 0, 0, modele->x_min, modele->x_max, 0.0, 0 );

  /* Combien de graphes forment la base de décomposition ? */
  OBJ_MettreEntier_U( DECOMP_GRAPHE, DG_NBR_BASE, infos_decomposition->nombre_bases );

  /* Affichage de l'ascenseur, en fonction du nombre de graphes de base */
  ARB_ascenseur( DECOMP_GRAPHE, DG_GRF_AVANT, DG_FOND_ASC, DG_GRF_BOUTON, DG_GRF_APRES, ASC_VERTICAL,
		 infos_decomposition->nombre_bases, 1, 0, ROOT, 0, 0, METTRE );

  /* Affichage du graphe de base concerné */
  if ( 0 == infos_decomposition->nombre_bases ) {
  } else {
    struct element_base *base;

    /* On récupère les informations sur le vecteur de base indiqué */
    if ( i_vecteur_actif >= infos_decomposition->nombre_bases ) {
      (void) fprintf( stderr, "%s,%s l.%u : Index de graphe de base incorrect [%lu, max %lu]\n",
		      __FILE__, __FUNCTION__, __LINE__, i_vecteur_actif, infos_decomposition->nombre_bases - 1 );
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      i_vecteur_actif = 0;
    }
    base = infos_decomposition->base + i_vecteur_actif;

    /* On les affiche */
    I_GRF_Decomposition_MettreVecteurBase( base );
  }
}

/* -------- Afficher les informations au vecteur de base [graphe] indiqué -------- */

static void I_GRF_Decomposition_MettreVecteurBase(struct element_base *vecteur_base)
{
  if ( NULL == vecteur_base ) {
    /* Ne devrait jamais arriver... */

    OBJ_MettreNombre( DECOMP_GRAPHE, DG_COEFFICIENT, 0.0, -1.0 );
    OBJ_CopierTexte( DECOMP_GRAPHE, DG_GRAPHE, composante_inexistante, LG_MAX_NOM_SPECTRE );
    return;
  }

  /* Le coefficient obtenu */
  OBJ_MettreNombre( DECOMP_GRAPHE, DG_COEFFICIENT,
		    VAR_LireValeur( &(vecteur_base->coefficient), VARIABLE_LIRE_VALEUR_AJUSTEE ), vecteur_base->erreur );

  /* Le nom du graphe correspondant */
  if ( vecteur_base->graphe != NULL ) {
    OBJ_CopierTexte( DECOMP_GRAPHE, DG_GRAPHE, vecteur_base->graphe->nom, LG_MAX_NOM_SPECTRE );
  } else {
    OBJ_CopierTexte( DECOMP_GRAPHE, DG_GRAPHE, aucun_graphe, LG_MAX_NOM_SPECTRE );
  }
}

/* -------- Préparation de l'arbre de la fenêtre de décomposition -------- */

static int I_GRF_Decomposition_PreparerArbre( unsigned int numero_arbre )
{
  /* On convertit en sous-menu ce qu'il faut */
  OBJ_ChangerType( numero_arbre, DCMP_METHODE, G_SSMENU );
  OBJ_MettreMenu ( numero_arbre, DCMP_METHODE, MODELE_DECOMPOSITION_NOMBRE_DE_METHODES, txt_methodes,
		   titre_menu_methode, I_GRF_Decomposition_ChangerDeMethode );

  /* Autres changements de types */
  OBJ_ChangerType( numero_arbre, DG_NBR_BASE, G_NOMBRE );

  if ( FAUX == arbres[ numero_arbre ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( numero_arbre, NULL, I_GRF_Decomposition_GererArbre, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return -1;
    }
  }

  /* On agrandit le fond pour que ce soit plus beau */
  OBJ_agrandit( numero_arbre, ROOT, 0, 1);

  /* On repositionne les objets */
  OBJ_deplace( numero_arbre, DG_TXT_XMIN, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_XMIN    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_TXT_XMAX, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_XMAX    , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_AVEC_ERREUR, 0, +4, DEPLACE_PIXELS );

  OBJ_deplace( numero_arbre, DG_TXT_COEFF   , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_COEFFICIENT , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_UTILISER    , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_TXT_UTILISER, 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_TXT_GRAPHE  , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, DG_GRAPHE      , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( numero_arbre, DG_TXT_BASE_1, 0, +500, DEPLACE_RELATIF );
  OBJ_deplace( numero_arbre, DG_NBR_BASE  , 0, +500, DEPLACE_RELATIF );
  OBJ_deplace( numero_arbre, DG_TXT_BASE_2, 0, +500, DEPLACE_RELATIF );

  OBJ_deplace( numero_arbre, DG_AJOUTER_GRF, 0, +500, DEPLACE_RELATIF );

  ARB_EnjoliverAscenseur( numero_arbre, DG_GRF_AVANT, DG_FOND_ASC, DG_GRF_BOUTON, DG_GRF_APRES, ASC_VERTICAL, 0 );

  /*  OBJ_deplace(numero_arbre, , 0, 2, DEPLACE_PIXELS); */

#ifndef _DEVELOPPE_
  OBJ_active( numero_arbre, DG_AVEC_ERREUR, FORCE_NON );
#endif

  /* On s'occupe des traductions évenuelles */
#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( numero_arbre, DG_AVEC_ERREUR , txt_avec_erreur );
  OBJ_TraduireTexte( numero_arbre, DG_TXT_COEFF   , txt_coefficient );
  OBJ_TraduireTexte( numero_arbre, DG_TXT_GRAPHE  , txt_graphe );
  OBJ_TraduireTexte( numero_arbre, DG_TXT_UTILISER, txt_utiliser );
  OBJ_TraduireTexte( numero_arbre, DG_TXT_BASE_1  , txt_base_1 );
  OBJ_TraduireTexte( numero_arbre, DG_TXT_BASE_2  , txt_base_2 );
#endif
  OBJ_TraduireTexte( numero_arbre, DG_OTER_BASE   , txt_oter_base ); /* Ô, infaisable dans le RSC... */

  arbres[ numero_arbre ].drapeaux.prepare = VRAI;

  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Aides pour la gestion des infos de décomposition ~-~-~-~-~-~-~-~-~ */

struct modele_ajuste *I_ModeleDecomposition_ChoisirCourant(struct graphe *a_decomposer)
{
  int reponse = I_GRF_DCMP_APPEL_NV_MODELE;
  struct modele_ajuste *modele;

  if ( NULL == a_decomposer ) return NULL;

  /* On cherche si un modèle contiendrait déjà ce graphe source */
  for ( modele = premier_modele ; modele != NULL ; modele = modele->suivant ) {
    if ( a_decomposer == modele->experience ) {
      /* Modèle trouvé : on le renvoie tel quel... */
      return modele;
    }
  }

  /* Aucun modèle ne contient l'expérience */
  if ( premier_modele != NULL ) {
    /* Cela vaut la peine de poser la question... */
    reponse = BI_ERR_alerte( NULL, &Q_GRF_DCMP_NouveauModele, VRAI );
    if ( I_GRF_DCMP_APPEL_ANNULER == reponse ) {
      /* Annulation : pas de modèle trouvé... */
      return NULL;
    }

    if ( I_GRF_DCMP_APPEL_REPRISE == reponse ) {
      OPTION_A_FAIRE( "Proposer de choisir un modèle" );
    }
  }

  /* Si on est là, on veut créer un nouveau modèle, de type décomposition */
  modele = I_ModeleDecomposition_Nouveau( 0, a_decomposer );
  return modele;
}

struct modele_ajuste *I_ModeleDecomposition_Premier(void)
{
  return premier_modele;
}

/* -------- Libérer les informations de la décomposition -------- */

static void I_GRF_Decomposition_LibererInfos(ModeleAjustement *infos)
{
  if ( NULL == infos ) return;

#ifdef __ancien__
  /* On libère le contenu */
  free(infos->a_1); free(infos->b_1); free(infos->c_1);
  free(infos->a_2); free(infos->b_2); free(infos->c_2);

  /* On efface les graphes associés */
  GRF_eff_isole( infos->residu ); infos->residu = NULL;
  GRF_eff_isole( infos->c_lineaire ); infos->c_lineaire = NULL;

  /* On efface la structure elle-même */
  free( infos ); infos = NULL;
#endif
}

/* ----------------- Gestion d'un clic souris (décomposition d'un spectre) ------------ */

static int I_GRF_Decomposition_ClicFenetre(Fenetre *fenetre, int bouton, int n_clic __attribute__((__unused__)),
					   int etat __attribute__((__unused__)), int x, int y __attribute__((__unused__)))
{
  ModeleAjustement *modele;
  Graphe *graphe;
  struct decomposition *decomposition;

  double echelle_X, echelle_Y, x_min, x_max, y_min, y_max, x_reel;

#ifdef _TOS_
  int handle;

  wind_get(fenetre->handle, WF_TOP, &handle);
  if (n_clic != fenetre->handle) {
    wind_set(fenetre->handle, WF_TOP); /* Fenêtre non active : on l'active donc */
    return BI_EVENEMENT_TRAITE;
  }
#endif

  /* On récupère les infos. */
  graphe = (Graphe *)           FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  modele = (ModeleAjustement *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( ( NULL == graphe ) || ( NULL == modele ) ) return BI_EVENEMENT_TRAITE;

  decomposition = modele->infos_modele.decomposition;

  /* On calcule l'échelle globale - pour avoir une meilleure superposition */
  x_min = graphe->x_min;  x_max = graphe->x_max;
  y_min = graphe->y_min;  y_max = graphe->y_max;

  if ( decomposition != NULL ) {
    struct element_base *base;
    unsigned long i;

    for ( i = 0, base = decomposition->base ; i < decomposition->nombre_bases ; i++, base++ ) {
      if ( NULL == base->graphe ) continue;

      if ( base->graphe->x_min < x_min) x_min = base->graphe->x_min;
      if ( base->graphe->x_max > x_max) x_max = base->graphe->x_max;
      if ( base->graphe->y_min < y_min) y_min = base->graphe->y_min;
      if ( base->graphe->y_max > y_max) y_max = base->graphe->y_max;
    }
  }

  TRC_Echelles( fenetre->w_larg, fenetre->w_haut, x_min, x_max, y_min, y_max, &echelle_X, &echelle_Y );
  if ( 0.0 == echelle_X ) return BI_EVENEMENT_TRAITE;

  x_reel = TRC_ConvertitX_EvP( x, x_min, 0, echelle_X );
  if ( 1 == bouton ) { /* Clic gauche : on met la borne inférieure */
    modele->x_min = x_reel;
  } else if ( 2 == bouton ) { /* Clic droit : on met la borne supérieure */
    modele->x_max = x_reel;
  } else {
    return BI_EVENEMENT_TRAITE;
  }

  if ( modele->x_max < modele->x_min ) {
    double x = modele->x_max; modele->x_max = modele->x_min; modele->x_min = x;
  }
  OBJ_MettreIntervalle( DECOMP_GRAPHE, DG_XMIN, DG_XMAX, 0, 0, modele->x_min, modele->x_max, 0.0, 0 );

  /* On fait le calcul [se charge de redessiner] */
  I_GRF_Decomposition_LancerAjustement( modele, FAUX );

  return BI_EVENEMENT_TRAITE;
}

/* -------- Gestion du clavier sur la fenêtre de décomposition -------- */

static int CLV_decompose(Fenetre *fenetre, unsigned int touche, int etat)
{
  int retour;
  ModeleAjustement *modele;
  Graphe *spectre __attribute__((__unused__));
  Message message;

  if ( etat & K_CTRL ) { /* Aucun raccourci avec Ctrl ! */
    return BI_EVENEMENT_NON_TRAITE;
  }

  spectre = (Graphe *)     FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  modele   = (ModeleAjustement *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == modele ) return BI_EVENEMENT_NON_TRAITE;

  switch ( touche )
  {
#ifdef __ancien__
  case 'c':
/*     infos->methode = MODELE_DECOMPOSITION_MOINDRES_CARRES; */
    met_chaine(DECOMP_GRAPHE, DCMP_METHODE, LG_MAX_NOM_METHODE,
	       txt_methodes[DCMP_METH_MDR_CARR], TXT_COPIER);
/*     GRF_DecomposerSurBase(spectre, infos); */
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case 'l':
    infos->methode = DCMP_METH_LINEAIRE;
    met_chaine(DECOMP_GRAPHE, DCMP_METHODE, LG_MAX_NOM_METHODE,
	       txt_methodes[DCMP_METH_LINEAIRE], TXT_COPIER);
    GRF_DecomposerSurBase(spectre, infos);
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case 'm':
    infos->methode = DCMP_METH_MOYENNE;
    met_chaine(DECOMP_GRAPHE, DCMP_METHODE, LG_MAX_NOM_METHODE,
	       txt_methodes[DCMP_METH_MOYENNE], TXT_COPIER);
    GRF_DecomposerSurBase(spectre, infos);
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case 's':
    if (infos->methode == DCMP_METH_SUIVI) /* On a déjà fait ce qu'il fallait... */
      return BI_EVENEMENT_TRAITE;

    infos->methode = DCMP_METH_SUIVI;
    met_chaine(DECOMP_GRAPHE, DCMP_METHODE, LG_MAX_NOM_METHODE,
	       txt_methodes[DCMP_METH_SUIVI], TXT_COPIER);
    GRF_DecomposerSurBase(spectre, infos);
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case 'u':         /* moindres carrés avec a+b=1 */
    infos->methode = DCMP_METH_MDR_CARR_SOMME1;
    met_chaine(DECOMP_GRAPHE, DCMP_METHODE, LG_MAX_NOM_METHODE,
	       txt_methodes[DCMP_METH_MDR_CARR_SOMME1], TXT_COPIER);
    GRF_DecomposerSurBase(spectre, infos);
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case 'v':
    infos->methode = DCMP_METH_VARIANCE;
    met_chaine(DECOMP_GRAPHE, DCMP_METHODE, LG_MAX_NOM_METHODE,
	       txt_methodes[DCMP_METH_VARIANCE], TXT_COPIER);
    GRF_DecomposerSurBase(spectre, infos);
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;
#endif

  case 'x':
    retour = ARB_SimulerClic( DECOMP_GRAPHE, DG_XMIN, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) {
      return retour;
    }
    return BI_EVENEMENT_TRAITE;

  case 'X':
    retour = ARB_SimulerClic( DECOMP_GRAPHE, DG_XMAX, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) {
      return retour;
    }
    return BI_EVENEMENT_TRAITE;

  case ' ': /* Fermer la fenêtre */
    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = 0; message.fenetre.y = 0;
    message.fenetre.larg = 0; message.fenetre.haut = 0; /* Dans le cas où != 0, fermeture forcée... */
    retour = (fenetre->message)(fenetre, &message);
    if ( retour != BI_EVENEMENT_NON_TRAITE ) {
      return retour;
    }
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE; /* Touche non traitée -> voir ailleurs ! */
}

/* Sauvegarde de l'historique */

static void I_GRF_Decomposition_Historique(Graphe *source, ModeleAjustement *modele)
{
  HIST_Titre( 1, "D"e_AIGU"composition lin"e_AIGU"aire d'un graphe", "DCMP" );
  HIST_Texte( 1, 0, "Graphe d"e_AIGU"compos"e_AIGU" : %s\n", source->nom );
  HIST_Texte( 1, 0, "Combinaison obtenue : %s\n", modele->modele->nom );

  HIST_Texte( 2, 0, "R"e_AIGU"sidu : %s\n", modele->residu->nom );
#ifdef __ancien___
  HIST_Texte( 2, 0, "Premier spectre de base : %s\n", infos->ref_1->nom );

  HIST_Texte( 3, 2, "Coefficient associ"e_AIGU" : " );
  if ( DCMP_METH_SUIVI == infos->methode ) { /* Suivi du coefficient */
    HIST_Texte( 3, 0, "graphe %s\n", infos->alpha->nom );
  } else if ( DCMP_METH_LINEAIRE == infos->methode ) { /* Évolution linéaire du coefficient */
    HIST_Texte( 3, 0, "%.10g * X + %.10g\n", infos->coef_1, infos->coef_2 );
  } else {
    HIST_Texte( 3, 0, "%.10g - m"e_AIGU"thode : %s\n", infos->coef_1, txt_methodes[ infos->methode ] );
  }

  HIST_Texte( 2, 0, "Second spectre de base : %s\n", infos->ref_2->nom );

  HIST_Texte( 3, 2, "Coefficient associ"e_AIGU" : " );
  if ( DCMP_METH_SUIVI == infos->methode ) { /* Suivi du coefficient */
    HIST_Texte( 3, 0, "graphe %s\n", infos->beta->nom );
  } else if ( DCMP_METH_LINEAIRE == infos->methode ) { /* Évolution linéaire du coefficient */
    HIST_Texte( 3, 0, "%.10g * X + %.10g\n", -infos->coef_1, 1.0 - infos->coef_2 );
  } else {
    HIST_Texte( 3, 0, "%.10g - m"e_AIGU"thode : %s\n", infos->coef_2, txt_methodes[ infos->methode ] );
  }
#endif
}
