/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996-2016 Emmanuel CURIS
   Création : 18 décembre 2016 par reprise de vue_3d.c

   Affichage 3d d'une molécule à l'aide d'Open-GL
   Ne fonctionne que sous X-Window dans l'état actuel des choses

   -------- Historique --------
   25 décembre 2016 : déplacé la gestion des paramètres de la vue 3D
                      dans son propre fichier (3d_changer_vue.c)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "graphes.h"
#include "exafs.h"

#include "atomes.h"
#include "chemins.h"
#include "couches.h"
#include "cristallo.h"
#include "feff.h" /* Pour ajout au modèle FEFF */
#include "fichiers.h" /* Pour export POV */
#include "interface.h"
#include "modeles.h"
#include "science.h"  /* Pour M_PI, si pas défini */
#include "tracer.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

static unsigned int couche_active = 0;

#define OPT_3D_DISTANCE         0
#define OPT_3D_DIST_CHAINE      1
#define OPT_3D_DIST_RAYON       2
#define OPT_3D_ANGLE            3
#define OPT_3D_ANGLE_CHAINE     4
#define OPT_3D_CHOISIR          6 /* Séparateur jute avant */
#define OPT_3D_EFFACER          7
#define OPT_3D_AJOUTE_FEFF      9 /* Séparateur juste avant */
#define OPT_3D_EFFACE_FEFF     10
#define OPT_3D_EXPORT_POV      12 /* Séparateur juste avant */
#define OPT_3D_IMAGE           13
#define OPT_3D_DIST_ABS        15 /* Séparateur juste avant */
#define OPT_3D_CREER_MAILLE    17 /* Séparateur juste avant */
#define OPT_3D_EFFACER_MAILLE  18
#define OPT_3D_ISOLER_MOLECULE 19
#define OPT_3D_LIER            21 /* Séparateur juste avant */
#define OPT_3D_LIER_CHAINE     22
#define OPT_3D_LIER_RADIAL     23
#define OPT_3D_OTER_LIAISON    24

#define NOMBRE_OPTIONS_VUE_3D  25

/* Les options de visualisation des atomes */
#define TXT_VOIR_ATOMES         0
#define TXT_VOIR_ATOMES_REDUITS 1
#define TXT_NE_PAS_VOIR_ATOMES  2

#define NOMBRE_MODE_VISION_ATOMES 3

/* Les options de visualisation des liaisons */

#define TXT_VOIR_LIAISON_SIMPLE    0
#define TXT_VOIR_LIAISON_MULTIPLE  1
#define TXT_NE_PAS_VOIR_LIAISONS   2

#define NOMBRE_MODE_VISION_LIAISONS 3

/* Les modes de sélection d'atomes */
#define MODE_NORMAL   0 /* Sélection d'un atome sans plus */
#define MODE_DISTANCE 1 /* Sélection d'un second atome puis calcul de la distance */
#define MODE_ANGLE    2 /* Sélection d'un 2e et d'un 3e atome puis calcul de l'angle */

static short int mode_fonction = MODE_NORMAL; /* Mode de commandes pour la vue 3D */
static int commande_clic = OPT_3D_CHOISIR; /* Commande à utiliser */
static Atome *atm_precedent = NULL, *atome_3 = NULL;

#include "int_principale_molecule.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void V3D_AfficherChemin(void);

/* ---- Modification de la vue en 3D  */

static void ASC_atomes_3d(Fenetre *fenetre, int type, int ss_type, int position);
static void RDS_atomes_3d(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);
static void RDS_graphes  (Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);
static int  MSG_atomes_3d(Fenetre *fenetre, Message *message);
static int  CLC_atomes_3d(Fenetre *fenetre, int bouton, int n_clic, int etat, int x, int y);
static int  CLV_atomes_3d(Fenetre *fenetre, unsigned int touche, int etat);
static int  GER_vue_3d   (unsigned int objet, int x, int y);

static int I_V3D_PreparerArbre(unsigned int numero_arbre);

static boolean params_affichage(void);
static int GER_params_affichage(unsigned int objet, int x, int y);
static void prepare_params_affichage(void);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

void I_MOL_Visualisation3D(void)
{
  unsigned long nbr_molecules = 0, num_molecule = 0;
  Molecule *molecule = NULL;

  if ( NULL == premiere_molecule ) return; /* Aucune molécule à afficher… */
  if ( NULL == mol_courante ) {
    AVERTISSEMENT( "Pas de molécule courante [1re = %p]", premiere_molecule );
    /* Il y a des molécules, mais pas de courante : on y va donc... */
    MOL_aller( premiere_molecule );
  }

  if ( fen_vue_3d != NULL ) {
    /* Fenêtre déjà créée ! */
    FEN_premierplan( fen_vue_3d );
    return;
  }

  /* On crée la fenêtre principale
     (appelera la routine de préparation de l'arbre, si nécessaire) */
  fen_vue_3d = FEN_CreerFenetreTriple( TOUS_SF_ASC|ASC_HORIZONTAL, mol_courante->nom,
				       AFF_STRUCTURE, BI_FEN_ARBRE_COIN_HG, I_V3D_PreparerArbre,
				       BI_FEN_ARBRE_DROITE,
				       RDS_graphes, RDS_atomes_3d, ASC_atomes_3d, CLV_atomes_3d,
				       ( int  (*)(Fenetre *, int, int, int, int, int) ) NULL, MSG_atomes_3d,
				       ( void (*)(Fenetre *, int, int, int, int) ) NULL,
				       ( int  (*)(Fenetre *) ) NULL);
  if ( NULL == fen_vue_3d ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de créer la fenêtre triple\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return;
  }
  FEN_infos( fen_vue_3d, txt_infos_aucun_atome_choisi  );
  fen_vue_3d->msg_masque.fermeture = VRAI; /* Messages traités */
  FEN_GL_evntsgl( fen_vue_3d,
		  ( void (*)(Fenetre *, int, int, unsigned int, unsigned int) ) NULL, /* Pas de dessin post-OpenGL actuellement */
		  ( void (*)(Fenetre *, int, int, int, int) ) NULL, /* Pas de pointeur */
		  CLC_atomes_3d,
		  ( int (*)(Fenetre *, int, int, int, int) ) NULL,  /* Pas de déplacement actuellement */
		  ( int (*)(Fenetre *) ) NULL );  /* Pas d'animation actuellement */

  /* On ouvre la fenêtre ! */
  FEN_Ouvrir( fen_vue_3d );

#ifdef AVEC_OPEN_GL_
  if ( VRAI == BI_OpenGL_actif ) { /* OpenGL est bien présent */
    glClearColor( 1.0, 1.0, 1.0, 0.0 ); /* Fond blanc */
    glColorMaterial( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
    glEnable( GL_LIGHTING );       /* On active la lumière pour avoir de la vraie 3D      */
    glEnable( GL_LIGHT0 );         /* Une seule lampe, qui sera à la position de l'oeil   */
    glEnable( GL_DEPTH_TEST );     /* On effectue les tests de profondeur (faces cachées) */
    glEnable( GL_COLOR_MATERIAL ); /* glColor change la couleur du matériau -> 3D colorée */
    glEnable( GL_CULL_FACE );      /* On autorise l'élimination des faces inutiles        */

    liste_molecule = glGenLists( 3 ); /* On génère les listes d'affichage Open/GL : molécule + liaisons */
    if ( 0 == liste_molecule ) {
      (void) fprintf( stderr, "%s,%s l.%u : Erreur Open/GL : plus de liste disponible\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      liste_liaisons = 0;
      liste_chemin = 0;
    } else {
      liste_liaisons = liste_molecule + 1;
      liste_chemin   = liste_molecule + 2;

      /* On prépare l'affichage de la molécule courante */
      glNewList( liste_molecule, GL_COMPILE);
      dessine_molecule( mol_courante, mode_vue3d );
      glEndList(); /* On a fini le dessin */
#ifdef _DEVELOPPE_
/*      printf("Erreur OpenGL/fin de la liste : %s\n", gluErrorString( glGetError())); */
#endif
      /* Affichage des liaisons ! */
      glNewList( liste_liaisons, GL_COMPILE );
      dessine_liaisons( mol_courante, mode_vue3d );
      glEndList(); /* On a fini le dessin */    
    }
  } else {
    liste_molecule = 0; /* Pas de liste Open-GL, donc ! */
    liste_liaisons = 0;
    liste_chemin = 0;
  }
#endif

  /* S'occuper de l'ascenseur en fonction du nombre de molécules en mémoire */
  for ( molecule = premiere_molecule ; molecule ; molecule = molecule->suivante ) {
    nbr_molecules++;
    if ( molecule == mol_courante ) num_molecule = nbr_molecules;
  }

  if ( nbr_molecules < 2 ) { /* Une seule molécule (voire aucune) */
    FEN_ascenseur( fen_vue_3d, ASC_HORIZONTAL, 0, 1000 );
  } else {
    float position, taille;

    taille   = 1000.0 / ((float) nbr_molecules);
    position = 1000.0 * ((float) (num_molecule - 1)) / ((float) nbr_molecules - 1);
    FEN_ascenseur( fen_vue_3d, ASC_HORIZONTAL, (int) position, (int) taille );
  }
}

/* -------- Gestion de l'ascenseur : passage d'une molécule à l'autre -------- */

static void ASC_atomes_3d(Fenetre *fenetre, int type, int ss_type, int position)
{
  Molecule *nv_molecule = NULL;

#ifdef _DEVELOPPE_
  printf( "ASC_atomes_3d : fenetre = %p, type = %d, ss_type = %d, position = %d\n",
	  (void *) fenetre, type, ss_type, position );
#endif

  if ( WM_ARROWED == type ) {
    switch ( ss_type )
    {
    case WA_LFPAGE: /* 10 molécules précédentes */
      if ( mol_courante != premiere_molecule ) { /* Il y en a avant */
	Molecule *mol;
	unsigned int i = 0;

	nv_molecule = premiere_molecule;
	mol = nv_molecule;
	while ( i < 10 ) {
	  mol = mol->suivante;
	  if ( ( NULL == mol ) ||
	       ( mol == mol_courante ) ) { /* Il n'y a pas 10 molécules, ou on est avant la 10e */
	    mol = NULL;
	    break;
	  }
	  i++;
	}

	if ( mol != NULL ) { /* Il faut continuer le décalage */
	  while ( mol->suivante != mol_courante ) {
	    mol = mol->suivante;
	    nv_molecule = nv_molecule->suivante;
	  }
	}
      }
      break;

    case WA_RTPAGE: /* 10 molécules suivantes */
      if ( mol_courante->suivante != NULL ) { /* On n'est pas sur la dernière */
	unsigned int i = 0;

	for ( nv_molecule = mol_courante ;
	      ( i < 10 ) && ( nv_molecule->suivante != NULL ) ;
	      i++, nv_molecule = nv_molecule->suivante );
      }
      break;

    case WA_LFLINE: /* Molécule précédente */
      if ( mol_courante != premiere_molecule ) { /* Il y en a avant */
	nv_molecule = premiere_molecule;
	while ( nv_molecule->suivante != mol_courante )
	  nv_molecule = nv_molecule->suivante;
      }
      break;

    case WA_RTLINE: /* Molécule suivante */
      nv_molecule = mol_courante->suivante;
      break;
    }
  } else { /* Changement de position de l'ascenseur */
    OPTION_A_FAIRE( "Glissement de l'ascenseur dans la vue 3D..." )
  }

  /* On passe sur la nouvelle molécule */
  MOL_aller( nv_molecule );
}

/* -------- Gestion des clics de la souris
   Idée : le bouton gauche sert à sélectionner un atome
          le bouton du milieu sert à déplacer un atome (à voir comment exactement) 
          le bouton droit appelle un menu déroulant qui propose diverses options
*/

static int CLC_atomes_3d(Fenetre *fenetre, int bouton, int n_clic __attribute__((__unused__)), int etat, int x, int y)
{
  if ( NULL == fenetre ) return FAUX;

  if ( 2 == bouton ) { /* Clic avec le bouton du milieu = bouton 2 */
    int option;

    option = SS_menu_texte( menu_contextuel_molecule, NOMBRE_OPTIONS_VUE_3D,
			    titre_menu_contextuel, -1, ROOT );
    if ( option < 0 ) /* Appel au menu annulé */
      return VRAI;

    switch ( option )
    {
    case OPT_3D_DISTANCE: /* Passer en mode "distance entre atomes" */
    case OPT_3D_DIST_CHAINE:
    case OPT_3D_DIST_RAYON:
    case OPT_3D_LIER:
    case OPT_3D_LIER_CHAINE:
    case OPT_3D_LIER_RADIAL:
      atm_precedent = NULL;
      atome_3 = NULL;
      mode_fonction = MODE_DISTANCE;
      break;

    case OPT_3D_ANGLE:
    case OPT_3D_ANGLE_CHAINE:
      atm_precedent = NULL;
      atome_3 = NULL;
      mode_fonction = MODE_ANGLE;
      break;

    case OPT_3D_CHOISIR: /* Dans tous ces cas, on agit sur un atome choisi à la fois */
    case OPT_3D_EFFACER:
    case OPT_3D_AJOUTE_FEFF:
    case OPT_3D_EFFACE_FEFF:
    case OPT_3D_DIST_ABS:
      atm_precedent = NULL;
      atome_3 = NULL;
      mode_fonction = MODE_NORMAL;
      break;

    case OPT_3D_EXPORT_POV: /* Exporter la vue au format POV */
      export_pov( mol_courante, mode_vue3d );
      option = commande_clic; /* Cette commande ne modifie pas l'état : elle ne se répète pas */
      break;

    case OPT_3D_CREER_MAILLE: /* Créer une nouvelle maille */
      {
	int i, j, k; /* Maille à créer */
	int reponse;

	/* On demande les index de la maille */
	reponse = CRIS_DemanderIndicesMaille( &Q_IndicesCreerMaille, &i, &j, &k );
	if ( reponse == (int) Q_IndicesCreerMaille.bouton_ok ) {
	  if ( MOL_creer_maille(mol_courante, i, j, k) != NULL ) {
#ifdef AVEC_OPEN_GL_
	    if ( liste_molecule ) {
	      glNewList( liste_molecule, GL_COMPILE );
	      dessine_molecule( mol_courante, mode_vue3d );
	      glEndList(); /* On a fini le dessin */
	    }
#endif
	    FEN_redessin( fenetre, BI_RDS_GL ); /* N'a pas d'intérêt sinon ! */
	  }
	}
      }
      option = commande_clic; /* Cette commande ne modifie pas l'état : elle ne se répète pas */
      break;

#ifdef _DEVELOPPE_
    case OPT_3D_EFFACER_MAILLE: /* Effacer une maille */
      {
	int i, j, k; /* Maille à créer */
	int reponse;

	/* On demande les index de la maille */
	reponse = CRIS_DemanderIndicesMaille( &Q_IndicesEffacerMaille, &i, &j, &k );
	if ( reponse == (int) Q_IndicesEffacerMaille.bouton_ok ) {
	  OPTION_A_FAIRE( "Repérer et effacer la maille demandée..." );
	}
      }
      break;

    case OPT_3D_ISOLER_MOLECULE: /* Isoler une entité moléculaire */
      atm_precedent = NULL;
      atome_3 = NULL;
      mode_fonction = MODE_NORMAL;
      break;
#endif
    }
    commande_clic = option;

    return VRAI;
  }

  if ( ( 3 == bouton ) &&
       ( OPT_3D_DIST_RAYON == commande_clic ) ) {
    Atome *atome;

    if ( NULL == mol_courante ) return VRAI; /* Aucune molécule, donc aucun atome, donc... */

#ifdef _DEVELOPPE_
    printf("Clic avec le bouton 3\n");
#endif
    atome = ATM_trouver( fenetre, mol_courante, x, y, mode_vue3d );
    if ( NULL == atome ) {
#ifdef _DEVELOPPE_
      printf("Aucun atome trouvé à cet endroit\n");
#endif
    } else
      atm_precedent = atome; /* Nouvel atome depuis lequel on rayonne */
    return VRAI;
  }

  if ( 1 == bouton ) { /* Clic avec le bouton gauche */
    Atome *atome;

    if ( NULL == mol_courante ) return VRAI; /* Aucune molécule, donc aucun atome, donc... */

    atome = ATM_trouver( fenetre, mol_courante, x, y, mode_vue3d );
    if ( NULL == atome ) {
#ifdef _DEVELOPPE_
      printf("Aucun atome trouvé à cet endroit\n");
#endif
      return VRAI;
    }

    switch (mode_fonction)
    {
    case MODE_NORMAL:
      switch (commande_clic)
      {
      case OPT_3D_CHOISIR: /* On choisit l'atome cliqué comme atome courant */
	if (etat & K_CTRL) { /* On le choisit comme absorbeur */
	  if ( FAUX == atome->marques.absorbeur ) { /* Pas utilisé comme absorbeur : on le fait donc */
	                                 /* v----------------------------v */
	    int reponse = form_alert(1,"[2][---- Choisir un absorbeur ----"
				          "|  L'atome indiqu"e_AIGU" n'est pas"
				          "|un absorbeur. Voulez-vous"
				          "|cr"e_AIGU"er un nouvel absorbeur ?][ Oui | Non ]");
	    if ( 1 == reponse ) { /* On veut un nouvel absorbeur */
	      Absorbeur *absorbeur = creer_absorbeur( atome, NULL, 100.0 );

	      if ( absorbeur != NULL ) {
		I_MOL_EditerAbsorbeur( absorbeur );
		ABS_aller( absorbeur, VRAI ); /* À CONTRÔLER ! */
	      }
	    }
	  } else {
	    unsigned int i;

	    for (i = 0 ; i < mol_courante->nbr_absorbeurs ; i++) {
	      if ( mol_courante->absorbeurs[i].absorbeur == atome ) {
		MOL_AllerAbsorbeur(mol_courante, i);
	      }
	    }
	  }
	  ATM_aller(atome);
	} else
	  ATM_aller(atome);
	break;

      case OPT_3D_EFFACER: /* Effacer l'atome choisi [Si CTRL-clic, pas de demande de confirmation] */
	ATM_efface( mol_courante, atome, (etat & K_CTRL) != 0 );
	/* Si on efface le dernier atome de la dernière molécule, on efface la dernière molécule
	   On ferme alors la fenêtre de vue en 3D, donc fenetre n'a plus de sens. */
	return VRAI;

      case OPT_3D_AJOUTE_FEFF: /* Ajouter au modèle FEFF */
	if ( mol_courante->absorbeurs != NULL ) {
	  Absorbeur *absorbeur;

	  absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
	  if ( NULL == absorbeur->modeles ) {
	    (void) fprintf(stderr, "Clic / enlever du modèle : pas de modèle pour l'absorbeur actif !\n");
	    return VRAI;
	  }
	  if ( absorbeur->modele_actif->type == MODELE_EXAFS_CHEMINS ) {
	    FEFF_AjouterAtome( absorbeur->modele_actif->infos_modele.feff, atome, -1, VRAI );
	  }
	}
	break;

      case OPT_3D_EFFACE_FEFF: /* Ôter du modèle FEFF */
	if ( mol_courante->absorbeurs != NULL ) {
	  Absorbeur *absorbeur;

	  absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
	  if ( NULL == absorbeur->modeles ) {
	    (void) fprintf(stderr, "Clic / enlever du modèle : pas de modèle pour l'absorbeur actif !\n");
	    return VRAI;
	  }
	  if ( absorbeur->modele_actif->type == MODELE_EXAFS_CHEMINS) {
	    FEFF_EnleverAtome(absorbeur->modele_actif->infos_modele.feff, atome, VRAI);
	    if ( OBJ_choisit( AFF_STRUCTURE, AST_FEFF, QUESTION ) == VRAI ) {
#ifdef AVEC_OPEN_GL_
	      if ( liste_molecule ) {
		glNewList( liste_molecule, GL_COMPILE );
		dessine_molecule( mol_courante, mode_vue3d );
		glEndList(); /* On a fini le dessin */
	      }
#endif
	      FEN_redessin( fenetre, BI_RDS_GL ); /* N'a pas d'intérêt sinon ! */
	    }
	  }
	}
	break;

      case OPT_3D_DIST_ABS: /* Modifier la distance à l'absorbeur */
	if ( mol_courante->absorbeurs != NULL ) {
	  Absorbeur *absorbeur;

	  absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
	  if ( eloigne_absorbeur(absorbeur->absorbeur, atome) == VRAI ) { /* On a changé la distance */
#ifdef AVEC_OPEN_GL_
	    if (liste_molecule) {
	      glNewList(liste_molecule, GL_COMPILE);
	      dessine_molecule(mol_courante, mode_vue3d);
	      glEndList(); /* On a fini le dessin */
	    }
#endif
	    FEN_redessin(fenetre, BI_RDS_GL); /* N'a pas d'intérêt sinon ! */
	  }
	} else {
	  (void) fprintf(stderr, "Pas d'absorbeur sélectionné ou disponible\n");
	}
	break;

      case OPT_3D_ISOLER_MOLECULE:  /* Isoler l'entité moléculaire à laquelle appartient l'atome cliqué */
	{
	  Molecule *nv_molecule = NULL;

	  nv_molecule = MOL_isoler(mol_courante, atome);
	  if (nv_molecule != NULL) { /* Devient la nouvelle molécule */
	    MOL_aller(nv_molecule);
	  }
	  commande_clic = OPT_3D_CHOISIR; /* Mode choix d'un atome pour être tranquille */
	}
	break;
      }
      break;

    case MODE_DISTANCE:
      if (atm_precedent == NULL)
	atm_precedent = atome;
      else if ( (commande_clic == OPT_3D_LIER) || (commande_clic == OPT_3D_LIER_CHAINE) || (commande_clic == OPT_3D_LIER_RADIAL) ) {
	/* On veut créer une liaison */
	Liaison *liaison;

	liaison = creer_liaison(mol_courante, atm_precedent, atome, 1.0, LIAISON_COVALENTE);
	if (liaison != NULL) {
#ifdef AVEC_OPEN_GL_
	  if (liste_liaisons) {
	    glNewList(liste_liaisons, GL_COMPILE);
	    dessine_liaisons(mol_courante, mode_vue3d);
	    glEndList(); /* On a fini le dessin */
	  }
#endif
	  FEN_redessin(fenetre, BI_RDS_GL); /* N'a pas d'intérêt sinon ! */
	}

	if (commande_clic == OPT_3D_LIER)
	  atm_precedent = NULL;
	else if (commande_clic == OPT_3D_LIER_CHAINE)
	  atm_precedent = atome;
      } else {
	char chaine[200];
	double d;

	d = ATM_distance(atm_precedent, atome);
	sprintf(chaine, "Distance %s(%f, %f, %f) - %s(%f, %f, %f) : %f "ANGSTROM,
		ATM_donnees[atm_precedent->num_atomique - 1].symbole, atm_precedent->position.x, atm_precedent->position.y, atm_precedent->position.z,
		ATM_donnees[atome->num_atomique - 1].symbole, atome->position.x, atome->position.y, atome->position.z, d);
	FEN_infos(fenetre, chaine);
	printf("Distance entre les 2 atomes : %f angstrom\n", d);
	if (commande_clic == OPT_3D_DISTANCE) /* Distance entre deux atomes, point */
	  atm_precedent = NULL;
	else if (commande_clic == OPT_3D_DIST_CHAINE)
	  atm_precedent = atome; /* Suivi d'une suite de distances */
      }
      break;

    case MODE_ANGLE:
      if (atm_precedent == NULL)
	atm_precedent = atome; /* Atome B */
      else if (atome_3 == NULL)
	atome_3 = atome; /* Atome A, au sommet car 2e choisi */
      else { /* Atome C, 3e choisi */
	double angle;

	angle = ATM_angle3( atome_3, atm_precedent, atome );
	printf("Angle entre les 3 atomes : %f°\n", angle);

	if (commande_clic == OPT_3D_ANGLE) { /* On recommence à zéro */
	  atm_precedent = NULL; atome_3 = NULL;
	} else {
	  atm_precedent = atome_3; atome_3 = atome;
	}
      }
      break;
    }
  }

  return VRAI;
}

/* -------- Gestion du clavier --------
  1, 9 : rotation autour de l'axe des Z
  4, 6 : rotation autour de l'axe des X
  2, 8 : rotation autour de l'axe des Y
  +    : zoom
  -    : zoom inverse
  A    : informations internes / molécule
  e    : bascule pour l'affichage des chemins équivalents
  G    : grouper les ligands en fragments
  i    : informations sur le chemin affiché
  I    : exporter une image (PNG)
  L    : chercher les liaisons dans la molécule
 */

static int CLV_atomes_3d(Fenetre *fenetre, unsigned int touche, int etat)
{
  double distance, latitude, longitude/*, inclinaison*/;
  double x, y, z; /* Coordonnées dans le repère du point visé */
  double pas;

  switch (touche)
  {
  case Fleche_Haut: /* Flèche vers le haut : site absorbeur précédent */
    break;

  case Fleche_Gauche: /* Flèche vers la gauche : molécule précédente = simule flèche gauche ascenseur */
    ASC_atomes_3d(fenetre, WM_ARROWED, WA_LFLINE, 0);
    return BI_EVENEMENT_TRAITE;

  case Fleche_Droite: /* Flèche vers la droite : molécule suivante = simule flèche droite de l'ascenseur */
    ASC_atomes_3d(fenetre, WM_ARROWED, WA_RTLINE, 0);
    return BI_EVENEMENT_TRAITE;

  case Fleche_Bas: /* Flèche vers la gauche : site absorbeur suivant */
    break;

  case ' ': /* On appuie sur Espace */
    if ( I_V3D_ChangerVue( mol_courante ) == VRAI)
      FEN_redessin( fenetre, BI_RDS_GL );
    return BI_EVENEMENT_TRAITE;

  case '1': /* Rotation de 15° négative, axe des X */
    y = mol_courante->pos_oeil.y - mol_courante->point_vise.y;
    z = mol_courante->pos_oeil.z - mol_courante->point_vise.z;
    if ( ( etat & K_LSHIFT ) || ( etat & K_RSHIFT ) )
      pas = 1.0;
    else
      pas = 15.0;
    mol_courante->pos_oeil.y  = y * cos(pas * M_PI / 180.0) - z * sin(pas * M_PI / 180.0);
    mol_courante->pos_oeil.z  = y * sin(pas * M_PI / 180.0) + z * cos(pas * M_PI / 180.0);
    mol_courante->pos_oeil.y += mol_courante->point_vise.y;
    mol_courante->pos_oeil.z += mol_courante->point_vise.z;
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

  case '2': /* Rotation de 15° vers le bas, axe des Z */
    x = mol_courante->pos_oeil.x - mol_courante->point_vise.x;
    y = mol_courante->pos_oeil.y - mol_courante->point_vise.y;
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) )
      pas = 1.0;
    else
      pas = 15.0;
    mol_courante->pos_oeil.x  = x * cos(pas * M_PI / 180.0) - y * sin(pas * M_PI / 180.0);
    mol_courante->pos_oeil.y  = x * sin(pas * M_PI / 180.0) + y * cos(pas * M_PI / 180.0);
    mol_courante->pos_oeil.x += mol_courante->point_vise.x;
    mol_courante->pos_oeil.y += mol_courante->point_vise.y;
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

  case '4': /* Rotation de 15° vers la gauche, axe des Y */
    x = mol_courante->pos_oeil.x - mol_courante->point_vise.x;
    z = mol_courante->pos_oeil.z - mol_courante->point_vise.z;
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) )
      pas = 1.0;
    else
      pas = 15.0;
    mol_courante->pos_oeil.x  = x * cos(pas * M_PI / 180.0) - z * sin(pas * M_PI / 180.0);
    mol_courante->pos_oeil.z  = x * sin(pas * M_PI / 180.0) + z * cos(pas * M_PI / 180.0);
    mol_courante->pos_oeil.x += mol_courante->point_vise.x;
    mol_courante->pos_oeil.z += mol_courante->point_vise.z;
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

  case '6': /* Rotation de 15° vers la droite, axe des Y */
    x = mol_courante->pos_oeil.x - mol_courante->point_vise.x;
    z = mol_courante->pos_oeil.z - mol_courante->point_vise.z;
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) )
      pas = 1.0;
    else
      pas = 15.0;
    mol_courante->pos_oeil.x  =   x * cos(pas * M_PI / 180.0) + z * sin(pas * M_PI / 180.0);
    mol_courante->pos_oeil.z  = - x * sin(pas * M_PI / 180.0) + z * cos(pas * M_PI / 180.0);
    mol_courante->pos_oeil.x += mol_courante->point_vise.x;
    mol_courante->pos_oeil.z += mol_courante->point_vise.z;
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

  case '8': /* Rotation de 15° vers le haut, axe des Z */
    x = mol_courante->pos_oeil.x - mol_courante->point_vise.x;
    y = mol_courante->pos_oeil.y - mol_courante->point_vise.y;
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) )
      pas = 1.0;
    else
      pas = 15.0;
    mol_courante->pos_oeil.x  =   x * cos(pas * M_PI / 180.0) + y * sin(pas * M_PI / 180.0);
    mol_courante->pos_oeil.y  = - x * sin(pas * M_PI / 180.0) + y * cos(pas * M_PI / 180.0);
    mol_courante->pos_oeil.x += mol_courante->point_vise.x;
    mol_courante->pos_oeil.y += mol_courante->point_vise.y;
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

  case '9': /* Rotation de 15° positive, axe des X */
    y = mol_courante->pos_oeil.y - mol_courante->point_vise.y;
    z = mol_courante->pos_oeil.z - mol_courante->point_vise.z;
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) )
      pas = 1.0;
    else
      pas = 15.0;
    mol_courante->pos_oeil.y  =   y * cos(pas * M_PI / 180.0) + z * sin(pas * M_PI / 180.0);
    mol_courante->pos_oeil.z  = - y * sin(pas * M_PI / 180.0) + z * cos(pas * M_PI / 180.0);
    mol_courante->pos_oeil.y += mol_courante->point_vise.y;
    mol_courante->pos_oeil.z += mol_courante->point_vise.z;
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

  case '-': /* On s'éloigne : -5 A*/
    x = mol_courante->pos_oeil.x - mol_courante->point_vise.x;
    y = mol_courante->pos_oeil.y - mol_courante->point_vise.y;
    z = mol_courante->pos_oeil.z - mol_courante->point_vise.z;
    distance = sqrt(x * x + y * y + z * z);
    if (distance > 0.0)
      latitude = acos(z / distance);
    else
      latitude = 0.0;
    longitude = atan2(y, x);
    distance += 5.0;
    x = distance * sin(latitude) * cos(longitude);
    y = distance * sin(latitude) * sin(longitude);
    z = distance * cos(latitude);
    mol_courante->pos_oeil.x = x + mol_courante->point_vise.x;
    mol_courante->pos_oeil.y = y + mol_courante->point_vise.y;
    mol_courante->pos_oeil.z = z + mol_courante->point_vise.z;
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

  case '+': /* On s'approche : + 5 A */
    x = mol_courante->pos_oeil.x - mol_courante->point_vise.x;
    y = mol_courante->pos_oeil.y - mol_courante->point_vise.y;
    z = mol_courante->pos_oeil.z - mol_courante->point_vise.z;
    distance = sqrt(x * x + y * y + z * z);
    if (distance > 5.0) { /* On peut encore s'approcher */
      latitude = acos(z / distance);
      longitude = atan2(y, x);
      distance -= 5.0;
      x = distance * sin(latitude) * cos(longitude);
      y = distance * sin(latitude) * sin(longitude);
      z = distance * cos(latitude);
      mol_courante->pos_oeil.x = x + mol_courante->point_vise.x;
      mol_courante->pos_oeil.y = y + mol_courante->point_vise.y;
      mol_courante->pos_oeil.z = z + mol_courante->point_vise.z;
      FEN_redessin(fenetre, BI_RDS_GL);
    }
    return BI_EVENEMENT_TRAITE;

#ifdef _DEVELOPPE_
  case 'a':
    if ( etat & K_CTRL ) return BI_EVENEMENT_NON_TRAITE; /* Pour pouvoir passer en mode ajustement... */
    MOL_Controle_Affiche_Structure(mol_courante, 100);
    return BI_EVENEMENT_TRAITE;
#endif

  case 'e':
    if ( etat & K_CTRL ) return BI_EVENEMENT_NON_TRAITE; /* Pour pouvoir passer en mode extraction... */
    if ( mode_vue3d & V3D_MODELE_EQUIV )
      mode_vue3d &= ~V3D_MODELE_EQUIV;
    else 
      mode_vue3d |= V3D_MODELE_EQUIV;
    V3D_AfficherChemin();
    FEN_redessin(fenetre, BI_RDS_GL);
    return BI_EVENEMENT_TRAITE;

#ifdef _DEVELOPPE_
  case 'g':
    if ( etat & K_CTRL ) return BI_EVENEMENT_NON_TRAITE; /* Pour pouvoir passer en mode graphique... */
    FRG_chercher_ligands(mol_courante);
    return BI_EVENEMENT_TRAITE;
#endif

  case 'i':
    printf("CLV_atomes_3d : À FAIRE : informations sur le chemin\n");
    return BI_EVENEMENT_TRAITE;

#ifdef AVEC_PNG_
  case 'I': /* On sauve une image de la vue en 3D */
    sv_image(fenetre);
    return BI_EVENEMENT_TRAITE;
#endif

  case 'L': /* On cherche les liaisons dans la molécule */
    MOL_chercher_liaisons(mol_courante);
    return BI_EVENEMENT_TRAITE;

  case 127: /* Delete = effacer la molécule */
    OPTION_A_FAIRE( "Effacer la molécule" );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE; /* Non traité : on renvoie FAUX pour pouvoir essayer les menus */
}

/* -------- Messages pour la fenêtre -------- */

static int MSG_atomes_3d(Fenetre *fenetre, Message *message)
{
  if ( message->type == WM_CLOSED ) { /* On veut fermer la fenêtre */
#ifdef _DEVELOPPE_
/*     printf("On ferme la fenêtre de vue en 3D\n"); */
#endif
#ifdef AVEC_OPEN_GL_
    /* On libère les listes Open-GL */
    if (BI_OpenGL_actif == VRAI) /* Pas de vue 3D de toute façon, alors... */
      glDeleteLists(liste_molecule, 3);
    liste_molecule = 0;
    liste_chemin = 0;
    liste_liaisons = 0;
#endif

    FEN_Detruire( fenetre );
    fen_vue_3d = NULL; /* Plus de vue 3d... Assez logique */
    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "Message non traité [%d]", message->type );
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Dessin 3D d'une molécule. --------
   Quand la fonction est appelée, l'écran OpenGL a été mis en fond noir
   Le réglage du type de projection se fait par d'autres fonctions -> il n'y
   a pas a s'en préoccuper ------ */

static void RDS_atomes_3d(Fenetre *fenetre, int x __attribute__((__unused__)), int y __attribute__((__unused__)), unsigned int larg __attribute__((__unused__)), unsigned int haut __attribute__((__unused__)))
{
#ifdef AVEC_OPEN_GL_
  GLfloat pos_lampe[4] = {0.0, 0.0, 0.0, 1.0};
#endif

  if (premiere_molecule == NULL) {
    TRC_Message( fenetre, "Aucune mol"e_AIGU"cule en mémoire actuellement" );
    return;
  }

#ifdef AVEC_OPEN_GL_
  if (BI_OpenGL_actif == FAUX) {
    TRC_Message( fenetre, "Open-GL n'est pas disponible sur cette machine" );
    return;
  } 

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(mol_courante->angle2, 1.3333333, 5, 500.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(mol_courante->pos_oeil.x, mol_courante->pos_oeil.y, mol_courante->pos_oeil.z, /* Position de l'oeil */
  	    mol_courante->point_vise.x, mol_courante->point_vise.y, mol_courante->point_vise.z, /* Pointe vers... */
  	    mol_courante->x_h , mol_courante->y_h , mol_courante->z_h); /* Vecteur vertical */
  printf( "Caractéristiques retenues : Oeil en (%g, %g, %g) et vise (%g, %g, %g) ; Vecteur vertical (%g, %g, %g)\n", 
	  mol_courante->pos_oeil.x, mol_courante->pos_oeil.y, mol_courante->pos_oeil.z,
	  mol_courante->point_vise.x, mol_courante->point_vise.y, mol_courante->point_vise.z,
	  mol_courante->x_h, mol_courante->y_h, mol_courante->z_h );

  /* On positionne la lumière */
  pos_lampe[ 0 ] = mol_courante->pos_oeil.x;
  pos_lampe[ 1 ] = mol_courante->pos_oeil.y;
  pos_lampe[ 2 ] = mol_courante->pos_oeil.z;
  glLightfv(GL_LIGHT0, GL_POSITION, pos_lampe); /* Position */
  pos_lampe[ 0 ] = mol_courante->point_vise.x;
  pos_lampe[ 1 ] = mol_courante->point_vise.y;
  pos_lampe[ 2 ] = mol_courante->point_vise.z;
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, pos_lampe); /* Point éclairé */
  
  /* On trace les axes - en rouge */
  glDisable(GL_LIGHTING);
  glShadeModel(GL_FLAT);
  glLineWidth(1.0); /* Épaisseur décente... */
  if ( OBJ_choisit(PARAMS_VUE_3D, P3D_AXE_ABSOLU, QUESTION) == VRAI ) {
    glColor3ub(255, 0, 0); /* On passe en rouge */
    glBegin(GL_LINES);
      glVertex3f(0.0, 0.0, 0.0); /* Axe des X */
      glVertex3f(1000.0, 0.0, 0.0);
      glVertex3f(0.0, 0.0, 0.0); /* Axe des Y */
      glVertex3f(0.0, 1000.0, 0.0);
      glVertex3f(0.0, 0.0, 0.0); /* Axe des Z */
      glVertex3f(0.0, 0.0, 1000.0);
    glEnd();
  }

  /* On trace les axes centrés sur le barycentre - en vert */
  if ( OBJ_choisit(PARAMS_VUE_3D, P3D_AXE_RELATIF, QUESTION) == VRAI ) {
    OPTION_A_FAIRE("Fonction pour tracer les axes 3D avec OpenGL");
#ifdef ancien
    glColor3ub(0, 255, 0); /* On passe en vert */
    glBegin(GL_LINES);
      glVertex3f(mol_courante->x_g, mol_courante->y_g, mol_courante->z_g); /* Axe des X */
      glVertex3f(1000.0, mol_courante->y_g, mol_courante->z_g);
      glVertex3f(mol_courante->x_g, mol_courante->y_g, mol_courante->z_g); /* Axe des Y */
      glVertex3f(mol_courante->x_g, 1000.0, mol_courante->z_g);
      glVertex3f(mol_courante->x_g, mol_courante->y_g, mol_courante->z_g); /* Axe des Z */
      glVertex3f(mol_courante->x_g, mol_courante->y_g, 1000.0);
    glEnd();
#endif
  }

  /* On affiche la maille, si désiré et a un sens */
  if ( ( mol_courante->maille != NULL ) && /* Les paramètres sont connus */
       (OBJ_choisit(PARAMS_VUE_3D, PV3_VOIR_MAILLE, QUESTION) == VRAI ) ) /* Et on demande le tracé */ {
    Vecteur3D v_a, v_b, v_c;
    GLfloat x_100, x_010, y_010, x_001, y_001, z_001, x_110;

    CRIS_VecteursMaille( mol_courante->maille, &v_a, &v_b, &v_c );
    x_100 = v_a.x;
    x_010 = v_b.x; y_010 = v_b.y;
    x_001 = v_c.x; y_001 = v_c.y; z_001 = v_c.z;
    x_110 = v_a.x + v_b.x;

    glLineWidth(2.0); /* Épaisseur double pour la maille */
    glColor3ub(0, 255, 255); /* Bleu ciel, par défaut */
    glBegin(GL_LINE_LOOP); /* Face basse */
      glVertex3i(0, 0, 0);           /* Point (0, 0, 0) */
      glVertex3f(x_100, 0.0  , 0.0); /* Point (1, 0, 0) */
      glVertex3f(x_110, y_010, 0.0); /* Point (1, 1, 0) */
      glVertex3f(x_010, y_010, 0.0); /* Point (0, 1, 0) */
    glEnd();
    glBegin(GL_LINE_LOOP); /* Face haute */
      glVertex3f(x_001,                 y_001        , z_001); /* Point (0, 0, 1) */
      glVertex3f(x_100 + x_001,         y_001        , z_001); /* Point (1, 0, 1) */
      glVertex3f(x_100 + x_010 + x_001, y_010 + y_001, z_001); /* Point (1, 1, 1) */
      glVertex3f(x_010 + x_001        , y_010 + y_001, z_001); /* Point (0, 1, 1) */
    glEnd();
    glBegin(GL_LINES); /* Face avant */
      glVertex3i(0, 0, 0);
      glVertex3f(x_001, y_001, z_001);
      glVertex3f(x_100, 0.0  , 0.0);
      glVertex3f(x_100 + x_001, y_001, z_001);
      glVertex3f(x_110, y_010, 0.0);
      glVertex3f(x_100 + x_010 + x_001, y_010 + y_001, z_001);
      glVertex3f(x_010, y_010, 0.0);
      glVertex3f(x_010 + x_001, y_010 + y_001, z_001);
    glEnd();
  }

  /* On affiche la molécule */
  if (liste_molecule == 0)
    dessine_molecule(mol_courante, mode_vue3d);
  else {
    glCallList(liste_molecule);
#ifdef _DEVELOPPE_
/*    printf("Erreur OpenGL/fin de l'appel de la liste : %s\n", gluErrorString( glGetError()));*/
#endif
  }

  /* Reste, ici, à s'occuper des liaisons entre fragments
     et des absorbeurs en-dehors des fragments */
  if (liste_liaisons == 0)
    dessine_liaisons(mol_courante, mode_vue3d);
  else
    glCallList(liste_liaisons);

  /* On s'occupe maintenant du tracé des chemins si nécessaire */
  if ( mode_vue3d & V3D_MODELE_SPECIAL ) { /* Il faut tracer les trucs spécifiques du modèle */
    if ( NULL == mol_courante->absorbeurs )
      return; /* Aucun absorbeur */
    if ( NULL == mol_courante->absorbeurs[ mol_courante->absorbeur_courant ].modeles )
      return; /* Aucun modèle */

    if (liste_chemin) { /* On a pu créer la liste des chemins */
      glCallList(liste_chemin);
    }/* else
	dessine_chemin( &(mol_courante->absorbeurs[num_absorbeur - 1]), num_chemin); */
  }
#else /* OpenGl non défini */
# ifdef _DEVELOPPE_
  printf("RDS_atomes_3d : fenetre = %p [OpenGL absent]\n", (void *) fenetre);
# endif
#endif
}

/* -------- Dessin des graphes dans la partie de la fenêtre à cet usage -------- */

static void RDS_graphes(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Absorbeur *absorbeur;
  Graphe *graphe = NULL;
  InfosRegion2D region;
  struct modele_ajuste *modele;
  double echelle_X = 0.0, echelle_Y = 0.0;
  int retour;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  if (mol_courante == NULL) {
    TRC_Message( fenetre, "Pas de mol"e_AIGU"cule en m"e_AIGU"moire" );
    return;
  }

  if ( NULL == mol_courante->absorbeurs ) {
    TRC_Message( fenetre, "Pas d'absorbeur dans cette mol"e_AIGU"cule\nou aucun absorbeur s"e_AIGU"lectionn"e_AIGU );
    return;
  }

  absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
  if ( NULL == absorbeur->modeles ) {
    TRC_Message( fenetre, "Pas de mod"e_GRAVE"le pour l'atome absorbeur actif" );
    return;
  }
  modele = absorbeur->modele_actif;

  if (NULL == modele->experience) { /* Pas de spectre expérimental => sans intérêt ! */
    TRC_Message( fenetre, "Pas de spectre de r"e_AIGU"f"e_AIGU"rence" );
    return;
  }

  if (OBJ_choisit(AFF_STRUCTURE, AST_EXPERIENCE, QUESTION) == VRAI) { /* On veut le voir */
    graphe = modele->experience; /* Spectre de référence, en trait continu noir */

    if ( (graphe->x_min == graphe->x_max) || (graphe->y_min == graphe->y_max) ) {
      TRC_Message( fenetre, "Graphe exp"e_AIGU"rimental : constant en X ou en Y !" );
      return;
    }

    /* On prépare la région 2D : aide pour les dessins après en fait */
    retour = TRC_Z2D_Initialiser( &region, fenetre, modele->experience );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la région 2D [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return;
    }

    /* On trace le graphe à modéliser dans la région 2D et on récupère les échelles au passage */
    /* On trace les barres d'erreur, pour ce spectre [mais en Y seulement] */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, modele->experience, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, AXE_Y,
			    &echelle_X, &echelle_Y );
  }

  /* On s'intéresse au graphe modèle, s'il existe - tracé en rouge continu, sans barres d'erreur
     avec la même échelle, évidemment... */

  if ( (OBJ_choisit(AFF_STRUCTURE, AST_THEORIE, QUESTION) == VRAI) /* On veut le voir */
       && (modele->modele != NULL) ) { /* et il existe */
    if (graphe == NULL) { /* On n'a pas affiché l'expérimental, donc il faut un repère */
      graphe = modele->modele;

      retour = TRC_Z2D_Initialiser( &region, fenetre, modele->modele );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la région 2D [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	return;
      }
    }

    /* On récupère les échelles, car peuvent ne pas avoir été calculées auparavant... */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, modele->modele, ROUGE, BI_CONTINU, BI_MARQUE_AUCUNE, 0,
			    &echelle_X, &echelle_Y );
  }

  /* On s'intéresse au graphe résidu, s'il existe - en bleu continu, sans barres d'erreur, même échelle */
  if (modele->residu) {
    if (graphe == NULL) { /* On n'a rien affiché, donc il faut un repère */
      graphe = modele->residu;

      retour = TRC_Z2D_Initialiser( &region, fenetre, modele->residu );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la région 2D [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	return;
      }
    }

    /* On récupère les échelles, car peuvent ne pas avoir été calculées auparavant... */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, modele->residu, BLEU, BI_CONTINU, BI_MARQUE_AUCUNE, 0,
			    &echelle_X, &echelle_Y );
  }

  /* On s'intéresse à la contribution du chemin affiché - tracée en vert, sans barres d'erreur */
  if ( OBJ_choisit(AFF_STRUCTURE, AST_CHEMINS, QUESTION) == VRAI ) { /* On veut voir les chemins */
    Modele_FEFF *modele_feff;
    Couche *couche_affichee = NULL;
    unsigned int i;

    if ( modele->type != MODELE_EXAFS_CHEMINS ) {
      (void) fprintf( stderr, "%s,%s l.%u : Le modèle proposé n'est pas analysable en chemins\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return;
    }
    modele_feff = modele->infos_modele.feff;
    if ( NULL == modele_feff ) {
      (void) fprintf( stderr, "%s,%s l.%u : Aucun modèle FEFF associé !\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return;
    }
    if ( NULL == modele_feff->couches ) {
      (void) fprintf( stderr, "%s,%s l.%u : Aucune couche dans le modèle FEFF !\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return;
    }

    /* On repère la couche affichée actuellement */
    for ( couche_affichee = modele_feff->couches, i = 0;
	  i < couche_active ; i++, couche_affichee = couche_affichee->suivante ) {
      if ( NULL == couche_affichee ) {
	(void) fprintf( stderr, "%s,%s l.%u : Numéro de couche active absurde [%u]\n",
			__FILE__, __FUNCTION__, __LINE__, couche_active );
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	couche_active = 0;
	couche_affichee = modele_feff->couches;
	break;
      }
    }

    /* La couche est-elle utilisée dans le modèle affiché ? */
    if ( FAUX == couche_affichee->utilisee ) {
      TRC_Message( fenetre, "Ce chemin n'est pas pris en compte dans le mod"e_GRAVE"le" );
      return;
    }

    /* On récupère le signal de cette couche */
    graphe = couche_affichee->signal;
    if ( NULL == graphe ) {

      /* Contribution comme un graphe isolé... */
      retour = COUCHE_CalculerContribution( modele_feff, couche_affichee, modele->modele );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la contribution de la couche...\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return;
      }

      /* Le graphe à dessiner est donc celui-là */
      graphe = couche_affichee->signal;
    }

    /* On dessine cette contribution, en vert */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, graphe, VERT, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );
  }
}

/* ~~~~~~~~ Fonctions ne concernant que l'affichage ~~~~~~~~ */

/* -------- Choisit le chemin à afficher, en fonction de la molécule et du modèle -------- */

void VUE3D_ChoisirChemin(void)
{
  Absorbeur *absorbeur;
  Chemin *chemin_affiche;

  couche_active = 0;
  OBJ_active(AFF_STRUCTURE, AST_CHEMINS, FORCE_NON);
  OBJ_active(AFF_STRUCTURE, AST_CHEM_AVANT, FORCE_NON);
  OBJ_active(AFF_STRUCTURE, AST_CHEM_APRES, FORCE_NON);
  OBJ_active(AFF_STRUCTURE, AST_CHEMIN_NOM, FORCE_NON);

  if (NULL == mol_courante)  {
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Pas de mol"e_AIGU"cule", TXT_COPIER);
    return;
  }
  if ( NULL == mol_courante->absorbeurs ) {
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Pas d'absorbeur", TXT_COPIER);
    return;
  }
  absorbeur = mol_courante->absorbeurs +  mol_courante->absorbeur_courant;
  if (NULL == absorbeur->modele_actif) {
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Pas de mod"e_GRAVE"le", TXT_COPIER);
    return;
  }

  if (MODELE_EXAFS_CHEMINS != absorbeur->modele_actif->type) {
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Mod"e_GRAVE"le sans chemins", TXT_COPIER);
    return;
  }
  if (NULL == absorbeur->modele_actif->infos_modele.feff) {
    (void) fprintf(stderr, "VUE3D_ChoisirChemin : modèle d'ajustement FEFF sans modèle FEFF !\n");
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Pas de mod"e_GRAVE"le FEFF", TXT_COPIER);
    return;
  }

  if (NULL == absorbeur->modele_actif->infos_modele.feff->couches) {
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Mod"e_GRAVE"le sans couche", TXT_COPIER);
    return;
  }

  OBJ_active(AFF_STRUCTURE, AST_CHEMINS, FORCE_OUI);
  OBJ_active(AFF_STRUCTURE, AST_CHEM_APRES, absorbeur->modele_actif->infos_modele.feff->nbr_couches > 1);
  OBJ_active(AFF_STRUCTURE, AST_CHEMIN_NOM, FORCE_OUI);

  chemin_affiche = absorbeur->modele_actif->infos_modele.feff->couches->chemin;
  if (NULL == chemin_affiche) {
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Couche sans chemin", TXT_COPIER);
    return;
  }

  met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, chemin_affiche->nom, TXT_COPIER);
#ifdef AVEC_OPEN_GL_
  if (liste_chemin != 0) {
    glNewList(liste_chemin, GL_COMPILE);
    dessine_chemin(absorbeur->absorbeur, chemin_affiche,
		   (mode_vue3d & V3D_MODELE_EQUIV) ? (~0U) : (0) );
    glEndList();
  }
#endif
}

/* -------- Trouver le chemin à afficher et l'afficher -------- */

static void V3D_AfficherChemin(void)
{
  Absorbeur *absorbeur;
  Couche *couche;
  Chemin *chemin;
  unsigned int i;

  if ( NULL == mol_courante )  {
    (void) fprintf( stderr, "%s,%s l.%u : appel sans molécule définie !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return;
  }
  if ( NULL == mol_courante->absorbeurs ) {
    (void) fprintf( stderr, "%s,%s l.%u : appel avec la molécule courante sans absorbeur !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return;
  }

  absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
  if (NULL == absorbeur->modele_actif) {
    (void) fprintf( stderr, "%s,%s l.%u : appel avec l'absorbeur actif sans modèle !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return;
  }
  if ( absorbeur->modele_actif->type != MODELE_EXAFS_CHEMINS ) {
    (void) fprintf( stderr, "%s,%s l.%u : modèle actif pas de type FEFF !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return;
  }
  if ( NULL == absorbeur->modele_actif->infos_modele.feff ) {
    (void) fprintf( stderr, "%s,%s l.%u : modèle actif sans infos FEFF !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return;
  }

  if ( NULL == absorbeur->modele_actif->infos_modele.feff->couches ) { /* Pas de couches dans le modèle */
    OBJ_active(AFF_STRUCTURE, AST_CHEM_AVANT, FORCE_NON );
    OBJ_active(AFF_STRUCTURE, AST_CHEMIN_NOM, FORCE_NON );
    OBJ_active(AFF_STRUCTURE, AST_CHEM_APRES, FORCE_NON );
    met_chaine(AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Pas de couche", TXT_COPIER);
    return;
  }

  OBJ_active( AFF_STRUCTURE, AST_CHEM_AVANT, 0 < couche_active );
  for ( couche = absorbeur->modele_actif->infos_modele.feff->couches, i = 0;
	i < couche_active ;
	i++, couche = couche->suivante) {
    if ( NULL == couche ) {
      (void) fprintf( stderr, "%s,%s l.%u : couche_active [%u] > au nombre de couches [%u] !\n",
		      __FILE__, __FUNCTION__, __LINE__,
		      couche_active, absorbeur->modele_actif->infos_modele.feff->nbr_couches );
      VUE3D_ChoisirChemin();
      return;
    }
  }
  OBJ_active( AFF_STRUCTURE, AST_CHEM_APRES, couche->suivante != NULL );

  chemin = couche->chemin;
  if ( NULL == chemin ) {
    met_chaine( AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, "Couche sans chemin", TXT_COPIER );
    return;
  }
  met_chaine( AFF_STRUCTURE, AST_CHEMIN_NOM, LG_MAX_TXT_CHEMIN, chemin->nom, TXT_COPIER );

#ifdef AVEC_OPEN_GL_
  if ( liste_chemin != 0 ) {
    glNewList( liste_chemin, GL_COMPILE );
    dessine_chemin( absorbeur->absorbeur, chemin,
		    (mode_vue3d & V3D_MODELE_EQUIV) ? (~0U) : (0) );
    glEndList();
  }
#endif
}

/* -------- Paramètres de l'affichage : qu'afficher, et où -------- */

static int GER_vue_3d(unsigned int objet, int x, int y)
{
  boolean etat;
  int nv_type;

#ifdef _DEVELOPPE_
  printf("GER_vue_3d : objet = %u, x = %d, y = %d\n", objet, x, y);
#endif

  switch (objet)
  {
  case AST_TYPE_ATOME:
    nv_type = SS_menu_texte( txt_vue_atomes, 3, titre_menu_vue_atomes,
			     AFF_STRUCTURE, AST_TYPE_ATOME );
    met_chaine(AFF_STRUCTURE, AST_TYPE_ATOME, 35, txt_vue_atomes[nv_type], TXT_COPIER);
    if (TXT_VOIR_ATOMES == nv_type) {
      mode_vue3d |= V3D_VOIR_ATOMES; /* Atomes visibles */
      mode_vue3d &= ~V3D_ATOME_REDUIT; /* Taille normale */
    } else if (TXT_VOIR_ATOMES_REDUITS == nv_type) {
      mode_vue3d |= V3D_VOIR_ATOMES | V3D_ATOME_REDUIT; /* Atomes visibles, taille réduite */
    } else {
      mode_vue3d &= ~V3D_VOIR_ATOMES; /* Atomes invisibles */
    }
#ifdef AVEC_OPEN_GL_
    if (liste_molecule) {
      glNewList(liste_molecule, GL_COMPILE);
      dessine_molecule(mol_courante, mode_vue3d);
      glEndList(); /* On a fini le dessin */
    }
#endif
    FEN_redessin(fen_vue_3d, BI_RDS_GL);
    break;

  case AST_VUE_TOUT: /* On veut voir tous les atomes    */
    etat = OBJ_choisit(AFF_STRUCTURE, objet, FORCE_OUI);
    if (etat == FORCE_OUI) {
      OBJ_choisit(AFF_STRUCTURE, AST_VISIBLES, FORCE_NON);
      OBJ_choisit(AFF_STRUCTURE, AST_FEFF, FORCE_NON);
    }
    mode_vue3d &= ~( V3D_VOIR_MODELE | V3D_VOIR_SELECTION );
#ifdef AVEC_OPEN_GL_
    if (liste_molecule) {
      glNewList(liste_molecule, GL_COMPILE);
      dessine_molecule(mol_courante, mode_vue3d);
      glEndList(); /* On a fini le dessin */
    }
#endif
    FEN_redessin(fen_vue_3d, BI_RDS_GL);
    break;

  case AST_VISIBLES: /* On veut voir les atomes marqués */
    etat = OBJ_choisit(AFF_STRUCTURE, AST_VISIBLES, ECHANGE);
    if (etat == FORCE_OUI) {
      OBJ_choisit(AFF_STRUCTURE, AST_VUE_TOUT, FORCE_NON);
      mode_vue3d |= V3D_VOIR_SELECTION;
    } else {
      mode_vue3d &= ~V3D_VOIR_SELECTION;
    }
#ifdef AVEC_OPEN_GL_
    if (liste_molecule) {
      glNewList(liste_molecule, GL_COMPILE);
      dessine_molecule(mol_courante, mode_vue3d);
      glEndList(); /* On a fini le dessin */
    }
#endif
    FEN_redessin(fen_vue_3d, BI_RDS_GL);
    break;

  case AST_FEFF:     /* On veut voir les atomes inclus dans un modèle FEFF */
    etat = OBJ_choisit(AFF_STRUCTURE, AST_FEFF, ECHANGE);
    if (etat == FORCE_OUI) {
      mode_vue3d |= V3D_VOIR_MODELE;
      OBJ_choisit(AFF_STRUCTURE, AST_VUE_TOUT, FORCE_NON);
    } else {
      mode_vue3d &= ~V3D_VOIR_MODELE;
    }
#ifdef AVEC_OPEN_GL_
    if (liste_molecule) {
      glNewList(liste_molecule, GL_COMPILE);
      dessine_molecule(mol_courante, mode_vue3d);
      glEndList(); /* On a fini le dessin */
    }
#endif
    FEN_redessin(fen_vue_3d, BI_RDS_GL);
    break;

  case AST_REINIT: /* On réinitialise la vue */
    OBJ_choisit(AFF_STRUCTURE, AST_FEFF, FORCE_OUI);
    MOL_cherche_barycentre(mol_courante, FAUX);
    MOL_initialise_vue(mol_courante);
#ifdef AVEC_OPEN_GL_
    if (liste_molecule) {
      glNewList(liste_molecule, GL_COMPILE);
      dessine_molecule(mol_courante, mode_vue3d);
      glEndList(); /* On a fini le dessin */
    }
#endif
    FEN_redessin(fen_vue_3d, BI_RDS_GL);
    OBJ_choisit(AFF_STRUCTURE, AST_FEFF, FORCE_NON);
    break;

  case AST_ABS_AVANT:
    if ( mol_courante->absorbeur_courant > 0 ) {
      MOL_AllerAbsorbeur(mol_courante, mol_courante->absorbeur_courant - 1);
    }
    break;

  case AST_ABS_APRES:
    if ( mol_courante->absorbeur_courant < mol_courante->nbr_absorbeurs - 1) {
      MOL_AllerAbsorbeur(mol_courante, mol_courante->absorbeur_courant + 1);
    }
    break;

  case AST_EXPERIENCE: /* Doit-on afficher le spectre expérimental ? */
  case AST_THEORIE: /* idem, pour le graphe théorique ? */
    OBJ_choisit(AFF_STRUCTURE, objet, ECHANGE);
    FEN_redessin(fen_vue_3d, BI_RDS_UTIL);
    break;

  case AST_CHEMINS:
    etat = OBJ_choisit(AFF_STRUCTURE, AST_CHEMINS, ECHANGE);
    OBJ_active(AFF_STRUCTURE, AST_CHEM_AVANT, etat);
    OBJ_active(AFF_STRUCTURE, AST_CHEM_APRES, etat);
    OBJ_active(AFF_STRUCTURE, AST_CHEMIN_NOM, etat);
    if ( VRAI == etat ) { /* On veut voir les chemins */
      V3D_AfficherChemin();
    }
    FEN_redessin(fen_vue_3d, BI_RDS_TOUT); /* Car affichage du signal du chemin éventuellement */
    break;

  case AST_CHEM_AVANT:
    if (couche_active > 0) {
      couche_active--;
      V3D_AfficherChemin();
      FEN_redessin(fen_vue_3d, BI_RDS_TOUT);
    }
    break;

  case AST_CHEMIN_NOM: /* Affiche une liste des chemins */
    if (mol_courante->absorbeurs[mol_courante->absorbeur_courant].modele_actif != NULL) {
      struct modele_ajuste *modele;

      modele = mol_courante->absorbeurs[mol_courante->absorbeur_courant].modele_actif;
      if (MODELE_EXAFS_CHEMINS == modele->type) {
	Couche *couche;

	/* On appelle le "sélecteur" */
	couche = COUCHE_Choisir(modele->infos_modele.feff, "Couche "" afficher", AFF_STRUCTURE, AST_CHEMIN_NOM);
	if ( NULL != couche ) {
	  unsigned int i;
	  Couche *avant;

	  for (i = 0, avant = modele->infos_modele.feff->couches ; couche != avant ; avant = avant->suivante, i++);
	  couche_active = i;
	  V3D_AfficherChemin();
	  FEN_redessin(fen_vue_3d, BI_RDS_TOUT); /* Car contribution du chemin... */
	}
      }
    }
    break;

  case AST_CHEM_APRES:
    couche_active++;
    V3D_AfficherChemin();
    FEN_redessin(fen_vue_3d, BI_RDS_TOUT);
    break;

  case AST_DIVERS: /* Que doit-on afficher comme "annexes" ? */
    if (params_affichage() == VRAI) { /* Vue modifiée */
#ifdef AVEC_OPEN_GL_
      if (liste_molecule) {
	glNewList(liste_molecule, GL_COMPILE);
	dessine_molecule(mol_courante, mode_vue3d);
	glEndList(); /* On a fini le dessin */
      }
#endif
      FEN_redessin(fen_vue_3d, BI_RDS_GL);
    }
    break;
  }

  return BI_EVENEMENT_TRAITE;
}

static int I_V3D_PreparerArbre(unsigned int numero_arbre __attribute__((__unused__)))
{
  int retour;

  if ( FAUX == arbres[ AFF_STRUCTURE ].drapeaux.init ) {
    OPTION_A_FAIRE( "Documentation arbre en fenêtre vue 3D" );

    /* Doc encore à rédiger
       Routine de gestion avancée : GER_vue_3d
       Pas de gestion clavier particulière
     */
    retour = ARB_Initialiser( AFF_STRUCTURE, NULL, GER_vue_3d, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%sl.%u : ERREUR à l'initialisation de l'arbre [code %d]\n",
		      __FILE__, __FUNCTION__, __LINE__, retour );
      return - 1;
    }
  }

  /* Position des objets */
  OBJ_deplace( AFF_STRUCTURE, AST_TYPE_ATOME  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( AFF_STRUCTURE, AST_TYPE_LIAISON, 0, +3, DEPLACE_PIXELS );
  
  /* ---- Sous-menus associés à cet arbre */
  /* Choix de la façon de représenter les atomes */
  retour = OBJ_MettreMenu( AFF_STRUCTURE, AST_TYPE_ATOME, 
			   NOMBRE_MODE_VISION_ATOMES, txt_vue_atomes, titre_menu_vue_atomes,
			   ( void (*)(unsigned int,unsigned int,unsigned long) ) NULL );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : Problème dans la préparation du sous-menu de visualisation des atomes [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
  }

  /* Choix de la façon de représenter les liaisons */
  retour = OBJ_MettreMenu( AFF_STRUCTURE, AST_TYPE_LIAISON, 
			   NOMBRE_MODE_VISION_LIAISONS, txt_vue_liaisons, titre_menu_vue_liaisons,
			   ( void (*)(unsigned int,unsigned int,unsigned long) ) NULL );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : Problème dans la préparation du sous-menu de visualisation des liaisons [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
  }

  /* if ( mode_vue3d & V3D_VOIR_ATOMES ) { */
  /*   if ( mode_vue3d & V3D_ATOME_REDUIT ) /\* Vue réduite des absorbeurs *\/ */
  /*     met_chaine( AFF_STRUCTURE, AST_TYPE_ATOME, 35, txt_vue_atomes[ TXT_VOIR_ATOMES_REDUITS ], TXT_COPIER ); */
  /*   else */
  /*     met_chaine( AFF_STRUCTURE, AST_TYPE_ATOME, 35, txt_vue_atomes[ TXT_VOIR_ATOMES ], TXT_COPIER ); */
  /* } else { /\* On refuse de voir les atomes *\/ */
  /*   met_chaine(AFF_STRUCTURE, AST_TYPE_ATOME, 35, txt_vue_atomes[ TXT_NE_PAS_VOIR_ATOMES ], TXT_COPIER ); */
  /* } */

  arbres[ AFF_STRUCTURE ].drapeaux.prepare = VRAI;
  return 0;
}


/* -------- Gestion des paramètres de la vue 3D (quels éléments afficher ?) ------ */

static boolean params_affichage(void)
{
  int retour;
  int fc_x, fc_z, Fc_x, Fc_z;

  if (arbres[PARAMS_VUE_3D].drapeaux.prepare == FAUX) prepare_params_affichage();

  /* On mémorise les anciennes valeurs en cas d'annulation */
  fc_x = OBJ_LireEntier_S( PARAMS_VUE_3D, P3D_X_PTE_SPH );
  fc_z = OBJ_LireEntier_S(PARAMS_VUE_3D, P3D_Z_PTE_SPH);
  Fc_x = OBJ_LireEntier_S(PARAMS_VUE_3D, P3D_X_GDE_SPH);
  Fc_z = OBJ_LireEntier_S(PARAMS_VUE_3D, P3D_Z_GDE_SPH);

#ifdef _UNIX_ /* On refait car peut avoir changé en cas de problème */
  OBJ_choisit(PARAMS_VUE_3D, P3D_DIRECT, BI_GL_direct == VRAI);
  OBJ_choisit(PARAMS_VUE_3D, P3D_INDIRECT, BI_GL_direct == FAUX);
#endif

  retour = ARB_Bloquant( PARAMS_VUE_3D, "Param"e_GRAVE"tres de dessin 3D", fen_vue_3d, 0 );
  if (retour == P3D_ANNULER) {
    OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_X_PTE_SPH, fc_x );
    OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_Z_PTE_SPH, fc_z );
    OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_X_GDE_SPH, Fc_x );
    OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_Z_GDE_SPH, Fc_z );
    return FAUX;
  }

#ifdef _UNIX_
  if (OBJ_choisit(PARAMS_VUE_3D, P3D_DIRECT, QUESTION) == VRAI)
    BI_GL_direct = VRAI;
  else
    BI_GL_direct = FAUX;
#endif

  /* Vérification d'éventuelles valeurs absurdes */
  if (OBJ_LireEntier_U( PARAMS_VUE_3D, P3D_X_PTE_SPH ) < 2) OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_X_PTE_SPH, 8 );
  if (OBJ_LireEntier_U( PARAMS_VUE_3D, P3D_Z_PTE_SPH ) < 2) OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_Z_PTE_SPH, 8 );
  if (OBJ_LireEntier_U( PARAMS_VUE_3D, P3D_X_GDE_SPH ) < 2) OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_X_GDE_SPH, 8 );
  if (OBJ_LireEntier_U( PARAMS_VUE_3D, P3D_Z_GDE_SPH ) < 2) OBJ_MettreEntier_U( PARAMS_VUE_3D, P3D_Z_GDE_SPH, 8 );

  return VRAI;
}

static int GER_params_affichage(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  int couleur;

  switch (objet)
  {
  case P3D_COL_MAILLE: /* Choix des couleurs */
  case P3D_COL_ABSOLU:
  case P3D_COL_RELATIF:
  case P3D_COL_CHEMIN:
    couleur = OBJ_couleur(PARAMS_VUE_3D, objet, -1, COULEUR_FOND, QUESTION);
    couleur = GER_ssmenu(CHX_COULEUR, PARAMS_VUE_3D, objet, couleur - 1);
    if (couleur > 0) /* On a changé de couleur */
    {
      OBJ_couleur(PARAMS_VUE_3D, objet, couleur, COULEUR_FOND, FORCE_OUI);
    }
    return VRAI;
  }

  return FAUX;
}

static void prepare_params_affichage(void)
{
  if ( FAUX == arbres[ PARAMS_VUE_3D ].drapeaux.init ) {
    int retour;

    OPTION_A_FAIRE( "Documentation : boîte paramètres d'affichage 3D" );
    retour = ARB_Initialiser( PARAMS_VUE_3D, NULL, GER_params_affichage, ( int(*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR à l'initialisation\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }
#ifndef _DEVELOPPE_
  OBJ_active(PARAMS_VUE_3D, P3D_COL_MAILLE, FORCE_NON);
  OBJ_active(PARAMS_VUE_3D, P3D_COL_ABSOLU, FORCE_NON);
  OBJ_active(PARAMS_VUE_3D, P3D_COL_RELATIF, FORCE_NON);
  OBJ_active(PARAMS_VUE_3D, P3D_COL_CHEMIN, FORCE_NON);
  OBJ_active(PARAMS_VUE_3D, P3D_DIRECT, FORCE_NON);
  OBJ_active(PARAMS_VUE_3D, P3D_INDIRECT, FORCE_NON);
#endif

  arbres[PARAMS_VUE_3D].drapeaux.prepare = VRAI;
}
