/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2001, Emmanuel CURIS
   Cration              : 10 avril 2001
   Dernire modification : 

   Modules gnraux
   Calcul matriciel, avec gestion de la mmoire optimise

   Autant que possible, les matrices sont stockes par ligne
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "systeme.h"
#include "erreurs.h"

#include "matrices.h"

/* -------- Variables locales ----------- */

static const char *txt_types_matrice[] = {
  "Quelconque",
  "Sym"e_AIGU"trique",
  "Antisym"e_AIGU"trique",
  "Triangulaire sup"e_AIGU"rieure",
  "Triangulaire inf"e_AIGU"rieure",
  "Diagonale"
};

/* -------- Fonctions locales ----------- */

static unsigned long MAT_NombreValeurs(unsigned int n_lignes, unsigned int n_colonnes, int type);

/* -------- Ralisation pratique -------- */

/* -------- Nombre de valeurs mmorises, en fonction du type et de la taille -------- */

static unsigned long MAT_NombreValeurs(unsigned int n_lignes, unsigned int n_colonnes, int type)
{
  unsigned long nbr_vals = 0;

  switch ( type )
  {
  case MATRICE_QUELCONQUE:
    nbr_vals = (unsigned long) n_lignes * (unsigned long) n_colonnes;
    break;

  case MATRICE_SYMETRIQUE: /* On ne stocke qu'un triangle, y compris la diagonale */
  case MATRICE_TRIANGULAIRE_SUP: /* La matrice est forcment carre, on prend n_lignes */
  case MATRICE_TRIANGULAIRE_INF:
    nbr_vals = (n_lignes * (unsigned long) (n_lignes + 1)) >> 1;
    break;

  case MATRICE_ANTISYMETRIQUE: /* On ne stocke que le triangle sans la diagonale, qui vaut zro */
    nbr_vals = (n_lignes * (unsigned long) (n_lignes - 1)) >> 1;
    break;

  case MATRICE_DIAGONALE: /* On ne stocke que la diagonale */
    nbr_vals = n_lignes;
    break;

  default:
    (void) fprintf(stderr, "MAT_NombreValeurs : type inconnu (%d) !\n", type);
  }

  return nbr_vals;
}

/* -------- Rservation d'une matrice -------- */

Matrice *creer_matrice(unsigned int n_lignes, unsigned int n_colonnes, int type)
{
  Matrice *matrice;
  unsigned long nbr_elements = 0, i;

  if ( (type != MATRICE_QUELCONQUE) && (n_lignes != n_colonnes) ) {
    (void) MAT_ERREUR(MAT_TYPE_IMPOSSIBLE, txt_types_matrice[type]);
    return NULL;
  }

  if ( (n_lignes == 0) || (n_colonnes == 0) ) {
    fprintf(stderr, "creer_matrice : dimensions absurdes, %u lignes et %u colonnes\n", n_lignes, n_colonnes);
    return NULL;
  }
  if ( (n_lignes == 1) && (n_colonnes == 1) )
    type = MATRICE_DIAGONALE; /* Pas besoin de se casser la tte */

  nbr_elements = MAT_NombreValeurs(n_lignes, n_colonnes, type);
  if (nbr_elements == 0) {
    (void) fprintf(stderr, "creer_matrice : Trop peu d'lments...\n");
    return NULL;
  }

  matrice = (Matrice *) malloc(sizeof(Matrice));
  if (matrice == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "matrices.c", "creer_matrice", "matrice");
    return NULL;
  }
  matrice->type = type;
  matrice->n_lignes = n_lignes;
  matrice->n_colonnes = n_colonnes;
  matrice->valeurs = NULL;

  matrice->valeurs = (double *) malloc(sizeof(double) * nbr_elements);
  if (matrice->valeurs == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "matrices.c", "creer_matrice", "matrice->valeurs");
    free(matrice);
    return NULL;
  }

  /* On initialise  zro les valeurs de la matrice */
  for ( i = 0 ; i < nbr_elements ; i++ )
    matrice->valeurs[ i ] = 0.0;

  return matrice;
}

/* -------- Copier une matrice -------- */

Matrice *copier_matrice(Matrice *matrice)
{
  Matrice *copie;
  unsigned long nombre_valeurs = 0;

  if (NULL == matrice) return NULL;

  copie = creer_matrice(matrice->n_lignes, matrice->n_colonnes, matrice->type);
  if ( NULL == copie ) return NULL;

  nombre_valeurs = MAT_NombreValeurs(copie->n_lignes, copie->n_colonnes, copie->type);
  memcpy( copie->valeurs, matrice->valeurs, sizeof(double) * nombre_valeurs);

  return copie;
}

/* -------- Dtruire une matrice cre au pralable -------- */

void effacer_matrice(Matrice *matrice)
{
  if (matrice == NULL) return;

  free(matrice->valeurs);
  free(matrice);
}

/* -------- Changer le type d'une matrice -------- */

int MAT_ChangerType(Matrice *matrice, int nouveau_type)
{
  Matrice mat_temp;
  unsigned int i, j;

#ifdef _DEVELOPPE_
  if ( NULL == matrice ) {
    (void) fprintf(stderr, "%s,%s l.%u : Matrice passe NULLE\n",
		   __FILE__, __FUNCTION__, __LINE__);
    abort();
  }
#endif

  if ( nouveau_type == matrice->type ) return 0; /* Rien  faire ! */

#ifdef _DEVELOPPE_
/*   printf("%s : Matrice avant = \n", __FUNCTION__); */
/*   MAT_AfficherMatrice( matrice ); */
#endif

  mat_temp = *matrice;
  mat_temp.type = nouveau_type;
  mat_temp.valeurs = (double *) calloc( MAT_NombreValeurs( matrice->n_lignes, matrice->n_colonnes, nouveau_type ), sizeof( double ) );
  if ( NULL == mat_temp.valeurs ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante pour faire le changement de type !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On recopie les anciennes valeurs */
  for ( i = 0 ; i < matrice->n_lignes ; i++ ) {
    for ( j = 0 ; j < matrice->n_colonnes ; j++ ) {
      met_element( &mat_temp, i, j, lit_element( matrice, i, j ) );
    }
  }

  /* Plus besoin de l'ancienne matrice */
  free( matrice->valeurs );

  /* Mais on met la nouvelle  la place ! */
  matrice->valeurs = mat_temp.valeurs;
  matrice->type = nouveau_type;

  /* Tout s'est bien pass */
#ifdef _DEVELOPPE_
/*   printf("%s : Matrice aprs = \n", __FUNCTION__); */
/*   MAT_AfficherMatrice( matrice ); */
#endif

  return 0;
}

/* -------- Accs en lecture  un lment de la matrice -------- 
   Attention, les lments sont indexs  partir de zro
*/

double lit_element(Matrice *matrice, unsigned int i, unsigned int j)
{
  if (matrice == NULL) return 0.0;

  if ( (i >= matrice->n_lignes) || (j >= matrice->n_colonnes) ) {
    (void) MAT_ERREUR(MAT_ELEMENT_INEXISTANT, i, j);
#ifdef _DEVELOPPE_
    fprintf(stderr, "Matrice %u par %u, lment demand (%u, %u) !\n",
	    matrice->n_lignes, matrice->n_colonnes, i + 1, j + 1);
    abort();
#endif
    return 0.0;
  }

  switch (matrice->type)
  {
  case MATRICE_QUELCONQUE:
#ifdef _DEVELOPPE_
/*     printf("%s : (%u, %u) soit index = %u\n", __FUNCTION__, i, j, i * matrice->n_colonnes + j ); */
#endif
    return matrice->valeurs[i * matrice->n_colonnes + j];

  case MATRICE_SYMETRIQUE: /* Partie triangulaire infrieure stocke */
    if (j > i) {
      unsigned int k = i; i = j; j = k;
    }
    return matrice->valeurs[ MAT_TINF_element(i, j) ];

  case MATRICE_ANTISYMETRIQUE: /* Triangle infrieur, sans la diagonale, stock */
    if (i == j) /* Diagonale */
      return 0.0;
    else if (j > i) /* Triangle suprieur */
      return -(matrice->valeurs[ ((i * (i-1)) >> 1) + j ]);
    else 
      return matrice->valeurs[ ((j * (j-1)) >> 1) + i ];

  case MATRICE_TRIANGULAIRE_SUP:
    if (j < i) /* Triangle infrieur */
      return 0.0;
    else
      return matrice->valeurs[ MAT_TINF_element(j, i) ]; /* On stocke en fait la transpose... */

  case MATRICE_TRIANGULAIRE_INF:
    if (j > i)
      return 0.0;
    else
      return matrice->valeurs[ MAT_TINF_element(i, j) ];

  case MATRICE_DIAGONALE:
    if (j == i)
      return matrice->valeurs[i];
    else /* Hors diagonale */
      return 0.0;
  }

  return 0.0; /* Type inconnu */
}

/* -------- Mmorise un lment dans la matrice indique -------- */

void met_element(Matrice *matrice, unsigned int i, unsigned int j, double valeur)
{
  if (matrice == NULL) return;

  if ( (i >= matrice->n_lignes) || (j >= matrice->n_colonnes) ) {
#ifdef _DEVELOPPE_
    fprintf(stderr, "met_element - Matrice %u par %u, lment demand (%u, %u) [%g] !\n",
	    matrice->n_lignes, matrice->n_colonnes, i + 1, j + 1, valeur);
    abort();
#endif
    (void) MAT_ERREUR(MAT_ELEMENT_INEXISTANT, i, j);
    return;
  }

  switch (matrice->type)
  {
  case MATRICE_QUELCONQUE:
    matrice->valeurs[i * matrice->n_colonnes + j] = valeur;
    break;

  case MATRICE_SYMETRIQUE: /* Partie triangulaire infrieure stocke */
    if (j > i) {
      unsigned int k = i; i = j; j = k;
    }
    matrice->valeurs[ MAT_TINF_element(i, j) ] = valeur;
    break;

  case MATRICE_ANTISYMETRIQUE: /* Triangle infrieur, sans la diaogonale, stock */
    if (i == j) { /* Diagonale */
      if (valeur != 0.0)
	(void) MAT_ERREUR(MAT_VAL_ABSURDE, i, j, txt_types_matrice[ MATRICE_ANTISYMETRIQUE ]);
      return;
    }
    else if (j > i) /* Triangle suprieur */
      matrice->valeurs[ ((j * (j-1)) >> 1) + i ] = -valeur;
    else 
      matrice->valeurs[ ((j * (j-1)) >> 1) + i ] = valeur;
    break;

  case MATRICE_TRIANGULAIRE_SUP:
    if (j < i) { /* Triangle infrieur */
      if (valeur != 0.0)
	(void) MAT_ERREUR(MAT_VAL_ABSURDE, i, j, txt_types_matrice[ MATRICE_TRIANGULAIRE_SUP ]);
      return;
    }
    else
      matrice->valeurs[ MAT_TINF_element(j, i) ] = valeur; /* On stocke en fait la transpose... */
    break;

  case MATRICE_TRIANGULAIRE_INF:
    if (j > i) {
      if (valeur != 0.0)
	(void) MAT_ERREUR(MAT_VAL_ABSURDE, i, j, txt_types_matrice[ MATRICE_TRIANGULAIRE_INF ]);
      return;
    }
    else
      matrice->valeurs[ MAT_TINF_element(i, j) ] = valeur;
    break;

  case MATRICE_DIAGONALE:
    if (j == i)
      matrice->valeurs[i] = valeur;
    else if (valeur != 0.0) { /* Hors diagonale */
      (void) MAT_ERREUR(MAT_VAL_ABSURDE, i, j, txt_types_matrice[ MATRICE_DIAGONALE ]);
      return;
    }
    break;
  }
}

#ifdef _DEVELOPPE_
/* -------- Pour contrles : affichage d'une matrice -------- */

void MAT_AfficherMatrice(Matrice *matrice)
{
  unsigned int i, j;

  if (matrice == NULL) return;

  for (i = 0 ; i < matrice->n_lignes ; i++) {
    printf("( ");
    for (j = 0 ; j < matrice->n_colonnes ; j++) {
      printf("%10g\t", lit_element(matrice, i, j));
    }
    printf(")\n");
  }
}
#endif
