/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996-2002, Emmanuel CURIS
   Cration de ce fichier : 14 juillet 2002

   Routines gnriques pour les ajustements

   tude du rsultat d'un ajustement par la mthode de Monte-Carlo
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "serie.h"

#include "couches.h"
#include "matrices.h"
#include "modeles.h"
#include "statistiques.h"
#include "tracer.h"		/* Pour la bote  moustaches... */

#include "montecarlo.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int MonteCarlo_StockeVariables(struct modele_ajuste *modele, Graphe *residu, Graphe **valeurs, unsigned int nbr_valeurs,
				      unsigned long tirage);
static int MonteCarlo_TiragesFaits   (struct modele_ajuste *modele, int (*fct_modele)(struct modele_ajuste *),
				      Serie_Graphes *tirages, Graphe **valeurs, unsigned int nbr_valeurs, Graphe *residu);
static int MonteCarlo_TiragesGarder  (struct modele_ajuste *modele, int (*fct_modele)(struct modele_ajuste *),
				      unsigned long nbr_tirages, Graphe **valeurs, unsigned int nbr_valeurs, Graphe *residu);
static int MonteCarlo_TiragesTemps   (struct modele_ajuste *modele, int (*fct_modele)(struct modele_ajuste *),
				      unsigned long nbr_tirages, Graphe **valeurs, unsigned int nbr_valeurs, Graphe *residu);
static int MonteCarlo_TireSpectre    (Graphe *tirage, Graphe *reference, boolean err_moyenne);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Interface visible de l'extrieur --------------------------\
 |  Renvoie la srie des valeurs (NULL en cas d'erreur)                |
 |                                                                     |
 | modele  : le modle ajust que l'on veut tudier                    |
 | tirages : si non-NULL, utilise les graphes comme spectres pr-tirs |
 | nbr_tirages    : le nombre de tirages  effectuer                   |
 | garder_tirages : VRAI si l'on veut mmoriser les spectres tirs     |
 | faire_matrice  : VRAI si l'on veut calculer la matrice de covariance|
 \-------------------------------------------------------------------- */

struct serie *MODELE_MonteCarlo(struct modele_ajuste *modele, Serie_Graphes *tirages, unsigned long nbr_tirages,
				boolean garder_tirages, boolean faire_matrice)
{
  Graphe **val_variables, *val_residu;
  Graphe *vx_experience;
  Serie_Graphes *valeurs;
  char *nom_serie;
  int retour;
  unsigned int i, nbr_variables;
  int (*fct_Modele)(struct modele_ajuste *) = ( int (*)(struct modele_ajuste *) ) NULL;

  if (NULL == modele) return NULL;
  if ( (NULL == tirages) && (0 == nbr_tirages) ) {
    (void) fprintf(stderr, "%s : aucun tirage demand !\n", __FUNCTION__);
    return NULL;
  }

  if ( VRAI == modele->drapeaux.en_cours ) {
    (void) fprintf(stderr, "MODELE_MonteCarlo : ajustement en cours sur le modle choisi !\n");
    return NULL;
  }

  /* On mmorise le vrai graphe exprimental
     Sert  rien de mmoriser la rfrence : est refaite avec MODELE_PreparerAjustement ! */
  vx_experience = modele->experience;

  /* On verifie et prepare pour les ajustements */
  retour = MODELE_VALIDITE_PreparerModele( modele );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s : la preparation du modle a chou [code %d].\n",
		    __FUNCTION__, retour );
    return NULL;
  }

  /* Dcompte des variables ajustes */
  nbr_variables = MODELE_CompterVariables( modele );
  if ( 0 == nbr_variables ) {
    (void) fprintf( stderr, "MODELE_MonteCarlo : le modle ne contient aucune variable ajustable !\n");
    MODELE_TerminerAjustement( modele, FAUX );
    return NULL;
  }

  /* On prpare les variables d'ajustement */
  retour = MODELE_InitVariables(modele);
  if (retour != 0) {
    (void) fprintf(stderr, "MODELE_MonteCarlo : l'initialisation du modle a chou [code %d].\n", retour);
    MODELE_TerminerAjustement(modele, FAUX);
    return NULL;
  }

  /* Table des graphes, contenant les variables dans l'ordre */
  val_variables = (Graphe **) malloc( sizeof(Graphe *) * nbr_variables );
  if (NULL == val_variables) {
    (void) fprintf(stderr, "MODELE_MonteCarlo : impossible de crer la table des graphes\n");
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Modeles/montecarlo.c", "MODELE_MonteCarlo", "val_variables");
    MODELE_TerminerAjustement(modele, FAUX);
    return NULL;
  }

  /* On gnre la srie qui contiendra les valeurs des variables */
  nom_serie = TXT_genere_chaine( txt_nom_serie_resultats, modele->nom_modele );
  valeurs = SERIE_GRF_CreerSerie( NULL, nom_serie );
  free( nom_serie ); nom_serie = NULL;
  if (NULL == valeurs) {
    (void) fprintf(stderr, "MODELE_MonteCarlo : impossible de crer la srie.\n");
    free( val_variables ); val_variables = NULL;
    MODELE_TerminerAjustement(modele, FAUX);
    return NULL;
  }

  if (tirages != NULL) { /* On utilise des tirages dj existants */
    nbr_tirages = tirages->nbr_graphes;
  }

  /* On cre les graphes correspondants */
  /* 1. pour le(s) rsidu(s) */
  val_residu = Serie_nv_graphe(valeurs, NULL, nbr_tirages, GR_EPREUVE, GRF_SERIE, "Monte-Carlo : r"e_AIGU"sidus des ajustements");
  if ( NULL == val_residu ) {
    (void) fprintf(stderr, "MODELE_MonteCarlo : impossible de gnrer le graphe de valeurs pour le rsidu !\n");
    free( val_variables ); val_variables = NULL;
    MODELE_TerminerAjustement(modele, FAUX);
    return NULL;
  }
  val_residu->drapeaux.protege = VRAI;

  /* 2. pour les variables proprement dites */
  for (i = 0 ; i < nbr_variables ; i++) {
    char *titre, *nom_variable, *unite;

    nom_variable = MODELE_InfosVariable(modele, i);
    titre = TXT_genere_chaine("Monte-Carlo : var. %u [%s]", (i + 1), nom_variable);
    val_variables[ i ] = Serie_nv_graphe( valeurs, (i > 0) ? (val_variables[i - 1]) : val_residu, nbr_tirages,
					  GR_EPREUVE, GRF_SERIE, titre );
    GRF_ChangerLegende( val_variables[ i ], AXE_Y, nom_variable, VRAI );
    free( nom_variable ); nom_variable = NULL;
    free( titre ); titre = NULL;
    if ( NULL == val_variables[ i ] ) {
      unsigned int j;

      (void) fprintf(stderr, "MODELE_MonteCarlo : impossible de gnrer les graphes de valeurs !\n");
      for (j = 0 ; j < i ; j++) {
	GRF_efface(val_variables[j], VRAI, VRAI);
      }
      free( val_variables ); val_variables = NULL;
      MODELE_TerminerAjustement(modele, FAUX);
      return NULL;
    }

    /* On met des bornes en Y provisoires, pour les affichages [fondes sur la valeur trouve dans l'ajustement] */
    val_variables[ i ]->y_min =  MODELE_ValeurVariable( modele, i, VARIABLE_LIRE_VALEUR_MINIMALE );
    val_variables[ i ]->y_max =  MODELE_ValeurVariable( modele, i, VARIABLE_LIRE_VALEUR_MAXIMALE );

    /* On met les bonnes units en Y */
    unite = MODELE_UniteVariable( modele, i );
    GRF_ChangerUnite( val_variables[ i ], AXE_Y, unite ) ;
    free( unite ); unite = NULL;

    /* Le graphe est protg pendant les calculs */
    val_variables[ i ]->drapeaux.protege = VRAI;
  }

  /* 3. La srie des valeurs est protge... */
  valeurs->drapeaux.protegee = VRAI;

  switch (modele->type)
  {
  case MODELE_EXAFS_CHEMINS:
    fct_Modele = NC_CalculeEXAFS;
    break;

  case MODELE_GENERIQUE:
    fct_Modele = MODELE_Generique_CalculerModele;
    break;

  default:
    fct_Modele = MODELE_CalculerModele;
    break;
  }

  /* tude Monte-Carlo proprement dite */
  if (tirages != NULL) { /* On utilise les tirages proposs */
    retour = MonteCarlo_TiragesFaits(modele, fct_Modele, tirages, val_variables, nbr_variables, val_residu);
  } else if ( VRAI == garder_tirages ) {
    retour = MonteCarlo_TiragesGarder(modele, fct_Modele, nbr_tirages, val_variables, nbr_variables, val_residu);
  } else {
    retour = MonteCarlo_TiragesTemps(modele, fct_Modele, nbr_tirages, val_variables, nbr_variables, val_residu);
  }

  /* On remet les vrais graphes modles */
  modele->experience = vx_experience;

  /* Bornes pour les graphes des variables & des rsidus */
  GRF_bornes( val_residu, NE_PAS_FAIRE, DESORDONNE );
  val_residu->drapeaux.protege = FAUX;
  if ( val_residu->affichage->complement.infos_bmoustache != NULL ) {
    GRF_eff_isole( val_residu->affichage->complement.infos_bmoustache->henry );
    val_residu->affichage->complement.infos_bmoustache->henry = NULL;
    free( val_residu->affichage->complement.infos_bmoustache );
    val_residu->affichage->complement.infos_bmoustache = NULL;
  }

  for ( i = 0 ; i < nbr_variables ; i++ ) {
    GRF_bornes( val_variables[ i ], NE_PAS_FAIRE, DESORDONNE);
    val_variables[ i ]->drapeaux.protege = FAUX;

    /* On efface les infos d'affichage de type bote  moustache, pour tre sr que les affichages soient correctement faits ensuite */
    if ( val_variables[ i ]->affichage->complement.infos_bmoustache != NULL ) {
      GRF_eff_isole( val_variables[ i ]->affichage->complement.infos_bmoustache->henry );
      val_variables[ i ]->affichage->complement.infos_bmoustache->henry = NULL;
      free( val_variables[ i ]->affichage->complement.infos_bmoustache );
      val_variables[ i ]->affichage->complement.infos_bmoustache = NULL;
    }
  }
  valeurs->drapeaux.protegee = FAUX;

  /* Fin du modle */

  /* Calcul de la matrice de covariance, si demand */
  if ( VRAI == faire_matrice ) {
    (void) MODELE_CreerMatriceCovariance(modele);
    if ( NULL != modele->covariance ) {
      unsigned int i;

      for (i = 0 ; i < modele->nbr_variables ; i++) {
	double mu_i;
	unsigned int j;

	mu_i = esperance( val_variables[i] );
	for (j = 0 ; j < i ; j++) {
	  double mu_j;

	  mu_j = esperance( val_variables[j] );
	  modele->covariance->valeurs[ MAT_TINF_element(i, j) ] = covariance( val_variables[i], val_variables[j], mu_i, mu_j);
	}
	modele->covariance->valeurs[ MAT_TINF_element(i, i) ] = variance( val_variables[i], mu_i);
      }
    }
  }

  /* Tout s'est bien pass */
  free( val_variables ); val_variables = NULL; /* Plus besoin de lui */
  MODELE_TerminerAjustement(modele, FAUX); /* Affichage inutile puisque les valeurs sont dans les graphes... */
  return valeurs;
}

/* -------- tude des Monte-Carlo, sur de nouveaux tirages, non conservs -------- */

static int MonteCarlo_TiragesTemps(struct modele_ajuste *modele, int (*fct_modele)(struct modele_ajuste *),
				   unsigned long nbr_tirages, Graphe **valeurs, unsigned int nbr_valeurs, Graphe *residu)
{
  Graphe *tirage, *vx_reference;
  unsigned long i;

  /* On prpare le tirage */
  tirage = GRF_nv_isole(modele->reference->nombre_point, modele->reference->type, modele->reference->genre_donnees);
  if (NULL == tirage) {
    (void) fprintf(stderr, "MonteCarlo_TiragesTemps : impossible de gnrer le graphe des tirages !\n");
    return -1;
  }
  memcpy(tirage->x, modele->reference->x, sizeof(double) * (modele->reference->nombre_point));
  tirage->x_min = modele->reference->x_min;
  tirage->x_max = modele->reference->x_max;
  tirage->erreur_x = modele->reference->erreur_x;
  tirage->erreur_y = modele->reference->erreur_y;

  vx_reference = modele->reference; /* Utilise dans les tirages */

/*  modele->experience = tirage; */
  modele->reference = tirage;

  for (i = 0 ; i < nbr_tirages ; i++) {
#ifdef _DEVELOPPE_
    (void) printf("MonteCarlo_TiragesTemps : tirage %lu / %lu\r", i + 1, nbr_tirages);
    (void) fflush(stdout);
#endif
    /* On gnre le spectre alatoire */
    MonteCarlo_TireSpectre(tirage, vx_reference, FAUX); /* Erreurs exprimentales uniquement */

    /* On ajuste */
    modele->iteration = 0; /* On recommence un tour...
			      Sinon, on risque de dpasser le maximum d'itrations autoris (100 000) */
    MODELE_AjusterModele(modele, fct_modele);

    /* Lecture & mmorisation des valeurs trouves */
    MonteCarlo_StockeVariables(modele, residu, valeurs, nbr_valeurs, i);
  }

  /* Nettoyage final */
  tirage->erreur_x = NULL; /* Pour ne pas effacer les erreurs de la rfrence ! */
  tirage->erreur_y = NULL;
  GRF_eff_isole(tirage);
  modele->reference = NULL;

  /* Tout s'est bien pass */
  return 0;
}

/* -------- tude des Monte-Carlo, sur de nouveaux tirages, conservs dans une srie spciale -------- */

static int MonteCarlo_TiragesGarder(struct modele_ajuste *modele, int (*fct_modele)(struct modele_ajuste *),
				    unsigned long nbr_tirages, Graphe **valeurs, unsigned int nbr_valeurs, Graphe *residu)
{
  Graphe *vx_reference, *tirage_n = NULL;
  Serie_Graphes *tirages;
  unsigned long i;

  tirages = SERIE_GRF_CreerSerie( GRF_QuelleSerie(valeurs[0]), txt_nom_serie_tirages);
  if (NULL == tirages) {
    (void) fprintf(stderr, "MonteCarlo_TiragesGarder : impossible de gnrer la srie des tirages !\n");
    return -1;
  }
  tirages->drapeaux.homogene = VRAI;
  tirages->drapeaux.protegee = VRAI;

  vx_reference = modele->reference; /* Utilise pour chaque tirage */

  for (i = 0 ; i < nbr_tirages ; i++) {
    char *titre;

    titre = TXT_genere_chaine("Monte-Carlo, tirage %l", i);
    tirages->drapeaux.protegee = FAUX; /* On dprotge juste le temps de crer un nouveau graphe */
    tirage_n = Serie_nv_graphe(tirages, tirage_n,
			       vx_reference->nombre_point, vx_reference->type, vx_reference->genre_donnees, titre);
    free(titre);
    tirages->drapeaux.protegee = VRAI;
    if (NULL == tirage_n ) {
      (void) fprintf(stderr, "MonteCarlo_TiragesGarder : impossible de gnrer le tirage %lu !\n", i + 1);
      break;
    }
    tirage_n->drapeaux.protege = VRAI;
    tirage_n->erreur_x = vx_reference->erreur_x; /* Pour la dure de l'ajustement seulement */
    tirage_n->erreur_y = vx_reference->erreur_y;

    memcpy(tirage_n->x, vx_reference->x, sizeof(double) * (vx_reference->nombre_point) );
    tirage_n->x_min = vx_reference->x_min;
    tirage_n->x_max = vx_reference->x_max;

    /* On gnre le spectre alatoire */
    MonteCarlo_TireSpectre(tirage_n, vx_reference, FAUX);

    /* On ajuste */
    /*    modele->experience = tirage_n; */
    modele->reference  = tirage_n;
    modele->iteration = 0; /* On recommence un tour... */
    MODELE_AjusterModele(modele, fct_modele);

    /* Lecture & mmorisation des valeurs trouves */
    MonteCarlo_StockeVariables(modele, residu, valeurs, nbr_valeurs, i);

    tirage_n->erreur_x = NULL; /* Pour ne pas avoir de lien vers des erreurs dans un autre graphe ! */
    tirage_n->erreur_y = NULL;
    tirage_n->drapeaux.protege = FAUX;
  }

  modele->reference = NULL; /* Sinon, sera (mal) effac si on relance un ajustement ! */

  /* Il faudrait s'occuper des barres d'erreur ! */
  printf("MonteCarlo_TiragesGarder :  FAIRE - conserver les barres d'erreur\n");

  /* Tout s'est bien pass */
  return 0;
}

/* -------- tude des Monte-Carlo, sur des tirages dj effectus -------- */

static int MonteCarlo_TiragesFaits(struct modele_ajuste *modele, int (*fct_modele)(struct modele_ajuste *),
				   Serie_Graphes *tirages, Graphe **valeurs, unsigned int nbr_valeurs, Graphe *residu)
{
  Graphe *tirage_n;
  unsigned long i;

  for (i = 0, tirage_n = tirages->premier_graphe ; i < tirages->nbr_graphes ; i++, tirage_n = tirage_n->suivant) {
    /* On adapte les anciens modles et rsidus au nouveau nombre de points */
    GRF_AdapteTaille(modele->modele, tirage_n->nombre_point);
    GRF_AdapteTaille(modele->residu, tirage_n->nombre_point);

    /* On leur prpare les X, identiques  ceux du tirage */
    memcpy(modele->modele->x, tirage_n->x, sizeof(double) * (tirage_n->nombre_point) ); 
    memcpy(modele->residu->x, tirage_n->x, sizeof(double) * (tirage_n->nombre_point) );

    /* Rfrence & exprience : le tirage considr */
/*     modele->experience = tirage_n; */
    modele->reference  = tirage_n;
    modele->iteration = 0; /* On recommence un tour... */

    /* Ajustement */
    MODELE_AjusterModele(modele, fct_modele);

    /* Lecture & mmorisation des valeurs trouves */
    MonteCarlo_StockeVariables(modele, residu, valeurs, nbr_valeurs, i);
  }
  modele->reference = NULL; /* Sinon, sera effac lors du prochain ajustement */

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Tirage d'un spectre  partir de la rfrence ---------\
 | Suppose une distribution gaussienne de chaque point            |
 |         l'indpendance de chaque point (au sens probabiliste)  |
 | Attention, aucune vrification de la validit des paramtres ! |
 \--------------------------------------------------------------- */

static int MonteCarlo_TireSpectre(Graphe *tirage, Graphe *reference, boolean err_moyenne)
{
  unsigned long i;

  if ( NULL == reference->erreur_y ) {
    (void) fprintf(stderr, "MonteCarlo_TireSpectre : pas d'erreur disponible, tirage impossible !\n");
    return -2;
  }

  if ( GRF_FCT_REELLE == reference->genre_donnees ) {
    double *y_ref, *y_erreur, *y_tire;
    int *ddl;

    for (y_ref = reference->y, y_tire = tirage->y, y_erreur = reference->erreur_y->y, ddl = reference->erreur_y->z.ddl, i = 0 ;
	 i < reference->nombre_point ; i++, y_ref++, y_tire++, y_erreur++, ddl++ ) {
      double erreur;

      /* Tirage gaussien N(,s), s : erreur ;  : valeur de la rfrence */
      erreur = *y_erreur;
      if (VRAI == err_moyenne) { /* On veut tirer avec les erreurs sur la moyenne */
	if (*ddl > 0) { /* Ce sont les erreurs exprimentales ! */
	  erreur /= sqrt( (double) (*ddl) );
	}
      } else { /* On veut les erreurs exprimentales */
	if (*ddl < 0) { /* Ce sont les sur la moyenne ! */
	  erreur *= sqrt( (double) (*ddl) );
	}
      }
      *y_tire = gaussienne() * (*y_erreur);
      *y_tire += *y_ref;
    }

  } else if ( GRF_FCT_COMPLEXE == reference->genre_donnees ) {
    (void) fprintf(stderr, "MonteCarlo_TireSpectre : tirage pour un spectre complexe  FAIRE !\n");
  } else {
    (void) fprintf(stderr, "MonteCarlo_TireSpectre : tirage impossible pour ce type de spectre !\n");
    return -1;
  }

  GRF_bornes(tirage, NE_PAS_FAIRE, DESORDONNE);
  return 0;
}

/* -------- Stocke les rsultats d'un ajustement dans les graphes adquats -------- */

static int MonteCarlo_StockeVariables(struct modele_ajuste *modele, Graphe *residu, Graphe **valeurs, unsigned int nbr_variables,
				      unsigned long tirage)
{
  unsigned int i;

  residu->y[tirage] = modele->erreur; /* Erreur  l'issue de l'ajustement */
  for (i = 0 ; i < nbr_variables ; i++) {
    valeurs[ i ]->y[ tirage ] = MODELE_ValeurVariable( modele, i, VARIABLE_LIRE_VALEUR_AJUSTEE );
    MODELE_ReinitialiseVariable( modele, i ); /* Pour qu'au tour suivant, on parte du mme point initial ! */
  }

  return 0;
}

/* ~-~-~-~-~-~-~-~- Interface graphique utilisateur ~-~-~-~-~-~-~-~- */

#define LGMAX_TXT_DISTRIBUTION 16

static int GER_montecarlo(unsigned int objet, int x, int y);

static void prep_arbre_montecarlo(void)
{
/*   OBJ_deplace(PARAM_MONTECARLO, MC_TXT_Y, 0, +500, DEPLACE_RELATIF); */
/*   OBJ_deplace(PARAM_MONTECARLO, MC_X, 0, +500, DEPLACE_RELATIF); */

#ifndef _DEVELOPPE_ /* Version de dveloppement uniquement */
  OBJ_active(PARAM_MONTECARLO, PMC_TYPE_DISTRIB, FORCE_NON);
  OBJ_active(PARAM_MONTECARLO, MC_X_CHX_DISTRIB, FORCE_NON);
  OBJ_active(PARAM_MONTECARLO, MC_REF_GRAPHE, FORCE_NON);
#endif

  arbres[PARAM_MONTECARLO].drapeaux.prepare = VRAI;
  arbres[PARAM_MONTECARLO].fonction.arbre = GER_montecarlo;
}

static int GER_montecarlo(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  switch (objet)
  {
  case MC_CONSERVER_GRF:
  case MC_CALC_MU:
  case MC_CALC_COVAR:
  case MC_CALC_ERR_TF:
    OBJ_choisit(PARAM_MONTECARLO, objet, ECHANGE);
    return VRAI;

  case MC_NV_SPECTRES:
  case MC_REF_GRAPHE:
    OBJ_choisit(PARAM_MONTECARLO, objet, FORCE_OUI);
    OBJ_active(PARAM_MONTECARLO, MC_CHX_MODELE, objet == MC_REF_GRAPHE);
    OBJ_active(PARAM_MONTECARLO, MC_CONSERVER_GRF, objet == MC_NV_SPECTRES);
    OBJ_active(PARAM_MONTECARLO, MC_NBR_TIRAGES, objet == MC_NV_SPECTRES);
    return VRAI;

  case MC_TF_RAPIDE:
  case MC_TF_NORMAL:
    OBJ_choisit(PARAM_MONTECARLO, objet, FORCE_OUI);
    return VRAI;

  case PMC_TYPE_DISTRIB: /* Choix de la loi de probas pour les Y */
    OPTION_A_FAIRE( "Choix de la distribution de tirage [en Y]" );
    return VRAI;

  case MC_X_CHX_DISTRIB: /* Choix de la loi de probas pour les X */
    OPTION_A_FAIRE( "Choix de la distribution de tirage [en X]" );
    return VRAI;
  }

  return FAUX;
}

void MODELE_InterfaceMonteCarlo(struct modele_ajuste *modele)
{
  Serie_Graphes *serie_valeurs;
  unsigned long nbr_tirages;

  if (NULL == modele) { /* On essaye d'en trouver un */
    if ( VRAI == arbres[PARAMS_MODELE].drapeaux.ouvert ) { /* On choisit celui affich pour paramtrage */
      modele = (struct modele_ajuste *) ARB_LireComplement( PARAMS_MODELE );
    } else { /* On cherche un modle pour le graphe courant */
      modele = MODELE_ChercherModele(graphe_courant);
    }

    if ( NULL == modele ) {
      (void) fprintf(stderr, "MODELE_InterfaceMonteCarlo : impossible de trouver un modle  ajuster !\n");
      return;
    }
  }

  /* On prpare l'arbre */
  if (FAUX == arbres[PARAM_MONTECARLO].drapeaux.prepare)
    prep_arbre_montecarlo();

  /* On l'appelle */
  if ( ARB_Bloquant( PARAM_MONTECARLO, "Mod"e_AIGU"lisation par Monte-Carlo", principale, 0 ) == MC_ANNULER)
    return;

  /* On en dduit les paramtres Monte-Carlo */
  nbr_tirages = OBJ_LireEntier_U( PARAM_MONTECARLO, MC_NBR_TIRAGES );

  /* On lance le Monte-Carlo proprement dit */
  serie_valeurs = MODELE_MonteCarlo(modele, NULL, nbr_tirages, OBJ_choisit(PARAM_MONTECARLO, MC_CONSERVER_GRF, QUESTION),
				    OBJ_choisit(PARAM_MONTECARLO, MC_CALC_COVAR, QUESTION));
  if ( NULL == serie_valeurs ) {
    return;
  }

  /* Traitement des rsultats
     1. Calcul des moyennes obtenues pour chaque variable */
  if ( VRAI == OBJ_choisit(PARAM_MONTECARLO, MC_CALC_MU, QUESTION) ) {
    Graphe *variable;

    (void) printf(" --- Valeurs et incertitudes obtenues ---\n");
    for (variable = serie_valeurs->premier_graphe ; variable != NULL; variable = variable->suivant) {
      double mu_j;

      (void) printf("Variable : %s\n", variable->nom);
      mu_j = esperance(variable);
      calc_param_stat(variable);
      (void) printf(" ==> %f  %g\n", mu_j, sqrt(variance(variable, mu_j)) );
    }
  }
}
