/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2002, Emmanuel CURIS
   Ce fichier : création le 14 septembre 2002

   Module d'ajustement par un modèle à n couches

   Combiner plusieurs couches semblables en une couche moyenne
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h> /* Pour fabs */

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"

#include "atomes.h"
#include "chemins.h"
#include "couches.h"
#include "feff.h"

/* ~~~~~~~~ Variables locales à ce fichier ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int NC_CHEM_Combiner_Successifs(struct modele_feff *modele, struct infos_combiner_couches *parametres);
static int NC_CHEM_Combiner_Total     (struct modele_feff *modele, struct infos_combiner_couches *parametres);

static boolean chemins_semblables(Atome *absorbeur, Chemin *chemin1, Chemin *chemin2, double d_lim, double angle_lim, boolean Z_flou);

/* ~~~~~~~~ Réalisation pratique ~~~~~~~~ */

/* -------- Interface utilisateur des combinaisons -------- */

int NC_CombinerCouches(struct modele_feff *modele, struct infos_combiner_couches *parametres)
{
  int retour = 0;
  unsigned int nbr_uniques = 0;

  /* Contrôles préliminaires */
  if ( ( NULL == modele) || ( NULL == parametres ) ) {
    AVERTISSEMENT( "Paramètres absurdes passés : modele = %p, parametres = %p", (void *) modele, (void *) parametres );
    return -1;
  }
  if ( NULL == modele->couches ) {
    AVERTISSEMENT( "Aucune couche dans le modèle indiqué [%p] !", (void *) modele );
    return -2;
  }
  if ( 1 == modele->nbr_couches ) {
    (void) fprintf( stderr, "NC_CombinerCouches : une seule couche, pas de combinaison à chercher !\n" );
    return 0;
  }

  printf( "--------- Combinaison automatique des chemins d'un absorbeur --------\n" );
  printf( "Nombre de chemins à contrôler : %d\n", modele->nbr_couches );

  if ( parametres->drapeaux & COMBINER_COUCHES_TOTAL ) {
    retour = NC_CHEM_Combiner_Total     ( modele, parametres );
  } else {
    retour = NC_CHEM_Combiner_Successifs( modele, parametres );
  }
  if ( retour != 0 ) {
    (void) fprintf( stderr, "NC_CombinerCouches : erreur dans combine_successifs ou combine_total [code : %d]\n", retour );
    return retour;
  }

  printf("%u groupes de chemins et %u chemins uniques trouvés (soit %u chemins, contre %d initialement)\n",
	 modele->nbr_couches - nbr_uniques, nbr_uniques, modele->nbr_couches, 0);

  return retour;
}

/* -------- Combiner les chemins d'un absorbeur donné --------

   La routine se charge aussi d'activer la couche associée, si nécessaire
   Elle repère des chemins SUCCESSIFS similaires et les combine en un seul
   Le chemin le plus "central" est modifié, les autres sont inchangés

   Renvoie la table des nouvelles couches (NULL en cas d'erreur) [à libérer avec free()]
           le nombre de ces nouvelles couches est dans nbr_couches
*/

static int NC_CHEM_Combiner_Successifs(struct modele_feff *modele, struct infos_combiner_couches *parametres)
{
  unsigned int nombre_couches_avant __attribute__((__unused__));
  int retour = 0;
  Chemin *reference __attribute__((__unused__)), *chemin = NULL;
  Couche *premiere_couche = NULL, *precedente = NULL;

  /* Contrôles préliminaires */
  if ( ( NULL == modele ) || ( NULL == parametres ) )return -1;
  if ( NULL == modele->couches ) {
    AVERTISSEMENT( "Aucune couche dans le modèle traité ! [%p]", (void *) modele );
    return -2;
  }

  parametres->nombre_uniques = 0;

  premiere_couche = modele->couches;
  nombre_couches_avant = modele->nbr_couches;

  /* Le modèle n'a pour l'instant aucune couche... */
  modele->nbr_couches = 0;
  modele->couches = NULL;

  /* On traite couche par couche */
  while ( premiere_couche != NULL ) {
    unsigned int nbr_semblables = 0;
    Couche *suivante;

    chemin = premiere_couche->chemin;
    if ( NULL == chemin ) { /* Pas de chemin associée : conservée telle quelle */
      Couche *suivante;

      suivante = premiere_couche->suivante;
      premiere_couche->suivante = NULL; /* On l'isole de la chaîne */
      COUCHE_Inserer( modele, precedente, premiere_couche ); /* On la garde telle qu'elle dans le modèle */
      precedente = premiere_couche;
      premiere_couche = suivante;
      continue; /* On passe à la couche suivante */
    }

#ifdef _DEVELOPPE_
    printf( "*** Nouvelle référence : %s, R = %f, N = %f\n",
	    chemin->nom, chemin->longueur_reelle, VAR_LireValeur( &(premiere_couche->N), VARIABLE_LIRE_VALEUR_INITIALE ) );
#endif

    suivante = premiere_couche->suivante;
    /* On compte le nombre de couches semblables */
    while ( chemins_semblables( modele->absorbeur, suivante->chemin, chemin,
				parametres->distance_limite, parametres->angle_limite, parametres->drapeaux & COMBINER_COUCHES_TOLERANCE_Z ) == VRAI) {
      nbr_semblables++;
      if ( 0 == ( parametres->drapeaux & COMBINER_COUCHES_DISTANCE_TOTALE ) ) {
	/* On compare les chemins successifs, pas 1er / dernier, pour la distance */
	chemin = suivante->chemin;
      }
      suivante = suivante->suivante;
    }

    if ( nbr_semblables > 0 ) { /* Il y a des combinaisons à faire */
      double R_moyen = 0.0, DW_estime = 0.0, N_total = 0.0;
      unsigned int i;
      Couche *couche, *nv_couche, *moyenne;

#ifdef _DEVELOPPE_
      printf("-- %u chemins semblables successifs trouvés --\n", nbr_semblables + 1);
#endif

      nv_couche = COUCHE_Creer( modele, precedente );
      if ( NULL == nv_couche ) {
	(void) fprintf(stderr, "combine_successifs : impossible de créer la couche moyenne\n");
	return -4;
      }

      /* Première passe : estimation de R moyen, de N total
         Travail avec les valeurs de la couche, comme ça si déjà moyenné, encore valable */
      for ( couche = premiere_couche , i = 0;
	    i <= nbr_semblables ;
	    i++, couche = couche->suivante ) {
	double N_chemin;

	N_chemin = VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_INITIALE );
	N_total += N_chemin;
	R_moyen += VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_INITIALE ) * N_chemin;
      }
      R_moyen /= N_total;

      /* Deuxième passe : quel chemin est le plus proche, estimation de DW² = variance sur R
         On en profite pour désactiver les chemins et afficher les infos */
      for ( couche = premiere_couche, moyenne = premiere_couche, i = 0 ;
	    i <= nbr_semblables ; i++, couche = couche->suivante ) {
	double delta_d;

	delta_d = VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_INITIALE ) - R_moyen;
	if ( fabs( VAR_LireValeur( &(moyenne->R), VARIABLE_LIRE_VALEUR_INITIALE ) - R_moyen ) > fabs( delta_d ) )
	  moyenne = couche; /* Cette couche a une longueur plus proche de la longueur moyenne */
	DW_estime += delta_d * delta_d * VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_INITIALE );

#ifdef _DEVELOPPE_
	printf( "%u : %s - R = %f, N = %f\n", i,
		couche->chemin->nom, VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_INITIALE ), VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_INITIALE ) );
#endif
      }
      DW_estime /= (N_total - 1.0);

      /* On initialise la couche avec les valeurs moyennes obtenues */
      nv_couche->N.valeur_initiale = N_total;
      nv_couche->R.valeur_initiale = R_moyen;
      if ( parametres->drapeaux & COMBINER_COUCHES_ESTIMER_DW ) {
	couche->Debye_Waller.valeur_initiale = DW_estime;
	if ( couche->Debye_Waller.min > DW_estime ) {
	  couche->Debye_Waller.min = DW_estime;
	}
      }

#ifdef _DEVELOPPE_
      printf("Nouveau chemin, moyen : %s, R = %f, N = %f, s² = %f\n", moyenne->nom, R_moyen, N_total, DW_estime);
#endif

      /* Phases et amplitudes */
      nv_couche->amplitude = moyenne->amplitude;
      nv_couche->phase = moyenne->phase;

      /* Chemin pour la couche : celui du moyen */
      nv_couche->chemin = moyenne->chemin;

      /* On ajoute les autres chemins comme équivalents */
      for ( couche = premiere_couche, i = 0 ; i <= nbr_semblables ; i++ ) {
	Couche *suivante;

	if ( couche != moyenne ) { /* Il faut ajouter son chemin dans les équivalents */
	  CHEMIN_AjouterEquivalent( nv_couche->chemin, couche->chemin );
	}

	/* On peut libérer la couche */
	suivante = couche->suivante;
	COUCHE_Liberer( couche, FAUX ); /* On garde phases et amplitudes ! */
	couche = suivante;
      }

      retour++;
      parametres->nombre_groupes++;

      /* On passe à la suite */
      premiere_couche = couche;
      precedente = nv_couche;
    } else { /* Couche sans équivalent : conservée telle qu'elle */
      Couche *suivante;

      suivante = premiere_couche->suivante;
      premiere_couche->suivante = NULL; /* On l'isole de la chaîne */
      COUCHE_Inserer( modele, precedente, premiere_couche ); /* On la garde telle qu'elle dans le modèle */
      precedente = premiere_couche;
      premiere_couche = suivante;

      parametres->nombre_uniques++;
    }

    reference = chemin; /* On saute tous ceux déjà combinés */
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Combiner les chemins d'un absorbeur donné --------

   La routine se charge aussi d'activer la couche associée, si nécessaire
   Elle repère des chemins similaires DANS TOUS LES CHEMINS et les combine en un seul

   Renvoie la table des nouvelles couches (NULL en cas d'erreur) [à libérer avec free()]
           le nombre de ces nouvelles couches est dans nbr_couches
*/

static int NC_CHEM_Combiner_Total(struct modele_feff *modele, struct infos_combiner_couches *parametres)
{
  unsigned int i, nbr_couches;
  Couche *couche, *reference, *precedente = NULL, **table_couches;

#ifdef _DEVELOPPE_
  printf( "entrée dans combine_total : modele = %p, parametres = %p\n",
	  (void *) modele, (void *) parametres );
#endif

  /* Contrôles préliminaires */
  if ( NULL == modele ) return -1;
  if ( NULL == modele->couches ) {
    (void) fprintf(stderr, "combine_successifs : aucune couche dans le modèle traité !\n");
    return -2;
  }
  if ( 1 == modele->nbr_couches ) {
    (void) fprintf(stderr, "combine_total : une seule couche dans le modèle traité !\n");
    return 0;
  }

  /* Préparation : liste des couches, isolées */
  table_couches = (Couche **) calloc( modele->nbr_couches, sizeof(Couche *) );
  if ( NULL == table_couches ) {
    (void) fprintf(stderr, "combine_total : impossible de créer la table initiale des couches\n");
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/combiner.c", "combine_total", "non_traites");
    return -3;
  }
  for ( i = 0, couche = modele->couches ;
	i < modele->nbr_couches ;
	i++ ) {
    table_couches[ i ] = couche;
    couche = couche->suivante;
    table_couches[ i ]->suivante = NULL; /* On défait le chaînage, inutile dans la suite ! */
  }
  nbr_couches = modele->nbr_couches;

  /* Plus de couches dans le modèle */
  modele->couches = NULL;
  modele->nbr_couches = 0;

  parametres->nombre_uniques = 0;

  /* On parcourt la table... */
  for ( i = 0 ; i < nbr_couches ; i++) {
    unsigned int nbr_semblables = 0, j;
    double R_moyen = 0.0, DW_estime = 0.0, N_total = 0.0;
    Couche *ref_interne;

    reference = table_couches[ i ];
    if ( NULL == reference ) /* Déjà regroupé auparavant */
      continue;

    if ( NULL == reference->chemin ) { /* Couche sans chemin, conservée telle qu'elle */
      COUCHE_Inserer( modele, precedente, reference );
      table_couches[ i ] = NULL; /* Normalement, ne devrait jamais resservir, mais c'est plus sûr */
      precedente = reference;
      continue;
    }

    ref_interne = reference;
    /* Première passe : on compte les semblables et prépare N et R */
    N_total = VAR_LireValeur( &(reference->N), VARIABLE_LIRE_VALEUR_INITIALE );
    R_moyen = VAR_LireValeur( &(reference->R), VARIABLE_LIRE_VALEUR_INITIALE ) * N_total;
    for ( j = i + 1 ; j < nbr_couches ; j++ ) {
      couche = table_couches[ j ];
      if ( NULL == couche ) continue; /* Déjà traitée */

      if ( chemins_semblables( modele->absorbeur, ref_interne->chemin, couche->chemin,
			       parametres->distance_limite, parametres->angle_limite, parametres->drapeaux & COMBINER_COUCHES_TOLERANCE_Z ) == VRAI ) {
	double N_chemin;

	nbr_semblables++;
	if ( 0 == ( parametres->drapeaux & COMBINER_COUCHES_DISTANCE_TOTALE ) ) {
	  /* On compare les chemins successifs, pas 1er / dernier, pour la distance */
	  ref_interne = couche;
	}
	N_chemin = VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_INITIALE );
	N_total += N_chemin;
	R_moyen += VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_INITIALE ) * N_chemin;
      }
    }

    ref_interne = reference;
    if ( nbr_semblables > 0 ) { /* Deuxième passe nécessaire */
      Couche *nv_couche, *moyenne, **anciennes;
      boolean N_variable = FAUX, R_variable = FAUX;
      unsigned long k;

#ifdef _DEVELOPPE_
      (void) printf("-- %u chemins semblables trouvés --\n", nbr_semblables + 1);
#endif
      nv_couche = COUCHE_Creer( modele, precedente );
      if ( NULL == nv_couche ) {
	(void) fprintf(stderr, "%s,%s l.%u : impossible de créer la couche moyenne\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -4;
      }

      /* Table des anciennes couches, à effacer */
      anciennes = (Couche **) malloc( sizeof( Couche * ) * ( nbr_semblables + 1 ) );
      if ( NULL == anciennes ) {
	ERR_Memoire( "Table des couches "a_GRAVE" effacer", sizeof( Couche * ) * ( nbr_semblables + 1 ) );
	COUCHE_Effacer( modele, nv_couche, VRAI, VRAI );
	return -10;
      }

      moyenne = reference;
      R_moyen /= N_total;

      for ( j = i ; j < nbr_couches ; j++ ) {
	couche = table_couches[ j ];
	if ( NULL == couche ) continue; /* Déjà traitée */

	if ( chemins_semblables( modele->absorbeur, ref_interne->chemin, couche->chemin,
				 parametres->distance_limite, parametres->angle_limite, parametres->drapeaux & COMBINER_COUCHES_TOLERANCE_Z ) == VRAI ) {
	  double delta_d;

	  delta_d = fabs( VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_INITIALE ) - R_moyen );
	  if ( 0 == ( parametres->drapeaux & COMBINER_COUCHES_DISTANCE_TOTALE ) ) {
	    /* On compare les chemins successifs, pas 1er / dernier, pour la distance */
	    ref_interne = couche;
	  }

	  if ( fabs( VAR_LireValeur( &(moyenne->R), VARIABLE_LIRE_VALEUR_INITIALE ) - R_moyen) > delta_d )
	    moyenne = couche; /* Ce chemin a une longueur plus proche de la longueur moyenne */
	  DW_estime += delta_d * delta_d * VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_INITIALE );

#ifdef _DEVELOPPE_
	  printf( "%u : %s - R = %f, N = %f, couche = %p\n", j + 1, couche->nom,
		  VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_INITIALE ), VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_INITIALE ), (void *) couche);
#endif

	  if ( VAR_LIBRE == couche->N.type )
	    N_variable = VRAI;
	  if ( VAR_LIBRE == couche->R.type )
	    R_variable = VRAI;
	}
      }
      DW_estime /= (N_total - 1.0);

      /* On initialise la nouvelle couche en conséquence */
#ifdef _DEVELOPPE_
      printf("Chemin moyen : %s, R = %f, N = %f, s² = %f\n", moyenne->nom, R_moyen, N_total, DW_estime);
#endif
      nv_couche->N.valeur_initiale = N_total;
      if ( FAUX == N_variable ) {
	nv_couche->N.type = VAR_CONSTANTE;
      } else {
	nv_couche->N.min = N_total - 1;
	nv_couche->N.min = N_total + 1;
      }
      nv_couche->R.valeur_initiale = R_moyen;
      if ( FAUX == R_variable ) {
	nv_couche->R.type = VAR_CONSTANTE;
      } else {
	nv_couche->R.min = R_moyen - 0.2; /* 0,2 A de variation autorisée */
	nv_couche->R.max = R_moyen + 0.2;
      }
      if ( parametres->drapeaux & COMBINER_COUCHES_ESTIMER_DW ) {
	nv_couche->Debye_Waller.valeur_initiale = DW_estime;
	if ( couche->Debye_Waller.min > DW_estime ) {
	  couche->Debye_Waller.min = DW_estime;
	}
      }
      nv_couche->chemin = moyenne->chemin;
      nv_couche->amplitude = moyenne->amplitude;
      nv_couche->phase = moyenne->phase;
      COUCHE_Nommer( nv_couche, moyenne->nom );

      /* Troisième passe : on met les chemins équivalents */
      ref_interne = reference;
      for ( j = i, k = 0 ; j < nbr_couches ; j++ ) {
	couche = table_couches[ j ];
	if ( NULL == couche ) continue; /* Déjà traitée */

	if ( chemins_semblables( modele->absorbeur, ref_interne->chemin, couche->chemin,
				 parametres->distance_limite, parametres->angle_limite, parametres->drapeaux & COMBINER_COUCHES_TOLERANCE_Z ) == VRAI ) {
#ifdef _DEVELOPPE_
	  printf( "%u : %s - chemin = %p sera équivalent à celui de référence interne %p\n",
		  j + 1, couche->nom, (void *) couche->chemin, (void*) ref_interne->chemin );
#endif
  	  CHEMIN_AjouterEquivalent( nv_couche->chemin, couche->chemin );
	  if ( 0 == ( parametres->drapeaux & COMBINER_COUCHES_DISTANCE_TOTALE ) ) {
	    /* On compare les chemins successifs, pas 1er / dernier, pour la distance */
	    ref_interne = couche;
	  }

 	  anciennes[ k++ ] = couche;
	  table_couches[ j ] = NULL;
	  /* Ne servira plus pour les autres groupes !
	     Mais pas encore effaçable, car peut servir pour la comparaison avec la couche suivante !
	   */
	}
      }

#ifdef _DEVELOPPE_
      /* Affichage de la table des couches */
      printf( "---- Table des %u anciennes couches ----\n", nbr_couches );
      for ( k = 0 ; k < nbr_couches ; k++ ) {
	printf( "%lu : %p\n", k, (void *) table_couches[ k ] );
      }
#endif

      /* Quatrième passe :  on efface les anciennes couches [pas avant car peuvent servir dans les successives... */
      printf( "-- Libération des couches : \n" );
      for ( k = 0 ; k <= nbr_semblables ; k++ ) {
	printf( " Couche %lu / %u [%p] {nouvelle = %p}\n",
		k, nbr_semblables, (void *) anciennes[ k ], (void *) nv_couche );
	anciennes[ k ]->chemin = NULL; /* Pour ne pas qu'il soit effacé ! */
	COUCHE_Liberer( anciennes[ k ], FAUX );
      }
      free( anciennes ); anciennes = NULL;

      parametres->nombre_groupes++;
      precedente = nv_couche;
    }  else {
      parametres->nombre_uniques++;

      /* On garde la couche telle qu'elle */
      COUCHE_Inserer( modele, precedente, reference );
      precedente = reference;
    }

    table_couches[ i ] = NULL; /* Normalement, ne devrait jamais resservir, mais c'est plus sûr */
  }

  free( table_couches ); table_couches = NULL;
  /* Tout s'est bien passé */
  return 0;
}

/* -------- Renvoie VRAI si deux chemins sont assimilables a priori -------- */

static boolean chemins_semblables(Atome *absorbeur, Chemin *chemin1, Chemin *chemin2, double d_limite, double angle_limite, boolean Z_flou )
{
  Atome *atome_1, *atome_2, *pre1, *pre2;
  unsigned int i;

  if ( (chemin1 == NULL) || (chemin2 == NULL) ) return FAUX;
  if (chemin1 == chemin2) return VRAI;

  /* Ont-ils le même nombre d'atomes ? */
  if (chemin1->nbr_atomes != chemin2->nbr_atomes) return FAUX;

  /* Les distances sont-elles compatibles ? Seuil à 0,05 A, ce qui fait pas mal... */
  if ( fabs(chemin1->longueur_reelle - chemin2->longueur_reelle) >= d_limite )
    return FAUX;
/*   printf("Longueurs réelles : %g et %g [limite : %g]\n", chemin1->longueur_reelle, chemin2->longueur_reelle, d_limite); */

  /* Passe-t-on par le même type d'atomes ? */
  for (i = 0 ; i < chemin1->nbr_atomes ; i++) {
    atome_1 = chemin1->atomes[i];
    atome_2 = chemin2->atomes[i];
    if ( atome_1->num_atomique != atome_2->num_atomique ) {
      if ( VRAI == Z_flou ) {
	/* Une tolérance Z ~ Z+1 ~ Z-1 est admise */
	if ( abs( (signed) atome_1->num_atomique - (signed) atome_2->num_atomique ) > 1 ) {
	  return FAUX;
	}
      } else {
	/* Aucune tolérance de Z */
	return FAUX;
      }
    }
  }

  /* On vérifie la géométrie des fragments successifs (tolérance définie par l'utilisateur) */
  if ( chemin1->nbr_atomes > 1 ) { /* Pas de sens pour les réflexions simples */
    Atome *post1, *post2;

    pre1 = absorbeur;
    pre2 = absorbeur;
    atome_1 = chemin1->atomes[0];
    atome_2 = chemin2->atomes[0];
    for (i = 1 ; i < chemin1->nbr_atomes - 1 ; i++) {
      double angle1, angle2;
      
      post1 = chemin1->atomes[i];
      post2 = chemin2->atomes[i];
      
      angle1 = ATM_angle3(pre1, atome_1, post1);
      angle2 = ATM_angle3(pre2, atome_2, post2);
      if ( fabs(angle1 - angle2) > angle_limite ) {
#ifdef _DEVELOPPE_
	printf("==>   Chemins différents car différence d'angle de %f°\n", angle1 - angle2);
#endif
	return FAUX;
      }

      pre1 = atome_1; atome_1 = post1;
      pre2 = atome_2; atome_2 = post2;
    }
  }

/*  printf("==> Chemins semblables : %s et %s\n", chemin1->nom, chemin2->nom); */
  return VRAI;
}

