/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2001, Emmanuel CURIS
   Ce fichier : cration le 17 septembre 1998

   Module d'ajustement par un modle  n couches

   Paramtrage des paramtres spcifiques  chaque couche de l'ajustement
   Fonctions de gestion des couches, au passage
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"

#include "atomes.h"
#include "chemins.h" /* Pour effacer_chemin() */
#include "couches.h"
#include "feff.h"

/* ~~~~~~~~ Variables locales  ce fichier ~~~~~~~~ */

static unsigned long nombre_couches = 0; /* Nombre de couches existantes */

static const Variable constante = {0, 0, 0, 0, VAR_CONSTANTE, {0}};

#include "couches.txt"

/* ~~~~~~~~ Fonctions locales  ce fichier ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique des diffrentes fonctions ~~~~~~~~ */

/* ~-~-~-~-~-~-~-~- Fonctions de cration/effacement des couches ~-~-~-~-~~-~-~-~- */

/* -------- Cration d'une nouvelle couche, mise aprs la couche actuellement dite --------\
 |   Renvoie NULL en cas d'erreur, l'adresse de la nouvelle couche sinon                     |
 |                                                                                           |
 | modele : le modle FEFF auquel doit tre ajoute cette couche                             |
 | avant  : la couche aprs laquelle on va crer cette couche (NULL : cration  la fin)     |
 \------------------------------------------------------------------------------------------ */

Couche *COUCHE_Creer(Modele_FEFF *modele, Couche *avant)
{
  int retour;
  Couche *couche;

  if ( NULL == modele ) return NULL; /* Pas de modle */

  /* Rservation de la mmoire */
  couche = (Couche *) malloc( sizeof(Couche) );
  if (couche == NULL) {
    (void) fprintf(stderr, "COUCHE_Creer : mmoire insuffisante [taille souhaite : %lu]\n", (unsigned long) sizeof(Couche));
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "N_couches/couches.c", "COUCHE_creer", "couche");
    return NULL;
  }

  /* Insertion dans le modle */
  couche->suivante = NULL; /* Pour prouver qu'elle est bien isole */
  retour = COUCHE_Inserer(modele, avant, couche);
  if ( 0 != retour ) {
    (void) fprintf(stderr, "COUCHE_Creer : erreur dans COUCHE_Inserer [code %d]\n", retour);
    free(couche);
    return NULL;
  }

  /* On fait l'init. de cette couche */
  sprintf(couche->nom, nom_couche_defaut, nombre_couches);

  couche->utilisee = VRAI; /* Par dfaut, sera prise en compte dans le modle */

  couche->N.valeur_initiale = 6.0; /* Nombre de voisins = dgnrescence */
  couche->N.min   = 5.5;
  couche->N.max   = 6.5;
  couche->N.delta = 0.001;
  couche->N.type  = VAR_LIBRE;
  couche->N.contenu.valeur = 6.0;

  couche->R.valeur_initiale = 1.8; /* Distance  l'absorbeur = longueur du chemin (en angstrom) */
  couche->R.min   = 1.5;
  couche->R.max   = 2.5;
  couche->R.delta = 0.0001;
  couche->R.type  = VAR_LIBRE;
  couche->R.contenu.valeur = 1.8;

  couche->Debye_Waller.valeur_initiale = 0.01; /* Facteur de Debye-Waller (en angstrom^2) */
  couche->Debye_Waller.min   = 0.001;
  couche->Debye_Waller.max   = 0.05;
  couche->Debye_Waller.delta = 0.00001;
  couche->Debye_Waller.type  = VAR_LIBRE;
  couche->Debye_Waller.contenu.valeur = 0.01;

  couche->Cumul_1 = constante; /* Premier cumulant   (dsordre) - fixe dans le cas gnral */
  couche->Cumul_2 = constante; /* Deuxime cumulant  (phases)   - fixe dans le cas gnral */
  couche->Cumul_3 = constante; /* Troisime cumulant (dsordre) - fixe dans le cas gnral */
  couche->Cumul_4 = constante; /* Quatrime cumulant (phases)   - fixe dans le cas gnral */
  couche->Delta_E = constante; /* Dcalage de E0, mme remarque que ci-dessous                   */
  couche->Gamma   = constante; /* Libre parcours moyen, cas 1 valeur par couche = non par dfaut */

  couche->signal    = NULL; /* Aucun signal associ */
  couche->amplitude = NULL;
  couche->phase     = NULL;

  couche->b_amp = NULL; /* Aucunes infos pour le lissage de l'amplitude */
  couche->c_amp = NULL;
  couche->d_amp = NULL;

  couche->b_pha = NULL; /* Aucunes infos pour le lissage de la phase    */
  couche->c_pha = NULL;
  couche->d_pha = NULL;

  couche->chemin = NULL; /* Aucun chemin associ */

  return couche;
}

/* -------- Effacer une couche ---------------------------\
 |  Renvoie 0 si tout s'est bien pass, != 0 sinon        |
 |                                                        |
 | modele_feff : le modle auquel appartient la couche    |
 | couche      : la couche que l'on veut effacer          |
 | forcer      : VRAI = on ne demande pas confirmation    |
 | eff_graphes : VRAI = on efface les graphes associs    |
 \------------------------------------------------------- */

int COUCHE_Effacer(struct modele_feff *modele, Couche *couche, boolean forcer, boolean eff_graphes)
{
  Couche *avant;

  if (NULL == modele) return -1;
  if (couche == NULL) return 0;

  if (forcer == FAUX) { /* On demande confirmation */
    int reponse;

    reponse = BI_ERR_alerte(NULL, &Q_EffacerCouche, VRAI, couche->nom);
    if (3 == reponse) return +1; /* On annule l'effacement */
    if (2 == reponse)
      eff_graphes = FAUX;
    else
      eff_graphes = VRAI;
  }

  /* ------ On retire la couche du chanage ------ */
#ifdef _DEVELOPPE_
  printf("COUCHE_Effacer : %s [%p]\n", couche->nom, (void *) couche);
#endif

  if (couche == modele->couches) {
    avant = NULL;
    modele->couches = couche->suivante;
  } else { /* On cherche quelle couche se trouve avant */
    avant = modele->couches;
    while ( avant->suivante != couche ) {
      avant = avant->suivante;
      if ( NULL == avant ) {
	(void) fprintf(stderr, "COUCHE_Effacer : la couche %s n'est pas dans le modle indiqu ! Effacemen annul.\n",
		       couche->nom);
	return -3;
      }
    }

    avant->suivante = couche->suivante; /* On reforme le chanage */
  }

  if (couche == couche_editee) { /* Nouvelle couche courante ? */
    if (couche->suivante) /* Il y a encore une couche aprs */
      COUCHE_Aller( modele, couche->suivante );
    else if ( avant != NULL )
      COUCHE_Aller( modele, avant );
    else
      COUCHE_Aller( modele, modele->couches );
  }

  /* ------ On vrifie que la couche n'est pas utilise ailleurs ------ */

  /* ------ On libre les donnes associes  la couche ------ */
  COUCHE_Liberer(couche, eff_graphes);

  /* Une couche de moins... */
  modele->nbr_couches--;

  return 0;
}

int COUCHE_Liberer(Couche *couche, boolean eff_graphes)
{
  /* 1. Phases et amplitudes, si ncessaire */
  if (VRAI == eff_graphes) { /* On veut effacer les phases et amplitudes associes */
    Graphe *tmp_amp, *tmp_phase;

    /* On vite un appel rcursif de l'effacement des graphes... */
    tmp_amp = couche->amplitude; couche->amplitude = NULL;
    tmp_phase = couche->phase; couche->phase = NULL;
    GRF_efface(tmp_amp, VRAI, VRAI);
    GRF_efface(tmp_phase, VRAI, VRAI);
  }

  /* 2. Chemins associs  la couche
     Rq. On n'efface pas de nouveau amplitudes et phases du chemin de rfrence, car sinon double effacement
     En revanche, on efface si ncessaire amplitudes et phases des chemins quivalents */
  CHEMIN_Effacer(couche, couche->chemin, VRAI, FAUX,
		 (eff_graphes == VRAI) ? (EFF_CHEMIN_GRAPHES_EQU) : (EFF_CHEMIN_SANS_GRAPHES));

  /* 3. Tableaux des coefficients des splines */
  free(couche->b_amp); free(couche->c_amp); free(couche->d_amp);
  free(couche->b_pha); free(couche->c_pha); free(couche->d_pha);

  /* 4. La couche elle-mme */
  free(couche);

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Insrer une couche dans un modle -------- */

int COUCHE_Inserer(struct modele_feff *modele, Couche *avant, Couche *couche)
{
  if ( (NULL == modele) || (NULL == couche) ) return -1;
  if (couche->suivante != NULL) {
    (void) fprintf(stderr, "COUCHE_Inserer : la couche n'est pas isole !\n");
    return -2;
  }

  /* Chanage */
  if (modele->couches == NULL) { /* C'est la premire couche que l'on cre */
    modele->couches = couche;
    modele->nbr_couches = 1;
    couche->suivante = NULL; /* C'est donc aussi la dernire           */
  } else { /* Il faut la placer juste aprs la couche indique */

    /* Reprage de la couche juste avant */
    if ( NULL == avant ) { /* On se place sur la dernire */
      for (avant = modele->couches ; avant->suivante != NULL ; avant = avant->suivante);
      avant->suivante = couche;
      couche->suivante = NULL;
    } else {
      Couche *c;

      /* On vrifie que la couche avant est bien dans le modle indiqu... */
      for (c = modele->couches ; c != avant ; c = c->suivante) {
	if (NULL == c) {
	  (void) fprintf(stderr, "COUCHE_creer : la couche 'avant' n'est pas dans le modle indiqu !\n");
	  avant = NULL;
	  break;
	}
      }

      if ( NULL == avant ) { /* On l'insre au dbut */
	couche->suivante = modele->couches;
	modele->couches = couche;
      } else {
	couche->suivante = avant->suivante;
	avant->suivante = couche;
      }
    }

    modele->nbr_couches++; /* Une couche de plus */
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Modifier le nom d'une couche -------- */

void COUCHE_Nommer(Couche *couche, const char *nom)
{
  if (NULL == couche) return;
  if (NULL == nom) return;

  strncpy(couche->nom, nom, LG_MAX_NOM_COUCHE);
  couche->nom[LG_MAX_NOM_COUCHE] = '\0';
}
