/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2002, Emmanuel CURIS
   Cration              : 17 septembre 1998

   Module d'ajustement par un modle  n couches

   Lancement de l'ajustement
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"

#include "atomes.h"   /* Pour les absorbeurs */
#include "couches.h"
#include "feff.h"     /* Pour le modle FEFF */
#include "filtrage.h" /* Pour les infos Nyquist */
#include "modeles.h"  /* Pour les modles */
#include "lissage.h"
#include "science.h"
#include "theorie.h"  /* Pour les modles de libre-parcours moyen */
#include "tracer.h"

#include "nc_lancer.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

static Couche *couche_affichee = NULL;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void NC_AffichageIteration(struct modele_ajuste *modele);
static void NC_AffichageCouche   (struct couche *couche);

static int prepare_arbre_ncouches(unsigned int);

void RDS_ncouches(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);
int  MSG_ncouches(Fenetre *fenetre, Message *message);

static int GER_ncouches(unsigned int objet, int x, int y);

static boolean verifie_deltae  (Variable *variable, char *couche, double de_min);
static boolean verifie_variable(Variable *variable, char *nom, char *couche);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

int NC_AjusterEXAFS(struct modele_ajuste *modele, Fenetre *fenetre)
{
  int retour;

  if ( NULL == modele ) return -1;
  if ( NULL == modele->experience ) return -2; /* Ne devrait jamais arriver */

  /* 1. Contrle ventuel du nombre de paramtres ajusts */
  if ( modele->experience->pointeur.nyquist != NULL ) { /* Si oui, on applique Shannon */
    unsigned long Nmax;
    Infos_Nyquist *nyquist = modele->experience->pointeur.nyquist;

    Nmax = (unsigned long) (2.0 * (modele->x_max - modele->x_min) * (nyquist->r_max - nyquist->r_min) ) / M_PI;
    if (Nmax + 2 < modele->nbr_variables) {
      int reponse;

      reponse = BI_ERR_alerte(NULL, &Q_ShannonDepasse, VRAI, Nmax + 2, modele->nbr_variables);
      if ( 2 == reponse ) {
	/* Ajustement annul : on le marque comme fini */
	modele->drapeaux.en_cours = FAUX;
	return -3;
      }
    }
    OBJ_active( RESULT_N_COUCHES, RNC_NYQUIST, FORCE_OUI );
    OBJ_MettreEntier_U( RESULT_N_COUCHES, RNC_NYQUIST, Nmax );
  } else {
    OPTION_A_FAIRE( "Contrle de Nyquist avec le R du plus long chemin comme rmax, 0 comme rmin" );
    OBJ_active( RESULT_N_COUCHES, RNC_NYQUIST, FORCE_NON );
  }

  /* 2. On a le modle de dbut, les rsidus ; les variables d'ajustement sont prtes : on ajuste */
#ifdef _DEVELOPPE_
  printf( " ==> %s <== On va lancer l'ajustement [ponderation : %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere );
#endif
  switch ( modele->algorithme )
  {
  case ALGO_PSEUDOGRADIENT:
    retour = MODELE_AjustementPG( modele, fenetre,
				  ( double (*)(struct modele_ajuste *, double) ) NULL,
				  NC_CalculeEXAFS, NC_IncrementsPG, NC_AffichageIteration );
    break;

  default:
    AVERTISSEMENT( "Algorithme inconnu [%d]", modele->algorithme );
    retour = -1;
  }

  /* 3. Ajustement fini : on fait ce qu'il faut, y compris l'affichage... */
  MODELE_TerminerAjustement( modele, (retour >=0) );

  /* 4. Contrle de la raison de la fin de l'ajustement */
  if ( retour < 0 ) { /* Ajustement fini anormalement : problme */
    (void) fprintf(stderr, "NC_AjusterEXAFS : l'ajustement n'a pas pu tre men  bien !\n");
    return retour;
  } else if ( retour > 0 ) { /* Ajustement fini car une condition d'arrt a t rencontre */
  }

  /* 5. Affichage des rsultats [dj fait pour les rsidus] */
  OPTION_A_FAIRE( "Afficher les rsultats de l'ajustement" );

  /* 6. Travail post-ajustement, si demand... */
/*   if ( OBJ_choisit(AJUSTE_GENERAL, NCG_MAT_CORR, QUESTION) == VRAI ) { */
/*     NC_covariance(infos); */
/*   } */

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Ajuster un modle -------- */

int NC_AjusterModele(struct modele_ajuste *modele)
{
  int code_erreur;
  unsigned int nbr_variables;

  if ( NULL == modele ) return -1;
  if ( NULL == modele->experience ) {
    (void) fprintf( stderr, "%s : le modle indiqu ne contient pas de spectre exprimental !\n", __FUNCTION__ );
    return -2;
  }

  /* Le modle "gnral" sera enregistr lors de l'appel de MODELE_PreparerAjustement
     mais on le fait un coup avant car on a besoin de x_min pour contrler la validit du DE0 */
  LMOD_EnregistrerModele( modele );
  /* On enregistre les paramtres spcifiques : couches */
  COUCHE_EnregistrerModele();

  /* On vrifie que le modle est utilisable */
  code_erreur = NC_ControleModele( modele->infos_modele.feff,
				   (modele->x_min > modele->experience->x_min) ? (modele->x_min) : (modele->experience->x_min),
				   &nbr_variables );
  if ( code_erreur != 0 ) {
    (void) fprintf( stderr, "%s : erreur dans NC_ControleModele [code %d]...\n",
		    __FUNCTION__, code_erreur );
    if ( VRAI ==  modele->drapeaux.isole ) { /* Effacer le modle cr localement */
      /* Comme c'est une copie locale d'un modle existant, pas de chanage ni d'infos  effacer... */
      free( modele ); modele = NULL;
    }
    return -7;
  }

  /* On prpare les graphes utiles au modle */
  code_erreur = MODELE_PreparerAjustement( modele, Q_MODELE_ANCIENS_RESULTATS_DEMANDER );
  if ( code_erreur != 0 ) {
    (void) fprintf( stderr, "%s : erreur dans MODELE_PreparerAjustement [code %d]...\n",
		    __FUNCTION__, code_erreur );
    if ( VRAI == modele->drapeaux.isole ) { /* Effacer le modle cr localement */
      /* Comme c'est une copie locale d'un modle existant, pas de chanage ni d'infos  effacer... */
      free( modele ); modele = NULL;
    }
    return -8;
  }

  /* On initialise le modle */
  code_erreur = NC_PreparerModele( modele->infos_modele.feff );
  if ( code_erreur != 0 ) {
    (void) fprintf( stderr, "N%s : erreur dans NC_PreparerModele [code %d]...\n",
		    __FUNCTION__, code_erreur );
    if ( modele->modele != NULL )
      modele->modele->drapeaux.protege = FAUX;
    if ( modele->residu != NULL )
      modele->residu->drapeaux.protege = FAUX;
    NC_LibereModele( modele, VRAI );
    if ( VRAI ==  modele->drapeaux.isole ) { /* Effacer le modle cr localement */
      /* Comme c'est une copie locale d'un modle existant, pas de chanage ni d'infos  effacer... */
      free( modele ); modele = NULL;
    }
    return -9;
  }

  /* Tout s'est bien pass */
  modele->nbr_variables = nbr_variables; /* Utilises dans l'ajustement... */
#ifdef _DEVELOPPE_
  printf( "Sortie normale de ==> %s <== [ponderation : %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere );
#endif
  return 0;
}

/* -------- Ajuster un spectre, sans connatre le modle -------- */

int NC_AjusterSpectre(struct graphe *reference)
{
  Absorbeur *absorbeur;
  Fenetre *fenetre;
  ModeleAjustement *modele = NULL;
  int code_erreur;

  if ( NULL == reference ) return -1;

  /* Le spectre doit tre du type convenable !
     Actuellement, modle FEFF ==> exafs (ou TF) */
  if ( ( reference->type != GR_TF ) && ( reference->drapeaux.exafs == FAUX ) 
    && ( reference->type != GR_SSTYPE ) ) { /* Graphes sans type peuvent servir  tout... */
    NC_ERREUR( TYPE_INCORRECT );
    (void) fprintf( stderr, "%s : graphe de type incorrect [%d, EXAFS = %d]\n",
		    __FUNCTION__, reference->type, reference->drapeaux.exafs);
    return -2;
  }

  /* On recherche le modle actif */
  if (NULL == mol_courante) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s : pas de molcule => pas de modle FEFF disponible\n", __FUNCTION__ );
#endif
    (void) BI_ERR_alerte( principale, &ERR_AucuneMolecule, VRAI );
    return -3;
  }

  if (NULL == mol_courante->absorbeurs) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s : pas d'absorbeur dans la molcule courante => pas de modle FEFF disponible\n", __FUNCTION__ );
#endif
    (void) BI_ERR_alerte( principale, &ERR_AucunAbsorbeur, VRAI );
    return -4;
  }

  absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
  if ( NULL == absorbeur->modele_actif ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s : pas de modle dans l'absorbeur courant => pas de modle FEFF disponible\n", __FUNCTION__ );
#endif
    (void) BI_ERR_alerte( principale, &ERR_AbsorbeurSansModele, VRAI );
    return -5;
  }
  if ( absorbeur->modele_actif->type != MODELE_EXAFS_CHEMINS ) {
    (void) fprintf(stderr, "%s,%s l.%u : modle courant pas de type FEFF...\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  if ( NULL == absorbeur->modele_actif->experience ) {
#ifdef _DEVELOPPE_
    (void) printf( "%s : modle courant sans spectre -> on lui assigne le spectre devant tre ajust\n", __FUNCTION__ );
#endif
    absorbeur->modele_actif->experience = reference;
  }

  if ( reference != absorbeur->modele_actif->experience ) {
    int reponse = -1;
    Fenetre *fenetre;
    struct modele_ajuste *modele_precedent;

    /* Est-ce que par hasard un modle temporaire existerait ? */
    fenetre = ARB_QuelleFenetre( arbres + RESULT_N_COUCHES, FAUX );
    modele_precedent = (struct modele_ajuste *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
    if ( modele_precedent != NULL  ) {
      if ( reference == modele_precedent->experience ) {
	/* On avait dj ajust ce modle */
	if ( VRAI == modele_precedent->drapeaux.isole ) {
	  /* On ajustait dj sur un modle temporaire : on continue !
	     ==> on perd les rsultats prcdents et on redemandera un temporaire */
	  reponse = 2;
	} else {
	  /* On ajustait un vrai modle ! */
	  reponse = BI_ERR_alerte( NULL, &Q_ChoisirModele, VRAI );
	  if ( 1 == reponse ) {
	    /* On ajuste sur le mme modle qu'avant */
	    modele = modele_precedent;
	    reponse = -1;
	  } else if ( 2 == reponse ) {
	    /* On ajuste sur le nouveau modle */
	    reponse = 1;
	  }
	}
      } else {
	/* L'ancien modle n'a aucun intrt pour nous a priori : mme cas que si pas de modle avant... */
	reponse = BI_ERR_alerte( NULL, &Q_RemplacerReference, VRAI );
      }
    } else {
      /* Aucun modle ajust prcdemment : que fait-on ? */
      reponse = BI_ERR_alerte( NULL, &Q_RemplacerReference, VRAI );
    }

    /* Doit-on remplacer le graphe de rfrence, ou faire un modle temporaire ? */
    if ( 3 == reponse ) {
      return +1; /* On annule... */
    }
    if ( 1 == reponse ) { /* On remplace */
      modele = absorbeur->modele_actif;
      reponse = MODELE_MettreReference( modele, reference, QUESTION );
      if ( reponse != 0 ) {
	AVERTISSEMENT( "ERREUR dans MODELE_MettreReference() --- code %d\n", reponse);
	return reponse;
      }
    } else if ( 2 == reponse ) { /* On fait un modle temporaire */
      /* Un modle isol puisque l'on travaille sur le graphe courant... */
      modele = MODELE_CreerModele( NULL, "Temporaire...", MODELE_EXAFS_CHEMINS, reference, FAUX );
      if ( NULL == modele ) {
	ERR_Memoire( "Modle local", sizeof ( *modele ) );
	return -6;
      }

      /* Mais avec le modle actif => on "copie" ses infos */
      modele->infos_modele.feff = absorbeur->modele_actif->infos_modele.feff;
      modele->drapeaux.isole = VRAI;

      modele->x_min = absorbeur->modele_actif->x_min;
      modele->x_max = absorbeur->modele_actif->x_max;
      modele->algorithme = absorbeur->modele_actif->algorithme;
      modele->drapeaux.pondere = absorbeur->modele_actif->drapeaux.pondere;
      modele->drapeaux.contraintes = absorbeur->modele_actif->drapeaux.contraintes;
    } else {
      /* On garde l'ancien modle : rien de spcial  faire... */
    }
  } else { /* On travaille directement sur le modle actif, puisque c'est son graphe que l'on ajuste... */
    modele = absorbeur->modele_actif;
  }

#ifdef _DEVELOPPE_
  printf("Modle choisi : %s\n", modele->nom_modele);
  printf("Exprience    : %s\n", modele->experience->nom);
  printf("Couches       : %u\n", modele->infos_modele.feff->nbr_couches);
#endif

  /* On prpare l'ajustement de ce modle nouvellement cr */
  code_erreur =  NC_AjusterModele( modele );
  if ( 0 != code_erreur ) {
    (void) fprintf( stderr, "%s : erreur dans NC_AjusterModele [code %d]\n",
		    __FUNCTION__, code_erreur );
    return code_erreur;
  }

  /* On ouvre la fentre contenant les rsultats */
  fenetre = NC_ouvre( modele );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "NC_AjusterSpectre : erreur dans NC_ouvre, impossible d'ouvrir la fentre\n");
    return -10;
  }

  /* On prpare les infos de chaque variable ajuste, en fonction de l'algorithme
     ATTENTION, si russi, l'ajustement est considr comme ayant dmarr
        ==> bien penser  le marquer comme arrt si annulation ensuite !
   */
  code_erreur = MODELE_InitVariables( modele );
  if ( 0 != code_erreur ) {
    (void) fprintf(stderr, "NC_AjusterSpectre : erreur dans MODELE_InitVariables, impossible d'ajuster\n");
    return -15;
  }

  if ( ( VRAI == reference->drapeaux.exafs ) || /* On ajuste un spectre EXAFS */
       ( (GR_SSTYPE == reference->type) && ( GRF_FCT_REELLE == reference->genre_donnees ) ) ) {
    NC_AjusterEXAFS( modele, fenetre );
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Contrle pralable d'un modle FEFF -------- */

int NC_ControleModele( Modele_FEFF *modele, double x_min, unsigned int *nbr_var )
{
  double de_min;
  unsigned int n_var = 0, n_couches = 0;
  Couche *couche;

  if ( NULL == modele ) return -1;

  if ( NULL == modele->couches ) {
    (void) BI_ERR_alerte(NULL, &ERR_ModeleSansCouche, FAUX);
    (void) fprintf(stderr, "NC_ControleModele : modle sans couche !\n");
    return -2;
  }

  /* Valeur minimale possible pour DE avec le k_min indiqu, pour ne pas avoir de k' ngatif */
  de_min = x_min * H_BARRE * M_V_ANGSTROM;
  de_min *= de_min;
  de_min /= -2.0 * MASSE_ELECTRON * EV_V_JOULE;

  /* Contrle couche par couche */
  for (couche = modele->couches ; couche != NULL ; couche = couche->suivante) {
    if ( FAUX == couche->utilisee ) continue;

    if ( NULL == couche->amplitude ) { /* Pas d'amplitude ou pas de dphasage */
      (void) BI_ERR_alerte(NULL, &ERR_CheminSansAmplitude, FAUX, couche->nom);
      couche->utilisee = FAUX;
      continue;
    }
    if ( NULL == couche->phase ) { /*  ==> inutilisable dans l'ajustement */
      (void) BI_ERR_alerte(NULL, &ERR_CheminSansPhase, FAUX, couche->nom);
      couche->utilisee = FAUX;
      continue;
    }

    n_couches++; /* Une couche utilise en plus */
    /* Vrification des variables */
    if ( verifie_variable(&(couche->N), "N", couche->nom) == VAR_LIBRE )
      n_var++;
    if ( verifie_variable(&(couche->R), "R", couche->nom) == VAR_LIBRE )
      n_var++;
    if ( verifie_variable(&(couche->Debye_Waller), "Facteur de Debye-Waller", couche->nom) == VAR_LIBRE )
      n_var++;
    if ( verifie_variable(&(couche->Cumul_1), "Premier Cumulant", couche->nom) == VAR_LIBRE )
      n_var++;
    if ( verifie_variable(&(couche->Cumul_2), "Deuxi"e_GRAVE"me cumulant", couche->nom) == VAR_LIBRE )
      n_var++;
    if ( verifie_variable(&(couche->Cumul_3), "Troisi"e_GRAVE"me cumulant", couche->nom) == VAR_LIBRE )
      n_var++;
    if ( verifie_variable(&(couche->Cumul_4), "Quatri"e_GRAVE"me cumulant", couche->nom) == VAR_LIBRE )
      n_var++;

    /* Les deux autres ne servent que dans des cas particuliers */
    if ( (FAUX == modele->drapeaux.DE0_general) &&
	 (verifie_deltae(&(couche->Delta_E), couche->nom, de_min) == VAR_LIBRE) )
      n_var++;
    if ( (FAUX == modele->drapeaux.GAM_general) && (modele->type_lpm != MODELE_LPM_INCLUS) &&
	 (verifie_variable(&(couche->Gamma), GAMMA_MAJUSCULE, couche->nom) == VAR_LIBRE) )
      n_var++;
  }

  if ( 0 == n_couches ) {
    (void) BI_ERR_alerte(NULL, &ERR_ModeleSansCouche, FAUX);
    (void) fprintf(stderr, "NC_ControleModele : aucune couche retenue dans le modle !\n");
    return -5;
  }

  /* Variables gnrales ? */
  if ( verifie_variable( &(modele->S0_2), "S0"DEUX_EXPOSANT, "Variables g"e_AIGU"n"e_AIGU"rales") == VAR_LIBRE )
    n_var++; 
  if ( (VRAI == modele->drapeaux.DE0_general) &&
       (verifie_deltae( &(modele->Delta_E), "Variables g"e_AIGU"n"e_AIGU"rales", de_min ) == VAR_LIBRE) )
    n_var++;

  /* Le cas du libre parcours moyen */
  if ( MODELE_LPM_GRAPHE == modele->type_lpm ) {
    if ( NULL == modele->lpm ) {
      (void) fprintf(stderr, "NC_ControleModele : graphe du libre parcours moyen absent !\n");
      modele->type_lpm = MODELE_LPM_INCLUS;
    }
  } else if ( modele->type_lpm != MODELE_LPM_INCLUS ) {
    if ( (VRAI == modele->drapeaux.GAM_general) &&
	 (verifie_variable(&(modele->Gamma), GAMMA_MAJUSCULE, "Variables g"e_AIGU"n"e_AIGU"rales") == VAR_LIBRE) )
      n_var++;
  }

  if (0 == n_var) { /* On prvient, mais on continue quand mme... */
    (void) fprintf(stderr, "NC_ControleModele : attention, modle sans variable ajustable.\n");
  }

  /* On renvoie le nombre de variables dans l'ajustement */
  if ( nbr_var != NULL )
    *nbr_var = n_var;

  return 0;
}

/* -------- Prparation d'un modle : on prpare phases et amplitudes --------
   Remarque : Suppose que le modle a t vrifi avec NC_ControleModele avant
 */

int NC_PreparerModele(Modele_FEFF *modele)
{
  Couche *couche;

  if ( NULL == modele ) return -1;
  if ( NULL == modele->couches ) { /* Devrait jamais arriver quand on est l! */
    (void) BI_ERR_alerte(NULL, &ERR_ModeleSansCouche, FAUX);
    (void) fprintf(stderr, "%s,%s l.%u : modle sans couche\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  for (couche = modele->couches ; couche != NULL ; couche = couche->suivante) {
    if ( FAUX == couche->utilisee ) continue;

    if ( couche->c_amp == NULL ) { /* Les splines n'ont pas dj t calcules */
      int retour;
      
      retour = SPL_tt_prep( couche->amplitude, &(couche->b_amp), &(couche->c_amp), &(couche->d_amp) );
      if ( retour < 0) {
	if ( NULL == couche->amplitude ) {
	  (void) LIS_ERREUR( retour, "<Pas de graphe amplitude>" );
	} else {
	  (void) LIS_ERREUR(retour, couche->amplitude->nom);
	} 
	couche->utilisee = FAUX;
	return -3; /* Erreur en prparant la spline des amplitudes */
      }

      retour = SPL_tt_prep( couche->phase, &(couche->b_pha), &(couche->c_pha), &(couche->d_pha) );
      if ( retour < 0) {
	if ( NULL == couche->phase ) {
	  (void) LIS_ERREUR( retour, "<Pas de graphe de phase>" );
	} else {
	  (void) LIS_ERREUR( retour, couche->phase->nom );
	}
	
	couche->utilisee = FAUX;
	free(couche->b_amp); couche->b_amp = NULL;
	free(couche->c_amp); couche->c_amp = NULL;
	free(couche->d_amp); couche->d_amp = NULL;
	return -4; /* Erreur en prparant la spline des phases */
      }
    }

    couche->amplitude->drapeaux.protege = VRAI;
    couche->phase->drapeaux.protege = VRAI;
  }

  /* ventuellement, coefficients des splines pour le lpm */
  /* On prpare pour le libre parcours moyen */
  if ( MODELE_LPM_GRAPHE == modele->type_lpm ) {
    int retour;

    if (modele->b_lpm != NULL) free(modele->b_lpm); /* Libration des anciens coefficients */
    if (modele->c_lpm != NULL) free(modele->c_lpm);
    if (modele->d_lpm != NULL) free(modele->d_lpm);

    retour = SPL_tt_prep(modele->lpm, &(modele->b_lpm), &(modele->c_lpm), &(modele->d_lpm));
    if ( retour != 0 ) {
      (void) LIS_ERREUR( retour, modele->lpm->nom);
      modele->type_lpm = MODELE_LPM_INCLUS; /* On suppose que le libre parcours moyen est inclus aux amplitudes */
    } else
      modele->lpm->drapeaux.protege = VRAI;
  }

  if ( modele->phase_central != NULL ) {
    modele->phase_central->drapeaux.protege = VRAI;
  }

  return 0; /* Tout s'est bien pass */
}

/* -------- Fin d'ajustement : libration de la mmoire (mais pas forcment des rsultats) -------- */

int NC_LibereModele(struct modele_ajuste *modele, boolean graphes)
{
  Couche *couche;
  Modele_FEFF *modele_feff;

  if ( NULL == modele ) return -1;
  modele_feff = modele->infos_modele.feff;

  /* Libration des coefficients de splines
     Dprotection des graphes */
  for (couche = modele_feff->couches ; couche != NULL ; couche = couche->suivante) {
    if ( NULL != couche->amplitude ) {
      free(couche->b_amp); couche->b_amp = NULL;
      free(couche->c_amp); couche->c_amp = NULL;
      free(couche->d_amp); couche->d_amp = NULL;
      couche->amplitude->drapeaux.protege = FAUX;
    }

    if ( NULL != couche->phase ) {
      free(couche->b_pha); couche->b_pha = NULL;
      free(couche->c_pha); couche->c_pha = NULL;
      free(couche->d_pha); couche->d_pha = NULL;
      couche->phase->drapeaux.protege = FAUX;
    }
  }

  if ( modele_feff->lpm != NULL ) {
    modele_feff->lpm->drapeaux.protege = FAUX;
    free(modele_feff->b_lpm); modele_feff->b_lpm = NULL;
    free(modele_feff->c_lpm); modele_feff->c_lpm = NULL;
    free(modele_feff->d_lpm); modele_feff->d_lpm = NULL;
  }
  if ( modele_feff->phase_central != NULL ) {
    modele_feff->phase_central->drapeaux.protege = FAUX;
  }

  /* Libration ventuelle des graphes */
  if ( VRAI == graphes ) {
    if ( (modele->reference) != (modele->experience) ) {
      GRF_eff_isole( modele->reference );
    }
    /* On vrifie la protection : si protg, c'est (par exemple) que l'on ajuste sur le rsultat de cet ajustement ! */
    GRF_efface(modele->modele, VRAI, FAUX);
    modele->modele = NULL;
    GRF_efface(modele->residu, VRAI, FAUX);
    modele->residu = NULL;
  }

  return 0;
}

/* ------ Quand on a fini un ajustement, on met tout comme il faut ------ */

/* ~~~~~~~~ Les briques de l'ajustement ~~~~~~~~ */

/* -------- Ouverture de la fentre d'affichage des rsultats ----------\
 |   Renvoie cette fentre, NULL si l'ouverture n'a pas pu tre faite.  |
 |                                                                      |
 | modele : le modle  ajuster                                         |
 \--------------------------------------------------------------------- */

/*
   Le drapeau util_1 de la fentre vaut VRAI <=> il y a plusieurs infos d'ajustement.
              util_2 de la fentre vaut VRAI <=> au moins 1 spectre avec plusieurs absorbeurs
 */

Fenetre *NC_ouvre(struct modele_ajuste *modele)
{
  Fenetre *fenetre;

  if ( NULL == modele ) return NULL;

  /* On vrifie que la fentre n'est pas dj ouverte */
  fenetre = ARB_QuelleFenetre( arbres + RESULT_N_COUCHES, FAUX );
  if ( NULL == fenetre ) {/* Il faut ouvrir la fentre */
    /* On cre la fentre */
    fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, NULL, NULL,
				  RESULT_N_COUCHES, BI_FEN_ARBRE_GAUCHE, prepare_arbre_ncouches,
				  RDS_ncouches,
				  ( int (*)(Fenetre *, unsigned int, int) ) NULL,
				  ( int (*)(Fenetre *, int, int, int, int, int) ) NULL,
				  MSG_ncouches,
				  ( void (*)(Fenetre *, int, int, int, int) ) NULL,
				  ( int (*)(Fenetre *) ) NULL );
    if (fenetre == NULL) /* Problme  la cration de la fentre */
      return NULL;
    fenetre->msg_masque.fermeture = VRAI; /* Message pour la fermeture de la fentre */

    FEN_Ouvrir( fenetre );
  } else {
    struct modele_ajuste *vx_modele;

    /* On rcupre les infos de l'ajustement prcdent */
    vx_modele = (struct modele_ajuste *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
    if ( NULL != vx_modele ) { /* Il y en a bien... */
      /* On vrifie que l'ajustement prcdent est bien termin */
      if ( VRAI == vx_modele->drapeaux.en_cours ) {
	OPTION_A_FAIRE( "Utiliser une vraie bote de dialogue !" );
	/*                 v----------------------------v */
	form_alert(1, "[3][---- Ajustement en cours ----"
		   "| Impossible de faire plusieurs"
		   "|ajustement en m"e_CIRC"me temps ; un"
		   "|ajustement tourne d"e_AIGU"j"a_GRAVE".][Annuler]");
	(void) fprintf(stderr, "NC_ouvre : il y a dj un ajustement en cours, sur le modle %s\n",
		       vx_modele->nom_modele);
	return NULL;
      }

      /* Que faire des rsultats pour le modle prcdemment ajust ?
	 La question n'a de sens que si l'ancien modle tait temporaire !
	 Sinon, elle a dj t pose dans MODELE_PreparerAjustement [normalement]
      */
      if ( VRAI == vx_modele->drapeaux.isole ) {
#ifdef ancien
	int reponse;

	reponse = NC_ERREUR(RELANCER_AJUSTEMENT);
	if (reponse == 3) return NULL; /* On annule le lancement */

	if (2 == reponse) { /* On demande  effacer les anciens graphes */
	  boolean retour;

	  if ( vx_modele != modele ) { /* On change de modle */
	    /* Rq : si utiliss dans un autre modle, ils seront protgs donc... */
	    retour = GRF_efface(vx_modele->modele, VRAI, FAUX);
	    if ( VRAI == retour )
	      vx_modele->modele = NULL;
	    retour = GRF_efface(vx_modele->residu, VRAI, FAUX);
	    if ( VRAI == retour )
	      vx_modele->residu = NULL;
	  } else {
	  }
	}
#endif

	/* Si l'ancien modle tait un modle temporaire, et que ce n'est pas le mme que le nouveau, on le libre */
	if ( modele != vx_modele ) {
	  free( vx_modele ); vx_modele = NULL; /* Modle isol, plus besoin ; pas de chanage */
	}
      }

      /* On dtache au besoin l'ancien modle de l'arbre, pour tre tranquille */
      if ( modele != vx_modele ) {
	ARB_MettreComplement ( RESULT_N_COUCHES, NULL );
      }

      /* Remarque : inutile de librer les informations par MODELE_TerminerAjustement, car
	 elle a dj t appele en fin de minimisation
	 VOIR QUAND MME CE QUI SE PASSERA AVEC INTERROMPRE ACTIF !*/
      /* 	MODELE_TerminerAjustement( vx_modele, reponse == 2); */
    }

    /* On passe la fentre au premier plan */
    FEN_premierplan( fenetre );
  }
  FEN_titre( fenetre, titre_fenetre_ajustement );
  FEN_infos( fenetre, infos_fenetre_ajustement );

  /* On associe les infos pour l'ajustement  la fentre */
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_1, modele );
  ARB_MettreComplement ( RESULT_N_COUCHES, modele );

  /* On affiche dans l'arbre les valeurs initiales du modle actif */
  NC_AfficheValeurs( modele );
  OBJ_active(RESULT_N_COUCHES, RNC_ABS_AVANT, FORCE_NON);
  /* Premire couche ajuste */
  couche_affichee = modele->infos_modele.feff->couches;
  OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_AVANT, FORCE_NON);
  OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_APRES, FORCE_OUI);
  while ( FAUX == couche_affichee->utilisee ) {
    couche_affichee = couche_affichee->suivante;
    if ( NULL == couche_affichee ) {
      OBJ_active( RESULT_N_COUCHES, RNC_COUCHE_APRES, FORCE_NON );
      break;
    }
  }

/*   OBJ_active(RESULT_N_COUCHES, RNC_ABS_APRES, infos->nbr_modeles > 1); */

#ifdef _DEVELOPPE_
  printf( "Sortie normale de ==> %s <== [ponderation : %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere );
#endif
  return fenetre; /* L'ouverture a russi */
}

/* -------- Cas particulier de verifie_variable pour un Delta(E0) -------- */

static boolean verifie_deltae(Variable *variable, char *couche, double de_min)
{
  if (variable == NULL) return FAUX;

  if ( (variable->type == VAR_CONSTANTE) ||
       (variable->type == VAR_COPIE) )
    return VAR_CONSTANTE; /* Inutile pour l'ajustement */

  if (variable->min < de_min) {
    NC_ERREUR(DEMIN_INCORRECT, couche);
    variable->min = de_min;
    if (variable->min > variable->max)
      variable->max = de_min;
  }

  return verifie_variable(variable, DELTA_MAJUSCULE"E0", couche);
}

/* -------- Vrifie qu'une variable est correctement initialise, donc utilisable --------\
 |    Renvoie VAR_LIBRE si elle est utilisable et variable, FAUX sinon                    |
 |            VAR_CONSTANTE si elle est utilisable mais non variable                      |
 | nom, couche : nom de la variable et de la couche dans laquelle elle intervient         |
 \--------------------------------------------------------------------------------------- */

static boolean verifie_variable(Variable *variable, char *nom, char *couche)
{
  if (variable == NULL) return FAUX;

  if ( (variable->type == VAR_CONSTANTE) ||
       (variable->type == VAR_COPIE) )
    return VAR_CONSTANTE; /* Inutile pour l'ajustement */

  if (variable->type == VAR_LIBRE ) {
    if ( variable->min == variable->max ) {
      NC_ERREUR(INIT_INCORRECTE, couche, nom,
		variable->valeur_initiale, variable->min, variable->max);
      variable->type = VAR_CONSTANTE;
      return VAR_CONSTANTE;
    }

    if ( (variable->valeur_initiale < variable->min) ||
	 (variable->valeur_initiale > variable->max) ) {
      NC_ERREUR(INIT_INCORRECTE, couche, nom,
		variable->valeur_initiale, variable->min, variable->max);
      variable->valeur_initiale = (variable->min + variable->max) / 2.0;
    }

    variable->contenu.valeur = variable->valeur_initiale;
    return VAR_LIBRE; /* Variable utilisable dans l'ajustement */
  }

  return FAUX; /* Variable pas encore dfinie... */
}

/* ************************************************ *
 !              Interface graphique                 !
 * ************************************************ */

/* -------- Redessin de la fentre -------- */

void RDS_ncouches(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  struct modele_ajuste *modele;
  double echelle_X, echelle_Y;
  int retour;
  Graphe *graphe;
  InfosRegion2D region;

  if (fenetre == NULL) return;

  /* Prparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  modele = (struct modele_ajuste *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == modele ) {
    TRC_Message( fenetre, "Aucun mod"e_GRAVE"le -- redessin impossible" );
    return;
  }

  graphe = modele->experience;
  if (graphe == NULL) {
    TRC_Message( fenetre, "Aucun spectre exp"e_AIGU"rimental -- redessin impossible" );
    return; /* Peut arriver si bote d'alerte : redessin avant la fin de l'initialisation */
  }

  if ( graphe->x_max == graphe->x_min ) {
    TRC_Message( fenetre, "Graphe constant (en X) -- redessin impossible" );
    return;
  }
  if ( graphe->y_max == graphe->y_min ) {
    TRC_Message( fenetre, "Graphe constant (en Y) -- redessin impossible" );
    return;
  }

  /* On prpare la rgion 2D : aide pour les dessins aprs en fait */
  retour = TRC_Z2D_Initialiser( &region, fenetre, graphe );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la rgion 2D [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* On trace le graphe  modliser dans la rgion 2D et on rcupre les chelles au passage */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, graphe, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &echelle_X, &echelle_Y );

  /* On trace la courbe exprimentale - en noir */

  /* On trace le spectre calcul - en rouge */
  if ( ( OBJ_choisit(RESULT_N_COUCHES, RSNC_AJUSTE, QUESTION) == VRAI )
    && ( modele->modele != NULL ) ) { /* Cas o alerte -> redessin avant la fin de l'initialisation */
    /* Avec les mmes chelles */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, modele->modele, ROUGE, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );
  }

  /* On trace le spectre diffrence - en bleu */
  if ( ( OBJ_choisit(RESULT_N_COUCHES, RNC_RESIDU, QUESTION) == VRAI )
    && ( modele->residu != NULL ) ) {
    TRC_GrapheDansRegion2D( fenetre, &region, 0, modele->residu, BLEU, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );
  }
}

/* -------- Messages pour la fentre ------ */

int MSG_ncouches(Fenetre *fenetre, Message *message)
{
  struct modele_ajuste *modele;

  if (fenetre == NULL)
    return BI_EVENEMENT_TRAITE;

  modele = (struct modele_ajuste *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );

  if ( message->type == WM_CLOSED ) { /* Fermeture de la fentre */
    if ( (modele != NULL) && (modele->drapeaux.en_cours == VRAI) ) { /* Ajustement en cours : veut-on l'arrter ? */
      NC_ERREUR(AJUSTEMENT_EN_COURS); /* Impossible de fermer si un ajustement est en cours */
      return BI_EVENEMENT_TRAITE;
    }

    /* Si on est l : l'ajustement a bien t arrt, la fentre peut tre ferme
       On a dj appel MODELE_TerminerAjustement() en fin de minimisation... */

    /* On peut de nouveau effacer des couches */
/*     OBJ_active(PARAM_COUCHES, EFFACER_COUCHE, premiere_couche != NULL); */
/*     OBJ_active(PARAM_COUCHES, TTEFFACER_COUCHE, premiere_couche != NULL); */

    FEN_Detruire( fenetre );
    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "Message non trait [%d]", message->type );
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Affichage des valeurs du modle indiqu dans l'arbre des rsultats -------- */

void NC_AfficheValeurs(struct modele_ajuste *modele)
{
  if ( NULL == modele ) return;

  /* Infos gnriques pour tout type de modle */
  OBJ_MettreEntier_U( RESULT_N_COUCHES, RNC_NBR_PARAMS, modele->nbr_variables );

  NC_AffichageIteration(modele);
}

/* -------- Affichage des rsultats  chaque tour de l'ajustement -------- */

static void NC_AffichageIteration(struct modele_ajuste *modele)
{
  Modele_FEFF *feff;
  double erreur;

#ifdef _DEVELOPPE_
/*   printf("On entre dans ==> NC_AffichageIteration <==\n"); */
#endif

  /* Informations "techniques" */
  OBJ_MettreEntier_U( RESULT_N_COUCHES, RNC_ITERATION, modele->iteration + 1 ); /* Itration en cours (commence  0) */

  /* Erreurs sur le modle */
  erreur = modele->erreur;
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_ERR_ABSOLUE, erreur, 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_ERR_PONDEREE, erreur / (double) (modele->reference->nombre_point), 0.0 );

  erreur = MODELE_CalculerErreur(modele, MODELE_ERREUR_CARRES, FAUX, VRAI);
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_ERR_RELATIVE, erreur, 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_ERR_REL_POND, erreur / (double) (modele->reference->nombre_point), 0.0 );

  /* valeurs des variables gnrales */
  /* Infos spcifiques d'un type de modle */
  switch (modele->type)
  {
  case MODELE_EXAFS_CHEMINS:
    feff = modele->infos_modele.feff;
    if (NULL == feff) return; /* Ne devrait jamais arriver */

    /*  OBJ_MettreEntier_U(RESULT_N_COUCHES, RNC_NCOUCHES, nbr_couches); *//* Nombre de couches dans le modle */

    /* 1. Variables gnrales, si ncessaire */
    OBJ_MettreNombre( RESULT_N_COUCHES, RNC_VAL_S02, VAR_LireValeur( &(feff->S0_2), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
    if ( VRAI == feff->drapeaux.DE0_general ) {
      OBJ_active( RESULT_N_COUCHES, RNC_DELTAE_GENE, FORCE_OUI);
      OBJ_MettreNombre( RESULT_N_COUCHES, RNC_DELTAE_GENE, VAR_LireValeur( &(feff->Delta_E), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0);
    } else {
      OBJ_active( RESULT_N_COUCHES, RNC_DELTAE_GENE, FORCE_NON);
    }
    if ( VRAI == feff->drapeaux.GAM_general ) {
      OBJ_active ( RESULT_N_COUCHES, RNC_GAMMA_GENE, FORCE_OUI);
      OBJ_MettreNombre( RESULT_N_COUCHES, RNC_GAMMA_GENE , VAR_LireValeur( &(feff->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
    } else {
      OBJ_active ( RESULT_N_COUCHES, RNC_GAMMA_GENE, FORCE_NON);
    }

    /* 2. Variables de la couche affiche */
    NC_AffichageCouche(couche_affichee);
    break;

  default:
    (void) fprintf(stderr, "NC_AfficheValeurs : pas d'affichage dtaill pour un modle de type %d\n", modele->type);
  }

#ifdef _DEVELOPPE_
/*   printf("Sortie normale de ==> NC_AffichageIteration <==\n"); */
#endif
}

/* -------- Affichage des valeurs des variables d'une couche -------- */

static void NC_AffichageCouche(struct couche *couche)
{
  if (NULL == couche) return;

  met_chaine ( RESULT_N_COUCHES, RNC_NOM_COUCHE  , 23, couche->nom, TXT_COPIER);
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_VAL_N       , VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_DEBYE_WALLER, VAR_LireValeur( &(couche->Debye_Waller), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_VAL_R       , VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_DELTAE_COUCH, VAR_LireValeur( &(couche->Delta_E), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_GAMMA_COUCHE, VAR_LireValeur( &(couche->Gamma), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_CUMUL1      , VAR_LireValeur( &(couche->Cumul_1), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_CUMUL_3     , VAR_LireValeur( &(couche->Cumul_3), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_CUMUL2      , VAR_LireValeur( &(couche->Cumul_2), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
  OBJ_MettreNombre( RESULT_N_COUCHES, RNC_CUMUL4      , VAR_LireValeur( &(couche->Cumul_4), VARIABLE_LIRE_VALEUR_AJUSTEE ), 0.0 );
}

/* -------- Gestion de l'arbre d'affichage des rsultats ------------ */

static int GER_ncouches(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  struct modele_ajuste *modele;
  Couche *couche = NULL;

  modele = (struct modele_ajuste *) ARB_LireComplement( RESULT_N_COUCHES );
  if (NULL == modele) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de modle associ  l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return BI_EVENEMENT_TRAITE;
  }

  switch (objet)
  {
  case RNC_COUCHE_AVANT: /* On affiche les valeurs pour la couche prcdente */
    /* Premire couche utilise */
    couche = modele->infos_modele.feff->couches;
    while (FAUX == couche->utilisee) {
      couche = couche->suivante;
      if ( NULL == couche ) { /* Aucune couche utilise ! */
	OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_AVANT, FORCE_NON);
	OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_APRES, FORCE_NON);
	return BI_EVENEMENT_TRAITE;
      }
    }
    OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_APRES, FORCE_OUI);
    if ( couche == couche_affichee ) {
      OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_AVANT, FORCE_NON);
    } else {
      Couche *active_avant;

      do {
	active_avant = couche;
	couche = couche->suivante;
	if (NULL == couche) {
	  return BI_EVENEMENT_TRAITE;
	}

	while (FAUX == couche->utilisee)
	  couche = couche->suivante;
      } while (couche != couche_affichee);

      couche_affichee = active_avant;
      couche = couche_affichee; /* Pour la suite ! */
    }
    NC_AffichageCouche(couche_affichee);
    return BI_EVENEMENT_TRAITE;

  case RNC_COUCHE_APRES: /* On affiche les valeurs pour la couche suivante */
    couche = couche_affichee;
    do {
      couche = couche->suivante;
      if (couche == NULL) {
	OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_APRES, FORCE_NON);
	return BI_EVENEMENT_TRAITE;
      }
    } while (FAUX == couche->utilisee);

    OBJ_active(RESULT_N_COUCHES, RNC_COUCHE_AVANT, FORCE_OUI);
    couche_affichee = couche;
    NC_AffichageCouche(couche_affichee);
    return BI_EVENEMENT_TRAITE;

  case RNC_RESIDU:
  case RSNC_AJUSTE:
    OBJ_choisit(RESULT_N_COUCHES, objet, ECHANGE);
    FEN_redessin( ARB_QuelleFenetre(arbres + RESULT_N_COUCHES, FAUX), BI_RDS_UTIL );
    return BI_EVENEMENT_TRAITE;

  case RNC_ARRETER: /* On force l'arrt de l'ajustement - pour l'instant, brutal... */
    modele->drapeaux.arret_force = VRAI;
    return BI_EVENEMENT_TRAITE;

  case RNC_CONTINUER: /* On relance l'ajustement. Suppose que les infos sont encore l,
		         donc on en fait que remettre les itrations  zro et relancer.
			 Comme termine_ncouches a t appel a priori, il faut rinit. les couches.
			 Pour l'instant, suppose une seule structure d'infos
		      */
    OPTION_A_FAIRE( "Reprise d'un ajustement interrompu" );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Prparation de l'arbre d'affichage des rsultats -------- */

static int prepare_arbre_ncouches(unsigned int numero_arbre __attribute((__unused__)))
{
  int retour;

  retour = ARB_Initialiser( RESULT_N_COUCHES, NULL, GER_ncouches, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'initialiser l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
#ifdef AVEC_TRADUCTION_
#endif

  arbres[RESULT_N_COUCHES].drapeaux.prepare = VRAI; /* Prparation termine */
  return 0;
}
