/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2003, Emmanuel CURIS

   Opérations diverses sur les graphes, plus générales que l'EXAFS lui-même
   (aussi bien mathématiques que statistiques simples : tri en particulier)
   (les opérations statistiques avancées sont dans ../Stats)

   Intégration d'une portion de graphe
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "graphes.h"
#include "operations.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double INT_Trapezes_R(Graphe *graphe, unsigned long premier, unsigned long nombre, double *erreur);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Graphes à valeurs réelles ~-~-~-~-~-~-~-~- */ 

/* -------- Intégration d'un graphe réel renvoie une constante, aire sous la courbe -------- */

double GRF_Integrer_R(Graphe *graphe, double x_min, double x_max, int methode, double *erreur)
{
  boolean bornes_inversees __attribute__((__unused__)) = FAUX;
  double aire = 0.0;
  unsigned long nbr_points, idx_premier;

  ENTREE_FONCTION("GRF_Integrer_R", "graphe=%p x_min=%g x_max=%g methode=%d erreur=%p", graphe, x_min, x_max, methode, erreur);

  /* Vérifications préliminaires */
  if ( NULL == graphe ) {
    SORTIE_FONCTION("GRF_Integrer_R", "%g [graphe = NULL]", 0.0);
    return 0.0;
  }
  if ( FAUX == GRF_EstFonctionReelle( graphe ) ) {
    (void) fprintf(stderr, "GRF_Integrer_R : Le graphe ne contient pas une fonction à valeurs réelles [%s]\n", graphe->nom);
    SORTIE_FONCTION("GRF_Integrer_R", "%g [graphe non réel]", 0.0);
    return 0.0;
  }

  /* Controles des bornes */
  if ( x_min == x_max ) { /* Bornes égales -> intégrale nulle par définition */
    if ( erreur != NULL ) *erreur = 0.0;
    SORTIE_FONCTION("GRF_Integrer_R", "%g [bornes égales]", 0.0);
    return 0.0;
  }
  if ( x_min > x_max ) { /* Intégrale(a->b) = -Intégrale(b->a) et on s'arrange pour avoir toujours a<b */
    double x;

    x = x_min ; x_min = x_max ; x_max = x;
    bornes_inversees = VRAI;
  }
  if ( ( x_max <= graphe->x_min ) || ( x_min >= graphe->x_max ) ) {
    (void) fprintf(stderr, "GRF_Integrer_R : Intervalle d'intégration [%g, %g] hors du graphe [%g, %g]\n", x_min, x_max, graphe->x_min, graphe->x_max);
    SORTIE_FONCTION("GRF_Integrer_R", "%g [hors-graphe]", 0.0);
    return 0.0;
  }
  if ( x_max > graphe->x_max ) x_max = graphe->x_max; /* On se restreint à la partie connue */
  if ( x_min < graphe->x_min ) x_min = graphe->x_min;

  /* On repère l'intervalle réellement utilisable */
  nbr_points = GRF_intervalle( graphe, x_min, x_max, &idx_premier );
  if ( 0 == nbr_points ) { /* Intervalle trop petit par rapport à la précision */
    (void) fprintf(stderr, "GRF_Integrer_R : Intervalle d'intégration trop petit [%g, %g]\n", x_min, x_max);
    SORTIE_FONCTION("GRF_Integrer_R", "%g [0 points]", 0.0);
    return 0.0;
  }

  /* Calcul proprement dit, en fonction de la méthode */
  switch ( methode )
  {
  case INTG_TRAPEZE:
    aire = INT_Trapezes_R( graphe, idx_premier, nbr_points, erreur );
    break;

  case INTG_RECTANGLE:
  case INTG_MEDIAN:
  case INTG_SIMPSON:
  case INTG_BODE:
  case INTG_SPLINE:
    (void) fprintf(stderr, "GRF_Integrer_R : À FAIRE --- Intégration par la méthode %d\n", methode);
    break;

  default:
    (void) fprintf(stderr, "GRF_Integrer_R : Méthode inconnue [%d]\n", methode);
    SORTIE_FONCTION("GRF_Integrer_R", "%g [méthode inconnue %d]", 0.0, methode);
    return 0.0;
  }

  /* Calcul fini : on renvoie les valeurs obtenues */
  SORTIE_FONCTION("GRF_Integrer_R", "%g", aire);
  return aire;
}

/* -------- Méthode des trapèzes [SUPPOSE LES X TRIÉS] -------- */

static double INT_Trapezes_R(Graphe *graphe, unsigned long premier, unsigned long nombre, double *erreur)
{
  double aire = 0.0, err_aire = 0.0, aire_i;
  unsigned long i;

  /* Calcul de l'aire par la méthode des trapèzes */
  for ( i = 1 ; i < nombre ; i++ ) {
    aire_i  = graphe->y[ premier + i - 1 ] + graphe->y[ premier + i ];  /* B + b */
    aire_i *= graphe->x[ premier + i ] - graphe->x[ premier + i - 1 ];  /* (B+b) * h */

    aire += aire_i;
  }
  aire /= 2.0; /* h*(B+b)/2 = aire du trapèze, mais on divise par 2 seulement à la fin : plus rapide ! */

  if ( NULL == erreur ) return aire; /* On ne s'intéresse pas aux erreurs */

  /* On s'occupe des erreurs éventuelles */
  if ( graphe->erreur_x != NULL ) {
    if ( graphe->erreur_y != NULL ) {
      printf("INT_Trapezes_R : À FAIRE --- Estimer l'erreur sur l'intégrale [erreurs en X]\n");
    } else {
      printf("INT_Trapezes_R : À FAIRE --- Estimer l'erreur sur l'intégrale [erreurs en X & Y]\n");
    }
  } else if ( graphe->erreur_y != NULL ) {
    printf("INT_Trapezes_R : À FAIRE --- Estimer l'erreur sur l'intégrale [erreurs en Y]\n");
  }

  /* On a fini */
  *erreur = err_aire;
  return aire;
}

