/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Création : 9 février 1999

   Calculs statistiques divers

   Calcul d'une analyse de variance à un facteur au sens large [donc régression linéaire & non-linéaire Y=f(X) incluses]
          des tests a posteriori de cette analyse
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "graphes.h"
#include "proba.h"
#include "statistiques.h"
#include "tri.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double       ANVA_1F_CalculerMoyennePredite(struct infos_ANVA_1F *infos_ANVA, double x);
static int          ANVA_1F_CalculsRL             (struct infos_ANVA_1F *infos_ANVA);
static int          ANVA_1F_EstimerParametres     (struct infos_ANVA_1F *infos_ANVA);
static unsigned int ANVA_1F_NombreParametres      (struct infos_ANVA_1F *infos_ANVA);

boolean ANVA_1F_EstEquilibree(struct infos_ANVA_1F *infos_ANVA)
{
  unsigned long i, n_valeurs;

  if ( ( NULL == infos_ANVA ) || ( 0 == infos_ANVA->facteur.nbr_modalites ) || ( NULL == infos_ANVA->facteur.modalites ) ) return FAUX;

  /* Nombre de valeurs dans la première modalité */
  if ( NULL == infos_ANVA->facteur.modalites[ 0 ].valeurs ) {
    n_valeurs = 0;
  } else {
    n_valeurs = infos_ANVA->facteur.modalites[ 0 ].valeurs->nombre_point;
  }

  /* On compare au nombre dans les autres */
  for ( i = 1 ; i < infos_ANVA->facteur.nbr_modalites ; i++ ) {
    unsigned long n;

    if ( NULL == infos_ANVA->facteur.modalites[ i ].valeurs ) {
      n = 0;
    } else {
      n = infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point;
    }

    if ( n != n_valeurs ) {
      /* Le nombre de valeurs est différent dans au moins une modalité : plan non-équilibré */
      return FAUX;
    }
  }

  /* Si on est là, toutes les modalités ont le même nombre de valeurs */
  if ( 0 == n_valeurs ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : Attention, toutes les modalités sont vides !\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
  }
  return VRAI;
}

/* -------- Calculs initiaux --------
    Renvoie 0    si tout s'est bien passé
            != 0 en cas de problème

    Les moyennes, variances, sommes par modalité sont stockées dans les modalités & dans le graphe final
    Les autres sommes sont dans la structuve infos_ANVA
 */

int ANVA_1F_Calculs(struct infos_ANVA_1F *infos_ANVA)
{
  double moyenne = 0.0, ecart;
  int retour;
  unsigned int i;
  unsigned long j;

  if ( NULL == infos_ANVA ) return -1;

  /* Les ddl faciles : terme factoriel */
  infos_ANVA->ddl_F = infos_ANVA->facteur.nbr_modalites - 1;

  /* Pour chaque modalité : calcul de la moyenne et de la contribution à la somme des carrés résiduelle */
  infos_ANVA->QR = 0.0;
  infos_ANVA->ddl_R = 0;
  infos_ANVA->ddl_T = 0;
  for ( i = 0 ; i < infos_ANVA->facteur.nbr_modalites ; i++ ) {
    retour = ANVA_MODALITE_CalculsInitiaux( infos_ANVA->facteur.modalites + i );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur dans ANVA_MODALITE_CalculsInitiaux() code %d\n",
		     __FILE__, __FUNCTION__, __LINE__, retour );
      if ( retour < 0 ) return -2;  /* Erreur grave ! */
      continue;			/* La modalité ne peut pas être utilisée pour les calculs */
    }

    /* On en profite pour calculer les éléments de la résiduelle */
    infos_ANVA->QR    += infos_ANVA->facteur.modalites[ i ].SCR;
    infos_ANVA->ddl_R += infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point - 1;

    /* Ainsi que la moyenne globale */
    infos_ANVA->ddl_T += infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point;
    moyenne += infos_ANVA->facteur.modalites[ i ].moyenne * (double) infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point;
  }

  /* Contrôle des degrés de liberté totaux [utilisés pour la moyenne globale] */
  if ( 0 == infos_ANVA->ddl_T ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune mesure ! Rien n'a de sens !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    infos_ANVA->ST = 0.0; infos_ANVA->QT = 0.0;
    infos_ANVA->SF = 0.0; infos_ANVA->QF = 0.0; infos_ANVA->ddl_F = 0;
    infos_ANVA->SR = 0.0; infos_ANVA->QR = 0.0; infos_ANVA->ddl_R = 0;

    infos_ANVA->rapport = -1.0;
    return +2;
  }
  infos_ANVA->ddl_T--;		/* Nombre de ddl totaux : nombre de mesures - 1 */

  if ( 0 == infos_ANVA->ddl_T ) {
    (void) fprintf(stderr, "%s, %s l.%u : Une seule mesure en tout et pour tout ! Toutes les variances sont nulles\n",
		   __FILE__, __FUNCTION__, __LINE__);
    infos_ANVA->ST = 0.0; infos_ANVA->QT = 0.0;
    infos_ANVA->SF = 0.0; infos_ANVA->QF = 0.0; infos_ANVA->ddl_F = 0;
    infos_ANVA->SR = 0.0; infos_ANVA->QR = 0.0; infos_ANVA->ddl_R = 0;

    infos_ANVA->moyenne = moyenne;  /* La moyenne est donc la seule valeur connue ! */

    infos_ANVA->rapport = -2.0;
    return +1;
  }

  /* Calculer la moyenne globale des modalités */
  moyenne /= (double) (infos_ANVA->ddl_T + 1);
  infos_ANVA->moyenne = moyenne;

  /* Calculer la somme des carrés totaux et, au passage, la factorielle */
  infos_ANVA->QF = 0.0;
  infos_ANVA->QT = 0.0;
  for ( i = 0 ; i < infos_ANVA->facteur.nbr_modalites ; i++ ) {
    if ( NULL == infos_ANVA->facteur.modalites[ i ].valeurs ) continue;	/* Modalité sans valeur : ne compte pas ! */

    /* Terme pour la somme des carrés factorielle */
    ecart = infos_ANVA->facteur.modalites[ i ].moyenne - moyenne;
    infos_ANVA->QF += ecart * ecart * (double) infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point;

    /* Termes pour la somme des carrés totale */
    for ( j = 0 ; j < infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point ; j++ ) {
      ecart = infos_ANVA->facteur.modalites[ i ].valeurs->y[ j ] - moyenne;
      infos_ANVA->QT += ecart * ecart;
    }

    /* On en profite pour estimer l'effet, pour chaque modalité [si qualitatif] */
    infos_ANVA->facteur.modalites[ i ].effet = ecart;
  }

  /* Contrôle des ddl */
  if ( 0 == infos_ANVA->ddl_F ) {
    (void) fprintf(stderr, "%s, %s l.%u : Une modalité : variance factorielle nulle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    infos_ANVA->SF = 0.0;
    infos_ANVA->rapport = 0.0;
    infos_ANVA->proba_critique = 1.0;
    return +2;
  }
  if ( 0 == infos_ANVA->ddl_R ) {
    (void) fprintf(stderr, "%s, %s l.%u : Une seule mesure pour chaque modalité : variance résiduelle nulle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    infos_ANVA->SR = 0.0;
    if ( 0.0 == infos_ANVA->SF ) {
      /* On a plusieurs fois la même valeur : aucun effet du "facteur" */
      infos_ANVA->rapport = 0.0;
      infos_ANVA->proba_critique = 1.0;
    } else {
      /* On a au moins une valeur différente : effet du "facteur" */
      infos_ANVA->rapport = HUGE_VAL;
      infos_ANVA->proba_critique = 0.0;
    }
    return +3;
  }

  /* On calcule, à partir des degrés de liberté, les variances */
  infos_ANVA->SR = infos_ANVA->QR / infos_ANVA->ddl_R;
  infos_ANVA->SF = infos_ANVA->QF / infos_ANVA->ddl_F;
  infos_ANVA->ST = infos_ANVA->QT / infos_ANVA->ddl_T;

  /* On calcule le rapport */
  if ( 0.0 == infos_ANVA->SR ) {
    /* Pas de variance résiduelle : tout est expliqué par le facteur */
    if ( 0.0 == infos_ANVA->SF ) {
      /* On a plusieurs fois la même valeur : aucun effet du "facteur" */
      infos_ANVA->rapport = 0.0;
      infos_ANVA->proba_critique = 1.0;
    } else {
      /* On a au moins une valeur différente : effet du "facteur" */
      infos_ANVA->rapport = HUGE_VAL;
      infos_ANVA->proba_critique = 0.0;
    }
  } else {
    infos_ANVA->rapport = infos_ANVA->SF / infos_ANVA->SR;
    infos_ANVA->proba_critique = PROBA_FdR_FisherSnedecor( infos_ANVA->rapport, (double) infos_ANVA->ddl_F, (double) infos_ANVA->ddl_R );
  }

  if ( infos_ANVA->facteur.proprietes & ANVA_FACTEUR_QUANTITATIF ) {
    retour = ANVA_1F_CalculerModele( infos_ANVA );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Problème dans le calcul du modèle ! [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }
  }

  /* Calculs terminés sans souci */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Calculs dans le cas d'une ANVA sur variable quantitative ~-~-~-~-~-~-~-~-~ */

int ANVA_1F_CalculerModele(struct infos_ANVA_1F *infos_ANVA)
{
  int retour;
  unsigned long i;

  if ( NULL == infos_ANVA ) return -1;
  if ( FAUX == ( infos_ANVA->facteur.proprietes & ANVA_FACTEUR_QUANTITATIF ) ) {
    AVERTISSEMENT( "Appel pour un facteur qualitatif ! [%s]", infos_ANVA->facteur.nom );
    return +1;			/* Erreur, mais mineure */
  }

  /* On estime les paramètres du modèle */
  retour = ANVA_1F_EstimerParametres( infos_ANVA );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur en estimant les paramètres [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -5;
  }

  /* On en déduit le graphe avec les valeurs modèles */
  if ( infos_ANVA->resultat != NULL ) {
    if ( infos_ANVA->resultat->nombre_point != infos_ANVA->facteur.nbr_modalites ) {
      /* C'est que l'on a changé le nombre de modalités entre-temps... */
      /* On efface donc pour le récréer ci-dessous */
      GRF_eff_isole( infos_ANVA->resultat ); infos_ANVA->resultat = NULL;
    }
  }
  if ( NULL == infos_ANVA->resultat ) {	 /* Pas else car on peut avoir effacé un graphe déjà existant */
    infos_ANVA->resultat = GRF_nv_isole( infos_ANVA->facteur.nbr_modalites, GR_SSTYPE, GRF_FCT_REELLE );
    if ( NULL == infos_ANVA->resultat ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de créer le graphe des résultats !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -10;
    }
  }

  /* Pour chaque modalité, on calcule la valeur de Y prédite par le modèle */
  for ( i = 0, infos_ANVA->QE = 0.0, infos_ANVA->QM = 0.0 ; i < infos_ANVA->facteur.nbr_modalites ; i++ ) {
    double x, ecart;

    x = infos_ANVA->facteur.modalites[ i ].valeur;
    infos_ANVA->resultat->x[ i ] = x;
    infos_ANVA->resultat->y[ i ] = ANVA_1F_CalculerMoyennePredite( infos_ANVA, x );

    /* On en profite pour calculer les sommes de carrés */
    /* a. Prédits par le modèle : entre moyenne prédite/modalité et moyenne globale */
    ecart = infos_ANVA->resultat->y[ i ] - infos_ANVA->moyenne;
    infos_ANVA->QM += ecart * ecart * (double) infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point;

    /* b. Écarts au modèle : entre moyenne prédite/modaltié et moyenne observée/modalité */
    ecart = infos_ANVA->resultat->y[ i ] - infos_ANVA->facteur.modalites[ i ].moyenne;
    infos_ANVA->QE += ecart * ecart * (double) infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point;
  }

  /* On calcule les degrés de liberté de chaque terme */
  infos_ANVA->ddl_M = ANVA_1F_NombreParametres( infos_ANVA ) - 1;  /* Nombre de paramètres - 1 */
  infos_ANVA->ddl_E = infos_ANVA->ddl_F - infos_ANVA->ddl_M;

  /* On en déduit les variances associées */
  if ( infos_ANVA->ddl_M > 0 ) {
    infos_ANVA->SM = infos_ANVA->QM / (double) infos_ANVA->ddl_M;
  } else {
    infos_ANVA->SM = -1.0;	/* Le modèle est complètement fixé : la variance expliquée par le modèle est infinie ! */
  }

  if ( infos_ANVA->ddl_E > 0 ) {
    infos_ANVA->SE = infos_ANVA->QE / (double) infos_ANVA->ddl_E;
  } else {
    infos_ANVA->SE = 0.0;	/* Il n'y a aucune latitude pour vérifier la validité du modèle : la variance expliquée par les écarts est nulle */
  }

  /* Et on en déduit les rapports */
  if ( infos_ANVA->SR > 0.0 ) {
    infos_ANVA->RE   = infos_ANVA->SE / infos_ANVA->SR;
    infos_ANVA->pc_E = PROBA_FdR_FisherSnedecor( infos_ANVA->RE, (double) infos_ANVA->ddl_E, (double) infos_ANVA->ddl_R );

    infos_ANVA->RM   = infos_ANVA->SM / infos_ANVA->SR;
    infos_ANVA->pc_M = PROBA_FdR_FisherSnedecor( infos_ANVA->RM, (double) infos_ANVA->ddl_M, (double) infos_ANVA->ddl_R );
  } else {
    /* Pas de variance résiduelle : on utilise alors les écarts au modèle comme estimation de la variance... */
    if ( infos_ANVA->SE > 0.0 ) {
      infos_ANVA->RE   = 1.0;
      infos_ANVA->pc_E = PROBA_FdR_FisherSnedecor( 1.0, (double) infos_ANVA->ddl_E, (double) infos_ANVA->ddl_E );

      infos_ANVA->RM   = infos_ANVA->SM / infos_ANVA->SE;
      infos_ANVA->pc_M = PROBA_FdR_FisherSnedecor( infos_ANVA->RM, (double) infos_ANVA->ddl_M, (double) infos_ANVA->ddl_E );
    } else {
      /* Pas de variance due aux écarts non-plus : modèle parfait ! */
      infos_ANVA->RE   = 0.0;
      infos_ANVA->pc_E = 1.0;

      infos_ANVA->RM   = -1.0;
      infos_ANVA->pc_M = -2.0;
    }
  }

  /* On trie le graphe par X croissants, pour faciliter le dessin */
  GRF_Trier( infos_ANVA->resultat, GRF_TRI_RAPIDE );

  /* On en déduit les bornes, en X et en Y */
  GRF_bornes( infos_ANVA->resultat, ORDONNE, DESORDONNE );

  /* Pour info : contrôles */
  printf("==> QF - (QE + QM) = %g\n", infos_ANVA->QF - (infos_ANVA->QE + infos_ANVA->QM) );

  /* On a terminé, tout s'est bien passé */
  return 0;
}

/* -------- Calculer la valeur prédite par le modèle pour le X indiqué -------- */

static double ANVA_1F_CalculerMoyennePredite(struct infos_ANVA_1F *infos_ANVA, double x)
{
  if ( NULL == infos_ANVA->modele.rl ) return 0.0;  /* Pas de modèle défini ! */

  switch ( infos_ANVA->type_modele )
  {
  case ANVA_MODELE_DROITE:
    return infos_ANVA->modele.rl->pente * x + infos_ANVA->modele.rl->o_a_lo;

  case ANVA_MODELE_POLYNOME:
    OPTION_A_FAIRE( "Modèle polynomial" );
    return 0.0;

  case ANVA_MODELE_CL:
    OPTION_A_FAIRE( "Modèle de combinaison linéaire" );
    return 0.0;

  case ANVA_MODELE_NL:
    OPTION_A_FAIRE( "Modèle non-linéaire quelconque" );
    return 0.0;

  default:
    AVERTISSEMENT( "Modèle inconnu [%u]", infos_ANVA->type_modele );
  }

  /* On ne devrait être ici qu'en cas de problème sur un modèle ! */
  return 0.0;
}

/* -------- Calculs dans le cas d'une régression linéaire de base -------- */

static int ANVA_1F_CalculsRL(struct infos_ANVA_1F *infos_ANVA)
{
  double S_covariance = 0.0;
  unsigned long i, j;

  /* On calcule la somme des X : donne la moyenne */
  infos_ANVA->modele.rl->moyenne_X = 0.0;
  for ( i = 0 ; i < infos_ANVA->facteur.nbr_modalites ; i++ ) {
    if ( infos_ANVA->facteur.modalites[ i ].valeurs != NULL ) {
      double x;

      x = infos_ANVA->facteur.modalites[ i ].valeur;
      infos_ANVA->modele.rl->moyenne_X += x * (double) infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point;

      printf("Modalité %lu : X = %g, %lu valeurs\n", i, x, infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point );
    }
  }
  infos_ANVA->modele.rl->moyenne_X /= (double) ( infos_ANVA->ddl_T + 1 );

  /* On calcule la somme des écarts pour X : donne la variance ; la somme des produits (X-E(X))*(Y-E(Y)) : donne la covariance */
  infos_ANVA->modele.rl->SCE_X = 0.0;
  for ( i = 0 ; i < infos_ANVA->facteur.nbr_modalites ; i++ ) {
    if ( infos_ANVA->facteur.modalites[ i ].valeurs != NULL ) {
      double x, ecart_X;

      x = infos_ANVA->facteur.modalites[ i ].valeur;
      ecart_X = x - infos_ANVA->modele.rl->moyenne_X;
      for ( j = 0 ; j < infos_ANVA->facteur.modalites[ i ].valeurs->nombre_point ; j++ ) {
	infos_ANVA->modele.rl->SCE_X += ecart_X * ecart_X;
	S_covariance += ( ecart_X ) * ( infos_ANVA->facteur.modalites[ i ].valeurs->y[ j ] - infos_ANVA->moyenne );
      }
    }
  }

  /* Calcul de la pente */
  if ( infos_ANVA->modele.rl->SCE_X <= 0.0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Une seule valeur en X, impossible de déterminer une droite !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }
  infos_ANVA->modele.rl->pente = S_covariance / infos_ANVA->modele.rl->SCE_X;

  /* Calcul de l'ordonnée à l'origine */
  infos_ANVA->modele.rl->o_a_lo = infos_ANVA->moyenne - infos_ANVA->modele.rl->pente * infos_ANVA->modele.rl->moyenne_X;

  /* Affichage des contrôles */
#ifdef _DEVELOPPE_
  printf( "X     : moyenne = %g ---- SCE = %g\n", infos_ANVA->modele.rl->moyenne_X, infos_ANVA->modele.rl->SCE_X );
  printf( "(X,Y) : SPE     = %g donc cov = %g\n", S_covariance, S_covariance / (double) ( infos_ANVA->ddl_T + 1 ) );
  printf( "droite: Y = %g * X + %g\n", infos_ANVA->modele.rl->pente, infos_ANVA->modele.rl->o_a_lo ); 
#endif

  /* Terminé sans souci */
  return 0;
}

/* -------- Ajuster le modèle pour estimer ses paramètres ------ */

static int ANVA_1F_EstimerParametres(struct infos_ANVA_1F *infos_ANVA)
{
  if ( NULL == infos_ANVA->modele.rl ) {
    (void) fprintf(stderr, "%s, %s l.%u : Le modèle n'a pas encore été défini !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On effectue la minimisation en fonction du type de modèle considéré */
  switch ( infos_ANVA->type_modele )
  {
  case ANVA_MODELE_DROITE:
    return ANVA_1F_CalculsRL( infos_ANVA );

  case ANVA_MODELE_POLYNOME:
    OPTION_A_FAIRE( "Modèle polynomial" );
    return -2;

  case ANVA_MODELE_CL:
    OPTION_A_FAIRE( "Modèle de combinaison linéaire" );
    return -2;

  case ANVA_MODELE_NL:
    OPTION_A_FAIRE( "Modèle non-linéaire quelconque" );
    return -2;

  default:
    AVERTISSEMENT( "Modèle inconnu [%u]", infos_ANVA->type_modele );
  }

  /* On ne devrait être ici qu'en cas de problème sur un modèle ! */
  return -5;
}

/* -------- Nombre de paramètres dans le modèle indiqué -------- */

static unsigned int ANVA_1F_NombreParametres(struct infos_ANVA_1F *infos_ANVA)
{
  if ( NULL == infos_ANVA->modele.rl ) return 0;  /* Pas de modèle défini ! */

  switch ( infos_ANVA->type_modele )
  {
  case ANVA_MODELE_DROITE:
    return 2;

  case ANVA_MODELE_POLYNOME:
    OPTION_A_FAIRE( "Modèle polynomial" );
    return 0;

  case ANVA_MODELE_CL:
    OPTION_A_FAIRE( "Modèle de combinaison linéaire" );
    return 0;

  case ANVA_MODELE_NL:
    OPTION_A_FAIRE( "Modèle non-linéaire quelconque" );
    return 0;

  default:
    AVERTISSEMENT( "Modèle inconnu [%u]", infos_ANVA->type_modele );
  }

  /* On ne devrait être ici qu'en cas de problème sur un modèle ! */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Routines de gestion des modèles ~-~-~-~-~-~-~-~-~ */

/* -------- Libérer les informations du modèle -------- */

int ANVA_1F_LibererModele (struct infos_ANVA_1F *infos_ANVA);

/* -------- Réserver et initialiser les informations du modèle -------- */

int ANVA_1F_PreparerModele(struct infos_ANVA_1F *infos_ANVA, unsigned int type)
{
  if ( NULL == infos_ANVA ) return -1;
  if ( infos_ANVA->modele.rl != NULL ) {
    /* On a déjà des infos pour un modèle */
    if ( type == infos_ANVA->type_modele ) {
      /* Rien à faire en réalité ! */
      return 0;
    }
    AVERTISSEMENT( "Il y a déjà un modèle [type %u] défini. Veuillez le libérer d'abord", infos_ANVA->type_modele );
    return -1;
  }

  /* Si on est là, il faut bien réserver le modèle */
  switch( type )
  {
  case ANVA_MODELE_DROITE:
    infos_ANVA->modele.rl = ( struct modele_droite * ) malloc( sizeof( struct modele_droite ) );
    if ( NULL == infos_ANVA->modele.rl ) {
      ERR_Memoire( "Impossible de pr"e_AIGU"parer le mod"e_GRAVE"le", sizeof( struct modele_droite ) );
      return -4;
    }

    infos_ANVA->modele.rl->covariance = 0.0;
    infos_ANVA->modele.rl->moyenne_X  = 0.0;
    infos_ANVA->modele.rl->pente      = 0.0;
    infos_ANVA->modele.rl->o_a_lo     = 0.0;
    break;

  case ANVA_MODELE_POLYNOME:
    OPTION_A_FAIRE( "Réserver pour un polynôme" );
    break;

  case ANVA_MODELE_CL:
    OPTION_A_FAIRE( "Réserver pour une CL de fonctions de base" );
    break;

  case ANVA_MODELE_NL:
    OPTION_A_FAIRE( "Réserver pour un modèle non-linéaire" );
    break;

  default:
    AVERTISSEMENT( "Type de modèle inconnu [%u]", type );
    return -2;
  }

  /* Si on est là, tout s'est bien passé */
  infos_ANVA->type_modele = type;
  return 0;
}
