/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2003, Emmanuel CURIS
   Cration de ce fichier : 12 novembre 2003

   Module d'analyse statistique
   Travail avec les lois discrtes
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "systeme.h"

#include "math_utils.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#undef _DEVELOPPE_

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

#define PROBA_ReserverTable(nbr_valeurs)   probas = (double *) malloc( sizeof( double ) * (nbr_valeurs) );\
  if ( NULL == probas ) {\
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de rserver les %lu probabilits\n",\
		   __FILE__, __FUNCTION__, __LINE__, ( nbr_valeurs ) );\
    return NULL;\
  }

#ifdef _DEVELOPPE_
static int STAT_ControlerProbas(unsigned int nbr_valeurs, double *probabilites);
#endif

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

#ifdef _DEVELOPPE_
/* -------- Contrler que les valeurs indiques sont plausibles -------- */
static int STAT_ControlerProbas(unsigned int nbr_valeurs, double *probabilites)
{
  unsigned int i;
  double somme = 0.0;

  for ( i = 0 ; i < nbr_valeurs ; i++ ) {
    if ( ( probabilites[ i ] > 1.0 ) || ( probabilites[ i ] < 0.0 ) ) {
      (void) fprintf(stderr, "%s, %s l.%u : Valeur absurde, p(X=%i)=%g\n",
		     __FILE__, __FUNCTION__, __LINE__, i, probabilites[ i ]);
      return -1;
    }
    somme += probabilites[ i ];
  }

  if ( somme != 1.0 ) {
    if ( fabs( somme - 1.0 ) < 1.e-8 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreurs d'arrondi dans les probabilit... [%g au lieu de 1, D=%g]\n",
		     __FILE__, __FUNCTION__, __LINE__, somme, 1.0 - somme);
      return 0;
    }
    (void) fprintf(stderr, "%s, %s l.%u : La somme des probabilits ne vaut pas 1 [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, somme);
    return -2;
  }

  /* Valeur acceptables */
  return 0;
}
#endif

/* ~-~-~-~-~-~-~-~-~ Tirer des nombres suivant la loi discrte fini indique ~-~-~-~-~-~-~-~-~ */

/* -------- Tirage d'un seul nombre : indique le numro de la valeur -------- */

unsigned int STAT_TirerValeurDiscrete(unsigned int nbr_valeurs, double *probabilites)
{
#ifdef _DEVELOPPE_
  int retour;
#endif
  double valeur, somme = 0.0;
  unsigned int i = 0;

  if ( 1 == nbr_valeurs ) { /* Une seule valeur possible => on tire forcment celle-l ! */
    return 1.0;
  }
  if ( ( NULL == probabilites ) || ( 0 == nbr_valeurs ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas de table des probabilits indique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

#ifdef _DEVELOPPE_
  retour = STAT_ControlerProbas( nbr_valeurs, probabilites );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Somme des probabilits non-nulles ou valeur absurde\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return 0;
  }
#endif

  /* On tire un nombre, uniformment, entre 0 et 1 */
  valeur = uniforme();
/*   printf("%s : Valeur tire = %g (p[0]=%g)\n", __FUNCTION__, valeur, probabilites[ 0 ] ); */

  /* Si 0 <= valeur < p( X < X[0] ), alors on a tir X[ 0 ] */
  if ( valeur < probabilites[ 0 ] ) return 0;

  /* Si p( X <= X[i] ) <= valeur < p( X < X[i + 1] ), alors on a tir la valeur X[ i ] */
  somme = probabilites[ 0 ];
  while ( valeur > somme ) {
    i++;
    somme += probabilites[ i ];
  }
/*   printf("%s : Valeur tire = %g (p[%u]=%g/somme=%g)\n", __FUNCTION__, valeur, i, probabilites[ i ], somme ); */

  /* Ici, on a donc somme <= valeur et somme = p[0] + p[1] + ... + p[i] = p( X <= X[i] ) */
  return i;
}

/* ~-~-~-~-~-~-~-~-~ Lois usuelles ~-~-~-~-~-~-~-~-~ */

/* -------- Loi binomiale -------- */

double *PROBA_LoiBinomiale(unsigned long n, double p)
{
  double *probas;
  unsigned long i;

  if ( 0 == n ) {
    (void) fprintf(stderr, "%s, %s l.%u : Paramtre absurde [n = %lu] !\n",
		   __FILE__, __FUNCTION__, __LINE__, n);
    return NULL;
  }
  if ( ( p < 0.0 ) || ( p > 1.0 ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Probabilit absurde [p = %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, p);
    return NULL;
  }

  PROBA_ReserverTable( n + 1 );

  for ( i = 0 ; i <= n ; i++ ) {
    probas[ i ] = CoefficientBinomial( n, i ) * Puissance( p, i ) * Puissance( 1.0 - p, n - i );
  }

  return probas;
}

/* -------- Loi uniforme -------- */

double *PROBA_LoiUniformeDiscrete(unsigned long nbr_valeurs)
{
  double *probas;
  unsigned long i;

  if ( 0 == nbr_valeurs ) {
    (void) fprintf(stderr, "%s, %s l.%u : Valeur absurde !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  PROBA_ReserverTable( nbr_valeurs );

  for ( i = 0 ; i < nbr_valeurs ; i++ ) {
    probas[ i ] = 1.0 / (double) nbr_valeurs;
  }

  return probas;
}
