/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Cration : 9 fvrier 1999

   Calculs statistiques divers

   Calcul des paramtres statistiques d'une loi  partir d'un graphe
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "stats_utils.h"

#ifndef VERSION_SIMPLIFIEE_
# include "exafs.h"
#endif

#include "statistiques.h"
#include "science.h"     /* Pour la fonction B */
#include "tri.h"

#include "parametres.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

struct param_stat {
  unsigned long effectif;	/* Nombre de valeurs disponibles */
  double min, max;		/* Valeurs extrmes */
  double etendue;		/* tendue (= max-min) */

  double moyenne;		/* Moyenne arithmtique : estimation de l'esprance */
  double variance_ech;		/* Variance de l'chantillon */
  double variance_pop;		/* Variance estime dans la population */

  double mediane;		/* Mdiane de l'chantillon */
  double ecart_absolu_moyen;	/* carts absolus moyens  la mdiane */
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int STAT_Parametres_R(Graphe *graphe, struct param_stat *parametres);
static int STAT_Parametres_H(Graphe *graphe, struct param_stat *parametres);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Calcul des paramtres statistiques d'une loi ---------------------\
 |   Renvoie la valeur moyenne                                                |
 | Suppose que les Y reprsentent les diffrentes valeurs d'une mme mesure   |
 |             les X ne sont pas utiliss                                     |
 |                                                                            |
 | calcule la moyenne, la variance, les coefficients gamma1 & 2 de Fisher     |
 \--------------------------------------------------------------------------- */

double calc_param_stat(Graphe *graphe)
{
  int retour;
  struct param_stat parametres;

  if (graphe == NULL) return 0.0;

  /* Estimation de l'esprance et de la variance (grandeurs les plus usuelles) */
  switch ( graphe->genre_donnees )
  {
  case GRF_FCT_REELLE:
  case GRF_SERIE:
    retour = STAT_Parametres_R( graphe, &parametres );
    break;

  case GRF_HISTOGRAMME:
    retour = STAT_Parametres_H( graphe, &parametres );
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Calcul de paramtres statistiques impossible sur ce genre de graphe [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->genre_donnees );
    retour = -20;
    break;
  }

  if ( retour != 0 ) {
    if ( retour < 0 ) {		/* Erreur grave */
      return 0.0;
    }

    /* Problme partiel : certains rsultats seront douteux */
  }

  HIST_Titre(1, "Param"e_GRAVE"tres statistiques", "Stats");
  HIST_Texte(1, 0, "S"e_AIGU"rie de valeurs : %s\n", graphe->nom);
  HIST_Texte(2, 0, "Nombre de valeurs : %lu\n", graphe->nombre_point);
  HIST_Texte(3, 0, "Moyenne estim"e_AIGU"e   : m = %.20f\n", parametres.moyenne);
  HIST_Texte(3, 0, "Variance estim"e_AIGU"e  : s*"DEUX_EXPOSANT" = %.20f soit s* = %.20f\n", parametres.variance_pop, sqrt( parametres.variance_pop ));

#ifdef _UNIX_ /* Appel depuis un shell => affichage texte intressant */
  printf("Moyenne : %.6g, intervalle approch"e_AIGU" "a_GRAVE" 95%% : [%.5g, %.5g]\n", parametres.moyenne,
	 parametres.moyenne - 1.96 * sqrt( parametres.variance_pop / (graphe->nombre_point - 1)),
	 parametres.moyenne + 1.96 * sqrt( parametres.variance_pop / (graphe->nombre_point - 1)));
  printf("Variance : %.6g - "e_AIGU"cart-type : %.6g\n", parametres.variance_pop, sqrt( parametres.variance_pop ));

  printf("Valeurs extr"e_CIRC"mes        : %.10g et %.10g\n", parametres.min, parametres.max);
  printf("Estimation de l'"e_AIGU"tendue : %.10g\n", parametres.etendue );
  printf("M"e_AIGU"diane          : %.10g\n", parametres.mediane);
  printf(E_AIGU"cart absolu moyen : %.10g\n", parametres.ecart_absolu_moyen);
#endif

  HIST_Texte(3, 0, "Valeurs extr"e_CIRC"mes : min = %.10g ; MAX = %.10g\n", parametres.min, parametres.max);
  HIST_Texte(3, 0, E_AIGU"tendue            : %.10g\n", parametres.etendue);
  HIST_Texte(3, 0, "M"e_AIGU"diane          : %.10g\n", parametres.mediane);
  HIST_Texte(3, 0, E_AIGU"cart absolu moyen : %.10g\n", parametres.ecart_absolu_moyen);

  (void) BI_ERR_alerte(NULL, &ERR_affiche_valeurs, VRAI,
		       parametres.moyenne,
		       parametres.moyenne - 1.9599 * sqrt( parametres.variance_pop / (double) parametres.effectif ),
		       parametres.moyenne + 1.9599 * sqrt( parametres.variance_pop / (double) parametres.effectif ),
		       parametres.variance_pop, sqrt( parametres.variance_pop ),
/*		       gamma_1, gamma_2, */
		       parametres.min, parametres.max, parametres.etendue,
		       parametres.mediane, parametres.ecart_absolu_moyen );
  return parametres.moyenne;
}

/* ~-~-~-~-~-~-~-~- Calculs pour un graphe (X, Y) ou une srie : table des Y ~-~-~-~-~-~-~-~- */

static int STAT_Parametres_R(Graphe *graphe, struct param_stat *parametres)
{
  double *y, *y_trie;
  unsigned long N, i;

  /* Le nombre de valeurs est le nombre de points du graphe */
  parametres->effectif = graphe->nombre_point;

  /* Les valeurs extrmes sont dj connues */
  parametres->min = graphe->y_min;
  parametres->max = graphe->y_max;
  parametres->etendue = parametres->max - parametres->min;

  /* Estimation de l'esprance : moyenne arithmtique de l'chantillon */
  (void) STAT_Moyenne_D( graphe->y, graphe->nombre_point, &(parametres->moyenne) );

  /* Estimation de la variance */
  parametres->variance_pop = variance( graphe, parametres->moyenne );

  /* Variance vraie */
  parametres->variance_ech = ( parametres->variance_pop * (double) ( parametres->effectif - 1 ) ) / (double) ( parametres->effectif );

  /* Calcul de la mdiane : il faut les valeurs tries */
  y = (double *) calloc(graphe->nombre_point, sizeof(double));
  if ( NULL == y ) {
    SYS_ERREUR(MANQUE_MEMOIRE, "Stats/parametres.c", "calc_param_stat", "y");
    return +1;
  }
  memcpy(y, graphe->y, sizeof(double) * graphe->nombre_point);

  y_trie = TRI_Tableau( y, graphe->nombre_point );
  if (NULL == y_trie ) {
    free( y ); y = NULL;
    return +1;
  }
  y = y_trie;
  N = graphe->nombre_point - 1; /* Index du dernier point dans le tableau */

  if ( (N & 1) == 0 ) { /* N pair, donc N+1 = nombre de points impair */
    parametres->mediane = y[N >> 1];
  } else { /* Nombre pair de points */
    parametres->mediane  = y[ (N + 1) >> 1 ] + y[ (N - 1) >> 1];
    parametres->mediane /= 2.0;
  }

  /* Calcul de l'cart absolu moyen */
  for (i = 0, y_trie = y; i <= N ; i++, y_trie++) {
    parametres->ecart_absolu_moyen += fabs( *y_trie - parametres->mediane );
  }
  parametres->ecart_absolu_moyen /= N;

  /* Plus besoin des Y tris */
  free( y ); y = NULL;

  /* Calcul de la moyenne gomtrique = exp( moyenne des logarithmes ) */
  {
    double m_geo = 0.0;

    for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
      if ( 0.0 == graphe->y[ i ] ) {
	m_geo = 0.0;
	printf( "** Moyenne gomtrique nulle **\n" );
	break;
      } else {
	m_geo += log( fabs( graphe->y[ i ] ) );
      }
    }
    m_geo = exp( m_geo / (double) graphe->nombre_point );
    printf("** Moyenne gomtrique : %g **\n", m_geo );
  }

  /* Termin, sans incident */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Calculs pour un graphe histogramme ~-~-~-~-~-~-~-~- */

static int STAT_Parametres_H(Graphe *graphe, struct param_stat *parametres)
{
  unsigned long i;

  /* On calcule l'effectif total : histogramme mmoris en effectifs */
  parametres->effectif = 0;
  for ( i = 1 ; i < graphe->nombre_point ; i++ ) {
    parametres->effectif += graphe->z.effectifs[ i ];
  }
  /* Contrle : classes infinies */
  if ( ( 0 != graphe->z.effectifs[ 0 ] ) || ( 0 != graphe->z.effectifs[ graphe->nombre_point ] ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Classes extrmes non-vides, elles seront ngliges dans la suite !\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }
  /* Contrle : effectif non-nul */
  if ( 0 == parametres->effectif ) {
    (void) fprintf(stderr, "%s, %s l.%u : Effectif nul !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Valeurs extrmes (en supposant les classes ]-infini, x_min] et ]x_max, +infini[ vides) */
  parametres->min = graphe->x[ 0 ];
  parametres->max = graphe->x[ graphe->nombre_point - 1 ];

  /* tendue */
  parametres->etendue = parametres->max - parametres->min;

  /* Moyenne : en utilisant les milieux de classe */
  parametres->moyenne = 0.0;
  for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
    double milieu_classe;

    milieu_classe = ( graphe->x[ i ] + graphe->x[ i - 1 ] ) / 2.0;
    parametres->moyenne += milieu_classe * (double) graphe->z.effectifs[ i ];
  }
  parametres->moyenne /= (double) parametres->effectif;

  /* cart-type : en utilisant les milieux de classe */
  parametres->variance_ech = 0.0;
  for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
    double milieu_classe;
    double ecart;

    milieu_classe = ( graphe->x[ i ] + graphe->x[ i - 1 ] ) / 2.0;
    ecart = milieu_classe - parametres->moyenne;
    parametres->variance_ech += ecart * ecart * (double) graphe->z.effectifs[ i ];
  }
  if ( parametres->effectif > 1 ) {
    parametres->variance_pop  = parametres->variance_ech / (double) (parametres->effectif - 1);
  } else {
    parametres->variance_pop  = 0.0;
  }
  parametres->variance_ech /= (double) parametres->effectif;

  return 0;
}

/* -------- Calcul effectif de la moyenne ------------------------\
 | Utilise l'estimateur classique :                               |
 |                                                                |
 |        y1 + y2 + ... + yn                                      |
 | E(Y) = ------------------ (d'aprs la loi des grands nombres)  |
 |                n                                               |
 \--------------------------------------------------------------- */

double esperance(Graphe *graphe)
{
  double moyenne = 0.0, m_pond = 0.0;
  unsigned long i;
  register double *y;

  for (i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++) {
#ifdef _DEVELOPPE_
    m_pond += *y / (double) (graphe->nombre_point);
#endif
    moyenne += *y;
  }

#ifdef _DEVELOPPE_
  HIST_Texte(5, 5, "Moyenne : S(y/n) = %.20g, S(y)/n = %.20g\n", m_pond, moyenne / (double) (graphe->nombre_point));
#endif

  return moyenne / (double) (graphe->nombre_point);
}

/* -------- Calcul effectif de la moyenne ---------------------------------\
 | Utilise l'estimateur classique :                                        |
 |                        (y1 - E(Y)) + (y2 - E(Y)) + ... + (yn - E(Y)) |
 | V(Y) = E(Y) - E(Y) = ------------------------------------------------ |
 |                                            n - 1                        |
 \------------------------------------------------------------------------ */

double variance(Graphe *graphe, double esperance)
{
  double V = 0.0;
  unsigned long i;
  register double *y;

  for (i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++) {
    double delta = *y - esperance;
    V += delta * delta;
  }

  return V / (graphe->nombre_point - 1);
}

/* ~~~~~~~~ Estimation des liaisons entre deux grandeurs ~~~~~~~~ */

#ifndef VERSION_SIMPLIFIEE_

#define LGMAX_NOM_GRAPHES 34

typedef struct {
  Graphe *grf1, *grf2; /* les deux graphes affichs */
} Infos_liaison;

static int GER_liaison(unsigned int objet, int x, int y);

/* -------- Interface ------------------------------------------------\
 |   Renvoie, si on le demande, un graphe  nuage de points          |
 |                                                                    |
 | graphe1, graphe2 : les deux graphes dont on veut la liaison        |
 \------------------------------------------------------------------- */

Graphe *estime_liaison(Graphe *graphe1, Graphe *graphe2)
{
  Infos_liaison infos = {NULL, NULL};
  double covar, correl, moyenne_1, moyenne_2;
  double var1, var2;
  int reponse;
  extern const char *txt_aucun_choisi; /* Dans variables.c */
  unsigned long nbr_pts;
  Graphe *nuage = NULL;

  if ( (graphe2 != NULL) && (graphe1 == NULL) ) {
    graphe1 = graphe2; graphe2 = NULL;
  }
  
  if (graphe1 != NULL) {
    infos.grf1 = graphe1;
    met_chaine(ESTIME_CORREL, EC_GRAPHE_1, LGMAX_NOM_GRAPHES, graphe1->nom, TXT_COPIER);
  } else
    met_chaine(ESTIME_CORREL, EC_GRAPHE_1, LGMAX_NOM_GRAPHES, txt_aucun_choisi, TXT_COPIER);
  if (graphe2 != NULL) {
    infos.grf2 = graphe2;
    met_chaine(ESTIME_CORREL, EC_GRAPHE_2, LGMAX_NOM_GRAPHES, graphe2->nom, TXT_COPIER);
  } else
    met_chaine(ESTIME_CORREL, EC_GRAPHE_2, LGMAX_NOM_GRAPHES, txt_aucun_choisi, TXT_COPIER);
  OBJ_active(ESTIME_CORREL, EC_ESTIMER, (infos.grf1 != NULL) && (infos.grf2 != NULL) );

  if (arbres[ESTIME_CORREL].drapeaux.prepare == FAUX) {
    OBJ_deplace(ESTIME_CORREL, EC_TXT_GRAPHE_2, 0, +1, DEPLACE_PIXELS);
    OBJ_deplace(ESTIME_CORREL, EC_GRAPHE_2, 0, +1, DEPLACE_PIXELS);

    arbres[ESTIME_CORREL].fonction.arbre = GER_liaison;
    arbres[ESTIME_CORREL].drapeaux.prepare = VRAI;
  }

  ARB_MettreComplement( ESTIME_CORREL, &infos );
  reponse = ARB_Bloquant( ESTIME_CORREL, titre_arbre_correl, principale, 0 );
  ARB_MettreComplement( ESTIME_CORREL, NULL );

  if ( EC_ANNULER == reponse )
    return NULL;

  /* Normalement, si on arrive l, les deux graphes sont dfinis */
  moyenne_1 = esperance(infos.grf1);
  moyenne_2 = esperance(infos.grf2);
  covar = covariance(infos.grf1, infos.grf2, moyenne_1, moyenne_2);

  var1 = variance(infos.grf1, moyenne_1);
  var2 = variance(infos.grf2, moyenne_2);
  correl = sqrt( var1 * var2 );
  if (correl != 0.0) {
    correl = covar / correl;
  } else {
    EXF_ERREUR(DIV_ZERO, "Variance nulle", "r = 0.0"); /* L'une des var. est cte ps */
    correl = 0.0; /* donc forcment, il y a indpendance */
  }

  nbr_pts = infos.grf1->nombre_point;
  if (infos.grf2->nombre_point < nbr_pts) nbr_pts = infos.grf2->nombre_point;

  /* Ici, on regarde quelle est la probabilit que |R| > |r| trouv - doit tre maximale

                                                  n - 4
                                                  -----
                                                    2
                                          (1 - x)
     La densit de R est donn par f(x) = ------------- si |x| < 1, 0 sinon
                                            (1   n - 2)
                                           B(- ; -----)
                                            (2     2  )

     Vrai uniquement si couples gaussiens... */

  {
    double coef, x, somme;

    coef = beta(0.5, ((double) (nbr_pts - 2)) / 2.0 );
    x = fabs(correl);

    if ( (nbr_pts % 2) == 0 ) { /* n est pair : n = 2k + 4 */
      unsigned long Cnk = 1, i, k;
      double integrale, epsilon = 1.0, x_k = x;

      /* On a alors

         /|r|                                 2i + 1
         |           k        k    i     i |r|
	 |   (1 - x)  dx = Somme C  (-1)  ---------
         |                  i = 0  k        2i + 1
         /0

	 NB : calcul compar avec maple. Pour nbr_pts = 100, r = 1 : ok  4 chiffres
                                                             r = 5,8810^-2 : ok  5 chiffres
      */

      k = (nbr_pts - 4) / 2;
      /*      printf("k = %lu - x = %f\n", k, x); */
      integrale = x; /* Premier terme, toujours |r| */
      for (i = 1 ; i <= k ; i++) {
	epsilon *= -1.0;
	x_k *= x * x; /* Contient x^(2k+1) */
	Cnk = (k - i + 1) * Cnk / i;
	/*	printf("Cnk (48, %lu) = %lu, epsilon = %f\n", i, Cnk, epsilon); */
	integrale += ( (double) Cnk ) * epsilon * x_k / (double) (2 * i + 1);
      }

      /*      printf("intgrale : %.10g - fonction B : %.10g\n", integrale, coef); */
      somme = 2.0 * integrale / coef; /* *2 car correspond  -r < R < r */
    } else
      somme = 2.0; /* Pour montrer qu'il y a un problme */

    printf("P(|R| < |r|) = %.10g - doit tre minimale\n", somme);
    printf("P(|U| > |r|) = %.10g - doit tre maximale. Approximation LG.\n",
	   erfc( x * sqrt( (nbr_pts - 3) / 2.0 ) ) );
  }

  (void) BI_ERR_alerte(NULL, &ERR_affiche_covariance, VRAI, covar, correl);
#ifdef _UNIX_
  printf("Covariance : %.20g - Coefficient de corrlation linaire : %.20e\n", covar, correl);
#endif

  /* Ici, on cre un nuage de points  partir des donnes */
  nuage = GRF_nouveau(infos.grf2, nbr_pts, GR_SSTYPE, GR_NORMAL, "Nuage de points");
  if (nuage != NULL) {
    memcpy( nuage->x, infos.grf1->y, sizeof(double) * nbr_pts );
    memcpy( nuage->y, infos.grf2->y, sizeof(double) * nbr_pts );
    GRF_bornes(nuage, DESORDONNE, DESORDONNE);
    GRF_couleur( nuage, NOIR, 0, PM_CARRE );
  }

  return nuage;
}

static int GER_liaison(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  Infos_liaison *infos;
  Graphe *nv_graphe;

  infos = (Infos_liaison *) ARB_LireComplement( ESTIME_CORREL );

  switch (objet)
  {
  case EC_GRAPHE_1:
    nv_graphe = GRF_choisir(GR_TOUS, FAUX, VRAI, NULL);
    if ( (nv_graphe != NULL) && (nv_graphe != infos->grf1) && (nv_graphe != infos->grf2) ) {
      infos->grf1 = nv_graphe;
      met_chaine(ESTIME_CORREL, EC_GRAPHE_1, LGMAX_NOM_GRAPHES, nv_graphe->nom, TXT_COPIER);
      OBJ_active(ESTIME_CORREL, EC_ESTIMER, (infos->grf1 != NULL) && (infos->grf2 != NULL) );
    }
    return BI_EVENEMENT_TRAITE;

  case EC_GRAPHE_2:
    if (infos->grf1 != NULL)
      nv_graphe = GRF_choisir(infos->grf1->type, FAUX, VRAI, NULL);
    else
      nv_graphe = GRF_choisir(GR_TOUS, FAUX, VRAI, NULL);
    if ( (nv_graphe != NULL) && (nv_graphe != infos->grf1) && (nv_graphe != infos->grf2) ) {
      infos->grf2 = nv_graphe;
      met_chaine(ESTIME_CORREL, EC_GRAPHE_2, LGMAX_NOM_GRAPHES, nv_graphe->nom, TXT_COPIER);
      OBJ_active(ESTIME_CORREL, EC_ESTIMER, (infos->grf1 != NULL) && (infos->grf2 != NULL) );
    }
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

#endif /* VERSION_SIMPLIFIEE_ */

/* -------- Calcul de la covariance -------------\
 |  Renvoie la covariance observe               |
 |  On se cale sur la plus courte srie          |
 |                                               |
 | graphe1 : premire srie                      |
 | graphe2 : deuxime srie                      |
 | m1, m2  : moyennes calcules pour ces sries  |
 \---------------------------------------------- */

double covariance(Graphe *graphe1, Graphe *graphe2, double moyenne1, double moyenne2)
{
  unsigned long i, nbr_pts;
  register double *x, *y;
  double covar = 0.0;

  if ( (graphe1 == NULL) || (graphe2 == NULL) ) return 0.0;

  nbr_pts = graphe1->nombre_point;
  if (graphe2->nombre_point < nbr_pts) nbr_pts = graphe2->nombre_point;

  for (i = 0, x = graphe1->y, y = graphe2->y ; i < nbr_pts ; i++, x++, y++) {
    covar += (*x - moyenne1) * (*y - moyenne2);
  }
  /*                                      _n_
                                      1   \   (      _ )2 (      _ )2
    Covariance dans la population : -----  >  ( x  - x )  ( y  - y )
                                    n - 1 /__ (  i     )  (  i     )
                                          i=1
  */
  covar /= ( nbr_pts - 1 );

  return covar;
}
