/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'une description de configuration exprimentale
   Gestion des composants lmentaires d'un chantillon
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "chimie_utils.h" /* Analyse des formules chimiques... */
#include "absorption.h"
#include "echantillon.h"
#include "extrait.h" /* Pour cherche_e0 */
#include "fluorescence.h"
#include "graphes.h"
#include "interface.h" /* Pour le slecteur universel */
#include "lissage.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int COMPOSANT_Charger_I (void);
static int COMPOSANT_Copier_I  (unsigned long n_original);
static int COMPOSANT_Effacer_I (unsigned long n_original);
static int COMPOSANT_Nouveau_I (unsigned long n_avant);
static int COMPOSANT_Sauver_I  (unsigned long n_composant);

static const char       *COMPOSANT_TrouverNom(unsigned int arbre, unsigned long objet);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Gestion interne de la structure de donnes ~-~-~-~-~-~-~-~- */

struct composant *CreerComposant( struct composant *avant )
{
  struct composant *nouveau;

  /* Rservation */
  nouveau = ( struct composant * ) calloc( 1, sizeof( struct composant ) );
  if ( NULL == nouveau ) {
    (void) fprintf(stderr, "CreerComposant : Mmoire insuffisante\n");
    return NULL;
  }

  /* Chainage : Ajout aprs celui indiqu [NULL == dbut de la liste] */
  if ( NULL == avant ) {
    nouveau->suivant = premier_composant;
    premier_composant = nouveau;
  } else {
    nouveau->suivant = avant->suivant;
    avant->suivant = nouveau;
  }

  /* Initialisation */
  nouveau->densite = 1.0;
  nouveau->nom = strdup( "Sans nom" );

  /* Tout s'est bien pass */
  return nouveau;
}

int EffacerComposant(struct composant *composant, boolean forcer)
{
  if ( NULL == composant ) return -1;

  if ( FAUX == forcer ) {
    (void) fprintf(stderr, "%s,%s l.%u : Poser la question\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  /* On te du chanage */
  if ( composant == premier_composant ) {
    premier_composant = composant->suivant;
  } else {
    struct composant *avant;

    avant = premier_composant;
    while ( avant->suivant != composant ) {
      avant = avant->suivant;
      if ( NULL == avant ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR, composant hors-chanage\n",
		       __FILE__, __FUNCTION__, __LINE__);
	break;
      }
    }

    avant->suivant = composant->suivant;
  }
  composant->suivant = NULL;

  /* On l'enlve des compositions pouvant l'utiliser */
  if ( premiere_composition != NULL ) {
    struct matiere *matiere;

    for ( matiere = premiere_composition ; matiere != NULL ; matiere = matiere->suivante ) {
      (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Contrle de la composition %s\n",
		     __FILE__, __FUNCTION__, __LINE__, matiere->nom);
    }
  }

  /* On libre les tables inutiles */
  free( composant->nom );

  /* On libre le composant */
  free( composant );
  return 0;
}

/* ~-~-~-~-~-~-~-~- Gestion du contenu ~-~-~-~-~-~-~-~- */

/* -------- Dterminer la composition  partir de la formule chimique -------- */

int Composant_AnalyserFormule(struct composant *composant, const char *formule)
{
  double masse_molaire, controle = 0.0;
  unsigned int i;
  Composition *composition;

  if ( ( NULL == composant ) || ( NULL == formule ) ) {
    return -1;
  }

  /* On dcortique la formule */
  composition = Chimie_AnalyseFormule( formule );
  if ( NULL == composition ) {
    (void) fprintf(stderr, "Composant_AnalyserFormule : ERREUR dans Chimie_AnalyseFormule\n");
    return -2;
  }

  /* On calcule la masse molaire associe */
  masse_molaire = Chimie_MasseMolaire( composition );
  if ( 0.0 == masse_molaire ) {
    (void) fprintf(stderr, "Composant_AnalyserFormule : masse molaire nulle [formule %s]\n", formule);
    return -3;
  }
  composant->masse_molaire = masse_molaire;

  /* On recopie le rsultat */
  for ( i = 0 ; i < PLUTONIUM ; i++ ) {
    composant->atomes[ i ] = composition->atomes[ i ] * ATM_Z_vers_masse( i + 1, ATM_MASSE_MOLAIRE ) / masse_molaire;
  }

  /* Controles finaux */
  for ( ; i < 120 ; i++ ) {
    if ( composition->atomes[ i ] > 0.0 ) {
      (void) fprintf(stderr, "Composant_AnalyserFormule : lment ignor, Z = %d [%s], stoechio = %g, f. massique = %g\n", i + 1, ATM_Z_vers_symbole( i + 1 ),
		     composition->atomes[ i ], composition->atomes[ i ] * ATM_Z_vers_masse( i + 1, ATM_MASSE_MOLAIRE ) / masse_molaire );
    }
  }
  for ( i = 0 ; i < PLUTONIUM ; i++ ) {
    controle += composant->atomes[ i ];
  }
  if ( controle != 1.0 ) {
    (void) fprintf(stderr, "Composant_AnalyserFormule : la somme des fractions massiques ne vaut pas 1 [%g, diffrence %g]\n", controle, controle - 1.0);
  }

  /* Termin, tout s'est bien pass */
  return 0;
}

/* -------- Modifier le nom -------- */

int Composant_ChangerNom(struct composant *composant, char *nom)
{
  if ( NULL == composant ) return -1;
  if ( NULL == nom ) return -2;

  /* On libre l'ancien nom */
  free( composant->nom );

  /* On installe une _copie_ du nouveau */
  composant->nom = strdup( nom );

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Combien d'lments sont impliqus dans la formule considre ? -------- */

int Composant_NombreElements(struct composant *composant)
{
  unsigned int nombre = 0, i;

  if ( NULL == composant ) return -1;

  for ( i = 0 ; i < ZMAX_TABLE_CONSTITUANTS ; i++ ) {
    if ( composant->atomes[ i ] > 0.0 ) {
      nombre++;
    }
  }

#ifdef _DEVELOPPE_
  printf("Composant_NombreElements : %d lments trouvs\n", nombre);
#endif
  return nombre;
}

/* -------- Estimer la densit d'un composant -------- */

int Composant_EstimerDensite(struct composant *composant)
{
  double somme = 0.0, densite = 0.0;
  unsigned int i;

  if ( NULL == composant ) return -1;

  for ( i = 0 ; i < ZMAX_TABLE_CONSTITUANTS ; i++ ) {
    double f_massique;

    f_massique = composant->atomes[ i ];
    if ( f_massique > 0.0 ) {
      somme += composant->atomes[ i ];
      densite += composant->atomes[ i ] * ATM_Z_vers_masse(i, ATM_MASSE_VOLUMIQUE);
    }
  }

  if ( somme <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Composition vide ou absurde\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }
  densite /= somme;

  composant->densite = densite;

  /* Le calcul s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Choix d'un composant dans la liste existante ~-~-~-~-~-~-~-~- */

struct composant *ChoisirComposant(Fenetre *fenetre_mere )
{
  InfosSelecteur infos_selecteur;
  struct composant *choix = NULL;
  int retour;

  /* Initialisation du slecteur */
  infos_selecteur.nouveau  = COMPOSANT_Nouveau_I;
  infos_selecteur.copier   = COMPOSANT_Copier_I;
  infos_selecteur.modifier = NULL;
  infos_selecteur.sauver   = COMPOSANT_Sauver_I;
  infos_selecteur.charger  = COMPOSANT_Charger_I;
  infos_selecteur.effacer  = COMPOSANT_Effacer_I;
  infos_selecteur.compter  = Composant_Compter;
  infos_selecteur.objet_v_texte = COMPOSANT_TrouverNom;

  retour = I_SelecteurUniversel( "Composant", &infos_selecteur, fenetre_mere );
  if ( retour == 0 ) {
    choix = Composant_N_Vers_C( NULL, infos_selecteur.obj_choisi );
  }

  return choix;
}

/* -------- Trouver le composant demand -------- */

unsigned long Composant_C_Vers_N(struct composant *premier, struct composant *composant)
{
  unsigned long i;
  struct composant *idx_composant;

  if ( NULL == premier ) premier = premier_composant;
  if ( NULL == composant ) return 0;

  for ( i = 0, idx_composant = premier ; idx_composant != composant ; idx_composant = idx_composant->suivant, i++ ) {
    if ( NULL == idx_composant ) {
      (void) fprintf(stderr, "%s,%s l.%u : Composant %s non-trouv dans la liste\n",
		     __FILE__, __FUNCTION__, __LINE__, composant->nom);
      return 0;
    }
  }

  /* On a trouv */
  return i;
}

struct composant *Composant_N_Vers_C(struct composant *premier, unsigned long n)
{
  struct composant *composant;
  unsigned long i;

  if ( NULL == premier ) premier = premier_composant;
  if ( 0 == n ) return premier;

  for ( i = 0, composant = premier ; i < n ; i++, composant = composant->suivant ) {
    if ( NULL == composant ) {
      (void) fprintf(stderr, "%s,%s l.%u : n=%lu, seulement %lu composants\n",
		     __FILE__, __FUNCTION__, __LINE__, n + 1, Composant_Compter() );
      return NULL;
    }
  }

  /* On a trouv */
  return composant;
}

int COMPOSANT_Charger_I(void)
{
  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE\n",
		 __FILE__, __FUNCTION__, __LINE__);

  /* Tout s'est bien pass */
  return 0;
}

int COMPOSANT_Copier_I(unsigned long n_original)
{
  struct composant *original;

  original = Composant_N_Vers_C( premier_composant, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans COMPOSANT_N_Vers_C()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE\n",
		 __FILE__, __FUNCTION__, __LINE__);

  return 0;
}

int COMPOSANT_Effacer_I(unsigned long n_original)
{
  struct composant *original;

  original = Composant_N_Vers_C( premier_composant, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans COMPOSANT_N_Vers_C\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  EffacerComposant( original, FAUX );

  return 0;
}

int COMPOSANT_Nouveau_I(unsigned long n_avant)
{
  struct composant *composant, *avant;

  /* On cherche o le placer dans la liste */
  if ( NULL == premier_composant ) {
    avant = NULL;
  } else {
    avant = Composant_N_Vers_C( premier_composant, n_avant );
  }

  /* On le cre */
  composant = CreerComposant( avant );
  if ( NULL == composant ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans CreerComposant()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  return 0;
}

int COMPOSANT_Sauver_I(unsigned long n_composant)
{
  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, n_composant);

  /* Tout s'est bien pass */
  return 0;
}

unsigned long Composant_Compter(void)
{
  unsigned long i;
  struct composant *d;

  for ( d = premier_composant, i = 0 ; d != NULL ; d = d->suivant, i++);
  return i;
}

const char *COMPOSANT_TrouverNom(unsigned int arbre __attribute__((__unused__)), unsigned long objet)
{
  struct composant *d;

  d = Composant_N_Vers_C( premier_composant, objet );
  if ( NULL == d ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans COMPOSANT_N_Vers_C\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  return d->nom;
}

/* ~-~-~-~-~-~-~-~- Calcul du spectre d'absorption ~-~-~-~-~-~-~-~- */

int Composant_CalculerMu(struct composant *composant, struct graphe *graphe, double f_massique)
{
  unsigned int Z;

  ENTREE_FONCTION("Composant_CalculerMu", "composant=%p graphe=%p f_massique=%g", composant, graphe, f_massique);

  if ( ( NULL == composant ) || ( NULL == graphe) ) {
    SORTIE_FONCTION("Composant_CalculerMu","%d [ERREUR paramtres]", -1);
    return -1;
  }
  if ( ( f_massique <= 0.0 ) || ( f_massique > 1.0 ) ) {
    (void) fprintf(stderr, "Composant_CalculerMu : fraction massique absurde [%g]\n", f_massique);
    SORTIE_FONCTION("Composant_CalculerMu","%d [ERREUR fraction massique]", -2);
    return -2;
  }

  /* Spectre utilisateur ? On l'interpole alors
     Ce spectre doit contenir (mu/rho) */
  if ( composant->mu_total != NULL ) {
    unsigned long i;
    double *x, *y;

    DANS_FONCTION("Spectre total fourni [%p]", composant->mu_total);

    for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
      *y += f_massique * GRF_LireValeur( composant->mu_total, *x, INTERPOLE_LINEAIRE );
    }

    /* Calcul fini, pour cette mthode : tout s'est bien pass */
    SORTIE_FONCTION("Composant_CalculerMu","%d", 0);
    return 0;
  }

  /* On calcule la contribution de chacun des constituants */
  for ( Z = 0 ; Z < ZMAX_TABLE_CONSTITUANTS ; Z++ ) {
    if ( composant->atomes[ Z ] <= 0.0 ) continue; /* lment absent */

    if ( composant->mu[ Z ] != NULL ) {
      Graphe *mu_seuil;
      double mu;
      double *x, *y;
      int seuil, E_seuil, Z_seuil;
      unsigned long i;

      mu_seuil = composant->mu[ Z ];
      DANS_FONCTION("Spectre de seuil fourni [%p, Z=%d]", composant->mu_total, Z+1);

      /* Pour quel seuil le graphe a-t-il t prvu ? */
      E_seuil = cherche_e0( mu_seuil, FAUX );
      Z_seuil = TH_IdentifierSeuil( E_seuil, 10.0, &seuil, FAUX );
      if ( Z_seuil != (signed) ( Z + 1 ) ) {
	(void) fprintf(stderr, "Composant_CalculerMu : Attention, la dtection automatique de seuil indique un seuil d'un autre lment [%d vs %d] !\n", Z_seuil, Z+1);
	(void) fprintf(stderr, "                     :  FAIRE --- Mthode de secours pour attribuer le seuil\n");
	/* Attention, seuils lgers : joue sur 1 ou 2 eV comparable aux effets redox => on ne peut qu'avertir, pas interdir */
/* 	SORTIE_FONCTION("Composant_CalculerMu","%d [ERREUR dtection de seuil]", -1); */
/* 	return -1; */
      }

      /* On ajoute la contribution du seuil ainsi dfini */
      for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
	if ( *x < mu_seuil->x_min ) {
	  /* Avant : on peut supposer que l'on n'absorbe rien [autres seuils grs aprs] */
	  mu = 0.0;
	} else if ( *x > mu_seuil->x_max ) {
	  double mu_fin, mu_calc, delta;
	  /* Aprs : on peut supposer que le modle normal suffit [plus d'EXAFS]
	     Mais il faut quand mme corriger pour viter d'avoir un dcrochement :
	     si on a calcul avec un (0) diffrent de McMaster, en particulier...
	  */
	  mu_fin = mu_seuil->y[ mu_seuil->nombre_point - 1 ]; /* Suppose le graphe tri par X croissant, ce qui est normal... */
	  mu_calc = TH_CalculerMu( Z + 1, mu_seuil->x_max, MUM_SELON_MCMASTER, seuil );
	  delta = mu_fin - mu_calc;
#ifdef _DEVELOPPE_
/*  	  printf("%s : correction de mu = %g [POURQUOI ?]\n", __FUNCTION__, delta); */
#endif

	  mu = TH_CalculerMu( Z + 1, *x, MUM_SELON_MCMASTER, seuil );
	  mu += delta; /* Correction pour viter le dcrochement... */
	} else {
	  mu = GRF_LireValeur( mu_seuil, *x, INTERPOLE_LINEAIRE );
	}

	*y += f_massique * composant->atomes[ Z ] * mu;
      }

      /* On ajoute la contribution des autres seuils, plus bas */
      TH_AjouterMum( graphe, Z + 1, f_massique * composant->atomes[ Z ], MUM_SELON_MCMASTER, FAUX );
    } else {
      DANS_FONCTION("Spectre  simuler [Z=%d]", Z+1);

      TH_AjouterMum( graphe, Z + 1, f_massique * composant->atomes[ Z ], MUM_SELON_MCMASTER, VRAI );
    }
  }

  /* Calcul fini, sans encombre */
  SORTIE_FONCTION("Composant_CalculerMu","%d", 0);
  return 0;
}

/* -------- Calcul pour une nergie donne -------- */

double Composant_CalculerMu_1p(struct composant *composant, double f_massique, double energie)
{
  double mu_composant = 0.0;
  unsigned int Z;

  if ( NULL == composant ) return -1.0;
  if ( ( f_massique <= 0.0 ) || ( f_massique > 1.0 ) ) {
    (void) fprintf(stderr, "Composant_CalculerMu : fraction massique absurde [%g]\n", f_massique);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -2;
  }

  /* L'utilisateur a-t-il fourni un spectre (mu/rho) ? */
  if ( composant->mu_total != NULL ) { /* Oui */
    mu_composant = GRF_LireValeur( composant->mu_total, energie, INTERPOLE_LINEAIRE );
    return mu_composant * f_massique;
  }

  /* Si on est l : on doit utiliser les valeurs tabules ! */
  for ( Z = 0 ; Z < ZMAX_TABLE_CONSTITUANTS ; Z++ ) {
    double mu_Z = 0.0;

    if ( composant->atomes[ Z ] <= 0.0 ) continue; /* lment absent */

    if ( composant->mu[ Z ] != NULL ) {	/* (mu/rho) pour l'lment donn par l'utilisateur */
      Graphe *mu_seuil;
      int seuil, E_seuil, Z_seuil;
      unsigned int i;

      mu_seuil = composant->mu[ Z ];

      /* Pour quel seuil le graphe a-t-il t prvu ? */
      E_seuil = cherche_e0( mu_seuil, FAUX );
      Z_seuil = TH_IdentifierSeuil( E_seuil, 10.0, &seuil, FAUX );
      if ( Z_seuil != (signed) ( Z + 1 ) ) {
	(void) fprintf(stderr, "Composant_CalculerMu : Attention, la dtection automatique de seuil indique un seuil d'un autre lment [%d vs %d] !\n", Z_seuil, Z+1);
	OPTION_A_FAIRE( "Mthode de secours pour attribuer le seuil" );
	/* Attention, seuils lgers : joue sur 1 ou 2 eV comparable aux effets redox => on ne peut qu'avertir, pas interdir */
      }

      /* On ajoute la contribution du seuil ainsi dfini */
      if ( energie < mu_seuil->x_min ) {
	/* Avant : on peut supposer que l'on n'absorbe rien [autres seuils grs aprs] */
	mu_Z = 0.0;
      } else if ( energie > mu_seuil->x_max ) {
	double mu_fin, mu_calc, delta;
	/* Aprs : on peut supposer que le modle normal suffit [plus d'EXAFS]
	   Mais il faut quand mme corriger pour viter d'avoir un dcrochement :
	   si on a calcul avec un (0) diffrent de McMaster, en particulier...
	*/
	mu_fin = mu_seuil->y[ mu_seuil->nombre_point - 1 ]; /* Suppose le graphe tri par X croissant, ce qui est normal... */
	mu_calc = TH_CalculerMu( Z + 1, mu_seuil->x_max, MUM_SELON_MCMASTER, seuil );
	delta = mu_fin - mu_calc;
#ifdef _DEVELOPPE_
/*  	  printf("%s : correction de mu = %g [POURQUOI ?]\n", __FUNCTION__, delta); */
#endif

	mu_Z = TH_CalculerMu( Z + 1, energie, MUM_SELON_MCMASTER, seuil );
	mu_Z += delta; /* Correction pour viter le dcrochement... */
      } else {
	mu_Z = GRF_LireValeur( mu_seuil, energie, INTERPOLE_LINEAIRE );
      }

      /* On ajoute la contribution des autres seuils, plus bas */
      for ( i = seuil + 1 ; i < NOMBRE_SEUILS ; i++ ) {
	mu_Z += TH_CalculerMu( Z + 1, energie, MUM_SELON_MCMASTER, i );
      }
    } else {
      mu_Z = TH_CalculerMu( Z + 1, energie, MUM_SELON_MCMASTER, ABSORPTION_TOTALE );
    }

    mu_Z *= composant->atomes[ Z ];
    mu_composant += mu_Z;
  }

  return mu_composant * f_massique;
}

/* ~-~-~-~-~-~-~-~- Calcul de la section efficace de diffusion ~-~-~-~-~-~-~-~- */

int Composant_CalculerTau(struct composant *composant, struct graphe *graphe, double f_massique, boolean elastique, int methode)
{
  int retour;
  unsigned int Z;

  ENTREE_FONCTION("Composant_CalculerTau", "composant=%p graphe=%p f_massique=%g elastique=%d methode=%d", composant, graphe, f_massique, elastique, methode);

  if ( ( NULL == composant ) || ( NULL == graphe) ) {
    SORTIE_FONCTION("Composant_CalculerTau","%d [ERREUR paramtres]", -1);
    return -1;
  }
  if ( ( f_massique <= 0.0 ) || ( f_massique > 1.0 ) ) {
    (void) fprintf(stderr, "Composant_CalculerTau : fraction massique absurde [%g]\n", f_massique);
    SORTIE_FONCTION("Composant_CalculerTau","%d [ERREUR fraction massique]", -2);
    return -2;
  }

  /* On calcule la contribution de chacun des constituants */
  for ( Z = 0 ; Z < PLUTONIUM ; Z++ ) {
    if ( composant->atomes[ Z ] <= 0.0 ) continue; /* lment absent */

    if ( VRAI == elastique ) {
      retour = TH_DiffusionElastique( Z + 1, graphe, methode, f_massique );
    } else {
      retour = TH_DiffusionInelastique( Z + 1, graphe, methode, f_massique );
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "Composant_CalculerTau : ERREUR dans le calcul de la diffusion [code %d]\n", retour);
    }
  }

  /* Calcul fini, sans encombre */
  SORTIE_FONCTION("Composant_CalculerTau","%d", 0);
  return 0;
}

/* ~-~-~-~-~-~-~-~- Calcul du spectre de fluorescence ~-~-~-~-~-~-~-~- */

int Composant_CalculerFluo(struct composant *composant, struct graphe *graphe, double f_massique, double energie, double mu_total)
{
  int Z;

  if ( ( NULL == composant ) || ( NULL == graphe) ) return -1;
  if ( ( f_massique <= 0.0 ) || ( f_massique > 1.0 ) ) {
    (void) fprintf(stderr, "Composant_CalculerFluo : fraction massique absurde [%g]\n", f_massique);
    return -2;
  }
  if ( mu_total <= 0.0 ) {
    (void) fprintf(stderr, "Composant_CalculerFluo : mu total absurde [%g]\n", mu_total);
    return -3;
  }

  /* On calcule la contribution de chacun des constituants */
  for ( Z = 0 ; Z < PLUTONIUM ; Z++ ) {
    if ( composant->atomes[ Z ] <= 0.0 ) continue; /* lment absent */

    FLUO_SimulerSpectre( graphe, energie, Z + 1, composant->mu[ Z ], composant->atomes[ Z ] * f_massique / mu_total);
  }

  /* Calcul fini, sans encombre */
  return 0;
}
