/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2002, Emmanuel CURIS
   Cration : 12 fevrier 2002

   Calculs theoriques divers

   Effet d'une degradation de l'echantillon
   (Plus generalement : influence d'une variation de concentration avec le temps
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"

#include "exafs.h"
#include "lissage.h"
#include "theorie.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_TXT_NOM_REFERENCE 37 /* Spectre du compos  idal  */
#define LGMAX_TXT_NOM_EVOL_CONC 37 /* Vitesse de dgradation */
#define LGMAX_TXT_NOM_FINAL     37 /* Spectre final, en cas de transformation */
#define LGMAX_TXT_NOM_PROFIL    37 /* Profil de concentration spatiale */

typedef struct {
  Graphe *reference;
  Graphe *profil;
  Graphe *concentration;
  Graphe *final;
} Infos_Degrade;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

Graphe *TH_deplacement_faisceau(Graphe *reference, Graphe *profil, unsigned int n_scans);

static Graphe *TH_DisparitionEchantillon(Graphe *ideal, Graphe *cinetique, double t_mesure, double t_attente, double t0);
static Graphe *TH_ReactionEchantillon   (Graphe *ideal, Graphe *produit, Graphe *cinetique, double t_mesure, double t_attente, double t0);

static int GER_arbre_degrade    (unsigned int objet, int x, int y);
static void    prepare_arbre_degrade(void);
static void    arbre_sortie_ok      (Infos_Degrade *infos);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Dplacement du faisceau (ponctuel) sur l'chantillon --------\
 |   Renvoie le dernier graphe parmi les x crs, contenant le rsultat  |
 |                                                                       |
 | reference = le graphe  idal  de l'espce tudie                   |
 |             contient en particulier la suite de X de mesures          |
 | profil    = le profil de concentration                                |
 | n_scans   = le nombre de spectres enregistrs sur ce profil           |
 \---------------------------------------------------------------------- */

Graphe *TH_deplacement_faisceau(Graphe *reference, Graphe *profil, unsigned int n_scans)
{
  double pas; /* Dplacement unitaire, entre chaque point de mesure */
  int retour;
  unsigned int j;
  Graphe *graphe = NULL;
  double *b_profil, *c_profil, *d_profil;

  if ( (reference == NULL) || (profil == NULL) || (n_scans == 0) ) return NULL;

  /* Les X du profils sont supposs en microns, au moins en unit cohrente */
  pas = (profil->x_max - profil->x_min) / (double) ( reference->nombre_point * n_scans );

  retour = SPL_tt_prep(profil, &b_profil, &c_profil, &d_profil);
  if ( retour != 0) {
    (void) LIS_ERREUR( retour, profil->nom );
    return NULL; /* Problme dans les splines */
  }

  for (j = 0 ; j < n_scans ; j++) {   /* Boucle sur les spectres */
    unsigned long i;
    double *x, *y, *ref_x, *ref_y, x_profil;
    char *titre;

    titre = TXT_genere_chaine("Passage %d sur %d", (int) j + 1, (int) n_scans);
    /* A REMPLACER PAR CRATION DE GRAPHES  LA SUITE LES UNS DES AUTRES */
    graphe = GRF_nouveau(reference, reference->nombre_point, reference->type, GR_NORMAL, titre);
    free(titre);
    if (graphe == NULL) {
      free(b_profil); free(c_profil); free(d_profil);
      return NULL;
    }

    for (i = 0, x = graphe->x, ref_x = reference->x, y = graphe->y, ref_y = reference->y ;
	 i < graphe->nombre_point ; i++, x++, y++, ref_x++, ref_y++) {
      *x = *ref_x;

      x_profil = profil->x_min + pas * (i + j * reference->nombre_point); 

      *y = *ref_y * SPL_tt_calc(profil, x_profil, b_profil, c_profil, d_profil);
    }

    graphe->x_min = reference->x_min;
    graphe->x_max = reference->x_max;
    GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
  }

  free(b_profil); free(c_profil); free(d_profil);

  return graphe;
}

/* -------- Degradation, modele simple --------\
 |   Renvoie le graphe resultat
 |
 | seuil : le spectre XANES ideal, sans degradation
 | a ,b  : parametres de la droite de degradation, C = a*t + b
 | t0    : le temps initial
 | pas   : une mesure tous les ...
 \------------------------------------------------------------ */

Graphe *TH_degrade_lineaire(Graphe *seuil, double a, double b, double t0, double pas)
{
  unsigned long i;
  double *y;
  Graphe *resultat;

  if (seuil == NULL) return NULL;

  resultat = GRF_Copier( seuil, NULL, FAUX );
  if (resultat == NULL) return NULL;

  for (i = 0, y = resultat->y ; i < seuil->nombre_point ; i++, y++) {
    double facteur;

    facteur = b + a * (t0 + pas * (double) i);
    if (facteur <= 0.0)
      facteur = 0.0; /* Plus de produit */

    *y *= facteur;
  }
  GRF_bornes(resultat, NE_PAS_FAIRE, DESORDONNE);

  return resultat;
}

/* -------- Dgradation, profil de concentration quelconque -------- */

struct graphe *TH_Transformation(struct graphe *ideal, struct graphe *cible, struct graphe *profil_concentration,
				 double t_mesure, double t_attente, unsigned long nbr_passes)
{
  char *nom;
  double t0, temps_mesure, temps_retour;
  unsigned long i;
  Graphe *resultat = NULL;

  if ( ( NULL == ideal ) || ( NULL == profil_concentration ) ) return NULL;
  if ( 0 == nbr_passes ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucun passage !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Combien de temps pour un spectre complet ? */
  temps_mesure = t_mesure * ( double ) ideal->nombre_point + t_attente * ( double ) ( ideal->nombre_point - 1 );

  /* Combien de temps, entre chaque spectre, pour revenir au dbut ?
     On suppose que t_attente est le temps de dplacement... */
  temps_retour = t_attente * ( double ) ( ideal->nombre_point - 1 );

  /* Calcul pour chaque passe */
  for ( i = 0 ; i < nbr_passes ; i++ ) {
    /* Temps initial */
    t0  = temps_retour * (double) i;
    if ( i > 0 ) {
      t0 += temps_mesure * (double) (i - 1);
    }

    if ( t0 > profil_concentration->x_max ) { /* On est au-del : aucun intrt */
      (void) fprintf(stderr, "%s,%s l.%u : On a termin la raction\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Calcul du signal */
    if ( NULL == cible ) {
      resultat = TH_DisparitionEchantillon( ideal, profil_concentration, t_mesure, t_attente, t0 );
    } else {
      resultat = TH_ReactionEchantillon( ideal, cible, profil_concentration, t_mesure, t_attente, t0 );
    }
    if ( NULL == resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le calcul\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* On donne un nom */
    nom = TXT_genere_chaine( "%s, "e_AIGU"volution : passage %u", ideal->nom, (i + 1) );
    GRF_titre( resultat, nom );
    free( nom ); nom = NULL;
  }

  /* Termin : on renvoie le dernier graphe cr */
  return resultat;
}

/* -------- Cintique : chantillon disparat -------- */

static Graphe *TH_DisparitionEchantillon(Graphe *ideal, Graphe *cinetique, double t_mesure, double t_attente, double t0)
{
  double t, attenuation;
  double *x, *y, *y_ideal;
  struct graphe *resultat;
  unsigned long i;

  /* Le rsultat aura les mmes X que l'idal */
  resultat = GRF_nv_TableX( ideal, ideal->type, ideal->genre_donnees, NULL, FAUX, ideal->x, ideal->nombre_point, VRAI );
  if ( NULL == resultat ) return NULL;

  /* Pour chaque point */
  for ( i = 0, x = resultat->x, y = resultat->y, y_ideal = ideal->y ;
	i < resultat->nombre_point ;
	i++, x++, y++, y_ideal++ ) {
    /* Quel est le temps [en dbut de mesure] ? */
    t  = t0;
    t += t_mesure  * (double) i; /* Mesure des points prcdents */
    t += t_attente * (double )i; /* Attente entre les points prcdents */

    /* On en dduit la concentration restante */
    attenuation = GRF_LireValeur( cinetique, t, INTERPOLE_LINEAIRE );

    /* Et donc la valeur mesure --- NGLIGE L'VOLUTION PENDANT LA MESURE */
    *y = *y_ideal * attenuation;
  }

  /* Termin : recalcul des bornes */
  GRF_bornes( resultat, NE_PAS_FAIRE, DESORDONNE );

  return resultat;
}

static Graphe *TH_ReactionEchantillon(Graphe *ideal, Graphe *produit, Graphe *cinetique, double t_mesure, double t_attente, double t0)
{
  double t, attenuation, signal_cible;
  double *x, *y, *y_ideal;
  struct graphe *resultat;
  unsigned long i;

  /* Le rsultat aura les mmes X que l'idal */
  resultat = GRF_nv_TableX( ideal, ideal->type, ideal->genre_donnees, NULL, FAUX, ideal->x, ideal->nombre_point, VRAI );
  if ( NULL == resultat ) return NULL;

  /* Pour chaque point */
  for ( i = 0, x = resultat->x, y = resultat->y, y_ideal = ideal->y ;
	i < resultat->nombre_point ;
	i++, x++, y++, y_ideal++ ) {
    /* Quel est le temps [en dbut de mesure] ? */
    t  = t0;
    t += t_mesure  * (double) i; /* Mesure des points prcdents */
    t += t_attente * (double )i; /* Attente entre les points prcdents */

    /* On en dduit la concentration restante */
    attenuation = GRF_LireValeur( cinetique, t, INTERPOLE_LINEAIRE );

    /* Quel est le signal cible pour l'nergie indique ? */
    signal_cible = GRF_LireValeur( produit, *x, INTERPOLE_LINEAIRE );

    /* Et donc la valeur mesure --- NGLIGE L'VOLUTION PENDANT LA MESURE */
    *y = *y_ideal * attenuation + signal_cible * ( 1.0 - attenuation );
  }

  /* Termin : recalcul des bornes */
  GRF_bornes( resultat, NE_PAS_FAIRE, DESORDONNE );

  return resultat;
}

/* -------- Interface graphique de paramtrage -------- */

Graphe *TH_evolution_temporelle(void)
{
  Graphe *resultat = NULL;
  unsigned int n_scans;
  int retour;
  Infos_Degrade infos = {NULL, NULL, NULL, NULL};

  infos.reference = graphe_courant;
  if (infos.reference == NULL) {
    met_chaine(EFFETS_TEMPS, ET_GRAPHE_1, LGMAX_TXT_NOM_REFERENCE, "Aucune r"e_AIGU"f"e_AIGU"rence", TXT_COPIER);
  } else {
    met_chaine(EFFETS_TEMPS, ET_GRAPHE_1, LGMAX_TXT_NOM_REFERENCE, infos.reference->nom, TXT_COPIER);
  }
  met_chaine(EFFETS_TEMPS, ET_PROFIL, LGMAX_TXT_NOM_PROFIL, "Aucun profil", TXT_COPIER);

  if (arbres[EFFETS_TEMPS].drapeaux.prepare == FAUX) {
    prepare_arbre_degrade();
  }
  arbre_sortie_ok(&infos);

  ARB_MettreComplement( EFFETS_TEMPS, &infos );
  retour = ARB_Bloquant( EFFETS_TEMPS, E_AIGU"volution temporelle", principale, 0 );
  ARB_MettreComplement( EFFETS_TEMPS, NULL );
  if ( ET_ANNULER == retour ) return NULL;

  /* Nombre de passages */
  n_scans = OBJ_LireEntier_U( EFFETS_TEMPS, ET_NBR_PASSAGE );
  if (n_scans == 0) return NULL;

  if ( VRAI == OBJ_choisit( EFFETS_TEMPS, ET_DISPARITION, QUESTION ) ) {
    resultat = TH_DisparitionEchantillon( infos.reference, infos.concentration,
					  OBJ_LireNombre( EFFETS_TEMPS, ET_TPS_MESURE, NULL ), OBJ_LireNombre( EFFETS_TEMPS, ET_TPS_DEPLACE, NULL ), 0.0 );
  }

  if ( OBJ_choisit(EFFETS_TEMPS, ET_DEPLACE, QUESTION) == VRAI ) {
    resultat = TH_deplacement_faisceau(infos.reference, infos.profil, n_scans);
  }

  return resultat;
}

static int GER_arbre_degrade(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  int etat;
  Infos_Degrade *infos;

  infos = (Infos_Degrade *) ARB_LireComplement( EFFETS_TEMPS );

  switch (objet)
  {
  case ET_GRAPHE_1: /* Choix du graphe de rfrence */
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_OUI);
    infos->reference = GRF_choisir(GR_BRUT, FAUX, VRAI, NULL);
    if (infos->reference == NULL) {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_REFERENCE, "Aucune r"e_AIGU"f"e_AIGU"rence", TXT_COPIER);
    } else {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_REFERENCE, infos->reference->nom, TXT_COPIER);
    }
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_NON);

    /* Peut-on sortir en faisant le calcul ? */
    arbre_sortie_ok(infos);
    return VRAI;

  case ET_DISPARITION: /* Le compos se transforme ou disparat avec le temps */
  case ET_TRANSFORME:
    OBJ_choisit(EFFETS_TEMPS, objet, ECHANGE);
    etat  = OBJ_choisit(EFFETS_TEMPS, ET_TRANSFORME, QUESTION);
    OBJ_active(EFFETS_TEMPS, ET_GRF_FINAL, etat);
    etat |= OBJ_choisit(EFFETS_TEMPS, ET_DISPARITION, QUESTION);
    OBJ_active(EFFETS_TEMPS, ET_CONCENTRATION, etat);
    OBJ_active(EFFETS_TEMPS, ET_INSTANTANE, etat);
    OBJ_active(EFFETS_TEMPS, ET_TPS_DEPLACE, etat);
    OBJ_active(EFFETS_TEMPS, ET_TPS_MESURE, etat);

    /* Peut-on sortir en faisant le calcul ? */
    arbre_sortie_ok(infos);
    return VRAI;

  case ET_CONCENTRATION: /* volution de la concentration avec le temps */
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_OUI);
    infos->concentration = GRF_choisir(GR_CINETIQUE, FAUX, VRAI, NULL);
    if (infos->concentration == NULL) {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_EVOL_CONC, "Aucun graphe de variation de concentration", TXT_COPIER);
    } else {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_EVOL_CONC, infos->concentration->nom, TXT_COPIER);
    }
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_NON);    

    /* Peut-on sortir en faisant le calcul ? */
    arbre_sortie_ok(infos);
    return VRAI;

  case ET_GRF_FINAL: /* Graphe du spectre du compos de fin de transformation */
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_OUI);
    infos->final = GRF_choisir(GR_BRUT, FAUX, VRAI, NULL);
    if (infos->final == NULL) {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_FINAL, "Aucun graphe final", TXT_COPIER);
    } else {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_FINAL, infos->final->nom, TXT_COPIER);
    }
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_NON);    

    /* Peut-on sortir en faisant le calcul ? */
    arbre_sortie_ok(infos);
    return VRAI;

  case ET_DEPLACE: /* Le faisceau drive avec le temps */
    etat = OBJ_choisit(EFFETS_TEMPS, ET_DEPLACE, ECHANGE);
    OBJ_active(EFFETS_TEMPS, ET_PROFIL, etat);
    OBJ_active(EFFETS_TEMPS, ET_DELTA_X, etat);

    /* Peut-on sortir en faisant le calcul ? */
    arbre_sortie_ok(infos);
    return VRAI;

  case ET_PROFIL: /* Profil de concentration suivant la courbe parcourue par le faisceau */
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_OUI);
    infos->profil = GRF_choisir(GR_PROFIL, FAUX, VRAI, NULL);
    if (infos->profil == NULL) {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_PROFIL, "Aucun profil", TXT_COPIER);
    } else {
      met_chaine(EFFETS_TEMPS, objet, LGMAX_TXT_NOM_PROFIL, infos->profil->nom, TXT_COPIER);
    }
    OBJ_choisit(EFFETS_TEMPS, objet, FORCE_NON);    

    /* Peut-on sortir en faisant le calcul ? */
    arbre_sortie_ok(infos);
    return VRAI;
  }

  return FAUX;
}

static void arbre_sortie_ok(Infos_Degrade *infos)
{
  int operation_demandee = FAUX;

  /* A-t-on une rfrence ? */
  if ( NULL == infos->reference ) {
    OBJ_active( EFFETS_TEMPS, ET_CALCULER, FORCE_NON );
    return;
  }

  /* Veut-on modliser une disparition ? */
  if ( VRAI == OBJ_choisit(EFFETS_TEMPS, ET_DISPARITION, QUESTION) ) {
    operation_demandee = VRAI;

    /* Il faut alors une concentration en fonction du temps */
    if ( NULL == infos->concentration ) {
      OBJ_active( EFFETS_TEMPS, ET_CALCULER, FORCE_NON );
      return;
    }
  }

  /* Veut-on modliser une transformation ? */
  if ( VRAI == OBJ_choisit(EFFETS_TEMPS, ET_TRANSFORME, QUESTION) ) {
    operation_demandee = VRAI;

    /* Il faut alors une concentration en fonction du temps... */
    if ( NULL == infos->concentration ) {
      OBJ_active( EFFETS_TEMPS, ET_CALCULER, FORCE_NON );
      return;
    }

    /* ... et une rfrence pour le compos d'arrive */
    if ( NULL == infos->final ) {
      OBJ_active( EFFETS_TEMPS, ET_CALCULER, FORCE_NON );
      return;
    }
  }

  /* Veut-on modliser un dplacement du faisceau sur l'chantillon ? */
  if ( VRAI == OBJ_choisit(EFFETS_TEMPS, ET_DEPLACE, QUESTION) ) {
    operation_demandee = VRAI;

    /* Il faut alors un profil de concentration */
    if ( NULL == infos->profil ) {
      OBJ_active( EFFETS_TEMPS, ET_CALCULER, FORCE_NON );
      return;
    }
  }

  /* Demande-t-on bien une opration ? */
  if ( FAUX == operation_demandee ) {
    OBJ_active( EFFETS_TEMPS, ET_CALCULER, FORCE_NON );
    return;
  }

  /* Si on est l, on peut faire les calculs */
  OBJ_active( EFFETS_TEMPS, ET_CALCULER, FORCE_OUI );
}

static void prepare_arbre_degrade(void)
{
  int retour;

  /* Prparation de l'arbre */
  retour = ARB_Initialiser( EFFETS_TEMPS, "menu_errsyst.html#degradation", GER_arbre_degrade, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) {
    return;
  }

  /* Amliorer les positions */
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_MESURE , 0, -1, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TPS_MESURE , 0, -1, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_INSTANTANE , 0, -1, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_PASSAGE, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_NBR_PASSAGE, 0, +1, DEPLACE_PIXELS);

  OBJ_deplace(EFFETS_TEMPS, ET_DISPARITION, 0, -3, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_DISP   , 0, -2, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TRANSFORME , 0, -3, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_TRANS  , 0, -2, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_FINAL  , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_GRF_FINAL  , 0, +2, DEPLACE_PIXELS);

  OBJ_deplace(EFFETS_TEMPS, ET_DEPLACE    , 0, -2, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_DEPLACE, 0, -2, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_VITESSE, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_DELTA_X    , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(EFFETS_TEMPS, ET_TXT_DERIVE2, 0, +2, DEPLACE_PIXELS);

  /* Fonctionnement... */
  arbres[EFFETS_TEMPS].drapeaux.prepare = VRAI;
}
