/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Calcul du signal obtenu en fluorescence
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "detecteur.h"
#include "echantillon.h"
#include "faisceau.h"
#include "graphes.h"
#include "lissage.h" /* pour GRF_LireValeur */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int FLUO_Init_Etendu   (Echantillon *echantillon, Faisceau *faisceau, double energie);
static int FLUO_Init_Polychromatique(Echantillon *echantillon, Faisceau *faisceau, double energie);
static int FLUO_Init_Ponctuel (Echantillon *echantillon, Faisceau *faisceau, double energie);
static int FLUO_SpectresIdeaux(Echantillon *echantillon, double energie);

static int FLUO_InitEtendu_H(Echantillon *echantillon, Faisceau *faisceau, Vecteur3D *u_faisceau, double energie, double delta_t);
static int FLUO_InitEtendu_G(Echantillon *echantillon, Faisceau *faisceau, Vecteur3D *u_faisceau, double energie);

static int FLUO_IndicesLimites(Echantillon *echantillon, Faisceau *faisceau, Vecteur3D *u_faisceau, double t,
			       unsigned int *ix_min, unsigned int *ix_max, unsigned int *iz_min, unsigned int *iz_max );

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Initialisation de l'échantillon ~-~-~-~-~-~-~-~- */

/* -------- Interface générique -------- */

int FLUO_InitialiserEchantillon(Echantillon *echantillon, Faisceau *faisceau, double energie)
{
  unsigned long i, nbr_valeurs;
  int retour = -2;

  if ( ( NULL == faisceau ) || ( NULL == echantillon ) ) return -1;

  /* On initialise le(s) spectre(s) de fluorescence idéaux */
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    Graphe *fluo;
    double *y;

    fluo = echantillon->valeurs.homogene.composition->fluo;
    for ( i = 0, y = fluo->y ; i < fluo->nombre_point ; i++, y++ ) {
      *y = 0.0;
    }
  } else {
    if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      Graphe *fluo;
      double *y;

      fluo = echantillon->valeurs.homogene.composition->fluo;
      for ( i = 0, y = fluo->y ; i < fluo->nombre_point ; i++, y++ ) {
	*y = 0.0;
      }
    } else {
      struct matiere *composition;
      double *y;

      for ( composition = premiere_composition ; composition != NULL ; composition = composition->suivante ) {
	if ( NULL == composition->fluo ) continue; /* Non-utilisé */

	for ( i = 0, y = composition->fluo->y ; i < composition->fluo->nombre_point ; i++, y++ ) {
	  *y = 0.0;
	}
      }
    }
  }

  /* On prépare la table des valeurs : nombre de photons absorbés en chaque point */
  nbr_valeurs = echantillon->nbr_x * echantillon->nbr_y * echantillon->nbr_z;
  if ( NULL == echantillon->nombre ) {
    echantillon->nombre = (double *) malloc( nbr_valeurs * sizeof( double ) );
    if ( NULL == echantillon->nombre ) {
      (void) fprintf(stderr, "FLUO_InitialiserEchantillon : Mémoire insuffisante [photons absorbés]\n");
      return -3;
    }
  }
  for ( i = 0 ; i < nbr_valeurs ; i++ ) {
    echantillon->nombre[ i ] = 0.0;
  }

  /* On remplit la table des valeurs [nombre de photons absorbés] */
  if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Faisceau polychromatique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    retour = FLUO_Init_Polychromatique( echantillon, faisceau, energie );
  } else {
    if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
      retour = FLUO_Init_Ponctuel( echantillon, faisceau, energie );
    } else {
    (void) fprintf(stderr, "%s,%s l.%u : Faisceau monochromatique étendu\n",
		   __FILE__, __FUNCTION__, __LINE__);
      retour = FLUO_Init_Etendu( echantillon, faisceau, energie );
    }
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation / fondamental\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* On complète avec la contribution des harmoniques */
  if ( faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    unsigned int i_h;
    Faisceau pseudo_faisceau;

    pseudo_faisceau = *faisceau;
    pseudo_faisceau.proprietes &= ~FAISCEAU_HARMONIQUES;
    for ( i_h = 0 ; i_h < faisceau->nbr_harmoniques ; i_h++ ) {
      pseudo_faisceau.intensite = faisceau->intensite * faisceau->i_harmoniques[ i_h ];
      pseudo_faisceau.g_E = faisceau->g_harmoniques[ i_h ];
      if ( 0.0 == pseudo_faisceau.intensite ) continue; /* Harmonique éteinte */

      if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
	retour = FLUO_Init_Polychromatique( echantillon, faisceau, energie );
      } else {
	if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
	  retour = FLUO_Init_Ponctuel( echantillon, &pseudo_faisceau, energie * (i_h + 2) );
	} else {
	  retour = FLUO_Init_Etendu( echantillon, &pseudo_faisceau, energie * (i_h + 2) );
	}
      }
    }
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Cas d'un faisceau polychromatique -------- */

static int FLUO_Init_Polychromatique(Echantillon *echantillon, Faisceau *faisceau, double energie)
{
  double e_vraie, *de, I_vraie, *gE;
  int retour;
  unsigned long i;
  Faisceau pseudo_faisceau;

  pseudo_faisceau = *faisceau;
  pseudo_faisceau.proprietes &= ~FAISCEAU_POLYCHROMATIQUE;
  pseudo_faisceau.nbr_harmoniques = 0;
  pseudo_faisceau.i_harmoniques = NULL; /* On utilise une nouvelle table d'harmoniques ! */
  pseudo_faisceau.g_harmoniques = NULL;

  /* Pour chacune des énergies enregistrées */
  for ( i = 0, de = faisceau->g_E->x, gE = faisceau->g_E->y ; i < faisceau->g_E->nombre_point ; i++, de++, gE++ ) {
    /* Énergie prise en compte */
    e_vraie = energie * ( 1.0 + *de );

    /* Intensité pour cette énergie [si FAISCEAU_DECROIT, pris en compte dans Monochromatique [par LireIntensite]] */
    I_vraie = faisceau->intensite * *gE;
    pseudo_faisceau.intensite = I_vraie;

    /* Contribution de cette énergie */
    if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
      retour = FLUO_Init_Ponctuel( echantillon, &pseudo_faisceau, e_vraie );
    } else {
      retour = FLUO_Init_Etendu( echantillon, &pseudo_faisceau, e_vraie );
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation / E=%g\n",
		     __FILE__, __FUNCTION__, __LINE__, e_vraie);
      return -1;
    }
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Cas d'un faisceau étendu -------- */

static int FLUO_Init_Etendu(Echantillon *echantillon, Faisceau *faisceau, double energie)
{
  double delta_t;
  double vx_intensite;
  int retour;
  Vecteur3D u_faisceau;

#ifdef _DEVELOPPE_
  if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Faisceau étendu, on ne devrait jamais être là !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1974;
  }
#endif

  /* Vecteur directeur de la droite */
  REPERE_VecteurNormal( &(faisceau->orientation), &u_faisceau );
  if ( 0.0 == u_faisceau.y ) {
    (void) fprintf(stderr, "FLUO_Init_Etendu : Faisceau parallèle à l'échantillon\n");
    return -2;
  }

  /* Intensité originelle du faisceau */
  vx_intensite = faisceau->intensite;

  /* Incrément pour passer d'une tranche à l'autre */
  delta_t = echantillon->pas_y / u_faisceau.y;

  /* Calcul proprement dit */
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    retour = FLUO_InitEtendu_H( echantillon, faisceau, &u_faisceau, energie, delta_t );
  } else {
    retour = FLUO_InitEtendu_G( echantillon, faisceau, &u_faisceau, energie );
  }
  /* On remet l'intensité originelle du faisceau [car le même peut servir à chaque tour, si projection constante !] */
  faisceau->intensite = vx_intensite;

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -3;
  }

  /* On ajoute la contribution aux spectres de fluorescence idéaux */
  retour = FLUO_SpectresIdeaux( echantillon, energie );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_SpectresIdeaux()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -4;
  }

  /* Initialisation terminée */
  return 0;
}

/* ------ Faisceau étendu, échantillon homogène ------ */

static int FLUO_InitEtendu_H(Echantillon *echantillon, Faisceau *faisceau, Vecteur3D *u_faisceau, double energie, double delta_t)
{
  double I0, dd;
  double mu, mu_total; /* mu : absorption photo-électrique seule ; mu_total : diffusion inclue */
  double f_absorbe, attenuation, n_absorbe;
  double x, z; /* Point d'impact du faisceau à une profondeur donné, dans le plan // à la face d'entrée */
  int retour;
  unsigned int i_x, i_y, i_z;
  unsigned int ix_min, ix_max, iz_min, iz_max;
  unsigned long n_case;
  struct matiere *composition __attribute__((__unused__));

  /* Composition de l'échantillon */
  composition = echantillon->valeurs.homogene.composition;

  /* Distance parcourue d'une tranche à l'autre */
  dd = fabs( delta_t );
  dd /= 10000.0; /* mu/rho est en cm^-1, les distances dans l'échantillon en microns... */

  /* On calcule la valeur du coefficient _linéaire_ d'absorption pour l'énergie indiquée */
  mu_total = ECH_CalculerMuTotal( echantillon, NULL, energie, &mu );

  /* Tranche de surface : on postule une profondeur de dy / 2 pour avoir quelque chose !
     On travaille avec le faisceau projeté => sa position (x,z) donne le point d'impact */
  /*   printf("%s : Cases remplies :\n", __FUNCTION__); */

  attenuation = exp( -mu_total * dd / 2.0 );
  f_absorbe = 1.0 - exp( -mu * dd / 2.0 );

  /* Limites de case données par le faisceau */
  retour = FLUO_IndicesLimites( echantillon, faisceau, u_faisceau, 0.0, &ix_min, &ix_max, &iz_min, &iz_max );
  if ( 0 == retour ) {
    /* On ne cherche à initialiser que pour les cases illuminées par le faisceau ! */
    for ( i_x = ix_min /*0*/ ; i_x <= ix_max /*echantillon->nbr_x - 1*/ ; i_x++ ) {
      x = -echantillon->larg / 2.0 + echantillon->pas_x * i_x;

      for ( i_z = iz_min /*0*/ ; i_z <= iz_max /*echantillon->nbr_z - 1*/ ; i_z++ ) {
	z = -echantillon->haut / 2.0 + echantillon->pas_z * i_z;

	I0 = FAISCEAU_LireIntensite( faisceau, x, z, 0.0 );
	if ( 0.0 == I0 ) continue; /* Ne contribue pas */

	/* Ici, le faisceau passe donc par ce point : on remplit donc la case avec ce qu'il faut */
	n_absorbe = I0 * f_absorbe;

	n_case = ECH_IJK_v_Case( echantillon, i_x, 0, i_z );
	echantillon->nombre[ n_case ] += n_absorbe;
	/*       printf("%lu %g %g %g %g\n", n_case, n_absorbe, attenuation, f_absorbe, I0); */
      }
    }
  }

  /* Puisque l'échantillon est homogène, c'est partout pareil ! */
  faisceau->intensite *= attenuation;

  /* Autres tranches */
  attenuation = exp( -mu_total * dd );
  f_absorbe = 1.0 - exp( -mu * dd );
  for ( i_y = 1 ; i_y < echantillon->nbr_y ; i_y++ ) {
    double t;

    t = echantillon->pas_y * i_y;

    /* Limites de case données par le faisceau à cette profondeur */
    retour = FLUO_IndicesLimites( echantillon, faisceau, u_faisceau, t, &ix_min, &ix_max, &iz_min, &iz_max );
    if ( retour != 0 ) continue; /* Le faisceau ne croise pas cette tranche */

    for ( i_x = ix_min /*0*/ ; i_x <= ix_max /*echantillon->nbr_x - 1*/ ; i_x++ ) {
      x  = -echantillon->larg / 2.0 + echantillon->pas_x * i_x;
      x -= u_faisceau->x * t; /* Le faisceau se décale au fur et à mesure */

      for ( i_z = iz_min /*0*/ ; i_z <= iz_max /*echantillon->nbr_z - 1*/ ; i_z++ ) {
	z  = -echantillon->haut / 2.0 + echantillon->pas_z * i_z;
	z -= u_faisceau->z * t; /* Le faisceau se décale au fur et à mesure */

	I0 = FAISCEAU_LireIntensite( faisceau, x, z, 0.0 );
	if ( 0.0 == I0 ) continue; /* Ne contribue pas */

	/* Ici, le faisceau passe donc par ce point : on remplit donc la case avec ce qu'il faut */
	n_absorbe = I0 * f_absorbe;

	n_case = ECH_IJK_v_Case( echantillon, i_x, i_y, i_z );
	echantillon->nombre[ n_case ] += n_absorbe;
/*  	printf("%lu %g %g %g %g\n", n_case, n_absorbe, attenuation, f_absorbe, I0); */
      }
    }

    /* Puisque l'échantillon est homogène, c'est partout pareil ! */
    faisceau->intensite *= attenuation;
  }

  /* Calcul sans histoire */
  return 0;
}

/* ------ Faisceau étendu, échantillon gradient ------ */

static int FLUO_InitEtendu_G(Echantillon *echantillon, Faisceau *faisceau, Vecteur3D *u_faisceau, double energie)
{
#ifdef ancien
  double mu_total, mu;
  double I0, attenuation, f_absorbe, n_absorbe;
  double dd;
  unsigned long n_case;
  Vecteur3D point_impact, point_entree;
#endif
  int retour;
  unsigned int i_x/*, i_y*/, i_z, ix_min, ix_max, iz_min, iz_max;
  Faisceau f_ponctuel;

  /* On travaille comme une juxtaposition de faisceaux ponctuels */
  /* a. Création du pseudo-faisceau ponctuel */
  f_ponctuel = *faisceau;
  f_ponctuel.suivant = NULL;
  f_ponctuel.largeur = 0.0; f_ponctuel.hauteur = 0.0;
  f_ponctuel.proprietes &= ~( FAISCEAU_INHOMOGENE | FAISCEAU_FOCALISE );
  f_ponctuel.proprietes |= FAISCEAU_PONCTUEL;
  f_ponctuel.g_espace = NULL;

  /* On se fonde sur le pas de l'échantillon : si faisceau plus finement défini, sert à rien ! */

  /* Limites de case données par le faisceau à cette profondeur */
  retour = FLUO_IndicesLimites( echantillon, faisceau, u_faisceau, 0.0, &ix_min, &ix_max, &iz_min, &iz_max );
  if ( retour != 0 ) {
    /* Le faisceau ne croise pas l'échantillon ! */
    (void) fprintf(stderr, "%s,%s l.%u : Le faisceau ne croise pas l'échantillon !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -578;
  }

  /* Pour chacune des cases illuminées, on suit le faisceau qui passe par son centre */
  for ( i_x = ix_min ; i_x <= ix_max ; i_x++ ) {
    for ( i_z = iz_min ; i_z <= iz_max ; i_z++ ) {
      /* Position d'entrée du faisceau : le centre de la case */
      ECH_IJK_v_XYZ( echantillon, i_x, 0, i_z, &(f_ponctuel.orientation.position.x), &(f_ponctuel.orientation.position.y),
		     &(f_ponctuel.orientation.position.z) );

      /* Intensité du faisceau en ce point
	 [Le faisceau a normalement déjà été corrigé des effets énergétiques !] */
      f_ponctuel.intensite = FAISCEAU_LireIntensite( faisceau, 0.0, 0.0, 0.0 );

      /* On suit le faisceau dans l'échantillon en remplissant les cases */
      FLUO_Init_Ponctuel( echantillon, &f_ponctuel, energie );
    }
  }

#ifdef ancien
  /* Distance parcourue d'une tranche à l'autre */
  dd = fabs( delta_t );
  dd /= 10000.0; /* mu/rho est en cm^-1, les distances dans l'échantillon en microns... */

/*   printf("%s : Cases remplies :\n", __FUNCTION__); */

  /* Méthode brutale : pas le plus efficace, mais devrait fonctionner aussi pour un échantillon quelconque ! */
  /* Tranche de surface : on postule une profondeur de dy / 2 pour avoir quelque chose !
     On travaille avec le faisceau projeté => sa position (x,z) donne le point d'impact
     On est homogène sur cette demi-tranche
  */
  for ( i_x = 0 ; i_x < echantillon->nbr_x ; i_x++ ) {
    point_impact.x = -echantillon->larg / 2.0 + echantillon->pas_x * i_x;

    for ( i_z = 0 ; i_z < echantillon->nbr_z ; i_z++ ) {
      point_impact.y = 0.0;
      point_impact.z = -echantillon->haut / 2.0 + echantillon->pas_z * i_z;

      I0 = FAISCEAU_LireIntensite( faisceau, point_impact.x, point_impact.z, 0.0 );
      if ( 0.0 == I0 ) continue; /* Ne contribue pas */

      /* Le faisceau passe par ce point : quelle est l'absorption, l'atténuation par ce point ? */
      /* 1. Coefficients linéiques d'absorption */
      mu_total = ECH_CalculerMuTotal( echantillon, &point_impact, energie, &mu );

      /* 2. Atténuation, absorption */
      attenuation = exp( -mu_total * dd / 2.0 );
      f_absorbe = 1.0 - exp( -mu * dd / 2.0 );

      /* Nombre de photons absorbés */
      n_absorbe = I0 * f_absorbe;

      /* On le mémorise */
      n_case = ECH_IJK_v_Case( echantillon, i_x, 0, i_z );
      echantillon->nombre[ n_case ] += n_absorbe;
/*       printf("%lu %g %g %g %g\n", n_case, n_absorbe, attenuation, f_absorbe, I0); */

      /* On s'occupe des autres couches */
      point_entree = point_impact;
      for ( i_y = 1 ; i_y < echantillon->nbr_y ; i_y++ ) {
	/* Atténuation due au point précédent */
	I0 *= attenuation;

	/* Nouveau point d'impact, couche ultérieure
	   On travaille en fait en supposant qu'un point mémorise l'absorption des demi-tranches avant et après
	   Or ces demi-tranches sont homogènes p. hyp. et d'épaisseur traversée dd
	   => un calcul simple suffit !
	   [REMARQUE : ainsi, la dernière couche compte un peu trop de photons car elle est comptée de largeur simple et non demi]
	*/
	point_impact = point_entree;
	point_impact.x += u_faisceau->x * (double) i_y * delta_t;
	point_impact.y += u_faisceau->y * (double) i_y * delta_t;
	point_impact.z += u_faisceau->z * (double) i_y * delta_t;

	/* 1. Absorption par ce nouveau point */
	mu_total = ECH_CalculerMuTotal( echantillon, &point_impact, energie, &mu );
#ifdef _DEVELOPPE_
	if ( 0.0 == mu_total ) {
	  (void) fprintf(stderr, "%s,%s l.%u : PAS d'absorbance ! [case %lu = (%u, %u, %u)]\n",
			 __FILE__, __FUNCTION__, __LINE__, n_case, i_x, i_y, i_z);
	  abort();
	}
#endif

	/* 2. Atténuation, absorption */
	attenuation = exp( -mu_total * dd );
	f_absorbe = 1.0 - exp( -mu * dd );

	/* 3. Nombre de photons absorbés */
	n_absorbe = I0 * f_absorbe;

	/* 4. Mémorisation */
	n_case = ECH_IJK_v_Case( echantillon, i_x, i_y, i_z );
	echantillon->nombre[ n_case ] += n_absorbe;

/*  	printf("%lu %g %g %g %g\n", n_case, n_absorbe, attenuation, f_absorbe, I0); */
      }
    }
  }
#endif

  /* Calcul sans histoire */
  return 0;
}

/* -------- Trouver les indices limites de case pour un faisceau indiqué -------- */

static int FLUO_IndicesLimites(Echantillon *echantillon, Faisceau *faisceau, Vecteur3D *u_faisceau, double t,
			       unsigned int *ix_min, unsigned int *ix_max, unsigned int *iz_min, unsigned int *iz_max )
{
  double x_min, x_max, z_min, z_max;

  /* On initialise aux valeurs de l'échantillon */
  *ix_min = 0; *ix_max = echantillon->nbr_x - 1;
  *iz_min = 0; *iz_max = echantillon->nbr_z - 1;

  /* Limites du faisceau */
  x_min = - faisceau->largeur/2.0 + u_faisceau->x * t; z_min = - faisceau->hauteur / 2.0 + u_faisceau->z * t;
  x_max =   faisceau->largeur/2.0 + u_faisceau->x * t; z_max =   faisceau->hauteur / 2.0 + u_faisceau->z * t;

  /* 1. Débordements grossiers : faisceau hors échantillon */
  if ( ( x_min >= echantillon->larg / 2.0 ) || ( x_max <= -echantillon->larg / 2.0 ) ) return +1;
  if ( ( z_min >= echantillon->haut / 2.0 ) || ( z_max <= -echantillon->haut / 2.0 ) ) return +1;

  /* 2. On cale les bords du faisceau sur l'échantillon */
  if ( x_min < -echantillon->larg / 2.0 ) x_min = -echantillon->larg / 2.0;
  if ( x_max >  echantillon->larg / 2.0 ) x_max =  echantillon->larg / 2.0;
  if ( z_min < -echantillon->haut / 2.0 ) z_min = -echantillon->haut / 2.0;
  if ( z_max >  echantillon->haut / 2.0 ) z_max =  echantillon->haut / 2.0;

  /* On travaille en y = 0 puisque c'est là qu'est le faisceau ! */
  /* a. Bornes inférieures */
  (void) ECH_XYZ_v_IJK( echantillon, x_min, 0.0, z_min, ix_min, NULL, iz_min );

  /* b. Bornes supérieures */
  (void) ECH_XYZ_v_IJK( echantillon, x_max, 0.0, z_max, ix_max, NULL, iz_max );

  return 0;
}

/* -------- Cas d'un faisceau ponctuel -------- */

static int FLUO_Init_Ponctuel(Echantillon *echantillon, Faisceau *faisceau, double energie)
{
  double I0;
  double distance, densite __attribute__((__unused__)) = 1.0;
  double mu = 0.0, mu_total = 0.0; /* mu : absorption photo-électrique seule ; mu_total : diffusion inclue */
  double n_absorbe;
  int retour;
  unsigned int i_x, i_y, i_z;
  unsigned long n_case;
  boolean est_homogene, sortie = FAUX;
  Vecteur3D point_entree, u_faisceau;

  /* Vecteur directeur unitaire de la droite */
  u_faisceau.x = sin( faisceau->orientation.theta );
  u_faisceau.y = cos( faisceau->orientation.theta ) * cos( faisceau->orientation.phi );
  if ( 0.0 == u_faisceau.y ) {
    (void) fprintf(stderr, "FLUO_Init_Ponctuel : Faisceau parallèle à l'échantillon\n");
    return -2;
  }
  u_faisceau.z = -cos( faisceau->orientation.theta ) * sin( faisceau->orientation.phi );

  /* Intensité du faisceau sur la face d'entrée */
  I0 = FAISCEAU_LireIntensite( faisceau, 0.0, 0.0, energie ); /* Effets de I(E) pris en comptes... */

  /* Calcul du coefficient d'absorption pour l'énergie indiquée */
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    est_homogene = VRAI;
    mu_total = ECH_CalculerMuTotal( echantillon, NULL, energie, &mu );
  } else {
    est_homogene = FAUX;
    if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      mu_total = ECH_CalculerMuTotal( echantillon, NULL, energie, &mu );
    }
  }

  /* Recherche de la case illuminée par le faisceau
     Attention, on ne peut pas utiliser ECH_XYZ_v_IJK qui donne les index pour la case BasGaucheDevant
     alors que l'on veut ici la case la plus proche du point d'impact
     16/07/2008 : repassé à ECH_XYZ_v_IJK() par cohérence avec la suite ; comprends plus la remarque ci-dessus !
        en effet, case est définie comme allant de (x, y, z) à (x + dx, y + dy, z + dz) donc le faisceau tape par définition
	dans la case (x, y, z)...
 */
  point_entree = faisceau->orientation.position;
  ECH_XYZ_v_IJK( echantillon, point_entree.x, point_entree.y, point_entree.z, &i_x, &i_y, &i_z );

/*   i_x = rint( ( point_entree.x + echantillon->larg / 2.0 ) / echantillon->pas_x ); */
/*   i_y = rint( point_entree.y / echantillon->pas_y ); */
/*   i_z = rint( ( point_entree.z + echantillon->haut / 2.0 ) / echantillon->pas_z ); */
#ifdef _DEVELOPPE_
/*   { */
/*     printf("%s -- Point d'entrée : (%g, %g, %g)=(%u,%u,%u) ; faisceau selon (%g, %g, %g)\n", __FUNCTION__, */
/* 	   point_entree.x, point_entree.y, point_entree.z, i_x, i_y, i_z, */
/* 	   u_faisceau.x, u_faisceau.y, u_faisceau.z );  */
/*   } */
#endif
  while ( FAUX == sortie ) {
    /* On calcule le coefficient d'absorption pour l'énergie indiquée */
    n_case = ECH_IJK_v_Case( echantillon, i_x, i_y, i_z );
    if ( FAUX == est_homogene ) {
      if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	densite = echantillon->valeurs.homogene.concentration[ n_case ];
      } else {
	mu_total = ECH_CalculerMuTotal( echantillon, &point_entree, energie, &mu );
      }
    }
#ifdef _DEVELOPPE_
/*     { */
/*       double x, y, z; */

/*       ECH_Case_v_XYZ( echantillon, n_case, &x, &y, &z ); */
/*       printf("En (%g, %g, %g)=(%u, %u, %u)=%lu : (...)\n", x, y, z, i_x, i_y, i_z, n_case ); */
/*     } */
    if ( n_case >= echantillon->nbr_x * echantillon->nbr_y * echantillon->nbr_z ) {
      (void) fprintf(stderr, "%s : Numéro de case ABSURDE [%lu, max %lu] ; x = %g [%u:%u] y=%u [%u] z=%g [%u:%u]\n", __FUNCTION__,
		     n_case, (unsigned long) echantillon->nbr_x * echantillon->nbr_y * echantillon->nbr_z,
		     point_entree.x, i_x, echantillon->nbr_x, i_y, echantillon->nbr_y, point_entree.z, i_z, echantillon->nbr_z);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return -1;
    }
#endif

    /* Distance parcourue dans cette case */
    distance = ECH_TrouverCaseSuivante( echantillon, &u_faisceau, &point_entree, &i_x, &i_y, &i_z, &sortie );
    distance /= 10000.0; /* Travail en microns, on veut des cm pour les µ/rho */

    /* Fraction de photons absorbés */
    n_absorbe = I0 * ( 1.0 - exp( - mu * distance ) );
    I0 *= exp( - mu_total * distance );

#ifdef _DEVELOPPE_
/*     printf( "(...) %g [d=%g] --> (%g, %g, %g)\n", n_absorbe, distance, point_entree.x, point_entree.y, point_entree.z ); */
#endif

    echantillon->nombre[ n_case ] += n_absorbe;
    /* On passe à la case suivante... [sauf si l'on est sorti] */
  }
  
  /* On ajoute la contribution aux spectres de fluorescence idéaux */
  retour = FLUO_SpectresIdeaux( echantillon, energie );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_SpectresIdeaux()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -4;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Absorption entre deux points supposés assez proche pour que le coefficient varie linéairement entre les deux -------- */

double FLUO_PhotonsAbsorbes(Echantillon *echantillon, Vecteur3D *point_entree, Vecteur3D *point_sortie, double distance,
				   double mu, double mu_total, double energie, double *attenuation)
{
  double fraction_photons_absorbes;

  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    fraction_photons_absorbes = 1.0 - exp( - mu * distance );
    *attenuation = exp( -mu_total * distance );
  } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    double d_entree, d_sortie, d_moyenne;

    /* Estimation de la densité moyenne */
    d_entree = ECH_LireDensite( echantillon, point_entree );
    d_sortie = ECH_LireDensite( echantillon, point_sortie );
    d_moyenne = ( d_entree + d_sortie ) / 2.0;

    fraction_photons_absorbes = 1.0 - exp( - mu * d_moyenne * distance );
    *attenuation = exp( -mu_total * d_moyenne * distance );
  } else {
    double mu_entree, mu_sortie;
    double mu_t_entree, mu_t_sortie;

    /* Coefficients d'absorption en chaque point */
    mu_t_entree = ECH_CalculerMuTotal( echantillon, point_entree, energie, &mu_entree );
    mu_t_sortie = ECH_CalculerMuTotal( echantillon, point_sortie, energie, &mu_sortie );

    /* Estimation du coefficient d'absorption moyen */
    mu = ( mu_entree + mu_sortie ) / 2.0;
    mu_total = ( mu_t_entree + mu_t_sortie ) / 2.0;

    fraction_photons_absorbes = 1.0 - exp( - mu * distance );
    *attenuation = exp( -mu * distance );
  }

  /* Calcul terminé */
  return fraction_photons_absorbes;
}

/* -------- Calcul du spectre de fluorescence [& diffusion] idéal pour chaque composition de l'échantillon -------- */

static int FLUO_SpectresIdeaux(Echantillon *echantillon, double energie_excitation)
{
  int retour = -2;

  /* On ne remet pas à 0 le spectre, pour ajouter les contributions de toutes les énergies [y compris harmoniques] */
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    retour = Matiere_CalculerFluo( echantillon->valeurs.homogene.composition, energie_excitation, FAUX );
  } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    retour = Matiere_CalculerFluo( echantillon->valeurs.homogene.composition, energie_excitation, FAUX );
  } else {
    struct matiere *composition;

    /* Le plus simple : on le fait pour chaque composition... */
    for ( composition = premiere_composition ; composition != NULL ; composition = composition->suivante ) {
      retour = Matiere_CalculerFluo( composition, energie_excitation, FAUX );
      if ( retour != 0 ) {
	AVERTISSEMENT( "Erreur dans Matiere_CalculerFluo() code %d", retour );
	return -20;
      }
    }
  }

  /* Préparation terminée */
  return retour;
}
