/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 2003, Emmanuel CURIS

   Calcul thorique de la fluorescence des lments
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "ascii_utils.h"
#include "chimie_utils.h"
#include "utils_tables.h"

#include "absorption.h"
#include "fluorescence.h" /* Dfinition des lignes de fluorescence */
#include "graphes.h"
#include "lissage.h" /* Pour GRF_LireValeur */
#include "theorie.h"
#include "science.h" /* Pour la conversion E <-> longueur d'onde */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define Z_MAX_FLUORESCENCE AMERICIUM

/* Rendements de fluorescence */

static float **FLUO_Rendements_Raies = NULL;

#define FICHIER_RENDEMENTS_K  PREFIXE_REP"/Lase/Raies/intensites_K.txt"
#define FICHIER_RENDEMENTS_L1 PREFIXE_REP"/Lase/Raies/intensites_L1.txt"
#define FICHIER_RENDEMENTS_L2 PREFIXE_REP"/Lase/Raies/intensites_L2.txt"
#define FICHIER_RENDEMENTS_L3 PREFIXE_REP"/Lase/Raies/intensites_L3.txt"

/* nergies des lignes d'mission [eV] */

static float **FLUO_Energies_Raies = NULL;

#define FICHIER_ENERGIES_K  PREFIXE_REP"/Lase/Raies/energies_K.asc"
#define FICHIER_ENERGIES_L1 PREFIXE_REP"/Lase/Raies/energies_L1.asc"
#define FICHIER_ENERGIES_L2 PREFIXE_REP"/Lase/Raies/energies_L2.asc"
#define FICHIER_ENERGIES_L3 PREFIXE_REP"/Lase/Raies/energies_L3.asc"

/* Largeurs naturelles des raies d'mission [eV] */

static float **FLUO_Largeurs_Raies = NULL;

#define FICHIER_LARGEURS_K  PREFIXE_REP"/Lase/Raies/largeurs_K.asc"
#define FICHIER_LARGEURS_L1 PREFIXE_REP"/Lase/Raies/largeurs_L1.asc"
#define FICHIER_LARGEURS_L2 PREFIXE_REP"/Lase/Raies/largeurs_L2.asc"
#define FICHIER_LARGEURS_L3 PREFIXE_REP"/Lase/Raies/largeurs_L3.asc"

/* Transitions Coster-Kronig */

static float **FLUO_TableCosterKronig = NULL;

#define FICHIER_COSTER_KRONIG PREFIXE_REP"/Lase/Raies/CosterKronig.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int FLUO_ChargerTableCK   (void);
static int FLUO_ChargerTableRaies(const char *f_energies, const char *f_rendements, const char *f_largeurs, unsigned int famille, unsigned int nbr_raies);
static int FLUO_CK_DebutTable    (int transition);

static int FLUO_Raie_Gauss  (Graphe *graphe, double energie, double intensite, double largeur);
static int FLUO_Raie_Lorentz(Graphe *graphe, double energie, double intensite, double largeur);
static int FLUO_ZDepart( int raie );

static inline double FLUO_CalculerMu  ( Graphe *mu_seuil, double energie, int Z, int seuil );
static inline int    FLUO_SimulerRaies(Graphe *graphe, int Z, int seuil, int premiere_raie, int derniere_raie, int premiere_lacune, double n_lacunes);

static inline int FLUO_SimulerRaies_K ( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_L1( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_L2( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_L3( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_M1( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_M2( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_M3( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_M4( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_M5( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_N1( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_N2( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_N3( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_N4( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_N5( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_N6( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_N7( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_O1( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_O2( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_O3( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_O4( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_O5( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_O6( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_O7( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_P1( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_P2( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_P3( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_Q1( Graphe *graphe, int Z, double f_massique );

#ifdef AVEC_TRANSURANIENS_
static inline int FLUO_SimulerRaies_P4( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_P5( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_Q2( Graphe *graphe, int Z, double f_massique );
static inline int FLUO_SimulerRaies_Q3( Graphe *graphe, int Z, double f_massique );
#endif

static int (*Table_Simulations[ NOMBRE_SEUILS ])(Graphe *, int, double) = {
  FLUO_SimulerRaies_K,
  FLUO_SimulerRaies_L1, FLUO_SimulerRaies_L2, FLUO_SimulerRaies_L3,
  FLUO_SimulerRaies_M1, FLUO_SimulerRaies_M2, FLUO_SimulerRaies_M3, FLUO_SimulerRaies_M4, FLUO_SimulerRaies_M5,
  FLUO_SimulerRaies_N1, FLUO_SimulerRaies_N2, FLUO_SimulerRaies_N3, FLUO_SimulerRaies_N4, FLUO_SimulerRaies_N5, FLUO_SimulerRaies_N6, FLUO_SimulerRaies_N7,
  FLUO_SimulerRaies_O1, FLUO_SimulerRaies_O2, FLUO_SimulerRaies_O3, FLUO_SimulerRaies_O4, FLUO_SimulerRaies_O5, FLUO_SimulerRaies_O6, FLUO_SimulerRaies_O7,
  FLUO_SimulerRaies_P1, FLUO_SimulerRaies_P2, FLUO_SimulerRaies_P3,
  FLUO_SimulerRaies_Q1
#ifdef AVEC_TRANSURANIENS_
  , FLUO_SimulerRaies_P4, FLUO_SimulerRaies_P5, FLUO_SimulerRaies_Q2, FLUO_SimulerRaies_Q3
#endif
};

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

int FLUO_InitialiserTables( void )
{
  int i, retour;

  /* Prparation de la table des tables */
  if ( NULL == FLUO_Energies_Raies ) {
    FLUO_Energies_Raies = (float **) calloc( NOMBRE_RAIES_FLUO, sizeof( float * ) );
    if ( NULL == FLUO_Energies_Raies ) {
      (void) fprintf(stderr, "FLUO_InitialiserTables : Mmoire insuffisante [FLUO_Energies_Raies]\n");
      return -2;
    }

    FLUO_Rendements_Raies = (float **) calloc( NOMBRE_RAIES_FLUO, sizeof( float * ) );
    if ( NULL == FLUO_Rendements_Raies ) {
      (void) fprintf(stderr, "FLUO_InitialiserTables : Mmoire insuffisante [FLUO_Rendements_Raies]\n");
      free( FLUO_Energies_Raies ); FLUO_Energies_Raies = NULL;
      return -2;
    }

    FLUO_Largeurs_Raies = (float **) calloc( NOMBRE_RAIES_FLUO, sizeof( float * ) );
    if ( NULL == FLUO_Largeurs_Raies ) {
      (void) fprintf(stderr, "FLUO_InitialiserTables : Mmoire insuffisante [FLUO_Largeurs_Raies]\n");
      free( FLUO_Energies_Raies ); FLUO_Energies_Raies = NULL;
      free( FLUO_Rendements_Raies ); FLUO_Rendements_Raies = NULL;
      return -2;
    }
  }

  /* On libre d'ventuelles anciennes valeurs */
  for ( i = 0 ; i < NOMBRE_RAIES_FLUO ; i++ ) {
    free( FLUO_Energies_Raies[ i ] ); FLUO_Energies_Raies[ i ] = NULL;
    free( FLUO_Largeurs_Raies[ i ] ); FLUO_Largeurs_Raies[ i ] = NULL;
    free( FLUO_Rendements_Raies[ i ] ); FLUO_Rendements_Raies[ i ] = NULL;
  }

  /* On charge les donnes pour les raies K */
  retour = FLUO_ChargerTableRaies( FICHIER_ENERGIES_K, FICHIER_RENDEMENTS_K, FICHIER_LARGEURS_K, RAIE_K_L1, RAIE_K_Q1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FLUO_InitialiserTables : ERREUR dans FLUO_ChargerTableRaies(), raies K\n");
    return -1;
  }

  /* On charge les donnes pour les raies L1 */
  retour = FLUO_ChargerTableRaies( FICHIER_ENERGIES_L1, FICHIER_RENDEMENTS_L1, FICHIER_LARGEURS_L1, RAIE_L1_L2, RAIE_L1_Q1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FLUO_InitialiserTables : ERREUR dans FLUO_ChargerTableRaies(), raies L1\n");
    return -1;
  }
  /* On charge les donnes pour les raies L2 */
  retour = FLUO_ChargerTableRaies( FICHIER_ENERGIES_L2, FICHIER_RENDEMENTS_L2, FICHIER_LARGEURS_L2, RAIE_L2_L3, RAIE_L2_Q1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FLUO_InitialiserTables : ERREUR dans FLUO_ChargerTableRaies(), raies L2\n");
    return -1;
  }
  /* On charge les donnes pour les raies L3 */
  retour = FLUO_ChargerTableRaies( FICHIER_ENERGIES_L3, FICHIER_RENDEMENTS_L3, FICHIER_LARGEURS_L3, RAIE_L3_M1, RAIE_L3_Q1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FLUO_InitialiserTables : ERREUR dans FLUO_ChargerTableRaies(), raies L3\n");
    return -1;
  }

  /* Initialisation russie */
  return 0;
}

static int FLUO_ChargerTableRaies(const char *f_energies, const char *f_rendements, const char *f_largeurs, unsigned int raie_1, unsigned int raie_n)
{
  FILE *fichier;
  unsigned int nbr_colonnes;
  unsigned int nbr_raies;
  unsigned long i, nbr_lignes;
  double *valeurs;

  /* On cre les tables ncessaires. Si problme, NULL, cas trait correctement ensuite... */
  nbr_raies = raie_n - raie_1 + 1;
  if ( 0 == nbr_raies ) return 0; /* Aucune raie  charger ! */

  for ( i = raie_1 ; i <= raie_n ; i++ ) {
    FLUO_Energies_Raies[ i ] = TH_Creer_Table( FLUO_ZDepart( i ), Z_MAX_FLUORESCENCE );
    if ( FLUO_Energies_Raies[ i ] == NULL ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_Creer_Table() / nergies, raie %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i);
    }

    FLUO_Largeurs_Raies[ i ] = TH_Creer_Table( FLUO_ZDepart( i ), Z_MAX_FLUORESCENCE );
    if ( FLUO_Largeurs_Raies[ i ] == NULL ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_Creer_Table() / largeur, raie %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i);
    }

    FLUO_Rendements_Raies[ i ] = TH_Creer_Table( FLUO_ZDepart( i ), Z_MAX_FLUORESCENCE );
    if ( FLUO_Rendements_Raies[ i ] == NULL ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_Creer_Table() / rendements, raie %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i);
    }
  }


  /* On commence par la table des nergies */
  fichier = TH_OuvrirFichier( f_energies, &nbr_lignes, &nbr_colonnes, &valeurs );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir %s [nergies]\n",
		   __FILE__, __FUNCTION__, __LINE__, f_energies);
  } else {
    /* On lit ces nergies */
    for ( i = 0 ; i < nbr_lignes ; i++ ) {
      int Z;
      unsigned int j;

      Z = TH_LireLigne( fichier, nbr_colonnes, valeurs, NULL );
      if ( Z < 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans TH_LireLigne() [ligne %lu]\n",
		       __FILE__, __FUNCTION__, __LINE__, i);
	(void) fclose( fichier );
	return -4;
      }

      /* Colonnes : transitions par ordre croissant, KL1, KL2, ... */
      for ( j = 0 ; j < nbr_colonnes ; j++ ) {
	if ( j < nbr_raies ) {
#ifdef _DEVELOPPE_
	  if ( ( isnan( valeurs[ j ] ) ) || ( isinf( valeurs[ j ] ) ) ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Valeur lue absurde [NaN/Inf], fichier=%s, lment=%d (%s), colonne=%u\n",
			   __FILE__, __FUNCTION__, __LINE__, f_energies, Z, ATM_Z_vers_symbole( Z ), j);
	    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
	    valeurs[ j ] = -1.0; /* Valeur inconnue ! */
	  }
#endif
	  TH_Table_Remplir( FLUO_Energies_Raies[ raie_1 + j ], Z, FLUO_ZDepart( raie_1 + j ), Z_MAX_FLUORESCENCE, valeurs[ j ] );
	}
      }
    }

    (void) fclose( fichier );
    free( valeurs ); /* donc plus besoin de la table des valeurs par ligne */
  }

  /* 2. Largeurs naturelles */
  fichier = TH_OuvrirFichier( f_largeurs, &nbr_lignes, &nbr_colonnes, &valeurs );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "FLUO_ChargerTableRaies : ERREUR dans FLUO_OuvrirFichier() [%s]\n", f_largeurs);
    return -1;
  }
  if ( nbr_colonnes < 2 ) {
    (void) fprintf(stderr, "FLUO_ChargerTableRaies : trop peu de colonnes, fichier inutilisable [%s]\n", f_largeurs);
    (void) fclose( fichier );
    return -3;
  }

  for ( i = 0 ; i < nbr_lignes ; i++ ) {
    int Z;
    unsigned int j;

    Z = TH_LireLigne( fichier, nbr_colonnes, valeurs, NULL );
    if ( Z < 0 ) {
      (void) fprintf(stderr, "FLUO_ChargerTableRaies : ERREUR dans TH_LireLigne() [%s ligne %lu]\n", f_largeurs, i);
      (void) fclose( fichier );
      return -4;
    }

    /* Colonnes : transitions par ordre croissant, KL1, KL2, ... */
    for ( j = 0 ; j < nbr_colonnes ; j++ ) {
      if ( j < nbr_raies ) {
	TH_Table_Remplir( FLUO_Largeurs_Raies[ raie_1 + j ], Z, FLUO_ZDepart( raie_1 + j ), Z_MAX_FLUORESCENCE, valeurs[ j ] );
      }
    }
  }

  (void) fclose( fichier );
  free( valeurs ); /* donc plus besoin de la table des valeurs par ligne */

  /* 3. Rendements */
  fichier = TH_OuvrirFichier( f_rendements, &nbr_lignes, &nbr_colonnes, &valeurs );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "FLUO_ChargerTableRaies : ERREUR dans FLUO_OuvrirFichier() [%s]\n", f_rendements);
    return -1;
  }
  if ( nbr_colonnes < 2 ) {
    (void) fprintf(stderr, "FLUO_ChargerTableRaies : trop peu de colonnes, fichier inutilisable [%s]\n", f_rendements);
    (void) fclose( fichier );
    return -3;
  }

  for ( i = 0 ; i < nbr_lignes ; i++ ) {
    int Z;
    unsigned int j;
    double i_totale, r_total;

    Z = TH_LireLigne( fichier, nbr_colonnes, valeurs, NULL );
    if ( Z < 0 ) {
      (void) fprintf(stderr, "FLUO_ChargerTableRaies : ERREUR dans TH_LireLigne() [%s ligne %lu]\n", f_rendements, i);
      (void) fclose( fichier );
      return -4;
    }

    /* Le fichier contient le rendement GLOBAL des raies K et les intensits relatives de ces raies
      On veut le rendement individuel de chaque raie : r(raie) = r(total) * I(raie)/I(total K)
    */
    r_total = valeurs[ 0 ] / 100.0; /* Rendement total : 1re colonne [en %, or on veut la fraction] */
    for (j = 0, i_totale = 0.0 ; j < nbr_colonnes - 1 ; j++ ) {
      i_totale += valeurs[ j + 1 ];
    }

    /* Colonnes : supposes w puis les colonnes KL1, KL2,... dans l'ordre */
    if ( i_totale > 0.0 ) { /* Il existe bien des raies */
      for ( j = raie_1 ; j <= raie_n ; j++ ) {
	if ( ( j - raie_1 + 1 ) < nbr_colonnes ) {
#ifdef _DEVELOPPE_
	  if ( ( isnan( valeurs[ j - raie_1 + 1 ] ) ) || ( isinf( valeurs[ j - raie_1 + 1 ] ) ) ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Valeur lue absurde [NaN/Inf], fichier=%s, lment=%d (%s), colonne=%u\n",
			   __FILE__, __FUNCTION__, __LINE__, f_rendements, Z, ATM_Z_vers_symbole( Z ), j - raie_1 + 1);
	    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
	    valeurs[ j ] = -1.0; /* Valeur inconnue ! */
	  }
#endif

	  TH_Table_Remplir( FLUO_Rendements_Raies[ j ], Z, FLUO_ZDepart( j ), Z_MAX_FLUORESCENCE, valeurs[ j - raie_1 + 1 ] * r_total / i_totale );
	}
      }
    }
  }

  (void) fclose( fichier );
  free( valeurs ); /* donc plus besoin de la table des valeurs par ligne */

#ifdef _DEVELOPPE_
/*   for ( i = HYDROGENE ; i <= URANIUM ; i++ ) { */
/*     printf("%2s : raie Ka1 = %8.2f eV LMH = %6.2f eV r = %g %%\n", ATM_Z_vers_symbole( i ), */
/* 	   (double) TH_Table_Lire( FLUO_Energies_Raies[ raie_1 ], i, FLUO_ZDepart( raie_1 ), Z_MAX_FLUORESCENCE ), */
/* 	   (double) TH_Table_Lire( FLUO_Largeurs_Raies[ raie_1 ], i, FLUO_ZDepart( raie_1 ), Z_MAX_FLUORESCENCE ), */
/* 	   100.0 * (double) TH_Table_Lire( FLUO_Rendements_Raies[ raie_1 ], i, FLUO_ZDepart( raie_1 ), Z_MAX_FLUORESCENCE ) ); */
/*   } */
#endif

  /* Chargement russi, tout s'est bien pass */
  return 0;
}

/* -------- Charger la table des rendements de transitions Coster-Kronig -------- */

static int FLUO_ChargerTableCK(void)
{
  FILE *fichier;
  unsigned long nbr_lignes;
  unsigned int i, nbr_colonnes;
  double *valeurs;

  /* On rserve la table des transitions */
  FLUO_TableCosterKronig = ( float ** ) calloc( NOMBRE_TRANSITIONS_CK, sizeof( float * ) );
  if ( NULL == FLUO_TableCosterKronig ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante pour rserver la table des tables\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* On rserve les tables pour chacune des transitions */
  for ( i = 0 ; i < NOMBRE_TRANSITIONS_CK ; i++ ) {
    int Z_min, Z_max = Z_MAX_FLUORESCENCE;

    Z_min = FLUO_CK_DebutTable( i );
    if ( Z_min >= Z_max ) continue;

    FLUO_TableCosterKronig[ i ] = TH_Creer_Table( Z_min, Z_max );
    if ( NULL == FLUO_TableCosterKronig[ i ] ) {
      (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante pour la table de transition Conster-Kronig num. %u\n",
		     __FILE__, __FUNCTION__, __LINE__, i);
    }
  }

  fichier = TH_OuvrirFichier( FICHIER_COSTER_KRONIG, &nbr_lignes, &nbr_colonnes, &valeurs );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_OuvrirFichier()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On lit, ligne  ligne */
  for ( i = 0 ; i < nbr_lignes ; i++ ) {
    int Z;
    unsigned int j;

    Z = TH_LireLigne( fichier, nbr_colonnes, valeurs, NULL );
    if ( Z < 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans TH_LireLigne() [ligne %u]\n",
		     __FILE__, __FUNCTION__, __LINE__, i);
      (void) fclose( fichier );
      return -4;
    }

    /* Colonnes : transitions CK par ordre croissant, en _pourcentages_ */
    for ( j = 0 ; j < nbr_colonnes ; j++ ) {
      if ( j < NOMBRE_TRANSITIONS_CK ) {
	TH_Table_Remplir( FLUO_TableCosterKronig[ j ], Z, FLUO_CK_DebutTable( j ), Z_MAX_FLUORESCENCE, valeurs[ j ] / 100.0 );
      }
    }
  }

  (void) fclose( fichier );
  return 0;
}

/* -------- Utilisation des tables -------- */

static int FLUO_CK_DebutTable(int transition)
{
  switch ( transition )
  {
  case CK_L1L2: return BORE;
  case CK_L1L3: case CK_L2L3: return AZOTE;

  case CK_M1M2: return ALUMINIUM;
  case CK_M1M3: case CK_M2M3: return PHOSPHORE;
  case CK_M1M4: case CK_M2M4: case CK_M3M4: return SCANDIUM;
  case CK_M1M5: case CK_M2M5: case CK_M3M5: case CK_M4M5: return CHROME;

  case CK_N1N2: return GALLIUM;
  case CK_N1N3: case CK_N2N3: return ARSENIC;
  case CK_N1N4: case CK_N2N4: case CK_N3N4: return YTTERBIUM;
  case CK_N1N5: case CK_N2N5: case CK_N3N5: case CK_N4N5: return MOLYBDENE;
  case CK_N1N6: case CK_N2N6: case CK_N3N6: case CK_N4N6: case CK_N5N6: return LANTHANE;
  case CK_N1N7: case CK_N2N7: case CK_N3N7: case CK_N4N7: case CK_N5N7: case CK_N6N7: return EUROPIUM;

  case CK_O1O2: return INDIUM;
  case CK_O1O3: case CK_O2O3: return ANTIMOINE;
  case CK_O1O4: case CK_O2O4: case CK_O3O4: return LUTETIUM;
  case CK_O1O5: case CK_O2O5: case CK_O3O5: case CK_O4O5: return TUNGSTENE;
  case CK_O1O6: case CK_O2O6: case CK_O3O6: case CK_O4O6: case CK_O5O6: return ACTINIUM;
  case CK_O1O7: case CK_O2O7: case CK_O3O7: case CK_O4O7: case CK_O5O7: case CK_O6O7: return AMERICIUM;

  case CK_P1P2: return THALLIUM;
  case CK_P1P3: case CK_P2P3: return BISMUTH;

#ifdef AVEC_TRANSURANIENS_
  case CK_P1P4: case CK_P2P4: case CK_P3P4: return LAWRENCIUM;
  case CK_P1P5: case CK_P2P5: case CK_P3P5: case CK_P4P5: return SEABORGIUM;
#endif

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Transition CK inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, transition);
    return -1;
  }
}

double FLUO_CosterKronig( int Z, int seuil_depart, int seuil_arrivee )
{
  int transition_CK = -1;
  double rendement_CK = 0.0;

  if ( seuil_depart >= seuil_arrivee ) {
    (void) fprintf(stderr, "%s,%s l.%u : Transition CK absurde [%s -> %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( seuil_depart ), TH_TexteSeuil( seuil_arrivee ) );
    return 0.0;
  }

  switch ( seuil_depart )
  {
  case SEUIL_K: return 0.0; /* Pas de Coster-Kronig pour un seuil K */
  case SEUIL_LI:
    if ( SEUIL_LII  == seuil_arrivee ) transition_CK = CK_L1L2;
    if ( SEUIL_LIII == seuil_arrivee ) transition_CK = CK_L1L3;
    break;
  case SEUIL_LII:
    if ( SEUIL_LIII == seuil_arrivee ) transition_CK = CK_L2L3;
    break;

  case SEUIL_MI:
    if ( seuil_arrivee <= SEUIL_MV ) transition_CK = CK_M1M2 + ( seuil_arrivee - SEUIL_MII );
    break;
  case SEUIL_MII:
    if ( seuil_arrivee <= SEUIL_MV ) transition_CK = CK_M2M3 + ( seuil_arrivee - SEUIL_MIII );
    break;
  case SEUIL_MIII:
    if ( seuil_arrivee <= SEUIL_MV ) transition_CK = CK_M3M4 + ( seuil_arrivee - SEUIL_MIV );
    break;
  case SEUIL_MIV:
    if ( SEUIL_MV == seuil_arrivee ) transition_CK = CK_M4M5;
    break;

  case SEUIL_NI:
    if ( seuil_arrivee <= SEUIL_NVII ) transition_CK = CK_N1N2 + ( seuil_arrivee - SEUIL_NII );
    break;
  case SEUIL_NII:
    if ( seuil_arrivee <= SEUIL_NVII ) transition_CK = CK_N2N3 + ( seuil_arrivee - SEUIL_NIII );
    break;
  case SEUIL_NIII:
    if ( seuil_arrivee <= SEUIL_NVII ) transition_CK = CK_N3N4 + ( seuil_arrivee - SEUIL_NIV );
    break;
  case SEUIL_NIV:
    if ( seuil_arrivee <= SEUIL_NVII ) transition_CK = CK_N4N5 + ( seuil_arrivee - SEUIL_NV );
    break;
  case SEUIL_NV:
    if ( SEUIL_NVII == seuil_arrivee ) transition_CK = CK_N6N7;
    break;

  case SEUIL_OI:
    if ( seuil_arrivee <= SEUIL_OVII ) transition_CK = CK_O1O2 + ( seuil_arrivee - SEUIL_OII );
    break;
  case SEUIL_OII:
    if ( seuil_arrivee <= SEUIL_OVII ) transition_CK = CK_O2O3 + ( seuil_arrivee - SEUIL_OIII );
    break;
  case SEUIL_OIII:
    if ( seuil_arrivee <= SEUIL_OVII ) transition_CK = CK_O3O4 + ( seuil_arrivee - SEUIL_OIV );
    break;
  case SEUIL_OIV:
    if ( seuil_arrivee <= SEUIL_OVII ) transition_CK = CK_O4O5 + ( seuil_arrivee - SEUIL_OV );
    break;
  case SEUIL_OV:
    if ( SEUIL_OVII == seuil_arrivee ) transition_CK = CK_O6O7;
    break;
  }

  if ( -1 == transition_CK ) {
    (void) fprintf(stderr, "%s,%s l.%u : Transition CK inconnue [%s -> %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( seuil_depart ), TH_TexteSeuil( seuil_arrivee ) );
    return 0.0;
  }

  if ( NULL == FLUO_TableCosterKronig ) {
    int retour;

    retour = FLUO_ChargerTableCK( );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_ChargerTableCK() -> transitions Coster-Kronig ngliges\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return 0.0;
    }
  }

  /* Maintenant que l'on sait quelle table utiliser, on lit la valeur dedans */
  rendement_CK = TH_Table_Lire( FLUO_TableCosterKronig[ transition_CK ], Z, FLUO_CK_DebutTable( transition_CK ), Z_MAX_FLUORESCENCE);
  if ( ( rendement_CK < 0.0 ) || ( rendement_CK > 2.0 ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Rendement CK absurde [%g; Z=%s, transition %s -> %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, rendement_CK, ATM_Z_vers_symbole( Z ), TH_TexteSeuil( seuil_depart ), TH_TexteSeuil( seuil_arrivee ) );
  }

  return rendement_CK;
}

double FLUO_QuelleEnergie(int Z, int raie, double *rendement, double *largeur)
{
  float energie = 0.0, rdmt = 0.0, lg = 0.0;
  int Z_min;

  if ( NULL == FLUO_Energies_Raies ) {
    int retour;

    retour = FLUO_InitialiserTables( );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_InitialiserTables() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }
  }

  if ( raie >= 0 ) { /* Valeurs directes */
    Z_min = FLUO_ZDepart( raie );
    if ( Z_min < 0 ) {
      (void) fprintf(stderr, "FLUO_QuelleEnergie : ERREUR dans FLUO_ZDepart()\n");
      return -5.0;
    }
    if ( Z_min > Z_MAX_FLUORESCENCE ) { /* lments non grs */
      rdmt = 0.0;
      lg = -2.0;
      energie = -2.0;
    } else {
      energie = TH_Table_Lire( FLUO_Energies_Raies[ raie ], Z, Z_min, Z_MAX_FLUORESCENCE );
      lg      = TH_Table_Lire( FLUO_Largeurs_Raies[ raie ], Z, Z_min, Z_MAX_FLUORESCENCE );
      rdmt    = TH_Table_Lire( FLUO_Rendements_Raies[ raie ], Z, Z_min, Z_MAX_FLUORESCENCE );
    }
  } else {
    (void) fprintf(stderr, "FLUO_QuelleEnergie :  FAIRE --- groupe de raies [%d]\n", raie);
  }

  if ( rendement != NULL ) *rendement = (double) rdmt;
  if ( largeur != NULL ) *largeur = (double) lg;
  return (double) energie;
}

/* -------- Donne les raies extrmes existant pour une lacune dans l'lment Z indiqu --------
 * renvoie 0 si on a pu trouver ces raies
 *
 * raie_1 = la raie de plus basse nergie (raie basse)
 * raie_n = la raie de plus haute nergie (raie haute)
 */

int FLUO_RaiesExtremes(int Z, int couche_lacune, int *raie_1, int *raie_n)
{
  int raie_basse = -1, raie_haute = -1;

  if ( Z <= HELIUM ) return -1; /* Pas de fluorescence */

  /* Raie de dbut : facile, existe toujours !
     Raie de fin : dpend de l'lment... Compte-tenu de la valence, on ne considre que les couches pleines */
  switch ( couche_lacune )
  {
  case SEUIL_K:
    if ( Z < LITHIUM ) return -1; /* Pas de raie K avant remplissage de la 2s! */
    raie_basse = RAIE_K_L1;
    if ( Z <= NEON ) raie_haute = RAIE_K_L3;
    else if ( Z < POTASSIUM ) raie_haute = RAIE_K_M3;
    else if ( Z < RUBIDIUM )  raie_haute = RAIE_K_N3; /* Inclue les raies 3d pour K, Ca : pas idal, mais trou =>
							difficile de faire mieux simplement */
    else if ( Z < CESIUM )    raie_haute = RAIE_K_O3; /* Mme remarques, raies 4d */
    else if ( Z < FRANCIUM )  raie_haute = RAIE_K_P3; /* Mme remarques, raies 4f, 5d */
    else raie_haute = RAIE_K_Q1;
    break;

  case SEUIL_LI:
    if ( Z < SODIUM ) return -1; /* Pas de raie L avant remplissage de la 3s! */
    raie_basse = RAIE_L1_M1; /* On nglige L1L2 et L1L3 : trop basse nergie... */
    if ( Z < POTASSIUM ) raie_haute = RAIE_L1_M3;
    else if ( Z < RUBIDIUM )  raie_haute = RAIE_L1_N3; /* Inclue les raies 3d pour K, Ca : pas idal, mais trou =>
							  difficile de faire mieux simplement */
    else if ( Z < CESIUM )    raie_haute = RAIE_L1_O3; /* Mme remarques, raies 4d */
    else if ( Z < FRANCIUM )  raie_haute = RAIE_L1_P3; /* Mme remarques, raies 4f, 5d */
    else raie_haute = RAIE_L1_Q1;
    /* Ainsi, ne tient pas compte de Coster-Kronig : raies L1 pures ! */
    break;

  case SEUIL_LII:
    if ( Z < SODIUM ) return -1; /* Pas de raie L avant remplissage de la 3s! */
    raie_basse = RAIE_L2_M1; /* On nglige L2L3 : trop basse nergie... */
    if ( Z < POTASSIUM ) raie_haute = RAIE_L2_M3;
    else if ( Z < RUBIDIUM )  raie_haute = RAIE_L2_N3; /* Inclue les raies 3d pour K, Ca : pas idal, mais trou =>
							  difficile de faire mieux simplement */
    else if ( Z < CESIUM )    raie_haute = RAIE_L2_O3; /* Mme remarques, raies 4d */
    else if ( Z < FRANCIUM )  raie_haute = RAIE_L2_P3; /* Mme remarques, raies 4f, 5d */
    else raie_haute = RAIE_L2_Q1;
    /* Ainsi, ne tient pas compte de Coster-Kronig : raies L1 pures ! */
    break;

  case SEUIL_LIII:
    if ( Z < SODIUM ) return -1; /* Pas de raie L avant remplissage de la 3s! */
    raie_basse = RAIE_L3_M1;
    if ( Z < POTASSIUM ) raie_haute = RAIE_L3_M3;
    else if ( Z < RUBIDIUM )  raie_haute = RAIE_L3_N3; /* Inclue les raies 3d pour K, Ca : pas idal, mais trou =>
							  difficile de faire mieux simplement */
    else if ( Z < CESIUM )    raie_haute = RAIE_L3_O3; /* Mme remarques, raies 4d */
    else if ( Z < FRANCIUM )  raie_haute = RAIE_L3_P3; /* Mme remarques, raies 4f, 5d */
    else raie_haute = RAIE_L3_Q1;
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Seuil %d [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, couche_lacune, TH_TexteSeuil( couche_lacune ) );
  }

  if ( raie_1 != NULL ) *raie_1 = raie_basse;
  if ( raie_n != NULL ) *raie_n = raie_haute;

  /* On a pu trouver les raies demandes */
  return 0;
}

static int FLUO_ZDepart( int raie )
{
  switch ( raie )
  {
  case RAIE_K_L1: return LITHIUM;

  case RAIE_K_L2:
  case RAIE_L1_L2:
    return BORE;

  case RAIE_K_L3:
  case RAIE_L1_L3:  case RAIE_L2_L3:
    return AZOTE;

  case RAIE_K_M1:
  case RAIE_L1_M1: case RAIE_L2_M1: case RAIE_L3_M1:
    return SODIUM;

  case RAIE_K_M2:
  case RAIE_L1_M2: case RAIE_L2_M2: case RAIE_L3_M2:
    return ALUMINIUM;

  case RAIE_K_M3:
  case RAIE_L1_M3: case RAIE_L2_M3: case RAIE_L3_M3:
    return PHOSPHORE;

  case RAIE_K_M4:
  case RAIE_L1_M4: case RAIE_L2_M4: case RAIE_L3_M4:
    return SCANDIUM;

  case RAIE_K_M5:
  case RAIE_L1_M5: case RAIE_L2_M5: case RAIE_L3_M5:
    return CHROME;

  case RAIE_K_N1:
  case RAIE_L1_N1: case RAIE_L2_N1: case RAIE_L3_N1:
  case RAIE_M1_N1: case RAIE_M2_N1: case RAIE_M3_N1: case RAIE_M4_N1: case RAIE_M5_N1:
    return POTASSIUM;

  case RAIE_K_N2:
  case RAIE_L1_N2: case RAIE_L2_N2: case RAIE_L3_N2:
  case RAIE_M1_N2: case RAIE_M2_N2: case RAIE_M3_N2: case RAIE_M4_N2: case RAIE_M5_N2:
    return GALLIUM;

  case RAIE_K_N3:
  case RAIE_L1_N3: case RAIE_L2_N3: case RAIE_L3_N3:
  case RAIE_M1_N3: case RAIE_M2_N3: case RAIE_M3_N3: case RAIE_M4_N3: case RAIE_M5_N3:
    return ARSENIC;

  case RAIE_K_N4:
  case RAIE_L1_N4: case RAIE_L2_N4: case RAIE_L3_N4:
  case RAIE_M1_N4: case RAIE_M2_N4: case RAIE_M3_N4: case RAIE_M4_N4: case RAIE_M5_N4:
    return YTTRIUM;

  case RAIE_K_N5:
  case RAIE_L1_N5: case RAIE_L2_N5: case RAIE_L3_N5:
  case RAIE_M1_N5: case RAIE_M2_N5: case RAIE_M3_N5: case RAIE_M4_N5: case RAIE_M5_N5:
    return MOLYBDENE;

  case RAIE_K_N6:
  case RAIE_L1_N6: case RAIE_L2_N6: case RAIE_L3_N6:
  case RAIE_M1_N6: case RAIE_M2_N6: case RAIE_M3_N6: case RAIE_M4_N6: case RAIE_M5_N6:
    return LANTHANE;

  case RAIE_K_N7:
  case RAIE_L1_N7: case RAIE_L2_N7: case RAIE_L3_N7:
  case RAIE_M1_N7: case RAIE_M2_N7: case RAIE_M3_N7: case RAIE_M4_N7: case RAIE_M5_N7:
    return 66;

  case RAIE_K_O1:
  case RAIE_L1_O1: case RAIE_L2_O1: case RAIE_L3_O1:
  case RAIE_M1_O1: case RAIE_M2_O1: case RAIE_M3_O1: case RAIE_M4_O1: case RAIE_M5_O1:
  case RAIE_N1_O1: case RAIE_N2_O1: case RAIE_N3_O1: case RAIE_N4_O1: case RAIE_N5_O1: case RAIE_N6_O1: case RAIE_N7_O1:
    return RUBIDIUM;

  case RAIE_K_O2:
  case RAIE_L1_O2: case RAIE_L2_O2: case RAIE_L3_O2:
  case RAIE_M1_O2: case RAIE_M2_O2: case RAIE_M3_O2: case RAIE_M4_O2: case RAIE_M5_O2:
  case RAIE_N1_O2: case RAIE_N2_O2: case RAIE_N3_O2: case RAIE_N4_O2: case RAIE_N5_O2: case RAIE_N6_O2: case RAIE_N7_O2:
    return INDIUM;

  case RAIE_K_O3:
  case RAIE_L1_O3: case RAIE_L2_O3: case RAIE_L3_O3:
  case RAIE_M1_O3: case RAIE_M2_O3: case RAIE_M3_O3: case RAIE_M4_O3: case RAIE_M5_O3:
  case RAIE_N1_O3: case RAIE_N2_O3: case RAIE_N3_O3: case RAIE_N4_O3: case RAIE_N5_O3: case RAIE_N6_O3: case RAIE_N7_O3:
    return ANTIMOINE;

  case RAIE_K_O4:
  case RAIE_L1_O4: case RAIE_L2_O4: case RAIE_L3_O4:
  case RAIE_M1_O4: case RAIE_M2_O4: case RAIE_M3_O4: case RAIE_M4_O4: case RAIE_M5_O4:
  case RAIE_N1_O4: case RAIE_N2_O4: case RAIE_N3_O4: case RAIE_N4_O4: case RAIE_N5_O4: case RAIE_N6_O4: case RAIE_N7_O4:
    return LUTETIUM;

  case RAIE_K_O5:
  case RAIE_L1_O5: case RAIE_L2_O5: case RAIE_L3_O5:
  case RAIE_M1_O5: case RAIE_M2_O5: case RAIE_M3_O5: case RAIE_M4_O5: case RAIE_M5_O5:
  case RAIE_N1_O5: case RAIE_N2_O5: case RAIE_N3_O5: case RAIE_N4_O5: case RAIE_N5_O5: case RAIE_N6_O5: case RAIE_N7_O5:
    return TUNGSTENE;

  case RAIE_K_O6:
  case RAIE_L1_O6: case RAIE_L2_O6: case RAIE_L3_O6:
  case RAIE_M1_O6: case RAIE_M2_O6: case RAIE_M3_O6: case RAIE_M4_O6: case RAIE_M5_O6:
  case RAIE_N1_O6: case RAIE_N2_O6: case RAIE_N3_O6: case RAIE_N4_O6: case RAIE_N5_O6: case RAIE_N6_O6: case RAIE_N7_O6:
    return ACTINIUM;

  case RAIE_K_O7:
  case RAIE_L1_O7: case RAIE_L2_O7: case RAIE_L3_O7:
  case RAIE_M1_O7: case RAIE_M2_O7: case RAIE_M3_O7: case RAIE_M4_O7: case RAIE_M5_O7:
  case RAIE_N1_O7: case RAIE_N2_O7: case RAIE_N3_O7: case RAIE_N4_O7: case RAIE_N5_O7: case RAIE_N6_O7: case RAIE_N7_O7:
    return AMERICIUM;

  case RAIE_K_P1:
  case RAIE_L1_P1: case RAIE_L2_P1: case RAIE_L3_P1:
  case RAIE_M1_P1: case RAIE_M2_P1: case RAIE_M3_P1: case RAIE_M4_P1: case RAIE_M5_P1:
  case RAIE_N1_P1: case RAIE_N2_P1: case RAIE_N3_P1: case RAIE_N4_P1: case RAIE_N5_P1: case RAIE_N6_P1: case RAIE_N7_P1:
  case RAIE_O1_P1: case RAIE_O2_P1: case RAIE_O3_P1: case RAIE_O4_P1: case RAIE_O5_P1: case RAIE_O6_P1: case RAIE_O7_P1:
    return CESIUM;

  case RAIE_K_P2:
  case RAIE_L1_P2: case RAIE_L2_P2: case RAIE_L3_P2:
  case RAIE_M1_P2: case RAIE_M2_P2: case RAIE_M3_P2: case RAIE_M4_P2: case RAIE_M5_P2:
  case RAIE_N1_P2: case RAIE_N2_P2: case RAIE_N3_P2: case RAIE_N4_P2: case RAIE_N5_P2: case RAIE_N6_P2: case RAIE_N7_P2:
  case RAIE_O1_P2: case RAIE_O2_P2: case RAIE_O3_P2: case RAIE_O4_P2: case RAIE_O5_P2: case RAIE_O6_P2: case RAIE_O7_P2:
    return THALLIUM;

  case RAIE_K_P3:
  case RAIE_L1_P3: case RAIE_L2_P3: case RAIE_L3_P3:
  case RAIE_M1_P3: case RAIE_M2_P3: case RAIE_M3_P3: case RAIE_M4_P3: case RAIE_M5_P3:
  case RAIE_N1_P3: case RAIE_N2_P3: case RAIE_N3_P3: case RAIE_N4_P3: case RAIE_N5_P3: case RAIE_N6_P3: case RAIE_N7_P3:
  case RAIE_O1_P3: case RAIE_O2_P3: case RAIE_O3_P3: case RAIE_O4_P3: case RAIE_O5_P3: case RAIE_O6_P3: case RAIE_O7_P3:
    return BISMUTH;

#ifdef AVEC_TRANSURANIENS_
  case RAIE_K_P4:
  case RAIE_L1_P4: case RAIE_L2_P4: case RAIE_L3_P4:
  case RAIE_M1_P4: case RAIE_M2_P4: case RAIE_M3_P4: case RAIE_M4_P4: case RAIE_M5_P4:
  case RAIE_N1_P4: case RAIE_N2_P4: case RAIE_N3_P4: case RAIE_N4_P4: case RAIE_N5_P4: case RAIE_N6_P4: case RAIE_N7_P4:
  case RAIE_O1_P4: case RAIE_O2_P4: case RAIE_O3_P4: case RAIE_O4_P4: case RAIE_O5_P4: case RAIE_O6_P4: case RAIE_O7_P4:
    return LAWRENCIUM;
#endif

  case RAIE_K_Q1:
  case RAIE_L1_Q1: case RAIE_L2_Q1: case RAIE_L3_Q1:
  case RAIE_M1_Q1: case RAIE_M2_Q1: case RAIE_M3_Q1: case RAIE_M4_Q1: case RAIE_M5_Q1:
  case RAIE_N1_Q1: case RAIE_N2_Q1: case RAIE_N3_Q1: case RAIE_N4_Q1: case RAIE_N5_Q1: case RAIE_N6_Q1: case RAIE_N7_Q1:
  case RAIE_O1_Q1: case RAIE_O2_Q1: case RAIE_O3_Q1: case RAIE_O4_Q1: case RAIE_O5_Q1: case RAIE_O6_Q1: case RAIE_O7_Q1:
  case RAIE_P1_Q1: case RAIE_P2_Q1: case RAIE_P3_Q1:
#ifdef AVEC_TRANSURANIENS_
  case RAIE_P4_Q1:
#endif
    return FRANCIUM;

  default:
    (void) fprintf(stderr, "FLUO_ZDepart : raie inconnue [%d]\n", raie);

    return -1;
  }
}

/* ~-~-~-~-~-~-~-~-~ Simuler un spectre de fluorescence ~-~-~-~-~-~-~-~-~ */

/* -------- Toutes les raies en une fois -------- */

int FLUO_SimulerSpectre(Graphe *graphe, double energie, int Z, Graphe *mu_seuil, double f_massique)
{
  double e_seuil, mu_Z;
  unsigned int j;
  int retour;

  if ( NULL == graphe ) return -1;
  if ( energie <= graphe->x_min ) {
    (void) fprintf(stderr, "%s,%s l.%u : nergie d'excitation [%g] plus basse que la rgion simule\n",
		   __FILE__, __FUNCTION__, __LINE__, energie);
    return +1; /* Erreur mineure */
  }

  /* Simulation de chacun des seuils possibles */
  for ( j = SEUIL_K ; j < NOMBRE_SEUILS ; j++ ) {
    /* Position du seuil */
    e_seuil = (double) TH_DonneesSeuil( Z, j, NULL, NULL );

    /* Si le seuil est avant la rgion simule, les raies de fluorescence associes aussi */
    if ( e_seuil < graphe->x_min ) {
      /* Comme les seuils sont classes par nergie dcroissante, ce sera pareil pour tous les suivants */
      return 0;
    }

    /* A-t-on un mu/rho particulier pour ce seuil ? */
    if ( mu_seuil != NULL ) { /* Le spectre indiqu est-il pour ce seuil ? */
      if ( ( e_seuil >= mu_seuil->x_min ) && ( e_seuil <= mu_seuil->x_max ) ) { /* Oui -> simulation avec ce mu */
	mu_Z = FLUO_CalculerMu( mu_seuil, energie, Z, j );
/* 	printf("%s : mu_Z = %g pour E = %g\n", __FUNCTION__, mu_Z, energie ); */
      } else { /* Autre seuil donn => on utilise les tables par dfaut */
	mu_Z = FLUO_CalculerMu( NULL, energie, Z, j );
      }
    } else {
      mu_Z = FLUO_CalculerMu( NULL, energie, Z, j );
    }
#ifdef _DEVELOPPE_
    if ( ( VRAI == isinf( mu_Z ) ) || ( VRAI == isnan( mu_Z ) ) || ( mu_Z < 0.0 ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Section efficace absurde [%g] : E=%g, Z=%d (%s), seuil = %d (%s)\n",
		     __FILE__, __FUNCTION__, __LINE__, mu_Z, energie, Z, ATM_Z_vers_symbole( Z ), j, TH_TexteSeuil( j ));
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      return -20;
    }
#endif

    /* Simulation du groupe de raies pour ce seuil */
    retour = Table_Simulations[ j ]( graphe, Z, mu_Z * f_massique );
    if ( retour < 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la simulation des raies %s [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( j ), retour );
      return -2;
    }
  }

  /* Tout s'est bien pass, calcul fini */
  return 0;
}


/* -------- Calcul de mu/rho pour le seul indiqu de l'lment donn, pour l'nergie prcise -------- */

static inline double FLUO_CalculerMu( Graphe *mu_seuil, double energie, int Z, int seuil )
{
  double mu;

  if ( NULL == mu_seuil ) {
    /* Pas de graphe mu/rho prcis => modle thorique */
    return TH_CalculerMu( Z, energie, MUM_SELON_MCMASTER, seuil );
  }

  if ( energie < mu_seuil->x_min ) {
    /* Avant : on peut supposer que l'on n'absorbe rien [autres seuils grs aprs] */
    mu = 0.0;
  } else if ( energie > mu_seuil->x_max ) {
    double mu_fin, mu_calc, delta;

    /* Aprs : on peut supposer que le modle normal suffit [plus d'EXAFS] */
    mu = TH_CalculerMu( Z, energie, MUM_SELON_MCMASTER, seuil );

    /* Correction pour viter le dcrochement... */
    mu_fin = mu_seuil->y[ mu_seuil->nombre_point - 1 ]; /* Suppose le graphe tri par X croissant, ce qui est normal... */
    mu_calc = TH_CalculerMu( Z, mu_seuil->x_max, MUM_SELON_MCMASTER, seuil );
    delta = mu_fin - mu_calc;

    mu += delta;
  } else {
    mu = GRF_LireValeur( mu_seuil, energie, INTERPOLE_LINEAIRE );
  }

  return mu;
}

/* -------- Simuler toutes les raies d'un seuil -------- */

static inline int FLUO_SimulerRaies(Graphe *graphe, int Z, int seuil, int premiere_raie, int derniere_raie, int premiere_lacune, double n_lacunes)
{
  int raie, retour;
  double e_raie, lg_raie, i_raie, e_seuil;

  /* Si on n'absorbe rien [ou trop peu compte-tenu de la prcision de la machine], rien  faire */
  if ( 0.0 == n_lacunes ) return +1;

  /* Si le seuil est avant l'nergie minimale, toutes les raies de fluorescence aussi
     => on ne les verra de toute faon pas dans le spectre simul */
  e_seuil = TH_DonneesSeuil( Z, seuil, NULL, NULL );
  if ( e_seuil < graphe->x_min ) return +2;

  /* On simule toutes les raies */
  for ( raie = premiere_raie ; raie <= derniere_raie ; raie++ ) {
    int couche_lacune;

    couche_lacune = ( raie - premiere_raie ) + premiere_lacune;

    e_raie = FLUO_QuelleEnergie( Z, raie, &i_raie, &lg_raie );
    if ( e_raie < graphe->x_min ) {
      /* -1.0 : raie inexistante ; on continue tout de mme car une suivante peut tre autorise [3d remplit aprs 4s p. ex.]
	 -2.0 : nergie de la transition inconnue */
      if ( -2.00 == e_raie ) {
	(void) fprintf(stderr, "%s : Raie d'nergie inconnue [%s, %s -> %s]\n",
		       __FUNCTION__, ATM_Z_vers_symbole( Z ), TH_TexteSeuil( couche_lacune ), TH_TexteSeuil( seuil ) );
      }

      /* La raie est en-dessous, donc toutes les raies secondaires aussi -> inutile de grer cette raie */
      continue;
    }
#ifdef _DEVELOPPE_
    if ( ( i_raie < 0.0 ) || ( i_raie > 1.0 ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Intensit de raie absurde [%g, raies : %d->%d, lacune en %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, i_raie, premiere_raie, derniere_raie, premiere_lacune);
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      continue;
    }
#endif

    i_raie *= n_lacunes;

    /* La raie est bien dans la rgion qui nous intresse */
    if ( e_raie <= graphe->x_max ) {
      if ( lg_raie <= 0.0 ) {
	(void) fprintf(stderr, "%s : Largeur de raie inconnue [%s, %s -> %s] ==> 1 eV arbitrairement\n",
		       __FUNCTION__, ATM_Z_vers_symbole( Z ), TH_TexteSeuil( couche_lacune ), TH_TexteSeuil( seuil ) );
	lg_raie = 1.0; /* Arbitrairement */
      }

      /* On simule la raie */
      retour = FLUO_SimulerRaie( graphe, e_raie, i_raie, lg_raie, FLUO_MODELE_LORENTZ );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Erreur en simulant la raie %d [E = %g, I=%g, d=%g]\n",
		       __FILE__, __FUNCTION__, __LINE__, raie, e_raie, i_raie, lg_raie);
	return -5;
      }
    } else {
#ifdef _DEVELOPPE_
       (void) printf( "%s : Raie %s -> %s ignore : %s, E = %g eV\n", __FUNCTION__,
 		     TH_TexteSeuil( seuil ), TH_TexteSeuil( couche_lacune ), ATM_Z_vers_symbole( Z ), e_raie );
#endif
    }

    /* On a alors cr des lacunes dans la couche de dpart => on simule la fluorescence secondaire de ces raies
       On doit le faire mme si la raie tait aprs, car les fluorescences secondaires peuvent tomber dans la zone ! */
    (void) Table_Simulations[ couche_lacune ]( graphe, Z, i_raie );
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies K -------- */

static int FLUO_SimulerRaies_K( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;

  /* On simule les raies K */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_K, RAIE_K_L1, RAIE_K_Q1, SEUIL_LI, f_absorbee );

  /* Calcul fini, tout s'est bien pass */
  return retour;
}

/* -------- Ensemble du spectre des raies L1 -------- */

static int FLUO_SimulerRaies_L1( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies L1  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_LI, RAIE_L1_L2, RAIE_L1_Q1, SEUIL_LII, f_absorbee );
  if ( retour != 0 ) return retour;

  /* Transitions Coster-Kronig L1 -> L2 */
  CK = FLUO_CosterKronig( Z, SEUIL_LI, SEUIL_LII );
  FLUO_SimulerRaies_L2( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig L1 -> L3 */
  CK = FLUO_CosterKronig( Z, SEUIL_LI, SEUIL_LIII );
  FLUO_SimulerRaies_L3( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies L2 -------- */

static int FLUO_SimulerRaies_L2( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies L2  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_LII, RAIE_L2_L3, RAIE_L2_Q1, SEUIL_LIII, f_absorbee );
  if ( retour != 0 ) return retour;

  /* Transitions Coster-Kronig L2 -> L3 */
  CK = FLUO_CosterKronig( Z, SEUIL_LII, SEUIL_LIII );
  FLUO_SimulerRaies_L3( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies L3 -------- */

static int FLUO_SimulerRaies_L3( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;

  /* On simule les raies L3 */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_LIII, RAIE_L3_M1, RAIE_L3_Q1, SEUIL_MI, f_absorbee );

  return retour;
}


/* -------- Ensemble du spectre des raies M1 -------- */

static int FLUO_SimulerRaies_M1( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies M1  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_MI, RAIE_M1_N1, RAIE_M1_Q1, SEUIL_NI, f_absorbee );
  if ( retour != 0 ) return retour;

  /* Transitions Coster-Kronig M1 -> M2 */
  CK = FLUO_CosterKronig( Z, SEUIL_MI, SEUIL_MII );
  FLUO_SimulerRaies_M2( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig M1 -> M3 */
  CK = FLUO_CosterKronig( Z, SEUIL_MI, SEUIL_MIII );
  FLUO_SimulerRaies_M3( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig M1 -> M4 */
  CK = FLUO_CosterKronig( Z, SEUIL_MI, SEUIL_MIV );
  FLUO_SimulerRaies_M4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig M1 -> M5 */
  CK = FLUO_CosterKronig( Z, SEUIL_MI, SEUIL_MV );
  FLUO_SimulerRaies_M5( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies M2 -------- */

static int FLUO_SimulerRaies_M2( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies M2  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_MII, RAIE_M2_N1, RAIE_M2_Q1, SEUIL_NI, f_absorbee );
  if ( retour != 0 ) return retour;

  /* Transitions Coster-Kronig M2 -> M3 */
  CK = FLUO_CosterKronig( Z, SEUIL_MII, SEUIL_MIII );
  FLUO_SimulerRaies_M3( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig M2 -> M4 */
  CK = FLUO_CosterKronig( Z, SEUIL_MII, SEUIL_MIV );
  FLUO_SimulerRaies_M4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig M2 -> M5 */
  CK = FLUO_CosterKronig( Z, SEUIL_MII, SEUIL_MV );
  FLUO_SimulerRaies_M5( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies M3 -------- */

static int FLUO_SimulerRaies_M3( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies M3  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_MIII, RAIE_M3_N1, RAIE_M3_Q1, SEUIL_NI, f_absorbee );
  if ( retour != 0 ) return retour;

  /* Transitions Coster-Kronig M3 -> M4 */
  CK = FLUO_CosterKronig( Z, SEUIL_MIII, SEUIL_MIV );
  FLUO_SimulerRaies_M4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig M3 -> M5 */
  CK = FLUO_CosterKronig( Z, SEUIL_MIII, SEUIL_MV );
  FLUO_SimulerRaies_M5( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies M4 -------- */

static int FLUO_SimulerRaies_M4( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies M4  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_MIV, RAIE_M4_N1, RAIE_M4_Q1, SEUIL_NI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig M4 -> M5 */
  CK = FLUO_CosterKronig( Z, SEUIL_MIV, SEUIL_MV );
  FLUO_SimulerRaies_M5( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies M5 -------- */

static int FLUO_SimulerRaies_M5( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;

  /* On simule les raies M5  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_MV, RAIE_M5_N1, RAIE_M5_Q1, SEUIL_NI, f_absorbee );

  return retour;
}

/* -------- Ensemble du spectre des raies N1 -------- */

static int FLUO_SimulerRaies_N1( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies N1  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_NI, RAIE_N1_O1, RAIE_N1_Q1, SEUIL_OI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig N1 -> N2 */
  CK = FLUO_CosterKronig( Z, SEUIL_NI, SEUIL_NII );
  FLUO_SimulerRaies_N2( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N1 -> N3 */
  CK = FLUO_CosterKronig( Z, SEUIL_NI, SEUIL_NIII );
  FLUO_SimulerRaies_N3( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N1 -> N4 */
  CK = FLUO_CosterKronig( Z, SEUIL_NI, SEUIL_NIV );
  FLUO_SimulerRaies_N4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N1 -> N5 */
  CK = FLUO_CosterKronig( Z, SEUIL_NI, SEUIL_NV );
  FLUO_SimulerRaies_N5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N1 -> N6 */
  CK = FLUO_CosterKronig( Z, SEUIL_NI, SEUIL_NVI );
  FLUO_SimulerRaies_N6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N1 -> N7 */
  CK = FLUO_CosterKronig( Z, SEUIL_NI, SEUIL_NVII );
  FLUO_SimulerRaies_N7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies N2 -------- */

static int FLUO_SimulerRaies_N2( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies N2  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_NII, RAIE_N2_O1, RAIE_N2_Q1, SEUIL_OI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig N2 -> N3 */
  CK = FLUO_CosterKronig( Z, SEUIL_NII, SEUIL_NIII );
  FLUO_SimulerRaies_N3( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N2 -> N4 */
  CK = FLUO_CosterKronig( Z, SEUIL_NII, SEUIL_NIV );
  FLUO_SimulerRaies_N4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N2 -> N5 */
  CK = FLUO_CosterKronig( Z, SEUIL_NII, SEUIL_NV );
  FLUO_SimulerRaies_N5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N2 -> N6 */
  CK = FLUO_CosterKronig( Z, SEUIL_NII, SEUIL_NVI );
  FLUO_SimulerRaies_N6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N2 -> N7 */
  CK = FLUO_CosterKronig( Z, SEUIL_NII, SEUIL_NVII );
  FLUO_SimulerRaies_N7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies N3 -------- */

static int FLUO_SimulerRaies_N3( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies N3  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_NIII, RAIE_N3_O1, RAIE_N3_Q1, SEUIL_OI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig N3 -> N4 */
  CK = FLUO_CosterKronig( Z, SEUIL_NIII, SEUIL_NIV );
  FLUO_SimulerRaies_N4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N3 -> N5 */
  CK = FLUO_CosterKronig( Z, SEUIL_NIII, SEUIL_NV );
  FLUO_SimulerRaies_N5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N3 -> N6 */
  CK = FLUO_CosterKronig( Z, SEUIL_NIII, SEUIL_NVI );
  FLUO_SimulerRaies_N6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N3 -> N7 */
  CK = FLUO_CosterKronig( Z, SEUIL_NIII, SEUIL_NVII );
  FLUO_SimulerRaies_N7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies N4 -------- */

static int FLUO_SimulerRaies_N4( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies N4 directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_NIV, RAIE_N4_O1, RAIE_N4_Q1, SEUIL_OI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig N4 -> N5 */
  CK = FLUO_CosterKronig( Z, SEUIL_NIV, SEUIL_NV );
  FLUO_SimulerRaies_N5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N4 -> N6 */
  CK = FLUO_CosterKronig( Z, SEUIL_NIV, SEUIL_NVI );
  FLUO_SimulerRaies_N6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N4 -> N7 */
  CK = FLUO_CosterKronig( Z, SEUIL_NIV, SEUIL_NVII );
  FLUO_SimulerRaies_N7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies N5 -------- */

static int FLUO_SimulerRaies_N5( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies N5  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_NV, RAIE_N5_O1, RAIE_N5_Q1, SEUIL_OI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig N5 -> N6 */
  CK = FLUO_CosterKronig( Z, SEUIL_NV, SEUIL_NVI );
  FLUO_SimulerRaies_N6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig N5 -> N7 */
  CK = FLUO_CosterKronig( Z, SEUIL_NV, SEUIL_NVII );
  FLUO_SimulerRaies_N7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies N6 -------- */

static int FLUO_SimulerRaies_N6( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies N5  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_NVI, RAIE_N6_O1, RAIE_N6_Q1, SEUIL_OI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig N5 -> N7 */
  CK = FLUO_CosterKronig( Z, SEUIL_NVI, SEUIL_NVII );
  FLUO_SimulerRaies_N7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return retour;
}

/* -------- Ensemble du spectre des raies N7 -------- */

static int FLUO_SimulerRaies_N7( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;

  /* On simule les raies N5  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_NV, RAIE_N6_O1, RAIE_N6_Q1, SEUIL_OI, f_absorbee );

  return retour;
}

/* -------- Ensemble du spectre des raies O1 -------- */

static int FLUO_SimulerRaies_O1( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies O1  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_OI, RAIE_O1_P1, RAIE_O1_Q1, SEUIL_PI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig O1 -> O2 */
  CK = FLUO_CosterKronig( Z, SEUIL_OI, SEUIL_OII );
  FLUO_SimulerRaies_O2( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O1 -> O3 */
  CK = FLUO_CosterKronig( Z, SEUIL_OI, SEUIL_OIII );
  FLUO_SimulerRaies_O3( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O1 -> O4 */
  CK = FLUO_CosterKronig( Z, SEUIL_OI, SEUIL_OIV );
  FLUO_SimulerRaies_O4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O1 -> O5 */
  CK = FLUO_CosterKronig( Z, SEUIL_OI, SEUIL_OV );
  FLUO_SimulerRaies_O5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O1 -> O6 */
  CK = FLUO_CosterKronig( Z, SEUIL_OI, SEUIL_OVI );
  FLUO_SimulerRaies_O6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O1 -> O7 */
  CK = FLUO_CosterKronig( Z, SEUIL_OI, SEUIL_OVII );
  FLUO_SimulerRaies_O7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies O2 -------- */

static int FLUO_SimulerRaies_O2( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies O2  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_OII, RAIE_O2_P1, RAIE_O2_Q1, SEUIL_PI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig O2 -> O3 */
  CK = FLUO_CosterKronig( Z, SEUIL_OII, SEUIL_OIII );
  FLUO_SimulerRaies_O3( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O2 -> O4 */
  CK = FLUO_CosterKronig( Z, SEUIL_OII, SEUIL_OIV );
  FLUO_SimulerRaies_O4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O2 -> O5 */
  CK = FLUO_CosterKronig( Z, SEUIL_OII, SEUIL_OV );
  FLUO_SimulerRaies_O5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O2 -> O6 */
  CK = FLUO_CosterKronig( Z, SEUIL_OII, SEUIL_OVI );
  FLUO_SimulerRaies_O6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O2 -> O7 */
  CK = FLUO_CosterKronig( Z, SEUIL_OII, SEUIL_OVII );
  FLUO_SimulerRaies_O7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies O3 -------- */

static int FLUO_SimulerRaies_O3( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies O3  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_OIII, RAIE_O3_P1, RAIE_O3_Q1, SEUIL_PI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig O3 -> O4 */
  CK = FLUO_CosterKronig( Z, SEUIL_OIII, SEUIL_OIV );
  FLUO_SimulerRaies_O4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O3 -> O5 */
  CK = FLUO_CosterKronig( Z, SEUIL_OIII, SEUIL_OV );
  FLUO_SimulerRaies_O5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O3 -> O6 */
  CK = FLUO_CosterKronig( Z, SEUIL_OIII, SEUIL_OVI );
  FLUO_SimulerRaies_O6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O3 -> O7 */
  CK = FLUO_CosterKronig( Z, SEUIL_OIII, SEUIL_OVII );
  FLUO_SimulerRaies_O7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies O4 -------- */

static int FLUO_SimulerRaies_O4( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies O4 directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_OIV, RAIE_O4_P1, RAIE_O4_Q1, SEUIL_PI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig O4 -> O5 */
  CK = FLUO_CosterKronig( Z, SEUIL_OIV, SEUIL_OV );
  FLUO_SimulerRaies_O5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O4 -> O6 */
  CK = FLUO_CosterKronig( Z, SEUIL_OIV, SEUIL_OVI );
  FLUO_SimulerRaies_O6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O4 -> O7 */
  CK = FLUO_CosterKronig( Z, SEUIL_OIV, SEUIL_OVII );
  FLUO_SimulerRaies_O7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies O5 -------- */

static int FLUO_SimulerRaies_O5( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies O5  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_OV, RAIE_O5_P1, RAIE_O5_Q1, SEUIL_PI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig O5 -> O6 */
  CK = FLUO_CosterKronig( Z, SEUIL_OV, SEUIL_OVI );
  FLUO_SimulerRaies_O6( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig O5 -> O7 */
  CK = FLUO_CosterKronig( Z, SEUIL_OV, SEUIL_OVII );
  FLUO_SimulerRaies_O7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies O6 -------- */

static int FLUO_SimulerRaies_O6( Graphe *graphe, int Z, double f_absorbee )
{
  double CK;
  int retour;

  /* On simule les raies O5  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_OVI, RAIE_O6_P1, RAIE_O6_Q1, SEUIL_PI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig O5 -> O7 */
  CK = FLUO_CosterKronig( Z, SEUIL_OVI, SEUIL_OVII );
  FLUO_SimulerRaies_O7( graphe, Z, f_absorbee * CK );

  /* Calcul fini, tout s'est bien pass */
  return retour;
}

/* -------- Ensemble du spectre des raies O7 -------- */

static int FLUO_SimulerRaies_O7( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;

  /* On simule les raies O5  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_OV, RAIE_O6_P1, RAIE_O6_Q1, SEUIL_PI, f_absorbee );

  return retour;
}

/* -------- Ensemble du spectre des raies P1 -------- */

static int FLUO_SimulerRaies_P1( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies P1  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_PI, RAIE_P1_Q1, RAIE_P1_Q1, SEUIL_QI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig P1 -> P2 */
  CK = FLUO_CosterKronig( Z, SEUIL_PI, SEUIL_PII );
  FLUO_SimulerRaies_P2( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig P1 -> P3 */
  CK = FLUO_CosterKronig( Z, SEUIL_PI, SEUIL_PIII );
  FLUO_SimulerRaies_P3( graphe, Z, f_absorbee * CK );

#ifdef AVEC_TRANSURANIENS_
  /* Transitions Coster-Kronig P1 -> P4 */
  CK = FLUO_CosterKronig( Z, SEUIL_PI, SEUIL_PIV );
  FLUO_SimulerRaies_P4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig P1 -> P5 */
  CK = FLUO_CosterKronig( Z, SEUIL_PI, SEUIL_PV );
  FLUO_SimulerRaies_P5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig P1 -> P6, P7 ngliges car aucun atome connu n'a ces couches remplies */
#endif

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies P2 -------- */

static int FLUO_SimulerRaies_P2( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;
  double CK; /* Rendement Coster-Kronig */

  /* On simule les raies P2  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_PII, RAIE_P2_Q1, RAIE_P2_Q1, SEUIL_QI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig P2 -> P3 */
  CK = FLUO_CosterKronig( Z, SEUIL_PII, SEUIL_PIII );
  FLUO_SimulerRaies_P3( graphe, Z, f_absorbee * CK );

#ifdef AVEC_TRANSURANIENS_
  /* Transitions Coster-Kronig P2 -> P4 */
  CK = FLUO_CosterKronig( Z, SEUIL_PII, SEUIL_PIV );
  FLUO_SimulerRaies_P4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig P2 -> P5 */
  CK = FLUO_CosterKronig( Z, SEUIL_PII, SEUIL_PV );
  FLUO_SimulerRaies_P5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig P2 -> P6, P7 ngliges car aucun atome connu n'a ces couches remplies */
#endif

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies P3 -------- */

static int FLUO_SimulerRaies_P3( Graphe *graphe, int Z, double f_absorbee )
{
#ifdef AVEC_TRANSURANIENS_
  double CK;
#endif
  int retour;

  /* On simule les raies P3  directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_PIII, RAIE_P3_Q1, RAIE_P3_Q1, SEUIL_QI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

#ifdef AVEC_TRANSURANIENS_
  /* Transitions Coster-Kronig P3 -> P4 */
  CK = FLUO_CosterKronig( Z, SEUIL_PIII, SEUIL_PIV );
  FLUO_SimulerRaies_P4( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig P3 -> P5 */
  CK = FLUO_CosterKronig( Z, SEUIL_PIII, SEUIL_PV );
  FLUO_SimulerRaies_P5( graphe, Z, f_absorbee * CK );
 
  /* Transitions Coster-Kronig P3 -> P6, P7 ngliges car aucun atome connu n'a ces couches remplies */
#endif

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

#ifdef AVEC_TRANSURANIENS_

/* -------- Ensemble du spectre des raies P4 -------- */

static int FLUO_SimulerRaies_P4( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;

  /* On simule les raies P4 directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_PIV, RAIE_P4_Q1, RAIE_P4_Q1, SEUIL_QI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig P3 -> P5 */
  CK = FLUO_CosterKronig( Z, SEUIL_PIV, SEUIL_PV );
  FLUO_SimulerRaies_P5( graphe, Z, f_absorbee * CK );

  /* Transitions Coster-Kronig P4 -> P6, P7 ngliges car aucun atome connu n'a ces couches remplies */

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

/* -------- Ensemble du spectre des raies P5 -------- */

static int FLUO_SimulerRaies_P5( Graphe *graphe, int Z, double f_absorbee )
{
  int retour;

  /* On simule les raies P4 directes  */
  retour = FLUO_SimulerRaies( graphe, Z, SEUIL_PV, RAIE_P5_Q1, RAIE_P5_Q1, SEUIL_QI, f_absorbee );
  if ( retour != 0 ) return retour; /* Simulation inutile => les CK aussi ! */

  /* Transitions Coster-Kronig P5 -> P6, P7 ngliges car aucun atome connu n'a ces couches remplies */

  /* Calcul fini, tout s'est bien pass */
  return 0;
}

#endif

static inline int FLUO_SimulerRaies_Q1( Graphe *graphe __attribute__((__unused__)), int Z __attribute__((__unused__)), double f_massique __attribute__((__unused__)) )
{
  /* Aucune transition possible pour les atomes connus */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Simuler une raie de fluorescence ~-~-~-~-~-~-~-~-~ */

int FLUO_SimulerRaie(Graphe *graphe, double energie, double intensite, double largeur, int methode)
{
  int retour = -2;

  if ( NULL == graphe ) return -1;

  switch ( methode )
  {
  case FLUO_MODELE_LORENTZ:
    retour = FLUO_Raie_Lorentz( graphe, energie, intensite, largeur );
    break;

  case FLUO_MODELE_GAUSS:
    retour = FLUO_Raie_Gauss( graphe, energie, intensite, largeur );
    break;

  default:
    (void) fprintf(stderr, "FLUO_SimulerRaie : Modle inconnu [%d]\n", methode);
    return -2;
  }

  /* Calcul termin */
  return retour;
}

/* -------- Modle de Lorentz --------
              (  (E - E0)) ; A  = intensite   : hauteur maximale
I(E) = A * exp(- -------- ) ; E0 = energie     : position du maximum
              (     s    ) ; 2 s racine(ln 2) : largeur  mi-hauteur
 */

static int FLUO_Raie_Gauss(Graphe *graphe, double energie, double intensite, double largeur)
{
  double s, I;
  double *x, *y;
  unsigned long i;

  /* Calcul des paramtres */
  s = largeur / (2.0 * sqrt( log( 2.0 ) ) );
  s = s * s;
  if ( 0.0 == s ) {
    (void) fprintf(stderr, "FLUO_Raie_Gauss : Raie de largeur nulle !\n");
    return -1;
  }

  /* Calcul de la fonction */
  for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    I = *x - energie;
    I = I * I;
    I = I / s;

    *y += intensite * exp( - I );
  }

  /* Calcul termin */
  return 0;
}

/* -------- Modle de Lorentz --------
            h       ; h/a = intensite : hauteur maximale
I(E) = ------------ ; E0   = energie   : position du maximum
       (E-E0) + a ; 2a   = largeur   : largeur  mi-hauteur
 */

static int FLUO_Raie_Lorentz(Graphe *graphe, double energie, double intensite, double largeur)
{
  double a, h, I;
  double *x, *y;
  unsigned long i;

  /* Calcul des paramtres */
  a = largeur / 2.0;
  a = a * a; /* N'intervient que par son carr */
  if ( 0.0 == a ) {
    (void) fprintf(stderr, "FLUO_Raie_Lorentz : Raie de largeur nulle !\n");
    return -1;
  }
  h = a * intensite;

  /* Calcul de la fonction */
  for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    I = *x - energie;
    I = I * I;
    I = h / (I + a);

    *y += I;
#ifdef _DEVELOPPE_
    if ( ( isinf( *y ) ) || ( isnan( *y ) ) ) {
      (void) fprintf( stderr, "%s, %s l.%u : y= +infini/NaN [I=%g, E=%g, i=%g, l=%g, x=%g]\n",
		      __FILE__, __FUNCTION__, __LINE__, I, energie, intensite, largeur, *x );
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      return -1;
    }
#endif
  }

  /* Calcul termin */
  return 0;
}
