/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modles structurels d'une station exprimentale
   Paramtrer l'chantillon [interface utilisateur]
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "detecteur.h"
#include "echantillon.h" /* Pour la composition ! */
#include "graphes.h"
#include "interface.h"
#include "science.h"		/* Pour M_PI */

#include "param_detecteur.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_FORME       13
#define LGMAX_NOM_COUPE       21
#define LGMAX_NOM_REPONSE     18
#define LGMAX_NOM_LINEARITE   30
#define LGMAX_NOM_COMPOSITION 14

struct detecteur *premier_detecteur = NULL;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void preparer_arbre(void);

static int GER_Parametrage(unsigned int objet, int x, int y);

static void InstallerAbsorbant(double longueur, struct matiere *composition, boolean activer);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

int DETECTEUR_Parametrer( Detecteur *detecteur )
{
  int retour;
  double r, theta, phi;

  if ( NULL == detecteur ) return -1;

  /* Prparation de l'arbre */
  if ( FAUX == arbres[ PARAMDETECTEUR ].drapeaux.prepare ) {
    preparer_arbre();
  }

  /* On installe les paramtres actuels */
  met_chaine( PARAMDETECTEUR, PD_NOM, 0, detecteur->nom, TXT_COPIER );

  OBJ_choisit( PARAMDETECTEUR, PD_INFINI, ( detecteur->proprietes & DETECTEUR_A_L_INFINI ) ? ( FORCE_OUI ) : ( FORCE_NON ) );

  REPERE_CartesienVersSpherique( &(detecteur->orientation.position), &r, &theta, &phi );
  OBJ_MettreNombre( PARAMDETECTEUR, PD_POS_R, r / 10000.0, 0.0 ); /* Interne en microns, affich en cm */
  OBJ_MettreAngle( PARAMDETECTEUR, PD_POSITION_THET, theta, 0.0 );
  OBJ_MettreAngle( PARAMDETECTEUR, PD_POSITION_PHI , phi, 0.0 );

  /* Mettre le point vers lequel pointe le dtecteur [dans la face d'entre]
     La direction est la droite passant par le centre du dtecteur = centre du repre local
     et normale au plan du dtecteur => axe des Y du repre local */
  {
    Vecteur3D u_droite, p_vise;

    /* Vecteur directeur de la droite */
    REPERE_VecteurNormal( &(detecteur->orientation), &u_droite );
/*     REPERE_CartesienVersSpherique( &u_droite, &r, &theta, &phi ); */
/*     printf("%s / mise : vecteur normal = (%g, %g, %g) [r=%g theta=%g{%g}, phi=%g{%g}], origine = (%g, %g, %g)\n", __FUNCTION__, */
/* 	   u_droite.x, u_droite.y, u_droite.z, r, detecteur->orientation.theta, theta, detecteur->orientation.phi, phi, */
/* 	   detecteur->orientation.position.x, detecteur->orientation.position.y, detecteur->orientation.position.z); */

    /* Point d'intersection */
    U3D_Inter_Droite_Plan_YCt( &(detecteur->orientation.position), &u_droite, 0.0, &p_vise );

    /* On met les valeurs, donnes en microns mais affiches en mm */
    OBJ_MettreNombre( PARAMDETECTEUR, RB_X_POINTE, p_vise.x / 1000.0, 0.0 );
    OBJ_MettreNombre( PARAMDETECTEUR, PD_Y_POINTE, p_vise.y / 1000.0, 0.0 );
    OBJ_MettreNombre( PARAMDETECTEUR, PD_Z_POINTE, p_vise.z / 1000.0, 0.0 );
  }

  OBJ_MettreNombre( PARAMDETECTEUR, PD_ANGLE_SOLIDE, detecteur->angle_solide, 0.0 );

  OBJ_choisit( PARAMDETECTEUR, PD_RESOLU, (detecteur->proprietes & DETECTEUR_RESOLU) ? (FORCE_OUI) : (FORCE_NON) );
  OBJ_choisit( PARAMDETECTEUR, PD_REP_NRJ, (detecteur->proprietes & DETECTEUR_NRJ_VARIABLE) ? (FORCE_OUI) : (FORCE_NON) );
  OBJ_choisit( PARAMDETECTEUR, PD_NON_LINEAIRE, (detecteur->proprietes & DETECTEUR_NON_LINEAIRE) ? (FORCE_OUI) : (FORCE_NON) );
  OBJ_choisit( PARAMDETECTEUR, PD_ENTIER, (detecteur->proprietes & DETECTEUR_ENTIER) ? (FORCE_OUI) : (FORCE_NON) );

  OBJ_choisit( PARAMDETECTEUR, PD_A_VIDE, (detecteur->proprietes & DETECTEUR_A_VIDE) ? (FORCE_OUI) : (FORCE_NON));
  OBJ_active ( PARAMDETECTEUR, PD_COUPS_A_VIDE, (detecteur->proprietes & DETECTEUR_A_VIDE) ? (FORCE_OUI) : (FORCE_NON));
  OBJ_MettreNombre( PARAMDETECTEUR, PD_COUPS_A_VIDE, detecteur->coups_a_vide, 0.0 );

  OBJ_choisit( PARAMDETECTEUR, PD_SURFACE, (detecteur->proprietes & DETECTEUR_SURFACE) ? (FORCE_OUI) : (FORCE_NON) );
  OBJ_choisit( PARAMDETECTEUR, PD_BRUITER, (detecteur->proprietes & DETECTEUR_BRUITE) ? (FORCE_OUI) : (FORCE_NON) );

  InstallerAbsorbant( detecteur->longueur, detecteur->composition, (detecteur->proprietes & DETECTEUR_ABSORBANT) ? (FORCE_OUI) : (FORCE_NON) );

  /* Appel de l'arbre */
  ARB_MettreComplement( PARAMDETECTEUR, detecteur );
  retour = ARB_Bloquant( PARAMDETECTEUR, titre_arbre_parametrage, principale, 0 );
  ARB_MettreComplement( PARAMDETECTEUR, NULL );

  if ( PD_ANNULER == retour ) return +1;

  /* On enregistre les paramtres */
  DET_NommerDetecteur( detecteur, lit_chaine( PARAMDETECTEUR, PD_NOM, 0, NULL, TXT_COPIER ) );

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_INFINI, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_A_L_INFINI;
  } else {
    detecteur->proprietes &= ~DETECTEUR_A_L_INFINI;
  }

  /* Position du dtecteur */
  r = OBJ_LireNombre( PARAMDETECTEUR, PD_POS_R, NULL ) * 10000.0; /* Affich en cm, interne en microns */
  theta = OBJ_LireAngle( PARAMDETECTEUR, PD_POSITION_THET, NULL );
  phi =  OBJ_LireAngle( PARAMDETECTEUR, PD_POSITION_PHI, NULL );
  REPERE_SpheriqueVersCartesien( r, theta, phi, &(detecteur->orientation.position) );

  /* Normale du plan local : dtermine  partir du point vis... */
  {
    Vecteur3D p_vise;

    p_vise.x = OBJ_LireNombre( PARAMDETECTEUR, RB_X_POINTE, NULL ) * 1000.0; /* Affich en mm, cod en microns */
    p_vise.y = OBJ_LireNombre( PARAMDETECTEUR, PD_Y_POINTE, NULL ) * 1000.0;
    p_vise.z = OBJ_LireNombre( PARAMDETECTEUR, PD_Z_POINTE, NULL ) * 1000.0;
    printf( "Vise (%g, %g, %g)", p_vise.x, p_vise.y, p_vise.z );

    p_vise.x -= detecteur->orientation.position.x; /* Vecteur directeur de la droite */
    p_vise.y -= detecteur->orientation.position.y;
    p_vise.z -= detecteur->orientation.position.z;
    U3D_Normalise( &p_vise, NULL ); /* Normalis */

    /* Le vecteur pointant, normalis, a les coordonnes indiques dans le repre absolu.
      En dduire les deux rotations  appliquer pour que ce vecteur soit le vecteur Y dans le repre d'arriv */
    REPERE_TrouverRotations_Y( &p_vise, &theta, &phi );
    if ( fabs( theta ) < 1.0e-8 ) theta = 0.0; /* Limite les erreurs d'arrondi sur les angles classiques */
    if ( fabs( phi ) < 1.0e-8 ) phi = 0.0;
    printf( ", normale = (%g, %g, %g) => t = %g, f=%g\n", p_vise.x, p_vise.y, p_vise.z, theta * 180.0/M_PI, phi * 180.0/M_PI );

    detecteur->orientation.theta = theta;
    detecteur->orientation.phi   = phi;

/*     printf("%s / lecture : vecteur normal = (%g, %g, %g) [r=%g theta=%g, phi=%g], origine = (%g, %g, %g)\n", __FUNCTION__, */
/* 	   p_vise.x, p_vise.y, p_vise.z, r, detecteur->orientation.theta, detecteur->orientation.phi, */
/* 	   detecteur->orientation.position.x, detecteur->orientation.position.y, detecteur->orientation.position.z); */
  }

  detecteur->angle_solide = OBJ_LireNombre( PARAMDETECTEUR, PD_ANGLE_SOLIDE, NULL );

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_RESOLU, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_RESOLU;
  } else {
    detecteur->proprietes &= ~DETECTEUR_RESOLU;
  }

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_REP_NRJ, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_NRJ_VARIABLE;
  } else {
    detecteur->proprietes &= ~DETECTEUR_NRJ_VARIABLE;
  }

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_NON_LINEAIRE, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_NON_LINEAIRE;
  } else {
    detecteur->proprietes &= ~DETECTEUR_NON_LINEAIRE;
  }

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_ENTIER, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_ENTIER;
  } else {
    detecteur->proprietes &= ~DETECTEUR_ENTIER;
  }

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_A_VIDE, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_A_VIDE;
    detecteur->coups_a_vide = OBJ_LireNombre( PARAMDETECTEUR, PD_COUPS_A_VIDE, NULL );
  } else {
    detecteur->proprietes &= ~DETECTEUR_A_VIDE;
  }

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_SURFACE, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_SURFACE;
  } else {
    detecteur->proprietes &= ~DETECTEUR_SURFACE;
  }

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_BRUITER, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_BRUITE;
  } else {
    detecteur->proprietes &= ~DETECTEUR_BRUITE;
  }

  if ( VRAI == OBJ_choisit( PARAMDETECTEUR, PD_ABSORBANT, QUESTION ) ) {
    detecteur->proprietes |= DETECTEUR_ABSORBANT;
    detecteur->longueur = OBJ_LireNombre( PARAMDETECTEUR, PD_LONGUEUR, NULL ); /* Contenu en cm, affichage en cm */
  } else {
    detecteur->proprietes &= ~DETECTEUR_ABSORBANT;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Gestion de l'arbre -------- */

static int GER_Parametrage(unsigned int objet, int x, int y)
{
  Detecteur *detecteur;
  int etat;

  detecteur = (Detecteur *) ARB_LireComplement( PARAMDETECTEUR );
  if ( NULL == detecteur ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de dtecteur !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_TRAITE;
  }

  switch ( objet )
  {
  case PD_ABSORBANT:
    etat = OBJ_choisit( PARAMDETECTEUR, objet, ECHANGE );
    if ( VRAI == etat ) { /* Choisir une composition, si ncessaire */
      if ( NULL == detecteur->composition ) {
	detecteur->composition = Matiere_ChoisirComposition( PARAMDETECTEUR, PD_COMPOSITION, LGMAX_NOM_COMPOSITION, NULL );
	if ( NULL == detecteur->composition ) {
	  InstallerAbsorbant( detecteur->longueur, detecteur->composition, FORCE_NON );
	  return BI_EVENEMENT_TRAITE;
	}
      }

      InstallerAbsorbant( detecteur->longueur, detecteur->composition, FORCE_OUI );
    } else {
      InstallerAbsorbant( detecteur->longueur, detecteur->composition, FORCE_NON );
    }
    return BI_EVENEMENT_TRAITE;

  case PD_COMPOSITION:
    detecteur->composition = Matiere_ChoisirComposition( PARAMDETECTEUR, PD_COMPOSITION, LGMAX_NOM_COMPOSITION, NULL );
    if ( NULL == detecteur->composition ) {
      OBJ_choisit( PARAMDETECTEUR, PD_ABSORBANT, FORCE_NON );
      OBJ_active( PARAMDETECTEUR, PD_COMPOSITION, FORCE_NON );
      OBJ_active( PARAMDETECTEUR, PD_LONGUEUR, FORCE_NON );
    }
    return BI_EVENEMENT_TRAITE;

  case PD_INFINI:
  case PD_RESOLU:
  case PD_ENTIER:
    OBJ_choisit( PARAMDETECTEUR, objet, ECHANGE );
    return BI_EVENEMENT_TRAITE;

  case PD_NON_LINEAIRE:
    etat = OBJ_choisit( PARAMDETECTEUR, PD_NON_LINEAIRE, ECHANGE );
    if ( VRAI == etat ) {
      if ( NULL == detecteur->r_signal ) {
	detecteur->r_signal = GRF_choisir( GR_REP_DETECTEUR, FAUX, VRAI, NULL );
	if ( NULL == detecteur->r_signal ) {
	  OBJ_choisit(  PARAMDETECTEUR, PD_NON_LINEAIRE, FORCE_NON );
	  return BI_EVENEMENT_TRAITE;
	}
      }
      OBJ_active( PARAMDETECTEUR, PD_GRF_NON_LIN, FORCE_OUI );
    } else {
      OBJ_active( PARAMDETECTEUR, PD_GRF_NON_LIN, FORCE_NON );
    }
    return BI_EVENEMENT_TRAITE;

  case PD_A_VIDE:
    etat = OBJ_choisit( PARAMDETECTEUR, objet, ECHANGE );
    OBJ_active( PARAMDETECTEUR, PD_COUPS_A_VIDE, etat );
    return BI_EVENEMENT_TRAITE;
  }

#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "GER_Parametrage : bouton non gr [objet=%u x=%d y=%d]\n", objet, x, y);
#endif
  return BI_EVENEMENT_NON_TRAITE;
}

static void InstallerAbsorbant(double longueur, struct matiere *composition, boolean activer)
{
  /* Si choisi, on active... */
  OBJ_choisit( PARAMDETECTEUR, PD_ABSORBANT, activer );
  OBJ_active( PARAMDETECTEUR, PD_LONGUEUR, activer );
  OBJ_active( PARAMDETECTEUR, PD_COMPOSITION, activer );

  /* On installe les paramtres */
  OBJ_MettreNombre( PARAMDETECTEUR, PD_LONGUEUR, longueur, 0.0 ); /* Contenu en cm, affichage en cm */
  if ( composition != NULL ) {
    met_chaine( PARAMDETECTEUR, PD_COMPOSITION, LGMAX_NOM_COMPOSITION, composition->nom, TXT_COPIER );
  } else {
    met_chaine( PARAMDETECTEUR, PD_COMPOSITION, LGMAX_NOM_COMPOSITION, texte_aucune_composition, TXT_COPIER );
  }
}

/* -------- Prparation de l'arbre -------- */

static void preparer_arbre(void)
{
  OBJ_deplace( PARAMDETECTEUR, PD_TXT_POS_THETA, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_POSITION_THET, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_TXT_POS_PHI  , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_POSITION_PHI , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_INFINI       , 0, +3, DEPLACE_PIXELS );

  OBJ_deplace( PARAMDETECTEUR, PD_TXT_Y_POINTE, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_Y_POINTE    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_TXT_Z_POINTE, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_Z_POINTE    , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_TXT_ANGLE   , 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_ANGLE_SOLIDE, 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_UNITE_ANGLE , 0, +3, DEPLACE_PIXELS );

  OBJ_deplace( PARAMDETECTEUR, PD_TXT_LARG , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_LARGEUR  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_TXT_HAUT , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_HAUTEUR  , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_TXT_FORME, 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_FORME    , 0, +3, DEPLACE_PIXELS );

  OBJ_deplace( PARAMDETECTEUR, PD_COUPE, 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( PARAMDETECTEUR, PD_TXT_FANO, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_FANO    , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( PARAMDETECTEUR, PD_TXT_MODELE   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_MOD_REP_NRJ  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_TXT_PARAM_REP, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMDETECTEUR, PD_PARAM_REP    , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( PARAMDETECTEUR, PD_GRF_NON_LIN, 0, +1, DEPLACE_PIXELS );

  arbres[ PARAMDETECTEUR ].fonction.arbre = GER_Parametrage;
  arbres[ PARAMDETECTEUR ].drapeaux.prepare = VRAI;
}

/* ~-~-~-~-~-~-~-~- Routines pour le slecteur universel ~-~-~-~-~-~-~-~- */

/* -------- Le chanage -------- */

int DET_ChainerDetecteur(Detecteur *avant, Detecteur *detecteur)
{
  if ( NULL == detecteur ) return -1;

  if ( NULL == premier_detecteur ) {
    premier_detecteur = detecteur;
    return 0;
  }

  /* Ajout en dbut de liste */
  if ( NULL == avant ) {
    detecteur->suivant = premier_detecteur;
    premier_detecteur = detecteur;

    return 0;
#ifdef _DEVELOPPE_
  } else {
    Detecteur *d;

    for ( d = premier_detecteur ; d != avant ; d = d->suivant ) {
      if ( NULL == d ) {
	(void) fprintf(stderr, "%s,%s l.%u : le dtecteur indiqu comme avant n'est pas dans la liste !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	abort();
      }
    }
#endif
  }


  /* Ajout aprs celui indiqu */
  detecteur->suivant = avant->suivant;
  avant->suivant = detecteur;

  return 0;
}

int DET_DechainerDetecteur(Detecteur *detecteur)
{
  Detecteur *avant;

  if ( NULL == detecteur ) return -1;

  if ( detecteur == premier_detecteur ) {
    premier_detecteur = detecteur->suivant;
    return 0;
  }

  for ( avant = premier_detecteur ; avant->suivant != detecteur ; avant = avant->suivant ) {
    if ( NULL == avant ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : dtecteur hors-chanage\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      return +1;
    }
  }

  avant->suivant = detecteur->suivant;
  return 0;
}

/* -------- Trouver le dtecteur demand -------- */

unsigned long DETECTEUR_D_Vers_N(Detecteur *premier, Detecteur *detecteur)
{
  Detecteur *idx_d;
  unsigned long i;

  if ( NULL == premier ) premier = premier_detecteur;
  if ( NULL == detecteur ) return 0;

  for ( i = 0, idx_d = premier ; idx_d != detecteur ; idx_d = idx_d->suivant, i++) {
    if ( NULL == idx_d ) {
      (void) fprintf(stderr, "%s,%s l.%u : Dtecteur non-trouv [%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, detecteur->nom);
      return 0;
    }
  }

  return i;
}


Detecteur *DETECTEUR_N_Vers_D(Detecteur *premier, unsigned long n)
{
  Detecteur *detecteur;
  unsigned long i;

  if ( NULL == premier ) premier = premier_detecteur;
  if ( 0 == n ) return premier;

  for ( i = 0, detecteur = premier ; i < n ; i++, detecteur = detecteur->suivant ) {
    if ( NULL == detecteur ) {
      (void) fprintf(stderr, "%s,%s l.%u : n=%lu, seulement %lu dtecteurs\n",
		     __FILE__, __FUNCTION__, __LINE__, n + 1, DETECTEUR_Compter() );
      return NULL;
    }
  }

  /* On a trouv */
  return detecteur;
}

int DETECTEUR_Charger_I(unsigned long n_detecteur)
{
  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, n_detecteur);

  /* Tout s'est bien pass */
  return 0;
}

int DETECTEUR_Copier_I(unsigned long n_original)
{
  Detecteur *original;

  original = DETECTEUR_N_Vers_D( premier_detecteur, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans DET_N_Vers_D\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE\n",
		 __FILE__, __FUNCTION__, __LINE__);

  return 0;
}

int DETECTEUR_Effacer_I(unsigned long n_original)
{
  Detecteur *original;

  original = DETECTEUR_N_Vers_D( premier_detecteur, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans DETECTEUR_N_Vers_D\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  DET_EffacerDetecteur( original, FAUX );

  return 0;
}

int DETECTEUR_Modifier_I(unsigned long n_original)
{
  Detecteur *original;

  original = DETECTEUR_N_Vers_D( premier_detecteur, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans DETECTEUR_N_Vers_D\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  DETECTEUR_Parametrer( original );

  return 0;
}

int DETECTEUR_Nouveau_I (unsigned long n_avant)
{
  Detecteur *detecteur, *avant;
  int retour;

  /* On le cre */
  detecteur = DET_CreerDetecteur( DETECTEUR_CHAMBRE, 0 );
  if ( NULL == detecteur ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans DET_CreerDetecteur()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On cherche o le placer dans la liste */
  avant = premier_detecteur;
  if ( NULL == avant ) {
    retour = DET_ChainerDetecteur( NULL, detecteur );
  } else {
    avant = DETECTEUR_N_Vers_D( premier_detecteur, n_avant );
    retour = DET_ChainerDetecteur( avant, detecteur );
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans DET_ChainerDetecteur()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    DET_EffacerDetecteur( detecteur, VRAI );
    return -1;
  }

  return 0;
}

int DETECTEUR_Sauver_I(unsigned long n_detecteur)
{
  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, n_detecteur);

  /* Tout s'est bien pass */
  return 0;
}

unsigned long DETECTEUR_Compter(void)
{
  unsigned long i;
  Detecteur *d;

  for ( d = premier_detecteur, i = 0 ; d != NULL ; d = d->suivant, i++);
  return i;
}

const char *DETECTEUR_TrouverNom(unsigned int arbre __attribute__((__unused__)), unsigned long objet)
{
  Detecteur *d;

  d = DETECTEUR_N_Vers_D( premier_detecteur, objet );
  if ( NULL == d ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans DETECTEUR_N_Vers_D\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  return d->nom;
}
