/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modles structurels d'une station exprimentale
   Paramtrer l'chantillon [interface utilisateur]
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "experience.h"
#include "fluorescence.h"
#include "graphes.h"
#include "interface.h"
#include "absorption.h" /* Pour les seuils */

#include "param_experience.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_TYPE        30
#define LGMAX_NOM_SEUIL       20
#define LGMAX_NOM_FAISCEAU    23
#define LGMAX_NOM_I0          23
#define LGMAX_NOM_ECHANTILLON 23
#define LGMAX_NOM_I1          23
#define LGMAX_NOM_COMPOSITION 30

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void InstallerType( int type );
static void ProposerROI(int element, int seuil);
static void preparer_arbre(void);

static int GER_Parametrage(unsigned int objet, int x, int y);
static void PE_ActiverAbsorbante  (boolean sens, struct matiere *composition);
static void PE_ActiverCartographie(boolean sens);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

int EXP_Parametrer( Experience *experience )
{
  struct matiere *composition;
  int retour;

  if ( NULL == experience ) return -1;

  /* Prparation de l'arbre */
  if ( FAUX == arbres[ LANCER_EXP ].drapeaux.prepare ) {
    preparer_arbre();
  }

  /* On installe les paramtres actuels */
  InstallerType( experience->methode );

  ARB_ascenseur( LANCER_EXP, LE_ORDRE_AVANT, LE_FND_ORDRE, LE_BOUTON_ORDRE, LE_ORDRE_APRES, ASC_HORIZONTAL, 5, 1, experience->ordre, ROOT, 0, 0, METTRE );
  OBJ_MettreEntier_U( LANCER_EXP, LE_VAL_ORDRE, (unsigned long) experience->ordre );

  OBJ_MettreIntervalle( LANCER_EXP, LE_E_DEBUT, LE_E_FIN, LE_PAS, 0, experience->E_debut, experience->E_fin, experience->pas_E, 0 );
  OBJ_MettreIntervalle( LANCER_EXP, LE_E_DEB_MESURE, LE_E_FIN_MES, LE_PAS_E_MESURE, 0,
			experience->Ef_debut, experience->Ef_fin, experience->pas_Ef, 0 );
  OBJ_MettreIntervalle( LANCER_EXP, LE_ROI_DEBUT, LE_ROI_FIN, 0, 0, experience->ROI_debut, experience->ROI_fin, 0.0, 0 );

  met_chaine( LANCER_EXP, LE_ELEMENT, 0, ATM_Z_vers_symbole( experience->Z ), TXT_COPIER );
  OBJ_ChoisirSeuil( LANCER_EXP, LESEUIL, LGMAX_NOM_SEUIL, experience->seuil, METTRE );

  if ( NULL == experience->faisceau ) {
    met_chaine( LANCER_EXP, LE_FAISCEAU, LGMAX_NOM_FAISCEAU, texte_pas_faisceau, TXT_COPIER );
  } else {
    met_chaine( LANCER_EXP, LE_FAISCEAU, LGMAX_NOM_FAISCEAU, experience->faisceau->nom, TXT_COPIER );
  }
  if ( NULL == experience->i0 ) {
    met_chaine( LANCER_EXP, LE_I0, LGMAX_NOM_FAISCEAU, texte_pas_i0, TXT_COPIER );
  } else {
    met_chaine( LANCER_EXP, LE_I0, LGMAX_NOM_FAISCEAU, experience->i0->nom, TXT_COPIER );
  }
  if ( NULL == experience->echantillon ) {
    met_chaine( LANCER_EXP, LE_ECHANTILLON, LGMAX_NOM_FAISCEAU, texte_pas_echantillon, TXT_COPIER );
  } else {
    met_chaine( LANCER_EXP, LE_ECHANTILLON, LGMAX_NOM_FAISCEAU, experience->echantillon->nom, TXT_COPIER );
  }
  if ( NULL == experience->i1 ) {
    met_chaine( LANCER_EXP, LE_I1, LGMAX_NOM_FAISCEAU, texte_pas_i1, TXT_COPIER );
  } else {
    met_chaine( LANCER_EXP, LE_I1, LGMAX_NOM_FAISCEAU, experience->i1->nom, TXT_COPIER );
  }

  PE_ActiverAbsorbante( experience->proprietes & EXPERIENCE_ABSORBANTE, experience->composition );
  PE_ActiverCartographie( ( experience->proprietes & EXPERIENCE_CARTOGRAPHIE ) != 0 );

  /* Appel de l'arbre */
  ARB_MettreComplement( LANCER_EXP, experience );
  retour = ARB_Bloquant( LANCER_EXP, titre_arbre_parametrage, principale, 0 );
  ARB_MettreComplement( LANCER_EXP, NULL );
  if ( LE_ANNULER == retour ) return +1;

  /* On enregistre les paramtres */
  experience->ordre = OBJ_LireEntier_U( LANCER_EXP, LE_VAL_ORDRE );

  experience->E_debut = OBJ_LireNombre( LANCER_EXP, LE_E_DEBUT, NULL );
  experience->E_fin   = OBJ_LireNombre( LANCER_EXP, LE_E_FIN, NULL );
  experience->pas_E   = OBJ_LireNombre( LANCER_EXP, LE_PAS, NULL );

  experience->Ef_debut = OBJ_LireNombre( LANCER_EXP, LE_E_DEB_MESURE, NULL );
  experience->Ef_fin   = OBJ_LireNombre( LANCER_EXP, LE_E_FIN_MES, NULL );
  experience->pas_Ef   = OBJ_LireNombre( LANCER_EXP, LE_PAS_E_MESURE, NULL );

  experience->ROI_debut = OBJ_LireNombre( LANCER_EXP, LE_ROI_DEBUT, NULL );
  experience->ROI_fin   = OBJ_LireNombre( LANCER_EXP, LE_ROI_FIN, NULL );

  experience->Z = OBJ_NumeroAtomique( LANCER_EXP, LE_ELEMENT );

  if ( OBJ_choisit( LANCER_EXP, LE_ABSORBE, QUESTION ) == VRAI ) {
    experience->proprietes |=  EXPERIENCE_ABSORBANTE;
  } else {
    experience->proprietes &= ~EXPERIENCE_ABSORBANTE;
  }

  if ( OBJ_choisit( LANCER_EXP, LE_CARTO, QUESTION ) == VRAI ) {
    experience->proprietes |=  EXPERIENCE_CARTOGRAPHIE;

    experience->x_min = OBJ_LireNombre( LANCER_EXP, LE_X_DEBUT, NULL );
    experience->x_max = OBJ_LireNombre( LANCER_EXP, LE_X_FIN  , NULL );
    experience->pas_x = OBJ_LireNombre( LANCER_EXP, LE_PAS_X  , NULL );

    experience->y_min = OBJ_LireNombre( LANCER_EXP, LE_Y_DEBUT, NULL );
    experience->y_max = OBJ_LireNombre( LANCER_EXP, LE_Y_FIN  , NULL );
    experience->pas_y = OBJ_LireNombre( LANCER_EXP, LE_Y_PAS  , NULL );
  } else {
    experience->proprietes &= ~EXPERIENCE_CARTOGRAPHIE;
  }

  /* Si on a modifier les bornes, le type,... il faut refaire les dessins ventuellement
     Il faut donc, en particulier, effacer d'ventuels mu(E) dj calculs ! */
  for ( composition = premiere_composition ; composition != NULL ; composition = composition->suivante ) {
    GRF_eff_isole( composition->mu_total ); composition->mu_total = NULL;
    GRF_eff_isole( composition->te_total ); composition->te_total = NULL;
    GRF_eff_isole( composition->ti_total ); composition->ti_total = NULL;
    GRF_eff_isole( composition->fluo ); composition->fluo = NULL;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Gestion de la boite de dialogue -------- */

static int GER_Parametrage(unsigned int objet, int x, int y)
{
  boolean etat;
  int nouveau;
  Experience *experience;
  InfosSelecteur infos_selecteur;

  experience = ( Experience * ) ARB_LireComplement( LANCER_EXP );
  if ( NULL == experience ) {
    (void) fprintf(stderr, "GER_Parametrage : experience=NULL\n");
    return BI_EVENEMENT_TRAITE;
  }

  switch ( objet )
  {
  case LE_TYPE_EXP:
    nouveau = SS_menu_texte( noms_types_experiences, NOMBRE_TYPE_EXPERIENCE, "Mthode exprimentale", LANCER_EXP, LE_TYPE_EXP );
    if ( ( nouveau >= 0 ) && ( nouveau != experience->methode ) ) {
      InstallerType( nouveau );
      experience->methode = nouveau;
    }
    return BI_EVENEMENT_TRAITE;

  case LE_ABSORBE:
    etat = OBJ_choisit( LANCER_EXP, objet, ECHANGE );
    if ( VRAI == etat ) {
      if ( NULL == experience->composition ) {
	experience->composition = Matiere_ChoisirComposition( LANCER_EXP, LE_COMPOSITION, LGMAX_NOM_COMPOSITION, NULL );
	if ( NULL == experience->composition ) {
	  PE_ActiverAbsorbante( FAUX, experience->composition );
	  return BI_EVENEMENT_TRAITE;
	}
      }

      PE_ActiverAbsorbante( VRAI, experience->composition );
    } else {
      PE_ActiverAbsorbante( FAUX, experience->composition );
    }
    return BI_EVENEMENT_TRAITE;

  case LE_COMPOSITION:
    experience->composition = Matiere_ChoisirComposition( LANCER_EXP, LE_COMPOSITION, LGMAX_NOM_COMPOSITION, NULL );
    PE_ActiverAbsorbante( VRAI, experience->composition );
    return BI_EVENEMENT_TRAITE;

  case LESEUIL:
    experience->seuil = OBJ_ChoisirSeuil( LANCER_EXP, LESEUIL, LGMAX_NOM_SEUIL, experience->seuil, ECHANGE );
    return BI_EVENEMENT_TRAITE;

  case LE_PROPOSER:
    if ( EXP_EstSpectroscopie( experience->methode ) ) {
      int Z;
      double E_seuil;

      Z = OBJ_NumeroAtomique( LANCER_EXP, LE_ELEMENT );
      E_seuil = (double) TH_DonneesSeuil( Z, experience->seuil, NULL, NULL );

      /* Spectre typique : on commence 100 eV avant le seuil, on s'arrte 1000 eV aprs */
      OBJ_MettreNombre( LANCER_EXP, LE_E_DEBUT, rint( E_seuil -  100.0 ), 0.0 );
      OBJ_MettreNombre( LANCER_EXP, LE_E_FIN  , rint( E_seuil + 1000.0 ), 0.0 );
      OBJ_MettreNombre( LANCER_EXP, LE_PAS, 2.0, 0.0 );

      /* ROI : raies de fluorescence associes */
      ProposerROI( Z, experience->seuil );
    } else {
      OBJ_MettreNombre( LANCER_EXP, LE_E_DEBUT, 14000.0, 0.0 ); /* Excitation  14 keV par dfaut */

      /* Intervalle de modlisation : entre 0,1 et 15 keV => spectre quasi-complet */
      OBJ_MettreNombre( LANCER_EXP, LE_E_DEB_MESURE, 100.0, 0.0 );
      OBJ_MettreNombre( LANCER_EXP, LE_E_FIN_MES   , 15000.0, 0.0 );
      OBJ_MettreNombre( LANCER_EXP, LE_PAS_E_MESURE, 2.0, 0.0 ); /* Avec un pas de 2 eV */
    }
    return BI_EVENEMENT_TRAITE;

  case LE_CARTO:
    etat = OBJ_choisit( LANCER_EXP, objet, ECHANGE );
    PE_ActiverCartographie( etat );
    return BI_EVENEMENT_TRAITE;

  case LE_FAISCEAU:
    infos_selecteur.nouveau  = FAISCEAU_Nouveau_I;
    infos_selecteur.copier   = FAISCEAU_Copier_I;
    infos_selecteur.modifier = FAISCEAU_Modifier_I;
    infos_selecteur.sauver   = NULL/*FAISCEAU_Sauver_I*/;
    infos_selecteur.charger  = NULL/*FAISCEAU_Charger_I*/;
    infos_selecteur.effacer  = FAISCEAU_Effacer_I;
    infos_selecteur.compter  = FAISCEAU_Compter;
    infos_selecteur.objet_v_texte = FAISCEAU_TrouverNom;
    nouveau = I_SelecteurUniversel( "Faisceau", &infos_selecteur, ARB_QuelleFenetre( arbres + LANCER_EXP, FAUX ) );
    if ( nouveau == 0 ) {
      experience->faisceau = FAISCEAU_N_Vers_F( NULL, infos_selecteur.obj_choisi );
      if ( NULL == experience->faisceau ) {
	met_chaine( LANCER_EXP, LE_FAISCEAU, LGMAX_NOM_FAISCEAU, "Pas de faisceau d"e_AIGU"fini", TXT_COPIER );
      } else {
	met_chaine( LANCER_EXP, LE_FAISCEAU, LGMAX_NOM_FAISCEAU, experience->faisceau->nom, TXT_COPIER );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case LE_I0:
    infos_selecteur.nouveau  = DETECTEUR_Nouveau_I;
    infos_selecteur.copier   = DETECTEUR_Copier_I;
    infos_selecteur.modifier = DETECTEUR_Modifier_I;
    infos_selecteur.sauver   = NULL/*DETECTEUR_Sauver_I*/;
    infos_selecteur.charger  = NULL/*DETECTEUR_Charger_I*/;
    infos_selecteur.effacer  = DETECTEUR_Effacer_I;
    infos_selecteur.compter  = DETECTEUR_Compter;
    infos_selecteur.objet_v_texte = DETECTEUR_TrouverNom;
    nouveau = I_SelecteurUniversel( "Dtecteur avant chantillon", &infos_selecteur, ARB_QuelleFenetre( arbres + LANCER_EXP, FAUX ) );
    if ( nouveau == 0 ) {
      experience->i0 = DETECTEUR_N_Vers_D( NULL, infos_selecteur.obj_choisi );
      if ( NULL == experience->i0 ) {
	met_chaine( LANCER_EXP, LE_I0, LGMAX_NOM_I0, "Pas de chambre I0 d"e_AIGU"finie", TXT_COPIER );
      } else {
	met_chaine( LANCER_EXP, LE_I0, LGMAX_NOM_I0, experience->i0->nom, TXT_COPIER );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case LE_ECHANTILLON:
    infos_selecteur.nouveau  = ECHANTILLON_Nouveau_I;
    infos_selecteur.copier   = ECHANTILLON_Copier_I;
    infos_selecteur.modifier = ECHANTILLON_Modifier_I;
    infos_selecteur.sauver   = NULL/*ECHANTILLON_Sauver_I*/;
    infos_selecteur.charger  = NULL/*ECHANTILLON_Charger_I*/;
    infos_selecteur.effacer  = ECHANTILLON_Effacer_I;
    infos_selecteur.compter  = ECHANTILLON_Compter;
    infos_selecteur.objet_v_texte = ECHANTILLON_TrouverNom;
    nouveau = I_SelecteurUniversel( "chantillon", &infos_selecteur, ARB_QuelleFenetre( arbres + LANCER_EXP, FAUX ) );
    if ( nouveau == 0 ) {
      experience->echantillon = ECHANTILLON_N_Vers_E( NULL, infos_selecteur.obj_choisi );
      if ( NULL == experience->echantillon ) {
	met_chaine( LANCER_EXP, LE_ECHANTILLON, LGMAX_NOM_ECHANTILLON, "Pas d'chantillon d"e_AIGU"fini", TXT_COPIER );
      } else {
	met_chaine( LANCER_EXP, LE_ECHANTILLON, LGMAX_NOM_ECHANTILLON, experience->echantillon->nom, TXT_COPIER );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case LE_I1:
    infos_selecteur.nouveau  = DETECTEUR_Nouveau_I;
    infos_selecteur.copier   = DETECTEUR_Copier_I;
    infos_selecteur.modifier = DETECTEUR_Modifier_I;
    infos_selecteur.sauver   = NULL/*DETECTEUR_Sauver_I*/;
    infos_selecteur.charger  = NULL/*DETECTEUR_Charger_I*/;
    infos_selecteur.effacer  = DETECTEUR_Effacer_I;
    infos_selecteur.compter  = DETECTEUR_Compter;
    infos_selecteur.objet_v_texte = DETECTEUR_TrouverNom;
    nouveau = I_SelecteurUniversel( "Dtecteur aprs chantillon", &infos_selecteur, ARB_QuelleFenetre( arbres + LANCER_EXP, FAUX ) );
    if ( nouveau == 0 ) {
      experience->i1 = DETECTEUR_N_Vers_D( NULL, infos_selecteur.obj_choisi );
      if ( NULL == experience->i1 ) {
	met_chaine( LANCER_EXP, LE_I1, LGMAX_NOM_I1, "Pas de chambre I1 d"e_AIGU"finie", TXT_COPIER );
      } else {
	met_chaine( LANCER_EXP, LE_I1, LGMAX_NOM_I1, experience->i1->nom, TXT_COPIER );
      }
    }
    return BI_EVENEMENT_TRAITE;
  }

#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "GER_Parametrage : bouton non gr [objet=%u, x=%d, y=%d]\n", objet, x, y);
#endif
  return BI_EVENEMENT_NON_TRAITE;
}

static void InstallerType( int type )
{
  OBJ_active( LANCER_EXP, LE_E_DEBUT, FORCE_OUI );

  if ( EXP_EstSpectroscopie( type ) == VRAI ) {
    OBJ_active( LANCER_EXP, LE_E_FIN, FORCE_OUI );
    OBJ_active( LANCER_EXP, LE_PAS  , FORCE_OUI );
    OBJ_active( LANCER_EXP, LESEUIL , FORCE_OUI );
  } else {
    OBJ_active( LANCER_EXP, LE_E_FIN, FORCE_NON );
    OBJ_active( LANCER_EXP, LE_PAS  , FORCE_NON );
    OBJ_active( LANCER_EXP, LESEUIL , FORCE_NON );
  }

  /* On adapte l'arbre */
  switch ( type )
  {
  case EXPERIENCE_XAS_TRANSMISSION:
    OBJ_active( LANCER_EXP, LE_E_DEB_MESURE, FORCE_NON );
    OBJ_active( LANCER_EXP, LE_E_FIN_MES   , FORCE_NON );
    OBJ_active( LANCER_EXP, LE_PAS_E_MESURE, FORCE_NON );

    OBJ_active( LANCER_EXP, LE_ELEMENT, FORCE_OUI );
    break;

  case EXPERIENCE_XAS_FLUORESCENCE:
    OBJ_active( LANCER_EXP, LE_E_DEB_MESURE, FORCE_OUI );
    OBJ_active( LANCER_EXP, LE_E_FIN_MES   , FORCE_OUI );
    OBJ_active( LANCER_EXP, LE_PAS_E_MESURE, FORCE_OUI );

    OBJ_active( LANCER_EXP, LE_ELEMENT, FORCE_OUI );
    break;

  case EXPERIENCE_XRF:
    OBJ_active( LANCER_EXP, LE_E_DEB_MESURE, FORCE_OUI );
    OBJ_active( LANCER_EXP, LE_E_FIN_MES   , FORCE_OUI );
    OBJ_active( LANCER_EXP, LE_PAS_E_MESURE, FORCE_OUI );

    OBJ_active( LANCER_EXP, LE_ELEMENT, FORCE_NON );
    break;

  default:
    (void) fprintf(stderr, "InstallerType : type d'exprience inconnu [%d]\n", type);
    return;
  }

  /* On installe le nom */
  met_chaine( LANCER_EXP, LE_TYPE_EXP, LGMAX_NOM_TYPE, noms_types_experiences[ type ], TXT_REMPLACE );
}

/* -------- En fonction du seuil, proposer l'intervalle pour la fluorescence -------- */

static void ProposerROI(int element, int seuil)
{
  double e_min = -1.0, e_max = -1.0, lg_raie, i_max;
  int raie, raie_basse, raie_haute, raie_max, retour;

  retour = FLUO_RaiesExtremes( element, seuil, &raie_basse, &raie_haute );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FLUO_RaiesExtremes() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* Valeurs extrmes */
  e_min = FLUO_QuelleEnergie( element, raie_basse, &i_max, NULL );
  e_max = FLUO_QuelleEnergie( element, raie_haute, NULL, NULL );

  /* Recherche de la raie la plus intense : ROI */
  raie_max = raie_basse;
  for ( raie = raie_basse ; raie <= raie_haute ; raie++ ) {
    double energie, intensite;

    energie = FLUO_QuelleEnergie( element, raie, &intensite, NULL );
    if ( intensite > i_max ) {
      i_max = intensite;
      raie_max = raie;
    }
    if ( e_min <= 0.0 ) { /* Raie basse inconnue => on essaye cette raie  la place */
      e_min = energie;
      raie_basse = raie;
    }
    if ( energie > e_max ) { /* Donc, logiquement, raie haute inconnue => on "remonte" */
      e_max = energie;
    }
  }

  /* On tient compte de la largeur ventuelle des raies : 10 eV, arbitrairement */
  e_min  = FLUO_QuelleEnergie( element, raie_basse, NULL, &lg_raie );
  e_min -= 4.0 * lg_raie + 1.0;

  e_max += 4.0 * lg_raie + 1.0;

  /* En-dessous de 50 eV, on peut s'attendre  des soucis ! */
  if ( e_min < 50.0 ) {
    if ( e_max > 50.0 ) {
      e_min = 50.0;
    } else if ( e_max <= 1.0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Problme dans la dtection automatique de la ROI\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      (void) BI_ERR_alerte( NULL, &ERR_ImpossibleTrouverROI, VRAI, ATM_Z_vers_symbole( element ), TH_TexteSeuil( seuil ) );
      return;
    } else if ( e_min < 1.0 ) {
      e_min = 0.5; /* On commence  1 eV */
    }
  }

  /* Zone de modlisation : [e_min - 10, e_max + 10] entiers */
  OBJ_MettreNombre( LANCER_EXP, LE_E_DEB_MESURE, rint( e_min ), 0.0 );
  OBJ_MettreNombre( LANCER_EXP, LE_E_FIN_MES, rint( e_max ), 0.0 );
  OBJ_MettreNombre( LANCER_EXP, LE_PAS_E_MESURE, 1.0, 0.0 ); /* Avec un pas de 1 eV */

  /* ROI : autour de la raie la plus intense */
  e_min = FLUO_QuelleEnergie( element, raie_max, NULL, &lg_raie );
  e_max = e_min + 4.0 * lg_raie + 1.0;
  e_min -= 4.0 * lg_raie - 1.0;

  /* Intervalle de modlisation : comme la ROI */
  OBJ_MettreNombre( LANCER_EXP, LE_ROI_DEBUT, rint( e_min ), 0.0 );
  OBJ_MettreNombre( LANCER_EXP, LE_ROI_FIN, rint( e_max ), 0.0 );
}

static void PE_ActiverAbsorbante( boolean sens, struct matiere *composition )
{
  OBJ_choisit( LANCER_EXP, LE_ABSORBE, sens );
  OBJ_active ( LANCER_EXP, LE_COMPOSITION, sens );
  if ( NULL == composition ) {
    met_chaine( LANCER_EXP, LE_COMPOSITION, LGMAX_NOM_COMPOSITION, texte_pas_composition, TXT_COPIER );
  } else {
    met_chaine( LANCER_EXP, LE_COMPOSITION, LGMAX_NOM_COMPOSITION, composition->nom, TXT_COPIER );
  }
}

static void PE_ActiverCartographie(boolean sens)
{
  OBJ_choisit( LANCER_EXP, LE_CARTO, sens );
  OBJ_active ( LANCER_EXP, LE_X_DEBUT, sens );
  OBJ_active ( LANCER_EXP, LE_X_FIN, sens );
  OBJ_active ( LANCER_EXP, LE_PAS_X, sens );
  OBJ_active ( LANCER_EXP, LE_Y_DEBUT, sens );
  OBJ_active ( LANCER_EXP, LE_Y_FIN, sens );
  OBJ_active ( LANCER_EXP, LE_Y_PAS, sens );
}

/* -------- Prparation de l'arbre : enjolivations et tout a -------- */

static void preparer_arbre(void)
{
  ARB_EnjoliverAscenseur( LANCER_EXP, LE_ORDRE_AVANT, LE_FND_ORDRE, LE_BOUTON_ORDRE, LE_ORDRE_APRES, ASC_HORIZONTAL, 0 );

  OBJ_deplace( LANCER_EXP, LE_TXT_E_FIN   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( LANCER_EXP, LE_E_FIN       , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( LANCER_EXP, LE_TXT_PAS     , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( LANCER_EXP, LE_PAS         , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( LANCER_EXP, LE_T_E_FIN_MES , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( LANCER_EXP, LE_E_FIN_MES   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( LANCER_EXP, LE_T_PAS_E_MES , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( LANCER_EXP, LE_PAS_E_MESURE, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( LANCER_EXP, LE_ROI_FIN   , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( LANCER_EXP, LE_TXT_I0, 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( LANCER_EXP, LE_I0    , 0, 500, DEPLACE_RELATIF );

  OBJ_deplace( LANCER_EXP, LE_I1    , 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( LANCER_EXP, LE_TXT_I1, 0, 500, DEPLACE_RELATIF );

#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_TYPE, texte_type_experience );

  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_ABSORBE, texte_experience_absorbante );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_ORDRE  , texte_ordre_simulation );

  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_E_DEBUT, texte_energie_initiale );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_E_FIN  , texte_energie_finale );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_PAS    , texte_pas );
  OBJ_TraduireTexte( LANCER_EXP, LETXTECITATION, texte_excitation );

  OBJ_TraduireTexte( LANCER_EXP, LETXTMESURE   , texte_mesure );
  OBJ_TraduireTexte( LANCER_EXP, LE_T_E_DEB_MES, texte_energie_initiale );
  OBJ_TraduireTexte( LANCER_EXP, LE_T_E_FIN_MES, texte_energie_finale );
  OBJ_TraduireTexte( LANCER_EXP, LE_T_PAS_E_MES, texte_pas );

  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_SEUIL, texte_seuil );
  OBJ_TraduireTexte( LANCER_EXP, LE_PROPOSER , texte_proposer );

  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_CARTO  , texte_cartographie );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_X_DEB  , texte_X_de );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_X_FIN  , texte_a );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_X_PAS  , texte_pas );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_Y_DEBUT, texte_Y_de );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_Y_FIN  , texte_a );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_Y_PAS  , texte_pas );

  OBJ_TraduireTexte( LANCER_EXP, LE_OK     , texte_accepter );
  OBJ_TraduireTexte( LANCER_EXP, LECALCULER, texte_calculer );
  OBJ_TraduireTexte( LANCER_EXP, LE_ANNULER, texte_annuler );

  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_FAISCEAU, texte_faisceau );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_I0      , texte_detecteur_i0 );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_ECHANT  , texte_echantillon );
  OBJ_TraduireTexte( LANCER_EXP, LE_TXT_I1      , texte_detecteur_i1 );
#endif

  arbres[ LANCER_EXP ].fonction.arbre = GER_Parametrage;
  arbres[ LANCER_EXP ].drapeaux.prepare = VRAI;
}
