/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modèles structurels d'une station expérimentale
   Paramétrer le faisceau incident [interface utilisateur]
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "faisceau.h"
#include "graphes.h"
#include "images.h"
#include "interface.h"
#include "science.h" /* pour M_PI, au cas-où */

#include "param_faisceau.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_COUPE   35
#define LGMAX_NOM_GE      35
#define LGMAX_NOM_IE      35
#define LGMAX_NOM_GE_HARM 27

#define MAX_NBR_HARMONIQUES 50 /* Arbitraire, et pour l'interface uniquement, mais il serait surprenant de nécessiter plus ! */

static unsigned long harmonique_visible = 0;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GER_ParametresFaisceau(unsigned int objet, int x, int y);
static void PrepareArbreParametresFaisceau(void);

static void ActiveDimensions( boolean sens );
static void ActiveHarmoniques(boolean sens, Faisceau *faisceau, boolean question );
static void ActiveInhomogene( boolean sens, Faisceau *faisceau );
static void ActivePolychrome( boolean sens, Faisceau *faisceau );
static void ActiveVariable  ( boolean sens, Faisceau *faisceau );

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Appel de la boîte de dialogue -------- */

int FAISCEAU_Parametrer(Faisceau *faisceau)
{
  double x_vise = 0.0, y_vise = 0.0, distance = 0.0;
  int retour;

  if ( NULL == faisceau ) {
    return -1;
  }

  if ( FAUX == arbres[ PARAMS_FAISCEAU ].drapeaux.prepare ) {
    PrepareArbreParametresFaisceau();
  }

  /* On associe le faiceau à l'arbre
     Au tout début, car GER_ParametresFaisceau en a besoin !
  */
  ARB_MettreComplement( PARAMS_FAISCEAU, faisceau );

  /* On installe les paramètres du faisceau actuel */
  OBJ_MettreNombre( PARAMS_FAISCEAU, PF_INTENSITE, faisceau->intensite, 0.0 );
  met_chaine( PARAMS_FAISCEAU, PF_NOM, 0, faisceau->nom, TXT_COPIER );

  /* Les dimensions sont codées en microns, mais affichées en cm : 1cm = 10 000 microns */
#ifdef _DEVELOPPE_
  printf("%s : [%s] %f×%f I=%g (%g, %g, %g)\n", __FUNCTION__, faisceau->nom, faisceau->largeur, faisceau->hauteur,
	 faisceau->intensite, faisceau->orientation.position.x, faisceau->orientation.position.y, faisceau->orientation.position.z);
#endif
  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
    GER_ParametresFaisceau( PF_PONCTUEL, 0, 0 );
  } else {
    GER_ParametresFaisceau( PF_RECTANGULAIRE, 0, 0 );
  }
  OBJ_MettreNombre( PARAMS_FAISCEAU, PF_LARGEUR, faisceau->largeur / 1.0e4, 0.0 );
  OBJ_MettreNombre( PARAMS_FAISCEAU, PF_HAUTEUR, faisceau->hauteur / 1.0e4, 0.0 );

  /* Les angles sont codés en radians, mais affichés en degrés */
  OBJ_MettreAngle( PARAMS_FAISCEAU, PF_THETA, faisceau->orientation.theta, 0.0 );
  OBJ_MettreAngle( PARAMS_FAISCEAU, PF_PHI  , faisceau->orientation.phi  , 0.0 );
  OBJ_MettreAngle( PARAMS_FAISCEAU, PF_ALPHA, faisceau->orientation.alpha, 0.0 );

  /* Recherche du point visé */
  if ( faisceau->orientation.position.y != 0.0 ) {
    double D;

    D = cos( faisceau->orientation.theta ) * cos( faisceau->orientation.phi );
    if ( 0.0 == D ) {
      (void) fprintf(stderr, "FAISCEAU_Parametrer : Faisceau parallèle à l'échantillon\n");
    } else {
      double t;

      t = - faisceau->orientation.position.y / D;
      if ( t < 0.0 ) {
	(void) fprintf(stderr, "FAISCEAU_Parametrer : ATTENTION, le faisceau ne pointe pas vers l'échantillon\n");
      }
      x_vise = faisceau->orientation.position.x + t * sin( faisceau->orientation.theta );
      y_vise = faisceau->orientation.position.z - t * cos( faisceau->orientation.theta ) * sin( faisceau->orientation.phi );

      distance = fabs( t );
    }
    printf("%s : point visé = (%g, 0, %g) [t=%g, p=%g]\n", __FUNCTION__, x_vise, y_vise,
	   faisceau->orientation.theta, faisceau->orientation.phi );
  } else {
    x_vise = faisceau->orientation.position.x;
    y_vise = faisceau->orientation.position.z;
    distance = fabs( faisceau->orientation.position.y );
  }
  OBJ_MettreNombre( PARAMS_FAISCEAU, PF_X_VISE, x_vise / 1.0e4, 0.0 ); /* Codage en microns, affichage en cm */
  OBJ_MettreNombre( PARAMS_FAISCEAU, PF_Y_VISE, y_vise / 1.0e4, 0.0 );

  /* Faisceau focalisé ? */
  OBJ_choisit( PARAMS_FAISCEAU, PF_FOCALISE, faisceau->proprietes & FAISCEAU_FOCALISE );

  /* Faisceau polychromatique ? */
  if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
    OBJ_choisit( PARAMS_FAISCEAU, PF_POLYCHROME, FORCE_OUI );
    ActivePolychrome( VRAI, faisceau );
  } else {
    OBJ_choisit( PARAMS_FAISCEAU, PF_POLYCHROME, FORCE_NON );
    ActivePolychrome( FAUX, faisceau );
  }

  /* Faisceau inhomogène ? */
  if ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) {
    OBJ_choisit( PARAMS_FAISCEAU, PF_INHOMOGENE, FORCE_OUI );
    ActiveInhomogene( VRAI, faisceau );
  } else {
    OBJ_choisit( PARAMS_FAISCEAU, PF_INHOMOGENE, FORCE_NON );
    ActiveInhomogene( FAUX, faisceau );
  }

  /* Faisceau polarisé ? */
  OBJ_choisit( PARAMS_FAISCEAU, PF_POLARISE, faisceau->proprietes & FAISCEAU_POLARISE );

  /* Faisceau avec des harmoniques ? */
  if ( faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    OBJ_choisit( PARAMS_FAISCEAU, PF_HARMONIQUES, FORCE_OUI );
    ActiveHarmoniques( VRAI, faisceau, VRAI );
  } else {
    OBJ_choisit( PARAMS_FAISCEAU, PF_HARMONIQUES, FORCE_NON );
    ActiveHarmoniques( FAUX, faisceau, FAUX );
  }

  /* Faisceau constant avec l'énergie ? */
  if ( faisceau->proprietes & FAISCEAU_DECROIT ) {
    OBJ_choisit( PARAMS_FAISCEAU, PF_VARIALE, FORCE_OUI );
    ActiveVariable( VRAI, faisceau );
  } else {
    OBJ_choisit( PARAMS_FAISCEAU, PF_VARIALE, FORCE_NON );
    ActiveVariable( FAUX, faisceau );
  }

  retour = ARB_Bloquant( PARAMS_FAISCEAU, titre_arbre_parametres, principale, 0 );
  ARB_MettreComplement( PARAMS_FAISCEAU, NULL );
  if ( PF_ANNULER == retour ) {
    return 0;
  }

  /* On enregistre les changements éventuels [Rq : graphe au vol] */
  if ( VRAI == OBJ_choisit( PARAMS_FAISCEAU, PF_VARIALE, QUESTION ) ) {
    faisceau->proprietes |= FAISCEAU_DECROIT;
  } else {
    faisceau->proprietes &= ~FAISCEAU_DECROIT;
  }

  if ( VRAI == OBJ_choisit( PARAMS_FAISCEAU, PF_HARMONIQUES, QUESTION ) ) {
    faisceau->proprietes |= FAISCEAU_HARMONIQUES;
    faisceau->i_harmoniques[ harmonique_visible ] = OBJ_LireNombre( PARAMS_FAISCEAU, PF_VAL_THETA, NULL );
  } else {
    faisceau->proprietes &= ~FAISCEAU_HARMONIQUES;
  }

  if ( VRAI == OBJ_choisit( PARAMS_FAISCEAU, PF_POLARISE, QUESTION ) ) {
    faisceau->proprietes |= FAISCEAU_POLARISE;
  } else {
    faisceau->proprietes &= ~FAISCEAU_POLARISE;
  }

  if ( VRAI == OBJ_choisit( PARAMS_FAISCEAU, PF_POLYCHROME, QUESTION ) ) {
    faisceau->proprietes |= FAISCEAU_POLYCHROMATIQUE;
  } else {
    faisceau->proprietes &= ~FAISCEAU_POLYCHROMATIQUE;
  }

  if ( VRAI == OBJ_choisit( PARAMS_FAISCEAU, PF_RECTANGULAIRE, QUESTION ) ) { /* Faisceau rectangulaire */
    faisceau->proprietes &= ~FAISCEAU_PONCTUEL;

    faisceau->largeur = 1.0e4 * OBJ_LireNombre( PARAMS_FAISCEAU, PF_LARGEUR, NULL ); /* Affichage en cm, codage en microns */
    faisceau->hauteur = 1.0e4 * OBJ_LireNombre( PARAMS_FAISCEAU, PF_HAUTEUR, NULL );

    faisceau->orientation.alpha = M_PI * OBJ_LireNombre( PARAMS_FAISCEAU, PF_ALPHA, NULL ) / 180.0;

    if ( VRAI == OBJ_choisit( PARAMS_FAISCEAU, PF_INHOMOGENE, QUESTION ) ) {
      faisceau->proprietes |= FAISCEAU_INHOMOGENE;
    } else {
      faisceau->proprietes &= ~FAISCEAU_INHOMOGENE;
    }

    if ( VRAI == OBJ_choisit( PARAMS_FAISCEAU, PF_FOCALISE, QUESTION ) ) {
      faisceau->proprietes |= FAISCEAU_FOCALISE;
    } else {
      faisceau->proprietes &= ~FAISCEAU_FOCALISE;
    }
  } else {
    faisceau->proprietes |= FAISCEAU_PONCTUEL;

    faisceau->proprietes &= ~( FAISCEAU_INHOMOGENE | FAISCEAU_FOCALISE );
  }

  faisceau->intensite = OBJ_LireNombre( PARAMS_FAISCEAU, PF_INTENSITE, NULL );
  FAISCEAU_Nommer( faisceau, lit_chaine( PARAMS_FAISCEAU, PF_NOM, 0, NULL, TXT_COPIER ) );

  /* Orientation & point visé: on s'arrange pour rester à la même distance... */
  {
    double theta, phi, x_vise, y_vise, t;
    double D;

    theta = OBJ_LireAngle( PARAMS_FAISCEAU, PF_THETA, NULL );
    phi   = OBJ_LireAngle( PARAMS_FAISCEAU, PF_PHI, NULL   );
    if ( 0.0 == cos( theta ) ) {
      (void) fprintf(stderr, "FAISCEAU_Parametrer : Faisceau parallèle à l'échantillon [theta = 90 degres]\n");
      theta = 0.0;
    }
    if ( 0.0 == cos( phi ) ) {
      (void) fprintf(stderr, "FAISCEAU_Parametrer : Faisceau parallèle à l'échantillon [phi = 90 degres]\n");
      phi = 0.0;
    }

    faisceau->orientation.theta = theta;
    faisceau->orientation.phi = phi;

    D = cos( faisceau->orientation.theta ) * cos( faisceau->orientation.phi );
    t = distance;
    x_vise = 1e4 * OBJ_LireNombre( PARAMS_FAISCEAU, PF_X_VISE, NULL ); /* Affichage en cm, codage en microns */
    y_vise = 1e4 * OBJ_LireNombre( PARAMS_FAISCEAU, PF_Y_VISE, NULL );

    faisceau->orientation.position.x = x_vise - t * sin( faisceau->orientation.theta );
    faisceau->orientation.position.y = -t * D;
    faisceau->orientation.position.z = y_vise + t * cos( faisceau->orientation.theta ) * sin( faisceau->orientation.phi );
  }

  /* Tout s'est bien passé */
#ifdef _DEVELOPPE_
  printf("%s : [%s] %f×%f I=%g (%g, %g, %g)\n", __FUNCTION__, faisceau->nom, faisceau->largeur, faisceau->hauteur,
	 faisceau->intensite, faisceau->orientation.position.x, faisceau->orientation.position.y, faisceau->orientation.position.z);
#endif
  return 0;
}

/* -------- Gestion de l'arbre -------- */

static int GER_ParametresFaisceau(unsigned int objet, int x, int y)
{
  Faisceau *faisceau;
  Graphe *nv_graphe;
  Image *nv_coupe;
  int etat;

  faisceau = (Faisceau *) ARB_LireComplement( PARAMS_FAISCEAU );
  if ( NULL == faisceau ) {
    (void) fprintf(stderr, "GER_ParametresFaisceau : pas de faisceau\n");
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_TRAITE;
  }

  switch ( objet )
  {
  case PF_PONCTUEL:
    OBJ_choisit( PARAMS_FAISCEAU, PF_PONCTUEL, FORCE_OUI );
    ActiveDimensions( FAUX );
    return BI_EVENEMENT_TRAITE;

  case PF_RECTANGULAIRE:
    OBJ_choisit( PARAMS_FAISCEAU, PF_RECTANGULAIRE, FORCE_OUI );
    ActiveDimensions( VRAI );
    return BI_EVENEMENT_TRAITE;

  case PF_FOCALISE:
  case PF_POLARISE:
    OBJ_choisit( PARAMS_FAISCEAU, objet, ECHANGE );
    return BI_EVENEMENT_TRAITE;

  case PF_INHOMOGENE:
    etat = OBJ_choisit( PARAMS_FAISCEAU, PF_INHOMOGENE, ECHANGE );
    ActiveInhomogene( etat, faisceau );
    return BI_EVENEMENT_TRAITE;

  case PF_CHOIX_COUPE:
    OBJ_choisit( PARAMS_FAISCEAU, objet, FORCE_OUI );
    {
      Image modele;

      modele.genre = IMG_GENRE_REELLE;
      modele.type  = IMG_TYPE_INTENSITES;

      nv_coupe = I_IMG_ChoisirUnique( &modele, I_FILTREIMAGE_GENRE | I_FILTREIMAGE_TYPE, FAUX, VRAI, NULL );
      if ( nv_coupe != NULL ) {
	int retour;

	(void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Association directe !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	retour = FAISCEAU_AssocierCoupe( faisceau, nv_coupe, VRAI, 10, 10 );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_AssocierCouper()\n",
			 __FILE__, __FUNCTION__, __LINE__);
	}
      }
    }
    OBJ_choisit( PARAMS_FAISCEAU, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case PF_POLYCHROME:
    etat = OBJ_choisit( PARAMS_FAISCEAU, objet, ECHANGE );
    ActivePolychrome( etat, faisceau );
    return BI_EVENEMENT_TRAITE;

  case PF_GE:
    nv_graphe = GRF_choisir( GR_RESOL_NRJ, FAUX, VRAI, NULL );
    if ( nv_graphe != NULL ) {
      int retour;

      retour = FAISCEAU_AssocierSpectre( faisceau, nv_graphe, FAUX );
      if ( 0 == retour ) {
	ActivePolychrome( VRAI, faisceau );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case PF_VARIALE:
    etat = OBJ_choisit( PARAMS_FAISCEAU, objet, ECHANGE );
    ActiveVariable( etat, faisceau );
    return BI_EVENEMENT_TRAITE;

  case PF_IE:
    nv_graphe = GRF_choisir( GR_I0, FAUX, VRAI, NULL );
    if ( nv_graphe != NULL ) {
      if ( faisceau->g_energie != NULL ) { /* Il y en avait un avant */
	if ( VRAI == faisceau->g_energie->drapeaux.isole ) {
	  GRF_eff_isole( faisceau->g_energie );
	}
      }
      faisceau->g_energie = nv_graphe;
      ActiveVariable( VRAI, faisceau );
    }
    return BI_EVENEMENT_TRAITE;

  case PF_HARMONIQUES:
    etat = OBJ_choisit( PARAMS_FAISCEAU, objet, ECHANGE );
    ActiveHarmoniques( etat, faisceau, VRAI );
    return BI_EVENEMENT_TRAITE;

  case PF_CHG_NR_HARMO:
    OBJ_choisit( PARAMS_FAISCEAU, objet, FORCE_OUI );
    {
      unsigned int nv_nombre;
      int reponse;
      struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_ENTIER_U, NULL, NULL,
					 BI_EDITABLE_MIN | BI_EDITABLE_MAX, 0.0, 0.0, 0.0,
					 0, NULL, { 0 }, { 0 }, { 0 } };

      nv_valeur.max = (double) MAX_NBR_HARMONIQUES;
      nv_valeur.valeur_initiale.entier_U = faisceau->nbr_harmoniques;
      reponse = BI_DemandeValeur( NULL, &Q_NombreHarmoniques, &nv_valeur );
      if ( reponse == (int) Q_NombreHarmoniques.bouton_annuler )
	nv_nombre = faisceau->nbr_harmoniques;
      else
	nv_nombre = nv_valeur.valeur_donnee.entier_U;

      if ( nv_nombre != faisceau->nbr_harmoniques ) {
	int retour;

	/* Mise à jour des tables */
	retour = FAISCEAU_CreerHarmoniques( faisceau, nv_nombre );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "GER_ : ERREUR dans FAISCEAU_CreerHarmoniques\n");

	  /* Contrôle des harmoniques affichées */
	  if ( harmonique_visible >= faisceau->nbr_harmoniques ) {
	    harmonique_visible = faisceau->nbr_harmoniques - 1;
	  }
	}

	ActiveHarmoniques( VRAI, faisceau, FAUX );
      }
    }
    OBJ_choisit( PARAMS_FAISCEAU, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case PF_HARM_AVANT:
  case PF_HARM_FOND:
  case PF_HARM_OUTON:
  case PF_HARM_APRES:
    {
      unsigned long num_harmonique;

      num_harmonique = ARB_ascenseur( PARAMS_FAISCEAU, PF_HARM_AVANT, PF_HARM_FOND, PF_HARM_OUTON, PF_HARM_APRES, ASC_HORIZONTAL,
				      faisceau->nbr_harmoniques, 1, harmonique_visible, objet, x, y, ECHANGE);
      if ( num_harmonique != harmonique_visible ) { /* On change d'harmonique */
	faisceau->i_harmoniques[ harmonique_visible ] = OBJ_LireNombre( PARAMS_FAISCEAU, PF_VAL_THETA, NULL );
	harmonique_visible = num_harmonique;
	ActiveHarmoniques( VRAI, faisceau, FAUX );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case PF_GE_HARMONIK:
    nv_graphe = GRF_choisir( GR_RESOL_NRJ, FAUX, VRAI, NULL );
    if ( nv_graphe != NULL ) {
      if ( faisceau->g_harmoniques[ harmonique_visible ] != NULL ) { /* Il y en avait un avant */
	if ( VRAI == faisceau->g_harmoniques[ harmonique_visible ]->drapeaux.isole ) {
	  GRF_eff_isole( faisceau->g_harmoniques[ harmonique_visible ] );
	}
      }
      faisceau->g_harmoniques[ harmonique_visible ] = nv_graphe;
      ActiveHarmoniques( VRAI, faisceau, FAUX );
    }
    return BI_EVENEMENT_TRAITE;
  }

  /* Objet non-trouvé */
  return BI_EVENEMENT_NON_TRAITE;
}

/* Activer ce qu'il faut pour un faisceau rectangulaire */

static void ActiveDimensions( boolean sens )
{
  OBJ_active( PARAMS_FAISCEAU, PF_LARGEUR, sens );
  OBJ_active( PARAMS_FAISCEAU, PF_HAUTEUR, sens );
  OBJ_active( PARAMS_FAISCEAU, PF_ALPHA  , sens );

  OBJ_active( PARAMS_FAISCEAU, PF_FOCALISE, sens );

  OBJ_active( PARAMS_FAISCEAU, PF_INHOMOGENE, sens );
  if ( FAUX == sens ) {
    OBJ_choisit( PARAMS_FAISCEAU, PF_INHOMOGENE, FORCE_NON );
    OBJ_choisit( PARAMS_FAISCEAU, PF_FOCALISE  , FORCE_NON );
  }

  OBJ_active( PARAMS_FAISCEAU, PF_CHOIX_COUPE, FORCE_NON );
  OBJ_active( PARAMS_FAISCEAU, PF_VUE_COUPE  , sens );
}

/* Activer ce qu'il faut pour un faisceau avec des harmoniques */

static void ActiveHarmoniques(boolean sens, Faisceau *faisceau, boolean question )
{
  /* Contrôler qu'il y a des harmoniques dans le faisceau */
  if ( VRAI == sens ) {
    if ( ( 0 == faisceau->nbr_harmoniques ) && ( VRAI == question ) ) {
      unsigned int nv_nombre;
      int reponse;
      struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_ENTIER_U, NULL, NULL,
					 BI_EDITABLE_MIN | BI_EDITABLE_MAX, 0.0, 0.0, 0.0,
					 0, NULL, { 0 }, { 0 }, { 0 } };

      nv_valeur.max = (double) MAX_NBR_HARMONIQUES;
      nv_valeur.valeur_initiale.entier_U = faisceau->nbr_harmoniques;
      reponse = BI_DemandeValeur( NULL, &Q_NombreHarmoniques, &nv_valeur );
      if ( reponse == (int) Q_NombreHarmoniques.bouton_annuler )
	nv_nombre = 0;
      else
	nv_nombre = nv_valeur.valeur_donnee.entier_U;

      if ( 0 == nv_nombre ) {
	sens = FAUX;
      } else { /* On essaye de créer les harmoniques souhaitées */
	int retour;

	retour = FAISCEAU_CreerHarmoniques( faisceau, nv_nombre );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "ActiveHarmoniques : ERREUR dans FAISCEAU_CreerHarmoniques\n");
	  sens = FAUX;
	}
      }
    }
  }
  if ( 0 == faisceau->nbr_harmoniques )
    sens = FAUX;

  /* Initialisation des textes */
  OBJ_MettreEntier_U( PARAMS_FAISCEAU, PF_NR_HARMONIK, faisceau->nbr_harmoniques );
  if ( faisceau->g_harmoniques != NULL ) {
    if ( NULL == faisceau->g_harmoniques[ harmonique_visible ] )
      met_chaine( PARAMS_FAISCEAU, PF_GE_HARMONIK, LGMAX_NOM_GE_HARM, texte_sans_gE, TXT_COPIER );
    else
      met_chaine( PARAMS_FAISCEAU, PF_GE_HARMONIK, LGMAX_NOM_GE_HARM, faisceau->g_harmoniques[ harmonique_visible ]->nom, TXT_COPIER );
    OBJ_MettreNombre( PARAMS_FAISCEAU, PF_VAL_THETA, faisceau->i_harmoniques[ harmonique_visible ], 0.0 );
  } else {
    met_chaine( PARAMS_FAISCEAU, PF_GE_HARMONIK, LGMAX_NOM_GE_HARM, texte_sans_harmonique, TXT_COPIER );
    OBJ_MettreNombre( PARAMS_FAISCEAU, PF_VAL_THETA, 0.0, 0.0 );
  }

  /* Activation des objets */
  OBJ_active( PARAMS_FAISCEAU, PF_CHG_NR_HARMO, sens );
  OBJ_active( PARAMS_FAISCEAU, PF_VAL_THETA   , sens );
  OBJ_active( PARAMS_FAISCEAU, PF_GE_HARMONIK , sens );

  if ( FAUX == sens ) {
    ARB_ascenseur( PARAMS_FAISCEAU, PF_HARM_AVANT, PF_HARM_FOND, PF_HARM_OUTON, PF_HARM_APRES, ASC_HORIZONTAL,
		   0, 1, 0, ROOT, 0, 0, METTRE );
  } else {
    ARB_ascenseur( PARAMS_FAISCEAU, PF_HARM_AVANT, PF_HARM_FOND, PF_HARM_OUTON, PF_HARM_APRES, ASC_HORIZONTAL,
		   faisceau->nbr_harmoniques, 1, harmonique_visible, ROOT, 0, 0, METTRE );
  }
}

/* Activer ce qu'il faut pour un faisceau inhomogène */

static void ActiveInhomogene( boolean sens, Faisceau *faisceau )
{
  OBJ_active( PARAMS_FAISCEAU, PF_CHOIX_COUPE, sens );

  if ( NULL == faisceau->g_espace ) {
    met_chaine( PARAMS_FAISCEAU, PF_CHOIX_COUPE, LGMAX_NOM_COUPE, texte_sans_coupe, TXT_COPIER );
  } else {
    met_chaine( PARAMS_FAISCEAU, PF_CHOIX_COUPE, LGMAX_NOM_COUPE, faisceau->g_espace->nom, TXT_COPIER );
  }
}

/* Activer ce qu'il faut pour un faisceau polychromatique */

static void ActivePolychrome( boolean sens, Faisceau *faisceau )
{
  OBJ_active( PARAMS_FAISCEAU, PF_GE, sens );
  OBJ_active( PARAMS_FAISCEAU, PF_AFF_GE, sens );

  OBJ_utilisateur( PARAMS_FAISCEAU, PF_AFF_GE, ( int (*)(PARMBLK *) ) NULL, (long) faisceau->g_E, METTRE );

  if ( NULL == faisceau->g_E ) {
    met_chaine( PARAMS_FAISCEAU, PF_GE, LGMAX_NOM_GE, texte_sans_gE, TXT_COPIER );
  } else {
    met_chaine( PARAMS_FAISCEAU, PF_GE, LGMAX_NOM_GE, faisceau->g_E->nom, TXT_COPIER );
  }
}

/* Activer ce qu'il faut pour un faisceau non constant avec E */

static void ActiveVariable( boolean sens, Faisceau *faisceau )
{
  OBJ_active( PARAMS_FAISCEAU, PF_IE, sens );
  OBJ_active( PARAMS_FAISCEAU, PF_AFF_IE, sens );

  OBJ_utilisateur( PARAMS_FAISCEAU, PF_AFF_IE, ( int (*)(PARMBLK *) ) NULL, (long) faisceau->g_energie, METTRE );

  if ( NULL == faisceau->g_energie ) {
    met_chaine( PARAMS_FAISCEAU, PF_IE, LGMAX_NOM_IE, texte_sans_iE, TXT_COPIER );
  } else {
    met_chaine( PARAMS_FAISCEAU, PF_IE, LGMAX_NOM_IE, faisceau->g_energie->nom, TXT_COPIER );
  }
}

/* -------- On prépare l'arbre : objets utilisateurs,... -------- */

static void PrepareArbreParametresFaisceau(void)
{
  int retour;

  /* Fonctions de redessin pour les objets utilisateurs */
  OBJ_utilisateur( PARAMS_FAISCEAU, PF_VUE_COUPE, USR_image, 0, ECHANGE );
  OBJ_utilisateur( PARAMS_FAISCEAU, PF_AFF_GE, USR_graphe, 0, ECHANGE );
  OBJ_utilisateur( PARAMS_FAISCEAU, PF_AFF_IE, USR_graphe, 0, ECHANGE );

  /* Initialisation de l'arbre */
  retour = ARB_Initialiser( PARAMS_FAISCEAU, NULL, GER_ParametresFaisceau, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'initialiser l'arbre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* Amélioration des positions des objets */
  ARB_EnjoliverAscenseur( PARAMS_FAISCEAU, PF_HARM_AVANT, PF_HARM_FOND, PF_HARM_OUTON, PF_HARM_APRES, ASC_HORIZONTAL, +2 );

  OBJ_deplace( PARAMS_FAISCEAU, PF_TXT_HAUT   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_FAISCEAU, PF_HAUTEUR    , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_FAISCEAU, PF_POLARISE, 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_FAISCEAU, PF_TXT_DIRECTION, 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_FAISCEAU, PF_THETA        , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_FAISCEAU, PF_TXT_ALPHA    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_FAISCEAU, PF_ALPHA        , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_FAISCEAU, PF_TXT_Y   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_FAISCEAU, PF_Y_VISE  , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_FAISCEAU, PF_TXT_COUPE  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_FAISCEAU, PF_CHOIX_COUPE, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_FAISCEAU, PF_TXT_GE, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_FAISCEAU, PF_GE    , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_FAISCEAU, PF_IE, 0, +1, DEPLACE_PIXELS );

  /* Séparateur entre les deux graphes */
  OBJ_deplace( PARAMS_FAISCEAU, 3, -1, 0, DEPLACE_PIXELS );

  arbres[ PARAMS_FAISCEAU ].drapeaux.prepare = VRAI;
}

/* ~-~-~-~-~-~-~-~- Routines pour le sélecteur universel ~-~-~-~-~-~-~-~- */

/* -------- Le chaînage -------- */

int FAISCEAU_ChainerFaisceau(Faisceau *avant, Faisceau *faisceau)
{
  if ( NULL == faisceau ) return -1;

  if ( NULL == premier_faisceau ) {
    premier_faisceau = faisceau;
    return 0;
  }

  /* Ajout en début de liste */
  if ( NULL == avant ) {
    faisceau->suivant = premier_faisceau;
    premier_faisceau = faisceau;

    return 0;
#ifdef _DEVELOPPE_
  } else {
    Faisceau *d;

    for ( d = premier_faisceau ; d != avant ; d = d->suivant ) {
      if ( NULL == d ) {
	(void) fprintf( stderr, "%s,%s l.%u : le faisceau indiqué comme avant n'est pas dans la liste !\n",
		        __FILE__, __FUNCTION__, __LINE__ );
	abort();
      }
    }
#endif
  }


  /* Ajout après celui indiqué */
  faisceau->suivant = avant->suivant;
  avant->suivant = faisceau;

  return 0;
}

/* -------- Trouver le faisceau demandé -------- */

unsigned long FAISCEAU_F_Vers_N(Faisceau *premier, Faisceau *faisceau)
{
  Faisceau *f;
  unsigned long i;

  if ( NULL == premier ) premier = premier_faisceau;
  if ( ( NULL == premier ) || ( NULL == faisceau ) ) return 0;
  if ( faisceau == premier ) return 0;

  for ( i = 0, f = premier ; f != NULL ; i++, f = f->suivant ) {
    if ( f == faisceau ) { /* Trouvé */
      return i;
    }
  }

  (void) fprintf(stderr, "%s,%s l.%u : Faisceau non trouvé [%s]\n",
		 __FILE__, __FUNCTION__, __LINE__, faisceau->nom);
  return 0;
}

Faisceau *FAISCEAU_N_Vers_F(Faisceau *premier, unsigned long n)
{
  Faisceau *faisceau;
  unsigned long i;

  if ( NULL == premier ) premier = premier_faisceau;
  if ( NULL == premier ) return NULL;
  if ( 0 == n ) return premier;

  for ( i = 0, faisceau = premier ; i < n ; i++, faisceau = faisceau->suivant ) {
    if ( NULL == faisceau ) {
      (void) fprintf(stderr, "%s,%s l.%u : n=%lu, seulement %lu détecteurs\n",
		     __FILE__, __FUNCTION__, __LINE__, n + 1, FAISCEAU_Compter() );
      return NULL;
    }
  }

  /* On a trouvé */
  return faisceau;
}

/* -------- Chargement, sauvegarde -------- */

int FAISCEAU_Charger_I(unsigned long n_faisceau)
{
  (void) fprintf(stderr, "%s,%s l.%u : À FAIRE [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, n_faisceau);

  /* Tout s'est bien passé */
  return 0;
}

int FAISCEAU_Copier_I(unsigned long n_original)
{
  Faisceau *original;

  original = FAISCEAU_N_Vers_F( premier_faisceau, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_N_Vers_F\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  (void) fprintf(stderr, "%s,%s l.%u : À FAIRE\n",
		 __FILE__, __FUNCTION__, __LINE__);

  return 0;
}

int FAISCEAU_Effacer_I(unsigned long n_original)
{
  Faisceau *original;

  original = FAISCEAU_N_Vers_F( premier_faisceau, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_N_Vers_F\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  FAISCEAU_Effacer( original );

  return 0;
}

int FAISCEAU_Modifier_I(unsigned long n_original)
{
  Faisceau *original;

  original = FAISCEAU_N_Vers_F( premier_faisceau, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_N_Vers_F\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  FAISCEAU_Parametrer( original );

  return 0;
}

int FAISCEAU_Nouveau_I(unsigned long n_avant)
{
  Faisceau *faisceau, *avant;
  int retour;

  /* On le crée */
  faisceau = FAISCEAU_Creer( 0.0, 0.0, 0, 0, 1.0 );
  if ( NULL == faisceau ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_Creer()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On cherche où le placer dans la liste */
  avant = premier_faisceau;
  if ( NULL == avant ) {
    retour = FAISCEAU_ChainerFaisceau( NULL, faisceau );
  } else {
    avant = FAISCEAU_N_Vers_F( premier_faisceau, n_avant );
    retour = FAISCEAU_ChainerFaisceau( avant, faisceau );
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_ChainerFaisceau()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FAISCEAU_Effacer( faisceau );
    return -1;
  }

  /* On le modifie */
  FAISCEAU_Parametrer( faisceau );

  return 0;
}

int FAISCEAU_Sauver_I(unsigned long n_faisceau)
{
  (void) fprintf(stderr, "%s,%s l.%u : À FAIRE [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, n_faisceau);

  /* Tout s'est bien passé */
  return 0;
}

unsigned long FAISCEAU_Compter(void)
{
  unsigned long i;
  Faisceau *d;

  for ( d = premier_faisceau, i = 0 ; d != NULL ; d = d->suivant, i++);
  return i;
}

const char *FAISCEAU_TrouverNom(unsigned int arbre __attribute__((__unused__)), unsigned long objet)
{
  Faisceau *d;

  d = FAISCEAU_N_Vers_F( premier_faisceau, objet );
  if ( NULL == d ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_N_Vers_F\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  return d->nom;
}
