/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modles structurels d'une station exprimentale
   Sauvegarde d'une exprience
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "bin_utils.h"

#include "experience.h"
#include "fichiers.h"
#include "graphes.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Sauvegarde / exprience --------
   Format (apparat avec la version 3.0) :

int             Type d'exprience modlise
int             Proprits de l'exprience
double          nergie d'excitation (dbut)
Si spectroscopie :
 double           nergie d'excitation (fin)
 double           Pas pour l'excitation
unsigned int    Ordre de modlisation
unsigned int    lment tudi
unsigned int    Seuil tudi
Si fluorescence :
 double           Dbut de l'intervalle de fluorescence
 double           Fin de l'intervalle de fluorescence
 double           Pas pour l'intervalle de fluorescence
 double           Dbut de la ROI
 double           Fin de la ROI

unsigned long   Nombre de faisceaux dfinis
<< Faisceau >>  Chacun des faisceaux
unsigned long   Numro du faisceau utilis dans l'exprience

unsigned long   Nombre de composants dfinis
<< Composant >> Chacun des composants

unsigned long   Nombre de compositions dfinies
<< Composition>> Chacune des compositions

unsigned long   Nombre d'chantillons
S'il y a des chantillons :
  << chantillon>> Chacun des chantillons
  unsigned long    Numro de l'chantillon utilis dans l'exprience

unsigned long   Nombre de dtecteurs
S'il y a des dtecteurs :
  << Dtecteurs >> Chacun des dtecteurs
  unsigned long    Numro du dtecteur utilis comme I0
  unsigned long    Numro du dtecteur utilis comme I1
*/

int EXP_SauverIsolee(const char *nom_fichier, Experience *experience)
{
  int retour;
  unsigned long nbr_objets, i, num_objet;
  FILE *fichier;

  if ( NULL == nom_fichier ) return -10;
  if ( NULL == experience  ) return -1;

  /* 0. Ouverture du fichier */
  fichier = fopen( nom_fichier, "wb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier );
    return -3;
  }

  /* 1. En-tte LASE dcrivant le format, la version,... */
  retour = SV_EnTete( fichier, FICHIER_EXPERIENCE );
  if ( FAUX == retour ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_EnTete()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -4;
  }

  /* Informations pratiques */
  fwrite( &(experience->methode), sizeof( int ), 1, fichier );    /* Type */
  fwrite( &(experience->proprietes), sizeof( int ), 1, fichier ); /* Proprits */

  fwrite( &(experience->E_debut), sizeof( double ), 1, fichier ); /* nergie d'excitation */
  if ( VRAI == EXP_EstSpectroscopie( experience->methode ) ) {
    fwrite( &(experience->E_fin), sizeof( double ), 1, fichier ); /* nergie d'excitation / fin du spectre */
    fwrite( &(experience->pas_E), sizeof( double ), 1, fichier ); /* nergie d'excitation / pas */
  }

  fwrite( &(experience->ordre), sizeof( unsigned int ), 1, fichier ); /* Ordre de modlisation */

  fwrite( &(experience->Z), sizeof( unsigned int ), 1, fichier );     /* lment tudi */
  fwrite( &(experience->seuil), sizeof( unsigned int ), 1, fichier ); /* Seuil tudi */

  if ( VRAI == EXP_UtiliseFluorescence( experience ) ) {
    fwrite( &(experience->Ef_debut), sizeof( double ), 1, fichier ); /* Dbut de l'intervalle de fluorescence */
    fwrite( &(experience->Ef_fin), sizeof( double ), 1, fichier );   /* Fin de l'intervalle de fluorescence */
    fwrite( &(experience->pas_Ef), sizeof( double ), 1, fichier );   /* Pas */

    fwrite( &(experience->ROI_debut), sizeof( double ), 1, fichier ); /* Dbut de la ROI */
    fwrite( &(experience->ROI_fin), sizeof( double ), 1, fichier );   /* Fin de la ROI */
  }

  /* Nombre de faisceaux disponibles */
  nbr_objets = FAISCEAU_Compter();
  fwrite( &nbr_objets, sizeof( unsigned long ), 1, fichier );

  /* Sauvegarde de tous ces faisceaux */
  for ( i = 0 ; i < nbr_objets ; i++ ) {
    Faisceau *faisceau;

    faisceau = FAISCEAU_N_Vers_F( NULL, i );
    retour = FAISCEAU_Sauver( fichier, faisceau );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_Sauver(), faisceau %lu sur %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i, nbr_objets);
      (void) fclose( fichier );
      return -5;
    }
  }

  /* Quel faisceau est utilis par l'exprience ? */
  num_objet = FAISCEAU_F_Vers_N( NULL, experience->faisceau );
  fwrite( &num_objet, sizeof( unsigned long ), 1, fichier );

  /* Nombre de composants disponibles */
  nbr_objets = Composant_Compter();
  fwrite( &nbr_objets, sizeof( unsigned long ), 1, fichier );

  /* Sauvegarde de tous ces composants */
  for ( i = 0 ; i < nbr_objets ; i++ ) {
    struct composant *composant;

    composant = Composant_N_Vers_C( NULL, i );
    retour = Composant_Sauver( fichier, composant );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans Composant_Sauver(), composant %lu sur %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i, nbr_objets);
      (void) fclose( fichier );
      return -5;
    }
  }

  /* Nombre de compositions disponibles */
  nbr_objets = COMPOSITION_Compter();
  fwrite( &nbr_objets, sizeof( unsigned long ), 1, fichier );

  /* On sauve ces compositions */
  for ( i = 0 ; i < nbr_objets ; i++ ) {
    struct matiere *composition;

    composition = COMPOSITION_N_Vers_C( NULL, i );
    retour = Composition_Sauver( fichier, composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans Composition_Sauver(), composition %lu sur %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i, nbr_objets);
      (void) fclose( fichier );
      return -6;
    }
  }

  /* Nombre d'chantillons disponibles */
  nbr_objets = ECHANTILLON_Compter();
  fwrite( &nbr_objets, sizeof( unsigned long ), 1, fichier );

  /* On sauve ces chantillons */
  for ( i = 0 ; i < nbr_objets ; i++ ) {
    Echantillon *echantillon;

    echantillon = ECHANTILLON_N_Vers_E( NULL, i );
    retour = ECHANTILLON_Sauver( fichier, echantillon );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans ECHANTILLON_Sauver(), chantillon %lu sur %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i, nbr_objets);
      (void) fclose( fichier );
      return -6;
    }
  }

  if ( nbr_objets > 0 ) {
    /* Quel chantillon sert pour l'exprience ? */
    num_objet = ECHANTILLON_E_Vers_N( NULL, experience->echantillon );
    fwrite( &num_objet, sizeof( unsigned long ), 1, fichier );
  }

  /* Combien de dtecteurs dfinis ? */
  nbr_objets = DETECTEUR_Compter();
  fwrite( &nbr_objets, sizeof( unsigned long ), 1, fichier );

  /* On sauve ces dtecteurs */
  for ( i = 0 ; i < nbr_objets ; i++ ) {
    Detecteur *detecteur;

    detecteur = DETECTEUR_N_Vers_D( NULL, i );
    retour = DETECTEUR_Sauver( fichier, detecteur );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans DETECTEUR_Sauver(), dtecteur %lu sur %lu\n",
		     __FILE__, __FUNCTION__, __LINE__, i, nbr_objets);
      (void) fclose( fichier );
      return -6;
    }
  }

  /* Quels dtecteurs servent pour l'exprience ? */
  if ( nbr_objets > 0 ) {
    /* Quel dtecteur sert pour I0 ? */
    num_objet = DETECTEUR_D_Vers_N( NULL, experience->i0 );
    fwrite( &num_objet, sizeof( unsigned long ), 1, fichier );

    /* Quel dtecteur sert pour I1 ? */
    num_objet = DETECTEUR_D_Vers_N( NULL, experience->i1 );
    fwrite( &num_objet, sizeof( unsigned long ), 1, fichier );
  }

  /* Termin */
  (void) fclose( fichier );

  /* La sauvegarde s'est bien passe */
  return 0;
}

/* -------- Sauvegarde d'un composant --------
 << chaine >> Nom du composant
 float        Densit du composant
 Pour chacun des numros atomiques :
   float        Fraction molaire
   Si > 0
    << Graphe ? >>  Coefficient d'absorption massique
 << Graphe ? >> Coefficient d'absorption massique global
 */

int Composant_Sauver(FILE *fichier, struct composant *composant)
{
  int retour;
  unsigned int i;

  if ( NULL == fichier ) return -1;
  if ( NULL == composant ) {
    (void) fprintf(stderr, "%s,%s l.%u : Composant = NULL !!\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
 }

  /* Le nom du composant (contient la formule aussi, si tout va bien !) */
  retour = BIN_SV_chaine( fichier, composant->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_chaine(\"%s\") [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, composant->nom, retour);
    return -10;
  }

  /* Densit */
  fwrite( &(composant->densite), sizeof( float ), 1, fichier );

  /* Atomes constitutifs */
  for ( i = 0 ; i < ZMAX_TABLE_CONSTITUANTS ; i++ ) {
    fwrite( composant->atomes + i, sizeof( float ), 1, fichier );

    if ( composant->atomes[ i ] > 0.0 ) {
      retour = SV_Graphe( fichier, composant->mu[ i ] );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Graphe() /rho pour Z=%u [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, i + 1, retour);
	return -11;
      }
    }
  }

  /* /rho global */
  retour = SV_Graphe( fichier, composant->mu_total );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Graphe() /rho global [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -11;
  }

  /* La sauvegarde a russi sans souci */
  return 0;
}

/* -------- Sauvegarde d'une composition ----

 << chaine >>     Nom de la composition
 double           Densit de cette composition
 unsigned int     Nombre de composants
 n * float        Fraction massique de chacun des composants
 n * unsigned int Numros des composants
 */

int Composition_Sauver(FILE *fichier, struct matiere *composition)
{
  int retour;

  if ( NULL == fichier ) return -1;
  if ( NULL == composition ) {
    (void) fprintf(stderr, "%s,%s l.%u : composition = NULL !!\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* Le nom de la composition */
  retour = BIN_SV_chaine( fichier, composition->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_chaine(\"%s\") [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, composition->nom, retour);
    return -10;
  }

  /* La densit */
  fwrite( &(composition->densite), sizeof( double ), 1, fichier );

  /* Le nombre de composants */
  fwrite( &(composition->nbr_composants), sizeof( unsigned int ), 1, fichier );

  if ( composition->nbr_composants > 0 ) {
    unsigned int i;

    /* La table des fractions massiques */
    fwrite( composition->fraction_massique, sizeof( float ), composition->nbr_composants, fichier );

    /* La table des constituants */
    for ( i = 0 ; i < composition->nbr_composants ; i++ ) {
      unsigned int numero_composant;

      numero_composant = Composant_C_Vers_N( NULL, composition->composants[ i ] );
      fwrite( &numero_composant, sizeof( unsigned int ), 1, fichier );
    }
  }

  /* On ne sauvegarde par les graphes, car recalculs au vol ! */

  /* Sauvegarde russie */
  return 0;
}

/* -------- Sauvegarde d'un dtecteur --------

 << chaine >>      Nom du dtecteur
 unsigned int      Proprits du dtecteur
 << Orientation >> Position et orientation du dtecteur
 double            Angle solide intercept
 2 * double        Dimensions du dtecteur
 double            Longueur du dtecteur (pour absorption)
   Si > 0 :
    unsigned int     Composition du dtecteur (pour absorption)
 double            Coups  vide
 << Graphe ? >>    Rponse avec l'intensit
 << Graphe ? >>    Rponse avec l'nergie
 << Image ? >>     Rponse avec la surface
 */

int DETECTEUR_Sauver(FILE *fichier, Detecteur *detecteur)
{
  int retour;

  if ( NULL == fichier ) return -1;
  if ( NULL == detecteur ) {
    (void) fprintf(stderr, "%s,%s l.%u : detecteur = NULL !!\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* Le nom du dtecteur */
  retour = BIN_SV_chaine( fichier, detecteur->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_chaine(\"%s\") [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, detecteur->nom, retour);
    return -10;
  }

  /* Proprits */
  fwrite( &(detecteur->proprietes), sizeof( unsigned int ), 1, fichier );

  /* Position, orientation */
  retour = BIN_SV_orientation( fichier, &(detecteur->orientation) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_orientation() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -11;
  }

  /* Angle solide intercept */
  fwrite( &(detecteur->angle_solide), sizeof( double ), 1, fichier );

  /* Dimensions du dtecteur */
  fwrite( &(detecteur->largeur), sizeof( double ), 1, fichier );
  fwrite( &(detecteur->hauteur), sizeof( double ), 1, fichier );

  /* Longueur traverse */
  fwrite( &(detecteur->longueur), sizeof( double ), 1, fichier );
  if ( detecteur->longueur > 0.0 ) {
    unsigned int num_composition;

    num_composition = COMPOSITION_C_Vers_N( NULL, detecteur->composition );
    fwrite( &num_composition, sizeof( unsigned int ), 1, fichier );
  }

  /* Coups  vide */
  fwrite( &(detecteur->coups_a_vide), sizeof( double ), 1, fichier );

  /* Rponse avec l'intensite */
  retour = SV_Graphe( fichier, detecteur->r_signal );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Graphe(), R(I) [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -13;
  }

  /* Rponse avec l'nergie */
  retour = SV_Graphe( fichier, detecteur->r_energie );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Graphe(), R(E) [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -13;
  }

  /* Rponse de la surface */
  retour = SV_Image( fichier, detecteur->r_surface );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Image(), R(x,y) [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -16;
  }

  /* Sauvegarde russie */
  return 0;
}

/* -------- Sauvegarde d'un chantillon --------

 << chaine >>      Nom de l'chantillon
 int               Type de l'chantillon
 3 * double        Dimensions de l'chantillon
 3 * unsigned int  Nombre de valeurs dans chaque direction
 unsigned int      Proprits de l'chantillon
 Si chantillon homogne
   unsigned int      Numro de la composition utilise
 Si chantillon gradient
   unsigned int      Numro de la composition utilise
   n * double        Corrections de densit en chaque point
 Si chantillon inhomogne
   n * int           Numros de composition (-1 = trou)
 */

int ECHANTILLON_Sauver(FILE *fichier, Echantillon *echantillon)
{
  int retour;
  unsigned long nbr_cases;

  if ( NULL == fichier ) return -1;
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : echantillon = NULL !!\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* Le nom de l'chantillon */
  retour = BIN_SV_chaine( fichier, echantillon->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_chaine(\"%s\") [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, echantillon->nom, retour);
    return -10;
  }

  /* Le type */
  fwrite( &(echantillon->type), sizeof( int ), 1, fichier );

  /* Les dimensions */
  fwrite( &(echantillon->larg), sizeof( double ), 1, fichier );
  fwrite( &(echantillon->haut), sizeof( double ), 1, fichier );
  fwrite( &(echantillon->epaisseur), sizeof( double ), 1, fichier );

  /* Les pas */
  fwrite( &(echantillon->nbr_x), sizeof( unsigned int ), 1, fichier );
  fwrite( &(echantillon->nbr_y), sizeof( unsigned int ), 1, fichier );
  fwrite( &(echantillon->nbr_z), sizeof( unsigned int ), 1, fichier );

  /* Les proprits */
  fwrite( &(echantillon->proprietes), sizeof( unsigned int ), 1, fichier );

  /* Nombre de cases  prendre en compte */
  nbr_cases = 1 + ECH_IJK_v_Case( echantillon, echantillon->nbr_x - 1, echantillon->nbr_y - 1, echantillon->nbr_z - 1 );

  /* Le contenu */
  if ( ECHANTILLON_HOMOGENE & echantillon->proprietes ) {
    unsigned int num_composition;

    /* Il suffit de sauver la composition ! */
    num_composition = COMPOSITION_C_Vers_N( NULL, echantillon->valeurs.homogene.composition );
    fwrite( &num_composition, sizeof( unsigned int ), 1, fichier );
  } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    unsigned int num_composition;

    /* La composition */
    num_composition = COMPOSITION_C_Vers_N( NULL, echantillon->valeurs.homogene.composition );
    fwrite( &num_composition, sizeof( unsigned int ), 1, fichier );

    /* La table des valeurs de correction de la densit */
    fwrite( echantillon->valeurs.homogene.concentration, sizeof( double ), nbr_cases, fichier );
  } else {
    int num_composition;
    unsigned long i;

    /* Composition en chaque point */
    for ( i = 0 ; i < nbr_cases ; i++ ) {
      struct matiere *composition;

      composition = echantillon->valeurs.inhomogene[ i ];
      if ( NULL == composition ) {
	num_composition = -1;
      } else {
	num_composition = (int) COMPOSITION_C_Vers_N( NULL, composition );
      }
      fwrite( &num_composition, sizeof( int ), 1, fichier );
    }
  }

  /* Sauvegarde russie */
  return 0;
}

/* -------- Sauvegarde d'un faisceau --------

 << chaine >>      Nom du faisceau
 double            Largeur du faisceau
 double            Hauteur du faisceau
 double            Intensit du faisceau
 unsigned int      Proprits du faisceau
 << orientation >> Orientation du faisceau
 Si focalis :
    << vecteur >>    Point focal
 << Graphe ? >>    Intensit avec l'nergie
 << Graphe ? >>    Polychromaticit (du fondamental)
 unsigned int      Nombre d'harmoniques
 Pour chaque harmonique :
   << float >>       Intensit relative au fondamental
   << Graphe ? >>    Polychromaticit
 << Image ? >>     Carte de rpartition spatiale
 */

int FAISCEAU_Sauver(FILE *fichier, Faisceau *faisceau)
{
  int retour;
  unsigned int i;

  if ( NULL == fichier ) return -1;
  if ( NULL == faisceau ) {
    (void) fprintf(stderr, "%s,%s l.%u : Faisceau = NULL !!\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
 }

  /* Le nom du faisceau */
  retour = BIN_SV_chaine( fichier, faisceau->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_chaine(\"%s\") [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, faisceau->nom, retour);
    return -10;
  }

  /* Les dimensions */
  fwrite( &(faisceau->largeur), sizeof( double ), 1, fichier );
  fwrite( &(faisceau->hauteur), sizeof( double ), 1, fichier );

  /* L'intensit */
  fwrite( &(faisceau->intensite), sizeof( double ), 1, fichier );

  /* Les proprits */
  fwrite( &(faisceau->proprietes), sizeof( unsigned int ), 1, fichier );

  /* L'orientation */
  retour = BIN_SV_orientation( fichier, &(faisceau->orientation) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_orientation() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -11;
  }

  /* Si ncessaire, le point focal */
  if ( FAISCEAU_FOCALISE & faisceau->proprietes ) {
    retour = BIN_SV_vecteur_3d( fichier, &(faisceau->point_focal) );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_point_focal() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return -12;
    }
  }

  /* Intensit avec l'nergie nominale */
  retour = SV_Graphe( fichier, faisceau->g_energie );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Graphe(), I(E) [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -13;
  }

  /* Spectre du faisceau  Monochomatique  */
  retour = SV_Graphe( fichier, faisceau->g_E );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Graphe(), g(E) [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -14;
  }

  /* Nombre d'harmoniques */
  fwrite( &(faisceau->nbr_harmoniques), sizeof( unsigned int ), 1, fichier );

  /* Les harmoniques */
  for ( i = 0 ; i < faisceau->nbr_harmoniques ; i++ ) {
    fwrite( faisceau->i_harmoniques + i, sizeof( float ), 1, fichier );

    retour = SV_Graphe( fichier, faisceau->g_harmoniques[ i ] );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Graphe(), g(E) harmonique %u/%u [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, i + 1, faisceau->nbr_harmoniques, retour);
      return -15;
    }
  }

  /* Carte de rpartition spatiale */
  retour = SV_Image( fichier, faisceau->g_espace );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_Image(), I(x,y) [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -16;
  }

  /* La sauvegarde a russi sans souci */
  return 0;
}
