/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS

   Module d'affichage graphique

   Boîtes à moustaches ("boxplots")
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"
#include "lissage.h"
#include "statistiques.h"
#include "tracer.h"

/* ~~~~~~~~ Variables locales à ce module ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales à ce module ~~~~~~~~ */

static int TRC_BM_Horizontale(Fenetre *fenetre, InfosRegion2D *region, struct graphe *valeurs, struct boite_moustache *infos, int position);
/* static int TRC_BM_Verticale  (Fenetre *fenetre, InfosRegion2D *region, struct graphe *valeurs, struct boite_moustache *infos, int position); */

/* ~~~~~~~~ Réalisation pratique des fonctions ~~~ */

/* ~-~-~-~-~-~-~-~- Préparation du tracé : détermination des paramètres importants ~-~-~-~-~-~-~-~- */

int TRC_GRF_PreparerBoiteAMoustache(struct graphe *valeurs, double f_inferieur, double f_superieur, double f_min, double f_max )
{
  double quartiles[ 3 ];
  int retour;
  struct boite_moustache *infos;
  struct graphe *f_repartition;

  /* Contrôles préliminaires */
  if ( NULL == valeurs ) return -1;
  if ( NULL == valeurs->affichage ) {
    AVERTISSEMENT( "Graphe sans infos d'affichage ! [%s]", valeurs->nom );
    return -2;
  }
  if ( valeurs->genre_donnees != GRF_SERIE ) {
    AVERTISSEMENT( "Graphe de genre incorrect : %d, il faudrait %d", valeurs->genre_donnees, GRF_SERIE );
    return -3;
  }

  /* Mise en place des valeurs par défaut, si besoin est */
  if ( ( f_inferieur < 0.0 ) || ( f_inferieur > 1.0 ) ) f_inferieur = 0.25;  /* Limites de la boîte : quartiles */
  if ( ( f_superieur < 0.0 ) || ( f_superieur > 1.0 ) ) f_superieur = 0.75;

  if ( ( f_min < 0.0 ) || ( f_min > 1.0 ) ) f_min = 0.10;  /* Position des moustaches : 1er et 9e déciles */
  if ( ( f_max < 0.0 ) || ( f_max > 1.0 ) ) f_max = 0.90;

  /* Si besoin est : réservation du complément */
  infos = valeurs->affichage->complement.infos_bmoustache;
  if ( NULL ==  infos ) {
    infos = (struct boite_moustache *) malloc( sizeof( struct boite_moustache ) );
    if ( NULL == infos ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de réserver les infos / boîte à moustache\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -10;
    }

    valeurs->affichage->complement.infos_bmoustache = infos;

    /* Initialisation */
    infos->elements = BM_AVEC_MEDIANE/*  | BM_AVEC_EXTREMES *//*  | BM_AVEC_ABERRANTES */ | BM_AVEC_MOYENNE;
    infos->elements |= BM_AVEC_IC;  /* Pour EDs de 3e année, en attendant l'interface ! */

    infos->largeur = 40;	/* Largeur par défaut, en pixels */

    infos->henry = NULL;
  } else {
    GRF_eff_isole( infos->henry );
    infos->henry = NULL;
  }

  /* On fait le remplissage de cette structure */
  /* a. Valeurs faciles */
  infos->moyenne    = esperance( valeurs );
  infos->ecart_type = sqrt( variance( valeurs, infos->moyenne ) );

  STAT_Quartiles( valeurs, STAT_MEDIANE_USUELLE, quartiles );
  infos->mediane = quartiles[ 1 ];
  if ( 0.25 == f_inferieur ) {
    infos->q1 = quartiles[ 0 ];
  }
  if ( 0.75 == f_superieur ) {
    infos->q3 = quartiles[ 2 ];
  }

  /* b. On estime la fonction de répartition */
  f_repartition = STAT_FdREmpirique( valeurs, STAT_FDR_EMPIRIQUE_BASE, 0 );
  if ( NULL == f_repartition ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de calculer la fonction de répartition empirique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -25;
  }

  /* c. On inverse X & Y de la fonction de répartition : donnera donc Y=la valeur connaissant X=le fractile */
  retour = GRF_PermuterXY( f_repartition );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR en permuttant X et Y, code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* d. On l'utilise pour avoir les valeurs cherchées */
  if ( f_inferieur != 0.25 ) {
    infos->q1 = GRF_LireValeur( f_repartition, f_inferieur, INTERPOLE_LINEAIRE );
  }
  if ( f_superieur != 0.75 ) {
    infos->q3 = GRF_LireValeur( f_repartition, f_superieur, INTERPOLE_LINEAIRE );
  }
  infos->d1 = GRF_LireValeur( f_repartition, f_min, INTERPOLE_LINEAIRE );
  infos->d9 = GRF_LireValeur( f_repartition, f_max, INTERPOLE_LINEAIRE );

  /* e. Plus besoin de la fonction de répartition */
  GRF_eff_isole( f_repartition ); f_repartition = NULL;

  /* Terminé, sans souci */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Tracé proprement dit ~-~-~-~-~-~-~-~- */

int TRC_BoiteAMoustache(Fenetre *fenetre, InfosRegion2D *region, struct graphe *valeurs, struct boite_moustache *infos, int position)
{
  int retour = 45;

  if ( ( NULL == region ) || ( NULL == infos ) ) return -1;
  if ( NULL == valeurs ) return 0;  /* Rien à faire : rien à tracer ! */

  /* Choix de la direction de tracé */
  if ( infos->elements & BM_VERTICALE ) {
    OPTION_A_FAIRE( "Boîte à moustache verticale" );
  } else {
    retour = TRC_BM_Horizontale( fenetre, region, valeurs, infos, position );
  }

  if ( retour != 0 ) {
    AVERTISSEMENT( "Problème dans le tracé : code %d", retour );
  }
  return retour;
}

static int TRC_BM_Horizontale(Fenetre *fenetre, InfosRegion2D *region, struct graphe *valeurs, struct boite_moustache *infos, int position)
{
  double echelle_X;
  int x_vrai_debut_boite, x_vrai_fin_boite, x_vrai;

  /* On s'assure que le style de dessin sera correct [la couleur est définie extérieurement] */
  TRC_Style( BI_CONTINU );
  TRC_Epaisseur( 1 );

  /* On calcule le facteur d'échelle : axe horizontal */
  echelle_X = AXE_Echelle( &(region->axe_X), region->x_BG, region->x_BG + region->largeur - 1 );

  /* On trace le rectangle de la boîte */
  x_vrai_debut_boite = TRC_ConvertitX_PvE( infos->q1, region->axe_X.min, region->x_BG, echelle_X );
  x_vrai_fin_boite   = TRC_ConvertitX_PvE( infos->q3, region->axe_X.min, region->x_BG, echelle_X );
  TRC_Rectangle( fenetre, x_vrai_debut_boite, position - ( infos->largeur >> 1 ), x_vrai_fin_boite - x_vrai_debut_boite + 1, infos->largeur );

  /* On trace la moustache inférieure */
  x_vrai = TRC_ConvertitX_PvE( infos->d1, region->axe_X.min, region->x_BG, echelle_X );
  TRC_LigneH( fenetre, x_vrai, position, x_vrai_debut_boite - x_vrai + 1 );
  TRC_LigneV( fenetre, x_vrai, position - ( infos->largeur >> 1 ), infos->largeur );

  /* On trace la moustache supérieure */
  x_vrai = TRC_ConvertitX_PvE( infos->d9, region->axe_X.min, region->x_BG, echelle_X );
  TRC_LigneH( fenetre, x_vrai_fin_boite, position, x_vrai - x_vrai_fin_boite + 1 );
  TRC_LigneV( fenetre, x_vrai, position - ( infos->largeur >> 1 ), infos->largeur );

  /* Les éléments supplémentaires, facultatifs */
  if ( infos->elements & BM_AVEC_MEDIANE ) {
    x_vrai = TRC_ConvertitX_PvE( infos->mediane, region->axe_X.min, region->x_BG, echelle_X );
    TRC_LigneV( fenetre, x_vrai, position - ( infos->largeur >> 1 ), infos->largeur );
  }
  if ( infos->elements & BM_AVEC_EXTREMES ) {
    x_vrai = TRC_ConvertitX_PvE( valeurs->y_min, region->axe_X.min, region->x_BG, echelle_X );
    TRC_Marque( fenetre, x_vrai, position, BI_MARQUE_LOSANGE, 10 );

    x_vrai = TRC_ConvertitX_PvE( valeurs->y_max, region->axe_X.min, region->x_BG, echelle_X );
    TRC_Marque( fenetre, x_vrai, position, BI_MARQUE_LOSANGE, 10 );
  }
  if ( infos->elements & BM_AVEC_ABERRANTES ) {
    unsigned long i;

    for ( i = 0 ; i < valeurs->nombre_point ; i++ ) {
      if ( ( valeurs->y[ i ] > infos->limite_sup ) || ( valeurs->y[ i ] < infos->limite_inf ) ) {
	/* La valeur est aberrante ! */
	x_vrai = TRC_ConvertitX_PvE( valeurs->y[ i ], region->axe_X.min, region->x_BG, echelle_X );
	TRC_Marque( fenetre, x_vrai, position, BI_MARQUE_ASTERISQUE, 10 );
      }
    }
  }
  if ( infos->elements & BM_AVEC_MOYENNE ) {
    x_vrai = TRC_ConvertitX_PvE( infos->moyenne, region->axe_X.min, region->x_BG, echelle_X );
    TRC_Marque( fenetre, x_vrai, position, BI_MARQUE_CERCLE, 10 );
  }
  if ( infos->elements & BM_AVEC_IC ) {
    int x_min, x_max;

    x_min = TRC_ConvertitX_PvE( infos->moyenne - 2.0 * infos->ecart_type, region->axe_X.min, region->x_BG, echelle_X );
    x_max = TRC_ConvertitX_PvE( infos->moyenne + 2.0 * infos->ecart_type, region->axe_X.min, region->x_BG, echelle_X );

    /* Les deux "parenthèses" */
    TRC_ArcEllipse( fenetre, x_min + 10, position, 10, (infos->largeur >> 1),  90, 270 );    
    TRC_ArcEllipse( fenetre, x_max - 10, position, 10, (infos->largeur >> 1), -90,  90 );    

    /* Le trait de liaison entre les deux */
    TRC_Style( BI_TIRET_COURT );
    TRC_LigneH( fenetre, x_min, position, x_max - x_min + 1 );
  }

  /* Aucun problème de tracé */
  return 0;
}
