/* Bibliothque d'interface
   (C) 1998-2002, Emmanuel CURIS

   Fonctions generales de traitement des chaines de caractere
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h> /* Pour is... */
#include <math.h>  /* Pour pow */

#include "deboguage.h"
#include "math_utils.h" /* Pour Puissance */
#include "txt_utils.h"

#ifndef M_PI
# define M_PI 3.14159265358979323846
#endif
#ifndef CHARGE_ELECTRON
# define CHARGE_ELECTRON 1.6021892e-19
#endif

/* -------- Fonctions locales -------- */

static double TXT_AnalyserExposantUnite(const char *unite, int *numerateur, int *denominateur, char const **fin_exposant);
static int    TXT_AnalyserMonomeUnite  (const char *unite, double *facteur, char const **fin_monome);
static int    TXT_TrouverUniteSimple   (const char *unite, double *facteur);

static void TXT_UniteOterExposant(char *unite);

static int TXT_MultiplierParUnite(char ***table_unites, int **exposants, unsigned int n_u_simples, const char *unite, int exposant);

static unsigned long TXT_LongueurUnite(char *unite, int exposant);

/* -------- Une unit est-elle simple ? -------- *
    Analyse une unit pour savoir si elle est de la forme
       unit            ex : eV, kg, mm,...     -> VRAI
       unit^qqchose    ex : mm^2, m, s^-1,... -> VRAI
    ou produit d'units ex : m/s, kW.h,...      -> FAUX
 * --------------------------------------------- */

int TXT_UniteEstSimple(const char *unite, int *exposant)
{
  unsigned long i, pos_exposant = 0;

  if ( NULL == unite ) {
    if ( exposant != NULL ) *exposant = 0;
    return VRAI;
  }

  if ( NULL == exposant ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de variable pour l'exposant !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return FAUX;
  }
  *exposant = 1;		/* Valeur par dfaut */

  /* Cas spcial : 1/unite */
  if ( ( '1' == unite[ 0 ] ) && ( '/' == unite[ 1 ] ) ) {
    int est_simple;

    est_simple = TXT_UniteEstSimple( unite + 2, exposant );
    if ( VRAI == est_simple ) {
      *exposant *= -1;
    }

    return est_simple;
  }

  /* On cherche un caractre de produit d'units : / . *  ou  */
  for ( i = 0 ; unite[ i ] != 0 ; i++ ) {
    if ( '/' == unite[ i ] ) return FAUX; /* Quotient, type m/s */
    if ( '' == unite[ i ] ) return FAUX; /* Produit, type ms */
    if ( '.' == unite[ i ] ) return FAUX; /* Produit, type m.s [. admis en remplacement de ] */
    if ( '*' == unite[ i ] ) return FAUX; /* Produit, type m*s [* admis mais incorrect] */
    if ( '' == unite[ i ] ) return FAUX; /* Produit, type ms [ admis mais incorrect] */
    if ( ( '^' == unite[ i ] ) || ( '' == unite[ i ] ) || ( '' == unite[ i ] ) ) {
      pos_exposant = i;		/* Exposant, type m^2 : on indique qu'il y a un exposant ! */
    }
  }

  /* Si on est l : unit simple, reste  s'occuper de l'exposant */
  if ( 0 == pos_exposant ) {	/* Une unit ne devrait pas pouvoir commencer par un exposant => pas d'exposant en ralit */
    return VRAI;
  }

  /* Exposants spciaux */
  if ( '' == unite[ pos_exposant ] ) { /* ^2 [on ne devrait pas rencontrer du ^22, ^23 ou aussi grand !] */
    *exposant = 2;
  } else if ( '' == unite[ pos_exposant ] ) { /* ^3 [on ne devrait pas rencontrer du ^22, ^23 ou aussi grand !] */
    *exposant = 3;
  } else {
    *exposant = (int) TXT_CNV_vNombre_L( unite + pos_exposant + 1, NULL );
  }

  /* Analyse termine, unit simple */
  return VRAI;
}

/* -------- Analyse le nom d'une unit et le simplifie --------
   Renvoie : le facteur multiplicatif pour passer  l'unit normalise
             le type de grandeur sur lequel on travaille
             la nouvelle unit aprs simplification

 unite            : l'unit  normaliser
 complete         : VRAI ==> simplification complte vers SI ( eV -> J p. ex.)
                    FAUX ==> simplification des multiples    (keV -> eV)
 type_unite       : le type de grandeur manipul
 unite_simplifiee : si non-NULL, contient l'unit aprs normalisation
 ---------------------------------------------------------------------- */

double TXT_NormaliserUnite(const char *unite, int complete __attribute__((__unused__)), int *type_unite, char *unite_simplifiee __attribute__((__unused__)))
{
  const char *pos_suite;
  double facteur = 0.0;
  int type;

  if ( NULL == unite ) {
    *type_unite = UNITE_AUCUNE;
    return 1.0; /* Sans unit */
  }

  while ( ' ' == *unite ) unite++;
  if ( 0 == *unite ) {
    *type_unite = UNITE_COMPTAGE;
    return 1.0; /* Sans unit */
  }

  /* Analyse progressive. Les parenthses ne sont autorises que pour encadrer les exposants ! */
  /* a. On cherche la fin de l'unit lmentaire de dbut */
  type = TXT_AnalyserMonomeUnite( unite, &facteur, &pos_suite );

  /* b. Si pos_suite != 0, c'est qu'il y a une opration => on analyse les termes suivants */
  while ( *pos_suite != 0 ) {
    double facteur_monome = 1.0;

    if ( ( '.' == *pos_suite ) || ( '' == *pos_suite ) ) { /* . : produit d'unit, m.s^-1 p. ex. */
      (void) TXT_AnalyserMonomeUnite( pos_suite + 1, &facteur_monome, &pos_suite );
      facteur *= facteur_monome;
      *type_unite = UNITE_AUTRE; /* Unit complexe : pas encore analyse*/
    } else if ( '/' == *pos_suite ) { /* / : quotient d'units, m/s p. ex. */
      (void) TXT_AnalyserMonomeUnite( pos_suite + 1, &facteur_monome, &pos_suite );
      if ( facteur_monome != 0.0 ) {
	facteur /= facteur_monome;
	*type_unite = UNITE_AUTRE; /* Unit complexe : pas encore analyse*/
      }
    } else {
      (void) fprintf(stderr, "%s, %s l.%u : Caractre interdit dans une unit [%s / %c] ignor\n",
		     __FILE__, __FUNCTION__, __LINE__, unite, *pos_suite);
      pos_suite++;
    }
  }

  /* Analyse termine */
  if ( type_unite != NULL ) *type_unite = type;
  return facteur;
}

/* -------- Analyse l'exposant d'une unit pour le simplifier -------- */

static double TXT_AnalyserExposantUnite(const char *unite, int *numerateur, int *denominateur, char const **fin_exposant)
{
  const char *pos_suite;
  int av_parenthese = FAUX;
  double exposant;

  if ( '(' == *unite ) {
    unite++;
    av_parenthese = VRAI;
  }
  *numerateur = TXT_CNV_vNombre_L( unite, &pos_suite );
  exposant = (double) *numerateur;
  if ( '/' == *pos_suite ) { /* Exposant fractionnaire ? */
    const char *pos_fin;

    *denominateur = TXT_CNV_vNombre_L( pos_suite + 1, &pos_fin );
    if ( *denominateur != 0 ) { /* Il existe bien : on avait bien un exposant fractionnaire */
      exposant /= (double) *denominateur;
      pos_suite = pos_fin; /* On est bien  la fin de l'exposant */
    }
  } else {
    *denominateur = 0;
  }

  /* Fermeture ventuelle de la parenthse */
  if ( VRAI == av_parenthese ) {
    if ( ')' == *pos_suite ) pos_suite++;
    else {
      (void) fprintf(stderr, "%s, %s l.%u : Parenthse non ferme dans l'exposant [%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, unite);
      /* Erreur mineure, on fait comme si la parenthse tait ferme... */
    }
  }

  /* On donne la position de la suite : 1er caractre non-analys */
  if ( fin_exposant != NULL ) *fin_exposant = pos_suite;

  /* Exposant calcul, calcul fini */
  return exposant;
}

/* -------- Analyser un monme : m, m, coups, eV, s^-1, M^1/3,...
     Renvoi le type de l'unit simple dtecte, le facteur de conversion vers l'unit interne correspondante
   -------- */

static int TXT_AnalyserMonomeUnite(const char *unite, double *facteur, char const **fin_monome)
{
  const char *fin;
  char *symbole;
  double fact_unite = 1.0;
  int type;

  /* Units  une lettre : dtection immdiate */
  if ( 0 == *(unite + 1) ) {
    *fin_monome = unite + 1;
    return TXT_TrouverUniteSimple(unite, facteur);
  }

  /* On cherche la fin : premier caractre non-alphabtique */
  fin = unite;
  *fin_monome = fin;
  while ( isalpha( *fin ) || ( '' == *fin ) || ( '' == *fin ) ) {
    fin++;
  }
  if ( fin == unite ) { /* Unit nulle : problme */
    (void) fprintf(stderr, "%s, %s l.%u : Unit non-alphanumrique, ignore [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, unite);
    return 0.0;
  }

  /* On isole le symbole proprement dit */
  symbole = (char *) malloc( fin - unite + 1 ); /* 0 final... */
  if ( NULL == symbole ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'isoler le monome [%s - %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, unite, fin);
    return 0.0;
  }
  strncpy( symbole, unite, fin - unite );
  symbole[ fin - unite ] = 0;

  /* Quel est le facteur entre cette unit et l'unit de rfrence */
  type = TXT_TrouverUniteSimple( symbole, &fact_unite );

  free( symbole ); symbole = NULL;

  /* On cherche un exposant ventuel : par quoi a-t-on fini ? */
  if ( '' == *fin ) { /* m p. ex. */
    fin++; /* l'exposant est lu */

    *facteur = fact_unite * fact_unite;
    if ( UNITE_LONGUEUR == type ) type = UNITE_SURFACE;
  } else if ( '' == *fin ) { /* m p. ex. */
    fin++; /* l'exposant est lu */

    *facteur = fact_unite * fact_unite * fact_unite;
    if ( UNITE_LONGUEUR == type ) type = UNITE_VOLUME;
  } else if ( '^' == *fin ) { /* exposant  lire */
    int numerateur, denominateur;
    double exposant;

    fin++;
    exposant = TXT_AnalyserExposantUnite( fin, &numerateur, &denominateur, &fin );
/*     printf("Exposant trouv : %s [%s] => %d/%d=%g\n", symbole, fin, numerateur, denominateur, exposant ); */
    *facteur = 1.0;
    if ( 0 == denominateur ) { /* Exposant entier : cas le plus frquent */
      if ( numerateur > 0 ) {
	*facteur = Puissance( fact_unite, (unsigned int) numerateur );
      } else if ( 0 == numerateur ) {
	(void) fprintf(stderr, "%s, %s l.%u : Exposant absurde [%s]\n",
		       __FILE__, __FUNCTION__, __LINE__, unite);
      } else {
	if ( fact_unite != 0.0 ) {
	  *facteur = Puissance( 1.0 / fact_unite, (unsigned int) -numerateur );
	}
      }

      /* Contrles de type */
      if ( 2 == numerateur ) {
	if ( UNITE_LONGUEUR == type ) {
	  type = UNITE_SURFACE;
	} else {
	  type = UNITE_AUTRE;
	}
      } else if ( 3 == numerateur ) {
	if ( UNITE_LONGUEUR == type ) {
	  type = UNITE_VOLUME;
	} else {
	  type = UNITE_AUTRE;
	}
      } else if ( -1 == numerateur ) {
	if ( UNITE_LONGUEUR == type ) {
	  type = UNITE_VERGENCE;
	} else if ( UNITE_TEMPS == type ) {
	  type = UNITE_PULSATION;
	} else {
	  type = UNITE_AUTRE;
	}
      }
    } else {
      if ( fact_unite != 0.0 ) {
	*facteur = pow( fact_unite, exposant );
      }
    }

  } else {
    *facteur = fact_unite;
  }

  /* On indique la fin retenue */
  *fin_monome = fin;

  /* On a termin l'analyse de ce monome */
  return type;
}

/* -------- Dtecte une unit simple et ses multiples -------- */

static int TXT_TrouverUniteSimple(const char *unite, double *facteur)
{
  double multiple = 1.0;
  int type = UNITE_COMPTAGE;

  /* Units  une lettre : comparaison directe  une table */
  if ( 0 == *(unite + 1) ) {
    switch ( *unite )
    {
      /* Units SI */
    case 'A': *facteur = 1.0; return UNITE_INTENSITE;/* ampre */
    case 'C': *facteur = 1.0; return UNITE_CHARGE; /* coulomb */
    case 'F': *facteur = 1.0; return UNITE_CAPACITE; /* farad */
    case 'g': *facteur = 1.0; return UNITE_MASSE;    /* gramme. On pourrait avoir g = acclration, mais priorit  la masse ! */
    case 'H': *facteur = 1.0; return UNITE_INDUCTANCE; /* henry */
    case 'J': *facteur = 1.0; return UNITE_ENERGIE; /* joule */
    case 'K': *facteur = 1.0; return UNITE_TEMPERATURE; /* kelvin */
    case 'm': *facteur = 1.0; return UNITE_LONGUEUR; /* mtre */
    case 'N': *facteur = 1.0; return UNITE_FORCE; /* newton */
    case 'S': *facteur = 1.0; return UNITE_CONDUCTANCE; /* siemens */
    case 's': *facteur = 1.0; return UNITE_TEMPS;    /* seconde */
    case 'T': *facteur = 1.0; return UNITE_MAGNETIQUE; /* tesla */
    case 'V': *facteur = 1.0; return UNITE_TENSION; /* volt */
    case 'W': *facteur = 1.0; return UNITE_PUISSANCE; /* watt */

      /* quivalences pour les units SI avec symbole grec */
    case 'd': *facteur = 1.0; return UNITE_VERGENCE; /* dioptrie : delta */
    case 'O': *facteur = 1.0; return UNITE_RESISTANCE; /* ohm : omga majuscule */

      /* Units courantes */
    case 'a': *facteur = 100.0;   return UNITE_SURFACE; /* are. 1 a = 100 m */
    case 'h': *facteur = 3600.0;  return UNITE_TEMPS; /* heure. 1 h = 3600 s */
    case 'j': *facteur = 86400.0; return UNITE_TEMPS; /* jour. 1 j = 24 h = 86400 s */
    case 'l': *facteur = 1.0e-3;  return UNITE_VOLUME;/* litre. 1 l = 1 dm */
    case 'L': *facteur = 1.0e-3;  return UNITE_VOLUME;/* litre. 1 L = 1 dm */
    case 'M': *facteur = 1.0e-3;  return UNITE_CONCENTRATION;/* molaire. 1 M = 1 mol/L */
    case 't': *facteur = 1.0e6;   return UNITE_MASSE; /* tonne mtrique. 1 t = 1000 kg */
    case 'u': *facteur = 1.66056e-24;return UNITE_MASSE; /* unit de masse atomique. 1 u = 1,66056e-27 kg */
    case '': *facteur = 1.0e-10; return UNITE_LONGUEUR; /* angstrm. 1  = 10^-10 m */
    case '': *facteur = 1.0e-6;  return UNITE_LONGUEUR; /* micron. 1  = 1 m */
    case '': *facteur = M_PI / 180.0; return UNITE_ANGLE; /* degr. 180 = pi radians */
    case '\'': *facteur = M_PI / 10800.0; return UNITE_ANGLE; /* minute d'angle. 60' = 1 */
    case '"': *facteur = M_PI / 648000.0; return UNITE_ANGLE; /* seconde d'angle. 60" = 1' */
    }

    (void) fprintf(stderr, "%s : unit inconnue, %s\n", __FUNCTION__, unite);
    *facteur = 1.0;
    return UNITE_COMPTAGE; /* Unit arbitraire non-reconnue */
  }

  /* Cas spcial : units  une lettre pouvant etre inaccessibles au clavier */
  if ( '\\' == unite[ 0 ] ) {
    if ( 'A' == unite[ 1 ] ) {	      /* \A = angstroem */
      *facteur = 1.0e-10; return UNITE_LONGUEUR;
    } else if ( 'm' == unite[ 1 ] ) { /* \m =  */
      if ( 0 == unite[ 2 ] ) {
	*facteur = 1.0e-6;  return UNITE_LONGUEUR;
      } else {
	type = TXT_TrouverUniteSimple( unite + 1, facteur );
	*facteur *= 1.0e-6;

	return type;
      }
    } else if ( 'O' == unite[ 1 ] ) { /* \O = Ohm */
      *facteur = 1.0; return UNITE_RESISTANCE;
    }
  }

  /* On compare aux units "longues" connues */
  /* Units SI */
  if ( 0 == strcmp( unite, "Bq" ) ) {
    *facteur = 1.0; return UNITE_ACTIVITE; /* becquerel */
  } else if ( 0 == strcmp( unite, "cd" ) ) {
    *facteur = 1.0; return UNITE_I_LUMIERE; /* candela */
  } else if ( 0 == strcmp( unite, "Gy" ) ) {
    *facteur = 1.0; return UNITE_DOSE; /* gray */
  } else if ( 0 == strcmp( unite, "Hz" ) ) {
    *facteur = 1.0; return UNITE_FREQUENCE; /* hertz */
  } else if ( 0 == strcmp( unite, "lm" ) ) {
    *facteur = 1.0; return UNITE_FLUX_LUM; /* lumen */
  } else if ( 0 == strcmp( unite, "Lx" ) ) {
    *facteur = 1.0; return UNITE_ECLAIREMENT; /* lux */
  } else if ( 0 == strcmp( unite, "m" ) ) {
    *facteur = 1.0; return UNITE_SURFACE; /* cas spcial, pour m^2 */
  } else if ( 0 == strcmp( unite, "m" ) ) {
    *facteur = 1.0; return UNITE_VOLUME; /* cas spcial, pour m^3 */
  } else if ( 0 == strcmp( unite, "mol" ) ) {
    *facteur = 1.0; return UNITE_QUANTITE; /* mole */
  } else if ( 0 == strcmp( unite, "Nt" ) ) {
    *facteur = 1.0; return UNITE_LUMINANCE; /* nit */
  } else if ( 0 == strcmp( unite, "Pa" ) ) {
    *facteur = 1.0; return UNITE_PRESSION; /* pascal */
  } else if ( 0 == strcmp( unite, "rad" ) ) {
    *facteur = 1.0; return UNITE_ANGLE; /* radians */
  } else if ( 0 == strcmp( unite, "sr" ) ) {
    *facteur = 1.0; return UNITE_ANGLE_SOLIDE; /* stradians */
  } else if ( 0 == strcmp( unite, "Sv" ) ) {
    *facteur = 1.0; return UNITE_EQUIVALENT_DOSE; /* sievert */
  } else if ( 0 == strcmp( unite, "Wb" ) ) {
    *facteur = 1.0; return UNITE_FLUX_MAG; /* weber */
  }

  /* Units non-SI traditionnelles (reconnues officiellement) */
  if ( 0 == strcmp( unite, "Ah" ) ) {
    *facteur = 3600.0; return UNITE_CHARGE; /* ampre-heure. 1 Ah = 3600 C */
  } else if ( 0 == strcmp( unite, "atm" ) ) {
    *facteur = 1.01325e5; return UNITE_PRESSION; /* atmosphre. 1 atm = 1.01325 bar */
  } else if ( 0 == strcmp( unite, "bar" ) ) {
    *facteur = 1.0e5; return UNITE_PRESSION; /* bar. 1 bar = 10^5 Pa */
  } else if ( 0 == strcmp( unite, "barn" ) ) {
    *facteur = 1.0e-28; return UNITE_SECTION_EFFICACE; /* barn. 1 barn = 10^-28 m */
  } else if ( 0 == strcmp( unite, "eV" ) ) {
    *facteur = CHARGE_ELECTRON; return UNITE_ENERGIE; /* lectron-volt. 1 eV = 1,610^-19 J */
  } else if ( 0 == strcmp( unite, "Gal" ) ) {
    *facteur = 0.01; return UNITE_ACCELERATION; /* gal. 1 Gal = 0,01 m/s */
  } else if ( 0 == strcmp( unite, "gon" ) ) {
    *facteur = M_PI / 200.0; return UNITE_ANGLE; /* grade. 100 gon = 90 */
  } else if ( 0 == strcmp( unite, "mille" ) ) {
    *facteur = 1852.0; return UNITE_LONGUEUR; /* mille marin. 1 mille = 1852 m */
  } else if ( 0 == strcmp( unite, "min" ) ) {
    *facteur = 60.0; return UNITE_TEMPS; /* minute. 1 min = 60 s */
  } else if ( ( 0 == strcmp( unite, "mmHg" ) ) || ( 0 == strcmp( unite, "mm Hg" ) ) ) {
    *facteur = 133.332; return UNITE_PRESSION; /* millimtre de mercure. 1 mm Hg = 133,332 Pa */
  } else if ( 0 == strcmp( unite, "noeud" ) ) {
    *facteur = 1852.0/3600.0; return UNITE_VITESSE; /* noeud. 1 noeud = 1 mille marin / heure */
  } else if ( 0 == strcmp( unite, "tex" ) ) {
    *facteur = 1.0e-6; return UNITE_MASSE_LINEIQUE; /* tex. 1 tex = 1e-6 kg/m */
  } else if ( 0 == strcmp( unite, "tr" ) ) {
    *facteur = 2.0 * M_PI; return UNITE_ANGLE; /* tour. 1 tour = 2 pi radians */
  } else if ( 0 == strcmp( unite, "Wh" ) ) {
    *facteur = 3600.0; return UNITE_ENERGIE; /* wattheure. 1 Wh = 3600 J */
  } else if ( 0 == strcmp( unite, "C" ) ) {
    *facteur = 0.0; return UNITE_TEMPERATURE; /* degr Celsius. Attention, la conversion n'est PAS multiplicative ! */
  }

  /* Si on est l : on essaye les multiples et sous-multiples, avec 1re lettre = code de sous-multiple */
  switch ( *unite )
  {
  case 'Y':
    multiple = 1.0e+24;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* Y = yotta = 10^24 */

  case 'Z':
    multiple = 1.0e+21;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* Z = zetta = 10^21 */

  case 'E':
    multiple = 1.0e+18;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* E = exa   = 10^18 */

  case 'P':
    multiple = 1.0e+15;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* P = peta  = 10^15 */

  case 'T':
    multiple = 1.0e+12;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* T = tra  = 10^12 */

  case 'G':
    multiple = 1.0e+9;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* G = giga  = 10^9 */

  case 'M':
    multiple = 1.0e+6;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* M = mga  = 10^6 */

  case 'k':
    multiple = 1.0e+3;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* k = kilo  = 10^3 */

  case 'h':
    multiple = 1.0e+2;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* h = hecto = 10^2 */

    /* Remarque : le dca n'est pas dtect ainsi, car prfixe da = deux caractres... */

  case 'd':
    multiple = 1.0e-1;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* d = dci  = 10^-1 */

  case 'c':
    multiple = 1.0e-2;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* c = centi = 10^-2 */

  case 'm':
    multiple = 1.0e-3;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* m = milli = 10^-3 */

  case '':
    multiple = 1.0e-6;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /*  = micro = 10^-6 */

  case 'n':
    multiple = 1.0e-9;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* n = nano  = 10^-9 */

  case 'p':
    multiple = 1.0e-12;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* p = pico  = 10^-12 */

  case 'f':
    multiple = 1.0e-15;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* f = femto = 10^-15 */

  case 'a':
    multiple = 1.0e-18;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* a = atto  = 10^-18 */

  case 'z':
    multiple = 1.0e-21;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* n = zepto = 10^-21 */

  case 'y':
    multiple = 1.0e-24;
    type = TXT_TrouverUniteSimple( unite + 1, facteur );
    break; /* y = yocto = 10^-24 */
  }

  /* Analyse termine */
  *facteur *= multiple;
  return type;
}

/* ~-~-~-~-~-~-~-~-~ Oprations pour modifier une unit intelligemment ~-~-~-~-~-~-~-~-~ */

/* -------- Unit rsultant d'un produit d'units ------------------------------------ *
     Renvoie la nouvelle unit [NULL en cas d'erreur]
             cette nouvelle unit est  librer avec free()

  unite_avant   : l'unit avant produit. NULL = sans unit
                  elle peut etre statique, une copie est toujours faite avec malloc()
		  => librer l'ancienne ventuellement !
  unite_produit : l'unit par laquelle on multiplie
  exposant      : exposant de cette unit
 * ----------------------------------------------------------------------------------- */

char *TXT_ProduitUnites(const char *unite_avant, const char *unite_produit, int exposant)
{
  char *unite_finale;
  char **table_produit;
  int *exposants_produit;
  unsigned int i, nbr_termes_produit, nbr_termes_ajoutes = 0;

  /* Contrles */
  if ( NULL == unite_produit ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune unit fournie !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Cas o il n'y a pas de produit  faire */
  if ( ( 0 == exposant ) || ( 0 == strlen( unite_produit ) ) ) {
    if ( unite_avant != NULL ) {
      /* On doit faire une copie pour viter des problmes de gestion mmoire */
      return strdup( unite_avant );
    } else {
      return NULL; /* Rsultat sans unit */
    }
  }
  if ( ( NULL == unite_avant ) && ( 1 == exposant ) ) {
    return strdup( unite_produit );
  }

  /* On dcompose l'unit produit en table d'units simples */
  nbr_termes_produit = TXT_DecomposerUnite( unite_produit, &table_produit, &exposants_produit );
  if ( 0 == nbr_termes_produit ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en dcomposant l'unit produit\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On applique l'exposant global */
  TXT_PuissanceUnite( table_produit, exposants_produit, nbr_termes_produit, exposant );

  if ( NULL != unite_avant ) {
    char **table_avant;
    int *exposants_avant, retour;
    unsigned int nbr_termes_avant;

    /* On dcompose l'unit d'avant en table de monomes */
    nbr_termes_avant = TXT_DecomposerUnite( unite_avant, &table_avant, &exposants_avant );
    if ( 0 == nbr_termes_avant ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR en dcomposant l'unit avant\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* On fait le produit terme  terme */
    for ( i = 0 ; i < nbr_termes_avant ; i++ ) {
      retour = TXT_MultiplierParUnite( &table_produit, &exposants_produit, nbr_termes_produit + nbr_termes_ajoutes, table_avant[ i ], exposants_avant[ i ] );
      if ( retour < 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR en faisant le produit [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	break;
      }
      nbr_termes_ajoutes += retour;
    }

    /* Plus besoin des tables d'avant */
    free( table_avant[ 0 ] ); table_avant[ 0 ] = NULL;
    free( table_avant ); table_avant = NULL;
    free( exposants_avant ); exposants_avant = NULL;
  }

  /* On recompose l'unit finale */
  unite_finale = TXT_RecomposerUnite( table_produit, exposants_produit, nbr_termes_produit + nbr_termes_ajoutes );

  /* On libre les tables */
  free( exposants_produit ); exposants_produit = NULL;
  free( table_produit[ 0 ] ); table_produit[ 0 ] = NULL;
  /* Attention, les units simples ajoutes en fin de table le sont avec strdup()
     alors que celles d'origine sont en un bloc dans [0] !*/
  for ( i = 0 ; i < nbr_termes_ajoutes ; i++ ) {
    free( table_produit[ nbr_termes_produit + i ] );
  }
  free( table_produit ); table_produit = NULL;

  /* Termin */
  return unite_finale;
}

/* -------- Convertir une table d'units simples en unit complexe -------- */

char *TXT_RecomposerUnite(char **unites_simples, int *exposants, unsigned int nbr_termes)
{
  unsigned int i;
  unsigned long lg_unite;
  char *unite_finale, *pos_ecriture;

  if ( 0 == nbr_termes ) return NULL;

  /* On calcule la place occupe par le texte de l'unit */
  if ( 1 == nbr_termes ) {
    lg_unite = TXT_LongueurUnite( unites_simples[ 0 ], exposants[ 0 ] );
  } else {
    unsigned long lg_terme;
    lg_unite = 0;

    for ( i = 0 ; i < nbr_termes ; i++ ) {
      /* Place de chaque unit individuelle */
      lg_terme = TXT_LongueurUnite( unites_simples[ i ], exposants[ i ] );
      if ( lg_terme > 0 ) {
	lg_unite += lg_terme + 1; /* Symbole multiplicatif */
      }
    }
    if ( lg_unite > 0 )
      lg_unite--;		/* Symbole multiplicatif final inutile */
  }

  /* Y a-t-il quelque chose  faire ? */
  if ( 0 == lg_unite ) return NULL;

  /* On rserve la place */
  unite_finale = (char *) malloc( sizeof( char ) * ( lg_unite + 1 ) ); /* NPO le 0 final */
  if ( NULL == unite_finale ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On crit */
  pos_ecriture = unite_finale;
  for ( i = 0 ; i < nbr_termes ; i++ ) {
    unsigned long lg_terme;

    if ( 0 == exposants[ i ] ) continue; /* Terme  ignorer */

    /* Longueur du terme  recopier [sans l'exposant] */
    lg_terme = TXT_LongueurUnite( unites_simples[ i ], 1 );

    /* On recopie l'unit simple */
    (void) strncpy( pos_ecriture, unites_simples[ i ], lg_terme );
    pos_ecriture += lg_terme;

    /* On met l'exposant */
    if ( 2 == exposants[ i ] ) {
      *pos_ecriture = '';
      pos_ecriture++;
    } else if ( 3 == exposants[ i ] ) {
      *pos_ecriture = '';
      pos_ecriture++;
    } else if ( exposants[ i ] != 1 ) {
      unsigned int lg_restante;
      *pos_ecriture = '^';
      pos_ecriture++;

      lg_restante = lg_unite + 1 - (pos_ecriture - unite_finale);

      TXT_CNV_L_vTexte( (long) exposants[ i ], pos_ecriture, &lg_restante );
      pos_ecriture += TXT_nbr_car_I( exposants[ i ] );
    }

    if ( i < (nbr_termes - 1) ) {
      *pos_ecriture = '';
      pos_ecriture++;
    }
  }

  *pos_ecriture = 0;

  /* Cration de la chaine termine */
  return unite_finale;
}

/* -------- Quelle est la longueur de la chaine pouvant contenir l'unit ? -------- */

static unsigned long TXT_LongueurUnite(char *unite, int exposant)
{
  unsigned long lg_unite;

  if ( ( NULL == unite ) || ( 0 == exposant ) ) return 0;

  lg_unite = strlen( unite ); /* unit elle-meme */
  if ( ( 2 == exposant ) || ( 3 == exposant ) ) {
    lg_unite++;		/*  et  */
  } else if ( exposant != 1 ) {
    lg_unite++;		/* ^ */
    lg_unite += TXT_nbr_car_I( exposant ); /* nombre */
  }

  return lg_unite;
}

/* -------- Dcomposer une unit en chacune des units produit -------- */

unsigned int TXT_DecomposerUnite(const char *unite, char ***monomes, int **exposants)
{
  int exp_total;
  unsigned int nbr_termes;
  char **table_monomes = NULL;
  unsigned long i;

  if ( NULL == unite ) return 0; /* Aucun terme */

  /* L'unit est-elle simple ? */
  if ( VRAI == TXT_UniteEstSimple( unite, &exp_total ) ) {
    nbr_termes = 1;
  } else {
    /* On compte le nombre de termes */
    nbr_termes = 0;

    /* On cherche un caractre de produit d'units : / . *  ou 
       ATTENTION, l'espace ne compte pas pour ne pas diviser des units comme "toise de Paris" */
    for ( i = 0 ; unite[ i ] != 0 ; i++ ) {
      if ( '/' == unite[ i ] ) nbr_termes++; /* Quotient, type m/s */
      if ( '' == unite[ i ] ) nbr_termes++; /* Produit, type ms */
      if ( '.' == unite[ i ] ) nbr_termes++; /* Produit, type m.s [. admis en remplacement de ] */
      if ( '*' == unite[ i ] ) nbr_termes++; /* Produit, type m*s [* admis mais incorrect] */
      if ( '' == unite[ i ] ) nbr_termes++; /* Produit, type ms [ admis mais incorrect] */
    }
  }

  /* On rserve les deux tables */
  table_monomes = (char **) malloc( sizeof( char * ) * nbr_termes );
  if ( NULL == table_monomes ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la table des monomes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

  *exposants = (int *) malloc( sizeof( int ) * nbr_termes );
  if ( NULL == *exposants ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la table des monomes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( table_monomes ); table_monomes = NULL;
    return 0;
  }

  /* On remplit les tables */
  table_monomes[ 0 ] = strdup( unite );
  if ( 1 == nbr_termes ) {
    *exposants[ 0 ] = exp_total;
  } else {
    unsigned int i_monome;
    int numerateur, denominateur;

    for ( i = 0, i_monome = 1 ; unite[ i ] != 0 ; i++ ) {
      if ( ( '/' == unite[ i ] ) || /* Quotient */
	   ( '' == unite[ i ] ) || ( '.' == unite[ i ] ) || /* Produit */
	   ( '*' == unite[ i ] ) || ( '' == unite[ i ] ) ) {
	table_monomes[ 0 ][ i ] = 0;
	table_monomes[ i_monome ] = table_monomes[ 0 ] + i + 1;

	/* On cherche l'exposant pour l'unit que l'on vient de finir */
	(void) TXT_AnalyserExposantUnite( table_monomes[ i_monome - 1 ], &numerateur, &denominateur, NULL );
	*exposants[ i_monome - 1 ] = numerateur;
	if ( denominateur != 1 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ATTENTION, dnominateur de l'unit ignor [%d, %s]\n",
			 __FILE__, __FUNCTION__, __LINE__, denominateur, table_monomes[ i_monome - 1 ]);
	}

	/* On remplira la suivant aprs */
	i_monome++;
      }
    }

    /* On cherche l'exposant pour la dernire unit simple */
    (void) TXT_AnalyserExposantUnite( table_monomes[ nbr_termes - 1 ], &numerateur, &denominateur, NULL );
    *exposants[ nbr_termes - 1 ] = numerateur;
    if ( denominateur != 1 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ATTENTION, dnominateur de l'unit ignor [%d, %s]\n",
		     __FILE__, __FUNCTION__, __LINE__, denominateur, table_monomes[ i_monome - 1 ]);
    }
  }

  /* On enlve les exposants dans les units simples */
  for ( i = 0 ; i < nbr_termes ; i++ ) {
    TXT_UniteOterExposant( table_monomes[ i ] );
  }

  /* La table des units simples est prete */
  *monomes = table_monomes;

  /* On a termin */
  return nbr_termes;
}

/* -------- Oter la partie de l'unit SIMPLE correspondant  l'exposant -------- */

static void TXT_UniteOterExposant(char *unite)
{
  for ( ; *unite != 0 ; unite++ ) {
    if ( '^' == *unite ) {
      *unite = 0;
      return;
    }

    if ( '' == *unite ) {
      *unite = 0;
      return;
    }

    if ( '' == *unite ) {
      *unite = 0;
      return;
    }
  }
}

/* -------- Mettre une unit, dcompose en units simples,  une puissance -------- */

int TXT_PuissanceUnite(char **u_simples, int *exposants, unsigned int n_termes, int exposant)
{
  unsigned int i;

  if ( 1 == exposant ) return 0; /* Rien  faire */

  if ( ( NULL == u_simples ) || ( NULL == exposants ) ) return -1;

  for ( i = 0 ; i < n_termes ; i++ ) {
    exposants[ i ] *= exposant;
  }

  /* Termin */
  return 0;
}

/* -------- Multiplier intelligemment par une unit --------
    Renvoie le nombre de termes ajouts dans la table [<0 : erreur]
 */

static int TXT_MultiplierParUnite(char ***table_unites, int **exposants, unsigned int n_u_simples, const char *unite, int exposant)
{
  unsigned int i;
  char **nv_table_unites;
  int *nv_table_exposants;

  /* On commence par reprer si l'unit existe dj dans la table */
  for ( i = 0 ; i < n_u_simples ; i++ ) {
    if ( 0 == strcmp( *table_unites[ i ], unite ) ) {
      /* L'unit existe dj : il suffit de modifier l'exposant */
      *exposants[ i ] += exposant;

      return 0;
    }
  }

  /* Il faut un nouveau terme dans le produit */
  nv_table_unites = (char **) realloc( *table_unites, sizeof(char *) * ( n_u_simples + 1 ) );
  if ( NULL == nv_table_unites ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de rserver pour la nouvelle unit [%s^%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, unite, exposant);
    return -1;
  }
  *table_unites = nv_table_unites;

  nv_table_exposants = (int *) realloc( *exposants, sizeof(int) * ( n_u_simples + 1 ) );
  if ( NULL == nv_table_exposants ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de rserver pour la nouvelle unit [%s^%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, unite, exposant);
    return -2;
  }
  *exposants = nv_table_exposants;

  /* On stocke la nouvelle unit et son exposant */
  nv_table_unites[ n_u_simples ]    = strdup( unite );
  nv_table_exposants[ n_u_simples ] = exposant;

  /* On a ajout un terme */
  return +1;
}
