/* Fonctions utilitaires gnrales
   Dveloppes pour LASE, mais plus gnriques
   (C) 1996 - 2003, Emmanuel CURIS

   Conversion symbole chimique -> numro atomique corresponddant
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "utils_3d.h"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void ProduitMatriciel(double *matrice, Vecteur3D *vecteur);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Oprations gomtriques de base ~-~-~-~-~-~-~-~- */

/* -------- De quelle direction est le plus proche le vecteur ? -------- */

inline int U3D_DirectionPreferentielle(Vecteur3D *vecteur)
{
  int retour;
  double c_max;

  if ( NULL == vecteur ) return -1;

  retour = U3D_PROPAGATION_SELON_X; /* Axe des X par dfaut */
  c_max = fabs( vecteur->x );

  if ( fabs( vecteur->y ) > c_max ) {
    retour = U3D_PROPAGATION_SELON_Y;
    c_max = fabs( vecteur->y );
  }

  if ( fabs( vecteur->z ) > c_max ) {
    retour = U3D_PROPAGATION_SELON_Z;
    c_max = fabs( vecteur->z );
  }

  if ( c_max == 0.0 ) { /* Vecteur nul */
    retour = U3D_PROPAGATION_NUL;
  }

  return retour;
}

/* -------- Calcul d'une distance [repre absolu] -------- */

inline double U3D_Distance(Vecteur3D *point_1, Vecteur3D *point_2)
{
  double dx, dy, dz;
  double d;

  if ( ( NULL == point_1 ) || ( NULL == point_2 ) ) return 0.0;

  dx = point_2->x - point_1->x;
  dy = point_2->y - point_1->y;
  dz = point_2->z - point_1->z;

  d = dx * dx + dy * dy + dz * dz;

  return sqrt( d );
}

/* -------- Normalisation d'un vecteur -------- */

inline int U3D_Normalise(Vecteur3D *vecteur, Vecteur3D_Erreurs *var_vecteur)
{
  double norme;

  if ( NULL == vecteur ) return -1;

  /* Comme il s'agit du mme vecteur, on ne peut pas utiliser U3D_ProduitScalaire : erreur incorrecte... */
  norme = vecteur->x * vecteur->x + vecteur->y * vecteur->y + vecteur->z * vecteur->z;
  norme = sqrt( norme );
  if ( 0.0 == norme ) {
    (void) fprintf(stderr, "U3D_Normalise : Vecteur nul\n");
    return +1; /* Vecteur nul, non normalisable --- Erreur mineure */
  }

  /* On adapte aussi les erreurs ? */
  if ( var_vecteur != NULL ) {
    (void) fprintf(stderr, "%s,%s l.%u : Adapter les erreurs\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  /* On adapte les coordonnes */
  vecteur->x /= norme;
  vecteur->y /= norme;
  vecteur->z /= norme;

  /* Normalisation russie */
  return 0;
}

/* -------- Calcul du produit scalaire -------- */

double U3D_ProduitScalaire(Vecteur3D *v_1, Vecteur3D *v_2,
			   Vecteur3D_Erreurs *err_1 __attribute__((__unused__)), Vecteur3D_Erreurs *err_2 __attribute__((__unused__)), double *erreur)
{
  double produit;

  if ( ( NULL == v_1 ) || ( NULL == v_2 ) ) return 0.0; /* Vecteur nul -> produit scalaire nul ! */

  /* On suppose le repre orthonorm */
  produit = v_1->x * v_2->x + v_1->y * v_2->y + v_1->z * v_2->z;

  if ( erreur != NULL ) {
    *erreur = 0.0;

    (void) fprintf(stderr, "%s,%s l.%u : OPTION  FAIRE --- Calculer l'erreur sur le produit scalaire\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  return produit;
}

/* -------- Conversion orientation -> vecteur --------
 |  Calcule les coordonnes du vecteur normal,  partir des angles de changement de repre
 |  [Quand repere sert  dfinir un faisceau, on obtient alors la direction de propagation]
 */

int REPERE_VecteurNormal( Orientation3D *repere, Vecteur3D *normale )
{
  double c_theta;

  if ( ( NULL == repere ) || ( NULL == normale ) ) return -1;

  c_theta = cos( repere->theta );

  normale->x =  sin( repere->theta );
  normale->y =  c_theta * cos( repere->phi );
  normale->z = -c_theta * sin( repere->phi );

  /* Conversion russie */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Changements de repres ~-~-~-~-~-~-~-~- */

/* -------- Repre absolu -> repre local --------
  Les oprations se font donc dans le sens direct, dcrit dans le .h
 */

int REPERE_AbsoluVersLocal(Vecteur3D *coord_absolues, Orientation3D *repere, Vecteur3D *coord_locales )
{
  double c_alpha, s_alpha, c_theta, s_theta, c_phi, s_phi;
  double matrice[ 9 ]; /* Les 9 coefficients de la matrice de passage */
 
  if ( ( NULL == coord_locales ) || ( NULL == repere ) || ( NULL == coord_absolues ) ) {
    return -1;
  }

  /* On effectue la translation "en sens direct" */
  coord_locales->x = coord_absolues->x - repere->position.x;
  coord_locales->y = coord_absolues->y - repere->position.y;
  coord_locales->z = coord_absolues->z - repere->position.z;

  /* Calcul des sinus & cosinus des angles */
  c_alpha = cos( repere->alpha ); s_alpha = sin( repere->alpha );
  c_theta = cos( repere->theta ); s_theta = sin( repere->theta );
  c_phi   = cos( repere->phi   ); s_phi   = sin( repere->phi   );

  /* Calcul de la matrice de passage [en lignes] */
  matrice[ 0 ] =   c_alpha * c_theta;
  matrice[ 1 ] = - c_alpha * s_theta * c_phi - s_alpha * s_phi;
  matrice[ 2 ] =   c_alpha * s_theta * s_phi - s_alpha * c_phi;

  matrice[ 3 ] =             s_theta;
  matrice[ 4 ] =             c_theta * c_phi;
  matrice[ 5 ] = -           c_theta * s_phi;

  matrice[ 6 ] =   s_alpha * c_theta;
  matrice[ 7 ] = - s_alpha * s_theta * c_phi + c_alpha * s_phi;
  matrice[ 8 ] =   s_alpha * s_theta * s_phi + c_alpha * c_phi;

  /* On effectue le produit */
  ProduitMatriciel( matrice, coord_locales );

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Repre local -> repre absolu --------
   Les oprations doivent donc tre effectues en sens inverse, soit
   1. Rotation de -phi autour de l'axe des X
   2. Rotation de -thta autour de l'axe des Z
   3. Rotation de -alpha autour de l'axe des Y
   4. Translation vers le repre original
 */

int REPERE_LocalVersAbsolu(Vecteur3D *coord_locales, Orientation3D *repere, Vecteur3D *coord_absolues)
{
  double c_alpha, s_alpha, c_theta, s_theta, c_phi, s_phi;
  double matrice[ 9 ]; /* Les 9 coefficients de la matrice de passage */
 
  if ( ( NULL == coord_locales ) || ( NULL == repere ) || ( NULL == coord_absolues ) ) {
    return -1;
  }

  /* Point de dpart */
  *coord_absolues = *coord_locales;

  /* Calcul des sinus & cosinus des angles */
  c_alpha = cos( repere->alpha ); s_alpha = sin( repere->alpha );
  c_theta = cos( repere->theta ); s_theta = sin( repere->theta );
  c_phi   = cos( repere->phi   ); s_phi   = sin( repere->phi   );

  /* Calcul de la matrice de passage [en lignes] */
  matrice[ 0 ] =   c_alpha * c_theta;
  matrice[ 1 ] =             s_theta;
  matrice[ 2 ] =   s_alpha * c_theta;

  matrice[ 3 ] = - c_alpha * s_theta * c_phi - s_alpha * s_phi;
  matrice[ 4 ] =             c_theta * c_phi;
  matrice[ 5 ] = - s_alpha * s_theta * c_phi + c_alpha * s_phi;

  matrice[ 6 ] =   c_alpha * s_theta * s_phi - s_alpha * c_phi;
  matrice[ 7 ] = -           c_theta * s_phi;
  matrice[ 8 ] =   s_alpha * s_theta * s_phi + c_alpha * c_phi;

  /* On effectue le produit */
  ProduitMatriciel( matrice, coord_absolues );

  /* On effectue la translation */
  coord_absolues->x += repere->position.x;
  coord_absolues->y += repere->position.y;
  coord_absolues->z += repere->position.z;

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Trouver les rotations convenables pour que le vecteur indiqu
   soit le vecteur Y (0 1 0) d'un repre local aprs transformation -------- */

int REPERE_TrouverRotations_Y(Vecteur3D *vect_Y, double *theta, double *phi)
{
  double v_theta, v_phi;
#ifdef _DEVELOPPE_
  if ( ( NULL == vect_Y ) || ( NULL == theta ) || ( NULL == phi ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Appel avec un paramtre NULL [Y=%p theta=%p phi=%p]\n",
		   __FILE__, __FUNCTION__, __LINE__, (void *) vect_Y, (void *) theta, (void *) phi );
    abort();
  }
  printf("%s : Vecteur = (X=%g, Y=%g, Z=%g)\n", __FUNCTION__, vect_Y->x, vect_Y->y, vect_Y->z );
#endif
  /* Les angles theta et phi doivent satisfaire
     0 = cT Y_x - sT sF Y_y + sT sF Y_z
     1 = sT Y_x + cT cF Y_y - cT sF Y_z
     0 =             sF Y_y +    cF Y_z
     ==> Systme donnant cos(theta) et tout a...
  */

  if ( 0.0 == vect_Y->y ) { /* alors phi = +/- pi/2 */
    if ( 0.0 == vect_Y->z ) { /* Vecteur sur l'axe des X */
      if ( vect_Y->x < 0.0 ) /* sT = -1.0 => - pi/2 */
	v_theta = -M_PI / 2.0;
      else /* sT = +1.0 => + pi/2 */
	v_theta = M_PI / 2.0;

      v_phi = 0.0; /* Rotation sans importance */
    } else if ( 0.0 == vect_Y->x ) { /* Vecteur sur l'axe des Z */
      if ( vect_Y->z < 0.0 )
	v_theta = 0.0;
      else
	v_theta = M_PI;
	
      v_phi = M_PI / 2.0;
    } else { /* alors phi = +/- pi/2 */
      v_phi = M_PI / 2.0;
      v_theta = - atan2( vect_Y->x, vect_Y->z );
    }
  } else {
    double delta, A, cT, sT;

    v_phi = - atan2( vect_Y->z, vect_Y->y );

    A = sin( v_phi ) * vect_Y->z - cos( v_phi ) * vect_Y->y;
    delta = vect_Y->x * vect_Y->x + A * A;
    if ( 0.0 == delta ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur, systme insoluble ?\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }

    cT = -A; /* Division par delta inutile, car on utilisera le quotient aprs ! */
    sT = vect_Y->x;

    v_theta = atan2( sT, cT );
  }

  *theta = v_theta; /* Les angles sont renvoys en RADIANS */
  *phi = v_phi;

  /* Le calcul a russi */
  return 0;
}

/* -------- produit en place de la matrice 3*3 en lignes par le vecteur colonne */

static void ProduitMatriciel(double *matrice, Vecteur3D *vecteur)
{
  double x0, y0, z0;

  /* Coordonnes originelles */
  x0 = vecteur->x;
  y0 = vecteur->y;
  z0 = vecteur->z;

  /* Produit matriciel */
  vecteur->x = matrice[ 0 ] * x0 + matrice[ 1 ] * y0 + matrice[ 2 ] * z0;
  vecteur->y = matrice[ 3 ] * x0 + matrice[ 4 ] * y0 + matrice[ 5 ] * z0;
  vecteur->z = matrice[ 6 ] * x0 + matrice[ 7 ] * y0 + matrice[ 8 ] * z0;
}

/* -------- Conversion absolu -> sphrique --------
   Les angles sont attendus en RADIANS
 */

int REPERE_CartesienVersSpherique( Vecteur3D *xyz, double *r, double *theta, double *phi )
{
  double d, Theta, Phi;

  if ( NULL == xyz ) return -1;

  d =  xyz->x * xyz->x + xyz->y * xyz->y + xyz->z * xyz->z;
  d = sqrt( d );
  if ( r  != NULL ) *r = d;

  if ( 0.0 == d ) {
    Theta = 0.0;
    Phi = 0.0;
  } else {
    Phi = acos( xyz->z / d );
    Theta = atan2( xyz->y, xyz->x );
  }

  /* Conversion russie */
  if ( theta != NULL ) *theta = Theta;
  if ( phi != NULL ) *phi = Phi;
  return 0;
}

/* -------- Conversion sphrique -> absolu -------- *
   Les angles sont attendus en RADIANS
 * ------------------------------------------------ */

int REPERE_SpheriqueVersCartesien( double r, double theta, double phi, Vecteur3D *xyz )
{
  if ( NULL == xyz ) return -1;
  if ( r < 0.0 ) {
    (void) fprintf(stderr, "REPERE_SpheriqueVersCartesien : R absurde [%g]\n", r);
    return -2;
  }

  if ( 0.0 == r ) {
    xyz->x = 0.0;
    xyz->y = 0.0;
    xyz->z = 0.0;

    return 0;
  }

  /* Changement de repre */
  xyz->x = r * sin( phi ) * cos( theta );
  xyz->y = r * sin( phi ) * sin( theta );
  xyz->z = r * cos( phi );

  /* Tout s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~-~-~- quation d'une droite, travail avec cette quation ~-~-~-~-~-~-~-~- */

/* -------- quation : vecteur directeur (2 - 1), la droite passe par 1 pour t=0 -------- */ 

inline int U3D_VecteurDirecteur(Vecteur3D *point_1, Vecteur3D *point_2, Vecteur3D *vecteur,
				Vecteur3D_Erreurs *err_1, Vecteur3D_Erreurs *err_2, Vecteur3D_Erreurs *err_vecteur)
{
  if ( ( NULL == point_1 ) || ( NULL == point_2 ) || ( NULL == vecteur ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Un des vecteurs n'est pas indiqu !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Coordonnes */
  vecteur->x = point_2->x - point_1->x;
  vecteur->y = point_2->y - point_1->y;
  vecteur->z = point_2->z - point_1->z;

  /* Erreur sur ce vecteur directeur si demand */
  if ( err_vecteur != NULL ) {
    /* On initialise */
    err_vecteur->var_x  = 0.0; err_vecteur->var_y  = 0.0; err_vecteur->var_z  = 0.0;
    err_vecteur->cov_xy = 0.0; err_vecteur->cov_xz = 0.0; err_vecteur->cov_yz = 0.0;

    /* Les matrices de covariance sont additives, les deux vecteurs tant supposs indpendants. */
    if ( err_1 != NULL ) {
      err_vecteur->var_x  += err_1->var_x;  err_vecteur->var_y  += err_1->var_y ; err_vecteur->var_z  += err_1->var_z;
      err_vecteur->cov_xy += err_1->cov_xy; err_vecteur->cov_xz += err_1->cov_xz; err_vecteur->cov_yz += err_1->cov_yz;
    }

    if ( err_2 != NULL ) {
      err_vecteur->var_x  += err_2->var_x;  err_vecteur->var_y  += err_2->var_y ; err_vecteur->var_z  += err_2->var_z;
      err_vecteur->cov_xy += err_2->cov_xy; err_vecteur->cov_xz += err_2->cov_xz; err_vecteur->cov_yz += err_2->cov_yz;
    }
  }

  return 0;
}

/* -------- On donne les deux points d'une droite, t : o se trouve le point correspondant ?
    Suppose les erreurs sur les deux points indpendantes
   -------- */

inline int U3D_PositionSurDroite_PP(Vecteur3D *origine, Vecteur3D *point_2, double t, Vecteur3D *position,
				    Vecteur3D_Erreurs *err_origine, Vecteur3D_Erreurs *err_point_2, double err_t,
				    Vecteur3D_Erreurs *err_position)
{
  if ( ( NULL == point_2 ) || ( NULL == origine ) || ( NULL == position ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Un des vecteurs n'est pas indiqu !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  position->x = origine->x + t * (point_2->x - origine->x);
  position->y = origine->y + t * (point_2->y - origine->y);
  position->z = origine->z + t * (point_2->z - origine->z);

  /* Calcul de l'erreur - en supposant les coordonnes indpendantes ! */
  if ( err_position != NULL ) {
    /* On initialise  zro pour tre tranquille */
    err_position->var_x  = 0.0; err_position->var_y  = 0.0; err_position->var_z = 0.0;
    err_position->cov_xy = 0.0; err_position->cov_xz = 0.0; err_position->cov_yz = 0.0;

    if ( err_t <= 0.0 ) { /* t parfaitement connu */
      /* Les matrices de covariance sont additives, les deux vecteurs tant supposs indpendants. */
      if ( err_point_2 != NULL ) {
	double a;

	a = t * t; /* (t)^2 */
	err_position->var_x += a * err_point_2->var_x; err_position->cov_xy += a * err_point_2->cov_xy;
	err_position->var_y += a * err_point_2->var_y; err_position->cov_yz += a * err_point_2->cov_yz;
	err_position->var_z += a * err_point_2->var_z; err_position->cov_xz += a * err_point_2->cov_xz; 
      }

      if ( err_origine != NULL ) {
	double a;

	a = (1.0 - t); a *= a; /* (1-t)^2 */
	err_position->var_x += a * err_origine->var_x; err_position->cov_xy += a * err_origine->cov_xy;
	err_position->var_y += a * err_origine->var_y; err_position->cov_yz += a * err_origine->cov_yz;
	err_position->var_z += a * err_origine->var_z; err_position->cov_xz += a * err_origine->cov_xz; 
      }
    } else {
      (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- cas d'une erreur sur t indique\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }

  }

  return 0;
}

/* -------- On donne l'quation d'une droite, t : o se trouve le point correspondant ? --------
    Suppose les erreurs sur les deux vecteurs indpendantes
    -------- */

inline int U3D_PositionSurDroite_VP(Vecteur3D *vecteur_directeur, Vecteur3D *origine, double t, Vecteur3D *position,
				    Vecteur3D_Erreurs *err_v_directeur, Vecteur3D_Erreurs *err_origine, double err_t,
				    Vecteur3D_Erreurs *err_position)
{
  if ( ( NULL == vecteur_directeur ) || ( NULL == origine ) || ( NULL == position ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Un des vecteurs n'est pas indiqu !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  position->x = origine->x + t * vecteur_directeur->x;
  position->y = origine->y + t * vecteur_directeur->y;
  position->z = origine->z + t * vecteur_directeur->z;

  /* Calcul de l'erreur - en supposant les coordonnes indpendantes ! */
  if ( err_position != NULL ) {
    /* On initialise  zro pour tre tranquille */
    err_position->var_x  = 0.0; err_position->var_y  = 0.0; err_position->var_z  = 0.0;
    err_position->cov_xy = 0.0; err_position->cov_xz = 0.0; err_position->cov_yz = 0.0;

    /* Les matrices de covariance sont additives, les deux vecteurs tant supposs indpendants. */
    if ( err_origine != NULL ) {
      err_position->var_x += err_origine->var_x; err_position->cov_xy += err_origine->cov_xy;
      err_position->var_y += err_origine->var_y; err_position->cov_yz += err_origine->cov_yz;
      err_position->var_z += err_origine->var_z; err_position->cov_xz += err_origine->cov_xz;
    }

    if ( err_t <= 0.0 ) { /* t parfaitement connu */
      if ( err_v_directeur != NULL ) {

	t *= t; /* t */
	err_position->var_x += t * err_v_directeur->var_x; err_position->cov_xy += t * err_v_directeur->cov_xy;
	err_position->var_y += t * err_v_directeur->var_y; err_position->cov_yz += t * err_v_directeur->cov_yz;
	err_position->var_z += t * err_v_directeur->var_z; err_position->cov_xz += t * err_v_directeur->cov_xz;
      }
    } else { /* On suppose t indpendant des deux autres vecteurs */
      if ( NULL == err_v_directeur ) {
	double var_t;

	var_t = err_t * err_t; /* On a pass l'cart-type */
	/* Si t est une VA et A un v. constant, alors E(tA) = E(t)A et S(tA) = V(t) A A', A' = transpose */
	err_position->var_x += var_t * vecteur_directeur->x * vecteur_directeur->x;
	err_position->var_y += var_t * vecteur_directeur->y * vecteur_directeur->y;
	err_position->var_z += var_t * vecteur_directeur->z * vecteur_directeur->z;

	err_position->cov_xy += var_t * vecteur_directeur->x * vecteur_directeur->y;
	err_position->cov_xz += var_t * vecteur_directeur->x * vecteur_directeur->z;
	err_position->cov_yz += var_t * vecteur_directeur->y * vecteur_directeur->z;
      } else {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Erreur sur t et sur le vecteur directeur\n",
		       __FILE__, __FUNCTION__, __LINE__);
      }
    }
  }

  return 0;
}

/* ~-~-~-~-~-~-~-~- Intersections de droites, de plans,... ~-~-~-~-~-~-~-~- */

/* -------- Intersection entre une droite et un plan d'quation Y=Cte --------
 |          Revient  projeter le point pt_droite sur le plan Y=Cte dans la direction v_directeur
 |
 | pt_droite   : un point par lequel passe la droite
 | v_directeur : vecteur directeur de la droite
 | y           : quation du plan sur lequel  projeter  le point pt_droite dans la direction v_directeur
 | intersection: coordonnes du point rsultat [intersection.y = y, forcment]
 */

int U3D_Inter_Droite_Plan_YCt(Vecteur3D *pt_droite, Vecteur3D *v_directeur, double y, Vecteur3D *intersection)
{
  double t;

  if ( ( NULL == pt_droite ) || ( NULL == v_directeur ) || ( NULL == intersection ) ) return -1;

  if ( 0.0 == v_directeur->y ) {
    (void) fprintf(stderr, "U3D_Inter_Droite_Plan_YCt : droite parallle au plan Y=Cte\n");
    return -2;
  }

  /* Systme  rsoudre :
     { pt_droite->x + t * v_directeur->x = intersection->x
     { pt_droite->y + t * v_directeur->y = intersection->y
     { pt_droite->z + t * v_directeur->z = intersection->z
     {                y                  = intersection->y

     donc t = (y - pt_droite->y) / v_directeur->y
  */
  t =  (y - pt_droite->y) / v_directeur->y;

  /* Et on en dduit donc le reste facilement... */
  intersection->x = pt_droite->x + t * v_directeur->x;
  intersection->y = y;
  intersection->z = pt_droite->z + t * v_directeur->z;

  /* Pas de problme */
  return 0;
}
