/* LASE : Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2001, Emmanuel CURIS
   Création : 22 mai 1998
   Dernière modification : 12 septembre 2001

   Tout ce qui concerne le X.A.N.E.S.

   Normalisation du XANES
   Récupéré de ../XANES.C et de ../EXF_GEM.C ( GER_xnorm() )
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "serie.h"

#include "interface.h" /* Pour RCT_pointeur */
#include "extrait.h"   /* Pour lisse_graphe() et MU1_calc() */
#include "tracer.h"
#include "xanes.h"

#include "norm_xanes.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_REFERENCE 20

/* ************ Fonctions locales, qui n'apparaissent qu'ici ******************************* */

static double facteur_norm(Graphe *signal, MU1_infos *infos, int mode, double *y_min);

static int GER_xnorm(unsigned int num_objet, int x, int y);
static int prepare_norm_xanes(unsigned int);

static int  CLC_xnorm(Fenetre *fenetre, int bouton, int n_clic, int etat, int x_clic, int y_clic);
static int  CLV_xnorm(Fenetre *fenetre, unsigned int touche, int etat);
static int  MSG_xnorm(Fenetre *fenetre, Message *message);
static void RDS_xnorm(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);

/* ~~~~~~~~~~~~ Normalisation en cherchant le point de fin de seuil par modélisation ~~~~~~~ */

/* ~-~-~-~- Réalisation de la normalisation sans interface ~-~-~-~- */

/* ------ Calcul de la valeur de normalisation
     Renvoie le facteur de normalisation

  signal : le spectre à normaliser
  infos  : les informations nécessaires à la normaliastion
  mode   : la façon dont on veut faire la normalisation
  y_min  : valeur prise égale à 0 dans la normalisation (doit être NON-NULL)
*/

static double facteur_norm(Graphe *signal, MU1_infos *infos, int mode, double *y_min)
{
  double facteur = 0.0, e_norm, e_norm2;
  register double *x, *y;
  int j;

  if (infos == NULL) return 0.0;
  if (y_min == NULL) return 0.0;

  e_norm = infos->e_0; /* Énergie de normalisation : en ce point, le spectre vaut 1 forcément */
  switch (mode)
  {
  case XNORM_MU0: /* On utilise un µ0, dont les paramètres sont dans infos */
    facteur = 0.0;
    for (j = infos->degre ; j > 0 ; j--) {
      facteur += infos->coefs[j];
      facteur *= e_norm;
    }
    facteur += infos->coefs[0];
    facteur -= signal->y_min;
    *y_min = signal->y_min;
    break;

  case XNORM_1_PT: /* Le point choisi vaut 1, le minimum vaut 0 */
    *y_min = signal->y_min;

    /* Interpolation linéaire entre les 2 points encadrant le E norm sélectionné
       pour estimer la valeur réelle au point choisi
       Suppose implicitement le graphe trié
    */
    x = signal->x;
    y = signal->y;
    
    if ( (e_norm > signal->x_max) || (e_norm < signal->x_min) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Point de normalisation %g en-dehors de l'autre spectre [%g, %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, e_norm, signal->x_max, signal->x_min);
      /* On ne peut pas afficher de message, car appelé dans le redessin
	 => donne un appel cyclique ! */
      return 0.0;
    }

    while (*x < e_norm) {
      x++; y++;
    }

    if (*x == e_norm) /* Il existe bien dans le spectre */
      facteur = *y;
    else { /* x est juste supérieur à la valeur voulu -> le x avant est juste inférieur */
      if (x == signal->x) { /* On est sur le premier point */
#ifdef _DEVELOPPE_
	fprintf(stderr, "xanes_normalise : premier point utilisé, devrait jamais arriver !\n");
	fprintf(stderr, "   [si on est là : e_norm >= signal->x_min, e_norm != signal->x_min et\n");
	fprintf(stderr, "    signal->x_min > e_norm !]\n");
#endif
	facteur = *y;
      } else { /* Interpolation linéaire */
	double y0 = *(y - 1);
	facteur = y0 + (e_norm - *(x - 1)) * ( ( *y - y0 ) / ( *x - *(x - 1) ) );
      }
    }
    break;

  case XNORM_2_PTS: /* Valeur en e_norm = 1, valeur en e2 = 0
		       Interpolation linéaire pour avoir les valeurs attendues avant normalisation */
    e_norm2 = infos->e_2;

    /* On cherche la valeur du y_min d'abord, soit e_norm2 */
    x = signal->x;
    y = signal->y;
    
    if ( (e_norm2 > signal->x_max) || (e_norm2 < signal->x_min) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Point de normalisation %g en-dehors de l'autre spectre [%g, %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, e_norm2, signal->x_max, signal->x_min);
      /* On ne peut pas afficher de message, car appelé dans le redessin
	 => donne un appel cyclique ! */
      return 0.0;
    }

    while (*x < e_norm2) {
      x++; y++;
    }

    if (*x == e_norm2) /* On est dessus */
      *y_min = *y;
    else { /* Interpolation linéaire, le point sur lequel on est est juste après */
      if ( *x == *(x - 1) ) {
	EXF_ERREUR(ERR_DOUBLONS);
	return 0.0;
      }

      *y_min = *(y - 1) + (*y - *(y - 1) ) * ( e_norm2 - *(x - 1) ) / ( *x - *(x - 1) );
    }

    /* On cherche maintenant la valeur sur laquelle on va normaliser */
    x = signal->x;
    y = signal->y;
    
    if ( (e_norm > signal->x_max) || (e_norm < signal->x_min) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Point de normalisation %g en-dehors de l'autre spectre [%g, %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, e_norm, signal->x_max, signal->x_min);
      /* On ne peut pas afficher de message, car appelé dans le redessin
	 => donne un appel cyclique ! */
      return 0.0;
    }

    while (*x < e_norm) {
      x++; y++;
    }

    {
      double y_max;

      if (*x == e_norm) /* On est dessus */
	y_max = *y;
      else { /* Interpolation linéaire, le point sur lequel on est est juste après */
	if ( *x == *(x - 1) ) {
	  EXF_ERREUR(ERR_DOUBLONS);
	  return 0.0;
	}
	
	y_max = *(y - 1) + (*y - *(y - 1) ) * ( e_norm2 - *(x - 1) ) / ( *x - *(x - 1) );
      }

      facteur = y_max - *y_min;
    }
    break;

#ifdef _DEVELOPPE_
  default:
    fprintf(stderr, "facteur_norm : mode inconnu (%d)\n", mode);
    return 0.0;
#endif
  }

  return facteur;
}

/* ------ Normalisation effective ------------------- */

Graphe *xanes_normalise(Graphe *signal, MU1_infos *infos, int mode, boolean nouveau)
{
  Graphe *resultat;
  double facteur = 0.0, y_min = 0.0;
  unsigned long i;
  register double *y, *err_y = NULL;
  char *titre; /* Titre du graphe normalisé */

  if (signal == NULL) return NULL;
  if (infos == NULL) return NULL;

  if (nouveau == VRAI) {
    resultat = GRF_Copier( signal, NULL, FAUX );
    if (resultat == NULL)
      return NULL;
  } else
    resultat = signal;

  if ( XNORM_MU0 == mode ) { /* On recalcule le polynome, car peut ne pas avoir été fait si normalisation automatique */
    XAS_MU01_CalculerMu0_I( signal, infos );
  }

  facteur = facteur_norm(signal, infos, mode, &y_min);
  if (facteur == 0.0) {
    EXF_ERREUR(DIV_ZERO, txt_err_fact_norm, txt_err_calc_annule);
    if (nouveau == VRAI)
      GRF_efface(resultat, VRAI, VRAI);
    return NULL;
  }

  /* Maintenant que tout est prêt, on normalise */
  if (resultat->erreur_y != NULL) {
    err_y = resultat->erreur_y->y;
  }
  for (i = 0, y = resultat->y ; i < resultat->nombre_point ; i++, y++) {
    *y -= y_min;
    *y /= facteur;

    if (err_y) {
      *err_y /= fabs(facteur);
      err_y++;
    }
  }
    
  titre = TXT_genere_chaine( txt_titre_grf_norm, signal->nom );
  GRF_titre(resultat, titre);
  free( titre ); titre = NULL;
  resultat->y_max = (resultat->y_max - y_min) / facteur;
  resultat->y_min = (resultat->y_min - y_min) / facteur;
  if (facteur < 0.0) {
    double x = resultat->y_max; resultat->y_max = resultat->y_min; resultat->y_min = x;
  }

  HIST_Titre(1, "Normalisation d'un seuil d'absorption X", "NORM");
  HIST_Texte(1, 0, "Spectre normalis"e_AIGU" : %s\n", resultat->nom);
  if ( infos->mu_0 != NULL ) {
    HIST_Texte(2, 0, "R"e_AIGU"f"e_AIGU"rence de contr"o_CIRC"le : %s\n", infos->mu_0->nom);
  }
  if ( mode == XNORM_MU0 ) {
    HIST_Texte(3, 0, E_AIGU"nergie de normalisation : %f\n", infos->e_0);
    HIST_Texte(2, 0, "M"e_AIGU"thode : mod"e_AIGU"le de "MU"0 (polyn"o_CIRC"me");
    HIST_Texte(5, 0, " de degr"e_AIGU" %d", infos->degre);
    if (infos->methode == ELIM_MU0_RAPIDE)
      HIST_Texte(4, 0, " -- moindres carr"e_AIGU"s simples");
    else
      HIST_Texte(4, 0, " -- moindres carr"e_AIGU"s pond"e_AIGU"r"e_AIGU"s par (x(i) - x(i-1)))");
    HIST_Texte(2, 0, ")\n");
    HIST_Texte(5, 0, "Intervalle utilis"e_AIGU" : [%f, %f]\n", infos->e_2, signal->x_max);
    HIST_Polynome(5, 5, 2, infos->coefs, infos->degre, 'a');
  } else if ( mode == XNORM_1_PT ) {
    HIST_Texte(2, 0, "M"e_AIGU"thode : minimum à 0, valeur 1 pour X = %f\n", infos->e_0);
  } else if ( mode == XNORM_2_PTS ) {
    HIST_Texte(2, 0, "M"e_AIGU"thode : aleur 0 pour X = %f, valeur 1 pour X = %f\n", infos->e_2, infos->e_0);
  }

  GRF_bornes(resultat->erreur_y, NE_PAS_FAIRE, DESORDONNE); /* Ont certainement changé ! */
  GRF_modifier(signal, resultat, FAUX); /* Même nombre de points */

  return resultat;
}

/* ------------ Ouverture de la fenêtre ---------------------------------------------------- */

void norm_xanes(Graphe *signal)
{
  Fenetre *fenetre;
  MU1_infos *infos = NULL;

  if (signal == NULL) return; /* Calcul n'a pas de raison d'être */

  if ( arbres[NORM_XANES].drapeaux.ouvert == VRAI ) { /* Arbre déjà ouvert */
    Graphe *vieux; /* Ancien spectre travaillé dans ce module */
    int reponse = EXF_ERREUR(DEJA_ACTIF); /* Module déjà ouvert */

    fenetre = ARB_QuelleFenetre( &(arbres[NORM_XANES]), FAUX );
    vieux = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
    infos = (MU1_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

    if ( (reponse == 1) || (vieux == signal) ) { /* On active seulement */
      FEN_premierplan(fenetre); /* Fenêtre au premier plan */
      return; /* Rien d'autre à faire */
    } else if (reponse == 2) { /* On rappelle, pour le nouveau graphe */
      if ( signal->drapeaux.protege ) {
	EXF_ERREUR(GRF_PROTEGE);
	return;
      }

      /* On délie les informations, par sécurité */
      FEN_GRF_LierGraphe( fenetre, NULL );
      FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

      vieux->drapeaux.protege = FAUX; /* On a fini de travailler avec */

      FEN_premierplan(fenetre); /* Fenêtre au premier plan */
    } else { /* On annule l'appel */
      return;
    }
  } else { /* Arbre non ouvert : on l'ouvre */
    if ( signal->drapeaux.protege ) {
      EXF_ERREUR(GRF_PROTEGE);
      return;
    }

    if ( arbres[NORM_XANES].drapeaux.prepare == FAUX ) {
      double e_norm;

      e_norm = cherche_e0(signal, FAUX); /* E norm = seuil, E2 100 eV après le seuil,
					    juste pour le premier appel, après on mémorise
					    pour gagner du temps de travail... */
      OBJ_MettreNombre( NORM_XANES, NXAN_ENORM, e_norm, 0.0 );
      OBJ_MettreNombre( NORM_XANES, NXAN_E2, e_norm + 100.0, 0.0 );
    }

    fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, txt_titre_fenetre, NULL,
				  NORM_XANES, BI_FEN_ARBRE_HAUT, prepare_norm_xanes,
				  RDS_xnorm, CLV_xnorm, CLC_xnorm, MSG_xnorm, RCT_pointeur,
				  (int (*)(Fenetre *)) NULL );
    if (fenetre == NULL) return; /* Rien à faire... */

    met_chaine(NORM_XANES, NX_REFERENCE, LGMAX_NOM_REFERENCE, txt_pas_reference, TXT_COPIER);

    fenetre->drapeaux.redessin = VRAI; /* Redessin si changement de taille */
    fenetre->msg_masque.fermeture = VRAI; /* Messages traités par MSG_xnorm */

    FEN_Ouvrir( fenetre );
  }
  
  if (infos == NULL) { /* Aucune info n'existe déjà */
    infos = (MU1_infos *) malloc( (size_t) sizeof(MU1_infos) );
    if (infos == NULL) {
      FEN_Detruire( fenetre );
      SYS_ERREUR(MANQUE_MEMOIRE, "XANES/normer.c", "norm_xanes", "infos"); /* Sinon dangereux sous X... */
      return;
    }
    infos->mu_0 = NULL; /* Pas de graphe de référence - s'il y en a un, il est stocké ici
			   car on n'utilise pas cette entrée autrement... */
  }

  /* On prépare les calculs... */
  infos->e_0 = OBJ_LireNombre( NORM_XANES, NXAN_ENORM, NULL ); /* Mêmes que la fois d'avant... */
  infos->e_2 = OBJ_LireNombre( NORM_XANES, NXAN_E2, NULL );
  infos->degre = OBJ_LireEntier_U( NORM_XANES, NXAN_DEGRE );
  if ( OBJ_choisit(NORM_XANES, NXAN_RAPIDE, QUESTION) )
    infos->methode = ELIM_MU0_RAPIDE;
  else if ( OBJ_choisit(NORM_XANES, NXAN_LENTE, QUESTION) )
    infos->methode = ELIM_MU0_LENT;
  infos->avec_e3 = FAUX; /* On n'utilise pas 3 points pour le polynôme */

  /* On calcule le polynôme de départ, si nécessaire */
  if ( VRAI == OBJ_choisit( NORM_XANES, NORM_MU0, QUESTION ) )
    XAS_MU01_CalculerMu0_I( signal, infos );

  /* On associe à la fenêtre */
  FEN_GRF_LierGraphe( fenetre, signal );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, infos );
}

/* -------------- Gestion du clic souris --------------------------------------------------- */

static int CLC_xnorm(Fenetre *fenetre, int bouton, int n_clic, int etat __attribute__((__unused__)), int x_clic, int y_clic)
{
  MU1_infos *infos;
  Graphe *graphe;
  double x_echelle;

#ifdef _TOS_
  int handle;

  wind_get(0, WF_TOP, &handle);
  if (handle != fenetre->handle) {
    wind_set(fenetre->handle, WF_TOP);
    return VRAI;
  }
#endif
  if ( n_clic > 1 ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Clic multiple, sans fonction\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return BI_EVENEMENT_NON_TRAITE;
  }

  if ( (x_clic < fenetre->w_x) || (x_clic > fenetre->w_x + (int) fenetre->w_larg)
    || (y_clic < fenetre->w_y) || (y_clic > fenetre->w_y + (int) fenetre->w_haut) ) {
    return BI_EVENEMENT_TRAITE; /* Clic en-dehors de la zone de travail */
  }

  graphe = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (MU1_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  
  x_echelle = (graphe->x_max - graphe->x_min) / ( (double) (fenetre->w_larg) );
  if (x_echelle) {
    if (bouton == 1) { /* Clic du bouton gauche -> E de normalisation */
      infos->e_0 = graphe->x_min + (x_clic - fenetre->w_x) * x_echelle;
      OBJ_MettreNombre( NORM_XANES, NXAN_ENORM, infos->e_0, 0.0 );
    } else { /* Bouton droit -> E2 */
      infos->e_2 = graphe->x_min + (x_clic - fenetre->w_x) * x_echelle;
      OBJ_MettreNombre( NORM_XANES, NXAN_E2, infos->e_2, 0.0 );

      XAS_MU01_CalculerMu0_I( graphe, infos ); /* On recalcule le polynome */
    }
    FEN_redessin(fenetre, BI_RDS_UTIL);     /* On redessine             */
  }

  return BI_EVENEMENT_TRAITE;
}

/* -------- Gestion du clavier ------------\
 |                                         |
 | espace = fin du calcul                  |
 |      1 = normalisation sur 1 point      |
 |      2 = normalisation sur 2 points     |
 |      m = normalisation avec µ0          |
 |    n,p = méthode utilisée pour µ0       |
 |      r = choisir un graphe de référence |
 |    +/- = degré du polynôme pour µ0      |
 \---------------------------------------- */

static int CLV_xnorm(Fenetre *fenetre, unsigned int touche, int etat)
{
  int retour;
  Message message;

  if ( etat & K_CTRL ) {
    return BI_EVENEMENT_NON_TRAITE;
  }

  switch (touche)
  {
  case ' ':
    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = 0; message.fenetre.y = 0;
    message.fenetre.larg = 0; message.fenetre.haut = 0; /* Dans le cas où != 0, fermeture forcée... */
    retour = (fenetre->message)(fenetre, &message);
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;

  case '+':
    retour = ARB_SimulerClic( NORM_XANES, NXAN_D_PLUS, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;
  case '-':
    retour = ARB_SimulerClic( NORM_XANES, NXAN_D_MOINS, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;

  case '1':
    retour = ARB_SimulerClic( NORM_XANES, NORM_1_POINT, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;
  case '2':
    retour = ARB_SimulerClic( NORM_XANES, NORM_2_POINTS, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;
  case 'm':
    retour = ARB_SimulerClic( NORM_XANES, NORM_MU0, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;

  case 'n':
    retour = ARB_SimulerClic( NORM_XANES, NXAN_RAPIDE, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;
  case 'p':
    retour = ARB_SimulerClic( NORM_XANES, NXAN_LENTE, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;

  case 'r':
    retour = ARB_SimulerClic( NORM_XANES, NX_REFERENCE, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE; /* Touche non traitée */
}

/* -------------- Messages reçus par la fenêtre de normalisation XANES --------------------- */

static int MSG_xnorm(Fenetre *fenetre, Message *message)
{
  Graphe *signal, *norm = NULL;
  MU1_infos *infos;
  double e_norm __attribute__((__unused__)) = -1.0;

  signal = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (MU1_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  if ( infos != NULL ) {
    e_norm = infos->e_0; /* Énergie de normalisation */
  }

  if ( message->type == WM_CLOSED ) { /* On veut fermer la fenêtre */
    int methode = XNORM_MU0, reponse;

    if ( ( NULL == signal ) || ( message->fenetre.haut != 0 ) ) /* Demande de fermeture forcée */
      reponse = REPONSE_FERMER; /* On ferme sans rien demander et sans rien faire */
    else { /* On demande s'il faut fermer */
      reponse = BI_ERR_alerte(fenetre, &Q_FermerFenetre, VRAI);
    }

    if ( REPONSE_ANNULER == reponse ) /* On annule la fermeture */
      return BI_EVENEMENT_TRAITE;

    /* On délie les informations, par sécurité */
    FEN_GRF_LierGraphe( fenetre, NULL );
    FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

    /* On ferme la fenêtre de toute facon */
    FEN_Detruire( fenetre );

    if ( REPONSE_FERMER == reponse ) { /* On ferme sans normaliser */
      free( infos ); infos = NULL;
      if ( signal != NULL ) /* Peut l'être si on force la fermeture pour cause d'effacement */
	signal->drapeaux.protege = FAUX;
      return BI_EVENEMENT_FEN_FERMEE;
    }

    /* Methode de normalisation ? */
    if ( OBJ_choisit(NORM_XANES, NORM_MU0, QUESTION) == VRAI ) {
      methode = XNORM_MU0;
    } else if ( OBJ_choisit(NORM_XANES, NORM_1_POINT, QUESTION) == VRAI ) {
      methode = XNORM_1_PT;
    } else if ( OBJ_choisit(NORM_XANES, NORM_2_POINTS, QUESTION) == VRAI ) {
      methode = XNORM_2_PTS;
    }

    /* On normalise le spectre */
    if ( REPONSE_SPECTRE_SEUL == reponse ) {
      norm = xanes_normalise( signal, infos, methode, OBJ_choisit( NORM_XANES, NORM_NOUVEAU, QUESTION ) );
    } else { /* On normalise toute la serie */
      Serie_Graphes *serie;
      Graphe **selection;
      unsigned long i, nbr_graphes;

      serie = GRF_QuelleSerie(signal); /* Travail sur la serie du signal */
      if (NULL == serie) {
	(void) fprintf(stderr, "MSG_xnorm : pas de serie pour le signal !\n");
	free( infos ); infos = NULL;
	return BI_EVENEMENT_FEN_FERMEE; /* Fini ! */
      }

      selection = GRF_ChoisirType(serie, signal->type, VRAI, &nbr_graphes);
      if (NULL == selection) {
	(void) fprintf(stderr, "MSG_xnorm : aucun graphe retenu dans la serie !\n");
	free( infos ); infos = NULL;
	return BI_EVENEMENT_FEN_FERMEE; /* Fini ! */
      }

      for (i = 0 ; i < nbr_graphes ; i++) {
	norm = xanes_normalise(selection[i], infos, methode, OBJ_choisit(NORM_XANES, NORM_NOUVEAU, QUESTION));
      }

      free( selection ); selection = NULL;
    }

    free( infos ); infos = NULL;
    GRF_aller(norm); /* Pour être sûr du redessin */
    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "Message non traité [%d]", message->type );
  return BI_EVENEMENT_NON_TRAITE;
}

/* ------------- Redessin de la fenêtre de normalisation X.A.N.E.S. -----------------------\
 |   Le clipping est déjà installé directement lors de l'appel                             |
 |                                                                                         |
 | fenetre    : structure associée à la fenêtre considérée                                 |
 | x, y       : coordonnées de la zone de redessin                                         |
 | larg, haut : dimensions de la zone de redessin                                          |
 \-------------------------------------------------------------------------------------­-- */

static void RDS_xnorm(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Graphe *graphe;
  InfosRegion2D region;
  MU1_infos *infos;
  double echelle_X, echelle_Y;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut);

  graphe = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  if ( NULL == graphe ) {
    return;
  }

  /* On prépare la région 2D : aide pour les dessins après en fait */
  region.x_BG = 0; region.y_BG = fenetre->w_haut;
  region.largeur = fenetre->w_larg;
  region.hauteur = fenetre->w_haut;

  region.axe_X.min = graphe->x_min;
  region.axe_X.max = graphe->x_max;

  region.axe_Y.min = graphe->y_min;
  region.axe_Y.max = graphe->y_max;

  /* On met à NULL ce qui ne sert à rien, pour éviter d'éventuelles mauvaises surprises ! */
  region.axe_X.legende = NULL; region.axe_X.unite   = NULL;
  region.axe_Y.legende = NULL; region.axe_Y.unite   = NULL;
  region.X_secondaire  = NULL;
  region.Y_secondaire  = NULL;

  /* On trace le graphe dans la région 2D et on récupère les échelles au passage */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, graphe, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &echelle_X, &echelle_Y );
  if ( ( echelle_X <= 0.0 ) || ( echelle_Y <= 0.0 ) ) {
    /* Problème dans les échelles : pas de dessin possible */
    return;
  }

  infos = (MU1_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return;

  if ( OBJ_choisit(NORM_XANES, NORM_MU0, QUESTION) ) {
    /* On trace le modèle de µ0 - en bleu */
    vsl_color(_PRG_station->handle, BLEU);
    vsl_type (_PRG_station->handle, 1);

    TRC_Polynome( fenetre, &region, BLEU, BI_CONTINU, BI_MARQUE_AUCUNE,
		  infos->degre, infos->coefs, graphe->nombre_point, graphe->x );
  }

  /* On met les barres */
  /* a. Énergie de normalisation [rouge] */
  if ( ( infos->e_0 >= graphe->x_min ) && ( infos->e_0 <= graphe->x_max ) ) {
    TRC_Couleur( ROUGE );
    TRC_LigneV( fenetre, TRC_ConvertitX_PvE( infos->e_0, graphe->x_min, 0, echelle_X ), 0, fenetre->w_haut );
  }

  /* b. 2e point de normalisation, s'il existe */
  if ( OBJ_choisit(NORM_XANES, NORM_1_POINT, QUESTION) == FAUX) { /* Pas sur 1 point */
    TRC_Couleur( VERT );
    TRC_LigneV( fenetre, TRC_ConvertitX_PvE( infos->e_2, graphe->x_min, 0, echelle_X ), 0, fenetre->w_haut );
  }

  /* S'il y a une référence, on l'affiche */
  if ( infos->mu_0 != NULL ) {
    int mode = XNORM_1_PT;
    double facteur = 0.0, y_min = 0.0;

    /* Quelle serait l'échelle verticale après normalisation ? */
    if ( OBJ_choisit(NORM_XANES, NORM_MU0, QUESTION) == VRAI )
      mode = XNORM_MU0;
    else if ( OBJ_choisit(NORM_XANES, NORM_1_POINT, QUESTION) == VRAI )
      mode = XNORM_1_PT;
    else if ( OBJ_choisit(NORM_XANES, NORM_2_POINTS, QUESTION) == VRAI )
      mode = XNORM_2_PTS;
    facteur = facteur_norm(graphe, infos, mode, &y_min);
    if ( 0.0 == facteur ) return; /* Rien à faire... */
#ifdef _DEVELOPPE_
/*    printf("Facteur de normalisation : %f\n", facteur); */
#endif

    /* Le spectre d'origine va de grf->y_min à grf_y_max. Après normalisation,
       il va de (grf->y_min - y_min) / facteur à (grf->y_max - y_min) / facteur
       si facteur > 0, en sens inverse sinon.*/
    echelle_Y /= fabs( facteur );

    /* On trace la courbe expérimentale - en gris foncé */
    if ( facteur > 0.0 ) {
      TRC_Graphe( fenetre, infos->mu_0, GRIS_FONCE, BI_CONTINU, BI_MARQUE_AUCUNE, 0, 0, fenetre->w_haut,
		  graphe->x_min, graphe->x_max, (graphe->y_min - y_min)/facteur, (graphe->y_max - y_min)/facteur,
		  echelle_X, echelle_Y );
    } else {
      TRC_Graphe( fenetre, infos->mu_0, GRIS_FONCE, BI_CONTINU, BI_MARQUE_AUCUNE, 0, 0, fenetre->w_haut,
		  graphe->x_min, graphe->x_max, (graphe->y_max - y_min)/facteur, (graphe->y_min - y_min)/facteur,
		  echelle_X, echelle_Y );
    }
  }
}

static int GER_xnorm(unsigned int num_objet, int x, int y)
{
  MU1_infos *infos;
  Graphe *brut;
  Fenetre *fenetre;
  int reponse;
  struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };

  fenetre = ARB_QuelleFenetre(arbres + NORM_XANES, FAUX); /* Fenêtre actuelle */
  brut  = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos = (MU1_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( ( NULL == brut ) || ( NULL == infos ) ) {
    return BI_EVENEMENT_TRAITE;
  }

  switch (num_objet)
  {
  case NXAN_RAPIDE: /* Méthode utilisée pour approcher l'absorption atomique */
    if ( infos->methode != ELIM_MU0_RAPIDE ) {
      OBJ_choisit(NORM_XANES, NXAN_RAPIDE, FORCE_OUI);
      infos->methode = ELIM_MU0_RAPIDE;
      XAS_MU01_CalculerMu0_I( brut, infos ); /* On recalcule  */
      FEN_redessin(fenetre, BI_RDS_UTIL);
    }
    return BI_EVENEMENT_TRAITE;

  case NXAN_LENTE:
    if ( infos->methode != ELIM_MU0_LENT) { /* On change de méthode */
      OBJ_choisit(NORM_XANES, NXAN_LENTE, FORCE_OUI);
      infos->methode = ELIM_MU0_LENT;
      XAS_MU01_CalculerMu0_I( brut, infos ); /* On recalcule  */
      FEN_redessin(fenetre, BI_RDS_UTIL);
    }
    return BI_EVENEMENT_TRAITE;

  case NXAN_ENORM: /* Énergie de normalisation */
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "Enorm =";
    nv_valeur.min = brut->x_min;
    nv_valeur.max = brut->x_max;
    nv_valeur.valeur_initiale.reel = infos->e_0;

    reponse = BI_DemandeValeur( fenetre, &Q_EnergieNormalisation, &nv_valeur );
    if ( Q_EnergieNormalisation.bouton_annuler == (unsigned int) reponse ) return VRAI; /* On annule */

    infos->e_0 = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( NORM_XANES, NXAN_ENORM, infos->e_0, 0.0 );
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case NXAN_E2: /* Valeur à utiliser pour E2 */
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "E2 =";
    nv_valeur.min = brut->x_min;
    nv_valeur.max = brut->x_max;
    nv_valeur.valeur_initiale.reel = infos->e_2;

    if ( OBJ_choisit(NORM_XANES, NORM_MU0, QUESTION) ) {
      reponse = BI_DemandeValeur( fenetre, &Q_NormalisationE2, &nv_valeur );
      if ( Q_NormalisationE2.bouton_annuler == (unsigned int) reponse ) return VRAI; /* On annule */

      XAS_MU01_CalculerMu0_I( brut, infos ); /* On recalcule le polynôme */
    } else {
      reponse = BI_DemandeValeur( fenetre, &Q_NormalisationPoint2, &nv_valeur );
      if ( Q_NormalisationPoint2.bouton_annuler == (unsigned int) reponse ) return VRAI; /* On annule */
    }
    infos->e_2 = nv_valeur.valeur_donnee.reel;

    OBJ_MettreNombre( NORM_XANES, NXAN_E2, infos->e_2, 0.0 );
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case NXAN_D_MOINS:
  case NXAN_D_FOND:
  case NXAN_DEGRE:
  case NXAN_D_PLUS:
    infos->degre = gere_ascenseur(NORM_XANES, NXAN_D_MOINS, NXAN_D_FOND, NXAN_DEGRE, NXAN_D_PLUS, ASC_HORIZONTAL,
				  1, 9, 1, infos->degre, num_objet, x, y);
    XAS_MU01_CalculerMu0_I( brut, infos ); /* On recalcule */
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case NORM_MU0:
  case NORM_1_POINT:
  case NORM_2_POINTS:
    OBJ_choisit(NORM_XANES, num_objet, FORCE_OUI);
    OBJ_active(NORM_XANES, NXAN_D_MOINS, num_objet == NORM_MU0);
    OBJ_active(NORM_XANES, NXAN_D_FOND, num_objet == NORM_MU0);
    OBJ_active(NORM_XANES, NXAN_DEGRE, num_objet == NORM_MU0);
    OBJ_active(NORM_XANES, NXAN_D_PLUS, num_objet == NORM_MU0);
    OBJ_active(NORM_XANES, NXAN_RAPIDE, num_objet == NORM_MU0);
    OBJ_active(NORM_XANES, NXAN_LENTE, num_objet == NORM_MU0);
    OBJ_active(NORM_XANES, NXAN_E2, num_objet != NORM_1_POINT);
    if (num_objet == NORM_MU0)
      XAS_MU01_CalculerMu0_I( brut, infos );
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case NX_REFERENCE:
    OBJ_choisit(NORM_XANES, NX_REFERENCE, FORCE_OUI);
    infos->mu_0 = GRF_choisir(GR_BRUT, FAUX, VRAI, NULL);
    if (infos->mu_0) {
      met_chaine(NORM_XANES, NX_REFERENCE, LGMAX_NOM_REFERENCE, infos->mu_0->nom, TXT_COPIER);
    } else {
      met_chaine(NORM_XANES, NX_REFERENCE, LGMAX_NOM_REFERENCE, txt_pas_reference, TXT_COPIER);
    }
    FEN_redessin(fenetre, BI_RDS_UTIL);
    OBJ_choisit(NORM_XANES, NX_REFERENCE, FORCE_NON);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Préparation de l'arbre de normalisation des spectres XANES -------- */

static int prepare_norm_xanes(unsigned int numero_arbre __attribute__((__unused__)))
{
  OBJ_deplace(NORM_XANES, NORM_2_POINTS, +1, 0, DEPLACE_PIXELS);
  OBJ_deplace(NORM_XANES, NORM_MU0     ,  0, +1, DEPLACE_PIXELS);

  OBJ_deplace(NORM_XANES, NXAN_E2      ,  0, +1, DEPLACE_PIXELS);

  OBJ_deplace(NORM_XANES, NXAN_LENTE   ,  0, +1, DEPLACE_PIXELS);

  OBJ_deplace(NORM_XANES, NXAN_D_PLUS  ,  0, +1, DEPLACE_PIXELS);
  OBJ_deplace(NORM_XANES, NXAN_D_FOND  , +1, +1, DEPLACE_PIXELS);
  OBJ_deplace(NORM_XANES, NXAN_D_MOINS ,  0, +1, DEPLACE_PIXELS);

  OBJ_agrandit(NORM_XANES, ROOT       ,  0, +1);
  OBJ_agrandit(NORM_XANES, NXAN_D_FOND, -2, 0);

#ifdef AVEC_TRADUCTION_ /* Traduction disponible */
  met_chaine(NORM_XANES, NORM_MU0, 0, txt_nx_avec_mu0, TXT_REMPLACE);
  met_chaine(NORM_XANES, NXAN_RAPIDE, 0, txt_nx_normal, TXT_REMPLACE);
  met_chaine(NORM_XANES, NXAN_LENTE, 0, txt_nx_pondere, TXT_REMPLACE);
  met_chaine(NORM_XANES, NORM_NOUVEAU, 0, txt_nx_nv_grf, TXT_REMPLACE);
#endif

  arbres[NORM_XANES].fonction.arbre = GER_xnorm;
  arbres[NORM_XANES].drapeaux.prepare = VRAI;

  return 0;
}
