/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2003, Emmanuel CURIS

   Sauvegarde des fichiers au format interne
   (Ce format conditionne le numéro de version)

   Version 1.0 :
 - sauvegarde du nom, des bornes, des X et Y, du type
 - reconnaissance : EMC\233 comme code d'en-tête
 - type complexe autorisé uniquement si type == GR_TF

   Version 1.1 :
 - sauvegarde des légendes, des unités
 - type de données (réelles, complexes) sauvé
 - sauvegarde de la couleur et du style
 - sauvegarde des infos de fenêtres pour une TF

   Version 1.2 :
 - on sauve les infos pour les fenêtres de TF asymétriques

   Version 1.3 : recommencé à zéro la sauvegarde
 - en-tête modifié : LASE et contient une info sur la taille des doubles
 - titre et légendes du graphe sauvées plus proprement
 - sauvegarde des barres d'erreur si elles existent

   Version 1.4 :
 - les barres d'erreurs sont aussi sauvées pour les graphes complexes

   Version 1.5 : réorganisé toute la sauvegarde
 - l'en-tête est 'LASE' codé sur un unsigned long
 - modifié la façon de détecter les double mal codés : on sauve 3.1415
 - ajout d'un flag dans l'en-tête : modèle 3d ou graphe isolé ?
 - première version de la sauvegarde d'un modèle 3d : 1 molécule, ses atomes & fragments, absorbeurs
 - corrigé la sauvegarde des infos de TF : type de type_fenetre et degre était double au lieu de int !
 - sauvegarde des graphes 3d : (x, y, z)

   Version 1.6 :
 - on ne sauve plus le degré pour la TF, puisqu'il n'a plus lieu d'être.

   Version 1.7 :
 - on sauve l'utilisation, ou non, d'un chemin pour l'ajustement
 - on sauve la contribution au spectre d'un chemin de diffusion
 - on sauve les infos de Nyquist d'un spectre
 - début pour la sauvegarde des matrices de covariance

   Version 1.8
 - on sauve la maille associée à la molécule

   Version 1.9
 - on sauve les drapeaux des fragments (maille ? maille à l'origine ?)

   Version 2.0
 - le rayon d'un atome est en float
 - on sauve le schéma de liaisons (pour les fragments)

   Version 2.1
 - on sauve les liaisons entre fragments aussi

   Version 2.2
 - matrice de covariance dissociée des infos de TF ou de TFI, sauvée en toute fin de fichier

   Version 2.3
 - graphe : sauvegarde des degrés de liberté pour les barres d'erreur
 - adaptation aux modèles : les modèles sont sauvés pour les absorbeurs

   Version 3.0: Gros changements dans la structure des graphes ; images, expériences,...
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "serie.h"

#include "atomes.h"
#include "c_formel.h"
#include "chemins.h"
#include "couches.h"
#include "cristallo.h" /* Pour sauvegarde de la maille */
#include "feff.h"
#include "fichiers.h"
#include "filtrage.h" /* Pour sauvegarde de TF_infos */
#include "images.h"
#include "matrices.h" /* Pour sauvegarde de la matrice de covariance */
#include "modeles.h"
#include "theorie.h" /* Pour sauvegarde des types de lpm / modèle FEFF */
#include "science.h" /* pour M_PI */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

static int version = VERSION;

#define VERIFIE(x, ch) if (n_ecrit < (x)) {\
                         (void) SYS_ERREUR(ERREUR_ECRITURE, "Fichier en cours", fichier, NULL); \
                         (void) fprintf(stderr, "VERIFIE, sauvegarde, %s\n", ch); \
                         (void) fclose(fichier); \
                         return FAUX; }

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int sv_absorbeur (FILE *fichier, Absorbeur *absorbeur);
static int sv_atome     (FILE *fichier, Atome *atome);
static int sv_chemin    (FILE *fichier, Chemin *chemin, boolean sv_amp, boolean sv_pha, boolean sv_signal);
static int sv_couche    (FILE *fichier, Couche *couche);
static int sv_fragment  (FILE *fichier, Fragment *fragment);
static int sv_liaison   (FILE *fichier, Liaison *liaison);
static int sv_modele    (FILE *fichier, struct modele_ajuste *modele);
static int sv_modele_feff(FILE *fichier, struct modele_feff *modele);
static int sv_variable  (FILE *fichier, Variable *variable);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Sauvegarde du modèle structural 3d.
            Existe uniquement depuis la version 1.5
   Format de fichier de base :

   << en_tête LASE classique 1.5 >>
   lg_titre  - unsigned int : nombre de caractères dans le titre
   <titre>   - n * char : titre de la molécule
   flg_cell  - char : VRAI si les paramètres de maille sont présents
      Si VRAI : a   - double : paramètre a
                b   - double : paramètre b
                c   - double : paramètre c
                a   - double : paramètre alpha 
                b   - double : paramètre beta
                c   - double : paramètre gamma
                grp - int    : groupe d'espace
   << les atomes de la molécule - finit par Z = 255, faux atome... >>   
   nbr_frag  - unsigned int : nombre de fragments
   << les fragments de la molécule >>
   nbr_abs   - unsigned int : nombre d'absorbeurs
   << les absorbeurs de la molécule >>
   nbr_liai  - unsigned int : nombre de liaisons dans la molécule, entre fragments
   << les liaisons de la molécule >>
*/

int MOL_SauverIsolee(const char *nom, Molecule *molecule)
{
  FILE *fichier;
  Atome *atome;
  Fragment *fragment;
  Liaison *liaison;
  /*  boolean flag = FAUX; */
  size_t n_ecrit = 0;
  unsigned long i;
  unsigned char Z = 255;
  unsigned int lg_titre;
  char flg_cell;

  OPTION_A_FAIRE("Déplacer dans Atomes/sauver.c");

  if ( NULL ==  molecule ) return VRAI;
  if ( NULL == molecule->premier ) {
    (void) fprintf(stderr, "MOL_sauver : Molécule sans atomes, sauvegarde annulée\n");
    return VRAI;
  }
  if ( NULL == nom ) { /* Pas de nom de fichier : on le demande */
    (void) fprintf(stderr, "%s, %s l.%u : Pas de nom de fichier indiqué !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    /*    flag = VRAI; */
  }

  fichier = fopen(nom, "wb");
  if (fichier == NULL) {
    (void) SYS_ERREUR(ERREUR_ECRITURE, nom, NULL, NULL);
    return FAUX;
  }

  HIST_Titre(1, "Sauvegarde", "E/S");
  HIST_Texte(1, 0, "Fichier : %s\n", nom);
  HIST_Texte(1, 0, "Mol"e_AIGU"cule sauv"e_AIGU" : %s\n", molecule->nom);

  if ( SV_EnTete( fichier, FICHIER_MOLECULE ) == FAUX )
    return FAUX;

  lg_titre = (unsigned int) strlen(molecule->nom);
  n_ecrit = fwrite(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "MOL_sauver - lg_titre");
  n_ecrit = fwrite(molecule->nom, sizeof(char), lg_titre, fichier);
  VERIFIE(lg_titre, "MOL_sauver - titre");

  /* ------ Sauvegarde des paramètres de maille ------ */
  if ( molecule->maille != NULL ) {
    OPTION_A_FAIRE( "Sauvegarder les paramètres de maille, nlle version" );
  }
  flg_cell = FAUX; /* FAUX <=> pas de paramètres de maille */
  n_ecrit = fwrite(&flg_cell, sizeof(char), 1, fichier);
  VERIFIE(1, "MOL_sauver - flg_cell");
  if ( flg_cell == VRAI ) {
    struct maille *maille;
    double angle;

    maille = molecule->maille;

    n_ecrit  = fwrite(&(maille->a), sizeof(double), 1, fichier);
    n_ecrit += fwrite(&(maille->b), sizeof(double), 1, fichier);
    n_ecrit += fwrite(&(maille->c), sizeof(double), 1, fichier);

    /* ATTENTION, Angles sauvés en degrés pour compatibilité avec l'ancien format
       À corriger dans la prochaine mise à jour du format de fichier */

    angle = maille->alpha * 180.0 / M_PI;
    n_ecrit += fwrite(&(angle), sizeof(double), 1, fichier);

    angle = maille->beta * 180.0 / M_PI;
    n_ecrit += fwrite(&(angle), sizeof(double), 1, fichier);

    angle = maille->gamma * 180.0 / M_PI;
    n_ecrit += fwrite(&(angle), sizeof(double), 1, fichier);
    VERIFIE(6, "MOL_sauver - paramètres de maille");

    n_ecrit  = fwrite(&(maille->type), sizeof(int), 1, fichier);
  }

  /* ------ On sauve les atomes, en mettant les index au passage ------ */
  for ( atome = molecule->premier, i = 1 ; atome != NULL ; i++, atome = atome->suivant ) {
    atome->index = i; /* Chaque atome n'apparaît qu'une fois dans la liste chaînée */
    if (sv_atome(fichier, atome) == FAUX)
      return FAUX;
  }

  /* Le dernier atome, spécial avec Z = 255 */
  n_ecrit = fwrite(&Z, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "MOL_sauver - fin de la liste des atomes");

  /* Maintenant, on sauve l'agencement de ces atomes dans la molécule
     et les infos complémentaires */
  /* On commence par les fragments */

  n_ecrit = fwrite(&(molecule->nbr_fragments), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "MOL_sauver - nombre de fragments");
  for (fragment = molecule->fragments; fragment ; fragment = fragment->suivant) {
    if ( sv_fragment(fichier, fragment) == FAUX)
      return FAUX;
  }

  /* On fait les absorbeurs */
  n_ecrit = fwrite(&(molecule->nbr_absorbeurs), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "MOL_sauver - nombre d'absorbeurs");
  for (i = 0 ; i < molecule->nbr_absorbeurs ; i++) {
    if ( sv_absorbeur(fichier, molecule->absorbeurs + i) == FAUX)
      return FAUX;
  }

  /* On sauve les liaisons entre fragments */
  n_ecrit = fwrite(&(molecule->nbr_liaisons), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "MOL_sauver - nombre de liaisons");
  for (i = 0, liaison = molecule->liaisons ; i < molecule->nbr_liaisons ; i++, liaison = liaison->suivante) {
    if ( sv_liaison(fichier, liaison) == FAUX)
      return FAUX;
  }

  (void) fclose(fichier);

  return VRAI; /* La sauvegarde a fonctionné */
}

/* ~~~~~~~~ Les briques de base pour la sauvegarde ~~~~~~~~ */

/* -------- Sauvegarde de l'en-tête --------

   Renvoie VRAI si a pu l'écrire, FAUX si problème.

   Format :
   'LASE'    - unsigned long, contrôle d'origine + taille & codage des entiers
   <version> - unsigned int
   3.1415    - double, contrôle taille & codage des doubles
   type      - int, type de données contenues dans le fichier [v. 3.0]
*/

inline int SV_EnTete(FILE *fichier, int type)
{
  size_t n_ecrit;
  const unsigned long en_tete = 'LASE';
  const double pi = 3.1415;

  if (fichier == NULL) return FAUX;

  n_ecrit = fwrite(&en_tete, sizeof(unsigned long), 1, fichier); /* Identificateur LASE */
  VERIFIE(1, "SV_EnTete - en_tete");
  n_ecrit = fwrite(&version, sizeof(unsigned int), 1, fichier);  /* Numéro de version   */
  VERIFIE(1, "SV_EnTete - version");
  n_ecrit = fwrite(&pi, sizeof(double), 1, fichier); /* Contrôle pour les doubles */
  VERIFIE(1, "SV_EnTete - Valeur de pi");

  n_ecrit = fwrite(&type, sizeof(int), 1, fichier); /* Quel type de fichier ? */
  VERIFIE(1, "SV_EnTete - type de fichier");
  return VRAI;
}

/* -------- Sauvegarde d'un graphe, avec flag de présence -------- */

inline int SV_Graphe(FILE *fichier, Graphe *graphe)
{
  short int drapeau;
  int n_ecrit, retour;

  if ( NULL == fichier ) return -1;

  if ( NULL == graphe ) {
    drapeau = FAUX;
    n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );
    if ( n_ecrit != 1 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de sauver le drapeau de présence / pas de graphe\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -2;
    }

    return 0;
  }

  drapeau = VRAI;
  n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );
  if ( n_ecrit != 1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de sauver le drapeau de présence / graphe %s\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->nom);
    return -3;
  }

  /* On sauve le graphe, puisqu'il y en a un */
  retour = GRF_Sauver( fichier, graphe );
  if ( FAUX == retour ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_Sauver() [graphe %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->nom);
    return -4;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Sauvegarde d'une image, avec flag de présence -------- */

inline int SV_Image(FILE *fichier, struct image *image)
{
  short int drapeau;
  int n_ecrit, retour = 1;

  if ( NULL == fichier ) return -1;

  if ( NULL == image ) {
    drapeau = FAUX;
    n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );
    if ( n_ecrit != 1 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de sauver le drapeau de présence / pas d'image\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -2;
    }

    return 0;
  }

  drapeau = VRAI;
  n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );
  if ( n_ecrit != 1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de sauver le drapeau de présence / image\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* On sauve l'image, puisqu'il y en a une */
  retour = IMG_Sauver( fichier, image );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_Sauver()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -4;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Sauvegarde des informations d'un absorbeur
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    lg_titre  - unsigned int, longueur du nom de l'absorbeur
    <nom de l'absorbeur>
    idx_atm   - unsigned long, numéro de l'atome absorbeur
    << variable : contribution >>
    << variable : S0_2 >>                                ]
    << variable : Delta_E >>                             ]
    nbr_chem  - unsigned int, nombre de chemins          ] v. 2.3 : DISPARAÎT, remplacé par les modèles
    << les nbr_chem chemins >>                           ]
    flg_exp   - unsigned char, 1 si spectre expérimental ]
       Si 1 : << spectre expérimental >>                 ]
    flg_th    - unsigned char, 1 si spectre théorique    ]
       Si 1 : << spectre théorique >>                    ]
    flg_res   - unsigned char, 1 si spectre résidu       ] v. 2.3 : DISPARAÎT, voir les modèles...
       Si 1 : << spectre résidu >>                       ]
    nbr_mod   - unsigned int : nombre de modèles   ] v. 2.3
   << les nbr_mod modèles d'ajustement >>          ] v. 2.3
*/

static int sv_absorbeur(FILE *fichier, Absorbeur *absorbeur)
{
  size_t n_ecrit;
  unsigned int lg_titre, i;
  struct modele_ajuste *modele;

  if (fichier == NULL) return FAUX;
  if (absorbeur == NULL) return FAUX;

  lg_titre = (unsigned int) strlen(absorbeur->nom);
  n_ecrit = fwrite(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_absorbeur");
  n_ecrit = fwrite(absorbeur->nom, sizeof(char), lg_titre, fichier);
  VERIFIE(lg_titre, "sv_absorbeur");

  n_ecrit = fwrite(&(absorbeur->absorbeur->index), sizeof(unsigned long), 1, fichier);
  VERIFIE(1, "sv_absorbeur");

  if (sv_variable(fichier, &(absorbeur->contribution)) == FAUX) return FAUX;

  n_ecrit = fwrite(&(absorbeur->nbr_modeles), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_absorbeur");

  for (i = 0, modele = absorbeur->modeles ; modele != NULL ; i++, modele = modele->suivant) {
    if ( FAUX == sv_modele(fichier, modele) ) {
      (void) fprintf(stderr, "sv_absorbeur : erreur dans sv_modele / modèle %u = %s\n", i, absorbeur->modeles[i].nom_modele);
      return FAUX;
    }
  }
  if (i != absorbeur->nbr_modeles) {
    (void) fprintf(stderr, "sv_absorbeur : %u absorbeurs indiqués dans le modèle, %u en réalité [%s] !\n",
		   absorbeur->nbr_modeles, i, absorbeur->nom);
#ifdef _DEVELOPPE_
    abort();
#endif
    (void) fclose(fichier);
    return FAUX;
  }

  return VRAI;
}

/* -------- Sauvegarde des informations d'un atome
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
   Z         - unsigned char, numéro atomique - m'étonnerait que l'on puisse avoir Z > 255 !
   x         - double }
   y         - double } coordonnées atome
   z         - double }
   rayon     - float  [ double avant la 2.0 ]
*/

static int sv_atome(FILE *fichier, Atome *atome)
{
  unsigned char Z;
  size_t n_ecrit;

  if (fichier == NULL) return FAUX;
  if (atome == NULL) return FAUX;

  Z = (unsigned char) (atome->num_atomique);
  n_ecrit  = fwrite(&Z, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_atome - numéro atomique");

  n_ecrit  = fwrite(&(atome->position.x), sizeof(double), 1, fichier);
  n_ecrit += fwrite(&(atome->position.y), sizeof(double), 1, fichier);
  n_ecrit += fwrite(&(atome->position.z), sizeof(double), 1, fichier);
  VERIFIE(3, "sv_atome - coordonnées");

  n_ecrit  = fwrite(&(atome->rayon), sizeof(float), 1, fichier);
  VERIFIE(1, "sv_atome - rayon");
  return VRAI;
}

/* -------- Sauvegarde d'un chemin
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    lg_titre - unsigned int, longueur du titre
    < le nom du chemin >
    n_atms - unsigned int, nombre d'atomes dans le chemin
    ix_atm - unsigned long, numéro des atomes - * n_atms
    << variable : multiplicité >>     } ]
    << variable : R >>                } ] v. 2.3 : DEVIENNENT DES DOUBLES
    << variable : dw_global >>        } ]
    flag : 1 si amplitude associée    }
      Si 1 : << graphe : amplitude >> }
    flag : 1 si déphasage associée    } de la couche s'il y en a une, du chemin sinon.
      Si 1 : << graphe : déphasage >> }
    flag : 1 si utilisé dans l'aj.    } ] DISPARAÎT DANS LA 2.3
    flag : 1 si contrib. associée     ]
      Si 1 : << graphe : contrib. >>  ] DEPUIS LA 1.7
    n_eq   - unsigned int, nombre de chemins équivalents ] DEPUIS LA 2.3
    << chemins équivalents >>                            ]
*/

static int sv_chemin(FILE *fichier, Chemin *chemin, boolean phase, boolean amplitude, boolean sv_signal)
{
  unsigned char drapeau;
  size_t n_ecrit;
  unsigned int i, lg_titre;

  if (fichier == NULL) return FAUX;
  if (chemin == NULL) return FAUX;

  lg_titre = (unsigned int) strlen(chemin->nom);
  n_ecrit = fwrite(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_chemin - longueur du nom");
  n_ecrit = fwrite(chemin->nom, sizeof(char), lg_titre, fichier);
  VERIFIE(lg_titre, "sv_chemin - nom");

  n_ecrit = fwrite(&(chemin->nbr_atomes), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_chemin - nombre d'atomes");

  n_ecrit = 0;
  for (i = 0 ; i < chemin->nbr_atomes ; i++)
    n_ecrit += fwrite( &(chemin->atomes[i]->index), sizeof(unsigned long), 1, fichier);
  VERIFIE(chemin->nbr_atomes, "sv_chemin / atomes");

  n_ecrit = fwrite(&(chemin->multiplicite), sizeof(double), 1, fichier);
  VERIFIE(1, "sv_chemin - multiplicité");
  n_ecrit = fwrite(&(chemin->longueur_reelle), sizeof(double), 1, fichier);
  VERIFIE(1, "sv_chemin - longueur réelle");
  n_ecrit = fwrite(&(chemin->dw_global), sizeof(double), 1, fichier);
  VERIFIE(1, "sv_chemin - DW global");

  if ( VRAI == amplitude )
    drapeau = (chemin->amplitude != NULL);
  else /* Déjà sauvée, dans la couche... */
    drapeau = FAUX;
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_chemin - drapeau / amplitude");
  if (drapeau == 1) {
    if (GRF_Sauver(fichier, chemin->amplitude) == FAUX) {
      (void) fprintf(stderr, "sv_chemin : erreur dans GRF_Sauver / amplitude\n");
      return FAUX;
    }
  }

  if ( VRAI == phase )
    drapeau = (chemin->phase != NULL);
  else /* Déjà sauvée, dans la couche... */
    drapeau = FAUX;
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_chemin - drapeau / phase");
  if (drapeau == 1) {
    if (GRF_Sauver(fichier, chemin->phase) == FAUX) {
      (void) fprintf(stderr, "sv_chemin : erreur dans GRF_Sauver / phase\n");
      return FAUX;
    }
  }

  if ( VRAI == sv_signal )
    drapeau = (chemin->exf_calcule != NULL);
  else
    drapeau = FAUX;
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_chemin");
  if (drapeau == 1) {
    if (GRF_Sauver(fichier, chemin->exf_calcule) == FAUX) {
      (void) fprintf(stderr, "sv_chemin : erreur dans GRF_Sauver / contribution\n");
      return FAUX;
    }
  }

  n_ecrit = fwrite(&(chemin->nbr_equivalents), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_chemin - nombre d'équivalents");
  for (i = 0 ; i < chemin->nbr_equivalents ; i++) {
    if ( FAUX ==  sv_chemin(fichier, chemin->equivalents[i], VRAI, VRAI, VRAI) ) {
      (void) fprintf(stderr, "sv_chemin : erreur dans sv_chemin / équivalent n° %u : %s\n",
		     i + 1, chemin->equivalents[i]->nom);
      return FAUX;
    }
  }

  return VRAI;
}

/* -------- Sauvegarde d'une couche d'un modèle en chemins
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    lg_titre  - unsigned long, longueur du nom
    <<titre>> - char * lg_titre, le nom
    drapeau   - unsigned char, 1 si la couche est utilisée dans l'ajustement 
    <<variable : N>>
    <<variable : R>>
    <<variable : Debye-Waller>>
    <<variable : Cumul_1>>
    <<variable : Cumul_3>>
    <<variable : Cumul_2>>
    <<variable : Cumul_4>>
    <<variable : Delta_E / couche>>
    <<variable : Gamma / couche>>
    drapeau   - unsigned char, 1 si la contribution existe
      Si drapeau == 1
        <<graphe : la contribution>>
    drapeau   - unsigned char, 1 si amplitude associée
      Si  drapeau == 1
        << graphe : amplitude >> }
    drapeau   - unsigned char, 1 si phase associée
      Si  drapeau == 1
        << graphe : phase >> }
    drapeau   - unsigned char, 1 si chemin associé
      Si  drapeau == 1
        << chemin : le chemin associé >>
*/

static int sv_couche(FILE *fichier, Couche *couche)
{
  unsigned long lg_titre;
  int n_ecrit;
  unsigned char drapeau;

  if (fichier == NULL) return FAUX;
  if (couche == NULL) return FAUX;

  lg_titre = strlen(couche->nom);
  n_ecrit = fwrite( &lg_titre, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, "sv_couche - longueur du nom");
  n_ecrit = fwrite( couche->nom, sizeof(char), lg_titre, fichier);
  VERIFIE( (int) (lg_titre), "sv_couche - nom");

  drapeau = (unsigned char) (couche->utilisee);
  n_ecrit = fwrite( &drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_couche - drapeau / couche utilisée");

  if ( FAUX == sv_variable(fichier, &(couche->N) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / N\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->R) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / R\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->Debye_Waller) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / DW\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->Cumul_1) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / 1er cumulant\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->Cumul_3) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / 3e cumulant\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->Cumul_2) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / 2e cumulant\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->Cumul_4) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / 4e cumulant\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->Delta_E) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / DE0\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(couche->Gamma) ) ) {
    (void) fprintf(stderr, "sv_couche : erreur dans sv_variable / Gamma\n");
    return FAUX;
  }

  drapeau = (couche->signal != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_couche - drapeau / graphe de contribution");
  if (VRAI == drapeau) {
    GRF_Sauver(fichier, couche->signal);
  }
  drapeau = (couche->amplitude != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_couche - drapeau / graphe d'amplitude ");
  if (VRAI == drapeau) {
    GRF_Sauver(fichier, couche->amplitude);
  }
  drapeau = (couche->phase != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_couche - drapeau / graphe de déphasage ");
  if (VRAI == drapeau) {
    GRF_Sauver(fichier, couche->phase);
  }

  drapeau = (couche->chemin != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_couche - drapeau / chemin associé ");
  if (VRAI == drapeau) {
    sv_chemin(fichier, couche->chemin, (couche->amplitude != couche->chemin->amplitude), /* Faut-il sauver l'amplitude ? */
	      (couche->phase != couche->chemin->phase), (couche->signal != couche->chemin->exf_calcule));
  }

  /* Tout s'est bien passé */
  return VRAI;
}

/* -------- Sauvegarde d'un fragment
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    lg_titre  - unsigned int, longueur du titre
    < le nom du fragment >
    flg_m     - char : le fragment est-il une maille ?
       Si oui : flg_o   - char : le fragment est-il la maille centrée à l'origine
    n_atms    - unsigned int : nombre d'atomes dans le fragment
    idx_atm   - unsigned long : index des atomes du fragment (n_atms fois)
    n_liaison - unsigned int : le nombre de liaisons qui suivent ]
   << liaison : les différentes liaisons >>                      ] depuis la version 2.0
*/

static int sv_fragment(FILE *fichier, Fragment *fragment)
{
  size_t n_ecrit;
  unsigned int i, lg_titre;
  char drapeau;
  Liaison *liaison;

  if (fichier == NULL) return FAUX;
  if (fragment == NULL) return FAUX;

  lg_titre = (unsigned int) strlen(fragment->nom);
  n_ecrit = fwrite(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_fragment - longueur de titre");
  n_ecrit = fwrite(fragment->nom, sizeof(char), lg_titre, fichier);
  VERIFIE(lg_titre, "sv_fragment - titre");

  drapeau = fragment->drapeaux.maille;
  n_ecrit = fwrite(&drapeau, sizeof(char), 1, fichier);
  VERIFIE(1, "sv_fragment - maille ?");
  if (drapeau == VRAI) /* Sinon, l'autre drapeau est forcément à zéro... */
  {
    drapeau = fragment->drapeaux.origine;
    n_ecrit = fwrite(&drapeau, sizeof(char), 1, fichier);
    VERIFIE(1, "sv_fragment - maille à l'origine ?");
  }

  n_ecrit = fwrite(&(fragment->nbr_atomes), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_fragment - nombre d'atomes");

  n_ecrit = 0;
  for (i = 0 ; i < fragment->nbr_atomes ; i++)
    n_ecrit += fwrite( &(fragment->atomes[i]->index), sizeof(unsigned long), 1, fichier);
  VERIFIE(fragment->nbr_atomes, "sv_fragment - les atomes");

  n_ecrit = fwrite( &(fragment->nbr_liaisons), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_fragment - nombre de liaisons");

  for (liaison = fragment->liaisons ; liaison != NULL ; liaison = liaison->suivante) {
    sv_liaison(fichier, liaison);
  }

  return VRAI;
}

/* -------- Sauvegarde d'une liaison
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    n_atomes     - unsigned short int, nombre d'atomes impliqués dans la liaison
    atome_1      - unsigned long, index du premier atome de la liaison
    atome_2      - unsigned long, index du deuxième atome de la liaison
      Si n_atomes > 2
         les index des autres atomes impliqués
    multiplicite - float, multiplicité de la liaison
    type         - unsigned short int, type de liaison considéré
*/

static int sv_liaison(FILE *fichier, Liaison *liaison)
{
  size_t n_ecrit;

  if (fichier == NULL) return FAUX;
  if (liaison == NULL) return FAUX;

#ifdef _DEVELOPPE_
  printf("sv_liaison : liaison %p [%s-%s]\n", (void *) liaison,
	 ATM_donnees[liaison->atome_1->num_atomique - 1].symbole, ATM_donnees[liaison->atome_2->num_atomique - 1].symbole);
#endif

  n_ecrit = fwrite(&(liaison->nbr_atomes), sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, "sv_liaison - nombre d'atomes");

  n_ecrit  = fwrite(&(liaison->atome_1->index), sizeof(unsigned long), 1, fichier);
  n_ecrit += fwrite(&(liaison->atome_2->index), sizeof(unsigned long), 1, fichier);
  VERIFIE(2, "sv_liaison - atomes principaux");

  if ( liaison->nbr_atomes > 2 ) {
    unsigned short int i;

    n_ecrit = 0;
    for (i = 0 ; i < liaison->nbr_atomes - 2 ; i++) {
      n_ecrit += fwrite(&(liaison->autres_atomes[i]->index), sizeof(unsigned long), 1, fichier);
    }
    VERIFIE( (unsigned int) (liaison->nbr_atomes - 2), "sv_liaison - atomes en trop");
  }

  n_ecrit = fwrite(&(liaison->multiplicite), sizeof(float), 1, fichier);
  VERIFIE(1, "sv_liaison - multiplicité");

  n_ecrit = fwrite(&(liaison->type), sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, "sv_liaison - type");

  return VRAI;
}

/* -------- Sauvegarde d'un modèle d'ajustement
     Renvoie  VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    type      - int, type de modèle
    lg_titre  - unsigned long, longueur du nom
    <<nom>>   - le nom (lg_titre caractères)
    x_min     - double, valeur minimale de l'intervalle d'ajustement
    x_max     - double, valeur maximale de l'intervalle d'ajustement
    drapeau   - unsigned char, 1 si le graphe expérimental est présent
      Si drapeau == 1
        <<graphe : graphe expérimental>>
    drapeau   - unsigned char, 1 si le graphe modèle est présent
      Si drapeau == 1
        <<graphe : graphe modèle>>
    drapeau   - unsigned char, 1 si le graphe des résidus est présent
      Si drapeau == 1
        <<graphe : graphe des résidus>>
    algo      - unsigned int, type d'algorithme utilisé
    max_iter  - unsigned int, nombre maximal d'itérations
    arret     - int, conditions d'arrêt
    err_min   - short int, -log10 de l'erreur minimale autorisée
    derr_min  - short int, -log10 de la variation minimale de l'erreur autorisée
    pondere   - unsigned char, 1 si ajustement pondéré par les barres d'erreur
    contraint - unsigned char, 1 si ajustement avec contraintes
      Si type == MODELE_EXAFS_CHEMINS
        <<modèle FEFF>>
*/

static int sv_modele(FILE *fichier, struct modele_ajuste *modele)
{
  size_t n_ecrit;
  unsigned char drapeau;
  unsigned long lg_titre;

  if (fichier == NULL) return FAUX;
  if (NULL == modele) return FAUX;

#ifdef _DEVELOPPE_
  printf("sv_modele : modele %p [%s / %d]\n", (void *) modele, modele->nom_modele, modele->type);
#endif
  n_ecrit = fwrite(&(modele->type), sizeof(int), 1, fichier);
  VERIFIE(1, "sv_modele - type");

  lg_titre = strlen(modele->nom_modele);
  n_ecrit = fwrite(&lg_titre, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, "sv_modele - longueur du nom");
  n_ecrit = fwrite(modele->nom_modele, sizeof(char), lg_titre, fichier);
  VERIFIE(lg_titre, "sv_modele - nom du mod"e_GRAVE"le");

  n_ecrit = fwrite(&(modele->x_min), sizeof(double), 1, fichier);
  VERIFIE(1, "sv_modele - x_min");
  n_ecrit = fwrite(&(modele->x_max), sizeof(double), 1, fichier);
  VERIFIE(1, "sv_modele - x_max");

  drapeau = (modele->experience != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_modele - drapeau / graphe expérimental");
  if ( VRAI == drapeau ) {
    GRF_Sauver(fichier, modele->experience);
  }
  drapeau = (modele->modele != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_modele - drapeau / graphe modèle");
  if ( VRAI == drapeau ) {
    GRF_Sauver(fichier, modele->modele);
  }
  drapeau = (modele->residu != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_modele - drapeau / graphe résidu");
  if ( VRAI == drapeau ) {
    GRF_Sauver(fichier, modele->residu);
  }

  n_ecrit = fwrite(&(modele->algorithme), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_modele - algorithme");
  n_ecrit = fwrite(&(modele->max_iteration), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_modele - nombre maximal d'itérations");
  n_ecrit = fwrite(&(modele->condition_arret), sizeof(int), 1, fichier);
  VERIFIE(1, "sv_modele - condition d'arrêt");
  n_ecrit = fwrite(&(modele->err_min), sizeof(short int), 1, fichier);
  VERIFIE(1, "sv_modele - erreur minimale autorisée");
  n_ecrit = fwrite(&(modele->derr_min), sizeof(short int), 1, fichier);
  VERIFIE(1, "sv_modele - variation minimale autorisée pour l'erreur");

  drapeau = modele->drapeaux.pondere;
  n_ecrit = fwrite( &drapeau, sizeof(unsigned char), 1, fichier );
#ifdef _DEVELOPPE_
  printf( " ==> %s <== [ponderation : %d / %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere, drapeau );
#endif
  VERIFIE(1, "sv_modele - ajustement pondéré ?");

  drapeau = modele->drapeaux.contraintes;
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_modele - ajustement contraint ?");

  /* Informations complémentaires éventuelles */
  switch (modele->type)
  {
  case MODELE_EXAFS_CHEMINS:
    if ( FAUX == sv_modele_feff(fichier, modele->infos_modele.feff) )
      return FAUX;
    break;

  default:
    fprintf(stderr, "sv_modele : À FAIRE, infos complémentaires / modèle de type %d\n", modele->type);
  }

  /* Tout s'est bien passé */
  return VRAI;
}

/* -------- Sauvegarde d'un modèle d'ajustement
     Renvoie  VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    seuil       - int, seuil modélisé [valeurs dans theorie.h]
    idx_atm     - unsigned long, atome absorbeur de ce modèle
    potentiels  - int * 9, table des potentiels FEFF
    <<variable : S0²>>
    <<variable : DE0 général>>
    <<variable : Gamma général>>
    DE0         - unsigned char, 1 si DE0 général
    GAM         - unsigned char, 1 si Gamma général
    nbr_couches - unsigned int, nombre de couches dans le modèle
    <<couche>> * nbr_couches
    ponderation - unsigned short int, pondération utilisée dans l'ajustement
    type_lpm    - unsigned short int, modèle de lpm utilisé
      Si type_lpm == MODELE_LPM_GRAPHE
        drapeau - unsigned char, 1 si graphe de lpm fourni
          Si drapeau == 1
            <<graphe : le lpm>>
      Si type_lpm == MODELE_LPM_GAMMA_ETA
        val_eta - double, valeur du paramètre eta
    drapeau     - unsigned char, 1 si phase de l'atome absorbeur séparée
      Si drapeau == 1
        <<graphe : le déphasage de l'atome central>>
*/

static int sv_modele_feff(FILE *fichier, struct modele_feff *modele)
{
  Couche *couche;
  size_t n_ecrit;
  unsigned char drapeau;

  if (fichier == NULL) return FAUX;
  if (NULL == modele) return FAUX;

#ifdef _DEVELOPPE_
  printf("sv_modele_feff : modele FEFF %p\n", (void *) modele);
#endif
  n_ecrit = fwrite(&(modele->seuil), sizeof(int), 1, fichier);
  VERIFIE(1, "sv_modele_feff - seuil");

  n_ecrit = fwrite(&(modele->absorbeur->index), sizeof(unsigned long), 1, fichier);
  VERIFIE(1, "sv_modele_feff - index de l'absorbeur");

  n_ecrit = fwrite(modele->potentiels, sizeof(int), 9, fichier);
  VERIFIE(9, "sv_modele_feff - potentiels");

  if ( FAUX == sv_variable(fichier, &(modele->S0_2) ) ) {
    (void) fprintf(stderr, "sv_modele_feff : erreur dans sv_variable / S0²\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(modele->Delta_E) ) ) {
    (void) fprintf(stderr, "sv_modele_feff : erreur dans sv_variable / Delta_E\n");
    return FAUX;
  }
  if ( FAUX == sv_variable(fichier, &(modele->Gamma) ) ) {
    (void) fprintf(stderr, "sv_modele_feff : erreur dans sv_variable / Gamma\n");
    return FAUX;
  }

  drapeau = modele->drapeaux.DE0_general;
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_modele_feff - drapeau / DE0 général");
  drapeau = modele->drapeaux.GAM_general;
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_modele_feff - drapeau / Gamma général");

  n_ecrit = fwrite(&(modele->nbr_couches), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, "sv_modele_feff - nombre de couches");
  for (couche = modele->couches ; couche != NULL ; couche = couche->suivante) {
    if ( FAUX == sv_couche(fichier, couche) ) {
      (void) fprintf(stderr, "sv_modele_feff : erreur dans sv_couche / couche %s\n", couche->nom);
      return FAUX;
    }
  }

  n_ecrit = fwrite(&(modele->ponderation), sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, "sv_modele_feff - pondération");

  n_ecrit = fwrite(&(modele->type_lpm), sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, "sv_modele_feff - type de lpm");
  if ( MODELE_LPM_GRAPHE == modele->type_lpm ) {
    drapeau = (modele->lpm != NULL);
    n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
    VERIFIE(1, "sv_modele_feff - drapeau / graphe de lpm");
    if (VRAI == drapeau)
      GRF_Sauver(fichier, modele->lpm);
  } else if ( MODELE_LPM_GAMMA_ETA == modele->type_lpm ) {
    n_ecrit = fwrite(&(modele->lpm_val_eta), sizeof(double), 1, fichier);
    VERIFIE(1, "sv_modele_feff - valeur de êta (modèle de lpm)");
  }

  drapeau = (modele->phase_central != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, "sv_modele_feff - drapeau / graphe de déphasage de l'absorbeur");
  if (VRAI == drapeau) {
    GRF_Sauver(fichier, modele->phase_central);
  }

  /* Tout s'est bien passé */
  return VRAI;
}

/* -------- Sauvegarde d'une variable
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde :
    valeur - double, valeur initiale
    type   - int
      Si type == VAR_LIBRE
         v_min - double }
         v_max - double } bornes de variation et pas autorisé
	 pas   - double }
      Si type == VAR_COPIE
        << la variable dont elle est la copie >>
      Si type == VAR_EXPRESSION ou VAR_RELATION
*/

static int sv_variable(FILE *fichier, Variable *variable)
{
  size_t n_ecrit;

  if (fichier == NULL) return FAUX;
  if (variable == NULL) return FAUX;

  n_ecrit = fwrite(&(variable->valeur_initiale), sizeof(double), 1, fichier);
  VERIFIE(1, "sv_variable");
  n_ecrit = fwrite(&(variable->type), sizeof(int), 1, fichier);
  VERIFIE(1, "sv_variable");
  switch (variable->type)
  {
  case VAR_CONSTANTE:
    return VRAI;

  case VAR_LIBRE:
    n_ecrit  = fwrite(&(variable->min), sizeof(double), 1, fichier);
    n_ecrit += fwrite(&(variable->max), sizeof(double), 1, fichier);
    n_ecrit += fwrite(&(variable->delta), sizeof(double), 1, fichier);
    VERIFIE(3, "sv_variable");
    return VRAI;

  case VAR_COPIE:
    if ( sv_variable(fichier, variable->contenu.original) == FAUX) return FAUX;
    return VRAI;
  }

  return FAUX; /* Normalement, on ne doit _pas_ arriver ici */
}
